#ifndef ATTRIBUTE_H
#define ATTRIBUTE_H

/// @file 
/// @author Jakub.Moscicki@cern.ch (kuba)

#include "AttributeTypes.h"
#include "boost/shared_ptr.hpp"
#include <string>
#include "POOLCore/Token.h"

namespace pool {
  class AttributeSpecification;
  class AttributeBaseValue;

  /**
   * Metadata attribute for collection and file catalog.
   *
   * Attribute may hold value of any type T.
   * Attribute has value semantics and may be used in STL container.
   * Attribute allows for typesafe access to stored value (of type T).
   * Attribute allows for generic (without knowledge of T) stringification.
   * 
   * Requirements: 
   * Type T should have value semantics (default constructor, copy and assignment).
   * This req. may be relaxed but it would require additional implementation efforts.
   *
   * Getting the value -- two alternative possibilities:
   *  - template<class T> T& value(); 
   *    drawback: if attribute empty then needs to throw.
   *  - template<class T> void getValue(T&v); 
   *
   */
  
  class Attribute
  {
  public:
    
    /// Empty attribute has no stored value and no type.
    Attribute();
    /// Destructor
    ~Attribute();
    
    Attribute(const Attribute& r);  
    Attribute& operator = (const Attribute& r);

    /// Compare the values and type of two attributes. Do not compare the name (full specification).
    /// Example: alist["A"] == alist["B"] if both attributes are int and have the same value BUT
    /// alist["A"] != alist["C"] if C is double, even if (truncated) double value of C is equal to the int value of A.
    /// FIXME: Should we compare the specifications too?
    bool operator == (const Attribute& rhs) const;
    bool operator != (const Attribute& rhs) const;       

    /// Name and type of the attribute. Throw bad_type if Attribute contains no value.
    const AttributeSpecification & spec() const;

    /// Shares the data with an other attribute
    void shareData( const Attribute& r );
    
    /// Set the value. Throw attribute_bad_type if T does not match the stored type. 
    template<class T> void setValue(const T&v);
    
    /// Copy the value into v. Throw attribute_bad_type if T does not match the stored type.
    template<class T> void getValue(T&v) const;  
    
    /// Get stringified value. Throw attribute_bad_type if Attribute contains no value.
    std::string getValueAsString() const; 
    
    /// Set value as string str. 
    /// Throw bad_conversion (?, FIXME) if str cannot be meaningfully converted to the type of the stored value.
    void setValueAsString(std::string str); 


    void print(std::ostream&) const;

    /* implementation without boost::any */

  private:   // friends interface

    /// Internal interface for constructing Attributes from value objects.
    /// Not exposed to general public but useful for implementors of other packages which need
    /// to have the control over the internal construction of attribute objects.
    friend Attribute AttributeValueConstructor(boost::shared_ptr<AttributeBaseValue> ap, const AttributeSpecification&spec);


    friend AttributeBaseValue& AttributeValueAccessor(Attribute& attr);

    friend const AttributeBaseValue& AttributeValueAccessor(const Attribute&attr);

    /**
     * Friendly attachment of the implementation. User never creates Attribute objects directly. 
     * Friends internally may use the following constructor to attach implementation to an attribute.
     * Performance optimization possibility: placement new operator to create attr impls.
     * If null pointer is passed as an implementation is a programming error -- guarded by assert.
     *
     * Shared pointer semantics apply. So you may have many Attribute objects pointing to the same implementation.
     * Neat for implementing multiple views on the same cache of the collection or file catalog.
     */
    
    Attribute(boost::shared_ptr<AttributeBaseValue> ap, const AttributeSpecification&);

  private:

    typename_string_t stored_typename() const;
    typename_string_t stored_name() const;
    
    boost::shared_ptr<AttributeBaseValue> m_attr; 
    
    const AttributeSpecification *m_spec;
  };


  // IMPLEMENTATION //

  /**
   * \implementation
   *  Polymorphic attribute interface.
   */

  class AttributeBaseValue
  {
  public:
    virtual ~AttributeBaseValue() {}

    //@name interface mapped to Attribute public interface
    //@{
    virtual std::string getValueAsString() const = 0; 
    virtual void setValueAsString(std::string str) = 0; 
    virtual bool isEqual(const AttributeBaseValue& rhs) const = 0;
    virtual void print(std::ostream&os) const = 0;
    //@}

    //@name direct access to data in memory
    //@{
    virtual void setValueFromData( const void* data ) = 0;
    virtual void cloneData( void** outputAddress ) const = 0;
    virtual void copyData( void* outputBuffer ) const = 0;
    virtual void* getMemoryAddress() = 0;
    virtual const void* getMemoryAddress() const = 0;
    //@}
  };
  
  /**
   * \implementation 
   * Static attribute interface with implementation.
   */
  template <class T>
  class AttributeTValue : public AttributeBaseValue
  {
  public:
    AttributeTValue() : m_data() {}
    AttributeTValue(const T&v) : m_data(v) {}
    T& value() { return m_data; }
    const T& value() const { return m_data; }
    void getValue(T&v) const { v = m_data; }
    void setValue(const T&v) { m_data = v; }
    std::string getValueAsString() const;
    void setValueAsString(std::string);
    bool isEqual(const AttributeBaseValue& rhs) const;
    void print(std::ostream&os) const;

    void setValueFromData( const void* data ) { m_data = *( static_cast< const T*>( data ) ); }
    void cloneData( void** outputAddress ) const { *outputAddress = new T(m_data); }
    void copyData( void* outputBuffer ) const { *( static_cast< T* >( outputBuffer ) ) = m_data; }
    void* getMemoryAddress();
    const void* getMemoryAddress() const;

  private:
    T m_data;
  };

  /**
   * special version for pool Token
   */
  template <>
  class AttributeTValue<Token> : public AttributeBaseValue
  {
  public:
    AttributeTValue() : m_data() {}
    AttributeTValue(const Token&v) : m_data(&v) {}
    Token& value() { return m_data; }
    const Token& value() const { return m_data; }
    void getValue(Token&v) const { m_data.setData(&v); }
    void setValue(const Token&v) { v.setData(&m_data); }
    std::string getValueAsString() const { return m_data.toString(); }
    void setValueAsString(std::string s) { m_data.fromString(s); }
    bool isEqual(const AttributeBaseValue& rhs) const;
    void print(std::ostream&os) const;

    void setValueFromData( const void* data ) {
      const Token& token = *( static_cast< const Token*>( data ) );
      token.setData( &m_data );
    }
    void cloneData( void** outputAddress ) const { *outputAddress = new Token( &m_data ); }

    void copyData( void* outputBuffer ) const { m_data.setData( static_cast< Token* >( outputBuffer ) ); }

    void* getMemoryAddress() { return &m_data; }
    const void* getMemoryAddress() const { return &m_data; }

  private:
    Token m_data;
  };

  /**
   * special version for bool: making sure that the size of bool corresponds to
   * the size of a well known plain type
   */
  template <>
  class AttributeTValue<bool> : public AttributeBaseValue
  {
  public:
    AttributeTValue() : m_data() {}
    AttributeTValue(const bool&v) : m_data() {m_data.m_bool=v;}
    bool& value() { return m_data.m_bool; }
    const bool& value() const { return m_data.m_bool; }
    void getValue(bool&v) const{ v = m_data.m_bool; }
    void setValue(const bool&v) { m_data.m_bool = v; }
    std::string getValueAsString() const;
    void setValueAsString(std::string);
    bool isEqual(const AttributeBaseValue& rhs) const;
    void print(std::ostream&os) const;

    void setValueFromData( const void* data ) { m_data.m_bool = *( static_cast< const bool*>( data ) ); }
    void cloneData( void** outputAddress ) const { *outputAddress = new bool(m_data.m_bool); }
    void copyData( void* outputBuffer ) const { *( static_cast< bool* >( outputBuffer ) ) = m_data.m_bool; }
    void* getMemoryAddress();
    const void* getMemoryAddress() const;

  private:
    union data_t
    {
      bool m_bool;
      unsigned int m_pad;
    };

    data_t m_data;
  };
  
}

#include "Attribute.inl"

#endif
