#include "AttributeTypes.h"
#include "AttributeExceptions.h"
#include "AttributeSpecification.h"

#include <cassert>
#include <sstream>

//// template definitions

template<class T> 
void pool::Attribute::setValue(const T& v)
{
  pool::AttributeTValue<T> *t = dynamic_cast< pool::AttributeTValue<T> * >(m_attr.get());

  if(t)
    t->setValue(v);
  else
    throw pool::attribute_bad_type(stored_name(), stored_typename(), pool::AttributeStaticTypeInfo<T>::type_name()); 
}

template<class T>
void pool::Attribute::getValue(T& v) const
{
  pool::AttributeTValue<T> *t = dynamic_cast< pool::AttributeTValue<T> * >(m_attr.get());

  if(t) t->getValue(v);
  else
    throw pool::attribute_bad_type(stored_name(), stored_typename(), pool::AttributeStaticTypeInfo<T>::type_name());
}


inline pool::Attribute::Attribute(const pool::Attribute& r) : m_attr(r.m_attr), m_spec(r.m_spec) {}

inline pool::Attribute& pool::Attribute::operator = (const pool::Attribute &r)
{
  m_attr = r.m_attr;
  m_spec = r.m_spec;
  return *this;
}

inline void
pool::Attribute::shareData(const pool::Attribute &r)
{
  m_attr = r.m_attr;
}


inline const pool::AttributeSpecification& pool::Attribute::spec() const
{
  if(m_spec) return *m_spec;
  else
    throw pool::attribute_bad_type(stored_name(), stored_typename(),"");
}

inline std::string pool::Attribute::getValueAsString() const
{
  if(m_attr)
    // FIXME: conversion error handling
    return m_attr->getValueAsString();
  else
    throw attribute_bad_type(stored_name(), stored_typename(),"");
}

inline void pool::Attribute::setValueAsString(std::string str)
{
  if(m_attr)
    // FIXME: conversion error handling
    m_attr->setValueAsString(str);
  else
    throw pool::attribute_bad_type(stored_name(),stored_typename(),"");
}

inline pool::typename_string_t pool::Attribute::stored_typename() const
{
  if(m_spec) return m_spec->type_name();
  else
    return "<void>";
}

inline std::string pool::Attribute::stored_name() const
{
  if(m_spec) return m_spec->name();
  else
    return "<unnamed>";
}

inline
pool::Attribute::Attribute(boost::shared_ptr<pool::AttributeBaseValue> ap,
			   const pool::AttributeSpecification&spec) : m_attr(ap), m_spec(&spec)
{
  // need .get() for Windows port:  the implicit conversion to
  // const int does not work there.

  assert(m_attr.get() != 0); // FIXME: maybe throw an internal error?
}

inline
void pool::Attribute::print(std::ostream&os) const
{
  m_attr->print(os);
}

inline
bool pool::Attribute::operator == (const Attribute& rhs) const
{
  return m_attr->isEqual(*rhs.m_attr);
}

inline
bool pool::Attribute::operator != (const Attribute& rhs) const
{
  return !((*this) == rhs);
}


// ----------- attribute implementation --------------

template<class T>
inline
std::string pool::AttributeTValue<T>::getValueAsString() const
{
  //FIXME: conversion error handling
  std::ostringstream os;
  os << m_data;
  return os.str();
}

namespace pool {
template<>
inline
std::string AttributeTValue<std::string>::getValueAsString() const
{
  return m_data;
}
}

template<class T>
inline
void pool::AttributeTValue<T>::setValueAsString(std::string s)
{
  //FIXME: conversion error handling
  std::istringstream is( s.c_str() );
  is >> m_data;
}

namespace pool {
template<>
inline
void AttributeTValue<std::string>::setValueAsString(std::string s)
{
  m_data = s;
}
}

template<class T>
bool pool::AttributeTValue<T>::isEqual(const AttributeBaseValue& rhs) const
{
  const AttributeTValue<T> *rp = dynamic_cast<const AttributeTValue<T> *>(&rhs);
  if(rp)
    return m_data == rp->m_data;
  else
    return false;
}

template<class T>
inline
void pool::AttributeTValue<T>::print(std::ostream&os) const
{
  os << m_data;
}

template<class T>
inline
void* pool::AttributeTValue<T>::getMemoryAddress()
{
  return &m_data;
}

template<class T>
inline
const void* pool::AttributeTValue<T>::getMemoryAddress() const
{
  return &m_data;
}


// ----------------------- bool

inline
std::string pool::AttributeTValue<bool>::getValueAsString() const
{
  std::ostringstream os;
  os << m_data.m_bool;
  return os.str();
}

inline
void pool::AttributeTValue<bool>::setValueAsString(std::string s)
{
  std::istringstream is( s.c_str() );
  is >> m_data.m_bool;
}

inline
bool pool::AttributeTValue<bool>::isEqual(const AttributeBaseValue& rhs) const
{
  const AttributeTValue<bool> *rp = dynamic_cast<const AttributeTValue<bool> *>(&rhs);
  if(rp)
    return m_data.m_bool == rp->m_data.m_bool;
  else
    return false;
}

inline
void pool::AttributeTValue<bool>::print(std::ostream&os) const
{
  os << m_data.m_bool;
}

inline
void* pool::AttributeTValue<bool>::getMemoryAddress()
{
  return &m_data.m_bool;
}

inline
const void* pool::AttributeTValue<bool>::getMemoryAddress() const
{
  return &m_data.m_bool;
}

inline
bool pool::AttributeTValue<pool::Token>::isEqual(const AttributeBaseValue& rhs) const
{
  const AttributeTValue<pool::Token> *rp = dynamic_cast<const AttributeTValue<pool::Token> *>(&rhs);
  if(rp)
    return m_data.equal(rp->m_data);
  else
    return false;
}

inline
void pool::AttributeTValue<pool::Token>::print(std::ostream&os) const
{
  os << getValueAsString();
}
