#ifndef ATTRIBUTELIST_H
#define ATTRIBUTELIST_H

/// @file 
/// @author Jakub.Moscicki@cern.ch (kuba)


# include "Attribute.h"
# include "AttributeListSpecification.h"
# include "boost/shared_ptr.hpp"

/**
 * AttributeList represent a logical row of attributes in a metadata table.
 */

namespace pool {
  class AttributeList
  {
  public:

    /// Construct attribute list according to the specification.
    /// The link to the specification is kept internally.
    /// User must make sure that the lifetime of the specification covers
    /// the lifetime of the attribute list.
    explicit AttributeList(const AttributeListSpecification&);

    /// Construct attribute list according to the specification.
    /// User is released from lifetime management issues for 
    /// the specification. 
    explicit AttributeList(boost::shared_ptr<const AttributeListSpecification> ptr);

    /// Construct empty attribute list with empty specification.
    AttributeList();

    /// Destructor
    ~AttributeList();

    /// Copy attribute lists.
    AttributeList(const AttributeList& rhs);
        
    /// @name Throw attribute_not_found if attribute not in the list.
    //@{
    const Attribute& operator[](std::string name) const;
    Attribute& operator[](std::string name);
    //@}

    /// Number of attributes.
    int size() const;
    
    /// @name Iterate in original push_back order.
    //@{
    typedef std::vector<Attribute>::iterator iterator;
    iterator begin();
    iterator end();

    typedef std::vector<Attribute>::const_iterator const_iterator;
    const_iterator begin() const;
    const_iterator end() const;
    //@}

    /// Two AttributeLists are equal if their specifications are equal and corresponding attributes are equal.
    bool operator == (const AttributeList& rhs) const;
    bool operator != (const AttributeList& rhs) const;    

    /// Pretty print.
    void print(std::ostream& os) const;

    /// Do not throw. Return empty specification if this default constructed.
    const AttributeListSpecification& attributeListSpecification() const;

  private:

    void initialize(const pool::AttributeListSpecification& spec); 

    boost::shared_ptr<const AttributeListSpecification> m_attr_spec_ptr;

    const AttributeListSpecification* m_attr_spec;
    
    std::vector<Attribute> m_attr_vector; // optimization for size possible: AttributeList could be just a proxy to the attribute vector which would be stored internally in the collection
  };
}

#endif
