#ifndef ATTRIBUTELISTSPECIFICATION_H
#define ATTRIBUTELISTSPECIFICATION_H

/// @file 
/// @author Jakub.Moscicki@cern.ch (kuba)

# include "AttributeSpecification.h"

# include "AttributeExceptions.h"

# include <map>
# include <vector>

namespace pool {

  class AttributeListSpecification;


  /**
   * Specifies the type and name of attributes 
   * 
   * Individual attributes are ordered by attribute name if accessed by operator[name]. 
   * 
   * Original order is available via operator[index]. Iterator iterates in original order.
   *
   * Desing assumption: once specification is constructed and started being used 
   * (by the attribute lists) it should not be mutated (changed) anymore.
   * If you do mutate the specification you may get unexpected behaviour from the 
   * lists which use it.
   */
  class AttributeListSpecification
  {
  private:
   
  public:
    
    /// Empty specification. Use push_back to add attributes.
    AttributeListSpecification();

    /// Destructor
    ~AttributeListSpecification();

    /// Throw attribute_duplicate if attribute already exists.
    /// Usage: spec.push_back("a","int")
    void push_back(std::string attr_name, typename_string_t type_name);
    
    /// Throw attribute_duplicate if attribute already exists.
    /// Usage: spec.push_back("a", typeid(int) )
    void push_back(std::string attr_name, const std::type_info& typeInfo );
    
    /// Throw attribute_duplicate if attribute already exists.
    /// Usage: spec.push_back<int>("a")
    template <class T> void push_back(std::string attr_name);
    
    /// @name Read specs.
    /// Throw attribute_not_found if index out of range or name not found.
    //@{
    const AttributeSpecification& operator[](std::string attr_name) const;
    const AttributeSpecification& operator[](size_t i) const;
    //@}
    
    /// Return index number of attr name.
    /// Return pool::IDX_NOT_FOUND if index out of range
    size_t index(std::string attr_name) const;

    /// Deprecated. Use equals instead.
    bool operator == (const AttributeListSpecification& rhs) const;
    bool operator != (const AttributeListSpecification& rhs) const;

    /// Compare the name and type of the attributes. The original push_back order is ignored.
    /// See also: std::equals algorithm.
    bool equals(const AttributeListSpecification& rhs) const;

    /// Return true if all attributes of rhs are present in this (rhs is subset of this).
    /// The original push_back order does not matter.
    /// See also: std::includes algorithm.
    bool includes(const AttributeListSpecification& rhs) const;

    /// Number of attributes.
    size_t size() const;

    /// Create a new specification by appending rhs specification to this,
    /// merging duplicate attributes. 
    /// If duplicate attributes have different type then throw attribute_duplicate exception.
    /// Original attribute order is maintained (operation is stable).
    AttributeListSpecification append_and_merge(const AttributeListSpecification& rhs) const;

    /// @name Iterate in sorted order (i.e. a unique order which does not depend on push_back order).
    /// Iterator points to a pair: (name, index). 
    /// Use operator[index] to access AttributeSpecification.
    //@{  
    typedef std::map<std::string, size_t>::iterator ordered_index_iterator;
    ordered_index_iterator ordered_index_begin();
    ordered_index_iterator ordered_index_end();

    typedef std::map<std::string, size_t>::const_iterator const_ordered_index_iterator; 
    const_ordered_index_iterator ordered_index_begin() const;
    const_ordered_index_iterator ordered_index_end() const;
    //@}

    /// @name Iterate in original push_back order. Iterator points to AttributeSpecification.
    //@{
    typedef std::vector<AttributeSpecification>::iterator iterator;
    iterator begin();
    iterator end();

    typedef std::vector<AttributeSpecification>::const_iterator const_iterator;
    const_iterator begin() const;
    const_iterator end() const;
    //@}

    /// Pretty print.
    void print(std::ostream&) const;

  private:

    std::map<std::string, size_t> m_specmap;
    std::vector<AttributeSpecification> m_index; 
  };
  
}

/// example: spec.push_back<int>("a")
template <class T> void pool::AttributeListSpecification::push_back(std::string attr_name)
{
  push_back(attr_name, typeid(T) );
}


#endif
