#ifndef COLLECTION_COLLECTION_H
#define COLLECTION_COLLECTION_H

#include "Reflex/Type.h"
#include "Collection/CollectionProxy.h"

namespace pool {

  /** 
    @class Collection Collection.h Collection/Collection.h

    @brief User interface to collections of objects stored in a database.
      
    A type-safe class used to store, retrieve and update
    collections of database objects and their associated 
    meta data. The class is templated according to the 
    type of data object that it stores.
  */

  template <class T>
  class Collection 
  {
  public:

    /**
      @class Iterator.

      @brief User interface for navigation of Collection objects.

      An iterator class for navigating over data objects stored in
      a Collection object.
    */

    class Iterator 
    {
    
    private:
    
      /**
        A friend prototype declaration of the Collection class so that 
        a collection can create its own iterator.
      */
      friend class Collection;
      friend class CollectionSvc;
      /**
         Default constructor.
      */
      Iterator();

      /**
        Constructor.
        @param proxy Proxy iterator
      */
      Iterator( const CollectionProxy::Iterator& proxy );

      /**
        Assignment operator: Not implemented.
      */
      Iterator& operator=( const Iterator& );
   
    public:
      
      /**
        Copy constructor.
      */
      Iterator( const Iterator& rhs );

      /** 
        Destructor.
      */
      ~Iterator() {}

      /**
        Constant data object reference accessor operator.
      */
      const T& operator*() const;

      /**
        Data object reference accessor operator.
      */
      T& operator*();
      
      /**
        Constant data object pointer accessor operator.
      */
      const T* operator->() const;
      
      /**
        Data object pointer accessor operator.
      */
      T* operator->();

      /**
        Retrieve next data object in collection.
      */
      bool next();
 
      /**
        Return meta data list for current data object.
      */
      const AttributeList& attributeList() const;

      /**
	Return the underlying reference
      */
      Ref<T>& ref();
      const Ref<T>& ref() const;

      /**
        Check if iterator is valid.
      */
      bool isValid() const;

    private:
      /// Seal class corresponding to this type
      const ROOT::Reflex::Type  m_thisClassType;
      /// Templated smart reference to current data object. 
      mutable Ref<T> m_ref;
      /// proxy iterator
      CollectionProxy::Iterator	m_proxy;
    };

    /** 
      Constuctor. 
        
      @param dataSvc POOL data service.
      @param type Database-specific type of collection. Possible types:
      - Explicit collections: "MySQLCollection", "RootCollection", "MemoryCollection"
      - Implicit collections: "ImplicitCollection"
      - Multi collections (collection of collections): "MySQLMultiCollection", "RootMultiCollection"
      .
      @param connection Location of database. See documentation of the concrete implementations for further information.
      @param persistentName Persistent name of collection.
      @param openMode Open mode for collection.
    */
    Collection( IDataSvc* dataSvc,
	        std::string type = "MemoryCollection",
		std::string connection = "a connection string",
		std::string persistentName = "a collection name",
		ICollection::OpenMode openMode = ICollection::CREATE_AND_OVERWRITE )
	  : m_proxy( dataSvc, type, connection, persistentName, openMode )
    { }
    
    /**
      Destructor.
    */
    ~Collection() { }

    /** 
      Return a reference to the collection's associated 
      meta data list specification object.
    */
    AttributeListSpecification& attributeListSpecification()
    {
       return m_proxy.attributeListSpecification();
    }

    /** 
      Return a constant reference to the collection's 
      associated meta data list specification object.
    */
    const AttributeListSpecification& attributeListSpecification() const 
    {
       return m_proxy.attributeListSpecification();
    }

    /// see CollectionProxy for description
    Iterator select( std::string primaryQuery = "",
		     std::string secondaryQuery = "",
		     std::string queryOptions = "" )
    {
       return Iterator( m_proxy.select( primaryQuery, secondaryQuery, queryOptions ) );
    }

    /**
      Add a data object and its associated meta data to the collection.

      @param pref Templated smart reference to the data object.
      @param attrs List of the data object's associated meta data. 
    */
    bool add( const Ref<T>& pref, const AttributeList& attrs );

    /**
      Commit last changes made to the collection.
    */
    bool commit() { return m_proxy.commit(); }

    /**
      Abort last changes made to collection before committing.
    */
    bool rollback() { return m_proxy.rollback(); }

    /**
      Close collection explicitly, aborting uncommitted changes.
    */
    bool close() { return m_proxy.close(); }

    /**
      Reopen collection explictly after it was closed.
    */
    bool open() { return m_proxy.open(); }

    /**
      Check if collection is open.
    */
    bool isOpen() const { return m_proxy.isOpen(); }
    
    /**
       Return persistent name of collection.
    */
    std::string name() const { return m_proxy.name(); }
    
    /**
       Return connection parameters of the collection
    */
    std::string connection() const { return m_proxy.connection(); }
    
    /**
       Return the type ID of the concrete implementation
    */
    std::string type() const { return m_proxy.type(); }
    
    /**
       Return the type ID of the concrete implementation
    */
    pool::ICollection::OpenMode openMode() const { return m_proxy.openMode(); }

  protected:
    /**
       Return the collection object in use
    */
    ICollection* collection() { return m_proxy.collection(); }
    const ICollection* collection() const { return m_proxy.collection(); }

    /**
       Return the dataSvc pointer
    */
    IDataSvc* dataSvc() { return m_proxy.dataSvc(); }
    const IDataSvc* dataSvc() const { return m_proxy.dataSvc(); }

  private:
    
    Collection( IDataSvc*, ICollection* );
    Collection( const CollectionProxy& proxy );
    
    /// Database-specific collection object.
    CollectionProxy	m_proxy;
	 
    friend class CollectionSvc;
    
  };
  
} // ns pool

#include "Collection.inl"

#endif

