#ifndef COLLECTION_COLLECTIONPROXY_H
#define COLLECTION_COLLECTIONPROXY_H

#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/ICollectionQuery.h"

#include "DataSvc/Ref.h"

namespace pool {

   class AttributeList;
   class AttributeListSpecification;
   
  /** 
    @class CollectionProxy CollectionProxy.h Collection/CollectionProxy.h

    @brief User interface to collections of objects stored in a database.
      
    A user class used to store, retrieve and update collections of database
    objects and their associated meta data. The class hides the management
    of tokens from the user, and its associated iterator allows for an
    explicit type checking for a given object and the creation of a Ref<T>
  */

  class CollectionProxy
  {
  public:

    /**
      @class Iterator.

      @brief User interface for navigation of Collection objects.

      An iterator class for navigating over data objects stored in
      a Collection object.
    */

    class Iterator 
    {
    
    private:
    
      /**
        A friend prototype declaration of the Collection class so that 
        a collection can create its own iterator.
      */
      friend class CollectionProxy;
      friend class CollectionSvc;
      /**
         Default constructor.
      */
      Iterator();

      /**
        Constructor.

        @param query Qeury object, results of which will be iterated over (can be NULL)
        @param cursor Cursor object, presumably returned by query. Only if  query is NULL the cursor will be deleted
        @param dataSvc POOL data service.
	
      */
      Iterator( ICollectionQuery* query,
		ICollectionCursor* cursor,
		IDataSvc* dataSvc );

      /**
        Assignment operator: Not implemented.
      */
      Iterator& operator=( const Iterator& );
   
    public:
      
      /**
        Copy constructor.
      */
      Iterator( const Iterator& rhs );

      /** 
        Destructor.
      */
      ~Iterator();

      /**
        Retrieve next data object in collection.
      */
      bool next();
 
      /**
        Return meta data list for current data object.
      */
      const AttributeList& attributeList() const;

     /**
	Checks if the object to which the iterator is currently
        pointing, is of type T
      */
      template<class T> bool isCurrentObjectOfType() const;

      /**
	Construct a reference for the current object
      */
      template<class T> Ref<T>	ref() const;

      IDataSvc*		dataSvc() const { return m_dataSvc; }

      /**
        Check if iterator is valid.
      */
      bool isValid() const;

    private:
      /// Create AttributeList object that is used to pass attribute values to the user
      void	makeAttributeList()  const;

      /// Read and set the main Token from the current cursor row
      void	getTokenFromCursor() const;
      
      /// Iteration query object
      mutable ICollectionQuery* 	m_query;
      /// Cursor for query result set
      mutable ICollectionCursor*	m_cursor; 
      /// POOL data service.
      IDataSvc* 			m_dataSvc;
      /// The underlying token
      mutable Token* 			m_token;
      /// flag indicating iterator has moved and attribute values need to be refreshed
      mutable bool			m_advancedCursor;
      /// AttributeListSpecification for the query result
      mutable AttributeListSpecification* m_attrListSpec;
      /// metadata values for the current row pointed to by the iterator
      mutable AttributeList*		m_attributeList;
      /// Flag indicating whether the current object is valid
      mutable bool 			m_valid;
    };

    /** 
      Constuctor. 
        
      @param dataSvc POOL data service.
      @param type Database-specific type of collection. Possible types:
      - Explicit collections: "MySQLCollection", "RootCollection", "MemoryCollection"
      - Implicit collections: "ImplicitCollection"
      - Multi collections (collection of collections): "MySQLMultiCollection", "RootMultiCollection"
      .
      @param connection Location of database. See documentation of the concrete implementations for further information.
      @param persistentName Persistent name of collection.
      @param openMode Open mode for collection.
    */
    CollectionProxy( IDataSvc* dataSvc,
	        std::string type = "MemoryCollection",
		std::string connection = "a connection string",
		std::string persistentName = "a collection name",
		ICollection::OpenMode openMode = ICollection::CREATE_AND_OVERWRITE );

    CollectionProxy( const CollectionProxy& rhs );
    
    /**
      Destructor.
    */
    ~CollectionProxy();

    /** 
      Return a reference to the collection's associated 
      meta data list specification object.
    */
    AttributeListSpecification& attributeListSpecification();

    /** 
      Return a constant reference to the collection's 
      associated meta data list specification object.
    */
    const AttributeListSpecification& attributeListSpecification() const;

    /**
      Create and initialize an iterator for the collection.

      @param primaryQuery Selection applied to object meta data (see AttributeList)
      @param secondaryQuery Selection applied to collection meta data (see MultiCollection)\n
                            The selection is applied only to the meta data associated with object collections (the leafs).\n
			    The returned iterator enters all subordinate multi collections.
      @param queryOptions Query options; Space separated choice of the following:
       - MySQLCollection: 
         - "FetchOne": The default; Selected rows are not cached.
         - "FetchAll": Selected rows are cached at initialization time.
         - "SELECT <attribute list>": where <attribute list> is a comma 
           separated list of attributes to be read in from persistence;
           "SELECT *" selects all attributes; "SELECT" with no arguments
           means only the Token is selected.
         - "NoToken": Flag to exclude Token selection from persistence.
       - RootCollection:
         - "SELECT <attribute list>" or "ATTRIBUTE_LIST <attribute list>": 
           where <attribute list> is a comma or space separated list of 
           attributes to be read in from persistence.
           "SELECT *" selects all attributes; "SELECT" with no arguments
           means only the Token is selected.
         - ROOT_SELECTION_SERVER: Performs selection on a remote server 
           (experimental):
           e.g. "ROOT_SELECTION_SERVER pcepsft01.cern.ch"\n
           On the remote server the simple server script 
           RootCollection/scripts/server.C has to be started. Type "root server.C+".
       - MultiCollection: In case of a MultiCollection ("RootMultiCollection","MySQLMultiCollection")
                          the options string is passed to the iterators of the selected object collections 
    */
    Iterator select( std::string primaryQuery = "",
		     std::string secondaryQuery = "",
		     std::string queryOptions = "" );

   /**
      Add a data object and its associated meta data to the collection.

      @param pref smart reference to the data object.
      @param attrs List of the data object's associated meta data. 
    */
    bool add( const RefBase& pref,
	      const AttributeList& attrs );

    /**
      Commit last changes made to the collection.
    */
    bool commit();

    /**
      Abort last changes made to collection before committing.
    */
    bool rollback();

    /**
      Close collection explicitly, aborting uncommitted changes.
    */
    bool close();

    /**
      Reopen collection explictly after it was closed.
    */
    bool open();

    /**
      Check if collection is open.
    */
    bool isOpen() const;
    
    /**
       Return persistent name of collection.
    */
    std::string name() const;
    
    /**
       Return connection parameters of the collection
    */
    std::string connection() const;
    
    /**
       Return the type ID of the concrete implementation
    */
    std::string type() const;
    
    /**
       Return the type ID of the concrete implementation
    */
    pool::ICollection::OpenMode openMode() const;

  protected:
    /**
       Return the collection object in use
    */
    ICollection* collection();
    const ICollection* collection() const;

    /**
       Return the dataSvc pointer
    */
    IDataSvc* dataSvc();
    const IDataSvc* dataSvc() const;

    /// create AttributeListSpecification for the open collection when user asks for one
    void	makeAttrListSpec() const;
    
  private:
    
    CollectionProxy( IDataSvc*, ICollection* );
    
    /// Database-specific collection object.
    mutable ICollection* 	m_collection;
    /// The data service
    IDataSvc* 		m_dataSvc;

    /// Is this collection freshly created? (means no schema defined yet)
    mutable bool	m_isNewCollection;

    /// Cached specification for the collection
    mutable AttributeListSpecification	*m_attrListSpec;

 friend class CollectionSvc;
    
  };
  
} // ns pool

#include "CollectionProxy.inl"

#endif

