#include "CollectionBase/CollectionService.h"
#include "CollectionBase/CollectionFactory.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/ICollectionCursor.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/FCEntry.h"

namespace pool { class ISession; }

#include "POOLCore/Exception.h"

#include "SealKernel/Context.h"

#include "boost/tokenizer.hpp"


DEFINE_SEAL_COMPONENT ( pool::CollectionService, "POOL/Services/CollectionService" )


pool::CollectionService::CollectionService( seal::Context* context ) 
  : seal::Service( context, classContextKey() ),
    m_fileCatalog(0),
    m_collectionCatalog(0)
{
  this->setDefaultFileCatalog();

  this->setDefaultCollectionCatalog();
}


pool::CollectionService::CollectionService( seal::Context* context, 
                                            const std::string& key )
  : seal::Service( context, key ),
    m_fileCatalog(0),
    m_collectionCatalog(0)
{
  this->setDefaultFileCatalog();

  this->setDefaultCollectionCatalog();
}


pool::CollectionService::~CollectionService()
{
  if ( m_fileCatalog )
  {
    delete m_fileCatalog;
    m_fileCatalog = 0;
  }

  if ( m_collectionCatalog )
  {
    delete m_collectionCatalog;
    m_collectionCatalog = 0;
  }
}


pool::ICollection*
pool::CollectionService::create( pool::ICollectionDescription& description, bool overwrite )
{
  if ( description.name().size() == 0 )  {
    std::string errorMsg = "Must specify name of collection in description input argument.";
    throw pool::Exception( errorMsg,
                           "CollectionService::create", 
                           "CollectionService" );
  }

  if ( description.type().size() == 0 )  {
    std::string errorMsg = "Must specify type of collection in description input argument.";
    throw pool::Exception( errorMsg,
                           "CollectionService::create", 
                           "CollectionService" );
  }

  if ( description.type() == "ImplicitCollection" )  {
    std::string errorMsg = 
         "Can only open a collection of type 'ImplicitCollection' for read transtions.";
    throw pool::Exception( errorMsg,
                           "CollectionService::create", 
                           "CollectionService" );
  }

  pool::ICollection::OpenMode openMode = overwrite? pool::ICollection::CREATE_AND_OVERWRITE : pool::ICollection::CREATE;
  return  pool::CollectionFactory::get()->create( description.type(), description, openMode );
}


pool::ICollection*
pool::CollectionService::createAndRegister( pool::ICollectionDescription& description,
                                            bool overwrite,
                                            std::string logicalName,
                                            MetaDataEntry* metadata )
{
  if ( description.name().size() == 0 )
  {
    std::string errorMsg = "Must specify name of collection in description input argument.";
    throw pool::Exception( errorMsg,
                           "CollectionService::createAndRegister", 
                           "CollectionService" );
  }

  if ( description.type().size() == 0 )
  {
    std::string errorMsg = "Must specify type of collection in description input argument.";
    throw pool::Exception( errorMsg,
                           "CollectionService::createAndRegister", 
                           "CollectionService" );
  }

  if ( description.type() == "ImplicitCollection" )
  {
    std::string errorMsg = "Cannot register a collection of type 'ImplicitCollection' in a collection catalog.";
    throw pool::Exception( errorMsg,
                           "CollectionService::createAndRegister", 
                           "CollectionService" );
  }

  return CollectionFactory::get()->createAndRegister( description,
						      m_collectionCatalog,
						      overwrite,
						      logicalName,
						      metadata );
}


bool
pool::CollectionService::registerExisting( const std::string& name,
                                           const std::string& type,
                                           std::string connection,
                                           std::string logicalName,
                                           MetaDataEntry* metadata )
{
  if ( name.size() == 0 )
  {
    std::string errorMsg = "Must specify name of collection as input.";
    throw pool::Exception( errorMsg,
                           "CollectionService::registerExisting", 
                           "CollectionService" );
  }

  if ( type.size() == 0 )
  {
    std::string errorMsg = "Must specify type of collection as input.";
    throw pool::Exception( errorMsg,
                           "CollectionService::registerExisting", 
                           "CollectionService" );
  }

  if ( type == "ImplicitCollection" )
  {
    std::string errorMsg = "Cannot register a collection of type 'ImplicitCollection' in a collection catalog.";
    throw pool::Exception( errorMsg,
                           "CollectionService::registerExisting",
                           "CollectionService" );
  }

  CollectionDescription description( name, type, connection );

  return pool::CollectionFactory::get()->registerExisting( description,
							   m_collectionCatalog,
							   logicalName,
							   metadata );
}


pool::ICollection* 
pool::CollectionService::handle( const std::string& name,
                                 const std::string& type,
                                 std::string connection,
                                 bool readOnly,
                                 pool::ISession* session ) const
{
  if ( ( type == "ImplicitCollection ") && (! readOnly ) )
  {
    std::string errorMsg = "Cannot open a collection of type 'ImplicitCollection' for updates.";
    throw pool::Exception( errorMsg,
                           "CollectionService::handle",
                           "CollectionService" );
  }

  pool::CollectionDescription description( name, type, connection );

  pool::ICollection::OpenMode openMode = readOnly? ICollection::READ : ICollection::UPDATE;

  pool::ICollection* collection
     = pool::CollectionFactory::get()->create( type, description, openMode, session );

  return collection;
}


pool::ICollection* 
pool::CollectionService::openWithPhysicalName( const std::string& physicalName, bool readOnly )
{
  return pool::CollectionFactory::get()->openWithPhysicalName( physicalName,
							       m_collectionCatalog,
							       readOnly );
}


pool::ICollection* 
pool::CollectionService::openWithLogicalName( const std::string& logicalName, bool readOnly )
{
  return pool::CollectionFactory::get()->openWithLogicalName( logicalName,
							      m_collectionCatalog,
							      readOnly );
}


pool::ICollection* 
pool::CollectionService::openWithGuid( const std::string& guid, bool readOnly )
{
   return pool::CollectionFactory::get()->openWithGuid( guid,
							m_collectionCatalog,
							readOnly );
}


bool
pool::CollectionService::exists( const std::string& name,
                                 const std::string& type,
                                 std::string connection,
                                 bool setForUpdate,
                                 bool checkChildFragments ) const
{
  return ( this->handle( name, type, connection )->exists( name, setForUpdate, checkChildFragments ) );
}


bool
pool::CollectionService::drop( const std::string& name,
                               const std::string& type,
                               std::string connection,
                               bool dropChildFragments,
                               bool ignoreExternalDependencies )
{
  return ( this->handle( name, type, connection, false )->drop( name, dropChildFragments, ignoreExternalDependencies ) );
}


bool
pool::CollectionService::rename( const std::string& oldName,
                                 const std::string& newName,
                                 const std::string& type,
                                 std::string connection )
{
  return ( this->handle( oldName, type, connection, false )->rename( oldName, newName ) );
}


bool
pool::CollectionService::grantToUser( const std::string& userName,
                                      pool::ICollection::Privilege privilege,
                                      const std::string& name,
                                      const std::string& type,
                                      std::string connection,
                                      bool grantForChildFragments )
{
  return ( this->handle( name, type, connection, false )->grantToUser( userName, 
                                                                       privilege, 
                                                                       name, 
                                                                       grantForChildFragments ) );
}


bool
pool::CollectionService::revokeFromUser( const std::string& userName,
                                         pool::ICollection::Privilege privilege,
                                         const std::string& name,
                                         const std::string& type,
                                         std::string connection,
                                         bool revokeForChildFragments )
{
  return ( this->handle( name, type, connection, false )->revokeFromUser( userName, 
                                                                          privilege, 
                                                                          name, 
                                                                          revokeForChildFragments ) );
}


bool
pool::CollectionService::grantToPublic( const std::string& name,
                                        const std::string& type,
                                        std::string connection,
                                        bool grantForChildFragments )
{
  return ( this->handle( name, type, connection, false )->grantToPublic( name, grantForChildFragments ) );
}


bool
pool::CollectionService::revokeFromPublic( const std::string& name,
                                           const std::string& type,
                                           std::string connection,
                                           bool revokeForChildFragments )
{
  return ( this->handle( name, type, connection, false )->revokeFromPublic( name, revokeForChildFragments ) );
}


bool 
pool::CollectionService::setDefaultFileCatalog()
{
  if( m_fileCatalog )
  {
    m_fileCatalog->disconnect();
    delete m_fileCatalog;
  }

  m_fileCatalog = new pool::IFileCatalog();

  const char* chr = getenv("POOL_WRITE_FILE_CATALOG");
  if(chr && strlen(chr))
  {
    m_fileCatalog->setWriteCatalog( chr );
  }
  else
  {
    m_fileCatalog->setWriteCatalog( "xmlcatalog_file:PoolFileCatalog.xml" );
  }

  chr = getenv("POOL_READ_FILE_CATALOGS");
  if(chr && strlen(chr))
  {
    std::string str = chr;
    typedef boost::tokenizer<boost::char_separator<char> > Tizer;
    boost::char_separator<char> sep(" ,");
    Tizer tizer( str, sep);
    for( Tizer::iterator token=tizer.begin();
         token!=tizer.end();
         ++token)
    {
      m_fileCatalog->addReadCatalog( *token );
    }
  }
  else
  {
    m_fileCatalog->addReadCatalog("xmlcatalog_file:PoolFileCatalog.xml");
  }
  
  m_fileCatalog->connect();

  return true;
}


bool
pool::CollectionService::setDefaultCollectionCatalog()
{
  if( m_collectionCatalog )
  {
    m_collectionCatalog->disconnect();
    delete m_collectionCatalog;
  }

  m_collectionCatalog = new pool::IFileCatalog();

  const char* chr = getenv("POOL_WRITE_COLLECTION_CATALOG");
  if(chr && strlen(chr))
  {
    m_collectionCatalog->setWriteCatalog( chr );
  }
  else
  {
    m_collectionCatalog->setWriteCatalog( "xmlcatalog_file:PoolCollectionCatalog.xml" );
  }

  chr = getenv("POOL_READ_COLLECTION_CATALOGS");
  if(chr && strlen(chr))
  {
    std::string str = chr;
    typedef boost::tokenizer<boost::char_separator<char> > Tizer;
    boost::char_separator<char> sep(" ,");
    Tizer tizer( str, sep);
    for( Tizer::iterator token=tizer.begin();
         token!=tizer.end();
         ++token)
    {
      m_collectionCatalog->addReadCatalog( *token );
    }
  }
  else
  {
    m_collectionCatalog->addReadCatalog("xmlcatalog_file:PoolCollectionCatalog.xml");
  }
  
  m_collectionCatalog->connect();

  return true;
}


bool
pool::CollectionService::setWriteFileCatalog( const std::string& name )
{
  m_fileCatalog->setWriteCatalog( name );

  return true;
}


bool
pool::CollectionService::addReadFileCatalog( const std::string& name )
{
  m_fileCatalog->addReadCatalog( name );

  return true;
}


pool::IFileCatalog*
pool::CollectionService::getFileCatalog( )
{
   return m_fileCatalog;
}



bool 
pool::CollectionService::setWriteCollectionCatalog( const std::string& name )
{
  m_collectionCatalog->setWriteCatalog( name );
  
  return true;
}


bool
pool::CollectionService::addReadCollectionCatalog( const std::string& name )
{
  m_collectionCatalog->addReadCatalog( name );

  return true;
}

pool::IFileCatalog*
pool::CollectionService::getCollectionCatalog( )
{
   return m_collectionCatalog;
}



pool::ICollectionCursor*
pool::CollectionService::getCatalogCollectionCursor( std::string collectionLevelQuery,
	                                             std::string rowLevelQuery,
	                                             std::vector<std::string>* tokenOutputList,
	                                             std::vector<std::string>* attribOutputList,
	                                             int rowCacheSize ) const
{
  return pool::CollectionFactory::get()->getCatalogCollectionCursor( m_collectionCatalog,
                                                                    collectionLevelQuery,
                                                                    rowLevelQuery,
                                                                    tokenOutputList,
                                                                    attribOutputList,
                                                                    rowCacheSize );
}
