#ifndef POOL_CACHELINK_H
#define POOL_CACHELINK_H 

#include "PersistencySvc/Placement.h"
#include "DataSvc/IRegisteredRef.h"
#include "DataSvc/SharedAnyPtr.h"

namespace pool    {

  // forward declarations
  class CacheEntry;
  class ICacheSvc;
  class IDataSvc;
  class AnyType;
  class Token;


  /** @class CacheLink
      Describes the binding with a particular cache. Acts as client of the bound
      cache for retrieve/write/update/delete objects.
      @author Giacomo Govi
      @date 12/08/2003
  */
  class CacheLink : virtual public IRegisteredRef {

    public:

    // default constructor
    CacheLink();

    // IDataSvc ctor
    CacheLink(IDataSvc*);

    // ICacheSvc ctor
    CacheLink(ICacheSvc*);

    // copy ctor
    CacheLink(const CacheLink&);

    // assignment operator
    CacheLink& operator=(const CacheLink&);

    // custom copy operator with client type checking
    CacheLink& copy(const AnyType&,const CacheLink&);

    // vitual dtor
    virtual ~CacheLink();

    // bool operator
    operator bool() const;

    // open a connection with the specified cache service 
    bool connect(ICacheSvc*);

    // link a specific cache line defined by a token, with client type checking
    bool bindCacheEntry(const AnyType&, const Token&);

    // link a specific cache line defined by an object pointer
    bool bindCacheEntry(const SharedAnyPtr&);

    // retrieve the object pointer from the bound cache line
    const SharedAnyPtr& getObject() const;

    // retrieve the object pointer as a specific type from the bound cache line
    //void* getObject(const std::type_info*) const;

    // retrieve the token from the bound cache line
    Token* getToken() const;

    // send a mark write request to the cache
    bool sendWriteRequest(const Placement&) const;

    // send a mark multi write request to the cache
    bool sendMultiWriteRequest(const Placement&);

    // send a mark update request to the cache
    bool sendUpdateRequest() const;

    // send a mark delete request to the cache
    bool sendDeleteRequest() const;

    // reset the object of the bound cache line
    bool resetObject() const;

    // retrieve the delete object policy of the cache
    bool deleteObjectFlag() const;

    // check if the object is loaded in the bound cache
    bool isObjectLoaded() const;

    // disconnect the cache svc in use  
    bool disconnect();

    // unlink the cache line in use
    bool unbind();

    // return the cache svc in use
    ICacheSvc* cacheSvc() const;

    // return the cache line in use
    CacheEntry* cacheEntry() const;

    // method to safely cut the connection to the specified cache svc
    void unregisterCacheSvc( ICacheSvc* cacheSvc ); 

    // method to safely cut the connection to the specified cache line
    void unregisterCacheEntry( CacheEntry* cacheEntry);
    
  private:
   
    // cache service in use
    ICacheSvc* m_cacheSvc;

    // cache line bound
    CacheEntry* m_cacheEntry;

    // empty default object
    SharedAnyPtr m_emptyObject;

  };
}

#endif
