// $Header: /cvs/POOL/pool/DataSvc/DataSvc/Ref.h,v 1.43 2004/01/13 16:00:20 govi Exp $
#ifndef POOL_REF_H
#define POOL_REF_H

#include "DataSvc/RefBase.h"
#include <assert.h>

namespace pool    {
    
  /** @class Ref Ref.h
      Smart pointer class for the handling of objects persistency  

      @author Giacomo Govi
      @date 10/10/2002
  */
  template <class T> class Ref : public RefBase {

    public:
    
    // default constructor
    Ref();

    // default constructor
    explicit Ref(DeletePolicy policy);

    // constructor holding the cache service pointer
    explicit Ref(IDataSvc* aDataSvc);

    // constructor holding the cache service pointer
    explicit Ref(IDataSvc& aDataSvc);

    // constructor holding the cache service reference
    explicit Ref(ICacheSvc& aCacheSvc);

    // from a real pointer
    Ref(IDataSvc* aDataSvc, T* anObject);

    // from a real pointer
    Ref(IDataSvc& aDataSvc, T* anObject);

    // from a real pointer
    Ref(ICacheSvc& aCacheSvc, T* anObject);

    // from a Token
    Ref(IDataSvc* aDataSvc, const Token& aToken);

    // from a Token
    Ref(ICacheSvc& aCacheSvc, const Token& aToken);

    // from an externalized token
    Ref(IDataSvc* aDataSvc, const std::string& s);

    // from an externalized token
    Ref(ICacheSvc& aCacheSvc, const std::string& s);

    // copy constructor
    Ref(const Ref<T>&); 

    // extended copy constructor
    template <class C> Ref(const Ref<C>&);  

    // destructor
    virtual ~Ref();

    // Assignment operator with real pointer 
    Ref<T>& operator=(T*);

    // assignment operator
    Ref<T>& operator=(const Ref<T>&);

    // extended assignment operator 
    template <class C> Ref<T>& operator=(const Ref<C>&);

    // copy operator for up/down casting 
    template <class C> Ref<T>& castDynamic(const Ref<C>&);

    // dereference operator
    T* operator->() const;

    // dereference operator
    T& operator*() const;

    // implicit type conversion
    operator T* () const;

    // return the real pointer
    T* ptr() const;

    // 'not' operator for consistency with pointers common use
    bool operator!() const;

    // equality operator between refs and pointers
    inline friend bool operator==(const Ref& aRef, T* aPtr){
      return aRef.object() == aPtr;
    }
    inline friend bool operator==(T* aPtr, const Ref& aRef){
      return aRef.object() == aPtr;
    }
    inline friend bool operator!=(const Ref& aRef, T* aPtr){
      return !(aRef==aPtr);
    }
    inline friend bool operator!=(T* aPtr, const Ref& aRef){
      return !(aPtr==aRef);
    }

    // equality operator between refs
    template <class C>
    bool operator==(const Ref<C>& aRef) const {
      return equals(aRef);
    }
    template <class C>
    bool operator!=(const Ref<C>& aRef) const {
      return !(this->operator==(aRef));     
    }

  private:

    // pointer with throw exception clause
    T* _ptr(bool throw_flag) const;
    
};

}

// default ctor
template <class T> pool::Ref<T>::Ref() : 
                                 RefBase(typeid(T)) {}

// default ctor
template <class T> pool::Ref<T>::Ref(DeletePolicy policy) : 
                                 RefBase(typeid(T),policy) {}

// cache service ctor
template <class T> pool::Ref<T>::Ref(IDataSvc* aDataSvc) : 
                                 RefBase(aDataSvc,typeid(T)) {}

// cache service ctor
template <class T> pool::Ref<T>::Ref(IDataSvc& aDataSvc) : 
                                 RefBase(&aDataSvc,typeid(T)) {}

// cache service ctor
template <class T> pool::Ref<T>::Ref(ICacheSvc& aCacheSvc) : 
                                 RefBase(&aCacheSvc,typeid(T)) {}

// ctor from a real pointer
template <class T> pool::Ref<T>::Ref(IDataSvc* aDataSvc, T* anObject) : 
                                 RefBase(aDataSvc,anObject,typeid(T)) {}

// ctor from a real pointer
template <class T> pool::Ref<T>::Ref(IDataSvc& aDataSvc, T* anObject) : 
                                 RefBase(&aDataSvc,anObject,typeid(T)) {}

// ctor from a real pointer
template <class T> pool::Ref<T>::Ref(ICacheSvc& aCacheSvc, T* anObject) : 
                                 RefBase(&aCacheSvc,anObject,typeid(T)) {}

template <class T> pool::Ref<T>::Ref(const Ref<T>& aRef) :
                                 RefBase(aRef) {}

template <class T> 
template <class C> pool::Ref<T>::Ref(const Ref<C>& aRef) :
                         RefBase(typeid(T)) {
  // compile-time type checking
  C* c = 0; T* t = c; assert(t==0);
  RefBase::_copyNoTypeCheck(aRef);
}

// ctor from a well specified Token
template <class T> pool::Ref<T>::Ref(IDataSvc* aDataSvc, const Token& aToken) :
                                 RefBase(aDataSvc,aToken,typeid(T)) {}

// ctor from a well specified Token
template <class T> pool::Ref<T>::Ref(ICacheSvc& aCacheSvc, const Token& aToken) :
                                 RefBase(&aCacheSvc,aToken,typeid(T)) {}

// ctor from an externalized Token
template <class T> pool::Ref<T>::Ref(IDataSvc* aDataSvc, const std::string& s):
                                 RefBase(aDataSvc,s, typeid(T)) {}

// ctor from an externalized Token
template <class T> pool::Ref<T>::Ref(ICacheSvc& aCacheSvc, const std::string& s):
                                 RefBase(&aCacheSvc,s, typeid(T)) {}

// destrctor
template <class T> pool::Ref<T>::~Ref() {}

// pointer assignment operator
template <class T> pool::Ref<T>& pool::Ref<T>::operator=(T* aPtr) {
  RefBase::_setPtr(aPtr);
  return *this;
}

template <class T> pool::Ref<T>& 
                            pool::Ref<T>::operator=(const Ref<T>& aRef){
  RefBase::_copyNoTypeCheck(aRef);
  return *this;  
}

template <class T> 
template <class C> pool::Ref<T>& 
                            pool::Ref<T>::operator=(const Ref<C>& aRef){
  C* c = 0; T* t(c); assert(t==0); 
  RefBase::_copyNoTypeCheck(aRef);
  return *this;  
}

template <class T> template <class C> pool::Ref<T>& 
                            pool::Ref<T>::castDynamic(const Ref<C>& aRef){
  copy(aRef);
  return *this;  
}

// dereference operator
template <class T> T* pool::Ref<T>::operator->() const {
  return _ptr(true);
}

// dereference operator
template <class T> T& pool::Ref<T>::operator*() const {
  return *_ptr(true);
}

// return the real pointer
template <class T> T* pool::Ref<T>::_ptr(bool throw_flag) const {
  return static_cast<T*>(_pointee(throw_flag)); 
}

// return the real pointer
template <class T> T* pool::Ref<T>::ptr() const  {
  return _ptr(false);  
}

// T* conversion operator
template <class T> pool::Ref<T>::operator T*() const {
  return _ptr(false);
}

// 'not' operator for consistency with pointers common use
template <class T> bool pool::Ref<T>::operator!() const {
  return _ptr(false)==0;
}

#endif  // POOL_REF_H















