#ifndef POOL_REFBASE_H
#define POOL_REFBASE_H

#include <string>
#include "POOLCore/Reference.h"
#include "DataSvc/SharedAnyPtr.h"
#include "DataSvc/CacheLink.h"

namespace pool    {
  
  /** @class RefBase RefBase.h
      Non-typed base class for the templated Ref<T>. 
      Responsible of the protocol of the interaction with the cache.
      @author Giacomo Govi
      @date 10/10/2002
  */
  class RefBase : public Reference {

    public:
      // predefined delete policy
      enum DeletePolicy {DONOT_DELETE, DO_DELETE}; 

    private:
      // default delete policy
      static DeletePolicy _defaultDeletePolicy(); 
  
    public:

      // default constructor
      RefBase();

      // ctor with type + delete policy
      explicit RefBase(const std::type_info& refType, 
              DeletePolicy policy=_defaultDeletePolicy());

      // ctor with data svc + type
      RefBase(IDataSvc* aDataSvc, const std::type_info& refType); 

      // ctor with cache svc + type
      RefBase(ICacheSvc* aCacheSvc, const std::type_info& refType);

      // explicit object constructor
      RefBase(IDataSvc* aDataSvc,
              const AnyPtr& anObjectPtr, 
              const std::type_info& refType);

      // explicit object constructor
      RefBase(ICacheSvc* aCacheSvc,
              const AnyPtr& anObjectPtr, 
              const std::type_info& refType);

      // token (stored object descriptor) constructor
      RefBase(IDataSvc* aDataSvc, const Token& aToken, 
                                  const std::type_info& refType);

      // token (stored object descriptor) constructor
      RefBase(ICacheSvc* aCacheSvc, const Token& aToken, 
                                  const std::type_info& refType);

      // string token constructor
      RefBase(IDataSvc* aDataSvc, const std::string& stringToken, 
                                  const std::type_info& refType);

      // string token constructor
      RefBase(ICacheSvc* aCacheSvc, const std::string& stringToken, 
                                  const std::type_info& refType);

      // copy constructor
      RefBase(const RefBase& aCopy);

      // destructor
      virtual ~RefBase();

      // RefBase assignment operator 
      RefBase& operator=(const RefBase&);

      // default copy operator (defined copy policy)
      void copy(const RefBase&);

      // copy the entire RefBase content, with type checking 
      void copyDeep(const RefBase&);

      // copy only the object pointer, with type checking 
      void copyShallow(const RefBase&);

      // check if the current refbase is referencing some object
      bool isNull() const;

      // method to check if the current ref is referencing an already 
      // loaded transient instance
      bool isOpen() const;

      // check if the current ref represents an entry in the related DB
      //bool isPersistent() const;

      // check the ref is bound to a cache
      bool isConnected() const;

      // connect the ref to a specified cache svc
      void setCacheSvc(ICacheSvc* aCacheSvc) const;   

      // retrieve the related cache svc
      ICacheSvc* cacheSvc() const;

      // retrieve the related cache line
      const CacheEntry* cacheEntry() const;

      // set the specified token as ref content
      void setToken(const Token& aToken);

      // return the associated token
      const Token* token() const;

      // return the string representation of the associated token
      const std::string toString() const;

      // creates a new placement object pointing to the same database address
      // of the embedded token
      Placement placement() const;

      // method to retrieve the object pointer
      const SharedAnyPtr& object() const; 

      // method to retrieve the type
      ROOT::Reflex::Type objectType() const;

      // method to retrieve the ref type
      ROOT::Reflex::Type refType() const;

      // copy from the specified Ref only the object - with type checking
      //void CopyPtr(const RefBase& aRefBase);

      // register for write 
      bool markWrite(const Placement& place) const;

      // register for write for more destinations 
      bool markMultiwrite(const Placement& place) const;

      // register for update 
      bool markUpdate() const;

      // register for delete
      bool markDelete() const;

      // register for write 
      bool mark_write(const Placement& place) const;

      // register for write for more destinations 
      bool mark_multiwrite(const Placement& place) const;

      // register for update 
      bool mark_update() const;

      // register for delete
      bool mark_delete() const;

      // comparison of the RefBase content:
      bool equals(const RefBase &rR) const;

      // reset the whole ref content
      void clear();

      // explicitly delete the object pointee
      void reset();

  protected:

      // copy operator 
      void _copyNoTypeCheck(const RefBase&);

      // return the object pointer
      void* _pointee(bool throw_flag) const;

      // method to set (or reset) the real pointer from a non-cache-bound Ref 
      //void _setPtr(const SharedAnyPtr& aPtr);

       // method to set (or reset) the real pointer from a non-cache-bound Ref 
      void _setPtr(const AnyPtr& aPtr);

  private:

      // copy operator 
      void _copyDeepNoTypeCheck(const RefBase&);

      // copy operator 
      void _copyShallowNoTypeCheck(const RefBase&);

      bool _compareCacheLink(const RefBase& aRefBase) const;

      // cache location associated to the ref
      mutable CacheLink m_cacheLink;
 
      // temporary record for user-created pointer 
      mutable SharedAnyPtr m_ptr;

  private:

      // reference class type
      AnyType m_refType;

      // pointer of the persistent object pointer
      // corrected for the Ref type offset
      mutable void* m_refTypePtr;

      // pointer of the persistent object pointer as it is
      mutable const void* m_fullTypePtr;
      
  };  
    
}

inline bool pool::RefBase::mark_write(const Placement& place) const {
  return markWrite(place);
}

inline bool pool::RefBase::mark_multiwrite(const Placement& place) const {
  return markMultiwrite(place);
}

inline bool pool::RefBase::mark_update() const {
  return markUpdate();
}

inline bool pool::RefBase::mark_delete() const {
  return markDelete();
}

#endif  // POOL_REFBASE_H















