// $Header: /cvs/POOL/pool/DataSvc/src/CacheTable.h,v 1.18 2003/10/13 14:48:52 govi Exp $
#ifndef POOL_CACHETABLE_H
#define POOL_CACHETABLE_H

#include "Hmap_kptr_table.h"
#include "Hmap_ktoken_table.h"
#include "Hset_list.h"

namespace pool    {

  class AnyPtr;
  class CacheEntryImp;
  class Token;

  /** @class CacheTable CacheTable.h
      Type-defined container for the cache lines bookkeeping. 
      Consists in a hash_set list of locations 
      and two hash_maps mapping the entries against their content in terms of 
      object pointer and tokens. Allows to search the cachelocation by object pointer or by token value.

      @author Giacomo Govi
      @date 25/07/2002
   **/
  class CacheTable {
  public:  
      // Object pointer table type
      typedef pool::Hmap_kptr_table<CacheEntryImp*> KeyPtrTable;

      // Token table type
      typedef pool::Hmap_ktoken_table<CacheEntryImp*> KeyTokenTable;

      // Entry list type
      typedef pool::Hset_list<KeyPtr> EntryList;
	  
    public:

      // Object pointer table iterator type
      typedef KeyPtrTable::const_iterator const_ptr_iterator;
 
      // Token table iterator type
      typedef KeyTokenTable::const_iterator const_token_iterator;

      // Entry list iterator type
      typedef EntryList::const_iterator const_list_iterator;      

      // default constructor
      CacheTable() : theTokenTable(),
	             thePtrTable(),
	             theEntryList() {}

      // constructor defining initial capacity
      CacheTable(unsigned int sz): 
		       theTokenTable(sz), 
	               thePtrTable(sz),
                       theEntryList(sz) {}

      // virtual dtor
      virtual ~CacheTable(){}

      // retrieve the cache location holding a specified object pointer
      CacheEntryImp* getEntry(const AnyPtr& aPtr);

      // retrieve the cache location holding a specified token
      CacheEntryImp* getEntry(const Token& aToken);

      // remove from the object ptr table the entry of a specified pointer
      bool unsetEntry(const AnyPtr& aPtr);
 
      // remove from the token table the entry of a specified token
      bool unsetEntry(const Token& aToken);

      // remove from all tables a specified cache location
      bool remove(const CacheEntryImp* aCacheEntryImp);

      // insert the specified cache line in the tables. 
      bool insert(const CacheEntryImp* aCacheEntryImp);

      // set a cache location for a specified Token in the Token table
      bool setEntry(const Token& aToken, CacheEntryImp* aCacheEntryImp);

      // set a cache location for a specified pointer in the object ptr table
      bool setEntry(const AnyPtr& aPtr, CacheEntryImp* aCacheEntryImp);

      // verify if a specified cache location is in the current cache
      bool find(const CacheEntryImp* aCacheEntryImp);

      // location list iterators
      const_list_iterator list_begin();
      const_list_iterator list_end();

      // object pointer table iterators
      const_ptr_iterator ptr_iter_begin(); 
      const_ptr_iterator ptr_iter_end();   
   
      // token table iterators
      const_token_iterator token_iter_begin();      
      const_token_iterator token_iter_end();      

      // delete all the entries from the tables
      void clear();

      // retrieve the number of elements of the cache line list.
      unsigned int size();

    private:

      //token table instance
      KeyTokenTable theTokenTable;

      // object pointer table instance
      KeyPtrTable thePtrTable;

      // location list instance
      EntryList theEntryList;

  };

}

#endif  // POOL_CACHETABLE_H


