//$Id: MySQLFileCatalog.cpp,v 1.94 2007/07/05 15:38:46 dirkd Exp $
/*
  Concrete implementation of the MySQL catalog. 
  @author: Zhen Xie
*/
#include "MySQLImpl.h"
#ifndef POOL_GUID_H
#include "POOLCore/Guid.h"
#endif
#ifndef POOL_FCMYSQLEXCEPTION_H
#include "MySQLCatalog/FCMySQLException.h"
#endif
#ifndef POOL_MYSQLURIPARSER_H
#include "MySQLURIParser.h"
#endif
#include "MySQLHelpers.h"
#include "POOLCore/PoolMessageStream.h"
#include "FileCatalog/FCSystemTools.h"
#include "./QueryParser.h"
namespace pool {
  
  MySQLFileCatalog::MySQLFileCatalog():m_pimpl(new MySQLFileCatalog::MySQLImpl){}
  MySQLFileCatalog::~MySQLFileCatalog(){}
  
  void MySQLFileCatalog::connect() {
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    MySQLURIParser p(m_url);
    try{
      p.parse();
      if ( !m_transactionsta.test(1) ){ //if connection flag is not set
 	mysqllog <<seal::Msg::Info<< "connect():Connecting to the catalog "<< p.path();
        m_pimpl->connect( p.host(), p.path(), p.username(), p.passwd(), p.hostport() );
	m_transactionsta.set(1);//set connection flag to 1
      }else{
	throw FCTransactionException("MySQLFileCatalog::connect","catalog is already connected");
      }
    }catch(const FCillegalContactStringException& er){
      throw er; 
    }catch(mysqlpp::BadQuery& er){
      throw FCconnectionException("MySQLFileCatalog::connect", er.error);
    }catch(...){
      throw;
    }
  }
  
  void MySQLFileCatalog::disconnect() const{
    try{
      m_pimpl->disconnect();
      if( m_transactionsta.test(1) ){ //if connection flag is set
	m_transactionsta.reset(1);//reset connection flag to 0
	m_transactionsta.reset(0);//reset transaction flag to 0
      }else{
	throw FCTransactionException("MySQLFileCatalog::disconnect","catalog is already disconnected");
      }
    }
    catch(mysqlpp::BadQuery& er){
      throw FCconnectionException("MySQLFileCatalog::disconnect",er.error);
    }
  }//disconnect
  
  void MySQLFileCatalog::start() const {
    if( m_transactionsta.test(1) ){ //if connected
      if( !m_transactionsta.test(0) ) { //if not already intransaction
	m_transactionsta.set(0); //set transaction flag
      }else{
	throw FCTransactionException("MySQLFileCatalog::start"," catalog is already in transaction, cannot start another transaction");
      }
    }else{ //if not connected
      throw FCTransactionException("MySQLFileCatalog::start"," catalog is not connected, cannot start transaction");
    }
  }
  
  void MySQLFileCatalog::commit(const FileCatalog::CommitMode cm) const{
    //check if in transaction 
    switch (cm){
    case FileCatalog::REFRESH:{break;}
    case FileCatalog::ONHOLD:{break;}
    }
    
    if( !m_transactionsta.test(0) || !m_transactionsta.test(1) ){
      throw FCTransactionException("MySQLFileCatalog::commit"," not in transaction state, cannot commit");
    }
    try{
      MySQLExecute e(m_pimpl->m_con);
      e("COMMIT");
      m_transactionsta.reset(0);//unset Intransaction flag
    }catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::commit",er.error); 
    }
  }
  
  void MySQLFileCatalog::rollback() const{
    //check if in transaction 
    if( !m_transactionsta.test(0)|| !m_transactionsta.test(1) ){
      throw FCTransactionException("MySQLFileCatalog::rollback","not in transaction state, cannot rollback");
    }
    try{
      MySQLExecute e(m_pimpl->m_con);
      e("rollback");
      m_transactionsta.reset(0);//unset Intransaction flag
    }catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::rollback",er.error); 
    }
  }
 
   void MySQLFileCatalog::registerPFN(const std::string& pfn,
                                      const std::string& filetype,
                                      FileCatalog::FileID& fid
				      ) const {
     Guid myuid;
     Guid::create(myuid);
     fid=myuid.toString();
     PFNEntry pentry(pfn,fid,filetype);          
     try{
       insertPFN(pentry);
     }catch (const FCduplicatePFNException& dup){
       pool::PoolMessageStream mysqllog( "MySQLFileCatalog::registerPFN", seal::Msg::Nil );
       mysqllog << seal::Msg::Error <<"PFN "<<pfn
		<<" is already registred"<<seal::endmsg;  
       throw dup;
     }catch(mysqlpp::BadQuery& er){
       throw FCMySQLQueryException("MySQLFileCatalog::registerPFN",er.error);
     } 
   }


  void MySQLFileCatalog::registerLFN(const std::string& pf, const std::string& lf) const {
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    if ( pf.empty() && !lf.empty() ) {
      mysqllog << seal::Msg::Info 
	       <<"MySQLFileCatalog::registerLFN: Logical filenames must be associated to physical filenames" 
	       <<seal::endmsg;
      return;
    }
    if (lf.empty() ) {
      mysqllog<< seal::Msg::Info
	      << "MySQLFileCatalog::registerLFN: You must specify a valid logical file name"
	      <<seal::endmsg;
      return;
    }
    MySQLconvertString c;
    PFNEntry pentry;
    try{
      unsigned int n=m_pimpl->selectfirstPFN(pentry,std::string("pfname=")+c(pf));
      if(n==0){
	mysqllog<< seal::Msg::Error<<"PFN "<<pf
		<<" is non-existent"<<seal::endmsg;
	throw FCnonexistentFileException("MySQLFileCatalog::registerLFN", pf);
      }
    }catch(mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::registerLFN",er.error);
    }
    LFNEntry lentry(lf,pentry.guid());
    try{
      insertLFN(lentry);
    }catch(const FCduplicateLFNException& dup){
      pool::PoolMessageStream mysqllog( "MySQLFileCatalog::registerLFN", seal::Msg::Nil );
      mysqllog << seal::Msg::Error <<"LFN "<<lf
	       <<" is already registred"<<seal::endmsg;  
      throw dup;
    }catch( mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::registerLFN",er.error);
    }
  }

  void MySQLFileCatalog::addReplicaPFN(const std::string& pf, const std::string& rpf) const
  {
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    if( pf.empty() || rpf.empty() ){ //if pf or rpf is not provided, do nothing
      mysqllog<<seal::Msg::Info <<"addReplicaPFN: empty filename specified"<<seal::endmsg;
      return;
    }
    MySQLconvertString c;
    try{
      PFNEntry p;
      unsigned int n=m_pimpl->selectfirstPFN(p,std::string("pfname=")+c(pf));
      if ( n!=0 ) {
        unsigned int m=m_pimpl->selectfirstPFN(p,std::string("pfname=")+c(rpf));
        if( m>0 ){
          mysqllog<<seal::Msg::Info<<"addReplicaPFN: replicated filename is already registered"<<seal::endmsg;
          return;
	}
	PFNEntry pentry(rpf,p.guid(),p.filetype());
	insertPFN(pentry);
      }else{ 
	//PFN not in the catalog
	throw FCnonexistentFileException("MySQLFileCatalog::addReplicaPFN",pf);
      }
    }catch(FCduplicatePFNException& per){
      throw per;
    }catch(mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::addReplicaPFN",er.error);
    } 
  }//addReplicaFilename
  
  void MySQLFileCatalog::addPFNtoGuid(const FileCatalog::FileID& guid, const std::string& pf, const std::string& /* filetype */ ) const{
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    if( pf.empty() ){ 
      mysqllog<<seal::Msg::Info <<"addPFNtoGuid: empty PFN specified"<<seal::endmsg;
      return;
    }
    if ( guid.empty() ) {
      mysqllog<<seal::Msg::Info <<"addPFNtoGuid: add PFN to a null guid" 
	      <<seal::endmsg;
      return;
    }
    MySQLconvertString c;
    PFNEntry presult;
    unsigned int n=m_pimpl->selectfirstPFN(presult,std::string("guid=")+c(guid));
    if (n==0){
      throw FCnonexistentFileException("MySQLFileCatalog::addPFNtoGuid", guid);
    }
    PFNEntry pentry(pf,guid,"");
    try{
      insertPFN(pentry);
    }catch( FCduplicatePFNException& per ){
      throw per;
    }catch( mysqlpp::BadQuery& er ){
      throw FCMySQLQueryException("MySQLFileCatalog::addPFNtoGuid",er.error);
    } 
  }

  void MySQLFileCatalog::renamePFN(const std::string& pf, 
                                   const std::string& newpf) const{
    if( newpf.length()>250 ) throw FCstringLimitException("MySQLFileCatalog::renamePFN",newpf); 
    MySQLconvertString c;
    std::string q("update t_pfn set pfname="+c(newpf)+" where pfname="+c(pf));
    try{
      MySQLExecute e(m_pimpl->m_con);
      e(q);
    }
    catch(mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::renamePFN",er.error);
    } 
  }//renamePFN

  void MySQLFileCatalog::lookupFileByPFN(const std::string& pf, 
                                    FileCatalog:: FileID& fid, std::string& ftype) const{
    MySQLconvertString c;
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    try{
      PFNEntry pentry;
      unsigned int n=m_pimpl->selectfirstPFN(pentry,std::string("pfname=")+c(pf));
      if( n==0 ){
        mysqllog<<seal::Msg::Info <<"PFN "<< pf<< " not found" <<seal::endmsg;
      	fid="";
        ftype="";
      }else{
        fid=pentry.guid();
        if( fid.empty() ){
          mysqllog<<seal::Msg::Info <<"Empty File ID " <<seal::endmsg;
        }
        ftype=pentry.filetype();
      }
    }catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::lookupFileByPFN",er.error);
    }
  }
  
  void MySQLFileCatalog::lookupFileByLFN(const std::string& lf, FileCatalog::FileID& fid) const {
    MySQLconvertString c;
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    try{
      LFNEntry lentry;
      unsigned int n=m_pimpl->selectfirstLFN(lentry,std::string("lfname=")+c(lf));
      if( n==0 ){
        mysqllog<<seal::Msg::Info <<"LFN "<< lf<< " not found" <<seal::endmsg;
        fid="";
      }else{
        fid=lentry.guid();
        if(fid.empty()){
	  mysqllog<<seal::Msg::Info <<"Empty File ID " <<seal::endmsg;
        }
      }
    }catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::lookupFileByLFN",er.error);
    }
  }//lookupFileByLFN

  void MySQLFileCatalog::lookupBestPFN(const FileCatalog::FileID& fid, 
				       const FileCatalog::FileOpenMode& omode, 
				       const FileCatalog::FileAccessPattern& amode,
				       std::string& pf,
                                       std::string& filetype) const{
    pool::PoolMessageStream mysqllog( "MySQLFileCatalog", seal::Msg::Nil );
    switch (omode){
    case FileCatalog::READ:{break;}
    case FileCatalog::WRITE:{break;}
    case FileCatalog::UPDATE:{break;}
    }
    switch (amode){
    case FileCatalog::SEQUENTIAL:{break;}
    case FileCatalog::RANDOM:{break;}
    case FileCatalog::PRANDOM:{break;}
    }
    MySQLconvertString c;
    try{
      PFNEntry pentry;
      unsigned int n=m_pimpl->selectfirstPFN(pentry,std::string("guid=")+c(fid));
      if( n==0 ){ 
	mysqllog<<seal::Msg::Info << "lookupBestPFN(): FileID "<<fid<<"not found in the catalog, return NULL"<<seal::endmsg;
	pf="";
	filetype="";
      }else{
	pf=pentry.pfname();
        filetype=pentry.filetype();
        if(pf.empty()){
          mysqllog<<seal::Msg::Info<<"lookupBestPFN(): empty PFN found"<<seal::endmsg;
	}
      }
    }catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::lookupBestPFN",er.error);
    }
  }
  void MySQLFileCatalog::insertPFN(PFNEntry& pentry) const{
    if( pentry.pfname().length()>250 || pentry.filetype().length()>250 ) throw FCstringLimitException("MySQLFileCatalog::insertPFN","");
    MySQLconvertString c;
    mysqlpp::Query q=m_pimpl->m_con->query();
    q<<"insert into t_pfn(pfname,guid,filetype) values(%0,%1,%2)";
    try{
      q.parse();
      q.execute(c(pentry.pfname()),c(pentry.guid()),c(pentry.filetype()));
    }catch(mysqlpp::BadQuery& er){
      if(er.error.find(std::string("Duplicate entry "))!=er.error.npos){
	throw FCduplicatePFNException("MySQLFileCatalog::insertPFN");
      }else{
	throw FCMySQLQueryException("MySQLFileCatalog::insertPFN",er.error);
      } 
    }
  }
  void MySQLFileCatalog::insertLFN(LFNEntry& lentry) const{
    if( lentry.lfname().length()>250 ) throw FCstringLimitException("MySQLFileCatalog::insertLFN","");
    MySQLconvertString c;
    std::string mylfn(lentry.lfname()); 
    std::string myguid(lentry.guid()); 
    mysqlpp::Query q=m_pimpl->m_con->query();
    q<<"insert into t_lfn(lfname,guid) values(%0,%1)";
    try{
      q.parse();
      q.execute(c(mylfn),c(myguid));
    }catch(mysqlpp::BadQuery& er){
      if(er.error.find(std::string("Duplicate entry "))!=er.error.npos){
	throw FCduplicateLFNException("MySQLFileCatalog::insertLFN");
      }else{
	throw FCMySQLQueryException("MySQLFileCatalog::insertLFN",er.error);
      }
    }
  }
  void MySQLFileCatalog::deletePFN( const std::string& pf ) const{
    MySQLconvertString c;
    try{
      bool metatable_exist=false;
      MySQLtableExist tbexist(m_pimpl->m_con);
      if( tbexist("t_meta") ){
	metatable_exist=true;
      }
      PFNEntry p;
      unsigned int n=m_pimpl->selectfirstPFN(p,std::string("pfname=")+c(pf));
      if( n==0 ){
        //do nothing
      }else{
	MySQLExecute e(m_pimpl->m_con);
	std::string myguid=p.guid();
	PFNEntry pe;
	unsigned int totguid=m_pimpl->selectfirstPFN(pe,std::string("guid=")+c(myguid));
	if( totguid == 1 ){ //if the only PFN with myguid, delete all LFNs  
	  e(std::string("delete from t_lfn where guid=")+c(myguid));
	  if( metatable_exist ){
	    e(std::string("delete from t_meta where guid=")+c(myguid));
	  }
	}
	e(std::string("delete from t_pfn where pfname=")+c(pf)); //delete PFN
      }
    }catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::deletePhysicalFilename",er.error);
    }
  }
  void MySQLFileCatalog::deleteLFN( const std::string& lf ) const{
    std::string myquery("delete from t_lfn where lfname=");
    MySQLconvertString c;
    MySQLExecute e(m_pimpl->m_con);
    myquery+=c(lf);
    try{
      e(myquery); 
    } catch (mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::deleteLFN",er.error);
    }
  }
  void MySQLFileCatalog::deleteEntry(const FileCatalog::FileID& guid) const{
    if( guid.empty() ) return;
    std::string qpf("delete from t_pfn where guid=");
    std::string qlf("delete from t_lfn where guid=");
    MySQLconvertString c;
    MySQLExecute e(m_pimpl->m_con);
    std::string myguid(c(guid));
    qpf+=myguid;
    qlf+=myguid;
    e(qpf);
    e(qlf);
  }
  bool MySQLFileCatalog::isReadOnly() const{
    return false;
  }
  /**empty, pfname=, pfname like, guid=, lfname=, lfname like
   **/
  bool MySQLFileCatalog::retrievePFN(const std::string& query, FCBuf<PFNEntry>& buf, const size_t& start ){
    buf.clear();
    std::string maxs=FCSystemTools::itostr( buf.maxsize() );
    std::string starts=FCSystemTools::itostr( start );
    std::string myquery;
    try{
      if( query.empty() ){
	myquery="select * from t_pfn"; //empty
      }else{
	QueryParser parser(query, "t_pfn");
	parser.parse();
	const std::string& tabs=parser.getTablesAsString();
	const std::string& whereClause=parser.parseResult();
	myquery="select t_pfn.* from "+tabs+" where "+whereClause;
	const std::vector<std::string>& s=parser.getTables();
	if( s.size()>1 ){
	  for(std::vector<std::string>::const_iterator it=s.begin();
	      it!=s.end(); ++it){
	    if( (*it)!="t_pfn" ){
	      myquery+=" AND t_pfn.guid="+(*it)+".guid";
	    }
	  }
	}
      }
      mysqlpp::Query q=m_pimpl->m_con->query();
      q<<myquery;
      q<<" limit "<<start<<" ," <<maxs;
      mysqlpp::Result res=q.store();
      if(res.size()==0) return false;
      mysqlpp::Result::iterator i;
      for(i=res.begin(); i!=res.end() ;++i){
	mysqlpp::Row row=*i;
	std::string mypfn(row.lookup_by_name("pfname"));
	std::string myguid(row.lookup_by_name("guid"));
	std::string filetype(row.lookup_by_name("filetype"));
	PFNEntry pentry(mypfn,myguid,filetype);
	buf.push_back( pentry );
      }
      res.purge();
    }catch(mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::retrievePFN",er.error);
    }
    return true;
  }
  bool MySQLFileCatalog::retrieveLFN(const std::string& query, FCBuf<LFNEntry>& buf, const size_t& start ){
    buf.clear();
    std::string maxs=FCSystemTools::itostr( buf.maxsize() );
    std::string starts=FCSystemTools::itostr( start );
    std::string myquery;
    try{
      if( query.empty() ){
	myquery="select * from t_lfn ";
      }else{
	QueryParser parser(query, "t_lfn");
	parser.parse();
	const std::string& tabs=parser.getTablesAsString();
	const std::string& whereClause=parser.parseResult();
	myquery="select t_lfn.* from "+tabs+" where "+whereClause;
	const std::vector<std::string>& s=parser.getTables();
	if( s.size()>1 ){
	  for(std::vector<std::string>::const_iterator it=s.begin();
	      it!=s.end(); ++it){
	    if( (*it)!="t_lfn" ){
	      myquery+=" AND t_lfn.guid="+(*it)+".guid";
	    }
	  }
	}
      } 
      mysqlpp::Query q=m_pimpl->m_con->query();
      q<<myquery;
      q<<" limit "<<start<<" ," <<maxs;
      mysqlpp::Result res=q.store();
      if(res.size()==0) return false;
      mysqlpp::Result::iterator i;
      for(i=res.begin(); i!=res.end() ;++i){
	mysqlpp::Row row=*i;
	std::string mylfn(row.lookup_by_name("lfname"));
	std::string myguid(row.lookup_by_name("guid"));
	LFNEntry lentry(mylfn,myguid);
	buf.push_back(lentry);
      }
      res.purge();
    }catch(mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::retrieveLFN",er.error);
    }
    return true;
  }
  
  bool MySQLFileCatalog::retrieveGuid(const std::string& query, FCBuf<FileCatalog::FileID>& buf, const size_t& start ){
    buf.clear();
    std::string maxs=FCSystemTools::itostr( buf.maxsize() );
    std::string starts=FCSystemTools::itostr( start );
    std::string myquery;
    try{
      if( query.empty() ){
	myquery="select distinct guid from t_pfn ";
      }else{
	QueryParser parser(query,"t_pfn");
	parser.parse();
	const std::string& tabs=parser.getTablesAsString();
	const std::string& whereClause=parser.parseResult();
	myquery="select distinct t_pfn.guid from "+tabs+" where "+whereClause;
	const std::vector<std::string>& s=parser.getTables();
	if( s.size()>1 ){
	  for(std::vector<std::string>::const_iterator it=s.begin();
	      it!=s.end(); ++it){
	    if( (*it)!="t_pfn" ){
	      myquery+=" AND t_pfn.guid="+(*it)+".guid";
	    }
	  }
	}
      }
      mysqlpp::Query q=m_pimpl->m_con->query();
      q<<myquery;
      q<<" limit "<<start<<" ," <<maxs;
      mysqlpp::Result res=q.store();
      if(res.size()==0) return false;
      mysqlpp::Result::iterator i;
      for(i=res.begin(); i!=res.end() ;++i){
	mysqlpp::Row row=*i;
	std::string myguid(row.lookup_by_name("guid"));
	buf.push_back(myguid);
      }
      res.purge();
    }catch(mysqlpp::BadQuery& er){
      throw FCMySQLQueryException("MySQLFileCatalog::retrieveGuid",er.error);
    }
    return true;
  }
} // ns pool 









