#ifndef POOL_TOKEN_H
#define POOL_TOKEN_H

#include "POOLCore/Guid.h"
#include <vector>
#include <string>
#include <utility>

/*
 *  POOL namespace
 */
namespace pool {

  /** @class Token Token.h POOLCore/Token.h
    *
    * Definition of the Token class. The token identifies 
    * in a unique way objects on the persistent storage.
    *
    * @author: M.Frank
    * @version 1.0
    */
  class Token   {

  public:
    typedef std::pair<int, int> OID_t;

  private:
    /// No copy allowed: put prototype to disable bit-wise copy.
    Token(const Token& copy);
    /// No assignment allowed: put prototype to disable bit-wise assignment.
    Token& operator=(const Token& copy);
  public:
    /// Standard Constructor
    Token();
    /// Constructor with data assignment
    explicit Token(const Token* source);
    /// Standard destructor: release all allocated resources.
    virtual ~Token();
    /// Operator < to allow ordering
    bool operator<(const Token& refToken) const  { return less(refToken);    }
    /// Operator > to allow ordering
    bool operator>(const Token& refToken) const  { 
      if ( less(refToken) ) return false;
      return !equal(refToken);
    }
    /// Equality operator
    bool operator==(const Token& refToken) const { return equal(refToken);   }
    /// Equality operator
    bool operator!=(const Token& refToken) const { return !equal(refToken);  }
    /// Fast token comparison: operator less
    virtual bool less(const Token& pTok) const;
    /// Fast token comparison: operator equals
    virtual bool equal(const Token& pTok) const;
    /// Release token: Decrease reference count and eventually delete.
    int release();
    /// Increase reference count
    int addRef()                              {  return ++m_refCount;         }
    /// Access database identifier
    const std::string& dbID() const           {  return m_dbID;               }
    /// Set database name
    Token& setDb(const std::string& db)       {  m_dbID = db; return *this;   }
    /// Access container identifier
    const std::string& contID() const         {  return m_cntID;              }
    /// Set container name
    Token& setCont(const std::string& cnt)    {  m_cntID = cnt; return *this; }
    /// Access database identifier
    const Guid& classID() const               {  return m_classID;            }
    /// Access database identifier
    Token& setClassID(const Guid& cl_id)      {  m_classID=cl_id;return *this;}
    /// Access technoliogy type
    int  technology() const                   {  return m_technology;         }
    /// Set technology type
    Token& setTechnology(int t)               {  m_technology=t;return *this; }
    /// Access object identifier
    const OID_t& oid() const                  {  return m_oid;                }
    /// Access object identifier
    OID_t& oid()                              {  return m_oid;                }
    /// Access token type
    int type() const                          {  return m_type;               }
    /// Retrieve the string representation of the token.
    virtual const std::string toString()  const;
    /// Retrieve token key
    virtual const std::string key()  const;
    /// Retrieve the string representation of the token.
    Token& fromString(const std::string & from);
    /// Set token information
    const Token& set(Token* pToken)  const;
    /// Set all the data part of the token
    const Token& setData( Token* pToken ) const;
  protected:
    /// Reference count
    int                   m_refCount;
    /// Technology identifier
    int                   m_technology;
    /// Database identifier
    std::string           m_dbID;
    /// Container identifier
    std::string           m_cntID;
    /// Object global identifier
    Guid                  m_classID;
    /// Persistent object identifier
    OID_t                 m_oid;
    /// Token type
    int                   m_type;
  };

  typedef std::vector<Token*> Tokens;

} // namespace pool

#endif /// POOL_TOKEN_H
