// $Header: /cvs/POOL/pool/POOLCore/src/Token.cpp,v 1.17 2007/01/31 18:00:03 frankb Exp $
//	====================================================================
//
//	Token.cpp
//	--------------------------------------------------------------------
//
//	Package   : Persistent Token to identify objects in the persistent
//              world.
//
//	Author    : Markus Frank
//
//	====================================================================
#define POOL_TOKEN_CPP 1

#include "POOLCore/Token.h"
#include <cstdio>
#include <climits>

static const char* fmt_Guid = "[CLID=";
static const char* fmt_tech = "[TECH=%08X]";
static const char* fmt_oid  = "[OID=%08X-%08X]";
static const int   KEY_MASK = (~0x0)<<CHAR_BIT;
static int s_numCount = 0;

namespace pool  {
  int numTokenInstances()  {     return s_numCount;   }
}

/// Standard Constructor(1): Empty constructor
pool::Token::Token() 
: m_refCount(1), m_technology(0), m_type(0) 
{
  m_classID = Guid::null();
  m_oid.first = m_oid.second = ~0x0;
  s_numCount++;
}

/// Copy constructor
pool::Token::Token(const Token& copy) 
{
  copy.set(this);
  m_oid.second = copy.m_oid.second;
  s_numCount++;
}

/// Copy constructor
pool::Token::Token(const Token* source) 
: m_refCount(1), m_technology(source->m_technology), 
  m_dbID(source->m_dbID), m_cntID(source->m_cntID),
  m_classID(source->m_classID), 
  m_oid(source->m_oid), 
  m_type(0) 
{
  s_numCount++;
}

pool::Token::~Token() 
{
  s_numCount--;
}

/// Release token: Decrease reference count and eventually delete.
int pool::Token::release()   {
  int cnt = --m_refCount;
  if ( 0 >= cnt )   {
    delete this;
  }
  return cnt;
}

/// Assignment operator
pool::Token& pool::Token::operator=(const Token& copy)  {
  if ( &copy != this )  {
    copy.setData(this);
  }
  return *this;
}

/// Equality operator
bool pool::Token::equal(const Token& copy)  const  {
  if ( &copy != this )  {
    if ( m_oid.second == copy.m_oid.second )    {
      if ( m_classID == copy.m_classID )  {
        if ( m_dbID == copy.m_dbID )    {
          if ( m_cntID == copy.m_cntID )    {
            return true;
          }
        }
      }
    }
    return false;
  }
  return true;
}

/// Operator to allow ordering
bool pool::Token::less(const Token& copy)  const  {
  if ( &copy != this )  {
    if ( m_oid.second < copy.m_oid.second )
      return true;
    else if ( m_oid.second > copy.m_oid.second )
      return false;

    int res = ::memcmp(&m_classID.Data1, &copy.m_classID.Data1, 16);
    if ( res != 0 )  {
      return (res < 0);
    }
    res = m_dbID.compare(copy.m_dbID);
    if ( res != 0 )  {
      return (res < 0);
    }
    res = m_cntID.compare(copy.m_cntID);
    if ( res != 0 )  {
      return (res < 0);
    }
  }
  return false;
}

const std::string pool::Token::toString()   const   {
  char text[128];
  std::string str = "[DB=" + m_dbID + "][CNT=" + m_cntID + "][CLID=";
  str += m_classID.toString() + "]";
  sprintf(text, fmt_tech, m_technology);
  str += text;
  sprintf(text, fmt_oid, m_oid.first, m_oid.second);
  str += text;
  return str;
}

pool::Token& pool::Token::fromString(const std::string& source)    {
  std::string tmp = source;
  for(char* p1 = (char*)tmp.c_str(); p1; p1 = ::strchr(++p1,'[')) {
    char* p2 = ::strchr(p1, '=');
    char* p3 = ::strchr(p1, ']');
    if ( p2 && p3 )   {
      char* val = p2+1;
      if ( ::strncmp("[DB=", p1, 4) == 0 )  {
        *p3 = 0;
        m_dbID = val;
      }
      else if ( ::strncmp("[CNT=", p1, 5) == 0 )  {
        *p3 = 0;
        m_cntID = val;
      }
      else if ( ::strncmp(fmt_oid, p1, 5) == 0 )  {
        ::sscanf(p1, fmt_oid, &m_oid.first, &m_oid.second);
      }
      else if ( ::strncmp(fmt_Guid, p1, 6) == 0 )  {
        m_classID.fromString( p1+6 );
      }
      else if ( ::strncmp(fmt_tech, p1, 6) == 0 )  {
        ::sscanf( p1, fmt_tech, &m_technology);
      }
      else    {
        *p3 = *p2 = 0;
      }
      *p3 = ']';
      *p2 = '=';
    }
  }
  return *this;
}

/// Retrieve the string representation of the token.
const std::string pool::Token::key()  const  {
  char text[1024];
  ::sprintf(text, "][TECH=%08X]", m_technology&KEY_MASK);
  std::string k = "[DB="+m_dbID+"][CNT="+m_cntID+"][CLID="+m_classID.toString()+text;
  return k;
}

const pool::Token& pool::Token::set(Token* pToken)  const  {
  pToken->m_dbID        = m_dbID;
  pToken->m_cntID       = m_cntID;
  pToken->m_classID     = m_classID;
  pToken->m_oid.first   = m_oid.first;
  pToken->m_technology  = m_technology;
  return *this;
}

const pool::Token& pool::Token::setData(Token* pToken)  const  {
  pToken->m_dbID        = m_dbID;
  pToken->m_cntID       = m_cntID;
  pToken->m_classID     = m_classID;
  pToken->m_oid.first   = m_oid.first;
  pToken->m_oid.second  = m_oid.second;
  pToken->m_technology  = m_technology;
  return *this;
}
