#include "DataTransform.h"
#include "StorageSvc/DbObjectCallBack.h"
#include "PersistencySvc/PersistencySvcException.h"
#include "POOLCore/PoolMessageStream.h"
#include <sstream>

pool::PersistencySvc::DataTransform::DataTransform():
  m_automaticCreationFlag( true ),
  m_callBackObjects(),
  m_callBackObjectsTransientType()
{}


pool::PersistencySvc::DataTransform::~DataTransform()
{
  for ( std::map< pool::Guid, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iClass = m_callBackObjects.begin();
        iClass != m_callBackObjects.end(); ++iClass ) {
    std::map< ROOT::Reflex::Type, pool::DataCallBack* >& callBackObjects = iClass->second;
    for ( std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBack = callBackObjects.begin();
          iCallBack != callBackObjects.end(); ++iCallBack ) {
      iCallBack->second->release();
    }
  }

  for ( std::map< ROOT::Reflex::Type, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iClass = m_callBackObjectsTransientType.begin();
        iClass != m_callBackObjectsTransientType.end(); ++iClass ) {
    std::map< ROOT::Reflex::Type, pool::DataCallBack* >& callBackObjects = iClass->second;
    for ( std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBack = callBackObjects.begin();
          iCallBack != callBackObjects.end(); ++iCallBack ) {
      iCallBack->second->release();
    }
  }
}


void
pool::PersistencySvc::DataTransform::setAutomaticCallBackCreation( bool flag )
{
  m_automaticCreationFlag = flag;
}


pool::DataCallBack&
pool::PersistencySvc::DataTransform::transformation( const pool::Guid& object_type,
                                                     const ROOT::Reflex::Type& shape_type )
{
  std::map< pool::Guid, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iCallBackObjectsForClass = m_callBackObjects.find( object_type );
  if ( iCallBackObjectsForClass == m_callBackObjects.end() ) {
    if ( ! m_automaticCreationFlag ) {
      std::ostringstream error;
      error << "Class \"" << object_type.toString() << "\" has no registered transformation";
      throw pool::PersistencySvcException( error.str(), "pool::DataTransform::transformation" );
    }
    pool::DbObjectCallBack cb( shape_type );
    registerTransformation( object_type, shape_type, cb );
    return transformation( object_type, shape_type );
  }
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >& callBackObjectsForClass = iCallBackObjectsForClass->second;
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBackObject = callBackObjectsForClass.find( shape_type );
  if ( iCallBackObject == callBackObjectsForClass.end() ) {
    if ( ! m_automaticCreationFlag ) {
      std::ostringstream error;
      error << "The persistent class \"" << shape_type.Name()
            << "\" has no registered transformation for the transient class \"" << object_type.toString() << "\"";
      throw pool::PersistencySvcException( error.str(), "pool::DataTransform::transformation" );
    }
    pool::DbObjectCallBack cb( shape_type );
    registerTransformation( object_type, shape_type, cb );
    return transformation( object_type, shape_type );    
  }
  pool::DataCallBack& cb = *( iCallBackObject->second );
  return cb;
}


pool::DataCallBack&
pool::PersistencySvc::DataTransform::transformation( const ROOT::Reflex::Type& object_type,
                                                     const ROOT::Reflex::Type& shape_type )
{
  std::map< ROOT::Reflex::Type, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iCallBackObjectsForClass = m_callBackObjectsTransientType.find( object_type );
  if ( iCallBackObjectsForClass == m_callBackObjectsTransientType.end() ) {
    if ( ! m_automaticCreationFlag ) {
      std::ostringstream error;
      error << "Class \"" << object_type.Name() << "\" has no registered transformation";
      throw pool::PersistencySvcException( error.str(), "pool::DataTransform::transformation" );
    }
    if ( object_type != shape_type ) {
      std::ostringstream error;
      error << "Cannot register a default transformation of a the persistent class \""
            << shape_type.Name() << "\" for the transient class \"" << object_type.Name() << "\"";
      throw pool::PersistencySvcException( error.str(), "pool::DataTransform::transformation" );
    }
    pool::DbObjectCallBack cb( object_type );
    registerTransformation( object_type, shape_type, cb );
    return transformation( object_type, shape_type );
  }
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >& callBackObjectsForClass = iCallBackObjectsForClass->second;
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBackObject = callBackObjectsForClass.find( shape_type );
  if ( iCallBackObject == callBackObjectsForClass.end() ) {
    if ( ! m_automaticCreationFlag ) {
      std::ostringstream error;
      error << "The persistent class \"" << shape_type.Name()
            << "\" has no registered transformation for the transient class \"" << object_type.Name() << "\"";
      throw pool::PersistencySvcException( error.str(), "pool::DataTransform::transformation" );
    }
    if ( object_type != shape_type ) {
      std::ostringstream error;
      error << "Cannot register a default transformation of a persistent shape \""
            << shape_type.Name() << "\" for class \"" << object_type.Name() << "\"";
      throw pool::PersistencySvcException( error.str(), "pool::DataTransform::transformation" );
    }
    pool::DbObjectCallBack cb( object_type );
    registerTransformation( object_type, shape_type, cb );
    return transformation( object_type, shape_type );
  }
  pool::DataCallBack& cb = *( iCallBackObject->second );
  return cb;
}


bool
pool::PersistencySvc::DataTransform::deregisterTransformation( const pool::Guid& object_type, 
                                                               const ROOT::Reflex::Type& shape_type )
{
  std::map< pool::Guid, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iCallBackObjectsForClass = m_callBackObjects.find( object_type );
  if ( iCallBackObjectsForClass == m_callBackObjects.end() ) return false;
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBackObjectForClass = iCallBackObjectsForClass->second.find( shape_type );
  if ( iCallBackObjectForClass == iCallBackObjectsForClass->second.end() ) return false;
  iCallBackObjectForClass->second->release();
  iCallBackObjectsForClass->second.erase( iCallBackObjectForClass );
  if ( iCallBackObjectsForClass->second.empty() ) {
    m_callBackObjects.erase( iCallBackObjectsForClass );
  }
  return true;
}


bool
pool::PersistencySvc::DataTransform::deregisterTransformation( const ROOT::Reflex::Type& object_type, 
                                                               const ROOT::Reflex::Type& shape_type )
{
  std::map< ROOT::Reflex::Type, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iCallBackObjectsForClass = m_callBackObjectsTransientType.find( object_type );
  if ( iCallBackObjectsForClass == m_callBackObjectsTransientType.end() ) return false;
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBackObjectForClass = iCallBackObjectsForClass->second.find( shape_type );
  if ( iCallBackObjectForClass == iCallBackObjectsForClass->second.end() ) return false;
  iCallBackObjectForClass->second->release();
  iCallBackObjectsForClass->second.erase( iCallBackObjectForClass );
  if ( iCallBackObjectsForClass->second.empty() ) {
    m_callBackObjectsTransientType.erase( iCallBackObjectsForClass );
  }
  return true;
}


bool
pool::PersistencySvc::DataTransform::registerTransformation( const pool::Guid& object_type,
                                                             const ROOT::Reflex::Type& shape_type,
                                                             const pool::DataCallBack& callback )
{
  std::map< pool::Guid, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iCallBackObjectsForClass = m_callBackObjects.find( object_type );
  if ( iCallBackObjectsForClass == m_callBackObjects.end() ) {
    iCallBackObjectsForClass = m_callBackObjects.insert( std::make_pair( object_type, std::map< ROOT::Reflex::Type, pool::DataCallBack* >() ) ).first;
  }
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >& callBackObjectsForClass = iCallBackObjectsForClass->second;
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBackObject = callBackObjectsForClass.find( shape_type );
  if ( iCallBackObject != callBackObjectsForClass.end() ) {
    return false;
  }

  pool::PoolMessageStream log( "PersistencySvc::DataTransform", seal::Msg::Nil );
  log << seal::Msg::Debug << "Registered a DataCallBack object for the transient class \"" << object_type.toString()
      << "\" to the persistent class \"" << shape_type.Name() << "\""
      << seal::endmsg;

  return callBackObjectsForClass.insert( std::make_pair( shape_type, callback.clone() ) ).second;
}


bool
pool::PersistencySvc::DataTransform::registerTransformation( const ROOT::Reflex::Type& object_type,
                                                             const ROOT::Reflex::Type& shape_type,
                                                             const pool::DataCallBack& callback )
{
  std::map< ROOT::Reflex::Type, std::map< ROOT::Reflex::Type, pool::DataCallBack* > >::iterator iCallBackObjectsForClass = m_callBackObjectsTransientType.find( object_type );
  if ( iCallBackObjectsForClass == m_callBackObjectsTransientType.end() ) {
    iCallBackObjectsForClass = m_callBackObjectsTransientType.insert( std::make_pair( object_type, std::map< ROOT::Reflex::Type, pool::DataCallBack* >() ) ).first;
  }
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >& callBackObjectsForClass = iCallBackObjectsForClass->second;
  std::map< ROOT::Reflex::Type, pool::DataCallBack* >::iterator iCallBackObject = callBackObjectsForClass.find( shape_type );
  if ( iCallBackObject != callBackObjectsForClass.end() ) {
    return false;
  }

  pool::PoolMessageStream log( "PersistencySvc::DataTransform", seal::Msg::Nil );
  log << seal::Msg::Debug << "Registered a DataCallBack object for the transient class \"" << object_type.Name()
      << "\" to the persistent class \"" << shape_type.Name() << "\""
      << seal::endmsg;

  return callBackObjectsForClass.insert( std::make_pair( shape_type, callback.clone() ) ).second;
}
