#include "GlobalTransaction.h"
#include "DatabaseHandler.h"
#include "PersistencySvc/ITransactionObserver.h"
#include "POOLCore/PoolMessageStream.h"

pool::PersistencySvc::GlobalTransaction::GlobalTransaction( pool::PersistencySvc::DatabaseRegistry& registry ):
  m_type( pool::ITransaction::UNDEFINED ),
  m_databases( registry ),
  m_observers()
{}


pool::PersistencySvc::GlobalTransaction::~GlobalTransaction()
{
  if ( this->isActive() ) {
    this->rollback();
  }
}


bool
pool::PersistencySvc::GlobalTransaction::registerObserver( pool::ITransactionObserver& observer )
{
  pool::ITransactionObserver* observerAddress = &observer;
  for ( std::list< pool::ITransactionObserver* >::const_iterator iObserver = m_observers.begin();
        iObserver != m_observers.end(); ++iObserver ) {
    if ( *iObserver == observerAddress ) return false;
  }
  m_observers.push_back( observerAddress );
  return true;
}


bool
pool::PersistencySvc::GlobalTransaction::deregisterObserver( pool::ITransactionObserver& observer )
{
  pool::ITransactionObserver* observerAddress = &observer;
  for ( std::list< pool::ITransactionObserver* >::iterator iObserver = m_observers.begin();
        iObserver != m_observers.end(); ++iObserver ) {
    if ( *iObserver == observerAddress ) {
      m_observers.erase( iObserver );
      return true;
    }
  }
  return false;
}


bool
pool::PersistencySvc::GlobalTransaction::start( pool::ITransaction::Type type )
{
  if ( this->isActive() || type == pool::ITransaction::UNDEFINED ) return false;
  unsigned int successfullDatabases = 0;
  bool result = false;
  for ( pool::PersistencySvc::DatabaseRegistry::iterator iDb = m_databases.begin();
        iDb != m_databases.end(); ++iDb ) {
    if ( ! (*iDb)->startTransaction() ) {
      pool::PoolMessageStream log( "PersistencySvc::GlobalTransaction" );
      log << seal::Msg::Error << "Could not start a transaction for the database with" << seal::endmsg
          << "FID = " << (*iDb)->fid() << seal::endmsg
          << "PFN = " << (*iDb)->pfn() << seal::endmsg;
      break;
    }
    ++successfullDatabases;
  }

  if ( successfullDatabases == m_databases.size() ) {
    result = true;
  }
  else { // at least one database failed. Rollback the ones we started
    pool::PersistencySvc::DatabaseRegistry::iterator iDb = m_databases.begin();
    while ( successfullDatabases > 0 ) {
      (*iDb)->rollBackTransaction();
      --successfullDatabases;
      ++iDb;
    }
  }

  if ( result ) {
    for ( std::list< pool::ITransactionObserver* >::iterator iTransactionObserver = m_observers.begin();
          iTransactionObserver != m_observers.end(); ++iTransactionObserver ) {
      pool::ITransactionObserver* pObserver = *iTransactionObserver;
      bool tempResult;
      if ( type == pool::ITransaction::READ ) {
        tempResult = pObserver->respondToStartRead();
      }
      else {
        tempResult = pObserver->respondToStartUpdate();
      }
      result = result && tempResult;
      if ( ! result ) {
        pool::PoolMessageStream log( "PersistencySvc::GlobalTransaction" );
        log << seal::Msg::Error << "Could not start a transaction for a transaction observer" << seal::endmsg;
        break;
      }
    }
  }

  if ( result ) m_type = type;
  return result;
}


bool
pool::PersistencySvc::GlobalTransaction::commit()
{
  if ( this->isActive() ) {
    bool OK = true;
    for ( pool::PersistencySvc::DatabaseRegistry::iterator iDb = m_databases.begin();
          iDb != m_databases.end(); ++iDb ) {
      bool bCommit = (*iDb)->commitTransaction(); // This has to be replaced with a two phase commit
      if ( ! bCommit ) {
        pool::PoolMessageStream log( "PersistencySvc::GlobalTransaction" );
        log << seal::Msg::Error << "Could not commit the transaction for the database with" << seal::endmsg
            << "FID = " << (*iDb)->fid() << seal::endmsg
            << "PFN = " << (*iDb)->pfn() << seal::endmsg;
      }
      OK = OK && bCommit;
    }
    
    // Loop over the the transaction observers.
    for ( std::list< pool::ITransactionObserver* >::iterator iTransactionObserver = m_observers.begin();
          iTransactionObserver != m_observers.end(); ++iTransactionObserver ) {
      bool tempResult;
      pool::ITransactionObserver* pObserver = *iTransactionObserver;
      tempResult = pObserver->respondToCommit();
      OK = OK && tempResult;
    }
    m_type = pool::ITransaction::UNDEFINED;
    return OK;
  }

  else return false;
}


bool
pool::PersistencySvc::GlobalTransaction::commitAndHold()
{
  if ( this->isActive() ) {
    bool OK = true;
    for ( pool::PersistencySvc::DatabaseRegistry::iterator iDb = m_databases.begin();
          iDb != m_databases.end(); ++iDb ) {
      bool bCommit = (*iDb)->commitAndHoldTransaction(); // This has to be replaced with a two phase commit
      if ( ! bCommit ) {
        pool::PoolMessageStream log( "PersistencySvc::GlobalTransaction" );
        log << seal::Msg::Error << "Could not commit and hold the transaction for the database with" << seal::endmsg
            << "FID = " << (*iDb)->fid() << seal::endmsg
            << "PFN = " << (*iDb)->pfn() << seal::endmsg;
      }
      OK = OK && bCommit;
    }
    
    // Loop over the the transaction observers.
    for ( std::list< pool::ITransactionObserver* >::iterator iTransactionObserver = m_observers.begin();
          iTransactionObserver != m_observers.end(); ++iTransactionObserver ) {
      bool tempResult;
      pool::ITransactionObserver* pObserver = *iTransactionObserver;
      tempResult = pObserver->respondToCommit();
      OK = OK && tempResult;
    }
    return OK;
  }

  else return false;
}


void
pool::PersistencySvc::GlobalTransaction::rollback()
{
  if ( this->isActive() ) {
    for ( pool::PersistencySvc::DatabaseRegistry::iterator iDb = m_databases.begin();
          iDb != m_databases.end(); ++iDb ) {
      (*iDb)->rollBackTransaction();
    }
    for ( std::list< pool::ITransactionObserver* >::iterator iTransactionObserver = m_observers.begin();
          iTransactionObserver != m_observers.end(); ++iTransactionObserver ) {
      pool::ITransactionObserver* pObserver = *iTransactionObserver;
      pObserver->respondToRollBack();
    }
    m_type = pool::ITransaction::UNDEFINED;
  }
}


bool
pool::PersistencySvc::GlobalTransaction::isActive() const
{
  if ( m_type == pool::ITransaction::UNDEFINED ) return false;
  else return true;
}


void
pool::PersistencySvc::GlobalTransaction::update()
{
  if ( m_type == pool::ITransaction::READ ) {
    for ( std::list< pool::ITransactionObserver* >::iterator iTransactionObserver = m_observers.begin();
          iTransactionObserver != m_observers.end(); ++iTransactionObserver ) {
      pool::ITransactionObserver* pObserver = *iTransactionObserver;
      pObserver->respondToUpdate();
    }
    m_type = pool::ITransaction::UPDATE;
  }
}


pool::ITransaction::Type
pool::PersistencySvc::GlobalTransaction::type() const
{
  return m_type;
}
