#ifndef PERSISTENCYSVC_PERSISTENCYSVC
#define PERSISTENCYSVC_PERSISTENCYSVC

// includes
#include "PersistencySvc/IPersistencySvc.h"

namespace pool {

  namespace PersistencySvc {

    // forward declarations
    class UserSession;

    /** @class PersistencySvc
     * 
     * A concrete implementation of the IPersistencySvc interface
     * 
     */
    class PersistencySvc : virtual public IPersistencySvc {
    public:
      /// Constructor
      explicit PersistencySvc( IFileCatalog& fileCatalog );
      
      /// Destructor
      ~PersistencySvc();

      /// Returns the file catalog in use
      IFileCatalog& fileCatalog();

      /// Set the attached file catalog
      void setFileCatalog( IFileCatalog& catalog );

      /// Returns the token validator in use
      ITokenValidator& tokenValidator();

      /// Set the attached object token validator
      void setTokenValidator( ITokenValidator& validator );

      /// Returns a reference to the IDataTransform object.
      IDataTransform& dataTransform();

      /// Sets the attached object data transform
      void setDataTransform( IDataTransform& dataTransform );

      /** Retrieves an object from persistent store and return with type information
       *  The handle to the reflection class is necessary to later delete the object.
       *
       * @param  transientType [IN]  reference to the type the corresponding transient class
       * @param  token         [IN]  reference to the token for the object
       * @param  handler       [IN]  optional data handler e.g. to interprete references
       *
       * @return std::pair< void*, const ROOT::Reflex::Type > The data and the class description.
       *
       * In case of failure zeros are returned.
       *
       */
      std::pair< void*, ROOT::Reflex::Type > readObject( const pool::Guid& transientType,
                                                         const Token& token,
                                                         DataHandler* handler = 0 );

      std::pair< void*, ROOT::Reflex::Type > readObject( const ROOT::Reflex::Type& transientType,
                                                         const Token& token,
                                                         DataHandler* handler = 0 );

      /** Retrieves an object from persistent store and return with type information
       *  The handle to the reflection class is necessary to later delete the object.
       *  The Guid of the transient class is assumed to be the classID of the token
       *
       * @param  token   [IN]  reference to the token for the object
       * @param  handler [IN]  optional data handler e.g. to interprete references
       *
       * @return std::pair< void*, ROOT::Reflex::Type > The data and the class description.
       *
       * In case of failure zeros are returned.
       *
       */
      std::pair< void*, ROOT::Reflex::Type > readObject( const Token& token,
                                                         DataHandler* handler = 0 );
      
      /**  registerForWrite registers an object for writing to the persistent medium
       *   The optional handler is only necessary if during the load phase 
       *   higher level interactions with the framework are necessary.
       *
       * @param  place        [IN]  the placement hint
       * @param  object       [IN]  pointer to transient object which will be written
       * @param  type         [IN]  reflection class description with the layout of transient object
       * @param  handler      [IN]  optional data handler e.g. to interprete references
       *
       * @return Token*   the token address of the persistent object. I case of failure 0 is returned.
       *
       */
      Token* registerForWrite( const Placement& place,
                               const void* object,
                               const ROOT::Reflex::Type& type,
                               DataHandler* handler = 0 );
      
      /**  updateObject updates an object which is already written in the persistent medium.
       *   higher level interactions with the framework are necessary.
       *
       * @param  object       [IN]  pointer to transient object which will be updates
       * @param  type         [IN]  reflection class description with the layout of transient object
       * @param  token        [IN]  reference to the token for the object
       * @param  handler      [IN]  optional data handler e.g. to interprete references
       *
       * @return bool   Indicates whether operation was successfull.
       *
       */
      bool updateObject( const void* object,
                         const ROOT::Reflex::Type& type,
                         const Token& token,
                         DataHandler* handler = 0 );
      
      /** Deletes an object from the persistent store
       *
       * @param  token        [IN]  reference to the token for the object
       *
       * @return bool   Indicates whether operation was successfull.
       *
       */
      bool destroyObject( const Token& token );
      
      /// Returns the underlying global session
      ISession& session();

      /// Registers a transaction observer
      bool registerObserver( ITransactionObserver& observer );

      /// Deregisters a transaction observer
      bool deregisterObserver( ITransactionObserver& observer );

    private:
      /// The session object.
      UserSession* m_session;

    };
  }
}

#endif // PERSISTENCYSVC_PERSISTENCYSVC
