#ifndef INCLUDE_PERSISTENCYSVC_PERSISTENCYSVCCONFIGURATION_H
#define INCLUDE_PERSISTENCYSVC_PERSISTENCYSVCCONFIGURATION_H

#include "PersistencySvc/DatabaseConnectionPolicy.h"

namespace pool {

  // forward declarations
  class IFileCatalog;
  class ITokenValidator;
  class IDataTransform;

  namespace PersistencySvc {
    /** @class PersistencySvcConfiguration
     *
     *  PersistencySvcConfiguration holds the file catalog reference,
     *  the token validator, the data transform and database connection policy objects,
     *  that are used in the various operations of the PersistencySvc classes.
     *
     */
    class PersistencySvcConfiguration {
    public:
      /// Constructor. Creates a dummy token validator and a default data transform
      PersistencySvcConfiguration();
      
      /// Destructor
      ~PersistencySvcConfiguration();
      
      /** Sets the attached file catalog
       *
       *  @param catalog   [IN] Reference to the file catalog
       */
      void setFileCatalog( IFileCatalog& catalog );
      
      /// Retrieves the file catalog in use.
      IFileCatalog& fileCatalog();
      const IFileCatalog& fileCatalog() const;
      
      /** Sets the attached object token validator
       *
       *  @param validator [IN] Reference to the external validator
       */
      void setTokenValidator( ITokenValidator& validator );
      
      /// Retrieves the token valitator in use
      ITokenValidator& tokenValidator();
      
      /** Sets the attached object data transform
       *
       *  @param transform [IN] Reference to the external data transform
       */
      void setDataTransform( IDataTransform& transform );
      
      /// Returns a reference to the IDataTransform object.
      IDataTransform& dataTransform();

      /// Sets the database connection policy
      void setDatabaseConnectionPolicy( const DatabaseConnectionPolicy& policy );
      
      /// Returns the database connection policy.
      const DatabaseConnectionPolicy& databaseConnectionPolicy() const;

    private:
      /// The file catalog
      IFileCatalog* m_fileCatalog;
      /// The token validator
      ITokenValidator* m_tokenValidator;
      /// Flag indicating the ownership of the token validator
      bool m_ownTokenValidator;
      /// The data transform
      IDataTransform* m_dataTransform;
      /// Flag indicating the ownership of the data transform
      bool m_ownDataTransform;
      /// The default database connection policy
      DatabaseConnectionPolicy m_policy;
    };
    
  }
}

#endif
