#include <iostream>
#include <stdexcept>
#include <memory>

#include "POOLCore/POOLContext.h"

#include "StorageSvc/IStorageSvcFactory.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbOption.h"
#include "StorageSvc/FileDescriptor.h"
#include "StorageSvc/poolDb.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/FCLeaf.h"
#include "FileCatalog/FCImpl.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/URIParser.h"


class InsertFileToCatalogApplication {
public:
  InsertFileToCatalogApplication( int argc, char* argv[] );
  ~InsertFileToCatalogApplication(){}
  bool parseArguments();
  void execute();
  void printSyntax();
private:
  const std::string executableName;
  std::string technologyName;
  std::string fcURL;
  std::vector< std::string > fileNames;
};

InsertFileToCatalogApplication::InsertFileToCatalogApplication( int argc, char* argv[] ):
  executableName( std::string( argv[0] ) ),
  technologyName( pool::ROOT_StorageType.storageName() ),
  fcURL( "" ),
  fileNames()
{
  for ( int i = 1; i < argc; ++i )
    fileNames.push_back( std::string( argv[i] ) );
}

bool
InsertFileToCatalogApplication::parseArguments()
{
  std::vector< std::string > theFiles;
  const std::vector< std::string >& args = fileNames;
  unsigned int excludedArgument = 0;
  for ( unsigned int iArg = 0; iArg < args.size(); ++iArg ) {
    if ( iArg > 0 && iArg == excludedArgument )
      continue;
    const std::string& arg = args[iArg];
    if ( arg == "-u" ) {
      unsigned int nextArgumentIndex = iArg + 1;
      if ( nextArgumentIndex < args.size() ) {
        excludedArgument = nextArgumentIndex;
        fcURL = args[nextArgumentIndex];
      }
    }
    else if ( arg == "-t" ) {
      unsigned int nextArgumentIndex = iArg + 1;
      if ( nextArgumentIndex < args.size() ) {
        excludedArgument = nextArgumentIndex;
        technologyName = args[nextArgumentIndex];
      }
    }
    else {
      theFiles.push_back( arg );
    }
  }

  fileNames = theFiles;

  // Check now the values.
  if ( fileNames.empty() ) return false;
  pool::DbType theType = pool::DbType::getType( technologyName );
  if ( theType.match( pool::TEST_StorageType ) ) {
    throw std::runtime_error( "Unsupported file type : " + technologyName );
  }

  return true;
}


void
InsertFileToCatalogApplication::execute()
{
  // Do the job here
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Warning );

  pool::IStorageSvcFactory* fac = pool::IStorageSvcFactory::get();
  if ( ! fac ) {
    throw std::runtime_error( "Could not retrieve an IStorageSvcFactory" );
  }

  pool::IStorageSvc* storageSvc = fac->create("pool_DbStorageSvc", 0, "StorageSvc");
  if ( ! storageSvc ) {
    throw std::runtime_error( "Could not create a StorageSvc object" );
  }
  storageSvc->addRef();
  void* pVoid = 0;
  pool::DbStatus sc = storageSvc->queryInterface( pool::IStorageExplorer::interfaceID(),&pVoid);
  pool::IStorageExplorer* storageExplorer = (pool::IStorageExplorer*)pVoid;
  if ( !( sc.isSuccess() && storageExplorer ) ) {
    storageSvc->release();
    throw std::runtime_error( "Could not retrieve a IStorageExplorer interface" );
  }

  bool allOK = true;

  // Retrieve the fids.
  std::vector< std::pair< std::string, std::string > > fidAndPfn;
  fidAndPfn.reserve( fileNames.size() );
  try {
    pool::Session* session = 0;
    long technologyId = pool::DbType::getType( technologyName ).majorType();
    if ( ! storageExplorer->startSession( poolDb::READ, technologyId, session ).isSuccess() ) {
      throw std::runtime_error( "Could not start a new session" );
    }
    if( technologyId == pool::ROOT_StorageType.majorType() ) {
       // Set ERROR_LEVEL ROOT domain option to kError (==2000)
       // to disable warnings about unknown classes when opening the file
       pool::DbOption        err_lvl("ERROR_LEVEL", "", 2000);
       storageExplorer->setDomainOption(session, err_lvl);
    }
    for ( std::vector< std::string >::const_iterator iFile = fileNames.begin();
          iFile != fileNames.end(); ++iFile ) {
      const std::string& pfn = *iFile;
      std::string fid = "";
      pool::FileDescriptor fd( fid, pfn );
      if ( ! storageExplorer->connect( session, poolDb::READ, fd ).isSuccess() ) {
        throw std::runtime_error( "Could not open the file \"" + pfn + "\"" );
      }
      if ( ! storageExplorer->dbParam( fd, "FID", fid ).isSuccess() ) {
        storageExplorer->disconnect( fd );
        throw std::runtime_error( "Could not retrieve the FID from file \"" + pfn + "\"" );
      }
      fidAndPfn.push_back( std::make_pair( fid, pfn ) );
      storageExplorer->disconnect( fd );
    }
    storageExplorer->endSession( session );
  }
  catch ( std::exception& error ) {
    std::cerr << error.what() << std::endl;
    allOK = false;
  }
  catch ( ... ) {
    std::cerr << "Funny error ..." << std::endl;
    allOK = false;
  }

  storageExplorer->release();
  storageSvc->release();

  // Open the file catalog and insert the pfn/fid/technology
  if ( allOK ) {
    pool::URIParser p( fcURL );
    p.parse();
    std::auto_ptr<pool::IFileCatalog> catalog( new pool::IFileCatalog );
    if ( ! catalog.get() ) 
      throw std::runtime_error( "Could not create a file catalog" );
    catalog->setWriteCatalog( p.contactstring() );
    catalog->connect();
    catalog->start();
    
    for ( std::vector< std::pair< std::string, std::string > >::const_iterator iFile = fidAndPfn.begin();
          iFile != fidAndPfn.end(); ++iFile ) {
      pool::PFNEntry entry( iFile->second, iFile->first, technologyName );
      dynamic_cast< pool::FCLeaf* >( catalog->getWriteCatalog() )->getImpl()->insertPFN( entry );
    }
    
    catalog->commit();
  }
}

void
InsertFileToCatalogApplication::printSyntax()
{
  std::cout << "Syntax : " << executableName << " [-u fileCatalog] [-t technologyType] files" << std::endl;
}




int main( int argc, char* argv[] )
{
  try {
    InsertFileToCatalogApplication app( argc, argv );
    if ( app.parseArguments() ) {
      app.execute();
    }
    else app.printSyntax();
  }
  catch ( std::exception& error ) {
    std::cerr << error.what() << std::endl;
    return 1;
  }
  catch ( ... ) {
    std::cerr << "Funny error ..." << std::endl;
    return 1;
  }
  return 0;
}
