#include "Python.h"
#include "structmember.h"
#include <iostream>
#include <string>
#include "POOLCore/Exception.h"
#include "FileCatalog/FCImpl.h"
#include "FileCatalog/FCQueryParser.h"
#include "PluginManager/PluginManager.h"
#include "FileCatalog/FCImplPluginFactory.h"
#include "FileCatalog/FCMetaImplPluginFactory.h"
#include "FileCatalog/FCException.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/FCBuf.h"
#include "FileCatalog/FCSystemTools.h"
#include "FileCatalog/URIParser.h"
#include "AttributeList/AttributeListSpecification.h"
#define PyFileCatalog_REFRESH 0
#define PyFileCatalog_ONHOLD 1
#define PyFileCatalog_NO_DELETE 0
#define PyFileCatalog_DELETE_REDUNDANT 1
#define PyFileCatalog_PFN 0
#define PyFileCatalog_LFN 1
#define PyFileCatalog_META 2
#define PyFileCatalog_GUID 3
#define PyFileCatalog_FileAccessPattern_SEQUENTIAL 0 
#define PyFileCatalog_FileAccessPattern_RANDOM 1
#define PyFileCatalog_FileAccessPattern_PRANDOM 2
#define PyFileCatalog_FileOpenMode_READ 0
#define PyFileCatalog_FileOpenMode_WRITE 1
#define PyFileCatalog_FileOpenMode_UPDATE 2

inline char*  _Text(const char* x) { return (char*)x; }
inline char** _Keys(const char** x) { return (char**)x; }

#if PY_MAJOR_VERSION<=2 && PY_MINOR_VERSION<5
typedef int PyFCsize_t;
#else
typedef Py_ssize_t PyFCsize_t;
#endif

static PyObject *PyFileCatalogError;

typedef struct {
  PyObject_HEAD
  pool::FCImpl* mycatalog;
  pool::FCMetaImpl* mymetacatalog;
} PyFCImpl;

static int 
PyFCImpl_init(PyObject* self, PyObject* args, PyObject* kwds)
{
  static const char* keys[]={"contactstr",NULL};
  std::string url;
  const char* contactstr=NULL;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("|s"), _Keys(keys), &contactstr)) 
    return -1;
  pool::URIParser par;
  pool::FCImpl* me=0;
  pool::FCMetaImpl* metame=0;
  try{
    if( contactstr==NULL ){
      par.parse();
    }else{
      par.setURI(contactstr);
      par.parse();
    }
    std::string uriprefix=par.prefix();
    url=par.url();
    seal::PluginManager* pm = seal::PluginManager::get();
    if ( ! pm ) {
      throw(pool::Exception("Could not get the plugin manager instance", 
                            "PyFCImpl_init", 
                            std::string("PyFileCatalog")));
    }
    pm->initialise();
    if(uriprefix.empty()||uriprefix=="XMLFileCatalog"){
      me=pool::FCImplPluginFactory::get()->create( "xmlcatalog" );
      metame=pool::FCMetaImplPluginFactory::get()->create("xmlcatalog_meta",me);
    }else{
      me=pool::FCImplPluginFactory::get()->create( uriprefix );
      metame=pool::FCMetaImplPluginFactory::get()->create(uriprefix+std::string("_meta"),me);
    }
    if( !me ){
      throw(pool::Exception("Could not load FileCatalogPlugin!", 
                            "PyFCImpl_init", 
                            std::string("PyFileCatalog")));
    }
    me->setURL(url);
    me->setType(uriprefix);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return -1;
  }
  ((PyFCImpl*)self)->mycatalog=me;
  ((PyFCImpl*)self)->mycatalog->setURL(url);
  ((PyFCImpl*)self)->mymetacatalog=metame;
  return 0;
}

static char PyFCImpl_doc[]=
"PyFCImpl() : contact string is picked up from POOL_CATALOG, if not set, default catalog file:PoolFileCatalog.xml is created"
"FileCatalog(contactstr): " ;

static void
PyFCImpl_dealloc(PyObject *self)
{
 delete ((PyFCImpl*)self)->mycatalog;
  if( ((PyFCImpl*)self)->mymetacatalog ) delete ((PyFCImpl*)self)->mymetacatalog;
  self->ob_type->tp_free(self);
}

static PyObject *
connect(PyObject *self)
{
  try{
    ((PyFCImpl*)self)->mycatalog->connect();
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static char connect_doc[]=
"connect() connect to the catalog";

static PyObject *
disconnect(PyObject *self)
{
  try{
    ((PyFCImpl*)self)->mycatalog->disconnect();
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static char disconnect_doc[]=
"disconnect() disconnect from the catalog";

static PyObject *
start(PyObject *self)
{
  try{
    ((PyFCImpl*)self)->mycatalog->start();
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char start_doc[]=
"start() start a transaction";

static PyObject *
commit(PyObject *self, PyObject *args)
{
  int mod=-99; 
  if(!PyArg_ParseTuple(args,_Text("|i"),&mod)) 
    return NULL;
  try{
    if( mod == -99 || mod==PyFileCatalog_REFRESH ) {
      ((PyFCImpl*)self)->mycatalog->commit(pool::FileCatalog::REFRESH);
    }else if( mod==PyFileCatalog_ONHOLD ) {
      ((PyFCImpl*)self)->mycatalog->commit(pool::FileCatalog::ONHOLD);
    }else{
      PyErr_SetString(PyExc_RuntimeError,"mod must be REFRESH,ONHOLD or None");
      return NULL;
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char commit_doc[]=
"commit(mode=REFRESH) commit the transaction with mode REFRESH(default) or ONHOLD"; 

static PyObject *
rollback(PyObject *self)
{
  try{
    ((PyFCImpl*)self)->mycatalog->rollback();
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char rollback_doc[]=
"rollback() rollback the transaction";

static PyObject *
registerPFN(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"pfn","filetype", NULL};
  char *pfn, *filetype;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("ss"), _Keys(keys), &pfn, &filetype)) 
    return NULL;
  std::string cpfn(pfn), cfiletype(filetype), fid;
  try{
    ((PyFCImpl*)self)->mycatalog->registerPFN(cpfn, cfiletype, fid);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  return Py_BuildValue(_Text("s"), fid.c_str());
}
static char registerPFN_doc[]=
"registerPFN(pfn, filetype) register a PFN with filetype. The generated FileID is returned.";

static PyObject *
registerLFN(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"pfn","lfn",NULL};
  char *pfn, *lfn;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("ss"), _Keys(keys), &pfn, &lfn)) 
    return NULL;
  std::string cpfn(pfn), clfn(lfn);
  try{
    ((PyFCImpl*)self)->mycatalog->registerLFN(cpfn, clfn);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char registerLFN_doc[]=
  "registerLFN(pfn,lfn) register a lfn with given pfn";

static PyObject *
addReplicaPFN(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"pfn","rpfn",NULL};
  char *pfn, *rpfn;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("ss"), _Keys(keys), &pfn, &rpfn)) 
    return NULL;
  std::string cpfn(pfn), crpfn(rpfn);
  try{
    ((PyFCImpl*)self)->mycatalog->addReplicaPFN(cpfn, crpfn);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char addReplicaPFN_doc[]=
"addReplicaPFN(pfn, replicapfn) add replica file name to the given pfn";

static PyObject *
addPFNtoGuid(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"guid", "pfn", "filetype", NULL};
  char *pfn, *guid, *filetype;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("sss"), _Keys(keys), &guid, &pfn, &filetype)) 
    return NULL;
  std::string cpfn(pfn), cguid(guid), cfiletype(filetype);
  try{
    ((PyFCImpl*)self)->mycatalog->addPFNtoGuid(cguid, cpfn, cfiletype);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char addPFNtoGuid_doc[]=
"addPFNtoGuid(guid, pfn, filetype) add PFN to the given guid";

static PyObject *
renamePFN(PyObject *self,PyObject *args, PyObject* kwds){
  static const char* keys[]={"pfn","newpfn",NULL};
  char *pfn, *npfn;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("ss"), _Keys(keys), &pfn, &npfn)) 
    return NULL;
  std::string cpfn(pfn), cnpfn(npfn);
  try{
    ((PyFCImpl*)self)->mycatalog->renamePFN(cpfn, cnpfn);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char renamePFN_doc[]=
"renamePFN('oldpfn','newpfn') rename the old pfn to the new one";

static PyObject *
lookupFileByPFN(PyObject *self, PyObject *args){
  char * pfn;
  if(!PyArg_ParseTuple(args,_Text("s"), &pfn) ) 
    return NULL;
  std::string cpfn(pfn),cfid,cftype;
  try{
    ((PyFCImpl*)self)->mycatalog->lookupFileByPFN(cpfn, cfid, cftype);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  return Py_BuildValue(_Text("{ssss}"),"guid",cfid.c_str(),"filetype",cftype.c_str());
}
static const char lookupFileByPFN_doc[]=
"lookupFileByPFN('pfn') lookup FileID by given PFN. The returned value is a dictionary {'guid':'guidval','filetype':'type'}";

static PyObject *
lookupFileByLFN(PyObject *self, PyObject *args){
  char * lfn;
  if(!PyArg_ParseTuple(args,_Text("s"), &lfn) ) 
    return NULL;
  std::string clfn(lfn),cfid;
  try{
    ((PyFCImpl*)self)->mycatalog->lookupFileByLFN(clfn, cfid);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  return Py_BuildValue(_Text("s"), cfid.c_str());
}
static const char lookupFileByLFN_doc[]=
"lookupFileByLFN('lfn') lookup FileID by given LFN. The guid is returned}";

static PyObject *
lookupBestPFN(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"fid",NULL};
  char *fid;
  //*omode, *amode; //dummy in python
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("z"), _Keys(keys), &fid)) 
    return NULL;
  std::string cfid(fid), cpfn, cfiletype;
  try{
    ((PyFCImpl*)self)->mycatalog->lookupBestPFN(cfid, pool::FileCatalog::READ, pool::FileCatalog::SEQUENTIAL, cpfn, cfiletype);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  return Py_BuildValue(_Text("{ssss}"),"pfn",cpfn.c_str(),"filetype",cfiletype.c_str());
}
static const char lookupBestPFN_doc[]=
"lookupBestPFN('pfn') return the first PFN found associated with given fid. The returned value is a dictionary {'pfn':'pfnval','filetype':'type'}";

static PyObject *
createMetaDataSpec(PyObject *self, PyObject *args, PyObject* kwds){
  static const char* keys[]={"metadataspec",NULL};
  PyObject* dict;
  if(!PyArg_ParseTupleAndKeywords(args,kwds,_Text("O"), _Keys(keys), &dict)) 
    return NULL;
  PyObject *name, *type; 
  pool::MetaDataEntry myspec;
  PyFCsize_t pos=0;
  try{
    while( PyDict_Next(dict, &pos, &name, &type) ){
      std::string cname(PyString_AsString(name));
      std::string ctype(PyString_AsString(type));
      myspec.addAttributeSpec( cname, ctype );
      //fixme descref name type?
    }
    if (myspec.size()>1){
      if( ((PyFCImpl*)self)->mymetacatalog){
        ((PyFCImpl*)self)->mymetacatalog->createMetaDataSpec(myspec);
      }
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char createMetaDataSpec_doc[]=
"createMetaDataSpec({'name1':'type1','namen':'typen'}) create metadata definition. guid is include implicitly and should not be defined here";

static PyObject *
updateMetaDataSpec(PyObject *self, PyObject *args){
  //static char* keys[]={"metadataspec, metamode",NULL};
  PyObject* dict;
  int metamode=-99;
  if(!PyArg_ParseTuple(args,_Text("O|i"),&dict,&metamode)) 
    return NULL;
  PyObject *name, *type; 
  pool::MetaDataEntry myspec;
  PyFCsize_t pos=0;
  try{
    while( PyDict_Next(dict, &pos, &name, &type) ){
      std::string cname(PyString_AsString(name));
      std::string ctype(PyString_AsString(type));
      myspec.addAttributeSpec( cname, ctype );
      //fixme descref name type?
    }
    if(metamode==-99 || metamode==PyFileCatalog_NO_DELETE){
      ((PyFCImpl*)self)->mymetacatalog->updateMetaDataSpec(myspec);
    }else if(metamode==PyFileCatalog_DELETE_REDUNDANT){
      ((PyFCImpl*)self)->mymetacatalog->updateMetaDataSpec(myspec,pool::FileCatalog::DELETE_REDUNDANT);
    }else{
      PyErr_SetString(PyExc_RuntimeError,"metamod must be NO_DELETE,DELETE_REDUNDANT or None");
      return NULL;
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char updateMetaDataSpec_doc[]=
"updateMetaDataSpec({'name1':'type1','namen':'typen'}, metamode=NO_DELETE) update the metadata definition. guid is include implicitly and should not be defined here.";

static PyObject *
getMetaDataSpec(PyObject *self){
  pool::MetaDataEntry meta;
  try{
    if( ((PyFCImpl*)self)->mymetacatalog ){
      ((PyFCImpl*)self)->mymetacatalog->getMetaDataSpec(meta);
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  PyObject* mydict=PyDict_New();
  const char *key, *type;
  pool::AttributeListSpecification spec=meta.spec();
  pool::AttributeListSpecification::iterator it;
  try{
    for(it=spec.begin(); it!=spec.end(); ++it){
      key=(*it).name().c_str();
      type=(*it).type_name().c_str();
      if( (*it).name() != "guid"){
        PyDict_SetItem(mydict, Py_BuildValue(_Text("s"),key), Py_BuildValue(_Text("s"), type));  
      }
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  return Py_BuildValue(_Text("O"),mydict);
}
static const char getMetaDataSpec_doc[]=
"getMetaDataSpec() return meta data specification of the catalog as a dictionary";

static PyObject *
dropMetaDataSpec(PyObject *self){
  try{
    if( ((PyFCImpl*)self)->mymetacatalog ){
      ((PyFCImpl*)self)->mymetacatalog->dropMetaDataSpec();
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char dropMetaDataSpec_doc[]=
"dropMetaDataSpec() drop the meta data and its definition of the catalog.";

static PyObject *
insertPFN (PyObject *self,  PyObject *args, PyObject *kwds){
  static const char* argnames[]={"pfname", "guid", "filetype", NULL};
  char *guid, *pfname, *filetype;  
  if(!PyArg_ParseTupleAndKeywords(args,kwds,_Text("sss"), _Keys(argnames), &pfname, &guid, &filetype)) 
    return NULL;
  std::string cpfname(pfname);
  std::string cguid(guid);
  std::string cfiletype(filetype);
  pool::PFNEntry pentry(cpfname,cguid,cfiletype);
  try{
    ((PyFCImpl*)self)->mycatalog->insertPFN(pentry);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static char insertPFN_doc[]=
"insertPFN(pfname, guid, filtype) insert PFN";

static PyObject *
insertLFN (PyObject *self,  PyObject *args, PyObject *kwds){
  static const char* argnames[]={"lfname", "guid",NULL};
  char *guid, *lfname;
  if(!PyArg_ParseTupleAndKeywords(args,kwds,_Text("ss"), _Keys(argnames), &lfname, &guid)) 
    return NULL;
  std::string cguid(guid);
  std::string clfname(lfname);
  try{
    pool::LFNEntry lentry(clfname, cguid);
    ((PyFCImpl*)self)->mycatalog->insertLFN(lentry);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char insertLFN_doc[]=
"insertLFN(lfname, guid) insert LFN";

static PyObject *
insertMetaData(PyObject *self, PyObject *args, PyObject* kwds ){
  char* fid; 
  static const char* keys[]={"fid","metadatadict",NULL};
  PyObject *dict;
  if(!PyArg_ParseTupleAndKeywords(args,kwds,_Text("sO"),_Keys(keys),&fid,&dict)) 
    return NULL;
  pool::MetaDataEntry meta;
  pool::AttributeListSpecification spec;
  try{
    //fixme: should keep only one copy of meta schema
    ((PyFCImpl*)self)->mymetacatalog->getMetaDataSpec(meta);
    spec=meta.spec();
    PyFCsize_t pos=0;
    PyObject *key, *value;
    while (PyDict_Next(dict,&pos,&key,&value)){
      std::string ckey=PyString_AsString(key);
      if( ckey!="guid"){
        std::string ctype=spec[ckey].type_name();
        if(ctype=="string"){
          std::string cvalue=PyString_AsString(value);
          meta.setAttributeValue(ckey,std::string(cvalue));
        }else if(ctype=="int"){
          int cvalue=PyInt_AsLong(value);
          meta.setAttributeValue(ckey,cvalue);
        }
      }else{
        std::string cfid(fid);
        meta.setGuidValue(cfid);
      }
    }
    ((PyFCImpl*)self)->mymetacatalog->insertMetaData(meta);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char insertMetaData_doc[]=
"insertMetaData('fid', {'name1':'val1','namen':'valn'}) insert metadata values of file with given fid.";

static PyObject *
deletePFN(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"pfn",NULL};
  char *pfn;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("s"), _Keys(keys), &pfn)) 
    return NULL;
  std::string cpfn(pfn);
  try{
    ((PyFCImpl*)self)->mycatalog->deletePFN(cpfn);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char deletePFN_doc[]=
"deletePFN(pfn) delete the pfn entry with given pfn";

static PyObject *
deleteLFN(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"lfn",NULL};
  char *lfn;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("s"), _Keys(keys), &lfn)) 
    return NULL;
  std::string clfn(lfn);
  try{
    ((PyFCImpl*)self)->mycatalog->deleteLFN(clfn);
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char deleteLFN_doc[]=
"deleteLFN(lfn) delete the lfn entry with given lfn";

static PyObject *
deleteMetaData(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"guid",NULL};
  char *guid;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("s"), _Keys(keys), &guid)) 
    return NULL;
  std::string cguid(guid);
  try{
    if( ((PyFCImpl*)self)->mymetacatalog ){
      ((PyFCImpl*)self)->mymetacatalog->deleteMetaData(cguid);
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char deleteMetaData_doc[]=
"deleteMetaData(guid) delete the MetaData entry with given guid";

static PyObject *
deleteEntry(PyObject *self,  PyObject *args, PyObject* kwds)
{
  static const char* keys[]={"guid",NULL};
  char *guid;
  if(!PyArg_ParseTupleAndKeywords(args,kwds, _Text("s"), _Keys(keys), &guid)) 
    return NULL;
  std::string cguid(guid);
  try{
    ((PyFCImpl*)self)->mycatalog->deleteEntry(cguid);
    if( ((PyFCImpl*)self)->mymetacatalog ){
      ((PyFCImpl*)self)->mymetacatalog->deleteMetaData(cguid);
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static const char deleteEntry_doc[]=
"deleteEntry(guid) delete all entries associated with given guid";

static PyObject *
isReadOnly(PyObject *self){
  bool result;
  try{
    result=((PyFCImpl*)self)->mycatalog->isReadOnly();
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }
  return Py_BuildValue(_Text("i"),result);
}
static const char isReadOnly_doc[]=
"isReadOnly() return true if catalog is readOnly";

static PyObject *
retrievePFN(PyObject *self, PyObject *args){
  bool result;
  char* query;
  size_t start;
  size_t cachesize;
  PyObject *mylist; 
  if(!PyArg_ParseTuple(args, _Text("sll"), &query, &start, &cachesize))
    return NULL;
  //std::cout<<query<<"#"<<start<<"#"<<cachesize<<std::endl;
  std::string cquery(query);
  pool::FCBuf<pool::PFNEntry> cmyBuf(cachesize);
  try{
    pool::FCQueryParser parser(cquery);
    parser.parse();
    if(parser.isMetaQuery()){
      result=((PyFCImpl*)self)->mymetacatalog->retrievePFN(cquery,cmyBuf,start);      
    }else{
      result=((PyFCImpl*)self)->mycatalog->retrievePFN(cquery,cmyBuf,start);      
    }
    //std::cout<<"result"<<result<<cmyBuf.size()<<std::endl;
    if(result){
      int n=0;
      size_t mysize=cmyBuf.size();
      mylist=PyList_New(int(mysize));
      for(size_t i=0; i<mysize; ++i){
        PyObject* mytuple=PyTuple_New(3);
        if(mytuple==NULL) return NULL;
        pool::PFNEntry pentry=cmyBuf[i];
        PyTuple_SetItem(mytuple,0,PyString_FromString(pentry.pfname().c_str()));
        PyTuple_SetItem(mytuple,1,PyString_FromString(pentry.guid().c_str()));
        PyTuple_SetItem(mytuple,2,PyString_FromString(pentry.filetype().c_str()));
        PyList_SetItem( mylist, n, mytuple);
        ++n;
      }
    }else{
      mylist=PyList_New(0);
    }

  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }  
  return Py_BuildValue(_Text("O"),mylist);
}
static const char retrievePFN_doc[]=
"retrievePFN( query, buffer_to_fill, startpos, |cachesize) return true if results are not empty. The buffer is a list of tuple, [(pfname1,guid1,filetype1),(pfname2,guid2,filetype2)]";

static PyObject *
retrieveLFN(PyObject *self, PyObject *args){
  bool result;
  char* query;
  size_t start;
  size_t cachesize;
  PyObject *mylist; 
  if(!PyArg_ParseTuple(args, _Text("sll"), &query, &start, &cachesize))
    return NULL;
  std::string cquery(query);
  pool::FCBuf<pool::LFNEntry> cmyBuf(cachesize);
  try{
    pool::FCQueryParser parser(cquery);
    parser.parse();
    if(parser.isMetaQuery()){
      result=((PyFCImpl*)self)->mymetacatalog->retrieveLFN(cquery,cmyBuf,start);      
    }else{
      result=((PyFCImpl*)self)->mycatalog->retrieveLFN(cquery,cmyBuf,start);      
    }
    if(result){
      int n=0;
      size_t mysize=cmyBuf.size();
      mylist=PyList_New(int(mysize));
      for(size_t i=0; i<mysize; ++i){
        PyObject* mytuple=PyTuple_New(2);
        if(mytuple==NULL) return NULL;
        pool::LFNEntry pentry=cmyBuf[i];
        PyTuple_SetItem(mytuple,0,PyString_FromString(pentry.lfname().c_str()));
        PyTuple_SetItem(mytuple,1,PyString_FromString(pentry.guid().c_str()));
        PyList_SetItem( mylist, n, mytuple);
        ++n;
      }
    }else{
      mylist=PyList_New(0);
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }  
  return Py_BuildValue(_Text("O"),mylist);
}
static const char retrieveLFN_doc[]=
"retrieveLFN( query, buffer_to_fill, startpos, |cachesize) return true if results are not empty";

static PyObject *
retrieveMetaData(PyObject *self, PyObject *args){
  bool result;
  char* query;
  size_t start;
  size_t cachesize;
  PyObject *mylist;
  if(!PyArg_ParseTuple(args, _Text("sll"), &query, &start, &cachesize))
    return NULL;
  std::string cquery(query);
  pool::FCBuf<pool::MetaDataEntry> cmyBuf(cachesize);
  try{
    result=((PyFCImpl*)self)->mymetacatalog->retrieveMetaData(cquery,cmyBuf,start);
    if(result){
      int n=0;
      size_t mysize=cmyBuf.size();
      mylist=PyList_New(int(mysize));
      for(size_t i=0; i<mysize; ++i){
        PyObject* mydict=PyDict_New();
        pool::AttributeListSpecification spec;
        pool::AttributeListSpecification::iterator it;
        pool::MetaDataEntry mentry=cmyBuf[i];
        spec=mentry.spec();
        for(it=spec.begin(); it!=spec.end(); ++it){
          std::string colname=(*it).name();
          std::string coltype=(*it).type_name();
          if(coltype=="string"){
            std::string value;
            mentry.getAttributeValue(colname,value);
            PyDict_SetItem(mydict, PyString_FromString(colname.c_str()), PyString_FromString(value.c_str()));
          }else if(coltype=="int"){
            int value;
            mentry.getAttributeValue(colname,value);
            PyDict_SetItem(mydict, PyString_FromString(colname.c_str()), PyLong_FromLong(value));
          }
        }
        PyList_SetItem( mylist, n, mydict);
        ++n;
      }
    }else{
      mylist=PyList_New(0);
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }  
  return Py_BuildValue(_Text("O"),mylist);
}
static const char retrieveMetaData_doc[]=
"retrieveMetaData( query, buffer_to_fill, startpos, |cachesize) return true if results are not empty";

static PyObject *
retrieveGuid(PyObject *self, PyObject *args){
  bool result;
  char* query;
  size_t start;
  size_t cachesize;
  PyObject *mylist; 
  if(!PyArg_ParseTuple(args, _Text("sll"), &query, &start, &cachesize))
    return NULL;
  std::string cquery(query);
  pool::FCBuf<pool::FileCatalog::FileID> cmyBuf(cachesize);
  try{
    pool::FCQueryParser parser(cquery);
    parser.parse();
    if(parser.isMetaQuery()){
      result=((PyFCImpl*)self)->mymetacatalog->retrieveGuid(cquery,cmyBuf,start);      
    }else{
      result=((PyFCImpl*)self)->mycatalog->retrieveGuid(cquery,cmyBuf,start);      
    }
    if(result){
      int n=0;
      size_t mysize=cmyBuf.size();
      mylist=PyList_New(int(mysize));
      for(size_t i=0; i<mysize; ++i){
        PyList_SetItem( mylist, n, PyString_FromString(cmyBuf[i].c_str()) );
        ++n;
      }
    }else{
      mylist=PyList_New(0);
    }
  }catch( pool::Exception& e ){
    PyErr_SetString(PyFileCatalogError, e.what() );
    return NULL;
  }  
  return Py_BuildValue(_Text("O"),mylist);
}
static const char retrieveGuid_doc[]=
"retrieveGuid( query, buffer_to_fill, startpos, |cachesize) return true if results are not empty";

static PyObject *
hasMetaDataCatalog(PyObject *self){
  bool hasmeta=true;
  if( !((PyFCImpl*)self)->mymetacatalog ){
    hasmeta=false;
  }
  return Py_BuildValue(_Text("i"),hasmeta);
}
static const char hasMetaDataCatalog_doc[]=
"hasMetaDataCatalog() returns 1 if metadata catalog implementation available";

static PyMemberDef PyFCImpl_members[]={
  {NULL, 0,0,0,NULL}
};

/******************************************************************************/
static PyMethodDef PyFCImpl_methods[] = {
  {_Text("connect"), (PyCFunction)connect, METH_NOARGS, _Text(connect_doc) },
  {_Text("disconnect"), (PyCFunction)disconnect, METH_NOARGS, _Text(disconnect_doc) },
  {_Text("start"), (PyCFunction)start, METH_NOARGS, _Text(start_doc) },
  {_Text("commit"), (PyCFunction)commit, METH_VARARGS, _Text(commit_doc) },
  {_Text("rollback"), (PyCFunction)rollback, METH_NOARGS, _Text(rollback_doc) },
  {_Text("registerPFN"), (PyCFunction)registerPFN, METH_KEYWORDS, _Text(registerPFN_doc) },
  {_Text("registerLFN"), (PyCFunction)registerLFN, METH_KEYWORDS, _Text(registerLFN_doc) },
  {_Text("lookupBestPFN"), (PyCFunction)lookupBestPFN, METH_KEYWORDS,_Text(lookupBestPFN_doc)},
  {_Text("addReplicaPFN"), (PyCFunction)addReplicaPFN, METH_KEYWORDS,_Text(addReplicaPFN_doc)},
  {_Text("addPFNtoGuid"), (PyCFunction)addPFNtoGuid, METH_KEYWORDS,_Text(addPFNtoGuid_doc)},
  {_Text("deleteLFN"), (PyCFunction)deleteLFN, METH_KEYWORDS, _Text(deleteLFN_doc)},
  {_Text("deletePFN"), (PyCFunction)deletePFN, METH_KEYWORDS, _Text(deletePFN_doc)},
  {_Text("deleteMetaData"), (PyCFunction)deleteMetaData, METH_KEYWORDS, _Text(deleteMetaData_doc)},
  {_Text("deleteEntry"), (PyCFunction)deleteEntry, METH_KEYWORDS, _Text(deleteEntry_doc)},
  {_Text("insertPFN"), (PyCFunction)insertPFN, METH_KEYWORDS, _Text(insertPFN_doc)},
  {_Text("insertLFN"), (PyCFunction)insertLFN, METH_KEYWORDS, _Text(insertLFN_doc)},
  {_Text("lookupFileByPFN"), (PyCFunction)lookupFileByPFN, METH_VARARGS,_Text(lookupFileByPFN_doc)},
  {_Text("lookupFileByLFN"), (PyCFunction)lookupFileByLFN, METH_VARARGS,_Text(lookupFileByLFN_doc)},
  {_Text("renamePFN"), (PyCFunction)renamePFN, METH_KEYWORDS, _Text(renamePFN_doc) },
  {_Text("getMetaDataSpec"), (PyCFunction)getMetaDataSpec, METH_NOARGS,_Text(getMetaDataSpec_doc)},
  {_Text("dropMetaDataSpec"), (PyCFunction)dropMetaDataSpec, METH_NOARGS, _Text(dropMetaDataSpec_doc) },
  {_Text("createMetaDataSpec"), (PyCFunction)createMetaDataSpec, METH_KEYWORDS, _Text(createMetaDataSpec_doc) },
  {_Text("updateMetaDataSpec"), (PyCFunction)updateMetaDataSpec, METH_VARARGS, _Text(updateMetaDataSpec_doc) },
  {_Text("insertMetaData"), (PyCFunction)insertMetaData, METH_KEYWORDS, _Text(insertMetaData_doc) },
  {_Text("isReadOnly"), (PyCFunction)isReadOnly, METH_NOARGS, _Text(isReadOnly_doc) },
  {_Text("retrievePFN"), (PyCFunction)retrievePFN, METH_VARARGS, _Text(retrievePFN_doc) },
  {_Text("retrieveLFN"), (PyCFunction)retrieveLFN, METH_VARARGS, _Text(retrieveLFN_doc) },
  {_Text("retrieveMetaData"), (PyCFunction)retrieveMetaData, METH_VARARGS, _Text(retrieveMetaData_doc) },
  {_Text("retrieveGuid"), (PyCFunction)retrieveGuid, METH_VARARGS, _Text(retrieveGuid_doc) },
  {_Text("hasMetaDataCatalog"), (PyCFunction)hasMetaDataCatalog,METH_NOARGS,_Text(hasMetaDataCatalog_doc) },
  {NULL, NULL, 0, NULL}        
};

static PyTypeObject PyFCImpl_Type = {
  PyObject_HEAD_INIT(0)
  0,                        /*ob_size*/
  _Text("FileCatalog.PyFCImpl"),                /*tp_name*/
  sizeof(PyFCImpl),        /*tp_basicsize*/
  0,                        /*tp_itemsize*/
  /* methods */
  PyFCImpl_dealloc, /*tp_dealloc*/
  0,                        /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0,                        /*tp_compare*/
  0,                        /*tp_repr*/
  0,                        /*tp_as_number*/
  0,                        /*tp_as_sequence*/
  0,                        /*tp_as_mapping*/
  0,                        /*tp_hash*/
  0,                      /*tp_call*/
  0,                      /*tp_str*/
  PyObject_GenericGetAttr, /*tp_getattro*/
  PyObject_GenericSetAttr, /*tp_setattro*/
  0,                      /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT,     /*tp_flags*/
  PyFCImpl_doc,  /*tp_doc*/
  0,                      /*tp_traverse*/
  0,                      /*tp_clear*/
  0,                      /*tp_richcompare*/
  0,                      /*tp_weaklistoffset*/
  0,                      /*tp_iter*/
  0,                      /*tp_iternext*/
  PyFCImpl_methods,                      /*tp_methods*/
  PyFCImpl_members,                      /*tp_members*/
  0,                      /*tp_getset*/
  0,                      /*tp_base*/
  0,                      /*tp_dict*/
  0,                      /*tp_descr_get*/
  0,                      /*tp_descr_set*/
  0,                      /*tp_dictoffset*/
  PyFCImpl_init,                      /*tp_init*/
  PyType_GenericAlloc,   /*tp_alloc*/
  PyType_GenericNew,      /*tp_new*/
  _PyObject_Del,          /*tp_free*/
  0,                      /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  PyFCImpl_dealloc /*tp_del*/
};

/* --------------------------------------------------------------------- */
/* Initialization function for the module (*must* be called initxx) */

//DL_EXPORT(void)
extern "C" void
#ifdef WIN32
initlcg_PyFileCatalog(void)
#else
initliblcg_PyFileCatalog(void)
#endif
{
  /* Create the module and add the functions */
#ifdef WIN32
  PyObject* this_module = Py_InitModule(_Text("lcg_PyFileCatalog"), PyFCImpl_methods);
#else
  PyObject* this_module = Py_InitModule(_Text("liblcg_PyFileCatalog"), PyFCImpl_methods);
#endif
  if (PyType_Ready(&PyFCImpl_Type) <0 ) return;
  PyObject* attr = (PyObject*)&PyFCImpl_Type;
  PyObject_SetAttrString(this_module, _Text("PyFCImpl"), attr);
  /* Add some error def to the module */
  PyFileCatalogError=PyErr_NewException(_Text("FileCatalog.PyFileCatalogError"),NULL,NULL);
  /* Add some symbolic constants to the module */
  PyModule_AddObject(this_module,_Text("PyFileCatalogError"),PyFileCatalogError);
  PyModule_AddIntConstant(this_module,_Text("REFRESH"),PyFileCatalog_REFRESH);
  PyModule_AddIntConstant(this_module,_Text("ONHOLD"),PyFileCatalog_ONHOLD);
  PyModule_AddIntConstant(this_module,_Text("NO_DELETE"),PyFileCatalog_NO_DELETE);
  PyModule_AddIntConstant(this_module,_Text("DELETE_REDUNDANT"),PyFileCatalog_DELETE_REDUNDANT);
  PyModule_AddIntConstant(this_module,_Text("PFN"),PyFileCatalog_PFN);
  PyModule_AddIntConstant(this_module,_Text("LFN"),PyFileCatalog_LFN);
  PyModule_AddIntConstant(this_module,_Text("META"),PyFileCatalog_META);
  PyModule_AddIntConstant(this_module,_Text("GUID"),PyFileCatalog_GUID);
  PyModule_AddIntConstant(this_module,_Text("SEQUENTIAL"),PyFileCatalog_FileAccessPattern_SEQUENTIAL);
  PyModule_AddIntConstant(this_module,_Text("RANDOM"),PyFileCatalog_FileAccessPattern_RANDOM);
  PyModule_AddIntConstant(this_module,_Text("PRANDOM"),PyFileCatalog_FileAccessPattern_PRANDOM);
  PyModule_AddIntConstant(this_module,_Text("READ"),PyFileCatalog_FileOpenMode_READ);
  PyModule_AddIntConstant(this_module,_Text("WRITE"),PyFileCatalog_FileOpenMode_WRITE);
  PyModule_AddIntConstant(this_module,_Text("UPDATE"),PyFileCatalog_FileOpenMode_UPDATE);
}
