//====================================================================
//
//	Package    : RootStorageSvc (The POOL project)
//
//	Author     : M.Frank
//====================================================================
#ifndef POOL_IROOTKEYIOHANDLER_H
#define POOL_IROOTKEYIOHANDLER_H

// Forward declarations
class TClass;
class TKey;

/*
 *  pool Namespace
 */
namespace pool {

  /** @class IRootKeyIOHandler IRootKeyIOHandler.h RootStorageSvc/IRootKeyIOHandler.h
    * Helper to perform keyed object I/O.
    *
    * By default the class CINT::IOHandler is instantiated for each
    * Keyed root container to perform the object I/O.
    * The default however only applies, if no plugin factory was found,
    * which wants to intercept the I/O mechanism.
    * Such a class must inherit from this interface and then can implement
    * the required object commit mechanism on write as wanted by the user.
    * POOL itself does only support the basic write/read operations in a
    * similar way as they are implemented e.g. through TObject::Write/Read
    *
    * @author  M.Frank
    * @date    1/2/2003
    * @version 1.0
    */
  class IRootKeyIOHandler {
  protected:
    /// Destructor (called only by sub-classes)
    virtual ~IRootKeyIOHandler()   {     }

  public:
    /// Release object.
    virtual void release() = 0;

    /// Write object by key: Specification of the object class can considerably speed-up
    /** Write object using ROOT Key mode
      * @param cl          [IN]    ROOT Class of the specified object
      * @param knam        [IN]    Key-name of the object to be written
      * @param obj         [IN]    Pointer to the object to be written.
      * @param opt         [IN]    Root write flags.
      * @param siz         [IN]    Optimal buffer size.
      *
      * @return Number of bytes written to file. 0 or negative on failure.
      */
    virtual int  write(TClass* cl, const char* knam, void* obj, int opt, int siz=0) const = 0;

    /// Write object by key: Specification of the object class can considerably speed-up
    /** Write object using ROOT Key mode
      * @param cl          [IN]    ROOT Class of the specified object
      * @param knam        [IN]    Key-name of the object to be written
      * @param title       [IN]    Title of the object to be written
      * @param obj         [IN]    Pointer to the object to be written.
      * @param opt         [IN]    Root write flags.
      * @param siz         [IN]    Optimal buffer size.
      *
      * @return Number of bytes written to file. 0 or negative on failure.
      */
    virtual int  write(TClass* cl, const char* knam, const char* title, void* obj, int opt, int siz=0) const = 0;

    /// Read object with key.
    /** Read object using ROOT Key mode
      * @param key         [IN]    Key of the object.
      * @param obj         [IN]    Location to store pointer to the object.
      *
      * @return Number of bytes read from file. 0 or negative on failure.
      */
    virtual int  read(TKey* key, void** obj) const = 0;

    /// Read object with key.
    /** Read object using ROOT Key mode
      * @param knam        [IN]    Key-name of the object.
      * @param obj         [IN]    Location to store pointer to the object.
      *
      * @return Number of bytes read from file. 0 or negative on failure.
      */
    virtual int  read(const char* knam, void** obj) const = 0;

    /// Destroy object with key.
    /** Destroy object using ROOT Key mode
      * @param knam        [IN]    Key-name of the object.
      *
      * @return Number of bytes read from file. 0 or negative on failure.
      */
    virtual int  destroy(const char* knam) const = 0;
  };
}
#endif // POOL_IROOTKEYIOHANDLER_H
