#include "POOLCore/Token.h"
#include "POOLCore/Reference.h"
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "RootStorageSvc/RefStreamer.h"
#include <stdexcept>

#include "TFile.h"
#include "TClass.h"
#include "TBuffer.h"

using namespace pool;

RefStreamer::RefStreamer(const TypeH& typ, TClass* cl, int vsn)  
: StreamerBase(typ, cl), m_version(vsn)
{
  m_refType = 0;
  switch(vsn)  {
    case 1:
      m_refType = DataCallBack::REFERENCE;
      break;
    case 2:
      m_refType = DataCallBack::TOKEN;
      break;
    default:
      throw std::runtime_error("Fatal error: unknown reference type for class "+DbReflex::fullTypeName(typ));
      break;
  }
}

void RefStreamer::put(TBuffer& b, void* obj)  {
  DataCallBack* caller = DbDataHandlerGuard::caller();
  switch ( m_refType )  {
    case DataCallBack::REFERENCE:
      {
        std::pair<unsigned int, unsigned int> oid(~0x0, ~0x0);
        if (caller) caller->notify(DataCallBack::PUT,m_refType,m_type,obj,&oid);
        b.WriteFastArray(&oid.first, 2);
      }
      break;
    case DataCallBack::TOKEN:
      {
        Token* t = (Token*)obj;
        if (caller) caller->notify(DataCallBack::PUT,m_refType,m_type,obj,&t->oid());
        b.WriteFastArray(&t->oid().first, 2);
      }
      break;
  }
}

void RefStreamer::get(TBuffer& b, void* obj)  {
  DataCallBack* caller = DbDataHandlerGuard::caller();
  switch ( m_refType )  {
    case DataCallBack::REFERENCE:    {
      std::pair<unsigned int, unsigned int> oid(~0x0, ~0x0);
      b.ReadFastArray(&oid.first, 2);
      if (caller) caller->notify(DataCallBack::GET,m_refType,m_type,obj,&oid);
      // std::cout << "Resolving ref:" << (void*)caller << std::endl;
    }
    break;
    case DataCallBack::TOKEN:        {
      Token* t = (Token*)obj;
      b.ReadFastArray(&t->oid().first, 2);
      if (caller) caller->notify(DataCallBack::GET,m_refType,m_type,t,&t->oid());
      // std::cout << "Serialize token:" << t->toString() << std::endl;
    }
    break;
  }
}

void RefStreamer::getLong(TBuffer& b, void* obj)  {
  DataCallBack* caller = DbDataHandlerGuard::caller();
  std::pair<unsigned long, unsigned long> oid(~0x0, ~0x0);
  Token* t = (Token*)obj;
  switch ( m_refType )  {
    case DataCallBack::REFERENCE:    {
      b.ReadFastArray(&oid.first, 2);
      if (caller) caller->notify(DataCallBack::GET,m_refType,m_type,obj,&oid);
    }
    break;
    case DataCallBack::TOKEN:        {
      b.ReadFastArray(&oid.first, 2);
      t->oid().first = oid.first;
      t->oid().second = oid.second;
      if (caller) caller->notify(DataCallBack::GET,m_refType,m_type,t,&t->oid());
      //std::cout << "Serialize token:" << t->toString() << std::endl;
    }
    break;
  }
}

void RefStreamer::operator()(TBuffer &b, void *obj)  {
  try {
    UInt_t start, count, tmp, checkSum;
    Int_t file_version = ((TFile*)b.GetParent())->GetVersion();
    if ( file_version >= 40000 )  {
      if ( b.IsReading() )  {
        // read the class version from the buffer
        /* Version_t vsn = */ b.ReadVersion(&start, &count, 0);
        switch(count) {
          case 22:              // These tokens were written as pair<long,long>
            b >> checkSum;      //
            getLong(b, obj);    // There was a bug in ROOT....
            break;              // see TBuffer::ReadFastArray(ulong*, int)
          case 14:              // Normal case: version:checksum+8 Bytes
            b >> checkSum;      //
            get(b, obj);        // Serialize the object
            break;
          case 10:              // Normal case: version:checksum+8 Bytes
            get(b, obj);        // Serialize the object
            break;
          case 8:               // Without checksum and version
            get(b, obj);        // Serialize the object
            break;
          default:              // No better idea
            get(b, obj);        // Serialize the object
            break;
        }
        // Check that the buffer position correesponds to the byte count
        b.CheckByteCount(start, count, m_root);
        return;
      }
      count = b.WriteVersion(m_root, true);
      put(b, obj);
      b.SetByteCount(count, true);
      return;
    }
    // Some very ugly cases used this:
    else if ( m_version == 2 )  {
      if ( b.IsReading() )  {
        Version_t vsn = b.ReadVersion(&start, &count, m_root);
        switch(vsn)  {
          case 2:
            b >> tmp;
            get(b, obj);
            b >> tmp;
            break;
          default:
            switch(m_version)  {
          case 2:
            b.SetBufferOffset(start+4);
            get(b, obj);
            return;
          default:
            get(b, obj);
            break;
            }
        }
        return;
      }
      count = b.WriteVersion(m_root, true);
      put(b, obj);
      b.SetByteCount(count, true);
      return;
    }
    if ( b.IsReading() )  {
      get(b, obj);
    }
    else  {
      put(b, obj);
    }
  }
  catch( const std::exception& e )    {
    std::string err = "Class:" + DbReflex::fullTypeName(m_type) + "> Exception in object I/O";
    err += e.what();
    throw std::runtime_error(err);
  }
  catch( ... )    {
    std::string err = "Class:" + DbReflex::fullTypeName(m_type) + "> Exception in object I/O";
    throw std::runtime_error(err);
  }
}
