//====================================================================
//    Root Database container implementation
//--------------------------------------------------------------------
//
//    Package    : pool (The POOL project)
//
//    Author     : M.Frank
//====================================================================
// Framework include files
#include "POOLCore/strcasecmp.h"
#include "StorageSvc/DbObject.h"
#include "StorageSvc/DbOption.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbColumn.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbInstanceCount.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/IClassLoader.h"
#include "StorageSvc/DbArray.h"
#include "StorageSvc/DbPrint.h"
#include "StorageSvc/DbTransaction.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "StorageSvc/DbOptionCallback.h"
#include "Reflex/Reflex.h"

// Local implementation files
#include "RootTreeContainer.h"
#include "RootDataPtr.h"
#include "RootDBase.h"
using namespace pool;

// Root include files
#include "RootSelect.h"
#include "TROOT.h"
#include "TTree.h"
#include "TFile.h"
#include "TLeaf.h"
#include "TBranch.h"

//===========================  START *** BRANCH PATCH
#include "StorageSvc/IClassHandler.h"
typedef IClassLoader* ClassHandler_t;
//===========================  END *** BRANCH PATCH

static UCharDbArray  s_char_Blob;
static IntDbArray   s_int_Blob;

RootTreeContainer::RootTreeContainer(IOODatabase* idb)
: DbContainerImp(idb), m_tree(0), m_type(0), m_dbH(POOL_StorageType), 
  m_rootDb(0), m_branchName()
{
  DbInstanceCount::increment(this);
}

/// Standard destructor
RootTreeContainer::~RootTreeContainer()   {
  DbInstanceCount::decrement(this);
  close();
}

/// Ask if a given shape is supported
DbStatus RootTreeContainer::isShapeSupported(const DbTypeInfo* typ) const  {
  return typ == m_type;
}

long long int RootTreeContainer::size()    {
  long long int s1;
  if (m_tree == 0) {
    return 0;
  } 
  else if ( m_branchName.empty() ) {
    s1 = m_tree->GetEntries();
  }
  else {
    TBranch * pBranch = m_tree->GetBranch(m_branchName.c_str());
    if (pBranch == 0) return 0;
    s1 = pBranch->GetEntries();
  }
  long long int s2 = DbContainerImp::size();
  return s1+s2;
}

/// Access branch by name
TBranch* RootTreeContainer::branch(const std::string& nam)  const  {
  Branches::const_iterator k;
  for(k=m_branches.begin(); k !=m_branches.end(); ++k) {
    if ( (*k).branch->GetName() == nam )  {
      return (*k).branch;
    }
  }
  return 0;
}

/// Execute object modification requests during a transaction
DbStatus RootTreeContainer::endTransaction(DbTransaction& refTr)  {
  DbStatus status = Success;
  switch(refTr.state()) {
  case Transaction::TRANSACT_COMMIT:
    break;
  case Transaction::TRANSACT_FLUSH:
    if ( m_tree )  {
      if (m_branchName.empty()) {
        m_tree->AutoSave();
      } else {
        status = finishTransAct();
      }
    }
    else {
      status = Error;
    }
    break;
  case Transaction::TRANSACT_START:
    break;
  case Transaction::TRANSACT_ROLLBACK:
    break;
  default:
    status = Error;
    break;
  }
  return status;
}

DbStatus RootTreeContainer::writeObject(TransactionStack::value_type& ent) {
  RootDataPtr p(0);
  RootDataPtr user(&ent.objH);
  const void* ptr = ent.objH ? ent.objH : ent.call->object();
  RootDataPtr context(ptr);
  DbStatus status = ent.call->start(DataCallBack::PUT, context.ptr, &user.ptr);
  if ( status.isSuccess() && ent.call ) {
    int icol;
    int num_bytes = 0;
    Branches::iterator k;
    DbDataHandlerGuard call_environment(ent.call, &m_dbH);
    for(k=m_branches.begin(), icol=0; k !=m_branches.end(); ++k, ++icol) {
      BranchDesc& dsc = (*k);
      DbStatus sc = ent.call->bind( DataCallBack::PUT,
                                    dsc.column,
                                    icol,
                                    user.ptr,
                                    &p.ptr);
      if ( sc.isSuccess() && sc.code() != DataCallBack::SKIP ) {
        int debug = 0;
        gDebug = debug;

        switch ( dsc.column->typeID() )    {
        case DbColumn::ANY:
        case DbColumn::POINTER:
          dsc.object            = p.ptr;
          p.ptr                 = &dsc.object;
          break;
        case DbColumn::BLOB:
          s_char_Blob.m_size    = p.blobSize();
          s_char_Blob.m_buffer  = (unsigned char*)p.blobData();
          dsc.object            = &s_char_Blob;
          p.ptr                 = &dsc.object;
          break;
        case DbColumn::TOKEN:
          s_int_Blob.m_size     = 2*p.linkSize(); 
          s_int_Blob.m_buffer   = p.linkData(); 
          dsc.object            = &s_int_Blob;
          p.ptr                 = &dsc.object;
          break;
        case DbColumn::STRING:
        case DbColumn::LONG_STRING:
          p.cptr                = p.string();
          break;
        case DbColumn::NTCHAR:
        case DbColumn::LONG_NTCHAR:
          p.ptr                 = p.deref();
          break;
        default:
          break;
        }
        if ( 0 == p.ptr )   {
          DbPrint err( m_name);
          err << DbPrint::ERROR 
              << "[RootTreeContainer] Could not write an object" 
              << DbPrint::endmsg;
          return Error;
        }
        //if (p.ptr != dsc.branch->GetAddress()) {
          dsc.branch->SetAddress(p.ptr);
        //}
        if (!m_branchName.empty()) {
          num_bytes = dsc.branch->Fill();
        }
      }
    }
    if (m_branchName.empty()) {
      num_bytes = m_tree->Fill();
    }
    for(k=m_branches.begin(); k !=m_branches.end(); ++k) {
      switch ( (*k).column->typeID() )    {
      case DbColumn::BLOB:
        s_char_Blob.release(false);
        break;
      case DbColumn::TOKEN:
        s_int_Blob.release(false);
        break;
      default:
        break;
      }
    }
    if ( num_bytes > 1 )  {
      m_rootDb->addByteCount(RootDBase::WRITE_COUNTER, num_bytes);
      return ent.call->end(DataCallBack::PUT, context.ptr);
    }
    DbPrint err( m_name);
    err << DbPrint::ERROR 
        << "[RootTreeContainer] Could not write an object" 
        << DbPrint::endmsg;
    return Error;
  }
  else {
    DbPrint err( m_name);
    err << DbPrint::ERROR << "[RootTreeContainer] "
        << "Could not start a callback sequence when writing an object" 
        << DbPrint::endmsg;
  }
  return status;
}

/// Fetch refined object address. Default implementation returns identity
DbStatus RootTreeContainer::fetch(const DbLink& linkH, DbLink& stmt)  {
  return DbContainerImp::fetch(linkH, stmt);
}

// Fetch next object address of the selection to set token
DbStatus RootTreeContainer::fetch(DbSelect& sel)  {
  if ( sel.criteria().length() == 0 || sel.criteria() == "*" )  {
    sel.link().second++;
    return DbContainerImp::fetch(sel.link(), sel.link());
  }
  DbSelect::Ptr<RootSelect>* stmt = 
    dynamic_cast<DbSelect::Ptr<RootSelect>* >(sel.statement());
  if ( stmt ) {
    RootSelect* selStmt = stmt->m_ptr;
    if ( selStmt )  {
      Branches::iterator k;
      int cur  = sel.link().second;
      Long64_t last = m_tree->GetEntries();
      for(k=m_branches.begin(); k != m_branches.end(); ++k)  {
        BranchDesc& dsc = (*k);
        int typ = dsc.column->typeID();
        switch ( typ )    {
        case DbColumn::TOKEN:
        case DbColumn::POINTER:
        case DbColumn::BLOB:
          dsc.branch->SetAddress(&dsc.buffer);
        default:
          break;
        }
      }
      selStmt->SetTree(m_tree);
      // loop on all selected entries
      for( ++cur; cur < last; ++cur) {
        m_tree->LoadTree(cur);
        selStmt->GetNdata();
        if ( selStmt->EvalInstance(0) != 0 ) {
          sel.link().second = cur;
          return Success;
        }
      }
    }
  }
  return Error;
}

DbStatus 
RootTreeContainer::loadObject(DataCallBack* call,DbLink& oid,DbAccessMode /* mode */) {
  RootDataPtr user(0);
  RootDataPtr context(0);
  DbStatus status = Error;
  int evt_id = oid.second;
  // Read data; disable branch after reading the information
  try {
    // gDebug = 2;
    //printBranches(m_tree);
    DbDataHandlerGuard call_environment(call, &m_dbH);
    status = call->start(DataCallBack::GET, context.ptr, &user.ptr);
    if ( status.isSuccess() ) {
      int numBytesBranch, icol = 0, numBytes = 0;
      Branches::iterator k;
      for(k=m_branches.begin(); k != m_branches.end(); ++k,++icol)  {
        BranchDesc& dsc = (*k);
        RootDataPtr p(0), q(0);
        DbStatus sc = call->bind(DataCallBack::GET,dsc.column,icol,user.ptr,&p.ptr);
        q.ptr = p.ptr;
        if ( sc.isSuccess() && sc != DataCallBack::SKIP ) {
          int typ = dsc.column->typeID();
          switch ( typ )    {
          case DbColumn::STRING:
          case DbColumn::LONG_STRING:
          case DbColumn::NTCHAR:
          case DbColumn::LONG_NTCHAR:
            break;
          case DbColumn::TOKEN:
            dsc.object = &s_int_Blob;
            p.ptr      = &dsc.object;
            dsc.branch->SetAddress(p.ptr);
            break;
          case DbColumn::BLOB:
            dsc.object = &s_char_Blob;
            p.ptr      = &dsc.object;
            dsc.branch->SetAddress(p.ptr);
            break;
          case DbColumn::ANY:
          case DbColumn::POINTER:
          default:
            dsc.object = p.ptr;
            dsc.branch->SetAddress(dsc.object);
            break;
          }
          numBytesBranch = dsc.branch->GetEntry(evt_id);
          numBytes += numBytesBranch;
          if ( numBytesBranch > 0 )     {
            switch ( typ )    {
            case DbColumn::STRING:
            case DbColumn::LONG_STRING:
              p.ptr = dsc.leaf->GetValuePointer();
              *q.str = p.c_str;
              break;
            case DbColumn::NTCHAR:
            case DbColumn::LONG_NTCHAR:
              p.ptr = dsc.leaf->GetValuePointer();
              ::strcpy(q.c_str, p.c_str);
              break;
            case DbColumn::BLOB:
              q.blob->adopt((char*)s_char_Blob.m_buffer,s_char_Blob.m_size);
              s_char_Blob.release(false);
              break;
            case DbColumn::TOKEN:  {
              q.links->clear();
              q.links->reserve(s_int_Blob.m_size/2);
              for(int *r=s_int_Blob.m_buffer, *s=r+s_int_Blob.m_size; r<s; r+=2)
                q.links->push_back(DbLink(*r,*(r+1)));
              s_int_Blob.release(false);
              }
              break;
            case DbColumn::POINTER:
              break;
            }
          }
        }
      }
      if ( numBytes > 1 )   {
        /// Update statistics
        m_rootDb->addByteCount(RootDBase::READ_COUNTER, numBytes);
        /// Terminate callback
        return call->end(DataCallBack::GET, user.ptr);
      }
    }
  }
  catch( const std::exception& e )    {
    DbPrint err(m_name);
    err << DbPrint::FATAL << "[RootTreeContainer] "
        << "STL C++ Exception: " << e.what() << DbPrint::endmsg;
  }
  catch (...)   {
    DbPrint err(m_name);
    err << DbPrint::FATAL << "[RootTreeContainer] "
        << "Unknown exception occurred. Cannot give more details." 
        << DbPrint::endmsg;
  }
  DbPrint log(m_name);
  log << DbPrint::INFO << "Cannot load entry No." << evt_id << "..." 
      << (m_branchName.empty() ? " Tree has " : " Branch has " ) 
      << size() << " Entries in total." << DbPrint::endmsg;
  return Error;
}

DbStatus RootTreeContainer::close()   {
  DbOptionCallback::getOptions(this, DbOptionCallback::_CONTAINER_ONCLOSE, m_name);
  m_dbH = DbDatabase(POOL_StorageType);
  if ( m_tree )   {
    //m_tree->Print();
  }
  for(Branches::iterator k=m_branches.begin(); k != m_branches.end(); ++k)  {
    BranchDesc& dsc = (*k);
    if ( dsc.buffer && dsc.clazz )  {
      // This somehow fails for templates.
      dsc.clazz->Destructor(dsc.buffer);
    }
  }
  m_branches.clear();
  m_rootDb = 0;
  m_tree = 0;
  return DbContainerImp::close();
}

DbStatus RootTreeContainer::open( const DbDatabase& dbH, 
                                  const std::string& nam, 
                                  const DbTypeInfo* info, 
                                  DbAccessMode mode)  
{
  DbPrint log(dbH.name());
  m_branches.clear();
  m_name = nam;
  if ( dbH.isValid() && info )    {
    const DbTypeInfo::Columns& cols = info->columns();
    DbTypeInfo::Columns::const_iterator i;
    std::string treeName(nam);
    for(std::string::iterator j = treeName.begin(); j != treeName.end(); ++j )    {
       if ( *j == '/' ) *j = '_';
    }
    if (cols.size() == 1) {
      std::string::size_type inx = nam.find('(');
      if (inx != std::string::npos) {
        std::string::size_type inx2 = nam.find(')');
        if (inx2 == std::string::npos || inx2 != nam.size()-1) {
          log << DbPrint::ERROR << "Misspecified branch name in " << m_name << "."
              << DbPrint::endmsg;
          return Error;
        }
        m_branchName = treeName.substr(inx+1, inx2-inx-1);
        treeName = treeName.substr(0, inx);
      }
    }
    IDbDatabase* idb = dbH.info();
    m_rootDb = dynamic_cast<RootDBase*>(idb);
    TDirectory::TContext dirCtxt(m_rootDb->file());
    m_tree = (TTree*)m_rootDb->file()->Get(treeName.c_str());
    DbOptionCallback::getOptions(this, DbOptionCallback::_CONTAINER_OPT, nam);

    bool hasBeenCreated = (m_branchName.empty() 
                             ? m_tree != 0 
                             : (m_tree && m_tree->GetBranch(m_branchName.c_str()) != 0));
    if ( hasBeenCreated && (mode&pool::READ || mode&pool::UPDATE) )   {
      int count;
      IClassHandler* hnd = 0;
      if ( !m_tree->InheritsFrom(TTree::Class()) )   {
        log << DbPrint::ERROR << "Cannot open the container " << m_name << " of type "
            << ROOTTREE_StorageType.storageName() << "." << DbPrint::endmsg
            << "The specified container is not a ROOT " << (m_branchName.empty() ? "Tree" : "Branch")
            << ", but rather of class " << m_tree->IsA()->GetName() << "."
            << DbPrint::endmsg;
        return Error;
      }
      m_branches.resize(cols.size());
      for(i = cols.begin(), count = 0; i != cols.end(); ++i, ++count )   {
        std::string nam  = (m_branchName.empty() ? (*i)->name() : m_branchName);
        const char* col_nam  = nam.c_str();
        TBranch* pBranch = m_tree->GetBranch(col_nam);
        if (!pBranch && m_branchName.empty()) {
          for ( std::string::iterator j = nam.begin(); j != nam.end(); ++j ) {
            if ( !::isalnum(*j) ) *j = '_';
          }
          col_nam  = nam.c_str();
          pBranch = m_tree->GetBranch(col_nam);
        }
        if ( pBranch )    {
          const DbColumn* c = *i;
          BranchDesc& dsc = m_branches[count];
          TClass* cl = 0;
          TLeaf* leaf = pBranch->GetLeaf(col_nam);
          switch ( (*i)->typeID() )    {
          case DbColumn::ANY:
          case DbColumn::BLOB:
          case DbColumn::TOKEN:
          case DbColumn::POINTER:
            hnd = loader()->nativeHandler(pBranch->GetClassName(), true);
            if ( hnd )  {
              cl = (TClass*)hnd->nativeClass();
            }
            if ( 0 == cl )  {
              log << DbPrint::DEBUG << "Cannot open the container " << m_name << " of type "
                  << ROOTTREE_StorageType.storageName()
                  << " Class " << pBranch->GetClassName() << " is unknown."
                  << DbPrint::endmsg;
              return Error;
            }
            dsc = BranchDesc(cl,pBranch,leaf,cl->New(),c,hnd);
            break;
          case DbColumn::CHAR:
          case DbColumn::UCHAR:
          case DbColumn::BOOL:
          case DbColumn::SHORT:
          case DbColumn::USHORT:
          case DbColumn::INT:
          case DbColumn::LONG:
          case DbColumn::UINT:
          case DbColumn::ULONG:
          case DbColumn::FLOAT:
          case DbColumn::DOUBLE:
          case DbColumn::STRING:
          case DbColumn::LONG_STRING:
          case DbColumn::NTCHAR:
          case DbColumn::LONG_NTCHAR:
            dsc.clazz = 0;
            dsc.leaf = leaf;
            dsc.branch = pBranch;
            dsc.buffer = 0;
            dsc.object = 0;
            dsc.column = *i;
            dsc.handler = 0;
            break;
          default:
            return Error;
          }
        }
        else  {
          log << DbPrint::WARNING << "Branch with name:" << col_nam 
              << " not present in container:" << m_name << " of type "
              << ROOTTREE_StorageType.storageName()
              << DbPrint::endmsg;
        }
      }
      log << DbPrint::DEBUG << "Opened container " << m_name << " of type "
          << ROOTTREE_StorageType.storageName()
          << DbPrint::endmsg;
      m_dbH = dbH;
      m_type = info;
      DbOptionCallback::getOptions(this, DbOptionCallback::_CONTAINER_ONOPEN, m_name);
      return Success;
    }
    else if ( !hasBeenCreated && mode&pool::CREATE )    {
      int count, defCompression=1, defSplitLevel=99, 
          defAutoSave=16*1024*1024, defBufferSize=16*1024;
      DbStatus res = Success;
      try   {
        DbOption opt1("DEFAULT_COMPRESSION","");
        DbOption opt2("DEFAULT_SPLITLEVEL","");
        DbOption opt3("DEFAULT_AUTOSAVE","");
        DbOption opt4("DEFAULT_BUFFERSIZE","");
        dbH.getOption(opt1);
        dbH.getOption(opt2);
        dbH.getOption(opt3);
        dbH.getOption(opt4);
        opt1._getValue(defCompression);
        opt2._getValue(defSplitLevel);
        opt3._getValue(defAutoSave);
        opt4._getValue(defBufferSize);
        if (!m_tree) {
          m_tree = new TTree(treeName.c_str(), treeName.c_str(), defSplitLevel);
          m_tree->SetDirectory(m_rootDb->file());
          // Autosave every mega byte...
          m_tree->SetAutoSave(defAutoSave);
        }
        //            - C : a character string terminated by the 0 character
        //            - B : an 8 bit signed integer (Char_t)
        //            - b : an 8 bit unsigned integer (UChar_t)
        //            - S : a 16 bit signed integer (Short_t)
        //            - s : a 16 bit unsigned integer (UShort_t)
        //            - I : a 32 bit signed integer (Int_t)
        //            - i : a 32 bit unsigned integer (UInt_t)
        //            - F : a 32 bit floating point (Float_t)
        //            - D : a 64 bit floating point (Double_t)
        m_branches.resize(cols.size());
        for (i = cols.begin(), count = 0; i != cols.end(); ++i, ++count )   {
          DbStatus iret = Success;
          BranchDesc& dsc = m_branches[count];
          switch ( (*i)->typeID() )    {
          case DbColumn::CHAR:       iret=addBranch(*i,dsc,"/B"); break;
          case DbColumn::UCHAR:
          case DbColumn::BOOL:       iret=addBranch(*i,dsc,"/B"); break;
          case DbColumn::SHORT:      iret=addBranch(*i,dsc,"/S"); break;
          case DbColumn::USHORT:     iret=addBranch(*i,dsc,"/s"); break;
          case DbColumn::INT:        iret=addBranch(*i,dsc,"/I"); break;
          case DbColumn::LONG:       iret=addBranch(*i,dsc,"/I"); break;
          case DbColumn::UINT:       iret=addBranch(*i,dsc,"/i"); break;
          case DbColumn::ULONG:      iret=addBranch(*i,dsc,"/i"); break;
          case DbColumn::FLOAT:      iret=addBranch(*i,dsc,"/F"); break;
          case DbColumn::DOUBLE:     iret=addBranch(*i,dsc,"/D"); break;
          case DbColumn::STRING:
          case DbColumn::LONG_STRING:
          case DbColumn::NTCHAR:
          case DbColumn::LONG_NTCHAR:
            iret=addBranch(*i,dsc,"/C");
            break;
          case DbColumn::TOKEN:      
            iret=addObject(*i,dsc,"IntDbArray", defSplitLevel, defBufferSize);
            break;
          case DbColumn::BLOB:
            iret=addObject(*i,dsc,"UCharDbArray", defSplitLevel, defBufferSize);
            break;
          case DbColumn::POINTER:
          case DbColumn::ANY:
            iret=addObject(*i,dsc,(*i)->typeName(), defSplitLevel, defBufferSize);
            break;
          default:
            return Error;
          }
          if ( !iret.isSuccess() )  {
            res = iret;
          }
        }
        if ( res.isSuccess() )    {
          log << DbPrint::DEBUG << "Opened container " << m_name << " of type "
              << ROOTTREE_StorageType.storageName()
              << DbPrint::endmsg;
          m_dbH  = dbH;
          m_type = info;
          DbOptionCallback::getOptions(this, DbOptionCallback::_CONTAINER_ONOPEN, m_name);
          return Success;
        }
        debugBreak(nam, "Cannot open ROOT container(Tree/Branch)", false);
        return res;
      }
      catch( const std::exception& e )    {
        debugBreak(nam, "Cannot open ROOT container(Tree/Branch)", e, false);
        res = Error;
      }
      catch (...)   {
        DbPrint err( m_name);
        err << DbPrint::FATAL << "Unknown exception occurred. Cannot give more details." 
            << DbPrint::endmsg;
        debugBreak(nam, "Cannot open ROOT container(Tree/Branch)");
        res = Error;
      }
    }
  }
  log << DbPrint::ERROR << "Cannot open container, invalid Database handle." 
      << DbPrint::endmsg;
  return Error;
}

// Define selection criteria
DbStatus  RootTreeContainer::select(DbSelect& sel)    {
  if ( 0 != m_tree )  {
    if( sel.criteria().length() == 0 || sel.criteria() == "*" )  {
      sel.link().second = -1;
      return Success;
    }
    else  {
      RootSelect* stmt = new RootSelect(sel.criteria().c_str(), m_tree);
      sel.setStatement(new DbSelect::Ptr<RootSelect>(stmt));
      sel.link().second = -1;
      return Success;
    }
  }
  return Error;
}

DbStatus  RootTreeContainer::addObject(const DbColumn* col,
                                       BranchDesc& dsc,
                                       const std::string& typ,
                                       int defSplitLevel, 
                                       int defBufferSize)
{
  IClassHandler* hnd = loader()->handler(typ, true);
  if ( 0 != hnd )  {
    try {
      dsc.buffer  = 0;
      dsc.object  = 0;
      dsc.column  = col;
      dsc.handler = hnd;
      dsc.clazz = (TClass*)hnd->nativeClass();
      if ( 0 != dsc.clazz )  {
        if ( dsc.clazz->GetStreamerInfo() )  {
          int split = dsc.clazz->CanSplit() ? defSplitLevel : 0;
          std::string nam  = (m_branchName.empty() ? col->name() : m_branchName);
          if (m_branchName.empty()) {
            for ( std::string::iterator j = nam.begin(); j != nam.end(); ++j )    {
              if ( !::isalnum(*j) ) *j = '_';
            }
          }
          dsc.branch  = m_tree->Branch(nam.c_str(),   // Branch name
                                       dsc.clazz->GetName(), // Object class
                                       (void*)&dsc.buffer,   // Object address
                                       defBufferSize, // Buffer size
                                       split);        // Split Mode (Levels)
          if ( dsc.branch )  {
            dsc.leaf = dsc.branch->GetLeaf(nam.c_str());
            dsc.branch->SetAutoDelete(kFALSE);
            // AUTO-DELETE is now OFF. 
            // This ensures, that all objects can be deleted
            // by the framework. Keep the created object in the
            // branch descriptor to allow selections
            return Success;
          }
        }
      }
    }
    catch( const std::exception& e )    {
      debugBreak(m_name, "Cannot attach ROOT object branch.", e);
    }
    catch (...)   {
      DbPrint err( m_name);
      err << DbPrint::FATAL << "Unknown exception occurred. Cannot give more details." 
          << DbPrint::endmsg;
      debugBreak(m_name, "Cannot attach ROOT object branch.", true);
    }
  }
  DbPrint log( m_name);
  log << DbPrint::ERROR << "Failed to open the container " << m_name << " of type "
      << ROOTTREE_StorageType.storageName()
      << " Class " << typ << " is unknown."
      << DbPrint::endmsg;
  return Error;
}

DbStatus 
RootTreeContainer::addBranch(const DbColumn* col,BranchDesc& dsc,const std::string& desc) {
  const char* nam  = (m_branchName.empty() ? col->name().c_str() : m_branchName.c_str());
  std::string  coldesc = col->name() + desc;
  char buff[32];
  dsc.branch = m_tree->Branch(nam, buff, coldesc.c_str(), 4096);
  if ( dsc.branch )  {
    dsc.leaf = dsc.branch->GetLeaf(nam);
    dsc.clazz = 0;
    dsc.column = col;
    dsc.buffer = 0;
    dsc.object = 0;
    dsc.handler = 0;
    return Success;
  }
  return Error;
}

/// Access options
DbStatus RootTreeContainer::getOption(DbOption& opt)  const  { 
  if ( m_tree )  {
    const char* n = opt.name().c_str();
    if ( !strcasecmp(n,"BRANCH") )  {
      TBranch* b = branch(opt.option());
      return opt._setValue((void*)b);
    }
    else if ( ::toupper(n[0])=='B' && opt.name().length() > 7 ) {
      TBranch* b = branch(opt.option());
      if ( b )  {
        switch(::toupper(n[7]))  {
        case 'B':
          if ( !strcasecmp(n+7,"BASKET_SIZE") )
            return opt._setValue(int(b->GetBasketSize()));
          break;
        case 'C':
          if ( !strcasecmp(n+7,"COMPRESSION_LEVEL") )
            return opt._setValue(int(b->GetCompressionLevel()));
          break;
        case 'E':
          if ( !strcasecmp(n+7,"ENTRIES") )
            return opt._setValue(int(b->GetEntries()));
          break;
        case 'M':
          if ( !strcasecmp(n+7,"MAX_BASKETS") )
            return opt._setValue(int(b->GetMaxBaskets()));
          break;
        case 'N':
          if ( !strcasecmp(n+7,"NLEAVES") )
            return opt._setValue(int(b->GetNleaves()));
          break;
        case 'S':
          if ( !strcasecmp(n+7,"SPLIT_LEVEL") )
            return opt._setValue(int(b->GetSplitLevel()));
          break;
        case 'T':
          if ( !strcasecmp(n+7,"TOTAL_SIZE") )
            return opt._setValue(double(b->GetTotalSize()));
          else if ( !strcasecmp(n+7,"TOTAL_BYTES") )
            return opt._setValue(double(b->GetTotBytes()));
          break;
        case 'Z':
          if ( !strcasecmp(n+7,"ZIP_BYTES") )
            return opt._setValue(double(b->GetZipBytes()));
          break;
        }
      }
    }
    else if ( !strcasecmp(n,"TREE") )  {
      return opt._setValue((void*)m_tree);
    }
    else if ( ::toupper(n[0]=='T') && opt.name().length() > 6 ) {
      switch(::toupper(n[5]))   {
      case 'B':
        if ( !strcasecmp(n+5,"BRANCH_IDX") )  {
          int idx = 0;
          opt._getValue(idx);
          TObjArray* arr = m_tree->GetListOfBranches();
          return opt._setValue((void*)arr->At(idx));
        }
        if ( !strcasecmp(n+5,"BRANCH_NAME") )  {
          const char* br_nam = 0;
          opt._getValue(br_nam);
          if ( br_nam )  {
            return opt._setValue((void*)m_tree->GetBranch(br_nam));
          }
          opt._setValue((void*)0);
        }
        if ( !strcasecmp(n+5,"BRANCH_IDX_NAME") )  {
          int idx = 0;
          opt._getValue(idx);
          TObjArray* arr = m_tree->GetListOfBranches();
          TBranch* br = (TBranch*)arr->At(idx);
          if ( br )  {
            return opt._setValue((char*)br->GetName());
          }
          opt._setValue((char*)0);
        }
        break;
      case 'E':
        if ( !strcasecmp(n+5,"ENTRIES") )
          return opt._setValue(int(m_tree->GetEntries()));
        break;
      case 'F':
        if ( !strcasecmp(n+5,"FILE_NUMBER") )
          return opt._setValue(int(m_tree->GetFileNumber()));
        break;
      case 'M':
        if ( !strcasecmp(n+5,"MAX_SIZE") )
          return opt._setValue((long long int)m_tree->GetMaxTreeSize());
        else if ( !strcasecmp(n+5,"MAX_VIRTUAL_SIZE") )
          return opt._setValue(int(m_tree->GetMaxVirtualSize()));
        break;
      case 'N':
        if ( !strcasecmp(n+5,"NBRANCHES") )
          return opt._setValue(int(m_tree->GetNbranches()));
        break;
      case 'T':
        if ( !strcasecmp(n+5,"TOTAL_BYTES") )
          return opt._setValue(double(m_tree->GetTotBytes()));
        break;
      case 'Z':
        if ( !strcasecmp(n+5,"ZIP_BYTES") )
          return opt._setValue(double(m_tree->GetZipBytes()));
        break;
      }
    }
  }
  return Error;  
}

/// Set options
DbStatus RootTreeContainer::setOption(const DbOption& opt)  { 
  if ( m_tree )  {
    const char* n = opt.name().c_str();
    if ( ::toupper(n[0]) == 'B' )  {
      TBranch* b = branch(opt.option());
      if ( b )  {
        switch(::toupper(n[7]))   {
        case 'A':
          if ( !strcasecmp(n+7,"AUTODELETE") )  {
            int val=0;
            opt._getValue(val);
            b->SetAutoDelete(val!=0);
            return Success;
          }
          break;
        case 'C':
          if ( !strcasecmp(n+7,"COMPRESSION_LEVEL") )  {
            int val=1;
            opt._getValue(val);
            b->SetCompressionLevel(val);
            return Success;
          }
          break;
        case 'P':
          if ( !strcasecmp(n+7,"PRINT") )  {
            b->Print();
            std::cout << std::endl;
            return Success;
          }
          break;
        default:
          break;
        }
      }
    }
    else if ( ::toupper(n[0]) == 'T' )  {
      switch(::toupper(n[5]))   {
      case 'A':
        if ( !strcasecmp(n+5,"AUTO_SAVE") )  {
          int val=1;
          opt._getValue(val);
          m_tree->SetAutoSave(val);
          return Success;
        }
        break;
      case 'M':
        if ( !strcasecmp(n+5,"MAX_SIZE") )  {
          long long int val=1;
          opt._getValue(val);
          m_tree->SetMaxTreeSize(val);
          return Success;
        }
        break;
      case 'P':
        if ( !strcasecmp(n+5,"PRINT") )  {
          m_tree->Print();
          std::cout << std::endl;
          return Success;
        }
        break;
      }
    }
  }
  return Error;  
}

DbStatus RootTreeContainer::finishTransAct()    {
  Branches::const_iterator k;
  for(k=m_branches.begin(); k !=m_branches.end(); ++k) {
    Long64_t branchEntries = k->branch->GetEntries();
    Long64_t treeEntries = m_tree->GetEntries();
    if (branchEntries > treeEntries) {
      TIter next(m_tree->GetListOfBranches());
      TBranch * b = 0;
      while((b = (TBranch*)next())){
        if (b->GetEntries() != branchEntries) {
          DbPrint log(m_name);
          log << DbPrint::ERROR << "Every branch must have the same number of entries." 
              << DbPrint::endmsg;
          return Error;
        }
      }
      m_tree->SetEntries(branchEntries);
      m_tree->AutoSave();
    } else if (branchEntries < treeEntries) {
      DbPrint log(m_name);
      log << DbPrint::ERROR << "Every branch must have the same number of entries." 
          << DbPrint::endmsg;
     return Error;
   }
 }
 return Success;
}
