// $Id: DbDefaultDataHandler.h,v 1.15 2007/11/20 18:10:03 frankb Exp $
//====================================================================
//  DbDefaultDataHandler: call to handle object manupulation at save/load time
//--------------------------------------------------------------------
//
//  Package    : StorageSvc  (The POOL project)
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBDEFAULTDATAHANDLER_H
#define POOL_DBDEFAULTDATAHANDLER_H 1

// Framework include files
#include "POOLCore/Token.h"
#include "StorageSvc/DataHandler.h"

/*
 *   POOL namespace declaration
 */
namespace pool {
  /// Forward declarations
  class Reference;
  class DbDatabase;
  class IStorageSvc;
  class IStorageSvcEx;

  /** @class DbDefaultDataHandler DbDefaultDataHandler.h StorageSvc/DbDefaultDataHandler.h
    *
    * Define default class for object handling during persistency operations.
    * This implementation only handles refrences in a way, that only reference
    * tokens are constrcuted on the fly.
    *
    * @author  Markus Frank 
    * @date    06/02/03
    * @version 1.0
    */
  class DbDefaultDataHandler : public DataHandler {
  private:
    /// Reference to current I/O context
    const DbDatabase*  m_ioContext;
  public:

    /// Standard constructor
    /** Handle object construction.
      * @return reference to constructed object
      */
    DbDefaultDataHandler();

    /// Standard destructor: release used resources
    virtual ~DbDefaultDataHandler();

    /// Quick identifier
    virtual unsigned int quickCast()  const          {  return 0xFEEDFEED; }

    /// Access the IO context
    const DbDatabase* ioContext() const              {  return m_ioContext;}

    /// Set the IO context
    virtual void setIOContext(const DbDatabase* dsc) {  m_ioContext = dsc; }

    /// Callback to handle on the fly object manipulation during save operations.
    /** @param data_type      [IN] DataType of the object to be processed
      * @param reflection     [IN] Reflection class information if availible
      * @param object_ptr     [IN] Pointer to the transient object.
      * @param data       [IN/OUT] Optional data for processing the request.
      *
      * @return DbStatus indicating success or failure.
      */
    virtual DbStatus onSave(int       data_type,      // data type
                            TypeH     reflection,     // Reflection information
                            void*     object_ptr,     // object ptr
                            void*     data);          // data pointer

    /// Callback to handle on the fly object manipulation during load operations.
    /** @param data_type      [IN] DataType of the object to be processed
      * @param reflection     [IN] Reflection class information if availible
      * @param object_ptr     [IN] Pointer to the transient object.
      * @param data       [IN/OUT] Optional data for processing the request.
      *
      * @return DbStatus indicating success or failure.
      */
    virtual DbStatus onLoad(int       data_type,      // data type
                            TypeH     reflection,     // Reflection information
                            void*     object_ptr,     // object ptr
                            void*     data);          // data pointer

    /// Internal call on saving references
    /** The default implementation is empty.
      * @param pRef           [IN] Pointer to reference object to be saved
      * @param pOid       [IN/OUT] Pointer to OID data for this reference
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus saveReference(Reference* pRef, Token::OID_t* pOid);

    /// Internal call on loading references
    /** The default implementation is empty.
      * @param pRef       [IN/OUT] Pointer to reference object to be loaded
      * @param pOid           [IN] Pointer to OID data for this reference
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus loadReference(Reference* pRef, Token::OID_t* pOid);

    /// Internal call on saving tokens
    /** The default implementation is empty.
      * @param pToken         [IN] Pointer to token object to be saved
      *                            The token OID must be valid on return!
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus saveToken(Token* pToken);

    /// Internal call on loading tokens
    /** The default implementation is empty.
      * @param pToken     [IN/OUT] Pointer to token object to be loaded
      *                            The token OID is valid!
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus loadToken(Token* pToken);

    /// Internal call on saving objects
    /** The default implementation is empty.
      * @param pObj           [IN] Pointer to object to be saved
      * @param data           [IN] Optional information to save the object.
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus saveObject(void* pObj, void* data);

    /// Internal call on loading objects
    /** The default implementation is empty.
      * @param pObj           [IN] Pointer to object to be loaded
      * @param data           [IN] Optional information to load the object.
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus loadObject(void* pObj, void* data);

    /// Internal call to create the token of a given reference
    /** The default implementation creates a token out of the oid.
      * @param pOid           [IN] Pointer to OID data for this reference
      * @param pToken        [OUT] Reference to pointer to store token
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus createToken(Token::OID_t* pOid, Token*& pToken);

    /// Internal call on saving the token of a given reference
    /** The default implementation creates the oid information out of the token.
      * @param pToken         [IN] Pointer to token to be saved
      * @param pOid       [IN/OUT] Pointer to OID data for this reference
      *
      * @return DbStatus code indicating success or failure
      */
    virtual DbStatus saveToken(const Token* pToken, Token::OID_t* pOid);
  };
}      // End namespace pool

#endif // POOL_DBDEFAULTDATAHANDLER_H
