// $Id: DbIter.h,v 1.17 2007/11/20 16:43:06 frankb Exp $
//====================================================================
//  DbIter definition
//--------------------------------------------------------------------
//
//  Package    : StorageSvc (The POOL project)
//
//  Description: Generic persistable object iterator
//
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBITER_H
#define POOL_DBITER_H

// Framework include files
#include "StorageSvc/DbObject.h"
#include "StorageSvc/DbContainer.h"
#include "StorageSvc/IDbCursor.h"
#include "StorageSvc/IOODatabase.h"
#include "StorageSvc/DataCallBack.h"

// STL include files
#include <memory>

/*
 *  POOL namespace declaration
 */
namespace pool  {

  /** @class DbIter<T> DbIter.h StorageSvc/DbIter.h
    *
    * Definition of the DbIter class (generic persistable object iterator).
    *
    * <P> Class:      DbIter
    * <P> Purpose:    
    * Object definition to access objects within a container.
    *
    * @author  Markus Frank                                               
    * @version 1                                                         
    */
  template <class T> class DbIter   {
    /// Handle to container object
    DbContainer       m_cnt;
    /// Handle to current object
    DbHandle<T>       m_obj;
    /// Iteration cursor
    IDbCursor*        m_cursor;
    /// Data access type (Default to READ)
    DbAccessMode      m_mode;

  private:
    /// Copy constructor
    DbIter(const DbIter& c) : m_obj(c.m_obj), m_cursor(c.m_cursor), 
      m_mode(c.m_mode)                        {                           }
    /// Equality operator
    bool operator==(const DbIter<T>& iter)   const 
    {    return m_obj.ptr() == iter.m_obj.ptr();                          }
    /// Assignment operator
    DbIter& operator=(const DbIter& copy)     {
      m_cnt     = copy.m_cnt;
      m_obj     = copy.m_obj;
      m_mode    = copy.m_mode;
      m_cursor  = copy.m_cursor;
      return *this;
    }
  public:
    /// Constructor with initializing arguments
    DbIter(DbAccessMode mode=pool::READ) : m_cursor(0),m_mode(mode) {       }
    /// Standard destructor
    virtual ~DbIter()                   {    releasePtr(m_cursor);    }
    /// Access to container handler
    const DbContainer& container() const{    return m_cnt;                    }
    /// Check validity
    bool isValid()  const               {    return m_obj.isValid();          }
    /// Dereference operator
    DbHandle<T>& operator->()           {    return m_obj;                    }
    /// Dereference operator
    DbHandle<T>& operator*()            {    return m_obj;                    }
    /// Access to current object handle
    DbHandle<T>& handle()               {    return m_obj;                    }
    /// Access to current object handle (CONST)
    const DbHandle<T>& handle() const   {    return m_obj;                    }
    /// Dereference operator
    T* object()  const                  {    return m_obj.ptr();              }
    /// Access to cursor token
    const Token* token() const {    return m_cursor ? m_cursor->token() : 0;  }
    /// Scan the container
    DbStatus scan(const DbContainer& cntH, const DbTypeInfo* typ);
    /// Retrieve next element
    DbStatus next();
  };

  /** @class DbPtrIter<T> DbIter.h StorageSvc/DbIter.h
    *
    * Description:
    * Definition of the DbPtrIter class (generic persistable object iterator).
    *  <P> Class:      DbPtrIter
    *  <P> Purpose:    
    * Object definition to access objects within a container.
    * The difference to the iterator above is, that objects do not have to be
    * allocated using the poolDb heap mechanism.
    * Hence they also do not support reference counting...
    *
    * @author  Markus Frank                                               
    * @version 1                                                         
    */
  template <class T> class DbPtrIter   {

  private:
    /// Handle to container object
    DbContainer       m_cnt;
    /// Handle to current object
    std::auto_ptr<T>  m_obj;
    /// Iteration cursor
    IDbCursor*        m_cursor;
    /// Data access type (Default to READ)
    DbAccessMode      m_mode;
    /// Data callback
    DataCallBack*     m_call;

  private:
    /// Copy constructor
    DbPtrIter(const DbPtrIter& c) : m_obj(0), m_cursor(c.m_cursor), 
      m_mode(c.m_mode)                        {                           }
    /// Equality operator
    bool operator==(const DbPtrIter<T>& iter)   const 
    {    return m_obj.get() == iter.m_obj.get();                          }
    /// Assignment operator
    DbPtrIter& operator=(const DbPtrIter& /* copy */)     {
      return *this;
    }
  protected:
    /// Constructor with initializing arguments
    DbPtrIter(DataCallBack* call, DbAccessMode mode=pool::READ) 
    : m_cursor(0), m_mode(mode), m_call(call) {                        }

  public:
    /// Constructor with initializing arguments
    DbPtrIter(const DataCallBack& call, DbAccessMode mode=pool::READ) 
    : m_cursor(0),m_mode(mode), m_call(call.clone()) {                        }
    /// Standard destructor
    virtual ~DbPtrIter()                {
      releasePtr(m_cursor);    
      deletePtr(m_call);    
    }
    /// Access to container handler
    const DbContainer& container() const{    return m_cnt;                    }
    /// Check validity
    bool isValid()  const               {    return m_obj.get() != 0;         }
    /// Dereference operator
    std::auto_ptr<T>& operator->()      {    return m_obj;                    }
    /// Dereference operator
    std::auto_ptr<T>& operator*()       {    return m_obj;                    }
    /// Access to current object handle
    std::auto_ptr<T>& handle()          {    return m_obj;                    }
    /// Access to current object handle (CONST)
    const std::auto_ptr<T>& handle() const{  return m_obj;                    }
    /// Dereference operator
    T* object()  const                  {    return m_obj.get();              }
    /// Access to user supplied call object
    DataCallBack* call() const          {    return m_call;                   }
    /// Access to cursor token
    const Token* token() const {    return m_cursor ? m_cursor->token() : 0;  }
    /// Scan the container
    DbStatus scan(const DbContainer& cntH, const DbTypeInfo* typ);
    /// Retrieve next element
    DbStatus next();
  };

  /// No void objects allowed on the user side!
  template <> class DbIter<DbObject>   {};

  /// No void objects allowed on the user side!
  template <> class DbPtrIter<DbObject>   {};

} // End namespace pool

/// Scan the container
template <class T> inline 
pool::DbStatus pool::DbIter<T>::scan(const DbContainer& cntH, 
                                     const DbTypeInfo* typ) 
{
  releasePtr(m_cursor);
  if ( cntH.isValid() )   {
    m_cursor = cntH.makeCursor(typ);
    m_cnt    = cntH;
  }
  m_obj = DbHandle<T>(cntH.type());
  return (m_cursor) ? Success : Error;
}

/// Retrieve next element of iteration
template <class T> inline pool::DbStatus pool::DbIter<T>::next()
{
  if ( m_cursor ) {
    DbObjectHandle<DbObject> objH(m_obj.ptr());
    objH._setType(m_obj.type());
    if ( m_cursor->next(m_cnt, objH, m_mode).isSuccess() ) {
      m_obj._setObject(static_cast<T*>(objH.ptr()));
      return Success;
    }
  }
  return Error;
}

/// Scan the container
template <class T> inline 
pool::DbStatus pool::DbPtrIter<T>::scan(const DbContainer& cntH, 
                                        const DbTypeInfo* typ) 
{
  releasePtr(m_cursor);
  if ( cntH.isValid() )   {
    m_cursor = cntH.makeCursor(typ);
    m_cnt    = cntH;
  }
  m_obj = std::auto_ptr<T>(0);
  return (m_cursor) ? Success : Error;
}

/// Retrieve next element
template <class T> inline pool::DbStatus pool::DbPtrIter<T>::next() {
  if ( m_cursor )  {
    DbStatus sc = m_cursor->next(m_cnt, m_call, m_mode);
    DbObject* pObj = const_cast<DbObject*>(m_call->object());
    m_obj = std::auto_ptr<T>(static_cast<T*>(pObj));
    return sc;
  }
  return Error;
}

#endif  // POOL_DBITER_H
