// $Id: DbObject.h,v 1.18 2007/11/20 16:43:06 frankb Exp $
//====================================================================
//  DbObject definition
//--------------------------------------------------------------------
//
//  Package    : (The POOL project)
//
//  Description: Generic persistable object
//
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBOBJECT_H
#define POOL_DBOBJECT_H 1

// Framework include files
#include "StorageSvc/DbObjectSwitch.h"
#include "StorageSvc/DbHandleBase.h"

/*
 * POOL namespace declaration
 */
namespace pool  {
  // forward declarations
  class DbObjectAccessor;
  class DataCallBack;
  class DbTypeInfo;
  class DbSession;
  class DbDomain;
  class DbDatabase;
  class DbContainer;
  class Token;
}

#ifdef POOL_HAVE_BASE

namespace pool  {
  /** @class DbObject DbObject.h StorageSvc/DbObject.h
    *
    * Definition of the DbObject class (generic persistable object).
    *
    * - Class:      Object
    * - Purpose:    
    *  Basic object for storage of identifiable persistent data.
    *  The DbObject is able to keep track of persistent links 
    *  between objects.
    *
    *
    * @author  Markus Frank                                               
    * @version 1.0
    */
  class DbObject    {
  protected:
    /// Persistent objects only live with reference counting!
    virtual ~DbObject();
  public:
    /// operator new for purely heap based objects
    void* operator new(size_t size);
    /// Need to overload operator delete after special instantiation
    void operator delete(void* ptr);
    /// operator new for the creation of objects using clustering hint
    void* operator new(size_t siz, DbContainer& cntH);
    /// Exception enabled compilation needs this delete operator
    void operator delete(void *ptr, DbContainer& cntH);
    /// operator new for the creation of objects using clustering hint
    void* operator new(size_t siz, DbContainer& cntH, DataCallBack* opt);
    /// Exception enabled compilation needs this delete operator
    void operator delete(void *ptr, DbContainer& cntH, DataCallBack* opt);
    /// Standard destructor
    DbObject();
  };
}
#else

/// operator new for the creation of objects using clustering hint
void* operator new(size_t siz, pool::DbContainer& cntH, const pool::DbTypeInfo* typ);
/// Exception enabled compilation needs this delete operator
void operator delete(void *ptr, pool::DbContainer& cntH, const pool::DbTypeInfo* typ);
/// operator new for the creation of objects using clustering hint
void* operator new(size_t siz, pool::DbContainer& cntH, pool::DataCallBack* opt);
/// Exception enabled compilation needs this delete operator
void operator delete(void *ptr, pool::DbContainer& cntH, pool::DataCallBack* opt);
#endif


namespace pool {

  /** @class DbObjectHandle DbObject.h StorageSvc/DbObject.h
    *
    * Description:
    * Base class for object handles.
    *
    * @author  M.Frank
    * @version 1.0
    */
  template <class USER> class DbObjectHandle : public DbHandleBase<USER> {
  private:
    typedef DbHandleBase<USER> Base;
    /// Pointer conversion
    const USER* makePtr(void* p)                {  return (const USER*)(p); }
  protected:
    /// Attach object to handle with type: may be accessed by sub-class only
    template<typename T> void _set(const T* ptr, const DbType& typ) {
      _setObject(ptr);
      _setType(typ);
    }
    /// Set object value with internal cast
    void _setWithCast(const DbObjectSwitch& c);
  public:
    /// Set handle type
    void _setType(const DbType& type) const    { Base::setType(type);       }
    /// Set object value
    template<class T> void _setObject(const T* p)  {
      if ( 0 == Base::ptr() && 0 == p ) return;
      _setWithCast(DbObjectSwitchImp<T,USER>(p, Base::m_ptr, Base::m_type));
    }
    void _setObject(const int /* null_obj */ )  { 
      if ( 0 == Base::ptr() ) return;
      _setWithCast(DbObjectSwitchImp<USER,USER>(0, Base::m_ptr, Base::m_type));
    }
  public:
    /// Standard destructor
    virtual ~DbObjectHandle()                 { _setObject(0);              }
    /// Standard Constructor
    DbObjectHandle()                          {                             }
    /// Constructor with storage type
    DbObjectHandle(const DbType& typ)         { _setType(typ);              }
    /// Constructor with object pointer
    DbObjectHandle(const USER* p)             { _setObject(p);              }
    /// Copy constructor
    DbObjectHandle(const DbObjectHandle<USER>& c) : DbHandleBase<USER>()
    { _set(c.ptr(),c.type());                                               }
    /// Generic assignment operator from base class
    operator USER*() const                        { return Base::ptr();     }
    /// Generic assignment operator
      DbObjectHandle<USER>& operator=(const int /* nuller */)
    { _setObject(0); return *this;                                          }
    /// Equality operator
    template <typename T> bool operator==(const DbHandleBase<T>& objH) const
    { return Base::type() == objH.type() && Base::ptr() == objH.ptr();      }
    /// Retrieve hosting container
    const DbContainer& containedIn() const;
    /// Access object oid
    DbLink& oid() const;
    /// Add persistent association entry
    DbStatus makeLink(const Token* pToken, DbLink& linkH) const;
    /// Validate an object association
    DbStatus getLink(const DbLink& lnkH, Token* p)  const;
    /// Open handle of object fully identified by its token
    DbStatus open(          const DbSession&    sessionH,
                            const std::string&  pfn,
                            const Token&        tokH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object fully identified by its token
    DbStatus open(          const DbDomain&     domH,
                            const std::string&  pfn,
                            const Token&        tokH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object originating from database dbH
    DbStatus open(          const DbDatabase&   dbH,
                            const Token&        tokenH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object originating from container cntH
    DbStatus open(          const DbContainer&  cntH,
                            const Token&        tokenH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object fully identified by its token
    static DbStatus openEx( const DbSession&   sessionH,
                            const std::string&  pfn,
                            const Token&        tokH, 
                            DataCallBack*       callH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object fully identified by its token
    static DbStatus openEx( const DbDomain&    domH,
                            const std::string&  pfn,
                            const Token&        tokH, 
                            DataCallBack*       callH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object originating from container cntH
    static DbStatus openEx( const DbDatabase&  dbH,
                            const Token&        tokenH, 
                            DataCallBack*       callH,
                            DbAccessMode        mod = pool::READ);
    /// Open handle of object originating from container cntH
    static DbStatus openEx( const DbContainer&  cntH,
                            const Token&        tokenH, 
                            DataCallBack*       callH,
                            DbAccessMode        mod = pool::READ);

    /// Context dependent open of object originating from database dbH
    DbStatus open(          const DbDatabase&   dbH,
                            const DbLink&       linkH,
                            DbAccessMode        mod = pool::READ);
    /// Context dependent open of object originating from container cntH
    DbStatus open(          const DbContainer&  cntH,
                            const DbLink&       linkH,
                            DbAccessMode        mod = pool::READ);
    /// Context dependent open of object originating from database dbH
    static DbStatus openEx( const DbDatabase&   dbH,
                            DataCallBack*       callH,
                            const DbLink&       linkH,
                            DbAccessMode        mod = pool::READ);
    /// Context dependent open of object originating from container cntH
    static DbStatus openEx( const DbContainer&  cntH,
                            DataCallBack*       callH,
                            const DbLink&       linkH,
                            DbAccessMode        mod = pool::READ);
  };

  template <class USER> class DbHandle : public DbObjectHandle< USER> {
    typedef DbObjectHandle< USER> Handle;
  public:
    /// Standard destructor
    virtual ~DbHandle()                           {                         }
    /// Default constructor
    DbHandle()                                    {                         }
    /// Constructor with storage type
    DbHandle(const DbType& typ)                   { Handle::_setType(typ);  }
    /// Constructor with object assignment
    DbHandle(const USER* obj)                     { Handle::_setObject(obj);}
    /// Copy constructor
    template <typename T> DbHandle(const DbObjectHandle<T>& c)
      : DbObjectHandle<USER>()                    
    { Handle::_set(c.ptr(),c.type());                                       }
    /// Copy constructor
    DbHandle(const DbHandle<USER>& c) : DbObjectHandle<USER>()
    { Handle::_set(c.ptr(),c.type());                                       }
    /// Generic assignment operator
    DbHandle<USER>& operator=(const USER* obj) {
      if ( Handle::ptr() != obj ) Handle::_setObject(obj);
      return *this;
    }
    /// Generic assignment operator
    template <typename T> 
    DbHandle<USER>& operator=(const DbObjectHandle<T>& c) {
      if ( Handle::ptr() != c.ptr() ) Handle::_set(c.ptr(), c.type());
      return *this;
    }
  };

  /// No void handles allowed: type must be polymorphic!
#ifndef POOL_HAVE_BASE
  template <> class DbHandle<DbObject>            {};
#endif
}

/*
 *
 *  Inline object handle implementation
 *
 */
#include "StorageSvc/DbObjectAccessor.h"

namespace pool {
#ifndef __no_inline


  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::open(const DbSession&    sessionH,
                                          const std::string&  pfn,
                                          const Token&        tokenH, 
                                          DbAccessMode        mod)
  { return DbObjectAccessor::open(this, 0, sessionH, pfn, tokenH, mod); }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::openEx(const DbSession&    sessionH,
                                            const std::string&  pfn,
                                            const Token&        tokenH,
                                            DataCallBack*       call, 
                                            DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, sessionH, pfn, tokenH, mod); }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::open(const DbDomain&     domH,
                                          const std::string&  pfn,
                                          const Token&        tokenH, 
                                          DbAccessMode        mod)
  { return DbObjectAccessor::open(this, 0, domH, pfn, tokenH, mod);     }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::openEx(const DbDomain&     domH,
                                            const std::string&  pfn,
                                            const Token&        tokenH,
                                            DataCallBack*       call, 
                                            DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, domH, pfn, tokenH, mod);     }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::open(const DbDatabase&   dbH,  
                                          const Token&        tokenH, 
                                          DbAccessMode        mod)
  { return DbObjectAccessor::open(this, 0, dbH, tokenH, mod);            }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::openEx(const DbDatabase&   dbH,  
                                            const Token&        tokenH,
                                            DataCallBack*       call, 
                                            DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, dbH, tokenH, mod);            }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::open(const DbContainer&  cntH,  
                                          const Token&        tokenH, 
                                          DbAccessMode        mod)
  { return DbObjectAccessor::open(this, 0,   cntH, tokenH, mod);         }

  /// Open handle
  template <> inline
  DbStatus DbObjectHandle<DbObject>::openEx(const DbContainer&  cntH,  
                                            const Token&        tokenH,
                                            DataCallBack*       call, 
                                            DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, cntH, tokenH, mod);           }

  // Context dependent open of object originating from database dbH
  template <> inline
  DbStatus DbObjectHandle<DbObject>::open(const DbDatabase&   dbH,
                                          const DbLink&       linkH,
                                          DbAccessMode        mod)
  { return DbObjectAccessor::open(this, 0, dbH, linkH, mod);             }

  // Context dependent open of object originating from container cntH
  template <> inline
  DbStatus DbObjectHandle<DbObject>::open(const DbContainer&  cntH,
                                          const DbLink&       linkH,
                                          DbAccessMode        mod)
  { return DbObjectAccessor::open(this, 0, cntH, linkH, mod);            }

  // Context dependent open of object originating from database dbH
  template <> inline
  DbStatus DbObjectHandle<DbObject>::openEx(const DbDatabase&  dbH,
                                            DataCallBack*      call, 
                                            const DbLink&      linkH,
                                            DbAccessMode       mod)
  { return DbObjectAccessor::open(0, call, dbH, linkH, mod);             }

  // Context dependent open of object originating from container cntH
  template <> inline
  DbStatus DbObjectHandle<DbObject>::openEx(const DbContainer& cntH,
                                            DataCallBack*      call, 
                                            const DbLink&      linkH,
                                            DbAccessMode       mod)
  { return DbObjectAccessor::open(0, call, cntH, linkH, mod);            }

  /// Attach object to handle
  template <class T> inline
  void DbObjectHandle<T>::_setWithCast(const DbObjectSwitch& cast) {
    DbObjectAccessor::switchPointer(cast);
  }

  /// Retrieve hosting container
  template <class T> inline
  const DbContainer& DbObjectHandle<T>::containedIn() const  
  { return DbObjectAccessor::containedIn(Base::ptr());                   }

  /// Access object oid
  template <class T> inline
  DbLink& DbObjectHandle<T>::oid() const  
  { return DbObjectAccessor::objectOid(Base::ptr());                     }

  /// Add persistent association entry
  template <class T> inline
  DbStatus DbObjectHandle<T>::makeLink(const Token* pToken, DbLink& linkH) const 
  { return DbObjectAccessor::makeObjectLink(Base::ptr(), pToken, linkH); }

  /// Validate an object association
  template <class T> inline
  DbStatus DbObjectHandle<T>::getLink(const DbLink& linkH, Token* pToken) const 
  { return DbObjectAccessor::getObjectLink(Base::ptr(), linkH, pToken);  }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::openEx( const DbSession&    sessionH,
                                      const std::string&  pfn,
                                      const Token&        tokenH,
                                      DataCallBack*       call, 
                                      DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, sessionH, pfn, tokenH, mod);  }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::open(const DbDomain&      domH,
                                   const std::string&   pfn,
                                   const Token&         tokenH, 
                                   DbAccessMode         mod) 
  {
    DbObjectHandle<DbObject> objH(Base::type());
    DbStatus status = objH.open(domH, pfn, tokenH, mod);
    _setObject(objH.ptr());
    return status;
  }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::openEx( const DbDomain&     domH,
                                      const std::string&  pfn,
                                      const Token&        tokenH,
                                      DataCallBack*       call, 
                                      DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, domH, pfn, tokenH, mod);      }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::open(  const DbDatabase&  dbH, 
                                   const Token&       tokenH,
                                   DbAccessMode       mod) 
  {
    DbObjectHandle<DbObject> objH(Base::type());
    DbStatus status = objH.open(dbH, tokenH, mod);
    _setObject(makePtr(objH.ptr()));
    return status;
  }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::openEx( const DbDatabase& dbH,
                                      const Token&      tokenH,
                                      DataCallBack*     call, 
                                      DbAccessMode      mod)
  { return DbObjectAccessor::open(0, call, dbH, tokenH, mod);            }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::open( const DbContainer&   cntH, 
                                    const Token&         tokenH, 
                                    DbAccessMode         mod) 
  {
    DbObjectHandle<DbObject> objH(Base::type());
    DbStatus status = objH.open(cntH,tokenH,mod);
    _setObject(objH.ptr());
    return status;
  }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::openEx(const DbContainer& cntH,
                                     const Token&       tokenH,
                                     DataCallBack*      call, 
                                     DbAccessMode       mod)
  { return DbObjectAccessor::open(0, call, cntH, tokenH, mod);          }

  /// Context dependent open of object originating from database dbH
  template <class T> inline
  DbStatus DbObjectHandle<T>::open(      const DbDatabase&   dbH,
                                         const DbLink&       linkH,
                                         DbAccessMode        mod)
  {
    DbObjectHandle<DbObject> objH(Base::type());
    DbStatus status = objH.open(dbH, linkH, mod);
    _setObject(makePtr(objH.ptr()));
    return status;
  }

  /// Open handle
  template <class T> inline
  DbStatus DbObjectHandle<T>::open( const DbSession&    sessionH,
                                    const std::string&  pfn,
                                    const Token&        tokenH, 
                                    DbAccessMode        mod) 
  {
    DbObjectHandle<DbObject> objH(Base::type());
    DbStatus status = objH.open(sessionH, pfn, tokenH, mod);
    _setObject(objH.ptr());
    return status;
  }

  // Context dependent open of object originating from container cntH
  template <class T> inline
  DbStatus DbObjectHandle<T>::open(      const DbContainer&  cntH,
                                         const DbLink&       linkH,
                                         DbAccessMode        mod)
  {
    DbObjectHandle<DbObject> objH(Base::type());
    DbStatus status = objH.open(cntH, linkH, mod);
    _setObject(makePtr(objH.ptr()));
    return status;
  }

  // Context dependent open of object originating from database dbH
  template <class T> inline
  DbStatus DbObjectHandle<T>::openEx(    const DbDatabase&   dbH,
                                         DataCallBack*       call, 
                                         const DbLink&       linkH,
                                         DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, dbH, linkH, mod);             }

  // Context dependent open of object originating from container cntH
  template <class T> inline
  DbStatus DbObjectHandle<T>::openEx(    const DbContainer&  cntH,
                                         DataCallBack*       call, 
                                         const DbLink&       linkH,
                                         DbAccessMode        mod)
  { return DbObjectAccessor::open(0, call, cntH, linkH, mod);            }

#endif
}       // End namespace pool
#endif  // POOL_DBOBJECT_H
