// $Id: DbReflex.h,v 1.14 2007/11/21 12:37:38 frankb Exp $
//====================================================================
//  DbDatabase and related class definitions
//--------------------------------------------------------------------
//
//  Package    : StorageSvc  (The POOL project)
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBREFLEX_H
#define POOL_DBREFLEX_H

// C++ include files
#include <string>
#include <vector>
#include <typeinfo>
#include <set>

// Framework include files
#include "Reflex/Type.h"
#include "POOLCore/Guid.h"

/*
 *  POOL namespace declaration
 */
namespace pool  {

  /** @class DbReflex DbReflex.h StorageSvc/DbReflex.h
    *
    * Description:
    * Handle transparent loading of reflection classes.
    *
    * @author  M.Frank
    * @version 1.0
    */
  class DbReflex  {
  private:
    /// Reflex class identifier
    typedef const ROOT::Reflex::Type TypeH;
    /// No public construction
    DbReflex() {}
    /// No public destruction
    ~DbReflex() {}
  public:
    /// Access classes by class name
    /** @param name     [IN]    type name
      *
      * @return  Pointer to reflection class if present.
      */
    static TypeH forTypeName(const std::string& name);

    /// Access classes by RTTI type information
    /** @param info     [IN]    referenece to RTTI typeid
      *
      * @return  Pointer to reflection class if present.
      */
    static TypeH forTypeInfo(const std::type_info& info);

    /// Access classes by Guid
    /** @param info     [IN]    Reference to Guid
      *
      * @return  Pointer to reflection class if present.
      */
    static TypeH forGuid(const Guid& info);

    /// Access full class name with scope
    /** @param type     [IN]    Reflection type 
      *
      * @return  String containing full scoped name
      */
    static std::string fullTypeName(const TypeH& type);

    /// Convert Guid to nomalized string (all char upper case!)
    /** @param id       [IN]    Guid reference
      *
      * @return  String containing full scoped name
      */
    static std::string guidString(const Guid& id);

    /// Determine Guid (normalized string form) from reflection type.
    /** @param type     [IN]    Reflection type 
      *
      * @return  String containing full scoped name
      */
    static Guid guid(const TypeH& id);

    /// Determine Guid (normalized string form) from reflection type.
    /** @param type     [IN]    Reflection type 
      *
      * @return  String containing full scoped name
      */
    static std::string guidString(const TypeH& id);

    /// Dump all known dictionaries
    static void dumpDictionaries();

    //--------------- Automatic dictionary loading:

    /// Enable automatic loading of dictionaries
    /** @param autoload  [IN]  true=enable autoloading, false=disable
    */
    static void setDictionaryAutoLoading( bool autoload );

    /// Allow to autoload this dictionary
    /** @param moduleName  [IN] name of the dictionary module that can be autoloaded
  (do not add "lib" and ".so" to the name) 
    */
    static void allowAutoLoadingFor(const std::string& moduleName);

    /// Load dictionary module (seal plugin module, shared library) that defines class "className"
    /// The name of the class is checked against seal module capabilities list
    /** @param className    [IN]  Class name which definition should be loaded from LCG dictionary.
      * @return   DbStatus code indicating success or failure.
      */
    static ROOT::Reflex::Type loadDictionaryForClass(const std::string &className);

 protected:
  };
}       // End namespace pool
#endif  // POOL_DBREFLEX_H
