// $Id: DbContainerImp.cpp,v 1.34 2007/11/20 18:10:05 frankb Exp $
//====================================================================
//
//  Package    : StorageSvc (The POOL project)
//
//  @author      M.Frank
//
//====================================================================

/// Framework include files
#include "StorageSvc/DbPrint.h"
#include "POOLCore/Token.h"
#include "StorageSvc/DbHeap.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbCallBack.h"
#include "StorageSvc/DbCursorImp.h"
#include "StorageSvc/DbTransaction.h"
#include "StorageSvc/DbObjectGuard.h"
#include "StorageSvc/DbInstanceCount.h"
#include "StorageSvc/DbContainerImp.h"
using namespace pool;

static const size_t GUARDSIZE = DbHeap::guardSize();
inline static DbObjectGuard* GUARD(const void* ptr)  {
  return (DbObjectGuard*) (((char*)ptr)-GUARDSIZE);
}

/// Standard Constructor
DbContainerImp::DbContainerImp(IOODatabase* idb)
: DbImplementation(idb), m_size(0), m_writeSize(0), m_name("UNKNOWN"),
  m_canUpdate(false),
  m_canDestroy(false)
{
  DbInstanceCount::increment(this);
  m_stackType = NONE;
}

/// Standard Destructor
DbContainerImp::~DbContainerImp() {
  DbInstanceCount::decrement(this);
  m_stack.clear();
}

/// Size of the container
long long int DbContainerImp::size()  {
  return m_writeSize;
}

/// Number of next record in the container (=size if no delete is allowed)
long long int DbContainerImp::nextRecordId()   {
  return size();
}

/// Access options
DbStatus DbContainerImp::getOption(DbOption& /* opt */) const { 
  return Error;  
}

/// Set options
DbStatus DbContainerImp::setOption(const DbOption& /* opt */){ 
  return Success;
}

/// Create iterator cursor to loop over container elements
IDbCursor* DbContainerImp::makeCursor(const DbTypeInfo* typ)  {
  return new DbCursorImp(typ);
}

/// Close the container and deallocate resources
DbStatus DbContainerImp::close()   {
  return clearStack();
}

/// In place allocation of raw memory for the transient object
void* DbContainerImp::allocate(unsigned long siz,DbContainer& cntH,DataCallBack* call)  {
  if ( call )  {
    DbObjectHandle<DbObject> objH(cntH.type());
    DbLink objLink(cntH.token()->oid().first, int(nextRecordId()));
    DbHeap::allocate(siz, &cntH, &objLink, &objH);
    if ( objH.ptr() )  {
      DbObjectGuard* g = GUARD(objH.ptr());
      g->addRef();
    }
    if ( m_stack.size() < m_size+1 )  {
      m_stack.resize(m_size+1024);
    }
    (*(m_stack.begin()+m_size)).assign(objH.ptr(), call, objLink, pool::WRITE);
    m_stackType |= pool::WRITE;
    m_writeSize++;
    m_size++;
    return objH.ptr();
  }
  throw std::bad_alloc();
}

/// In place allocation of raw memory for the transient object
DbStatus DbContainerImp::allocate(DbContainer& cntH,DataCallBack* call,DbLink& oid) {
  if ( call )  {
    if ( m_stack.size() < m_size+1 )  {
      m_stack.resize(m_size+1024);
    }
    oid.first  = cntH.token()->oid().first;
    oid.second = int(nextRecordId());
    (*(m_stack.begin()+m_size)).assign(0, call, oid, WRITE);
    m_stackType |= pool::WRITE;
    m_writeSize++;
    m_size++;
    return Success;
  }
  throw std::bad_alloc();
}

/// In place deletion of raw memory
DbStatus DbContainerImp::free(void* ptr,DbContainer& cntH,DataCallBack* /* opt_call */) {
  return DbHeap::free(ptr, &cntH);
}

/// Clear Transaction stack
DbStatus DbContainerImp::clearStack()   {
  static DbLink void_link(INVALID, INVALID);
  for(size_t i=0; i < m_size; ++i )  {
    _Transaction& t = *(m_stack.begin()+i);
    if ( t.objH )  {
      DbObjectGuard* g = GUARD(t.objH);
      DbObjectHandle<DbObject> objH(g->type());
      objH = t.objH;
      g->release();
    }
    releasePtr( t.call );
    t.assign(0, 0, void_link, NONE);
  }
  m_size = 0;
  m_writeSize = 0;
  m_stackType = NONE;
  return Success;
}

/// Execute object modification requests during a transaction
DbStatus DbContainerImp::commitTransaction(DbTransaction& refTr ) {
  DbStatus iret   = Success;
  DbStatus status = Success;
  TransactionStack::iterator i = m_stack.begin();
  for(size_t j=0; j < m_size; ++j, ++i )  {
    switch( (*i).action )  {
      case pool::DESTROY:
        status = destroyObject(*i);
        break;
      case pool::UPDATE:
        status = updateObject(*i);
        break;
      case pool::WRITE:
        status = writeObject(*i);
        break;
      default:
        status = Error;
        break;
    }
    if ( !status.isSuccess() ) {
      iret = status;
      DbPrint log( m_name);
      log << DbPrint::ERROR << "The Transaction cannot be committed..." 
          << " Container has " << size() << " Entries in total."
          << DbPrint::endmsg;
      break;
    }
  }
  clearStack();
  status = endTransaction(refTr);
  if ( !status.isSuccess() )  {
    iret = status;
  }
  return iret;
}

/// Execute object modification requests during a transaction
DbStatus DbContainerImp::endTransaction(DbTransaction& /* refTr */ ) {
  return Success;
}

/// Start/Commit/Rollback Database Transaction
DbStatus DbContainerImp::transAct(DbTransaction& refTr)   {
  DbStatus status = Success;
  switch(refTr.state()) {
  case Transaction::TRANSACT_COMMIT:
    status = commitTransaction(refTr);
    refTr.set(Transaction::TRANSACT_ENDED);
    break;
  case Transaction::TRANSACT_FLUSH:
    status = commitTransaction(refTr);
    refTr.set(Transaction::TRANSACT_ENDED);
    break;
  case Transaction::TRANSACT_START:
    refTr.set(Transaction::TRANSACT_ACTIVE);
    clearStack();
    break;
  case Transaction::TRANSACT_ROLLBACK:
    clearStack();
    refTr.set(Transaction::TRANSACT_ENDED);
    break;
  default:
    status = Error;
    break;
  }
  return status;
}

DbStatus 
DbContainerImp::save(const DbObjectHandle<DbObject>& objH)  {
  // Can only be done if no Transaction is ongoing...
  // i.e. exactly one object was allocated
  TransactionStack::iterator it = m_stack.begin();
  if ( m_writeSize == 1 )   {
    TransactionStack::value_type& entry = *it;
    if ( entry.objH == objH.ptr() )   {
      DbObjectGuard* g = GUARD(objH.ptr());
      DbObjectHandle<DbObject> oH(entry.objH);
      objH.oid() = oH.oid();
      g->addRef();
      DbStatus status = writeObject(entry);
      clearStack();
      return status;
    }
  }
  return Error;
}

DbStatus DbContainerImp::save(DbContainer&  /* cntH */,DataCallBack* call,DbLink& linkH) {
  // Only possible if no open transaction, i.e. No object was allocated
  if ( m_stack.empty() )  {
    _Transaction ent(0, call, linkH, WRITE);
    return writeObject(ent);
  }
  return Error;
}

// Update existing object in the container
DbStatus 
DbContainerImp::update(DbContainer& /* cntH */,DataCallBack* call,const DbObjectHandle<DbObject>& objH) {
  if ( m_canUpdate )  {
    if ( call )  {
      if ( objH.ptr() )  {
        DbObjectGuard* g = GUARD(objH.ptr());
        g->addRef();
      }
      if ( m_stack.size() < m_size+1 )  {
        m_stack.resize(m_size+1024);
      }
      (*(m_stack.begin()+m_size)).assign(objH.ptr(), call, objH.oid(), pool::UPDATE);
      m_stackType |= pool::UPDATE;
      m_size++;
      return Success;
    }
    return Error;
  }
  DbPrint log( m_name );
  log << DbPrint::ERROR << "The chosen implementation does not allow to "
      << "update existing objects."
      << "The requested action is not supported." << DbPrint::endmsg;
  return Error;
}

// Update existing object in the container
DbStatus DbContainerImp::update(DbContainer& /* cntH */,DataCallBack* call,const DbLink& linkH) {
  if ( m_canUpdate )  {
    if ( call )  {
      if ( m_stack.size() < m_size+1 )  {
        m_stack.resize(m_size+1024);
      }
      (*(m_stack.begin()+m_size)).assign(0, call, linkH, pool::UPDATE);
      m_stackType |= pool::UPDATE;
      m_size++;
      return Success;
    }
    return Error;
  }
  DbPrint log( m_name );
  log << DbPrint::ERROR << "The chosen implementation does not allow to "
      << "update existing objects."
      << "The requested action is not supported." << DbPrint::endmsg;
  return Error;
}

// Fetch next object address of the selection to set token
DbStatus DbContainerImp::fetch(DbSelect& sel) {
  if ( sel.criteria().length() == 0 || sel.criteria() == "*" )  {
    DbLink lnk = sel.link();
    while ( lnk.second < size() ) {
      if ( fetch(lnk, lnk).isSuccess() )  {
        sel.link() = lnk;
        return Success;
      }
      lnk.second++;
    }
    return Error;
  }
  DbPrint log( m_name );
  log << DbPrint::ERROR << "The chosen implementation does not allow to "
      << "refine container scans."
      << "The only valid selection criterium is: \"\" (empty string)"
      << DbPrint::endmsg;
  return Error;
} 

/// Destroy persistent object in the container; does not touch transient!
DbStatus DbContainerImp::destroy(const DbLink& linkH) {
  if ( m_canDestroy )  {
    if ( m_stack.size() < m_size+1 )  {
      m_stack.resize(m_size+1024);
    }
    (*(m_stack.begin()+m_size)).assign(0, 0, linkH, DESTROY);
    m_stackType |= DESTROY;
    m_size++;
    return Success;
  }
  DbPrint log( m_name );
  log << DbPrint::ERROR << "The chosen implementation does not allow to "
      << "destroy existing objects from the container."
      << "The requested action is not supported." << DbPrint::endmsg;
  return Error;
}

// Fetch refined object address. Default implementation returns identity
DbStatus DbContainerImp::fetch(const DbLink& linkH, DbLink& stmt)  {
  stmt.second = linkH.second;
  return linkH.second >= 0 && linkH.second < size() ? Success : Error;
}

// Interface Implementation: Find entry in container
DbStatus DbContainerImp::load(DataCallBack* call,
                              const DbLink& linkH,
                              DbLink&       oid,
                              DbAccessMode  mode,
                              bool          any_next)
{
  DbStatus sc = Error;
  oid.second = linkH.second;
  while ( oid.second < size() ) {
    sc = fetch(linkH, oid);
    if ( sc.isSuccess() )  {
      sc = loadObject(call, oid, mode);
      if ( sc.isSuccess() )  {
        return sc;
      }
    }
    if ( !any_next ) return Error;
    oid.second++;
  }
  if ( linkH.second < 0 || linkH.second <= size() ) {
    DbPrint log( m_name );
    log << DbPrint::DEBUG << "No objects passing selection criteria..." 
        << " Container has " << size() << " Entries in total." << DbPrint::endmsg;
  }
  return sc;
}

/// Find object by object identifier and load it into memory
DbStatus DbContainerImp::loadObject(DataCallBack* /* call */,
                                    DbLink&       /* oid  */,
                                    DbAccessMode  /* mode */)
{
  return Error;
}

