// $Id: DbInstanceCount.cpp,v 1.22 2008/01/11 13:17:14 frankb Exp $
//====================================================================
//  DbInstanceCount implementation
//--------------------------------------------------------------------
//
//  Package    : System (The POOL project)
//
//  Description: Allow for internal object tracing
//
//  @author      M.Frank
//====================================================================

#include "StorageSvc/DbInstanceCount.h"
#include "StorageSvc/DbObjectSet.h"
#include "StorageSvc/DbTypeInfo.h"
#include <iostream>
#include <iomanip>
#include <memory>
#include <cstdio>
#include <map>
namespace pool  {
  const std::string typeName(const std::type_info& typ);
  int numTokenInstances();
  void delete_DbObjectSetBase(void*) {  
  }
}
using namespace pool;

// Do not clutter global namespace
namespace {
  typedef DbInstanceCount::Counter                COUNT;
  typedef std::map<const std::type_info*, COUNT*> TypeCounter;
  typedef std::map<const std::string*,    COUNT*> StringCounter;
  typedef DbObjectSet<DbObjectSetBase>            ContCounter;
  static bool s_trace_instances = (0 != ::getenv("POOL_TRACE")) || (0 != ::getenv("POOLDB_TRACE"));

  ContCounter& contCounter() {
    static ContCounter       s_cntCounts(delete_DbObjectSetBase);
    return s_cntCounts;
  }
  static std::auto_ptr<TypeCounter>       s_typCounts(new TypeCounter());
  static std::auto_ptr<StringCounter>     s_strCounts(new StringCounter());
  static DbInstanceCount::Counter s_nullCount;
  static DbInstanceCount::Counter s_thisCount;
  static DbInstanceCount          s_counter;
  inline TypeCounter& types()      {    return *(s_typCounts.get());  }
  inline StringCounter& strings()  {    return *(s_strCounts.get());  }
}

DbObjectSetBase::DbObjectSetBase(dtor_t dt, const std::type_info& t) 
: m_destructor(dt), m_type(t)  {
  if ( m_destructor != delete_DbObjectSetBase ) {
    contCounter().add(this);
  }
}

DbObjectSetBase::~DbObjectSetBase()   {    
  std::set<void*>::iterator i=m_objects.begin();
  printOut();
  for( ; i != m_objects.end(); i = m_objects.begin() )  {
    void* obj = *i;
    m_objects.erase(i);
    if ( m_destructor )  (*m_destructor)(obj);
  }
  if ( m_destructor != delete_DbObjectSetBase ) {
    contCounter().remove(this);
  }
}

void DbObjectSetBase::printOut()  {
  if ( s_trace_instances && m_type != typeid(DbObjectSetBase) )  {
    std::cout << "Removal on unload of " << m_objects.size() << " instances of type:" 
              << typeName(m_type) << std::endl;       
  }
}

void DbObjectSetBase::removeObject(void* pObj)  {
  std::set<void*>::iterator i=m_objects.find(pObj);
  if ( i != m_objects.end() )  {
    m_objects.erase(i);
  }
}

/// Standard Constructor
DbInstanceCount::DbInstanceCount() {
  s_thisCount.increment();
}

/// Standard destructor
DbInstanceCount::~DbInstanceCount() {
  s_thisCount.decrement();
  if ( 0 == s_thisCount.value() )   {
    StringCounter::iterator i;
    TypeCounter::iterator   j;
    dump(s_trace_instances ? ALL : NONE);
    for(i=s_strCounts->begin(); i != s_strCounts->end(); ++i)
      delete (*i).second;
    for(j=s_typCounts->begin(); j != s_typCounts->end(); ++j)
      delete (*j).second;
    s_strCounts->clear();
    s_typCounts->clear();
  }
}

/// Check if tracing is enabled.
bool DbInstanceCount::doTrace()    {
  return s_trace_instances;
}

/// Enable/Disable tracing
void DbInstanceCount::doTracing(bool value)    {
  s_trace_instances = value;
}

/// Access counter object for local caching on optimizations
DbInstanceCount::Counter* DbInstanceCount::getCounter(const std::type_info& typ)  {
  Counter* cnt = s_trace_instances ? types()[&typ] : &s_nullCount;
  return (0!=cnt) ? cnt : types()[&typ] = new Counter();
}

/// Access counter object for local caching on optimizations
DbInstanceCount::Counter* DbInstanceCount::getCounter(const std::string& typ)  {
  Counter* cnt = s_trace_instances ? strings()[&typ] : &s_nullCount;
  return (0!=cnt) ? cnt : strings()[&typ] = new Counter();
}

/// Force dump of counters
void DbInstanceCount::dump(int typ)  {
  bool need_footer = false;
  if ( (typ & STRING) && s_strCounts.get() )    {
    if ( s_strCounts->begin() != s_strCounts->end() )   {
      StringCounter::const_iterator i;
      std::cout << "+--------------------------------------------------------------+" << std::endl;
      std::cout << "| I n s t a n c e   c o u n t e r s   b y    N A M E           |" << std::endl;
      std::cout << "+----------+-------+-------------------------------------------+" << std::endl;
      std::cout << "|   Total  | Diff  |      Type identifier                      |" << std::endl;
      std::cout << "+----------+-------+-------------------------------------------+" << std::endl;
      for(i=s_strCounts->begin(); i != s_strCounts->end(); ++i) {
        std::cout << "|" << std::setw(10) << (*i).second->total()
                  << "|" << std::setw(7)  << (*i).second->value()
                  << "|" << (*i).first << std::endl;       
      }
      need_footer = true;
    }
  }
  if ( (typ & TYPEINFO) && s_typCounts.get() )   {
    if ( s_typCounts->begin() != s_typCounts->end() )  {
      TypeCounter::const_iterator i;
      std::cout << "+--------------------------------------------------------------+" << std::endl;
      std::cout << "| I n s t a n c e   c o u n t e r s   b y    T Y P E I N F O   |" << std::endl;
      std::cout << "+----------+-------+-------------------------------------------+" << std::endl;
      std::cout << "|   Total  | Diff  |      Type identifier                      |" << std::endl;
      std::cout << "+----------+-------+-------------------------------------------+" << std::endl;
      for(i=s_typCounts->begin(); i != s_typCounts->end(); ++i) {
        std::cout << "|" << std::setw(10) << (*i).second->total()
                  << "|" << std::setw(7)  << (*i).second->value()
                  << "|" << typeName(*((*i).first)) << std::endl;       
      }
      need_footer = true;
    } 
  }
  if ( need_footer )  {
    std::cout << "|" << std::setw(10) << "<Unknown>"
              << "|" << std::setw(7)  << numTokenInstances()
              << "|" << "Token" << std::endl;       
    std::cout << "+----------+-------+-------------------------------------------+" << std::endl;
  }
}
