// $Id: DbObject.cpp,v 1.19 2007/11/20 16:43:09 frankb Exp $
//====================================================================
//  DbObject implementation
//--------------------------------------------------------------------
//
//  Package    : System (The POOL project)
//
//  Description: Generic data persistency
//
//  @author      M.Frank
//====================================================================

// Framework include files
#include "POOLCore/Token.h"
#include "StorageSvc/DbHeap.h"
#include "StorageSvc/DbObject.h"
#include "StorageSvc/DbSession.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbContainer.h"
#include "StorageSvc/DbObjectGuard.h"
#include "StorageSvc/DbInstanceCount.h"
#include <stdexcept>

using namespace pool;

static const size_t GUARDSIZE = DbHeap::guardSize();
inline static DbObjectGuard* GUARD(const void* ptr)  {
  return (DbObjectGuard*) (((char*)ptr)-GUARDSIZE);
}

#ifdef POOL_HAVE_BASE
#define HEAP(x) DbObject::x
namespace pool {
  void* DbObject::operator new(size_t size)   // New operator
  { return DbHeap::allocate(size, 0, 0, 0);         }
  void DbObject::operator delete(void* ptr)   // delete operator
  { DbHeap::free(ptr, 0);                           }
  DbObject::DbObject()   {  DbInstanceCount::increment(this);  }
  DbObject::~DbObject()  {  DbInstanceCount::decrement(this);  }
}
#else
#define HEAP(x) x
#endif

/// operator new for the creation of objects using clustering hint
void* HEAP(operator new)(size_t size, 
                         DbContainer& cntH,
                         const DbTypeInfo* typ) 
{ 
  return cntH.allocate(size, typ);
}

/// C++ exception enabled compilation needs this delete operator
void HEAP(operator delete)(void *ptr, 
                           DbContainer& cntH,
                           const DbTypeInfo* typ) 
{
  cntH.free(ptr, typ);
}

/// operator new for the creation of objects using clustering hint
void* HEAP(operator new)(size_t size, 
                         DbContainer& cntH,
                         DataCallBack* call)
{
  return cntH.allocate(size, call);
}

/// C++ exception enabled compilation needs this delete operator
void HEAP(operator delete)(void *ptr, 
                           DbContainer& cntH, 
                           DataCallBack* call)
{
  cntH.free(ptr, call);
}

/// Helper for handles
void DbObjectAccessor::switchPointer(const DbObjectSwitch& op)  {
  DbObjectSwitch::Data& data = const_cast<DbObjectSwitch::Data&>(op.m_data);
  void* ptr = *data.to;
  try     {
    *data.to = op.cast();
    if ( *data.to ) {
      DbObjectGuard* g = GUARD(*data.to);
      *data.typ = g->type();
      if ( !g->freeHandler() )  {
        g->installFreeHandler(op.destructor());
      }
      g->addRef();
    }
    if ( ptr )      {
      DbObjectGuard* g = GUARD(ptr);
      int cnt = g->release();
      if ( cnt <= 0 )   {
        DbObjectSwitch::Dtor call = g->freeHandler();
        if ( call )   {
          call(ptr);
          DbHeap::free(ptr, 0);
        }
        else    {
          DbHeap::free(ptr, 0);
//          unexpectedProblem(Error, 
//            "Missing object destructor: Severe memory problem.");
        }
      }
    }
  }
  catch(std::exception& e) {
    std::string msg = "StorageSvc> Tried to assign bad DbObject Pointer: Illegal assignment.\n";
    msg            += "StorageSvc> ";
    msg            += e.what();
    data.from = 0;
    switchPointer(op); 
    throw std::runtime_error(msg);
  }
  catch(...)    {
    const char* msg = "StorageSvc> Tried to assign bad DbObject Pointer: Illegal assignment!";
    badAssignMessage(msg);
    data.from = 0;
    switchPointer(op);
    throw std::runtime_error(msg);
  }
}

/// Access to hosting object
const DbContainer& DbObjectAccessor::containedIn(const DbObject* pObj)  {
  return DbHeap::container(pObj);
}


/// Add a link entry to the Database
DbStatus DbObjectAccessor::makeObjectLink(const DbObject* pObj,
                                          const Token* pTok, 
                                          DbLink& lnkH)
{
  if ( 0 != pTok )   {
    const DbDatabase& dbH = DbHeap::container(pObj).containedIn();
    return dbH.makeLink(pTok, lnkH);  
  }
  return Error;
}

/// Validate an association of an object
DbStatus DbObjectAccessor::getObjectLink( const DbObject* pObj,
                                          const DbLink& lnkH,
                                          Token* pTok)
{
  return DbHeap::container(pObj).containedIn().getLink(lnkH, pTok);
}

/// Validate an association of an object
DbLink& DbObjectAccessor::objectOid(const DbObject* pObj) {
  return GUARD(pObj)->oid();
}

/// Open handle
DbStatus DbObjectAccessor::open(DbObjectHandle<DbObject>* objH,
                                DataCallBack*             callH,
                                const DbSession&          sesH,
                                const std::string&        pfn,
                                const Token&              tokenH, 
                                DbAccessMode              mode)
{
  DbDomain domH(tokenH.technology());
  if ( domH.open(sesH, tokenH.technology(), mode).isSuccess() )   {
    return open(objH, callH, domH, pfn, tokenH, mode);
  }
  return Error;
}

/// Open handle
DbStatus DbObjectAccessor::open(DbObjectHandle<DbObject>* objH,
                                DataCallBack*             callH,
                                const DbDomain&           domH, 
                                const std::string&        pfn,
                                const Token&              tokenH, 
                                DbAccessMode              mode)
{
  if ( domH.type() == DbType(tokenH.technology()) ) {
    DbDatabase dbH(domH.type());
    const std::string& fn = (pfn.length()==0) ? tokenH.dbID() : pfn;
    if ( dbH.open(domH, fn, tokenH.dbID(), mode).isSuccess() )   {
      return open(objH, callH, dbH, tokenH, mode);
    }
  }
  return Error;
}

/// Open handle
DbStatus DbObjectAccessor::open(DbObjectHandle<DbObject>* objH,
                                DataCallBack*             callH,
                                const DbDatabase&         dbH, 
                                const Token&              tokenH, 
                                DbAccessMode              mode)
{
  Token tok;
  DbStatus sc = dbH.getLink(tokenH.oid(), &tok);
  if ( sc.isSuccess() )   {
    DbContainer cntH(tok.technology());
    if ( dbH.name() == tok.dbID() )   {
      const DbTypeInfo* db_info = dbH.objectShape(tok.classID());
      if ( cntH.open(dbH, tok.contID(), db_info, tok.technology(), mode).isSuccess() )   {
        return open(objH, callH, cntH, tokenH, mode);
      }
    }
  }
  return Error;
}

/// Open object with handle or callback
DbStatus DbObjectAccessor::open(DbObjectHandle<DbObject>* objH,
                                DataCallBack*             callH,
                                const DbDatabase&         dbH,  
                                const DbLink&             linkH, 
                                DbAccessMode              mod)
{
  Token tok;
  // Without a token , no shape is available.
  // Need to get it from the database description.
  if ( dbH.getLink(linkH, &tok).isSuccess() )  {
    const DbTypeInfo* typ = dbH.objectShape(tok.classID());
    if ( typ )  {
      tok.setClassID(typ->shapeID());
      return open(objH, callH, dbH, tok, mod);
    }
  }
  return Error;
}

/// Open object with handle or callback
DbStatus DbObjectAccessor::open(DbObjectHandle<DbObject>* objH,
                                DataCallBack*             callH,
                                const DbContainer&        cntH,
                                const DbLink&             linkH, 
                                DbAccessMode              mod)
{
  return open(objH, callH, cntH.containedIn(), linkH, mod);
}

/// Open handle
DbStatus DbObjectAccessor::open(DbObjectHandle<DbObject>* objH,
                                DataCallBack*             callH,
                                const DbContainer&        cntH, 
                                const Token&              tokenH, 
                                DbAccessMode              mode )
{
  if ( callH )  {       // This is the general POOL case
    return cntH.load(callH, tokenH.oid(), mode);
  }
  else if ( objH ) {    // Here we enter when dealing with object handles
    const DbTypeInfo* typ = cntH.containedIn().objectShape(tokenH.classID());
    return cntH.load(*objH, tokenH.oid(), typ, mode);
  }
  return Error;
}
