// $Id: DbReflex.cpp,v 1.19 2007/11/21 15:25:35 frankb Exp $
//====================================================================
//  Package    : StorageSvc (The POOL project)
//
//  @author      M.Frank
//====================================================================
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbTransform.h"
#include "PluginManager/PluginManager.h"
#include "PluginManager/Module.h"
#include "PluginManager/PluginError.h"
#include "StorageSvc/DbPrint.h"
#include "SealBase/SharedLibrary.h"

namespace pool {  void genMD5(const std::string& s, void* code);  }
using namespace ROOT::Reflex;
using namespace pool;

namespace {
  void normalizeGuidString(std::string& s)  {
    for(char* c = (char*)s.c_str(); *c; ++c)  {
      switch(*c) {
        case 'a':  *c = 'A'; break;
        case 'b':  *c = 'B'; break;
        case 'c':  *c = 'C'; break;
        case 'd':  *c = 'D'; break;
        case 'e':  *c = 'E'; break;
        case 'f':  *c = 'F'; break;
      }
    }
  }
}

typedef std::map<Guid,Type> GuidMap;
static GuidMap& guid_mapping()  {
  static GuidMap s_map;
  return s_map;
}

typedef std::map<Type,Guid> TypeMap;
static TypeMap& type_mapping()  {
  static TypeMap s_map;
  return s_map;
}

/// Dump all known dictionaries
void DbReflex::dumpDictionaries()  {
  std::string n;
  for(size_t i=0; i<Type::TypeSize(); ++i)  {
    n = Type::TypeAt(i).Name(SCOPED);
    std::cout << i << ": " << n << std::endl;
  }
}

/// Access to reflection information by type name
DbReflex::TypeH DbReflex::forTypeName(const std::string& name)  {
  // std::cout << "<-> DbReflex::forTypeName(" << name << ")" << std::endl;
  Type t = Type::ByName(name);
  return (t)? t : loadDictionaryForClass(name);
}

/// Access to reflection information by type info
DbReflex::TypeH DbReflex::forTypeInfo(const std::type_info& info)  {
  Type t = Type::ByTypeInfo(info);
  return (t)? t : forTypeName( DbTypeInfo::typeName(info) );
}

std::string DbReflex::fullTypeName(DbReflex::TypeH& typ)  {
  return typ.Name(SCOPED);
}

/// Convert Guid to nomalized string (all char upper case!)
std::string DbReflex::guidString(const Guid& id)  {
  std::string s = id.toString();
  normalizeGuidString(s);
  return s;
}

/// Convert Guid to nomalized string (all char upper case!)
Guid DbReflex::guid(DbReflex::TypeH& type)  {
  TypeMap::iterator i = type_mapping().find(type);
  if ( i != type_mapping().end() )  {
    return (*i).second;
  }
  if ( type )  {
    std::string s;
    PropertyList pl = type.Properties();
    if ( pl.HasProperty("ClassID") )  {
      s = pl.PropertyAsString("ClassID");
      Guid id(s);
      guid_mapping()[id] = type;
      type_mapping()[type] = id;
      return id;
    }
    else  {
      char buff[20];
      genMD5(fullTypeName(type),buff);
      s = ((Guid*)buff)->toString();
      guid_mapping()[*(Guid*)buff] = type;
      type_mapping()[type] = *(Guid*)buff;
      return *(Guid*)buff;
    }
  }
  return Guid::null();
}

/// Convert Guid to nomalized string (all char upper case!)
std::string DbReflex::guidString(DbReflex::TypeH& type)  {
  return guid(type).toString();
}

/// Access to reflection information by Guid
DbReflex::TypeH DbReflex::forGuid(const Guid& id)  {
  GuidMap& m = guid_mapping();
  GuidMap::iterator k = m.find(id);
  if ( k != m.end() )  {
    return (*k).second;
  }
  // Check the transformation map
  const DbTypeInfo* info = 0;
  if ( DbTransform::getShape(id, info).isSuccess() )  {
    Type t = info->clazz();
    if ( t )  {
      Guid g = guid(t);
      if ( ::memcmp(&g, &id, sizeof(Guid))==0 )  {
        return t;
      }
    }
  }
  // GUID not in the map....refresh the map for freshly loaded classes
  static bool first = true;
  if ( first )  {
    first = false;
    if ( doTrace() )  {
      std::cout << "WARNING: Transformation <-> Guid map looks incomplete for id:" 
        << id.toString() << std::endl;
    }
  }
  Type wanted;
  for(size_t j=0, n=Type::TypeSize(); j<n; ++j)  {
    TypeH t = Type::TypeAt(j);
    if ( t.IsClass() || t.IsStruct() )  {
      Guid g = guid(t);
      if ( ::memcmp(&g, &id, sizeof(Guid))==0 )  {
        wanted = t;
      }
    }
  }
  return wanted ? wanted : Type();
}


//------------------------  automatic dictionary loading


static bool        s_autoLoadDictionaries = false;
static std::set<std::string>  s_allowedDictionaries;
static std::set<std::string>  s_failedClasses;


/// Enable automatic loading of dictionaries
void DbReflex::setDictionaryAutoLoading( bool autoload )
{
  s_autoLoadDictionaries = autoload;
}

/// Allow to autoload this dictionary
/** @param moduleName  [IN] name of the dictionary module that can be autoloaded
  * (do not add "lib" and ".so" to the name) 
  */
void DbReflex::allowAutoLoadingFor(const std::string& moduleName)   {
  DbPrint log("autoDictLoad");
  std::string  libName = seal::SharedLibrary::libname( moduleName );
  log << DbPrint::DEBUG
    << "Adding library " << libName << " to the list of allowed dynamic dictionaries"
    << DbPrint::endmsg;
  s_allowedDictionaries.insert( libName );
  setDictionaryAutoLoading( true );
}


/// Load dictionary module (seal plugin module, shared library) that defines class "className"
/// The name of the class is checked against seal module capabilities list
Type DbReflex::loadDictionaryForClass(const std::string &className)   {
  DbPrint log("autoDictLoad"); 
  log << DbPrint::DEBUG
    << "loadDictionaryForClass: " << className
    << ", autoLoadDictionaries=" << s_autoLoadDictionaries << DbPrint::endmsg;
  if( !s_autoLoadDictionaries )
    return 0;
  if( s_failedClasses.size() && s_failedClasses.find(className) != s_failedClasses.end() ) {
    log << DbPrint::VERBOSE
      << "Attempt to autoload class " << className
      << " failed before, NOT trying again" <<  DbPrint::endmsg;
    return 0;
  }
  log << DbPrint::DEBUG
    << "Attempt to dynamically load dictionary for class " << className << DbPrint::endmsg;

  std::string  capability = "LCGReflex/" + className;  
  seal::PluginManager  *plugMgr = seal::PluginManager::get();
  plugMgr->initialise();
  seal::PluginManager::ModuleIterator  modIter = plugMgr->beginModules();
  while( modIter != plugMgr->endModules() ) {
    seal::Module  *module = *modIter;
    const std::string  lib_name = module->libraryName().name();
    if( s_allowedDictionaries.empty()
      || s_allowedDictionaries.find(lib_name) != s_allowedDictionaries.end() ) {
        seal::ModuleDescriptor *descr = module->cacheRoot();
        for(unsigned i=0; i<descr->children(); ++i) {
          if( descr->child(i)->token(0)=="Capability" 
            && descr->child(i)->token(1)==capability ) {
              log << DbPrint::INFO
                << "Capability " <<  capability << " found in module "
                << module->libraryName() <<  DbPrint::endmsg;
              try { module->load(); }
              catch( seal::PluginError &e ) {
                log << DbPrint::ERROR
                  << "Error loading library " << lib_name << ": "
                  << e.explainSelf() << DbPrint::endmsg;
                break;  //break from the inner loop and check other modules
              }
              Type t = Type::ByName(className);
              if( t ) {
                log << DbPrint::INFO
                  << "Class " << className << " dictionary ("
                  << lib_name << ") LOADED"
                  <<  DbPrint::endmsg;
                return t;
              }
              break;
            }
        }
      }
      modIter++;
  }
  log << DbPrint::ERROR
    << "Class " << className << " NOT FOUND in "
    <<  (!s_allowedDictionaries.empty() ? "allowed " : "")
    << "dictionaries" <<  DbPrint::endmsg;
  s_failedClasses.insert( className );
  return 0;
}
