#include "Collection/Collection.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Exception.h"
#include "POOLCore/Token.h"
#include "TestDictionary/SimpleTestClass.h"
#include "TestDictionary/TestClassPrimitives.h"

#include "SealKernel/Status.h"
#include "SealBase/SharedLibrary.h"

#include <stdexcept>

const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string DB_FILE_ESD = "TestFileESD.pool";
const std::string DB_FILE_AOD = "TestFileAOD.pool";
const pool::DbType DB_TYPE = pool::ROOTTREE_StorageType;
const std::string CONTAINER_NAME_ESD = "TestContainerESD";
const std::string CONTAINER_NAME_AOD = "TestContainerAOD";
const std::string COLLECTION_NAME_ESD = "TestCollectionESD";
const std::string COLLECTION_NAME_AOD = "TestCollectionAOD";
const std::string COLLECTION_NAME_TAG = "TestCollectionTag";
const std::string QUERY = "event < 7";
const unsigned int NUM_EVENTS = 20;

int main(int argc, const char *argv[])
{
  try
  { 
    seal::Status sc = seal::Status::SUCCESS;
    std::string collectionDb = "";
    unsigned int evtCounter = 0;

    if( argc < 2 )
    {
      std::cerr << "Collection_BackNavigate: Correct command line usage is: "
        << "test_Collection_BackNavigate <collection type> "
          << "[SEAL message level]" << std::endl;
      return 1;
    }

    std::string collectionType = std::string( argv[1] );
    if( collectionType == "RootCollection" )
    {
      collectionDb = "";
    }
    else if( collectionType == "MySQLCollection" || collectionType ==
      "MySQLltCollection" )
    {
      collectionDb = 
        "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression";
    }
    else if ( collectionType == "RelationalCollectionMySQL" )
    {
      collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
    }
    else if ( collectionType == "RelationalCollectionOracle" )
    {
      collectionDb = "oracle://devdb10.cern.ch/ral_writer";
    }
    else
    {
      std::cerr << "Collection_BackNavigate: Collection type " 
        << collectionType << "is not presently handled by this program." 
          << std::endl;
      return 1;
    }

    std::cout << "Collection_BackNavigate: Initializing the SEAL services." 
      << std::endl;
    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );

    std::cout << "Collection_BackNavigate: Declaring a POOL message stream."
      << std::endl;
    seal::Msg::Level msgLevel = seal::Msg::Info;
    if ( argc >2 )
    {
      if( std::string(argv[2]) == "Verbose" )
      {
        msgLevel = seal::Msg::Verbose;
      }
      else if ( std::string(argv[2]) == "Debug" )
      {
        msgLevel = seal::Msg::Debug;
      }
      else if ( std::string(argv[2]) == "Info" )
      {
        msgLevel = seal::Msg::Info;
      }
      else if ( std::string(argv[2]) == "Warning" )
      {
        msgLevel = seal::Msg::Warning;
      }
      else if ( std::string(argv[2]) == "Error" )
      {
        msgLevel = seal::Msg::Error;
      }
      else if ( std::string(argv[2]) == "Fatal" )
      {
        msgLevel = seal::Msg::Fatal;
      }
      else
      {
        std::cerr << "Collection_BackNavigate: Warning: Message level command "
          << "line argument not set to expected value. Using seal::Msg::Info." 
            << std::endl;
        return 1;
      }
    }
    pool::POOLContext::setMessageVerbosityLevel( msgLevel );
    pool::PoolMessageStream log( "Collection_BackNavigate", msgLevel );

    log << seal::Msg::Info << "Loading the shared libraries." << seal::endmsg;
    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    log << seal::Msg::Info << "Creating a file catalog." << seal::endmsg;
    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->setWriteCatalog( p.contactstring() );
    fileCatalog->addReadCatalog( p.contactstring() );

    log << seal::Msg::Info << "Establishing a connection to the file catalog." 
      << seal::endmsg;
    fileCatalog->connect();

    log << seal::Msg::Info << "Starting the file catalog transaction." 
      << seal::endmsg;
    fileCatalog->start();

    log << seal::Msg::Info << "Creating a data service instance." 
      << seal::endmsg;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );

    log << seal::Msg::Info << "Setting the database connection policy for "
      << "this session." << seal::endmsg;
    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting(pool::DatabaseConnectionPolicy::CREATE);
    policy.setWriteModeForExisting(pool::DatabaseConnectionPolicy::OVERWRITE);
    policy.setReadMode(pool::DatabaseConnectionPolicy::READ);
    dataSvc->session().setDefaultConnectionPolicy(policy);
    log << seal::Msg::Info << "Write mode for a non-existing database is "
      << "set to: " << policy.writeModeForNonExisting() << seal::endmsg;
    log << seal::Msg::Info << "Write mode for an existing database is set to: "
      << policy.writeModeForExisting() 
        << seal::endmsg;
    log << seal::Msg::Info << "Read mode for a database is set to: " 
      << policy.readMode() << seal::endmsg;

    log << seal::Msg::Info << "Creating a Placement object for the ESD "
      << "objects." << seal::endmsg;
    pool::Placement placementESD;
    placementESD.setDatabase( DB_FILE_ESD, pool::DatabaseSpecification::PFN );
    placementESD.setContainerName( CONTAINER_NAME_ESD );
    placementESD.setTechnology( DB_TYPE.type() );

    log << seal::Msg::Info << "Creating a Placement object for the AOD "
      << "objects." << seal::endmsg;
    pool::Placement placementAOD;
    placementAOD.setDatabase( DB_FILE_AOD, pool::DatabaseSpecification::PFN );
    placementAOD.setContainerName( CONTAINER_NAME_AOD );
    placementAOD.setTechnology( DB_TYPE.type() );

    log << seal::Msg::Info << "Creating primitive test attributes." 
      << seal::endmsg;
    TestClassPrimitives tcPrim;
    tcPrim.setNonZero();

    log << seal::Msg::Info << "Starting a database transaction for UPDATE." 
      << seal::endmsg;
    sc = dataSvc->transaction().start( pool::ITransaction::UPDATE );
    if( !sc.isSuccess() )
    {
      log << seal::Msg::Info << "Error when starting database transaction." 
        << seal::endmsg;
    }

    log << seal::Msg::Info << "Adding objects to the ESD file:" 
      << seal::endmsg;
    std::cout << std::endl;

    for (unsigned int evt = 1; evt <= NUM_EVENTS; evt++)
    {
      log << seal::Msg::Debug << "Events counted: " << evt << seal::endmsg;

      log << seal::Msg::Debug << "Creating a SimpleTestClass object with null "
        << "values." << seal::endmsg;
      pool::Ref<SimpleTestClass> 
        simpleTestClass( dataSvc, new SimpleTestClass() );

      if ( msgLevel <= seal::Msg::Debug )
      { 
        std::cout << "Collection_BackNavigate: The ESD object's data is: ";
        simpleTestClass->streamOut(std::cout);
        std::cout << std::endl;
      }

      log << seal::Msg::Debug << "Marking the object for writing in the ESD "
        << "database." << seal::endmsg;
      simpleTestClass.markWrite( placementESD );
      if( evt == 1 )
      {
        fileCatalog->commit();
	fileCatalog->start();
      }

      std::cout << std::endl;
    }

    log << seal::Msg::Info << "Creating an AOD collection for "
      << "CREATE_AND_OVERWRITE:" << seal::endmsg;
    log << seal::Msg::Info << "Collection type is " << collectionType 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection location is " << collectionDb 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection name is " << COLLECTION_NAME_AOD 
      << seal::endmsg;
    pool::Collection<SimpleTestClass>
      collectionWriteAOD( dataSvc, 
                          collectionType, 
                          collectionDb, 
                          COLLECTION_NAME_AOD, 
                          pool::ICollection::CREATE_AND_OVERWRITE );
    std::cout << std::endl;

    log << seal::Msg::Info << "Defining an attribute list specification for "
      << "the AOD collection." << seal::endmsg;
    pool::AttributeListSpecification& attribSpecAOD = 
      collectionWriteAOD.attributeListSpecification();
    attribSpecAOD.push_back( "event", "unsigned int" );
    attribSpecAOD.push_back( "ESD_obj_token", "string" );
    attribSpecAOD.push_back( "test_bool", "bool" );
    attribSpecAOD.push_back( "test_short", "short" );
    attribSpecAOD.push_back( "test_ushort", "unsigned short" );
    attribSpecAOD.push_back( "test_sshort", "short" );
    attribSpecAOD.push_back( "test_shortint", "short" );
    attribSpecAOD.push_back( "test_ushortint", "unsigned short" );
    attribSpecAOD.push_back( "test_sshortint", "short" );
    attribSpecAOD.push_back( "test_int", "int" );
    attribSpecAOD.push_back( "test_uint", "unsigned int" );
    attribSpecAOD.push_back( "test_sint", "int" );
    attribSpecAOD.push_back( "test_s", "int" );
    attribSpecAOD.push_back( "test_u", "unsigned int" );
    attribSpecAOD.push_back( "test_long", "long" );
    attribSpecAOD.push_back( "test_ulong", "unsigned long" );
    attribSpecAOD.push_back( "test_slong", "long" );
    attribSpecAOD.push_back( "test_longint", "long" );
    attribSpecAOD.push_back( "test_ulongint", "unsigned long" );
    attribSpecAOD.push_back( "test_slongint", "long" );
    attribSpecAOD.push_back( "test_float", "float" );
    attribSpecAOD.push_back( "test_double", "double" );
    attribSpecAOD.push_back( "test_string", "string" );
    std::cout << "Collection_BackNavigate: ";
    attribSpecAOD.print(std::cout);
    std::cout << std::endl;

    log << seal::Msg::Info << "Opening the ESD file as an implicit collection "
      << "for READ:" << seal::endmsg;
    log << seal::Msg::Info << "Collection type is ImplicitCollection" 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection location is PFN:" << DB_FILE_ESD
      << seal::endmsg;
    log << seal::Msg::Info << "Collection name is " << CONTAINER_NAME_ESD
      << seal::endmsg;
    pool::Collection<SimpleTestClass> 
      collectionReadESD( dataSvc, 
                         "ImplicitCollection", 
                         "PFN:"+DB_FILE_ESD,
                         CONTAINER_NAME_ESD, 
                         pool::ICollection::READ );
    std::cout << std::endl;

    log << seal::Msg::Info << "Getting an iterator over all objects in the "
      << "ESD collection." << seal::endmsg;
    pool::Collection<SimpleTestClass>::Iterator collIterESD = 
      collectionReadESD.select( "", "", "FetchOne" );
    std::cout << std::endl;

    log << seal::Msg::Info << "Looping over all objects in the ESD collection "
      << "and adding objects and attributes, including Tokens of the ESD "
        << "objects, to the new AOD collection:" << seal::endmsg;
    std::cout << std::endl;

    evtCounter = 0;
    while ( collIterESD.next() ) 
    {
      evtCounter++;
      log << seal::Msg::Debug << "Events counted: " << evtCounter 
        << seal::endmsg;

      log << seal::Msg::Debug << "New ESD object found." << seal::endmsg;

      if ( msgLevel <= seal::Msg::Debug )
      {
        std::cout << "Collection_BackNavigate: The ESD object's data is: ";
        collIterESD->streamOut( std::cout );
        std::cout << std::endl;
      }

      log << seal::Msg::Debug << "Creating a SimpleTestClass object."
        << seal::endmsg;
      pool::Ref<SimpleTestClass> 
        simpleTestClass( dataSvc, new SimpleTestClass() );

      log << seal::Msg::Debug << "Marking the object for writing in the AOD "
        << "database." << seal::endmsg;
      simpleTestClass.markWrite( placementAOD );
      if( evtCounter == 1 )
      {
        fileCatalog->commit();
	fileCatalog->start();
      }

      log << seal::Msg::Debug << "Setting the AOD object's data to non-zero "
        << "values." << seal::endmsg;
      simpleTestClass->setNonZero();

      if ( msgLevel <= seal::Msg::Debug )
      {
        std::cout << "Collection_BackNavigate: The AOD object's data is: ";
        simpleTestClass->streamOut(std::cout);
        std::cout << std::endl;
      }

      log << seal::Msg::Debug << "Setting the AOD object's attributes." 
        << seal::endmsg;
      pool::AttributeList attributeList( attribSpecAOD );
      attributeList[ "event" ].setValue( evtCounter );
      attributeList[ "ESD_obj_token" ].setValue( collIterESD.ref().toString() );
      attributeList[ "test_bool" ].setValue( tcPrim.m_bool );
      attributeList[ "test_short" ].setValue( tcPrim.m_short );
      attributeList[ "test_ushort" ].setValue( tcPrim.m_ushort );
      attributeList[ "test_sshort" ].setValue( tcPrim.m_sshort );
      attributeList[ "test_shortint" ].setValue( tcPrim.m_shortint );
      attributeList[ "test_ushortint" ].setValue( tcPrim.m_ushortint );
      attributeList[ "test_sshortint" ].setValue( tcPrim.m_sshortint );
      attributeList[ "test_int" ].setValue( tcPrim.m_int );
      attributeList[ "test_uint" ].setValue( tcPrim.m_uint );
      attributeList[ "test_sint" ].setValue( tcPrim.m_sint );
      attributeList[ "test_s" ].setValue( tcPrim.m_s );
      attributeList[ "test_u" ].setValue( tcPrim.m_u );
      attributeList[ "test_long" ].setValue( tcPrim.m_long );
      attributeList[ "test_ulong" ].setValue( tcPrim.m_ulong );
      attributeList[ "test_slong" ].setValue( tcPrim.m_slong );
      attributeList[ "test_longint" ].setValue( tcPrim.m_longint );
      attributeList[ "test_ulongint" ].setValue( tcPrim.m_ulongint );
      attributeList[ "test_slongint" ].setValue( tcPrim.m_slongint );
      attributeList[ "test_float" ].setValue( tcPrim.m_float );
      attributeList[ "test_double" ].setValue( tcPrim.m_double );
      attributeList[ "test_string" ].setValue( std::string("Test String") );
      std::cout << "Collection_BackNavigate: ";
      attributeList.print(std::cout);
      std::cout << std::endl;

      log << seal::Msg::Debug << "Adding the AOD object's Token and "
        << "attributes to the AOD collection." << seal::endmsg; 
      collectionWriteAOD.add( simpleTestClass, attributeList );

      std::cout << std::endl;
    }

    log << seal::Msg::Info << "Committing all changes made to the AOD "
      << "collection." << seal::endmsg;
    collectionWriteAOD.commit();
    std::cout << std::endl;

    log << seal::Msg::Info << "Closing the AOD collection." << seal::endmsg;
    collectionWriteAOD.close();
    std::cout << std::endl;

    log << seal::Msg::Info << "Closing the ESD collection." << seal::endmsg;
    collectionReadESD.close();
    std::cout << std::endl;

    log << seal::Msg::Info << "Creating an event tag collection for "
      << "CREATE_AND_OVERWRITE:" << seal::endmsg;
    log << seal::Msg::Info << "Collection type is " << collectionType 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection location is " << collectionDb 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection name is " << COLLECTION_NAME_TAG 
      << seal::endmsg;
    pool::Collection<SimpleTestClass>
      collectionWriteTag( dataSvc, 
                          collectionType, 
                          collectionDb,
                          COLLECTION_NAME_TAG, 
                          pool::ICollection::CREATE_AND_OVERWRITE );
    std::cout << std::endl;

    log << seal::Msg::Info << "Defining the attribute list specification for "
      << "the event tag collection." << seal::endmsg;
    pool::AttributeListSpecification& attribSpecTag = 
      collectionWriteTag.attributeListSpecification();
    attribSpecTag.push_back( "event", "unsigned int" );
    attribSpecTag.push_back( "ESD_obj_token", "string" );
    attribSpecTag.push_back( "AOD_obj_token", "string" );
    attribSpecTag.push_back( "test_bool", "bool" );
    attribSpecTag.push_back( "test_short", "short" );
    attribSpecTag.push_back( "test_int", "int" );
    attribSpecTag.push_back( "test_long", "long" );
    attribSpecTag.push_back( "test_float", "float" );
    attribSpecTag.push_back( "test_double", "double" );
    attribSpecTag.push_back( "test_string", "string" );
    log << seal::Msg::Debug << " ";
    attribSpecTag.print(std::cout);
    std::cout << std::endl;
    std::cout << std::endl;

    log << seal::Msg::Info << "Reopening the AOD collection for READ:" 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection type is " << collectionType 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection location is " << collectionDb 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection name is " << COLLECTION_NAME_AOD
      << seal::endmsg;
    pool::Collection<SimpleTestClass>
      collectionReadAOD( dataSvc, 
                         collectionType, 
                         collectionDb, 
                         COLLECTION_NAME_AOD,
                         pool::ICollection::READ );
    std::cout << std::endl;

    log << seal::Msg::Info << "Getting an iterator over all objects in the "
      << "AOD collection that satisfy a query." << seal::endmsg;
    pool::Collection<SimpleTestClass>::Iterator collIterAOD = 
      collectionReadAOD.select( QUERY, "", "FetchOne" );
    std::cout << std::endl;

    log << seal::Msg::Info << "Looping over all objects in the AOD "
      << "collection that satisfy the query:" << seal::endmsg;
    std::cout << std::endl;

    unsigned int evtNum = 0;
    bool testBool = false;
    short testShort = 0;
    int testInt = 0;
    long testLong = 0;
    float testFloat = 0.0;
    double testDouble = 0.0;
    std::string testString = "";
    evtCounter = 0;
    while ( collIterAOD.next() ) 
    {
      pool::Token* esdObjToken = new pool::Token;

      evtCounter++;
      log << seal::Msg::Debug << "Events counted: " << evtCounter 
        << seal::endmsg;

      log << seal::Msg::Debug << "New AOD object found." << seal::endmsg;

      if ( msgLevel <= seal::Msg::Debug )
      {
        std::cout << "Collection_BackNavigate: The AOD object's data is: ";
        collIterAOD->streamOut( std::cout );
        std::cout << std::endl;
      }

      log << seal::Msg::Debug << "The AOD object's attributes are: " 
        << seal::endmsg;
      const pool::AttributeList& attributeListAOD = 
        collIterAOD.attributeList();
      if ( msgLevel <= seal::Msg::Debug )
      {
        std::cout << "Collection_BackNavigate: ";
        attributeListAOD.print(std::cout);
        std::cout << std::endl;
      }

      log << seal::Msg::Debug << "Setting the AOD object's event tag "
        << "attributes." << seal::endmsg;
      pool::AttributeList attributeListTag( attribSpecTag );
      attributeListAOD[ "event" ].getValue( evtNum );
      attributeListTag[ "event" ].setValue( evtNum );
      std::string esdObjTokenString = "";
      attributeListAOD[ "ESD_obj_token" ].getValue( esdObjTokenString );
      attributeListTag[ "ESD_obj_token" ].setValue( esdObjTokenString );
      attributeListTag[ "AOD_obj_token" ].setValue( collIterAOD.ref().toString() );
      attributeListAOD[ "test_bool" ].getValue( testBool );
      attributeListTag[ "test_bool" ].setValue( testBool );
      attributeListAOD[ "test_short" ].getValue( testShort );
      attributeListTag[ "test_short" ].setValue( testShort );
      attributeListAOD[ "test_int" ].getValue( testInt );
      attributeListTag[ "test_int" ].setValue( testInt );
      attributeListAOD[ "test_long" ].getValue( testLong );
      attributeListTag[ "test_long" ].setValue( testLong );
      attributeListAOD[ "test_float" ].getValue( testFloat );
      attributeListTag[ "test_float" ].setValue( testFloat );
      attributeListAOD[ "test_double" ].getValue( testDouble );
      attributeListTag[ "test_double" ].setValue( testDouble );
      attributeListAOD[ "test_string" ].getValue( testString );
      attributeListTag[ "test_string" ].setValue( testString );
      log << seal::Msg::Debug << " ";
      attributeListTag.print(std::cout);
      std::cout << std::endl;

      log << seal::Msg::Debug << "Adding the AOD object's Token and event tag "
        << "attributes to the event tag collection." << seal::endmsg; 
      collectionWriteTag.add( collIterAOD.ref(), attributeListTag );

      esdObjToken->release();

      std::cout << std::endl;
    }

    log << seal::Msg::Info << "Committing all changes made to the event tag "
      << "collection." << seal::endmsg;
    collectionWriteTag.commit();
    std::cout << std::endl;

    log << seal::Msg::Info << "Closing the event tag collection." 
      << seal::endmsg;
    collectionWriteTag.close();
    std::cout << std::endl;

    log << seal::Msg::Info << "Closing the AOD collection." << seal::endmsg;
    collectionReadAOD.close();
    std::cout << std::endl;

    log << seal::Msg::Info << "Ending the database transaction." 
      << seal::endmsg;
    sc = dataSvc->transaction().commit();
    if( !sc.isSuccess() )
    {
      log << seal::Msg::Error << "Error when ending database transaction." 
        << seal::endmsg;
    }
    std::cout << std::endl;

    log << seal::Msg::Info << "Disconnecting from the database." 
      << seal::endmsg;
    delete dataSvc;
    std::cout << std::endl;

    log << seal::Msg::Info << "Creating a data service instance." 
      << seal::endmsg;
    dataSvc = pool::DataSvcFactory::instance( fileCatalog );
    dataSvc->session().setDefaultConnectionPolicy(policy);

    log << seal::Msg::Info << "Starting a database transaction for READ." 
      << seal::endmsg;
    sc = dataSvc->transaction().start( pool::ITransaction::READ );
    if( !sc.isSuccess() )
    {
      log << seal::Msg::Info << "Error when starting database transaction." 
        << seal::endmsg;
    }

    std::cout << std::endl;
    log << seal::Msg::Info << "Reopening the event tag collection for READ:" 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection type is " << collectionType 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection location is " << collectionDb 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection name is " << COLLECTION_NAME_TAG
      << seal::endmsg;
    pool::Collection<SimpleTestClass>
      collectionReadTag( dataSvc, 
                         collectionType, 
                         collectionDb, 
                         COLLECTION_NAME_TAG, 
                         pool::ICollection::READ );
    std::cout << std::endl;
    std::cout << std::endl;

    log << seal::Msg::Info << "Getting an iterator over all objects in the "
      << "event tag collection." << seal::endmsg;
    pool::Collection<SimpleTestClass>::Iterator collIterTag = 
      collectionReadTag.select( "", "", "FetchOne" );
    std::cout << std::endl;

    log << seal::Msg::Info << "Looping over all objects in the event tag "
      << "collection:" << seal::endmsg;
    std::cout << std::endl;

    std::string dbFileId = "";
    std::string bestPFN = "";
    std::string fileType = "";
    std::string contId = "";
    pool::Token::OID_t objId;
    evtCounter = 0;
    while ( collIterTag.next() ) 
    {
      pool::Token* esdObjToken = new pool::Token;
      pool::Token* aodObjToken = new pool::Token;

      evtCounter++;
      log << seal::Msg::Info << "Events counted: " << evtCounter 
        << seal::endmsg;

      log << seal::Msg::Debug << "New event tag object found." << seal::endmsg;

      log << seal::Msg::Debug << "Retrieving the event tag object's attribute "
        << "list." << seal::endmsg;
      const pool::AttributeList& attributeListTag = 
        collIterTag.attributeList();

      log << seal::Msg::Info << "The event tag object's attributes are:" 
        << seal::endmsg;
      if ( msgLevel <= seal::Msg::Info )
      {
        std::cout << std::endl;
        attributeListTag.print(std::cout);
        std::cout << std::endl;
      }

      log << seal::Msg::Debug << "Retrieving Tokens of the AOD and ESD "
        << "objects." << seal::endmsg;
      std::string esdObjTokenString = "";
      attributeListTag[ "ESD_obj_token" ].getValue( esdObjTokenString );
      esdObjToken->fromString( esdObjTokenString );
      std::string aodObjTokenString = "";
      attributeListTag[ "AOD_obj_token" ].getValue( aodObjTokenString );
      aodObjToken->fromString( aodObjTokenString );

      log << seal::Msg::Debug << "Retrieving references to the AOD and ESD "
        << "objects." << seal::endmsg;
      pool::Ref<SimpleTestClass> simpleTestClassAOD( dataSvc, *aodObjToken );
      pool::Ref<SimpleTestClass> simpleTestClassESD( dataSvc, *esdObjToken );

      if ( msgLevel <= seal::Msg::Info )
      {
        std::cout << "Collection_BackNavigate: The event tag object's "
          << "data is: ";
        collIterTag->streamOut( std::cout );
        std::cout << std::endl;
        std::cout << "Collection_BackNavigate: The AOD object's data is: ";
        simpleTestClassAOD->streamOut(std::cout);
        std::cout << std::endl;
        std::cout << "Collection_BackNavigate: The ESD object's data is: ";
        simpleTestClassESD->streamOut(std::cout);
        std::cout << std::endl;
      }

      dbFileId = collIterTag.ref().token()->dbID();
      log << seal::Msg::Info << "The event tag database file ID is " 
        << dbFileId << seal::endmsg;
      pool::IFCAction action;
      fileCatalog->setAction( action );
      action.lookupBestPFN( dbFileId, 
                            pool::FileCatalog::READ, 
                            pool::FileCatalog::SEQUENTIAL, 
                            bestPFN, 
                            fileType );
      log << seal::Msg::Info << "The event tag database physical file "
        << "name is " << bestPFN << seal::endmsg;
      log << seal::Msg::Info << "The event tag database file type is " 
        << fileType << seal::endmsg;
      contId = collIterTag.ref().token()->contID();
      log << seal::Msg::Info << "The event tag database container ID is " 
        << contId << seal::endmsg;
      objId = collIterTag.ref().token()->oid();
      log << seal::Msg::Info << "The event tag database object ID is " 
        << objId.first << "-" << objId.second << seal::endmsg;

      dbFileId = aodObjToken->dbID();
      log << seal::Msg::Info << "The AOD database file ID is " << dbFileId 
        << seal::endmsg;
      action.lookupBestPFN( dbFileId, 
                            pool::FileCatalog::READ, 
                            pool::FileCatalog::SEQUENTIAL, 
                            bestPFN, 
                            fileType );
      log << seal::Msg::Info << "The AOD database physical file name is " 
        << bestPFN << seal::endmsg;
      log << seal::Msg::Info << "The AOD database file type is " << fileType 
        << seal::endmsg;
      contId = aodObjToken->contID();
      log << seal::Msg::Info << "The AOD database container ID is " << contId 
        << seal::endmsg;
      objId = aodObjToken->oid();
      log << seal::Msg::Info << "The AOD database object ID is " 
        << objId.first << "-" << objId.second << seal::endmsg;

      dbFileId = esdObjToken->dbID();
      log << seal::Msg::Info << "The ESD database file ID is " << dbFileId 
        << seal::endmsg;
      action.lookupBestPFN( dbFileId, 
                            pool::FileCatalog::READ, 
                            pool::FileCatalog::SEQUENTIAL, 
                            bestPFN, 
                            fileType );
      log << seal::Msg::Info << "The ESD database physical file name is " 
        << bestPFN << seal::endmsg;
      log << seal::Msg::Info << "The ESD database file type is " << fileType 
        << seal::endmsg;
      contId = esdObjToken->contID();
      log << seal::Msg::Info << "The ESD database container ID is " << contId 
        << seal::endmsg;
      objId = esdObjToken->oid();
      log << seal::Msg::Info << "The ESD database object ID is " 
        << objId.first << "-" << objId.second << seal::endmsg;

      esdObjToken->release();
      aodObjToken->release();

      std::cout << std::endl;
    }

    log << seal::Msg::Info << "Closing the event tag collection." 
      << seal::endmsg;
    collectionReadTag.close();
    std::cout << std::endl;

    log << seal::Msg::Info << "Ending the database transaction." 
      << seal::endmsg;
    sc = dataSvc->transaction().commit();
    if( !sc.isSuccess() )
    {
      log << seal::Msg::Error << "Error when ending database transaction." 
        << seal::endmsg;
    }
    std::cout << std::endl;

    log << seal::Msg::Info << "Disconnecting from the database." 
      << seal::endmsg;
    delete dataSvc;
    std::cout << std::endl;

    log << seal::Msg::Info << "Committing all file catalog changes." 
      << seal::endmsg;
    fileCatalog->commit();
    std::cout << std::endl;

    log << seal::Msg::Info << "Disconnecting from the file catalog." 
      << seal::endmsg;
    fileCatalog->disconnect();
    delete fileCatalog;
    std::cout << std::endl;

    log << seal::Msg::Info << "Returning from main()" << seal::endmsg;

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::exception: " << exception.what() << std::endl;;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception caught in main()." << std::endl;;
    return 1;
  }
}
