#include "Collection/CollectionProxy.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFileCatalog.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"
#include "PoolBench/ProcessTime.h"
#include "util.h" // string conversions in ns Util

#include "SealKernel/Status.h"
#include "SealBase/SharedLibrary.h"

# include <iostream>
# include <fstream>
# include <stdexcept>

using namespace std;

const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";

int main(int argc, const char *argv[])
{
  try
  {
    seal::Status sc = seal::Status::SUCCESS;
    bool isInt = true;

    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
    pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

    pool::PoolMessageStream log( "Collection_ExplicitReadPerformance", 
      seal::Msg::Info );
    int msgLevel = seal::Msg::Info - 1;
    char* msgLevelGetEnv = ::getenv( "POOL_OUTMSG_LEVEL" );
    char msgLevelSetEnv[20] = "";
    if( msgLevelGetEnv )
    {
      msgLevel = atoi( msgLevelGetEnv );
      if( ( msgLevel < (seal::Msg::Nil-1) ) || 
	( msgLevel > (seal::Msg::Fatal-1) ) )
      {
	log << seal::Msg::Fatal << "Environment variable POOL_OUTMSG_LEVEL "
	    << "is set to " << msgLevelGetEnv << " which is out of required "
	      << " range of [" << (seal::Msg::Nil-1) << "," 
		<< (seal::Msg::Fatal-1) << "]" << seal::endmsg;
	return 1;
      }
    }
    else
    {
      log << seal::Msg::Warning << "Environment variable POOL_OUTMSG_LEVEL "
	<< "not set. Setting it to (seal::Msg::Info-1)" << seal::endmsg;
      std::sprintf( msgLevelSetEnv, "%i", 2 );
      ::setenv( "POOL_OUTMSG_LEVEL", msgLevelSetEnv, 1 );
    }

    if( argc < 5 )
    {
      log << seal::Msg::Fatal 
	<< "Usage: test_Collection_ExplicitReadPerformance "
	<< "<collection name prefix> "
	<< "<# of events> " 
	<< "<# of attributes> "
        << "<collection type (Root, MySQL, MySQLlt, RALMySQL or RALOracle)> "
	<< "Optional arguments (use - char to mask out those not needed): "
	<< "[Task to perform on query results: 0=empty loop, 1=type check, "
	<< "2=create ref<T>, 3=dereference ref<T>, 4=read attribute list, "
	<< "5=convert ref<T> to string DEFAULT=0] " 
	<< "[name of performance data output file DEFAULT = Test_Read.log] "
	<< "[selection query DEFAULT = \"\"] "
	<< "[query options: \"SELECT <attribute list>\", \"FetchOne\" "
        << "(no caching, DEFAULT, MySQL), \"FetchAll\" (caching, MySQL), "
	<< "\"NO_TOKEN\" (no Token selection, MySQL), "
        << "\"ROOT_SELECTION_SERVER\" (remote server selection, Root)]"
	<< "[benchmark options: attribute_list_constructor (create attribute "
	<< "list from iterator for every event) or pause (wait for a key "
        << "before doing the benchmark) DEFAULT = attribute_list_constructor] "
	<< seal::endmsg;
      return 1;
    }

    std::string basename = std::string(argv[1])+"_"+argv[2]+"_"+argv[3]+
      "_"+argv[4]+"_";

    std::string dbFile = basename;
    dbFile += "File.pool";
    std::string containerName = basename;
    containerName += "Container";
    std::string collectionName = basename; 
    collectionName += "Collection";
    std::string collectionType = argv[4]; 
    collectionType += "Collection";

    unsigned int event_number = 0;
    if(!Util::to_value(argv[2],event_number))
    {
      log << seal::Msg::Fatal << "Event number should be integer, got: " 
	<< argv[2]<< seal::endmsg;
      return 1;
    }

    unsigned int attr_number = 0;
    if(!Util::to_value(argv[3],attr_number))
    {
      log << seal::Msg::Fatal << "Attribute number should be integer, got: " 
	<< argv[3]<< seal::endmsg;
      return 1;
    }

    std::string collectionDb = "";
    char* dbConnectGetEnv;
    if( !strcmp( argv[4], "Root" ) )
    { 
      collectionDb = "";
    }
    else if( !strcmp( argv[4], "MySQL" ) || 
             !strcmp( argv[4], "MySQLlt" ) )
    {   
      dbConnectGetEnv = ::getenv( "POOL_MYSQL_CONNECT" );
      if( dbConnectGetEnv )
      {
	collectionDb = dbConnectGetEnv;
      }
      else
      {
	log << seal::Msg::Warning << "Environment variable POOL_MYSQL_CONNECT "
	  << "not set. Setting MySQL database connection to `" 
	    << "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression'"
	      << seal::endmsg;
	collectionDb = 
          "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression"; 
      }
    }
    else if( !strcmp( argv[4], "RALMySQL" ) )
    {   
      dbConnectGetEnv = ::getenv( "POOL_RAL_MYSQL_CONNECT" );
      if( dbConnectGetEnv )
      {
	collectionDb = dbConnectGetEnv;
      }
      else
      {
	log << seal::Msg::Warning << "Environment variable POOL_RAL_MYSQL_CONNECT "
	  << "not set. Setting RAL MySQL database connection to `" 
	    << "mysql://pcitdb59.cern.ch/ralcolls'"
	      << seal::endmsg;
	collectionDb = 
          "mysql://pcitdb59.cern.ch/ralcolls"; 
      }
    }
    else if( !strcmp( argv[4], "RALOracle" ) )
    {   
      dbConnectGetEnv = ::getenv( "POOL_RAL_ORACLE_CONNECT" );
      if( dbConnectGetEnv )
      {
	collectionDb = dbConnectGetEnv;
      }
      else
      {
	log << seal::Msg::Warning << "Environment variable POOL_RAL_ORACLE_CONNECT "
	  << "not set. Setting RAL Oracle database connection to `" 
	    << "oracle://devdb10.cern.ch/ral_writer'"
	      << seal::endmsg;
	collectionDb = 
          "oracle://devdb10.cern.ch/ral_writer"; 
      }
    }
    else
    {
      log << seal::Msg::Fatal << "Command line argument " << argv[4]
	<< " is not an expected database type." << seal::endmsg;
      return 1;
    }

    int whatToDo = 0;
    if( argc > 5 )
    {
      if( strcmp(argv[5], "-") )
      {   
	isInt = Util::to_value(argv[5], whatToDo);
	if( !isInt || (whatToDo<0) || (whatToDo>5) )
	{
	  log << seal::Msg::Fatal << "Task number to perform on query result "
	    << "must be a positive integer < 6. Got: " << argv[5] 
	      << seal::endmsg;
	  return 1;
	}
      }
    }

    std::string output_filename = "Test_Read.log";
    if(argc > 6)
      if( strcmp(argv[6], "-") )
	output_filename = argv[6];

    std::string select_query;

    if(argc > 7)
      select_query = argv[7];

    if(select_query=="-")
      select_query = "";

    std::string iterator_options;
    if(argc > 8)
      iterator_options = argv[8];
    if(iterator_options == "-")
      iterator_options = "FetchOne";
    if (iterator_options.find("NO_TOKEN") != std::string::npos)
    {
      if (whatToDo) whatToDo = 4;
    }

    std::string benchmark_options = "attribute_list_constructor";
    if(argc > 9)
      if( strcmp(argv[9], "-") ) benchmark_options = argv[9];

    bool BENCH_attribute_list_constructor = (benchmark_options.find(
      "attribute_list_constructor") != std::string::npos);

    bool BENCH_pause = (benchmark_options.find("pause") != std::string::npos);

    try
    {
      const std::string platformDependentName =
	seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->addReadCatalog( p.contactstring() );
    fileCatalog->connect();
    fileCatalog->start();

    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );

    pool::DatabaseConnectionPolicy policy;
    policy.setReadMode( pool::DatabaseConnectionPolicy::READ );
    dataSvc->session().setDefaultConnectionPolicy( policy );

    sc = dataSvc->transaction().start( pool::ITransaction::READ );

    if( !sc.isSuccess() )
    {
      log << seal::Msg::Fatal << "Error when starting database transaction." 
	<< seal::endmsg;
      return 1;
    }

    ////////////////////////////////////////////////////////////////////////

    std::cout << std::endl;
    log << seal::Msg::Info << "Collection_Read: Opening a collection object "   
      << "for READ" << seal::endmsg;
    log << seal::Msg::Info << "Collection type is " << collectionType 
      << seal::endmsg;
    log << seal::Msg::Info << "Collection location is " << collectionDb  
      << seal::endmsg;
    log << seal::Msg::Info << "Collection name is " << collectionName 
      << seal::endmsg;

    pool::CollectionProxy collection( dataSvc, collectionType,
      collectionDb, collectionName, pool::ICollection::READ );
    std::cout << std::endl;

    pool::AttributeListSpecification& attribSpec = 
      collection.attributeListSpecification();

    if( attr_number != attribSpec.size() )
    {
      log << seal::Msg::Warning << "Number of attributes in the collection is "
	<< "not the same as given at command line" << seal::endmsg;
      log << seal::Msg::Warning << "Collection has " << attribSpec.size() 
	<< seal::endmsg;
    }

    if( msgLevel < seal::Msg::Debug )
    {
      std::cout << "[OVAL] ";
      attribSpec.print( std::cout );
      std::cout << std::endl;
    }

    log << seal::Msg::Info << "Selection query: "
      << (select_query=="" ? "<empty>" : select_query) << seal::endmsg;

    if(BENCH_attribute_list_constructor)
      log << seal::Msg::Info << "Benchmark option: attribute_list_constructor " 
	<< seal::endmsg;

    if(BENCH_pause)
    {
      log << seal::Msg::Info << "Benchmark option: pause " << seal::endmsg;
      log << seal::Msg::Info << "% Press <Enter> to continue... " 
	<< seal::endmsg;
      std::cin.get();
      log << seal::Msg::Info << "Continuing..."  << seal::endmsg;
    }

    log << seal::Msg::Info << "Performing selection query."  << seal::endmsg;

    pool::ProcessTime t1;

    pool::CollectionProxy::Iterator collIter =
      collection.select( select_query, "", iterator_options );

    pool::ProcessTime t2;

    log << seal::Msg::Info << "Looping over selected events." << seal::endmsg;

    int col_cnt = 0;
    pool::AttributeList attributeList;
    while ( collIter.next() ) 
    {
      ++col_cnt;

      if(whatToDo == 0)
	continue;    
      else if (whatToDo == 1)
	collIter.isCurrentObjectOfType<SimpleTestClass>();
      else if (whatToDo == 2)
	collIter.ref<SimpleTestClass>();
      else if (whatToDo == 3)
	collIter.ref<SimpleTestClass>()->setNonZero();
      else if (whatToDo == 4)
	attributeList = collIter.attributeList();
      else
	std::cout << collIter.ref<SimpleTestClass>().toString() << std::endl;

      if( msgLevel < seal::Msg::Debug )
      {
        if (iterator_options.find("NoToken") == std::string::npos)
        {
	  std::cout << "[OVAL] The data is: ";
	  collIter.ref<SimpleTestClass>()->streamOut( std::cout );
	  std::cout << std::endl;
        }
      }

      if(BENCH_attribute_list_constructor)
      {
	if( msgLevel < seal::Msg::Debug )
	{
	  std::cout << "[OVAL] ";
	  attributeList.print( std::cout );
	  std::cout << std::endl;
	}
      }
    }

    pool::ProcessTime t3;

    std::cout << "Total number of events: " << event_number << std::endl;
    std::cout << "Total Number of attributes: " << attr_number << std::endl;
    std::cout << "Number of events satisfying query: " << col_cnt 
      << std::endl;

    if(whatToDo == 0)
      std::cout << "Performed empty loop." << std::endl;
    else if (whatToDo == 1)
      std::cout << "Performed type check." << std::endl;
    else if (whatToDo == 2)
      std::cout << "Performed Ref<T> creation." << std::endl;
    else if (whatToDo == 3)
      std::cout << "Performed Ref<T> dereferencing." << std::endl;
    else if (whatToDo == 4)
      std::cout << "Performed attribute list read." << std::endl;
    else
      std::cout << "Performed ref<T> stringification." << std::endl;

    std::cout << "Iterator creation time: "; 
    (t2-t1).printPrettyOn( std::cout ); 
    std::cout << std::endl;
    std::cout << "Loop iteration time: "; 
    (t3-t2).printPrettyOn( std::cout ); 
    std::cout << std::endl;
    std::cout << "Total time: "; 
    (t3-t1).printPrettyOn( std::cout ); 
    std::cout << std::endl;

    std::ofstream of(output_filename.c_str(),std::ios::out|std::ios::app);

    of << "Total number of events: " << event_number << std::endl;
    of << "Total number of attributes: " << attr_number << std::endl;
    of << "Number of events satisfying query: " << col_cnt << std::endl; 

    if(whatToDo == 0)
      of << "Performed empty loop." << std::endl;
    else if (whatToDo == 1)
      of << "Performed type check." << std::endl;
    else if (whatToDo == 2)
      of << "Performed Ref<T> creation." << std::endl;
    else if (whatToDo == 3)
      of << "Performed Ref<T> dereferencing." << std::endl;
    else if (whatToDo == 4)
      of << "Performed attribute list read." << std::endl;
    else
      of << "Performed ref<T> stringification." << std::endl;

    of << "Iterator creation time: "; 
    (t2-t1).printPrettyOn( of ); 
    of << std::endl;
    of << "Loop iteration time: "; 
    (t3-t2).printPrettyOn( of ); 
    of << std::endl;
    of << "Total time: "; 
    (t3-t1).printPrettyOn( of ); 
    of << std::endl;

    of << std::endl;

    of.close();

    collection.close();

    sc = dataSvc->transaction().commit();
    if( !sc.isSuccess() )
    {
      log << seal::Msg::Info << "Collection_Read: Error when ending database "
	<< "transaction." << seal::endmsg;
    }

    delete dataSvc;

    fileCatalog->disconnect();
    delete fileCatalog;

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
