#include "Collection/Collection.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"

#include "SealKernel/Status.h"
#include "SealBase/SharedLibrary.h"

#include <fstream>
#include <stdexcept>

const std::string OUT_FILE = "CollectionFilesInfo.log";
const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string COLLECTION_NAME = "TestCollection";

int main(int argc, const char *argv[])
{
  try
  {
    seal::Status sc = seal::Status::SUCCESS;
    std::string collectionDb = "";
    std::string query = "";

    if( argc < 2 )
    {
      std::cerr << "Collection_FileInfoRetrieve: Incorrect usage. Must "
       << "specify collection type as command line argument." << std::endl;
      return 1;
    }

    std::string collectionType = std::string( argv[1] );
    if( collectionType == "RootCollection" )
    {
      collectionDb = "";
      query = "run < 5";
    }
    else if( collectionType == "MySQLCollection" || collectionType ==
      "MySQLltCollection" )
    {
      collectionDb = 
        "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression";
      query = "run < 5";
    }
    else if ( collectionType == "RelationalCollectionMySQL" )
    {
      collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
    }
    else if ( collectionType == "RelationalCollectionOracle" )
    {
      collectionDb = "oracle://devdb10.cern.ch/ral_writer";
    }
    else
    {
      std::cerr << "Collection_FileInfoRetrieve: Collection type " 
        << collectionType << " is not presently handled by this program." 
          << std::endl;
      return 1;
    }

    std::ofstream outFileStrm;
    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Opening output file: "
      << OUT_FILE << std::endl;
    outFileStrm.open( OUT_FILE.c_str() );
    if( outFileStrm.bad() )
    {
      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Could not open file: " 
	<< OUT_FILE << std::endl;
    }

    std::cout << "[OVAL] Collection_FileInfoRetrieve: Initialising the "
      << "SEAL services." << std::endl;
    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
    pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Loading the shared "
      << "libraries." << std::endl;
    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname(TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Creating a file catalog."
      << std::endl;

    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->addReadCatalog( p.contactstring() );
    
    if ( ! fileCatalog ) 
      throw std::runtime_error( "Could not create a file catalog." );

    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Establishing a "
      << "connection to the file catalog." << std::endl;
    fileCatalog->connect();

    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Starting the file "
      << "catalog transaction." << std::endl;
    fileCatalog->start();
    
    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Creating a data service " 
      << "instance."<< std::endl;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );
    if ( dataSvc ) {
      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Setting the database "
	<< "connection policy for this session." << std::endl;
      pool::DatabaseConnectionPolicy policy;
      policy.setReadMode( pool::DatabaseConnectionPolicy::READ );
      dataSvc->session().setDefaultConnectionPolicy( policy );
      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Read mode for a "
	<< "database is set to " << policy.readMode() << std::endl;

      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Starting a database "
      << "transaction for READ." << std::endl;
      sc = dataSvc->transaction().start( pool::ITransaction::READ );
      if( !sc.isSuccess() )
      {
	std::cout << "[OVAL] Collecton_FileInfoRetrieve: Error when starting "
	  << "database transaction." << std::endl;
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Opening a collection "
	<< "object for READ: " << std::endl;
      std::cout << "[OVAL] Collection type is " << collectionType 
        << std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
        << std::endl;
      std::cout << "[OVAL] Collection name is " << COLLECTION_NAME 
        << std::endl;
      pool::Collection< SimpleTestClass > collection( dataSvc, collectionType,
        collectionDb, COLLECTION_NAME, pool::ICollection::READ );
      std::cout << std::endl;

      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Retrieving the "
        << "attribute list specification." << std::endl;
      const pool::AttributeListSpecification attribSpec = 
	collection.attributeListSpecification();
      std::cout << "[OVAL] ";
      attribSpec.print(std::cout);
      std::cout << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Creating a collection "
        << "iterator for server side query: " << query << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter = 
        collection.select( query, "", "FetchOne" );
      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Looping over all " 
	<< "collection elements." << std::endl;

      while ( collIter.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter->streamOut( std::cout );
	std::cout << std::endl;

	outFileStrm << "The data is: ";
	collIter->streamOut( outFileStrm );
	outFileStrm << std::endl;

	pool::AttributeList attributeList = collIter.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print( std::cout );
	std::cout << std::endl;

	attributeList.print( outFileStrm );
	outFileStrm << std::endl;

	std::string dbFileId = collIter.ref().token()->dbID();
	std::cout << "Database file ID is " << dbFileId << std::endl;
	outFileStrm << "Database file ID is " << dbFileId << std::endl; 

	std::string bestPFN;
	std::string fileType;
  
        pool::IFCAction action;
        fileCatalog->setAction( action );  
	action.lookupBestPFN( dbFileId, pool::FileCatalog::READ, 
                         pool::FileCatalog::SEQUENTIAL, bestPFN, fileType );
  
	std::cout << "[OVAL] Physical file name is " << bestPFN
	  << std::endl;
	std::cout << "[OVAL] File type is " << fileType << std::endl;
	outFileStrm << "Physical file name is " << bestPFN << std::endl;
	outFileStrm << "File type is " << fileType << std::endl;

	std::string contId = collIter.ref().token()->contID();
	std::cout << "[OVAL] Container ID is " << contId << std::endl;
	outFileStrm << "Container ID is " << contId << std::endl;

	long tech = collIter.ref().token()->technology();
	std::cout << "[OVAL] Technology is " << tech << std::endl;
	outFileStrm << "Technology is " << tech << std::endl;

	outFileStrm << std::endl;
      }

      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Closing the "
        << "collection." << std::endl;
      collection.close();

      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Ending the datbase "
		<< "transaction." << std::endl;
      sc = dataSvc->transaction().commit();
      if( !sc.isSuccess() )
	{
	  std::cout << "[OVAL] Collecton_FileInfoRetrieve: Error when ending "
		    << "database transaction." << std::endl;
	}

      std::cout << "[OVAL] Collecton_FileInfoRetrieve: Disconnecting from the "
		<< "database." << std::endl;
    }

    delete dataSvc;

    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Disconnecting from the "
      << "file catalog." << std::endl;
    fileCatalog->disconnect();
    delete fileCatalog;

    std::cout << "[OVAL] Collecton_FileInfoRetrieve: Returning from main()" 
      << std::endl;
    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
