#include "Collection/Collection.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"

#include "SealKernel/Status.h"
#include "SealBase/SharedLibrary.h"

#include <stdexcept>

const int NUM_FILES = 3;
const int NUM_RUNS_PER_FILE = 2;
const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string DB_FILE_PREFIX = "TestDbFile";
const std::string DB_FILE_SUFFIX = "pool";
const pool::DbType DB_TYPE = pool::ROOTTREE_StorageType;
const std::string CONTAINER_NAME_PREFIX = "TestContainer";
const std::string COLLECTION_NAME = "TestCollection";

int main(int argc, const char *argv[])
{
  try
  {
    seal::Status sc = seal::Status::SUCCESS;
    char fileNum[2];
    std::string collectionDb = "";

    if( argc < 2 )
    {
      std::cerr << "Collection_MultiFileWrite: Incorrect usage. Must specify "
	<< " collection type as command line argument." << std::endl;
      return 1;
    }

    std::string collectionType = std::string( argv[1] );
    if( collectionType == "RootCollection" )
    {
      collectionDb = "";
    }
    else if( collectionType == "MySQLCollection" || collectionType ==
      "MySQLltCollection" )
    {
      collectionDb = 
        "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression";
    }
    else if ( collectionType == "RelationalCollectionMySQL" )
    {
      collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
    }
    else if ( collectionType == "RelationalCollectionOracle" )
    {
      collectionDb = "oracle://devdb10.cern.ch/ral_writer";
    }
    else
    {
      std::cerr << "Collection_MultiFileWrite: Collection type " 
	<< collectionType << " is not presently handled by this program." 
          << std::endl;
      return 1;
    }

    std::cout << "[OVAL] Collection_MultiFileWrite: Initialising the SEAL"
      << "services." << std::endl;
    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
    pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

    std::cout << "[OVAL] Collection_MultiFileWrite: Loading the shared "
      << "libraries." << std::endl; 
    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    std::cout << "[OVAL] Collection_MultiFileWrite: Creating a file catalog: " 
      << std::endl;
    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->setWriteCatalog( p.contactstring() );

    if ( ! fileCatalog ) 
      throw std::runtime_error( "Could not create a file catalog." );

    std::cout << "[OVAL] Collection_MultiFileWrite: Establishing a connection "
      << "to the file catalog." << std::endl;
    fileCatalog->connect();

    std::cout << "[OVAL] Collection_MultiFileWrite: Starting the file catalog "
      << "transaction." << std::endl;
    fileCatalog->start();

    std::cout << "[OVAL] Collection_MultiFileWrite: Creating a data service "
      << "instance." << std::endl;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );

    std::cout << "[OVAL] Collection_MultiFileWrite: Setting the database "
      << "connection policy for this session." << std::endl;
    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
    policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
    dataSvc->session().setDefaultConnectionPolicy( policy );
    std::cout << "[OVAL] Collection_MultiFileWrite: Write mode for a "
      << "non-existing database is set to: " 
        << policy.writeModeForNonExisting() << std::endl;
    std::cout << "[OVAL] Collection_MultiFileWrite: Write mode for an "
      << "existing database is set to: " 
        << policy.writeModeForExisting() << std::endl;

    std::vector<std::string> dbFile(NUM_FILES);
    std::cout << "[OVAL] Collection_MultiFileWrite: Declaring " << NUM_FILES
      << " database files." << std::endl;
    for( int i=0; i<NUM_FILES; i++ )
    {
      std::sprintf(fileNum, "%i", i+1);
      dbFile[i] = DB_FILE_PREFIX;
      dbFile[i] += fileNum;
      dbFile[i] += ".";
      dbFile[i] += DB_FILE_SUFFIX;
    }

    std::vector<pool::Placement> placement(NUM_FILES);
    std::string contName;
    for( int i=0; i<NUM_FILES; i++ )
    {
      std::cout << "[OVAL] Collection_MultiFileWrite: Creating a Placement "
	<< "object for file " << dbFile[i] << std::endl;
      placement[i].setDatabase( dbFile[i], pool::DatabaseSpecification::PFN );
      std::sprintf(fileNum, "%i", i+1);
      contName = CONTAINER_NAME_PREFIX;
      contName += fileNum;
      placement[i].setContainerName( contName );
      placement[i].setTechnology( DB_TYPE.type() );
    }

    std::cout << "[OVAL] Collection_MultiFileWrite: Starting a database "
      << "transaction for UPDATE." << std::endl;
    sc = dataSvc->transaction().start( pool::ITransaction::UPDATE );
    if( !sc.isSuccess() )
    {
      std::cout << "[OVAL] Collection_MultiFileWrite: Error when starting "
	<< "database transaction." << std::endl;
    }

    std::cout << std::endl;
    std::cout << "[OVAL] Collection_MultiFileWrite: Creating a collection "
      << "object." << std::endl;
    std::cout << "[OVAL] Collection type is " << collectionType << std::endl;
    std::cout << "[OVAL] Collection location is " << collectionDb << std::endl;
    std::cout << "[OVAL] Collection name is " << COLLECTION_NAME << std::endl;
    pool::Collection< SimpleTestClass > collection( dataSvc, collectionType,
      collectionDb, COLLECTION_NAME, pool::ICollection::CREATE_AND_OVERWRITE );
    std::cout << std::endl;

    std::cout << "[OVAL] Collection_MultiFileWrite: Defining the attribute "
      << "list specification." << std::endl;
    pool::AttributeListSpecification& attribSpec = 
      collection.attributeListSpecification();
    attribSpec.push_back( "run", "int" );
    attribSpec.push_back( "event", "int" );
    std::cout << "[OVAL] ";
    attribSpec.print(std::cout);
    std::cout << std::endl;

    std::cout << std::endl;
    std::cout << "[OVAL] Collection_MultiFileWrite: Adding objects and "
      "attributes to the collection:" << std::endl;
    std::cout << std::endl;
    int runCounter = 1;
    for( int i = 0; i<NUM_FILES; i++ )
    {
      std::cout << "Writing to file " << dbFile[i] << std::endl;
      std::cout << std::endl;
      for (int run = runCounter; run < runCounter+NUM_RUNS_PER_FILE; run++)
      {
	for (int evt = 1; evt <= 5; evt++)
	{
	  std::cout << "[OVAL] Creating a SimpleTestClass object." 
            << std::endl;
	  pool::Ref<SimpleTestClass> 
	    simpleTestClass( dataSvc, new SimpleTestClass() );

	  std::cout << "[OVAL] Marking the object for writing." << std::endl;
	  simpleTestClass.markWrite( placement[i] );

	  std::cout << "[OVAL] Setting the object's data." << std::endl;
	  simpleTestClass->setNonZero();

	  std::cout << "[OVAL] The data is: ";
	  simpleTestClass->streamOut(std::cout);
	  std::cout << std::endl;

	  std::cout << "[OVAL] Setting the object's attributes." << std::endl;
	  pool::AttributeList attributeList(attribSpec);
	  attributeList[ "run" ].setValue( run );
	  attributeList[ "event" ].setValue( evt );

	  std::cout << "[OVAL] ";
	  attributeList.print(std::cout);
	  std::cout << std::endl;

	  std::cout << "[OVAL] Adding object and attributes to collection." 
	    << std::endl; 
	  collection.add( simpleTestClass, attributeList );

	  std::cout << std::endl;
	}
      }
      runCounter += 2;
    }

    std::cout << "[OVAL] Collection_MultiFileWrite: Committing all changes "
      << "made to the collection." << std::endl;
    collection.commit();

    std::cout << "[OVAL] Collection_MultiFileWrite: Closing the collection."
      << std::endl;
    collection.close();

    std::cout << "[OVAL] Collection_MultiFileWrite: Ending the database "
      << "transaction." << std::endl;
    sc = dataSvc->transaction().commit();
    if( !sc.isSuccess() )
    {
      std::cout << "[OVAL] Collection_MultiFileWrite: Error when ending "
	<< "database transaction." << std::endl;
    }

    std::cout << "[OVAL] Collection_MultiFileWrite: Committing all file "
      << "catalog changes." << std::endl;
    fileCatalog->commit();

    std::cout << "[OVAL] Collection_MultiFileWrite: Disconnecting from the "
      << "database." << std::endl;
    delete dataSvc;

    std::cout << "[OVAL] Collection_MultiFileWrite: Disconnecting from the "
      << "file catalog." << std::endl;
    fileCatalog->disconnect();
    delete fileCatalog;

    std::cout << "[OVAL] Collection_MultiFileWrite: Returning from main()"
      << std::endl;

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
