#include "Collection/Collection.h"
#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeListSpecification.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/Exception.h"
#include "TestDictionary/SimpleTestClass.h"
#include "TestDictionary/TestClassPrimitives.h"

#include "SealKernel/Status.h"
#include "SealBase/SharedLibrary.h"

#include <stdexcept>

const std::string TEST_LIBRARY = "TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string DB_FILE = "TestDbFile.pool";
const pool::DbType DB_TYPE = pool::ROOTTREE_StorageType;
const std::string CONTAINER_NAME = "TestContainer";
const std::string COLLECTION_NAME = "TestCollection1";
const unsigned int NUM_REF_ATTRIBUTES = 3;
const unsigned int NUM_RUNS = 2;
const unsigned int NUM_EVENTS = 5;

int main(int argc, const char *argv[])
{
  try
  { 
    seal::Status sc = seal::Status::SUCCESS;
    unsigned int runCounter = 0;
    std::string collectionDb = "";

    if( argc < 2 )
    {
      std::cerr << "Collection_Update: Incorrect usage. Must specify "
	<< " collection type as command line argument." << std::endl;
      return 1;
    }

    std::string collectionType = std::string( argv[1] );
    if( collectionType == "RootCollection" )
    {
      collectionDb = "";
    }
    else if( collectionType == "MySQLCollection" || collectionType ==
      "MySQLltCollection" )
    {
      collectionDb = 
        "mysql://regrTester:rT2003@lxshare070d.cern.ch/regression";
    }
    else if ( collectionType == "RelationalCollectionMySQL" )
    {
      collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
    }
    else if ( collectionType == "RelationalCollectionOracle" )
    {
      collectionDb = "oracle://devdb10.cern.ch/ral_writer";
    }
    else
    {
      std::cerr << "Collection_Update: Collection type " << collectionType
	<< " is not presently handled by this program." << std::endl;
      return 1;
    }

    std::cout << "[OVAL] Collection_Update: Initialising the SEAL services."
      << std::endl;
    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
    pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

    std::cout << "[OVAL] Collection_Update: Loading the shared libraries."
      << std::endl;
    try
    {
      const std::string platformDependentName =
        seal::SharedLibrary::libname( TEST_LIBRARY );
      seal::SharedLibrary::load( platformDependentName );
    }
    catch ( seal::SharedLibraryError* error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    std::cout << "[OVAL] Collection_Update: Creating a file catalog." 
      << std::endl;
    pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
    pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
    p.parse();
    fileCatalog->setWriteCatalog( p.contactstring() );
    fileCatalog->addReadCatalog( p.contactstring() );

    std::cout << "[OVAL] Collection_Update: Establishing a connection to "
      << "the file catalog." << std::endl;
    fileCatalog->connect();

    std::cout << "[OVAL] Collection_Update: Starting the file catalog "
      << "transaction." << std::endl;
    fileCatalog->start();

    std::cout << "[OVAL] Collection_Update: Creating a data service instance." 
      << std::endl;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );
    if ( dataSvc ) 
    {
      std::cout << "[OVAL] Collection_Update: Setting the database connection "
        << "policy for this session." << std::endl;
      pool::DatabaseConnectionPolicy policy;
      policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::UPDATE );
      policy.setReadMode( pool::DatabaseConnectionPolicy::UPDATE );
      dataSvc->session().setDefaultConnectionPolicy( policy );
      std::cout << "[OVAL] Collection_Update: Write mode for an existing "
        << "database is set to: " << policy.writeModeForExisting() 
          << std::endl;
      std::cout << "[OVAL] Collection_Update: Read mode for a database is "
        << "set to " << policy.readMode() << std::endl;

      std::cout << "[OVAL] Collection_Update: Creating a Placement object." 
        << std::endl;
      pool::Placement placement;
      placement.setDatabase( DB_FILE, pool::DatabaseSpecification::PFN );
      placement.setContainerName( CONTAINER_NAME );
      placement.setTechnology( DB_TYPE.type() );

      char attrNum[128];

      std::vector<std::string> refAttrName(NUM_REF_ATTRIBUTES);
      for(unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
      {
        std::sprintf(attrNum, "%i", i+1);
        refAttrName[i] = "test_ref_";
        refAttrName[i] += attrNum;
      }

      std::cout << "[OVAL] Collection_Update: Creating Placement objects" 
        << " to store the Ref attributes." << std::endl;
      std::vector<pool::Placement> placeRefAttr(NUM_REF_ATTRIBUTES);
      std::string contName;
      for(unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
      {
        std::sprintf(attrNum, "%i", i+1);
        placeRefAttr[i].setDatabase( DB_FILE, pool::DatabaseSpecification::PFN );
        contName = CONTAINER_NAME;
        contName += "Ref";
        contName += attrNum;
        placeRefAttr[i].setContainerName( contName );
        placeRefAttr[i].setTechnology( DB_TYPE.type() );
      }

      std::cout << "[OVAL] Collection_Update: Starting a database transaction "
		<< "UPDATE." << std::endl;
      sc = dataSvc->transaction().start( pool::ITransaction::UPDATE );
      if( !sc.isSuccess() )
      {
        std::cout << "[OVAL] Collection_Update: Error when starting "
          << "database transaction." << std::endl;
      }

      std::cout << "[OVAL] Collection_Update: Creating primitive test "
        << "attributes." << std::endl;
      TestClassPrimitives tcPrim;
      tcPrim.setNonZero();

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Update: Opening a collection object for "
        << "UPDATE." << std::endl;
      std::cout << "[OVAL] Collection type is " << collectionType
        << std::endl;
      std::cout << "[OVAL] Collection location is " << collectionDb 
        << std::endl;  
      std::cout << "[OVAL] Collection name is " << COLLECTION_NAME 
        << std::endl; 
      pool::Collection< SimpleTestClass > collection( dataSvc, collectionType,
	collectionDb, COLLECTION_NAME, pool::ICollection::UPDATE ); 
      std::cout << std::endl;

      std::cout << "[OVAL] Collection_Update: Retrieving the attribute list "
	<< "specification." << std::endl;
      const pool::AttributeListSpecification& attribSpec = 
        collection.attributeListSpecification();
      std::cout << "[OVAL] ";
      attribSpec.print( std::cout );
      std::cout << std::endl;

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Update: Creating a collection iterator." 
        << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter1 = 
        collection.select();

      std::cout << "[OVAL] Collection_Update: Looping over all existing "
	<< "collection elements:" << std::endl;
      std::cout << std::endl;

      while ( collIter1.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter1->streamOut( std::cout );
	std::cout << std::endl;

	pool::AttributeList attributeList = collIter1.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print(std::cout);
	std::cout << std::endl;

	attributeList["run"].getValue(runCounter);
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Update: Adding more objects and "
        << "attributes to the collection." << std::endl;
      std::cout << std::endl;
      for (unsigned int run = runCounter+1; run <= runCounter+NUM_RUNS; run++)
      {
	for (unsigned int evt = 1; evt <= NUM_EVENTS; evt++)
	{
	  std::cout << "[OVAL] Creating a SimpleTestClass object." 
            << std::endl;
	  pool::Ref<SimpleTestClass> 
          simpleTestClass( dataSvc, new SimpleTestClass() );

	  std::cout << "[OVAL] Marking the object for writing." << std::endl;
	  simpleTestClass.markWrite( placement );

	  std::cout << "[OVAL] Setting the object's data." << std::endl;
	  simpleTestClass->setNonZero();

	  std::cout << "[OVAL] The data is: ";
	  simpleTestClass->streamOut(std::cout);
	  std::cout << std::endl;

	  std::cout << "[OVAL] Setting the object's attributes." << std::endl;
	  pool::AttributeList attributeList(attribSpec);
	  attributeList[ "run" ].setValue( run );
	  attributeList[ "event" ].setValue( evt );
          attributeList[ "test_bool" ].setValue( tcPrim.m_bool );
          attributeList[ "test_short" ].setValue( tcPrim.m_short );
          attributeList[ "test_ushort" ].setValue( tcPrim.m_ushort );
	  attributeList[ "test_sshort" ].setValue( tcPrim.m_sshort );
	  attributeList[ "test_shortint" ].setValue( tcPrim.m_shortint );
	  attributeList[ "test_ushortint" ].setValue( tcPrim.m_ushortint );
	  attributeList[ "test_sshortint" ].setValue( tcPrim.m_sshortint );
	  attributeList[ "test_int" ].setValue( tcPrim.m_int );
	  attributeList[ "test_uint" ].setValue( tcPrim.m_uint );
	  attributeList[ "test_sint" ].setValue( tcPrim.m_sint );
	  attributeList[ "test_s" ].setValue( tcPrim.m_s );
	  attributeList[ "test_u" ].setValue( tcPrim.m_u );
	  attributeList[ "test_long" ].setValue( tcPrim.m_long );
	  attributeList[ "test_ulong" ].setValue( tcPrim.m_ulong );
	  attributeList[ "test_slong" ].setValue( tcPrim.m_slong );
	  attributeList[ "test_longint" ].setValue( tcPrim.m_longint );
	  attributeList[ "test_ulongint" ].setValue( tcPrim.m_ulongint );
	  attributeList[ "test_slongint" ].setValue( tcPrim.m_slongint );
	  attributeList[ "test_float" ].setValue( tcPrim.m_float );
	  attributeList[ "test_double" ].setValue( tcPrim.m_double );
	  attributeList[ "test_string" ].setValue(std::string("Test String"));
          for (unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
          {
   	    pool::Ref<SimpleTestClass> 
	      refAttr( dataSvc, new SimpleTestClass() );
            refAttr.markWrite( placeRefAttr[i] );
            refAttr->setNonZero();
            std::cout << "[OVAL] Ref attribute " << refAttrName[i] 
              << " has its data set to: ";
            refAttr->streamOut(std::cout);
            std::cout << std::endl;
            attributeList[refAttrName[i]].setValue( refAttr.toString() );
          }

	  std::cout << "[OVAL] ";
	  attributeList.print(std::cout);
	  std::cout << std::endl;

	  std::cout << "[OVAL]: Adding object and attributes to collection."
	    << std::endl; 
	  collection.add(simpleTestClass, attributeList);

	  std::cout << std::endl;
	}
      }

      std::cout << "[OVAL] Collection_Update: Rolling back last changes made "
        << "to the collection." << std::endl;
      collection.rollback();

      std::cout << "[OVAL] Collection_Update: Committing all changes made to "
        << "the collection." << std::endl;
      collection.commit();

      std::cout << "[OVAL] Collection_Update: Testing temporary closing of "
        << "collection." << std::endl;
      collection.close();
      std::cout << "[OVAL] Collection_Update: Collection is open? " 
        << collection.isOpen() << std::endl;

      std::cout << "[OVAL] Collection_Update: Testing reopening of collection."
	<< std::endl;
      collection.open();
      std::cout << "[OVAL] Collection_Update: Collection is open? "
	<< collection.isOpen() << std::endl;

      std::cout << "[OVAL] Collection_Update: Creating another collection "
	<< "iterator." << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter2 =  
        collection.select();
      std::cout << "[OVAL] Collection_Update: Looping over all collection "
        << "elements to verify that the new elements were erased:" 
          << std::endl;
      std::cout << std::endl;

      while ( collIter2.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter2->streamOut( std::cout );
	std::cout << std::endl;

	pool::AttributeList attributeList = collIter2.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print(std::cout);
	std::cout << std::endl;

	attributeList["run"].getValue(runCounter);
      }

      std::cout << std::endl;
      std::cout << "[OVAL] Collection_Update: Adding more objects and "
        << "attributes to the collection again:" << std::endl;
      std::cout << std::endl;
      for (unsigned int run = runCounter+1; run <= runCounter+NUM_RUNS; run++)
      {
	for (unsigned int evt = 1; evt <= NUM_EVENTS; evt++)
	{
	  std::cout << "[OVAL] Creating a SimpleTestClass object." 
	    << std::endl;
	  pool::Ref<SimpleTestClass> 
	  simpleTestClass( dataSvc, new SimpleTestClass() );

	  std::cout << "[OVAL] Marking the object for writing."
	    << std::endl;
	  simpleTestClass.markWrite( placement );

	  std::cout << "[OVAL] Setting the object's data." << std::endl;
	  simpleTestClass->setNonZero();

	  std::cout << "[OVAL] The data is: ";
	  simpleTestClass->streamOut(std::cout);
	  std::cout << std::endl;

	  std::cout << "[OVAL] Setting the object's attributes."
	    << std::endl;
	  pool::AttributeList attributeList(attribSpec);
	  attributeList[ "run" ].setValue( run );
	  attributeList[ "event" ].setValue( evt );
          attributeList[ "test_bool" ].setValue( tcPrim.m_bool );
          attributeList[ "test_short" ].setValue( tcPrim.m_short );
          attributeList[ "test_ushort" ].setValue( tcPrim.m_ushort );
	  attributeList[ "test_sshort" ].setValue( tcPrim.m_sshort );
	  attributeList[ "test_shortint" ].setValue( tcPrim.m_shortint );
	  attributeList[ "test_ushortint" ].setValue( tcPrim.m_ushortint );
	  attributeList[ "test_sshortint" ].setValue( tcPrim.m_sshortint );
	  attributeList[ "test_int" ].setValue( tcPrim.m_int );
	  attributeList[ "test_uint" ].setValue( tcPrim.m_uint );
	  attributeList[ "test_sint" ].setValue( tcPrim.m_sint );
	  attributeList[ "test_s" ].setValue( tcPrim.m_s );
	  attributeList[ "test_u" ].setValue( tcPrim.m_u );
	  attributeList[ "test_long" ].setValue( tcPrim.m_long );
	  attributeList[ "test_ulong" ].setValue( tcPrim.m_ulong );
	  attributeList[ "test_slong" ].setValue( tcPrim.m_slong );
	  attributeList[ "test_longint" ].setValue( tcPrim.m_longint );
	  attributeList[ "test_ulongint" ].setValue( tcPrim.m_ulongint );
	  attributeList[ "test_slongint" ].setValue( tcPrim.m_slongint );
	  attributeList[ "test_float" ].setValue( tcPrim.m_float );
	  attributeList[ "test_double" ].setValue( tcPrim.m_double );
	  attributeList[ "test_string" ].setValue(std::string("Test String"));
          for (unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
          {
   	    pool::Ref<SimpleTestClass> 
	      refAttr( dataSvc, new SimpleTestClass() );
            refAttr.markWrite( placeRefAttr[i] );
            refAttr->setNonZero();
            std::cout << "[OVAL] Ref attribute " << refAttrName[i] 
              << " has its data set to ";
            refAttr->streamOut(std::cout);
            std::cout << std::endl;
            attributeList[refAttrName[i]].setValue( refAttr.toString() );
          }

	  std::cout << "[OVAL] ";
	  attributeList.print(std::cout);
	  std::cout << std::endl;

	  std::cout << "[OVAL] Adding object and attributes to collection."
	    << std::endl; 
	  collection.add(simpleTestClass, attributeList);

	  std::cout << std::endl;
	}
      }

      std::cout << "[OVAL] Collection_Update: Committing all changes made to "
        << "the collection." << std::endl;
      collection.commit();

      std::cout << "[OVAL] Collection_Update: Trying to roll back last "
        << "changes made to collection after they were commited." << std::endl;
      collection.rollback();

      std::cout << "[OVAL] Collection_Update: Creating another collection "
        << "iterator." << std::endl;
      pool::Collection< SimpleTestClass >::Iterator collIter3 =  
	collection.select();
      std::cout << "[OVAL] Collection_Update: Looping over all collection "
	<< "elements to verify that the new elements were not rolled back:" 
          << std::endl;
      std::cout << std::endl;

      while ( collIter3.next() ) 
      {
	std::cout << "[OVAL] New element found." << std::endl;

	std::cout << "[OVAL] The data is: ";
	collIter3->streamOut( std::cout );
	std::cout << std::endl;

	pool::AttributeList attributeList = collIter3.attributeList();
	std::cout << "[OVAL] ";
	attributeList.print(std::cout);
	std::cout << std::endl;
      }

      std::cout << "[OVAL] Collection_Update: Closing the collection." 
        << std::endl;
      collection.close();

      std::cout << "[OVAL] Collection_Update: Ending the database transaction "
        << std::endl;
      sc = dataSvc->transaction().commit();
      if( !sc.isSuccess() )
      {
	std::cout << "[OVAL] Collection_Update: Error when ending database "
          << "transaction." << std::endl;
      }

      std::cout << "[OVAL] Collection_Update: Committing all file catalog "
	<< "changes." << std::endl;
      fileCatalog->commit();

      std::cout << "[OVAL] Collection_Update: Disconnecting from the database."
	<< std::endl;
    }
    delete dataSvc;

    std::cout << "[OVAL] Collection_Update: Disconnecting from the file "
      << "catalog." << std::endl;
    fileCatalog->disconnect();
    delete fileCatalog;

    std::cout << "[OVAL] Collection_Update: Returning from main()" 
      << std::endl;

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception caught in main(). " << std::endl;
    return 1;
  }
}
