#include "POOLCore/Token.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include <stdexcept>
#include <fstream>
#include <vector>
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "TestBase/TestBase.h"

#include "DataModel.h"

//#include "TROOT.h"
//#include "TClass.h"
//#include "TStreamerInfo.h"
//#include "TStreamerElement.h"

namespace pool  {


  class TestDriver : public TestBase  {

    private:
    static const std::string m_dbName1;
    static const std::string m_dbName2;
    static const DbType m_type1;
    static const DbType m_type2;
    char* mode; 
    public: 
    TestDriver(char* aMode) : TestBase("data/fileCatalog.xml"),mode(aMode){

      std::vector< std::string > libraries;
      libraries.push_back( "DataRegression_DataSvc_CMSClassesDict" );
      init( libraries );
      setMessageVerbosityLevel(seal::Msg::Verbose);
    };

    /// write objects to the persistent store
    void write(unsigned int ntimes){

      std::cout << "starting WRITE procedure\n";
      pool::DatabaseConnectionPolicy policy;
      policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
      policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
      m_dataSvc->session().setDefaultConnectionPolicy( policy );

      std::ofstream tokensFile("data/tokens.txt");

      Placement place1, place2;  
      place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
      place1.setContainerName("test1_cont_1");
      place1.setTechnology(m_type1.type());
      place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
      place2.setContainerName("test2_cont_1");
      place2.setTechnology(m_type2.type());

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

      unsigned int i;
      for (i = 0; i < ntimes; ++i )   {

        pool::Ref<ext::ooScopeObj> sc(m_dataSvc,new ext::ooScopeObj(i,"Parent"));
        pool::Ref<ext::ooScopeObj> child1(m_dataSvc,new ext::ooScopeObj(i,"Child1"));
        pool::Ref<ext::ooScopeObj> child2(m_dataSvc,new ext::ooScopeObj(i,"Child2"));
        sc.markWrite(place1);
        tokensFile << sc.toString() << std::endl;
        child1.markWrite(place2);
        child2.markWrite(place2);
        sc->objMap().insert(std::make_pair("Child1",child1));
        sc->objMap().insert(std::make_pair("Child2",child2));

      }

      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }

      std::cout << OVALTAG << "Written "<< i << " objects in the database\n";
      tokensFile.close();
      m_dataSvc->clearCache();
      return;
    }

    /// read objects from the persistent store
    void read() {

      std::cout << "starting READ procedure\n";
      std::vector<std::string> tokens;
      char buff[1024];
      std::ifstream tokensFile("data/tokens.txt",std::ios::in);
      if (tokensFile.bad()) {
        printError("test ERROR: Error opening file for read");
      }

      while(!tokensFile.getline(buff, sizeof(buff)).eof()) {
        std::cout << buff << '\n' ;
        std::string s(buff);
        tokens.push_back(s);
      }
      tokensFile.close();

      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      unsigned int i=0;
      std::vector<std::string>::iterator iter;
      for(iter=tokens.begin();iter!=tokens.end();++iter){

	pool::Ref<ext::ooScopeObj> sc(m_dataSvc,*iter);

        std::map<std::string,Ref<ext::ooScopeObj> >::iterator iobj = sc->objMap().find("Child1");
        if(iobj == sc->objMap().end()){
          printError("test ERROR: child not found!");
        } else {
          if(iobj->second->label()!= "Child1"){
            printError("test ERROR: child1 label different from expected.");
	    std::cout << OVALTAG << "Child1 label = " << iobj->second->label() << std::endl; 
          }
	  if(sc->id() != iobj->second->id()){
            printError("test ERROR: child1 id different from parent id.");
	    std::cout << OVALTAG << "Parent id=" << sc->id() << " Child1 id="<<iobj->second->id() << std::endl;
          }        
        }
        
        iobj = sc->objMap().find("Child2");
        if(iobj == sc->objMap().end()){
          printError("test ERROR: child not found!");
        } else {
          if(iobj->second->label()!= "Child2"){
            printError("test ERROR: child2 label different from expected.");
	    std::cout << OVALTAG << "Child2 label = " << iobj->second->label() << std::endl;
          } 
	  if(sc->id() != iobj->second->id()){
            printError("test ERROR: child2 id different from parent id.");
	    std::cout << OVALTAG << "Parent id=" << sc->id() << " Child2 id="<<iobj->second->id() << std::endl;
          }        
        }

        i++;
      }
         
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      std::cout << "Object read back: "<< i << std::endl;

      printM("Clear cache...");
      m_dataSvc->clearCache();
      return;
    }

    void execute(){
      if(mode){
        if(mode[0] == 'W'){
          write(10);
        } else if(mode[0] == 'R'){
          read();
        }
      }else {
	std::cout << "No mode provided\n";
      }
    }
  };

}

const std::string pool::TestDriver::m_dbName1 = "data/test1.root";
const std::string pool::TestDriver::m_dbName2 = "data/test2.root";
const pool::DbType pool::TestDriver::m_type1 = pool::ROOTTREE_StorageType;
const pool::DbType pool::TestDriver::m_type2 = pool::ROOTTREE_StorageType;

int main(int argn, char** args)     {
  if(argn>1){
    pool::TestDriver theTest(args[1]);
    theTest.run();    
  } else {
    std::cout << "Wrong number of arguments provided\n";
    return 1;
  }
  return 0;
}
