#include "POOLCore/Token.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include <stdexcept>
#include <fstream>
#include <vector>
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "TestBase/TestBase.h"

#include "DataModel.h"
#include "Reflex/Reflex.h"

#include "TROOT.h"
#include "TClass.h"
#include "TStreamerInfo.h"
#include "TStreamerElement.h"

namespace pool  {

  class TestDriver : public TestBase  {

    private:
    static const std::string m_dbName1;
    static const std::string m_dbName2;
    static const DbType m_type1;
    static const DbType m_type2;
    char* mode; 
    public: 
    TestDriver(char* aMode) : TestBase("data/fileCatalog.xml"),mode(aMode){

      std::vector< std::string > libraries;
      libraries.push_back( "DataRegression_DataSvc_SimpleEmbeddedRefsDict" );
      init( libraries );
    };

    /// write objects to the persistent store
    void write(unsigned int ntimes){

      std::cout << "starting WRITE procedure\n";
      pool::DatabaseConnectionPolicy policy;
      policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
      policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
      m_dataSvc->session().setDefaultConnectionPolicy( policy );

      std::ofstream tokensFile("data/tokens.txt");

      Placement place1, place2;  
      place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
      place1.setContainerName("test1_cont_1");
      place1.setTechnology(m_type1.type());
      place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
      place2.setContainerName("test2_cont_1");
      place2.setTechnology(m_type2.type());

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

      unsigned int i;
      for (i = 1; i <= ntimes; ++i )   {

        pool::Ref<ext::Base<int> > o1(m_dataSvc, new ext::B<int>(i));
        const ROOT::Reflex::Type c2 = 
                                 ROOT::Reflex::Type::ByName("ext::B<int>");
        if(!c2){
          printError("test ERROR: a required class is not on the dictionary");
          //return;
        }
        try{
          o1.mark_write(place1);
        } catch (const RefException& e){
	  std::cout << "failed writing o1\n";
          printException(e);        
        }
	pool::Ref<ext::A> o2(m_dataSvc, new ext::A(i));
        try{
          o2.mark_write(place2);
          tokensFile << o2.toString() << std::endl;
        } catch (const RefException& e){
	  std::cout << "failed writing o2\n";
          printException(e);        
        }
        o2->p.second = o1;    

      }
      printM("Commit transactions...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }
      gROOT->GetClass("pair<ext::C,pool::Ref<ext::Base<int> > >")->GetStreamerInfo()->ls();

      std::cout << "Written "<< i-1 << " objects in the database\n";
      tokensFile.close();
      m_dataSvc->clearCache();
      return;
    }

    /// read objects from the persistent store
    void read() {

      std::cout << "starting READ procedure\n";
      std::vector<std::string> tokens;
      char buff[1024];
      std::ifstream tokensFile("data/tokens.txt",std::ios::in);
      if (tokensFile.bad()) {
        printError("test ERROR: Error opening file for read");
      }

      while(!tokensFile.getline(buff, sizeof(buff)).eof()) {
        std::cout << buff << '\n' ;
        std::string s(buff);
        tokens.push_back(s);
      }
      tokensFile.close();

      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      unsigned int i=0;
      std::vector<std::string>::iterator iter;
      for(iter=tokens.begin();iter!=tokens.end();++iter){

        // Read first object
        pool::Ref<ext::A> ref(m_dataSvc,*iter);

        if(ref){
	  if(ref->p.first.d != ref->p.second->obj()){
            printError("test ERROR: data mismatch");
          } else {
	    std::cout << OVALTAG << "b="<< ref->p.first.d << std::endl;
	    std::cout << OVALTAG << "b ref="<< ref->p.second->obj() 
                                            << std::endl;
          }
        } else {
          printError("test ERROR: no ref valid found for token");
        }
        i++;
      }
        
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      std::cout << "Object read back: "<< i << std::endl;

      printM("Clear cache...");
      m_dataSvc->clearCache();
      return;
    }

    void execute(){
      if(mode){
        if(mode[0] == 'W'){
          write(10);
        } else if(mode[0] == 'R'){
          read();
        }
      }else {
	std::cout << "No mode provided\n";
      }
    }
  };

}

const std::string pool::TestDriver::m_dbName1 = "data/test1.root";
const std::string pool::TestDriver::m_dbName2 = "data/test2.root";
const pool::DbType pool::TestDriver::m_type1 = pool::ROOTTREE_StorageType;
const pool::DbType pool::TestDriver::m_type2 = pool::ROOTTREE_StorageType;

int main(int argn, char** args)     {
  if(argn>1){
    pool::TestDriver theTest(args[1]);
    theTest.run();    
  } else {
    std::cout << "Wrong number of arguments provided\n";
  }
  return 0;
}
