#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "CrossReferenceClass.h"
#include "RelatedClass.h"
#include "DataSvc_CrossReference.h"

#ifndef WIN32
# include <unistd.h>
#endif

pool::CrossReferenceTest::CrossReferenceTest(std::string& dbName, const DbType& storage_type)
      :TestBase(),m_dbName(dbName),m_type(storage_type),m_tokens(){
  m_fid = fileId(m_dbName,m_type);
  if(m_fid == ""){
    printError("test ERROR: cannot retrieve file id");
    return;
  }
}

/// write objects to the persistent store
// Do to the cross reference, the write operation cannot be perfomed in two nested transactions -
// (when the first is committed, the corrispondent cache location is resetted)
// This type of use case can be handled only trough a single transaction.
void pool::CrossReferenceTest::write(unsigned int NC){
 
  Placement place1, place2;
  place1.setDatabase(m_dbName,pool::DatabaseSpecification::PFN );
  place1.setContainerName("db_cont_1");
  place1.setTechnology(m_type.type());
  place2.setDatabase(m_dbName,pool::DatabaseSpecification::PFN );
  place2.setContainerName("db_cont_2");
  place2.setTechnology(m_type.type());

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  std::vector<pool::Ref<CrossReferenceClass> > refObj(NC);
  std::vector<pool::Ref<RelatedClass> > refRel(NC);
  std::vector<CrossReferenceClass*> ptrs(NC);
  std::vector<RelatedClass*> rptrs(NC);
  unsigned int i;
  for (i = 0; i < NC; ++i )   {
    char txt1[256], txt2[256];
    ::sprintf(txt1,"Hello [CrossReferenceClass(%d)]", i);
    ::sprintf(txt2,"Hello [RelatedClass(%d)]", i);

    pool::Ref<CrossReferenceClass> refCrossReference(m_dataSvc,ptrs[i]=new CrossReferenceClass);
    pool::Ref<RelatedClass> refRelated(m_dataSvc,rptrs[i]=new RelatedClass);        

    std::string str1 = "CrossReferenceClass_";
    str1 += txt1;
    refCrossReference->setMessage(str1);

    refObj[i] = refCrossReference;
    refRel[i] = refRelated;

    std::string str2 = "RelatedClass_";
    str2 +=txt2;
    refRelated->setMessage(str2);
 
    try{
      refObj[i].markWrite(place1);
      m_tokens.push_back(refObj[i].toString());
    } catch (const RefException& e){
      printException(e);
    }
    try{
      refRel[i].markWrite(place2);
    } catch (const RefException& e){
      printException(e);
    }
    refObj[i]->setRef(refRelated);
    refRel[i]->setRef(refCrossReference); 
  }
  printM("Commit transactions...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  printM("Acid read back...");
  for(i=0;i<NC;i++){
    const RelatedClass* rel = refObj[i]->related();
    if(!rel){
      printError("test ERROR: reference (1) to related class pointer is null");
      break;
    }
    bool comp = (*refObj[i] == *(rel->relObject()));
    if(!comp){
      printError("test ERROR: reference mismatch");
      break;
    } 
  }
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  //for cross-referencing object the embedded refs have to be explicitely reset
  // (otherwise the object are NOT deleted - reference counting never reaches 0)
  printM("Resetting refs...");
  std::vector<Ref<CrossReferenceClass> > tmpRef(NC);
  for(i=0;i<NC;i++){
    tmpRef[i] = refObj[i].ptr();
    refObj[i].reset(); 
    if(refObj[i].token() == 0) 
      printError("test ERROR: token of refObj has been deleted\n");
    if(refRel[i].token() == 0) 
      printError("test ERROR: token of related ref has been deleted\n");
  }
  
  printM("Checking refs integrity...");
  for(i=0;i<NC;i++){
    try{
     refObj[i]->related();
      printError("test ERROR: Exception was not throw after ref reset\n");
    } catch (const RefException& /* e */){  // expected behavor
    }
    try{
      tmpRef[i]->related();
      printError("test ERROR: Exception was not throw after ref reset\n");
    } catch (const RefException& /* e */){  // expected behavor
    }
  }

  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  printM("Reading refs again...");
  for(i=0;i<NC;i++){
    if(!refObj[i]){
      printError("test ERROR: ref is null");
      std::cout << "token="<<refObj[i].token()<<std::endl;
    }
    const RelatedClass* rel = refObj[i]->related();
    if(!rel){
      printError("test ERROR: reference (1) to related class pointer is null");
      break;
    }
    if(rel!=rptrs[i])
      printError("test ERROR: related pointer different from original");
    if(!refObj[i]->refRelated.token()) 
      printError("test ERROR: token of related is null");
    if(rel!=refRel[i].ptr()) 
      printError("test ERROR: embedded rel pointer different from expected");
    const CrossReferenceClass* cr = rel->relObject();
    if(refObj[i].ptr()!=cr)
      printError("test ERROR: embedded obj pointer different from expected");
    bool comp = (*refObj[i] == *(rel->relObject()));
    if(!comp){
      printError("test ERROR: reference mismatch");
      break;
    }
  } 
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }
  printM("Final reset...");
  for(i=0;i<NC;i++){
   refObj[i].reset(); 
  }
 
  return;
}

/// read objects from the persistent store
void pool::CrossReferenceTest::read(unsigned int n_expected){
  if(SharedAnyPtr::numberOfInstances()){
    SharedAnyPtr::dumpRegister();
  }
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }

  unsigned int i;    
  for ( i = 0; i<m_tokens.size(); ++i )   {

    // Read first object
    pool::Ref<CrossReferenceClass> refObj(m_dataSvc,m_tokens[i]);

    if(refObj){
      Ref<const RelatedClass> rel; 
      rel = refObj->related();
      if(!rel){
        printError("test ERROR: reference (2) to related class pointer is null");
        break;
      }
      bool comp = (*refObj == *(rel->relObject()));
      if(!comp){
        printError("test ERROR: reference mismatch");
        break;
      } 

      refObj.reset();

    } else {
      printError("test ERROR: reference to object requested is empty.");
    }
  }

  std::cout << OVALTAG << " " << i << " objects retrieved\n";

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }
      
  if(i!=n_expected){
    printError("test ERROR: number of object retrived different from expected");
  }

  return;
}

void pool::CrossReferenceTest::execute(){
  write(1000);
  read(1000);
}

int main(int,char**)     {
  typedef pool::CrossReferenceTest Test;
  std::string file = "test1.root";
  std::vector< std::string > libraries;
  libraries.push_back( "test_DataSvc_CrossReferenceDict" );
  Test::init( libraries );
  Test theTest(file, pool::ROOTKEY_StorageType);
  theTest.run();
  return 0;
}
