#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TestRefClasses/DerivedClass.h"
#include "TestRefClasses/DerivedTNamed.h"
#ifndef _WIN32
#include <unistd.h>
#endif
#include <stdexcept>

#include "Reflex/Type.h"

namespace pool  {

  class InheritanceTest : public TestBase {
  private:
    const std::string m_dbName1;
    const std::string m_dbName2;
    const DbType m_type1;
    const DbType m_type2;
    std::string m_fid1;
    std::string m_fid2;
    std::vector<std::string> m_tokens;

    bool compareExpected(unsigned int i,const DerivedClass& obj){
      OId oid(i);
      DerivedClass t(i,oid);   
      char txt1[256], txt2[256];
      ::sprintf(txt1,"Name1 %d", i);
      ::sprintf(txt2,"Title %d", i);
      TNamed* nam = new TNamed(txt1,txt2);
      t.setTNamed(nam);   
      bool comp = (t== obj);
      if(!comp){
        printError("test ERROR: object retrieved different from the expected one");
        printM("Expected:-->");
        t.dump();
        printM("Retrieved:-->");
        obj.dump();
      }
      return comp;
    }

  public: 
    InheritanceTest(std::string& dbName1, const DbType& storage_type1,
      std::string& dbName2, const DbType& storage_type2)
      :TestBase(),m_dbName1(dbName1),m_dbName2(dbName2),
      m_type1(storage_type1),m_type2(storage_type2){
        std::vector< std::string > libraries;
        libraries.push_back( "test_TestRefClassesDict" );
        init( libraries );       
        m_fid1 = fileId(m_dbName1,m_type1);
        m_fid2 = fileId(m_dbName2,m_type2);
        if(m_fid1 == "" || m_fid2 == ""){
          printError("test ERROR: cannot retrieve file id");
          return;
        }
      };

    /// write objects to the persistent store
    void write(unsigned int NC)       {

      Placement place1, place2;
      place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
      place1.setContainerName("test_cont_1");
      place1.setTechnology(m_type1.type());
      place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
      place2.setContainerName("test_cont_2");
      place2.setTechnology(m_type2.type());

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

      std::vector<pool::Ref<TestClass> > refTst(NC);
      unsigned int i;
      for (i = 0; i < NC; ++i )   {
        char txt1[256], txt2[256];
        ::sprintf(txt1,"Name1 %d", i);
        ::sprintf(txt2,"Title %d", i);

        OId id(i);
        pool::Ref<TestClass> reftmp(m_dataSvc,new DerivedClass(i,id));   

        //pool::Ref<TNamed> refNamed(m_dataSvc,new DerivedTNamed(txt1, txt2));
        pool::Ref<TNamed> refNamed(m_dataSvc,new TNamed(txt1,txt2));

        refTst[i] = reftmp;

        try{
          refNamed.markWrite(place1);
        } catch (const RefException& e){
          printException(e);        
        }
        try{
          refTst[i].markWrite(place2);
          m_tokens.push_back(refTst[i].toString());
        } catch (const RefException& e){
          printException(e);
        }
        refTst[i]->setTNamed(refNamed);   
      }


      printM("Commit transaction...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }

      std::cout << OVALTAG << "Written on disk "<< NC << " objects\n";

      printM("Acid read-back test:");
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }
      for(i=0;i<NC;i++){
        DerivedClass *deriv = dynamic_cast<DerivedClass*>(refTst[i].ptr());
        if(deriv){ 
          if(!compareExpected(i,*deriv)) break;
        } else {  
          printError("test ERROR: null pointer retrieved");
        }      
      }
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      printM("Clearing cache");
      m_dataSvc->clearCache();
      for(i=0;i<NC;i++){
        try{
          refTst[i]->named()->GetName();
          printError("test ERROR: Exception was not throw after cache clearing");
          break;
        } catch (const RefException& /* e */){  // expected behavor
        }
      }
      return;
    }

    /// read objects from the persistent store
    void read(unsigned int n_expected)        {

      // Starting a read transaction
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      unsigned int i;
      for ( i = 0; i<m_tokens.size() ; ++i )   {

        // Read first object
        pool::Ref<TestClass> refObj(m_dataSvc,m_tokens[i]);

        if( !refObj ){
          printError("test ERROR: reference to object requested is empty.");
        }

        DerivedClass* deriv = dynamic_cast<DerivedClass*>(refObj.ptr());
        deriv->doNothing(); // only check that the call is possible (otherwise will throw exception)

        if(!compareExpected(i,*deriv)) break;

        Ref<DerivedClass> refDer;
        refDer.castDynamic(refObj);
        //refDer = deriv;
        //refObj.castDynamic(refDer);
        refObj = refDer;
        (dynamic_cast<DerivedClass*>(refObj.ptr()))->doNothing();

        refObj = 0;
      }

      std::cout << OVALTAG 
        << " End of reading after " << i << " Objects." << std::endl;

      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      if(i!=n_expected)
        printError("test ERROR: number of object retrieved different from expected");

      printM("Clear cache...");
      m_dataSvc->clearCache();
      return;
    }

    void execute(){ 
      write(1000);
      read(1000);
    }
  };

}

int main(int, char**)     {
  typedef pool::InheritanceTest Test;
  std::string f1 = "type_checking1.root";
  std::string f2 = "type_checking2.root";
  Test theTest(f1, pool::ROOTTREE_StorageType, 
    f2, pool::ROOTTREE_StorageType);
  theTest.run();
  return 0;
}
