#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "DataSvc/ObjectDeletePolicy.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TestRefClasses/TestClass.h"
#include "TNamed.h"
#include <stdexcept>
#include <map>

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool  {

  class MultiWriteTest : public TestBase {
    private:
    const DbType m_type;
    std::map<std::string,pool::Guid> m_guidMap;    

    bool compareExpected(std::string label, unsigned int i, const TestClass& obj){
      OId oid(i);
      TestClass t(i,oid);   
      char txt1[256], txt2[256];
      ::sprintf(txt1,"Name1 %d", i);
      ::sprintf(txt2,"Title %d", i);
      TNamed* nam = new TNamed(txt1,txt2);
      t.setTNamed(nam);        
      t.setLabel(label);
      bool comp = (t== obj);
      if(!comp){
        printError("test ERROR: object retrieved different from the expected one");
        printM("Expected:-->");
        t.dump();
        printM("Retrieved:-->");
        obj.dump();
      }
      return comp;
    }

    bool compareNamed(unsigned int i, const TNamed& nam){   
      char txt1[256], txt2[256];
      ::sprintf(txt1,"Name1 %d", i);
      ::sprintf(txt2,"Title %d", i);
      TNamed t(txt1,txt2);        
      bool comp = (::strcmp(t.GetName(),nam.GetName())==0) 
                  && (::strcmp(t.GetTitle(),nam.GetTitle())==0);
      if(!comp){
        printError("test ERROR: object retrieved different from the expected one");
        printM("Expected:-->");
        std::cout << "Name="<< t.GetName() << std::endl;
        std::cout << "Title="<< t.GetTitle() << std::endl;
        printM("Retrieved:-->");
        std::cout << "Name="<< nam.GetName() << std::endl;
        std::cout << "Title="<< nam.GetTitle() << std::endl;
      }
      return comp;
    }

    public: 
    MultiWriteTest(const DbType& storage_type):TestBase(),m_type(storage_type),m_guidMap(){
      std::vector< std::string > libraries;
      libraries.push_back( "test_TestRefClassesDict" );
      init( libraries );       
    };

   /// write objects to the persistent store
   void write(unsigned int NC){
 
     printM("Starting write procedure"); 
     Placement place1, place2, place3;
     place1.setDatabase("MultiWrite1.root",pool::DatabaseSpecification::PFN );
     place1.setContainerName("test1_cont");
     place1.setTechnology(m_type.type());
     place2.setDatabase("MultiWrite2.root",pool::DatabaseSpecification::PFN );
     place2.setContainerName("test2_cont_1");
     place2.setTechnology(m_type.type());
     place3.setDatabase("MultiWrite2.root",pool::DatabaseSpecification::PFN );
     place3.setContainerName("test2_cont_2");
     place3.setTechnology(m_type.type());
     Placement place4, place5;
     place4.setDatabase("MultiWrite4.root",pool::DatabaseSpecification::PFN );
     place4.setContainerName("test4_cont");
     place4.setTechnology(m_type.type());
     place5.setDatabase("MultiWrite5.root",pool::DatabaseSpecification::PFN );
     place5.setContainerName("test5_cont");
     place5.setTechnology(m_type.type());

     // Start an update transaction
     if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
       printError("test ERROR: Could not start an UPDATE transaction");
     }  

     std::vector<pool::Ref<TestClass> > refTst(NC);
     unsigned int i;
     for (i = 0; i < NC; ++i )   {
       char txt1[256], txt2[256];
       ::sprintf(txt1,"Name1 %d", i);
       ::sprintf(txt2,"Title %d", i);

       pool::Ref<TNamed> refNamed(m_dataSvc,new TNamed(txt1, txt2));
       OId id(i);
       pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));           
       refTst[i] = reftmp;
 
       try{
         refNamed.markMultiwrite(place1);
         refNamed.markMultiwrite(place2);
         refNamed.markMultiwrite(place3);
         m_guidMap.insert(std::make_pair("TNamed",refNamed.token()->classID()));
       } catch (const RefException& e){
         printException(e);        
       }

       try{
         refTst[i].markMultiwrite(place4);
         refTst[i].markMultiwrite(place5);
         m_guidMap.insert(std::make_pair("TestClass",refTst[i].token()->classID()));
       } catch (const RefException& e){
         printException(e);
       }

       refTst[i]->setLabel("WRITE");
       refTst[i]->setTNamed(refNamed); // Assign object pointer to reference
     }

     printM("Commit transactions...");
     if ( ! m_dataSvc->transaction().commit() ) {
       printError("test ERROR: Could not commit the WRITE transaction." );
     }

     std::cout << OVALTAG << "Written on disk "<< NC << " objects\n";

     printM("Acid read-back test:");
     std::vector<pool::Ref<TestClass> > refs(NC);     
     for(i=0;i<NC;i++){
       refs[i] = refTst[i].ptr();
       try{
         if(!compareExpected("WRITE",i,*refTst[i])) break;
       } catch (const RefException& e){
         printException(e);
       }
     }

     printM("Clearing cache");
     m_dataSvc->clearCache();

     for(i=0;i<NC;i++){
       if(refTst[i]) printError("test ERROR: the clear cache did not apply");
       if(refs[i]->label()!="WRITE")
         printError("test ERROR: object found different from expected");
     }

     printM("Write procedure completed");
     return;
   }

   /// read objects from the persistent store
   void read(std::string dbName,std::string contName, 
                                unsigned int n_expected){

     printM("Starting read procedure"); 
     // Starting a read transaction
     if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
       printError("test ERROR: Could not start a READ transaction." );
     }

     std::map<std::string,pool::Guid>::iterator ig = m_guidMap.find("TestClass");
     if(ig == m_guidMap.end()){
       printError("test ERROR: Guid for TestClass is not available  ");
       return; 
     }
     Guid theGuid = ig->second;

     unsigned int i=0;
     for ( i = 0; ; ++i )   {
       Token *tok1 = new Token();
       // Since this example does not use collections,
       // we do it here by faking the tokens to be 
       // returned by the collections
       // 
       tok1->setDb(fileId(dbName,m_type));
       tok1->setCont(contName);
       tok1->setTechnology(m_type.type());
       tok1->setClassID(theGuid);
       tok1->oid().first  = 1;
       tok1->oid().second = i;
       std::string tokString1 = tok1->toString();
       tok1->release();

       // Read main object
       pool::Ref<TestClass> refObj(m_dataSvc,tokString1);
   
       if(!refObj  ){
         std::cout << OVALTAG 
		   << " In file "<< dbName << " End of reading after " 
           << i << " Objects." << std::endl;
         break;
       }

       if(!compareExpected("WRITE",i,*refObj)) break;

       bool okRef = false;

       Ref<const TNamed> refNam(m_dataSvc);
       refNam = refObj->named();
       if ( refNam )   {
         if ( ! (! refNam) )   {
           if ( ! refNam.isNull() )  {
             pool::Ref<const TNamed> other(m_dataSvc,refObj->named());
             if ( refObj->named() == other )   {
               if ( !(refObj->named() != other) )   {
                 const TNamed* t = other.ptr();
                 if ( refObj->named() == t )   {
                   t = refNam.ptr();
                   if ( t == other )   {
                     okRef = true;
                   }
                 }
               }
             }
           }
         }
       }
       if(!okRef)
         printError(
           "test ERROR: Something went awfully wrong with the refs...");

       std::string dbId = refNam.token()->dbID();
       std::string expId = fileId("MultiWrite2.root",m_type);
       if(dbId!= expId){
         printError("test ERROR: reference not found in the expected database");
	 std::cout << "Found    Id = "<< dbId << std::endl;
	 std::cout << "Expected Id = "<< expId << std::endl;
       }

       refObj = 0;
     }

     if ( ! m_dataSvc->transaction().commit() ) {
       printError("test ERROR: Could not commit the READ transaction." );
     }

     if(i!=n_expected)
       printError("test ERROR: number of objects retrieved different from expected");

     printM("Clear cache...");
     m_dataSvc->clearCache();
     printM("Read procedure completed");
     return;
   }

   /// read objects from the persistent store
   void readNamed(std::string dbName, std::string contName, 
                                      unsigned int n_expected){

     std::map<std::string,pool::Guid>::iterator ig = m_guidMap.find("TNamed");
     if(ig == m_guidMap.end()){
       printError("test ERROR: Guid for TNamed is not available  ");
       return; 
     }
     Guid theGuid = ig->second;

     printM("Starting read procedure of named objects"); 
     // Starting a read transaction
     if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
       printError("test ERROR: Could not start a READ transaction." );
     }

     unsigned int i=0;
     for ( i = 0; ; ++i )   {
       Token *tok1 = new Token();
       // Since this example does not use collections,
       // we do it here by faking the tokens to be 
       // returned by the collections
       // 
       tok1->setDb(fileId(dbName,m_type));
       tok1->setCont(contName);
       tok1->setTechnology(m_type.type());
       tok1->setClassID(theGuid);
       tok1->oid().first  = 1;
       tok1->oid().second = i;
       std::string tokString1 = tok1->toString();
       tok1->release();

       // Read main object
       pool::Ref<TNamed> refNam(m_dataSvc,tokString1);
   
       if(!refNam  ){
         std::cout << OVALTAG 
		   << " In file "<< dbName << " End of reading after " 
           << i << " Objects." << std::endl;
         break;
       }

       if(!compareNamed(i,*refNam)) break;

       refNam = 0;
     }

     if ( ! m_dataSvc->transaction().commit() ) {
       printError("test ERROR: Could not commit the READ transaction." );
     }

     if(i!=n_expected)
       printError("test ERROR: number of object retrieved different from expected.");

     printM("Clear cache...");
     m_dataSvc->clearCache();
     printM("Read procedure completed");
     return;
   }

    void execute(){
      write(100);
      read("MultiWrite4.root","test4_cont",100);
      read("MultiWrite5.root","test5_cont",100);
      readNamed("MultiWrite1.root","test1_cont",100);
      readNamed("MultiWrite2.root","test2_cont_1",100);
    }
  };

}

int main(int, char**)     {
  typedef pool::MultiWriteTest Test;
  pool::ObjectDeletePolicy::GLOBAL_DEFAULT = 
        pool::ObjectDeletePolicy::DONOT_DELETE;
  Test theTest(pool::ROOTTREE_StorageType);
  theTest.run();
  return 0;
}
