#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "DataSvc/ObjectDeletePolicy.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TestRefClasses/TestClass.h"
#include "TNamed.h"
#include <stdexcept>

//#include "../../DataSvc/src/CacheSvc.h"

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool  {

  class PtrOwnershipTest : public TestBase {
    private:
    const std::string m_dbName1;
    const std::string m_dbName2;
    const DbType m_type1;
    const DbType m_type2;
    std::string m_fid1;
    std::string m_fid2;
    std::vector<const TNamed*> namedList; 
    std::vector<const TestClass*> tclassList;
    pool::Guid m_testClassGuid;

    bool compareExpected(unsigned int i,const TestClass& obj){
        OId oid(i);
        TestClass t(i,oid);   
	char txt1[256], txt2[256];
        ::sprintf(txt1,"Name1 %d", i);
        ::sprintf(txt2,"Title %d", i);
        TNamed* nam = new TNamed(txt1,txt2);
        t.setTNamed(nam);        
        bool comp = (t== obj);
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
          printM("Expected:-->");
          t.dump();
          printM("Retrieved:-->");
          obj.dump();
        }
        return comp;
    }
    public: 
    PtrOwnershipTest(std::string& dbName1, const DbType& storage_type1,
                     std::string& dbName2, const DbType& storage_type2)
      :TestBase(),m_dbName1(dbName1),m_dbName2(dbName2),
       m_type1(storage_type1),m_type2(storage_type2),m_testClassGuid(""){
      std::vector< std::string > libraries;
      libraries.push_back( "test_TestRefClassesDict" );
      init( libraries );       
      m_fid1 = fileId(m_dbName1,m_type1);
      m_fid2 = fileId(m_dbName2,m_type2);
      if(m_fid1 == "" || m_fid2 == ""){
        printError("test ERROR: cannot retrieve file id");
        return;
      }
    };

    /// write objects to the persistent store
    void write(unsigned int NC)       {
 
      Placement place1, place2;
      place1.setDatabase(m_dbName1,pool::DatabaseSpecification::PFN );
      place1.setContainerName("test1_cont_1");
      place1.setTechnology(m_type1.type());
      place2.setDatabase(m_dbName2,pool::DatabaseSpecification::PFN );
      place2.setContainerName("test2_cont_1");
      place2.setTechnology(m_type2.type());

      // Start an update transaction
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
        printError("test ERROR: Could not start an UPDATE transaction");
      }  

      unsigned int i;
      for (i = 0; i < NC; ++i )   {
        char txt1[256], txt2[256];
        ::sprintf(txt1,"Name1 %d", i);
        ::sprintf(txt2,"Title %d", i);

        pool::Ref<TNamed> refNamed1(m_dataSvc,new TNamed(txt1, txt2));
        OId id(i);
        pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));        
        
        namedList.push_back(refNamed1.ptr());
        tclassList.push_back(reftmp);

        try{
          refNamed1.markWrite(place1);
        } catch (const RefException& e){
          printException(e);        
        }
        try{
          reftmp.markWrite(place2);
          m_testClassGuid = reftmp.token()->classID();
        } catch (const RefException& e){
          printException(e);
        }
        reftmp->setTNamed(refNamed1);   // Assign object pointer to reference
     }
      printM("Commit transaction...");
      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the WRITE transaction." );
      }

      printM("Clearing cache");
      m_dataSvc->clearCache();
      return;
    }

    /// read objects from the persistent store
    void read(unsigned int n_expected) {
      // Starting a read transaction
      if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
        printError("test ERROR: Could not start a READ transaction." );
      }

      unsigned int i;
      for ( i = 0;  ; ++i )   {
        Token *tok = new Token();
        // Since this example does not use collections,
        // we do it here by faking the tokens to be 
        // returned by the collections
        // 
        tok->setDb(m_fid2);
        tok->setCont("test2_cont_1");
        tok->setTechnology(m_type2.type());
        tok->setClassID(m_testClassGuid);
        tok->oid().first  = 1;
        tok->oid().second = i;

        // Read first object
      	pool::Ref<TestClass> refObj(m_dataSvc,*tok);
        tok->release();

        if( !refObj ){
          std::cout << OVALTAG 
            << " End of reading after " << i << " Objects." << std::endl;
          break;
        }

        if(!compareExpected(i,*refObj)) break;

        // Implicit object read and pointer conversion 
        const TNamed* pNam = refObj->named();   
 	char txt[256];
        ::sprintf(txt,"Name1 %d", i);
        bool comp = ::strcmp(txt,pNam->GetName())==0;
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
	  std::cout << "Expected: "<< txt << std::endl;
	  std::cout << "Retrieved: "<< pNam->GetName() << std::endl;
          break;
        }
        namedList.push_back(pNam);

        OId oid(i);
        Ref<TestClass> nRef;
        nRef = new TestClass(i,oid);

        Ref<TestClass> tref(m_dataSvc);
        if(!tref.cacheSvc()) 
          printError("test ERROR: conflicting constructors in Ref<T> class");

        tclassList.push_back(nRef);
        nRef->setTNamed(pNam);
        refObj = nRef;
        if(!compareExpected(i,*refObj)) break;
      }

      if ( ! m_dataSvc->transaction().commit() ) {
        printError("test ERROR: Could not commit the READ transaction." );
      }

      if(i!=n_expected)
        printError("test ERROR: number of objects retrieved different from expected"); 

      std::cout << "Clear cache..." << std::endl;
      m_dataSvc->clearCache();
      return;
    }

    void execute(){
      namedList.clear();
      tclassList.clear();
      unsigned int NC = 10;
      write(NC);
      for(unsigned int i=0;i<NC;i++){
       	char txt[256];
        ::sprintf(txt,"Name1 %d", i);
        bool comp = ::strcmp(txt,namedList[i]->GetName())==0;
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
	  std::cout << "Expected: "<< txt << std::endl;
	  std::cout << "Retrieved: "<< namedList[i]->GetName() << std::endl;
          break;
        }
      }
      for(unsigned int i=0;i<tclassList.size();i++){
        delete tclassList[i];
      }
      for(unsigned int i=0;i<tclassList.size();i++){
        delete namedList[i];
      }
      namedList.clear();
      tclassList.clear();
      read(NC);
      for(unsigned int i=0;i<NC;i++){
       	char txt[256];
        ::sprintf(txt,"Name1 %d", i);
        bool comp = ::strcmp(txt,namedList[i]->GetName())==0;
        if(!comp){
          printError("test ERROR: object retrieved different from the expected one");
	  std::cout << "Expected: "<< txt << std::endl;
	  std::cout << "Retrieved: "<< namedList[i]->GetName() << std::endl;
          break;
        }
      }
      for(unsigned int i=0;i<tclassList.size();i++){
        delete tclassList[i];
      }
      for(unsigned int i=0;i<tclassList.size();i++){
        delete namedList[i];
      }

    }
  };

}

int main(int, char**)     {
  typedef pool::PtrOwnershipTest Test;
  std::string f1 = "test1.root";
  std::string f2 = "test2.root";
  pool::ObjectDeletePolicy::GLOBAL_DEFAULT = 
        pool::ObjectDeletePolicy::DONOT_DELETE;
  Test theTest(f1, pool::ROOT_StorageType,f2,pool::ROOT_StorageType);
  theTest.run();
  return 0;
}
