#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include <stdexcept>
#include "RootPtrMemberClass.h"
#include "ObjectPtrMemberClass.h"
#include "DataSvc_PtrMember.h"

#ifndef WIN32
# include <unistd.h>
#endif

bool pool::PtrMemberTest::compareExpected(unsigned int i,const RootPtrMemberClass& obj){
  RootPtrMemberClass t(i);   
  char txt1[256], txt2[256];
  ::sprintf(txt1,"Name1 %d", i);
  ::sprintf(txt2,"Title %d", i);
  TNamed nam(txt1,txt2);
  t.setTNamed(&nam);
  bool comp = (t== obj);
  if(!comp){
    printError("test ERROR: object retrieved different from the expected one");
    printM("Expected:-->");
    t.dump();
    printM("Retrieved:-->");
    obj.dump();
  }
  return comp;
}
bool pool::PtrMemberTest::compareExpected(unsigned int i,const ObjectPtrMemberClass& obj){
  ObjectPtrMemberClass t(i);   
  OId oid(i);
  t.setExtId(&oid);
  bool comp = (t== obj);
  if(!comp){
    printError("test ERROR: object retrieved different from the expected one");
    printM("Expected:-->");
    t.dump();
    printM("Retrieved:-->");
    obj.dump();
  }
  return comp;
}

pool::PtrMemberTest::PtrMemberTest(
                       std::string& dbName, const DbType& storage_type) 
                       :TestBase(),m_dbName(dbName),m_type(storage_type),m_guidMap(){
  std::vector< std::string > libraries;
  libraries.push_back( "test_DataSvc_PtrMemberDict" );
  pool::TestBase::init( libraries );
  m_fid = fileId(m_dbName,m_type);
  if(m_fid == ""){
    printError("test ERROR: cannot retrieve file id");
  }
}

void pool::PtrMemberTest::write(unsigned int NC) { 

  Placement place;
  place.setDatabase(m_dbName,pool::DatabaseSpecification::PFN );
  place.setContainerName("test_cont_1");
  place.setTechnology(m_type.type());

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  std::vector<pool::Ref<RootPtrMemberClass> > refR(NC);
  std::vector<pool::Ref<ObjectPtrMemberClass> > refO(NC);
  std::vector<TNamed*> rids(NC);
  std::vector<OId*> oids(NC);
  unsigned int i;
  for (i = 0; i < NC; ++i )   {
    pool::Ref<RootPtrMemberClass> refRoot(m_dataSvc,new RootPtrMemberClass(i));        
    pool::Ref<ObjectPtrMemberClass> refObj(m_dataSvc,new ObjectPtrMemberClass(i));        
        
    refR[i] = refRoot;        
    refO[i] = refObj;        

    refR[i].markWrite(place);
    refO[i].markWrite(place);

    m_guidMap.insert(std::make_pair("RootPtrMemberClass",refR[i].token()->classID()));
    m_guidMap.insert(std::make_pair("ObjectPtrMemberClass",refO[i].token()->classID()));

    char txt1[256], txt2[256];
    ::sprintf(txt1,"Name1 %d", i);
    ::sprintf(txt2,"Title %d", i);
    rids[i] = new TNamed(txt1,txt2);
    refR[i]->setTNamed(rids[i]);
    oids[i] = new OId(i);
    refO[i]->setExtId(oids[i]);

  }
  printM("Commit transaction...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  std::cout << OVALTAG << "Written on disk 2*"<< NC << " objects\n";

  printM("Acid read-back test:");
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  for(i=0;i<NC;i++){
    if(!compareExpected(i,*refR[i])) break;
    if(!compareExpected(i,*refO[i])) break;
  }
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  for(i=0;i<NC;i++){
    delete rids[i];
    delete oids[i];
  }

  printM("Clearing cache");
  m_dataSvc->clearCache();
  return;
}

void pool::PtrMemberTest::read(unsigned int n_expected){
  // Starting a read transaction
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  unsigned int i;
  for ( i = 0;  ; ++i )   {
    Token *tok = new Token();
    tok->setDb(m_fid);
    tok->setCont("test_cont_1");
    tok->setTechnology(m_type.type());
    tok->oid().first  = 1;
    tok->oid().second = i;
    if(i%2 == 0){       
      tok->setClassID(m_guidMap["RootPtrMemberClass"]);
      // Read first object
      pool::Ref<RootPtrMemberClass> refObj(m_dataSvc,*tok);
      tok->release();

      if( !refObj ){
        std::cout << OVALTAG 
        << " End of reading after " << i << " Objects." << std::endl;
        break;
      }
 
      if(!compareExpected(refObj->id(),*refObj)) break;

      refObj = 0;
    } else {
      tok->setClassID(m_guidMap["ObjectPtrMemberClass"]);
      // Read first object
      pool::Ref<ObjectPtrMemberClass> refObj(m_dataSvc,*tok);
      tok->release();

      if( !refObj ){
        std::cout << OVALTAG 
        << " End of reading after " << i << " Objects." << std::endl;
        break;
      }
 
      if(!compareExpected(refObj->id(),*refObj)) break;

      refObj = 0;
    }

  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  if(i!=2*n_expected) 
    printError("test ERROR: number of read objects different from expected");

  printM("Clear cache...");
  m_dataSvc->clearCache();
  return;
}


int main(int, char**)     {
  typedef pool::PtrMemberTest Test;
  std::string file = "test1.root";
  Test theTest(file, pool::ROOTKEY_StorageType);
  theTest.run();
  return 0;
}
