#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "TestBase/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "TestRefClasses/TestClass.h"
#include <stdexcept>

#ifndef WIN32
# include <unistd.h>
#endif

namespace pool  {

  class RefAssignmentTest : public TestBase {
    private:
    const DbType m_type;
    Token* tok;

    public: 
    RefAssignmentTest(const DbType& storage_type)
                   :TestBase(),m_type(storage_type),tok(0){
      std::vector< std::string > libraries;
      libraries.push_back( "test_TestRefClassesDict" );
      init( libraries );       
    };


    void testAssignment(pool::Ref<TestClass>& aRef){
      // assignment to an empty Ref
      printM("Empty ref case");
      Ref<TestClass> empty;
      empty = aRef;
      if(empty.isNull() != aRef.isNull()){
        printError("test ERROR: RefBase::isNull() mismatch");
      }
      if(empty.cacheSvc() != aRef.cacheSvc()){
        printError("test ERROR: RefBase::cacheSvc() mismatch");
      }
      if(empty.cacheEntry() != aRef.cacheEntry()){
        printError("test ERROR: RefBase::cacheLocation() mismatch");
      }
      if(empty  != aRef){
        printError("test ERROR: RefBase::equals() mismatch");
      }      

      // assignment to a dataSvc bound Ref
      printM("DataSvc bound ref case");
      Ref<TestClass> boundRef(m_dataSvc);
      boundRef = aRef;
      if(boundRef.isNull() != aRef.isNull()){
        printError("test ERROR: RefBase::isNull() mismatch");
      }
      if(aRef.cacheSvc()){
        if(boundRef.cacheSvc() != aRef.cacheSvc()){
          printError("test ERROR: RefBase::cacheSvc() mismatch");
        }
        if(boundRef.cacheEntry() != aRef.cacheEntry()){
          printError("test ERROR: RefBase::cacheLocation() mismatch");
        }
      } else {
        if(!boundRef.cacheSvc()){
          printError("test ERROR: cache binding broken");
        }
        if(aRef){
          if(!boundRef.cacheEntry()){
            printError("test ERROR: cache location binding broken");
          }
        }
      }
      if(boundRef  != aRef){
        printError("test ERROR: RefBase::equals() mismatch");
      }      

      // assignment to a dataSvc bound (pointer) Ref
      printM("DataSvc + pointer ref case");
      Ref<TestClass> objRef(m_dataSvc, new TestClass);
      objRef = aRef;
      if(objRef.isNull() != aRef.isNull()){
        printError("test ERROR: RefBase::isNull() mismatch");
      }
      if(aRef.cacheSvc()){
        if(objRef.cacheSvc() != aRef.cacheSvc()){
          printError("test ERROR: RefBase::cacheSvc() mismatch");
        }
        if(objRef.cacheEntry() != aRef.cacheEntry()){
          printError("test ERROR: RefBase::cacheLocation() mismatch");
        }
      } else {
        if(!objRef.cacheSvc()){
          printError("test ERROR: cache binding broken");
        }
        if(aRef){
          if(!objRef.cacheEntry()){
            printError("test ERROR: cache location binding broken");
          }
        }
      }
      if(objRef  != aRef){
        printError("test ERROR: RefBase::equals() mismatch");
      }   

      // assignment to a dataSvc bound (token) Ref
      printM("DataSvc + token ref case");
      if(tok){
        Ref<TestClass> tokRef(m_dataSvc, *tok);
        tokRef = aRef;
        if(tokRef.isNull() != aRef.isNull()){
          printError("test ERROR: RefBase::isNull() mismatch");
        }
        if(aRef.cacheSvc()){
          if(tokRef.cacheSvc() != aRef.cacheSvc()){
            printError("test ERROR: RefBase::cacheSvc() mismatch");
          }
          if(tokRef.cacheEntry() != aRef.cacheEntry()){
            printError("test ERROR: RefBase::cacheLocation() mismatch");
          }
        } else {
          if(!tokRef.cacheSvc()){
            printError("test ERROR: cache binding broken");
          }
          if(aRef){
            if(!tokRef.cacheEntry()){
              printError("test ERROR: cache location binding broken");
            }
          }
        }
        if(tokRef  != aRef){
          printError("test ERROR: RefBase::equals() mismatch");
        }      
      }
    }

    void testOnWrite(){
      // testing ref assignment from a Ref created for write
      printM("testing assignment on write---------------------");
      m_dataSvc->transaction().start(pool::ITransaction::UPDATE);
      Ref<TestClass> ref(m_dataSvc,new TestClass);
      Placement place;
      place.setDatabase("DB.root", pool::DatabaseSpecification::PFN); 
      place.setContainerName("Cont1");
      place.setTechnology(pool::ROOTKEY_StorageType.type()); 
      testAssignment(ref);      
      ref.markWrite(place);
      tok = const_cast<Token*>(ref.token());
      tok->addRef();
      testAssignment(ref);      
      m_dataSvc->transaction().commit();
      m_dataSvc->session().disconnectAll();
      testAssignment(ref); 
      printM("-------------------------------------------------");     
    }

    void testOnRead(){
      printM("testing assignment on read------------------------");
      m_dataSvc->transaction().start(pool::ITransaction::READ);
      Ref<TestClass> ref(m_dataSvc,*tok);
      testAssignment(ref);      
      m_dataSvc->transaction().commit();
      m_dataSvc->session().disconnectAll();
      testAssignment(ref);      
      printM("-------------------------------------------------");     
    }

    void testUnboundRef(){
      printM("testing assignment of a cache-unbound ref------------");
      Ref<TestClass> ref;
      Ref<TestClass> r = ref;
      testAssignment(ref);      
      ref = new TestClass;
      testAssignment(ref);           
      printM("-------------------------------------------------");     
    }


    void execute(){
      testOnWrite();
      testOnRead();
      testUnboundRef();
      tok->release();
      tok = 0;
    }
  };

}

int main(int, char**)     {
  typedef pool::RefAssignmentTest Test;
  Test theTest(pool::ROOTTREE_StorageType);
  theTest.run();
  return 0;
}
