#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "POOLCore/Token.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include <stdexcept>
#include "TestRefClasses/TestClass.h"
#include "DataSvc_SimpleAppend.h"

#ifndef WIN32
# include <unistd.h>
#endif

bool pool::SimpleAppendTestBase::compareExpected(unsigned int i,const TestClass& obj){
  OId oid(i);
  TestClass t(i,oid);   
  bool comp = (t== obj);
  if(!comp){
    printError("test ERROR: object retrieved different from the expected one");
    printM("Expected:-->");
    t.dump();
    printM("Retrieved:-->");
    obj.dump();
  }
  return comp;
}

pool::SimpleAppendTestBase::SimpleAppendTestBase(
                       std::string& dbName, const DbType& storage_type) 
                       :TestBase(),m_dbName(dbName),m_type(storage_type){
  std::vector< std::string > libraries;
  libraries.push_back( "test_TestRefClassesDict" );
  pool::TestBase::init( libraries );
  m_fid = fileId(m_dbName,m_type);
  if(m_fid == ""){
    printError("test ERROR: cannot retrieve file id");
  }
}

void pool::SimpleAppendTestBase::write(unsigned int NC) { 

  printM("Start WRITE procedure");
  Placement place;
  place.setDatabase(m_dbName,pool::DatabaseSpecification::PFN );
  place.setContainerName("test_cont_1");
  place.setTechnology(m_type.type());

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  std::vector<pool::Ref<TestClass> > refTst(NC);
  unsigned int i;
  for (i = 0; i < NC; ++i )   {
    OId id(i);
    pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));        
        
    refTst[i] = reftmp;        

    try{
      refTst[i].markWrite(place);
    } catch (const RefException& e){
      printException(e);
    }

    Token* tok = const_cast<Token*>(refTst[i].token());
    tok->addRef();
    tokens.push_back(tok);

  }
  printM("Commit transaction...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();
  std::cout << OVALTAG << "Written on disk "<< NC << " objects\n";

  printM("Clearing cache");
  m_dataSvc->clearCache();
  return;
}

void pool::SimpleAppendTestBase::reWrite(unsigned int NC){
  printM("Start reWRITE procedure");
  Placement place;
  place.setDatabase(m_dbName,pool::DatabaseSpecification::PFN );
  place.setContainerName("test_cont_1");
  place.setTechnology(m_type.type());

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  
  std::vector<pool::Ref<TestClass> > refTst(NC);
  unsigned int i;
  for (i = 0; i < NC; ++i )   {
    OId id(i);
    pool::Ref<TestClass> reftmp(m_dataSvc,new TestClass(i,id));        
        
    refTst[i] = reftmp;               

    try{
      refTst[i].markWrite(place);
    } catch (const RefException& e){
      printException(e);
    }

    Token* tok =  const_cast<Token*>(refTst[i].token());
    tok->addRef();
    tokens.push_back(tok);
  }
  printM("Commit transaction...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  printM("Acid read-back test:");
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  for(i=0;i<NC;i++){
    if(!compareExpected(i,*refTst[i])) break;
  }
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();

  printM("Clearing cache");
  m_dataSvc->clearCache();
  return;
}

void pool::SimpleAppendTestBase::read(unsigned int n_expected){
  printM("Start READ procedure");
  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) {
    printError("test ERROR: Could not start a READ transaction." );
  }
  unsigned int i;
  for ( i = 0; i<tokens.size() ; ++i )   {

    // Read first object
    pool::Ref<TestClass> refObj(m_dataSvc,*tokens[i]);
    tokens[i]->release();

    if( !refObj ){
      std::cout << OVALTAG 
      << " End of reading after " << i << " Objects." << std::endl;
      break;
    }
 
    if(!compareExpected(refObj->id(),*refObj)) break;

    refObj = 0;
  }

  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the READ transaction." );
  }

  // disconnect 
  m_dataSvc->session().disconnectAll();

  if(i!=n_expected) 
    printError("test ERROR: number of read objects different from expected");

  printM("Clear cache...");
  m_dataSvc->clearCache();
  return;
}


int main(int argc, char** argv)     {
  pool::TestBase* theTest = 0;
  std::string f = "test_simpleAppend1.root";
  if(argc > 1){
    if(argv[1][0]=='F'){
      theTest = new pool::SimpleAppendTest(f, pool::ROOTKEY_StorageType);
    } else if(argv[1][0]=='W'){
      theTest = new pool::SimpleWriteTest(f, pool::ROOTKEY_StorageType);
    } else if(argv[1][0]=='U'){
      theTest = new pool::SimpleReWriteTest(f, pool::ROOTKEY_StorageType);
    } else if(argv[1][0]=='R'){
      theTest = new pool::SimpleReadTest(f, pool::ROOTKEY_StorageType);
    }
  } else {
    theTest = new pool::SimpleAppendTest(f, pool::ROOTKEY_StorageType);   
  }
  if(theTest){
    theTest->run();
    delete theTest;
  }
  return 0;
}
