#include "RefTestClass.h"

#include "StressTestBase/StressTestApp.h"
#include "StressTestBase/SimpleAsciiCollection.h"

#include "POOLCore/Token.h"
#include "FileCatalog/URIParser.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"

#include <iostream>
#include <fstream>
#include <string>
#ifndef _WIN32
#include <unistd.h>
#endif

using std::cout;
using std::endl;
using std::string;
using std::ifstream;
using std::ios;

using pool::Ref;
using pool::Token;
using pool::URIParser;
using pool::IDataSvc;
//using pool::CacheLocation;


class RefTestApp : public pool::StressTestApp {
public:
  RefTestApp(int argc, const char* argv[])
      :StressTestApp(argc,argv)
    {}

  int userMain();
  
  void CreateTests();
  void SimpleCreate(RefTestClass& c);
  void SimpleCreate(const RefTestClass& c);
  void IsNull();
  void FromToken();

  void CopyTests();
  void CopyConstructor();
  void AssignmentOperator();

  void OperatorTests();
  void Ptr();
  void Not();
  void Star();

  void EqualityTests();
  void Equality();
  void Inequality();

  void MiscTests();
  void ContextReset();
  void CacheLocation();
private:
};






int RefTestApp::userMain()
{
  CreateTests();
  CopyTests();
  OperatorTests();
  EqualityTests();
  MiscTests();
  
  return 0;
}






void RefTestApp::CreateTests()
{
  int firstInt = 1;
  int secondInt = 2;

  string step = "========== Ref Creation Tests =========="; ovalmsg(step);
  
  // I create some objects, then pass them to someone else.
  // Someone else creates a Ref of those objects- when they go
  // out of scope the object is destroyed- or is it?
  // (Note effectively 3 tests in this block- tests Ref ownership
  RefTestClass *pTestObj1 = new RefTestClass(firstInt);
  const RefTestClass *pConstTestObj1 = new RefTestClass(secondInt);

  SimpleCreate(*pTestObj1);
  SimpleCreate(*pConstTestObj1);

  step = "Read from my object after someone else has created a Ref (using a copy constructor!)";
  if (pTestObj1->data() == firstInt) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  step = "Read from my STATIC object after someone else 'might' have Refd it";
  if (pConstTestObj1->data() == secondInt) {
    step="PASSED "+step;
  } else {
    step="FAILED (Ref took ownership and then went out of scope) "+step;
  }
  ovalmsg(step);

  IsNull();
  FromToken();
}

void RefTestApp::SimpleCreate(RefTestClass& c)
{
  string step = "Breaks build! (good) Create const Ref, try to modify transient";

  //int check = c.data();
  static Ref<const RefTestClass> r(context(),new RefTestClass(c));
  //r->data(check+1);

  //if (r->data() != check) {
  //  step="PASSED "+step;
  //} else {
  //  step="FAILED "+step;
  //}
  ovalmsg(step);
}

void RefTestApp::SimpleCreate(const RefTestClass& c)
{
  string step = "Breaks build! (good) Create const Ref, try to modify STATIC transient";

  // Now c is const so
  // c->data(1);
  // doesn't work. In theory you shouldn't be able to set up a Ref
  // that allows you to modify c- but you can!

  // We need to be honest and if someone passes a const object, use a const Ref!

  //int check = c.data();
  static Ref<const RefTestClass> r(context(),&c);
  //r->data(check+1);

  //if (c.data() == check) {
  //  step="PASSED "+step;
  //} else {
  //  step="FAILED "+step;
  //}
  ovalmsg(step);
}

void RefTestApp::IsNull()
{
  string step = "Check is_null method";

  RefTestClass *c = new RefTestClass(1);

  static Ref<RefTestClass> r3;
  static Ref<RefTestClass> r(context(),c);
  if ( r3.isNull() ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if ( !(r.isNull()) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::FromToken()
{
  catalog().start();

  Ref<RefTestClass> r1(context(),new RefTestClass(1));

  string dbFile = "ConstRefTest.root";
  ifstream dbcheck (dbFile.c_str(),ios::in);
  ::unlink(dbFile.c_str());

  pool::Placement place1;
  place1.setDatabase(dbFile,pool::DatabaseSpecification::PFN);
  place1.setContainerName("cont_1");
  place1.setTechnology(pool::ROOTTREE_StorageType.type());

  string collFile = "ConstRefTest.coll";
  ::unlink(collFile.c_str());
  pool::SimpleAsciiCollection coll(collFile);

  context()->transaction().start( pool::ITransaction::UPDATE );
  r1.markWrite(place1);
  coll.add(r1.token()->toString());
  coll.write();
  context()->transaction().commit();

  catalog().commit();


  std::string step = "Create a Ref from the externalised token string";
  Ref<RefTestClass> r4(context(),coll.collection()[0]);
  if ( r4 == r1 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  step = "Check token-as-string returned from Ref";
  if ( r4.toString() == coll.collection()[0] ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
    cout << "From Ref " << r4.toString() << endl << "Externalised " << coll.collection()[0] << endl;
  }
  ovalmsg(step);

  step = "Check token ptr returned from Ref";
  const Token *tok = r4.token();
  if ( tok->toString() == coll.collection()[0] ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
    cout << "From Token " << tok->toString() << endl << "Externalised " << coll.collection()[0] << endl;
  }
  ovalmsg(step);

  step = "Check ptr to context returned from Ref (no longer applicable)";
  //if ( r4.dataSvc() == context() ) {
  //  step="PASSED "+step;
  //} else {
  //  step="FAILED "+step;
  //}
  ovalmsg(step);




}

void RefTestApp::CopyTests()
{
  string step = "========== Ref Copy Tests =========="; ovalmsg(step);

  CopyConstructor();
  AssignmentOperator();
}

void RefTestApp::CopyConstructor()
{
  string step = "Creating new Ref via Copy Constructor";

  RefTestClass *c = new RefTestClass(1);

  static Ref<RefTestClass> r(context(),c);
  static Ref<RefTestClass> s(r);

  if (r->data() == s->data()) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::AssignmentOperator()
{
  string step = "Creating new Ref via Assignment Operator";

  RefTestClass *c = new RefTestClass(1);

  static Ref<RefTestClass> r(context(),c);
  static Ref<RefTestClass> s = r;

  if (r->data() == s->data()) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::OperatorTests()
{
  string step = "========== Misc Operator Tests =========="; ovalmsg(step);

  Ptr();
  Not();
  Star();
}

void RefTestApp::Ptr()
{
  string step = "Check returned real pointer";

  RefTestClass *c = new RefTestClass(1);
  RefTestClass *pFoilObj = new RefTestClass(10);

  static Ref<RefTestClass> r1(context(),c);
  if ( (r1.ptr()) == c ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if (! ((r1.ptr()) == pFoilObj) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  delete pFoilObj;
}

void RefTestApp::Not()
{
  string step = "Check ! operator";

  RefTestClass *c = new RefTestClass(1);

  static Ref<RefTestClass> r3;
  static Ref<RefTestClass> r(context(),c);
  if ( !r3 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if ( !(!r) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::Star()
{
  string step = "Check star operator";

  RefTestClass *c = new RefTestClass(1);

  static Ref<RefTestClass> r(context(),c);
  if (*r == *c) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::EqualityTests()
{
  string step = "========== Misc Equality Tests =========="; ovalmsg(step);

  Equality();
  Inequality();
}

void RefTestApp::Equality()
{
  string step = "Check Ref == Ref equality";

  static Ref<RefTestClass> r1(context(),new RefTestClass(1));
  static Ref<RefTestClass> r2(context(),new RefTestClass(1));
  static Ref<RefTestClass> r3(context(),new RefTestClass(2));

  if ( r2 == r1 ) {
    step="FAILED "+step;
  } else {
    step="PASSED "+step;
  }
  if (! (r3 == r2) ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);
}

void RefTestApp::Inequality()
{
  static Ref<RefTestClass> r1(context(),new RefTestClass(1));
  static Ref<RefTestClass> r3(context(),new RefTestClass(2));

  RefTestClass *pTestObj = new RefTestClass(5);

  string step = "Check Ref != Ref equality";

  if ( r3 != r1 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  if ( r3 != r3 ) {
    step="FAILED "+step;
  } else {
    step="PASSED "+step;
  }
  ovalmsg(step);

  step = "Check Ref != Obj equality ";

  if ( r3 != pTestObj ){
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);

  step = "Check Obj != Ref equality";

  if ( pTestObj != r3 ) {
    step="PASSED "+step;
  } else {
    step="FAILED "+step;
  }
  ovalmsg(step);


  delete pTestObj;
}

void RefTestApp::MiscTests()
{
  string step = "========== Misc Tests =========="; ovalmsg(step);

  ContextReset();
  CacheLocation();
}

void RefTestApp::ContextReset()
{
  string step = "Check resetting of Ref context (no longer applicable)";

  //static Ref<RefTestClass> r4(context(),new RefTestClass(1));
  
  //IDataSvc* altContext = createContext();
  //r4.setDataSvc(altContext);
  
  //if ( r4.dataSvc() == context() ) {
  //  step="FAILED "+step;
  //} else {
  //  step="PASSED "+step;
  //}
  ovalmsg(step);
}

void RefTestApp::CacheLocation()
{
  string step = "TO DO !! Check Cache location query";

  //Ref<RefTestClass> r4(context(),new RefTestClass(1));

  //CacheLocation *loc = r4.cacheLocation();

  ovalmsg(step);
}

  
// low level main
int main(int argc, const char *argv[])
{
  std::vector<std::string> libraries;
  libraries.push_back("test_DataSvc_StaticRefDict");

  RefTestApp app(argc,argv);
  app.loadLib(libraries);
  app.run();

  return 0;
}
