#include "PluginManager/PluginManager.h"
#include "POOLCore/Exception.h"
#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"

#include "StorageSvc/DbType.h"

#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"

#include "DataSvc/IDataSvc.h"
#include "DataSvc/ICacheSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/Ref.h"

#include "TestDictionary/SimpleTestClass.h"
#include <stdexcept>


namespace pool  {
  int numTokenInstances();
}

int main( int, char** ) {
  try {
    const std::string databaseFile = "pool_test.root";
    const std::string container = "theContainer";
    const std::string fileCatalog = "FileCatalog.xml";
    const int numberOfObjects = 1000;

    std::cout << "Retrieving the plugin manager." << std::endl;
    seal::PluginManager* pm = seal::PluginManager::get();
    if ( ! pm ) {
      throw std::runtime_error( "Could not get the plugin manager instance" );
    }
    std::cout << "Initializing the plugin manager." << std::endl;
    pm->initialise();
    
    std::cout << "Loading the dictionary" << std::endl;
    const std::string dictlibrary = "test_TestDictionaryDict";
    try {  // This is temporary. SEAL should get better in this.
      seal::SharedLibrary::load( seal::SharedLibrary::libname( dictlibrary ) );
    }
    catch ( seal::SharedLibraryError *error) {
      throw std::runtime_error( error->explainSelf().c_str() );
    }

    std::cout << "Creating a file catalog object" << std::endl;
//     pool::IFileCatalogPluginFactory* fcf = pool::IFileCatalogPluginFactory::get();
//     if ( ! fcf ) {
//       throw std::runtime_error( "Could not retrieve the file catalog plugin factory" );
//     }
//     pool::IFileCatalog* fc = fcf->create( "XMLFileCatalog" );

    pool::IFileCatalog* fc = new pool::IFileCatalog();
    pool::URIParser p("file:" + fileCatalog);
    p.parse();
    fc->addReadCatalog( p.contactstring() );
    
    if ( ! fc ) {
      throw std::runtime_error( "Could not create a file catalog" );
    }
    fc->connect();
    fc->start();

    pool::ObjectDeletePolicy::GLOBAL_DEFAULT = pool::ObjectDeletePolicy::DONOT_DELETE;
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
    if ( ! dataSvc ) {
      throw std::runtime_error( "Could not get an IDataSvc object." );
    }

    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
    policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
    dataSvc->session().setDefaultConnectionPolicy( policy );
    
    pool::Placement placementHint;
    placementHint.setDatabase( databaseFile, pool::DatabaseSpecification::PFN );
    placementHint.setContainerName( container );
    placementHint.setTechnology( pool::ROOTKEY_StorageType.type() );

    pool::Ref< SimpleTestClass > myRef;

    for ( int i = 0; i < numberOfObjects; ++i ) {
      if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
	throw std::runtime_error( "Could not start the transaction" );
      }
      SimpleTestClass* data = new SimpleTestClass;
      data->data = i;
      pool::Ref< SimpleTestClass > r( dataSvc, data );
      myRef = r;
      myRef.markWrite( placementHint );
      if ( ! dataSvc->transaction().commit() ) {
	throw std::runtime_error( "Could not commit the transaction" );
      }
      delete data;
      //myRef.reset();
      myRef = 0;
    }

    fc->commit();
    std::cout << "[OVAL] Number of cached items : " << dataSvc->cacheSvc().cacheSize() << std::endl;
    delete dataSvc;
    delete fc;

  }
  catch( const pool::Exception& se ) {
    std::cerr << se.what() << std::endl;
    return 1;
  }
  catch( const std::exception& e ) {
    std::cerr << e.what() << std::endl;
    return 1;
  }
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
  return 0;
}
