#include "POOLCore/Token.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "DataSvc/IDataSvc.h"
#include "AttributeList/AttributeListSpecification.h"
#include "AttributeList/AttributeList.h"
#include "Collection/Collection.h"


#include "TestBase.h"
#include "ExampleClass.h"

#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include <vector>
#include <fstream>


/**
   Simple test driver of the explicit (MySQL and Root) collections 
   The Driver is based on Giacomo Govi's DataSvc_ReadWrite test,
   and includes fragments of his code.
*/


namespace pool
{

  class ExpliciteCollectionTest : public TestBase 
  {

  private:
    const std::string m_dbName1;
    const std::string m_dbName2;
    const DbType m_type1;
    const DbType m_type2;
    std::string m_collectionType;
    std::string m_connectionString;
    const std::string m_collectionName;
    std::string m_fid1;
    std::string m_fid2;

    void printAttributes( const AttributeList& attrList );

  public:
    ExpliciteCollectionTest( std::string& dbName1, const DbType& storage_type1,
                             std::string& dbName2, const DbType& storage_type2,
                             std::string& collType )
      :TestBase(),m_dbName1( dbName1 ),m_dbName2( dbName2 ),
       m_type1( storage_type1 ),m_type2( storage_type2 ), m_collectionType( collType ),
       m_connectionString( "" ),
       m_collectionName( std::string( "test_collection" ) )
    {
       if( collType != "RootCollection" ) {
	  if( ! ::getenv( "CORAL_AUTH_PATH" ) ) {
	     ::putenv( "CORAL_AUTH_PATH=/afs/cern.ch/sw/lcg/app/pool/db" ); 
	  }
	  if( ! ::getenv( "CORAL_DBLOOKUP_PATH" ) ) {
	     ::putenv( "CORAL_DBLOOKUP_PATH=/afs/cern.ch/sw/lcg/app/pool/db" );
	  }
	  //m_connectionString = "oracle://devdb10/ral_writer";
	  m_connectionString = "write_test";
      }

      std::vector< std::string > libraries;
      libraries.push_back( "test_ExplicitCollection_FunctionalityDict" );
      TestBase::init( libraries );
      m_fid1 = fileId( m_dbName1, m_type1 );
      m_fid2 = fileId( m_dbName2, m_type2 );
      if( m_fid1 == "" || m_fid2 == "" )
      {
        printError( "test ERROR: cannot retrieve file id" );
        return;
      }
    };

    void write();
    void read();
    void update();

    void execute()
    {
      std::cout << std::endl << "******* COLLECTION WRITE TEST *******" << std::endl;
      write();
      std::cout << std::endl << "******* COLLECTION READ TEST *******" << std::endl;
      read();
      std::cout << std::endl << "******* COLLECTION UPDATE TEST *******" << std::endl;
      update();
      std::cout << std::endl << "******* COLLECTION READ TEST *******" << std::endl;
      read();
    };
  };
}


void pool::ExpliciteCollectionTest::printAttributes( const AttributeList& attrList )
{
  const AttributeListSpecification& attrListSpec = attrList.attributeListSpecification();
  for( unsigned int i = 0; i < attrListSpec.size(); ++i )
  {
    const AttributeSpecification& attrSpec = attrListSpec[i];
    std::string attrName = attrSpec.name();
    std::string attrValue = attrList[attrName].getValueAsString();
    std::cout << attrName << " = " << attrValue << "\t";
  }
}  


void pool::ExpliciteCollectionTest::write()
{

  Placement place1, place2;
  place1.setDatabase( m_dbName1, pool::DatabaseSpecification::PFN );
  place1.setContainerName( "test1_cont_1" );
  place1.setTechnology( m_type1.type() );
  place2.setDatabase( m_dbName2, pool::DatabaseSpecification::PFN );
  place2.setContainerName( "test2_cont_1" );
  place2.setTechnology( m_type2.type() );

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) )  {
    printError( "test ERROR: Could not start an UPDATE transaction" );
  }

  printM( "Creating the collection." );
  Collection< ExampleClass >* coll = new Collection< ExampleClass >( m_dataSvc, m_collectionType,
                                                                     m_connectionString, m_collectionName,
                                                                     ICollection::CREATE_AND_OVERWRITE );
  if( ! coll )  {
    throw std::runtime_error( "Could not create the collection" );
  }

  printM( "Creating attribute list specification." );
  AttributeListSpecification& spec = coll->attributeListSpecification();
  spec.push_back( "attr1", "unsigned int" );
  spec.push_back( "attr2", "string" );
  spec.push_back( "attr3", "double" );

  unsigned int NC = 100;
  std::vector< pool::Ref< ExampleClass > > refTst( NC );
  printM( "Adding elements to the collection." );
  for ( unsigned int i = 0; i < NC; ++i )
  {
    char txt1[256], txt2[256], txt3[256];
    ::sprintf( txt1, "Name1 %d", i );
    ::sprintf( txt2, "Title %d", i );
    ::sprintf( txt3,"SimpleString %d", i );

    pool::Ref< TNamed > refNamed( m_dataSvc, new TNamed( txt1, txt2 ) );
    pool::Ref< ExampleClass > reftmp( m_dataSvc, new ExampleClass( i ) );
    refTst[i] = reftmp;
    refTst[i]->s = "ExampleClass_";
    refTst[i]->s += txt3;          

    try
    {
      refNamed.markWrite( place1 );
    }
    catch ( RefException e )
    {
      printException( e );
    }

    try
    {
      refTst[i].markWrite( place2 );
    }
    catch ( RefException e )
    {
      printException( e );
    }
    
    refTst[i]->setTNamed( refNamed ); // Assign object pointer to reference

    AttributeList attrList( spec );
    attrList[ "attr1" ].setValue( i );
    char s[ 20 ];
    sprintf( s, "%s_%d", "testString", i );
    attrList[ "attr2" ].setValue( std::string( s ) );
    attrList[ "attr3" ].setValue( double( i )/3 );
    coll->add( refTst[i], attrList );
    std::cout << "Example object added to the collection. ";
    std::cout << "Object's metadata: ";
    printAttributes( attrList );
    std::cout << std::endl;
  }

  printM("Committing the transaction...");
  if ( ! m_dataSvc->transaction().commit() ) {
    printError("test ERROR: Could not commit the WRITE transaction." );
  }

  printM( "Committing the collection." );
  coll->commit();
  printM( "Closing the collection." );
  coll->close();
  
  printM("Clearing cache");
  m_dataSvc->clearCache();

  delete coll;
  return;

}

void pool::ExpliciteCollectionTest::read()
{

  if ( ! m_dataSvc->transaction().start( pool::ITransaction::READ ) ) 
  {
    printError("test ERROR: Could not start a READ transaction." );
  }

  printM( "Opening the collection." );
  Collection< ExampleClass > coll( m_dataSvc, m_collectionType, m_connectionString,
                                   m_collectionName, ICollection::READ );

  //AttributeListSpecification& spec = coll.attributeListSpecification();
  printM( "Creating Multiple Fetches Iterator with empty query." );
  Collection< ExampleClass >::Iterator it = coll.select( "", "", "FetchAll" );

  while( it.next() )
  {
    std::cout << "Next object retrieved. ";
    std::cout << "Object's metadata: ";
    printAttributes( it.attributeList() );
    std::cout << std::endl;
  }

  printM( "Closing the collection." );
  coll.close();

  printM( "Opening the collection." );
  coll.open();

  std::string query( "(attr1 < 10 OR attr1 > 80) AND attr2 like '%5'" );
  if(m_collectionType=="RootCollection")
    query = "(attr1 < 10 || attr1 > 80) && strstr(attr2,\"5\")";
  std ::cout << "Creating Single Fetches Iterator with a query: \"" << query  << "\"" << std::endl;
  Collection< ExampleClass >::Iterator it2 = coll.select( query, "", "FetchOne" );

  while( it2.next() )
  {
    std::cout << "[OVAL] ";
    std::cout << "Next object retrieved. ";
    std::cout << "Object's metadata: ";
    printAttributes( it2.attributeList() );
    std::cout << std::endl;
  }

  printM( "Closing the collection." );
  coll.close();

  if ( ! m_dataSvc->transaction().commit() ) 
  {
    printError( "test ERROR: Could not commit the READ transaction." );
  }

  printM( "Clearing cache..." );
  m_dataSvc->clearCache();

  return;
}


void pool::ExpliciteCollectionTest::update()
{

  Placement place1, place2;
  place1.setDatabase( m_dbName1, pool::DatabaseSpecification::PFN );
  place1.setContainerName( "test1_cont_1" );
  place1.setTechnology( m_type1.type() );
  place2.setDatabase( m_dbName2, pool::DatabaseSpecification::PFN );
  place2.setContainerName( "test2_cont_1" );
  place2.setTechnology( m_type2.type() );

  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) )
  {
    printError( "test ERROR: Could not start an UPDATE transaction" );
  }

  // Change the default policy NOT to overwrite an existing file
  pool::DatabaseConnectionPolicy policy = m_dataSvc->session().defaultConnectionPolicy();
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::UPDATE );
  m_dataSvc->session().setDefaultConnectionPolicy( policy );

  printM( "Creating the collection." );
  Collection< ExampleClass >* coll = new Collection< ExampleClass >( m_dataSvc, m_collectionType, 
                                                                     m_connectionString, m_collectionName, 
                                                                     ICollection::UPDATE );
  if( ! coll )
  {
    throw std::runtime_error( "Could not create the collection" );
  }

  printM( "Creating attribute list specification." );
  AttributeListSpecification& spec = coll->attributeListSpecification();

  unsigned int NC = 50;

  std::vector< pool::Ref< ExampleClass > > refTst( NC );
  printM( "Adding elements to the collection." );
  for ( unsigned int i = 0; i < NC; ++i )
  {
    char txt1[256], txt2[256], txt3[256];
    ::sprintf( txt1, "Name1 %d", i+100 );
    ::sprintf( txt2, "Title %d", i+100 );
    ::sprintf( txt3, "SimpleString %d", i+100 );

    pool::Ref< TNamed > refNamed( m_dataSvc, new TNamed( txt1, txt2 ) );
    pool::Ref< ExampleClass > reftmp( m_dataSvc, new ExampleClass( i ) );
    refTst[i] = reftmp;
    refTst[i]->s = "ExampleClass_";
    refTst[i]->s += txt3;
    
    try
    {
      refNamed.markWrite( place1 );
    }
    catch ( RefException e )
    {
      printException( e );
    }

    try
    {
      refTst[i].markWrite( place2 );
    }
    catch ( RefException e )
    {
      printException( e );
    }
    
    refTst[i]->setTNamed( refNamed ); // Assign object pointer to reference

    AttributeList attrList( spec );
    attrList[ "attr1" ].setValue( i+100 );
    char s[ 20 ];
    sprintf( s, "%s_%d", "testString", i+100 );
    attrList[ "attr2" ].setValue( std::string( s ) );
    attrList[ "attr3" ].setValue( double( i+100 )/3 );
    coll->add( refTst[i], attrList );
    std::cout << "Example object added to the collection. ";
    std::cout << "Object's metadata: ";
    printAttributes( attrList );
    std::cout << std::endl;
    
    if( i == 30 )
    {
      printM( "Committing the collection." );
      coll->commit( );
      printM("Committing the transaction...");
      if ( ! m_dataSvc->transaction().commit() )
      {
        printError("test ERROR: Could not commit the UPDATE transaction." );
      }
      if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) )
      {
        printError( "test ERROR: Could not start an UPDATE transaction" );
      }
    }
  }

  printM( "Rolling back the collection." );
  coll->rollback();

  printM("Rolling back the transaction...");
  m_dataSvc->transaction().rollback();

  printM( "Closing the collection." );
  coll->close();

  printM("Clearing cache.");
  m_dataSvc->clearCache();

  delete coll;
  return;

}


namespace pool  {
  int numTokenInstances();
}

int main()
{
  {
    std::string collectionType( "RelationalCollection" );
    std::cout << "######################################" << std::endl;
    std::cout << "Testing " << collectionType << " ..." << std::endl;
    std::cout << "######################################" << std::endl;
    typedef pool::ExpliciteCollectionTest Test;
    std::string f1 = "test1.root";
    std::string f2 = "test2.root";
    Test theTest( f1, pool::ROOTTREE_StorageType,
                  f2, pool::ROOTKEY_StorageType,
                  collectionType );
    try
    {
      theTest.run();
    }
    catch( ... )
    {
      std::cout << "Unknown exception" << std::endl;
    }
  }

  {
    std::string collectionType( "RootCollection" );
    std::cout << "######################################" << std::endl;
    std::cout << "Testing " << collectionType << " ..." << std::endl;
    std::cout << "######################################" << std::endl;
    typedef pool::ExpliciteCollectionTest Test;
    std::string f1 = "test1.root";
    std::string f2 = "test2.root";
    Test theTest( f1, pool::ROOTTREE_StorageType,
                  f2, pool::ROOTKEY_StorageType,
                  collectionType );
    try
    {
      theTest.run();
    }
    catch( ... )
    {
      std::cout << "Unknown exception" << std::endl;
    }
  }
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
  return 0;
}


