#include <iostream>

#include "TestBase.h"

#include <stdexcept>
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/RefException.h"
#include "DataSvc/SharedAnyPtr.h"
#include "FileCatalog/IFileCatalog.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"

#include "FileCatalog/URIParser.h"
#include "FileCatalog/FCSystemTools.h"
#include "FileCatalog/IFCAction.h"
#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"
#include "POOLCore/POOLContext.h"

namespace pool
{
  int numTokenInstances();
}


std::string pool::TestBase::OVALTAG = "[OVAL]";

pool::TestBase::TestBase() : m_dataSvc( 0 ), m_fileCatalog( 0 ),
                             m_init_test( false ), nerr( 0 )
{
  try
  {
    pool::POOLContext::loadComponent( "SEAL/Services/Message Service" );
    pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );
    // initialize catalog
    URIParser pars( "xmlcatalog_file:test_catalog.xml" );
    pars.parse();

//     m_fileCatalog = FCSystemTools::createCatalog( pars.prefix() );

    m_fileCatalog = new pool::IFileCatalog();
    m_fileCatalog->setWriteCatalog( pars.contactstring() );
    
    if ( ! m_fileCatalog )
    {
      throw std::runtime_error( "Could not retrieve a file catalog" );
    }
    
    m_fileCatalog->connect();
    m_fileCatalog->start();

    // initialize dataSvc
    m_dataSvc = DataSvcFactory::instance( m_fileCatalog );
    m_init_test = true;
    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
    policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
    m_dataSvc->session().setDefaultConnectionPolicy( policy );
  } 
  catch ( std::exception e )
  {
    std::cout << "test ERROR:" << e.what() << std::endl;
    addError();
  }
}


pool::TestBase::~TestBase()
{
  delete m_dataSvc;
  m_fileCatalog->commit();
  m_fileCatalog->disconnect();
  delete m_fileCatalog;
}


std::string pool::TestBase::fileId( const std::string& pfn, const pool::DbType& technology )
{
  std::string fid, tech;
  // try to re-use an already existing file
  pool::IFCAction action;
  m_fileCatalog->setAction( action );
  
  action.lookupFileByPFN( pfn, fid, tech );
  // did not exist - so try register the file
  if( fid.length() == 0 )
  {
    pool::DbType majorType( technology.majorType() );
    action.registerPFN( pfn, majorType.storageName(), fid );
    if ( fid.length() == 0 )
      throw std::runtime_error( "failed to register file `" + pfn +"'" );
  }
  return fid;
}


void pool::TestBase::printException( pool::Exception& e )
{
  std::cout << "Exception: " << e.what() << std::endl;
  addError();
}

void pool::TestBase::init( const std::vector< std::string >& libraries )
{
  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
	iLibrary != libraries.end(); ++iLibrary ) 
  {
    const std::string& library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;
    try 
    {  // This is temporary. SEAL should get better in this.
      const std::string plaformDependentName = seal::SharedLibrary::libname( library );
      seal::SharedLibrary::load( plaformDependentName );
    }
    catch ( seal::SharedLibraryError *error)
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }
  }
}


void pool::TestBase::run()
{
  if( m_init_test )
  {
    try
    {
      execute();
    }
    catch( pool::Exception e )
    {
      std::cout << "The test has thrown an exception:" << e.what() << std::endl;
      addError();
    }
    catch( std::runtime_error e )
    {
      std::cout << "The test has thrown an exception:" << e.what() << std::endl;
      addError();
    }
    catch( std::exception e )
    {
      std::cout << "The test has thrown an exception:" << e.what() << std::endl;
      addError();
    }
    catch( ... )
    {
      std::cout << "The test has thrown an unknown exception." << std::endl;
      addError();
    }
  }
  dumpError();

}


void pool::TestBase::printM( const char* str )
{
  std::cout << str << std::endl;
}


void pool::TestBase::printE( pool::Exception& e )
{
  std::cout << "Exception: " << e.what() << std::endl;
}


void pool::TestBase::dLine()
{
  printM("----------------------------------------------------------------\n");
}


void pool::TestBase::sLine()
{
  printM("****************************************************************\n");
}


void pool::TestBase::printError( const char* str )
{
  printM( str );
  nerr++;
}


void pool::TestBase::dumpError() const
{
  std::cout << OVALTAG <<" Test result: " << nerr << " errors" << std::endl;
}


