#include <vector>
#include <stdexcept>

#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"

#include "POOLCore/POOLContext.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"

#include "StorageSvc/DbType.h"

#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"

#include "DataSvc/IDataSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/Ref.h"

#include "Collection/Collection.h"

#include "MyTestClasses.h"
#include "TestDriver.h"


namespace pool  {
  int numTokenInstances();
}

TestDriver::TestDriver():
  m_databaseFile( "pool_test.root" ),
  m_container( "theContainer" ),
  m_fileCatalog( "FileCatalog.xml" ),
  m_numberOfObjects( 1000 )
{
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

  std::cout << "Loading the dictionary" << std::endl;

  const std::string dictlibrary = "test_ImplicitCollection_EmbeddedRefsDict";

  try {
    seal::SharedLibrary::load( seal::SharedLibrary::libname( dictlibrary ) );
  }
  catch ( seal::SharedLibraryError *error) {
    throw std::runtime_error( error->explainSelf().c_str() );
  }
}


TestDriver::~TestDriver()
{
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}


void
TestDriver::writeData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }

  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  dataSvc->session().setDefaultConnectionPolicy( policy );

  fc->start();
  if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
    throw std::runtime_error( "Could not start the transaction" );
  }

  pool::Placement placementHint;
  placementHint.setDatabase( m_databaseFile, pool::DatabaseSpecification::PFN );
  placementHint.setContainerName( m_container );
  placementHint.setTechnology( pool::ROOTKEY_StorageType.type() );

  for ( int i = 0; i < m_numberOfObjects; ++i ) {
    MyBaseClass* myBaseClass = new MyBaseClass;
    myBaseClass->data = i + 1111;
    pool::Ref< MyBaseClass > myBaseClassRef( dataSvc, myBaseClass );

    MyDerivedClass* myDerivedClass = new MyDerivedClass;
    myDerivedClass->data = 2*i + 2222;
    myDerivedClass->extraData = 2*i + 1;
    pool::Ref< MyDerivedClass > myDerivedClassRef( dataSvc, myDerivedClass );

    MyOtherClass* myOtherClass = new MyOtherClass;
    myOtherClass->otherData = 3*i;
    pool::Ref< MyOtherClass > myOtherClassRef( dataSvc, myOtherClass );
    myOtherClassRef->refToBase = myBaseClassRef;
    myOtherClassRef->refToDerived = myDerivedClassRef;

    myBaseClassRef.markWrite( placementHint );
    myDerivedClassRef.markWrite( placementHint );
    myOtherClassRef.markWrite( placementHint );
  }

  if ( ! dataSvc->transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction" );
  }
  fc->commit();

  dataSvc->clearCache();
  delete dataSvc;
  delete fc;
}


void
TestDriver::readDataSequencially()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }
  else {
    fc->start();
    if ( ! dataSvc->transaction().start( pool::ITransaction::READ ) ) {
      throw std::runtime_error( "Could not start the transaction" );
    }

    std::vector< pool::Token* > tokens;
    tokens.reserve( 2 * m_numberOfObjects );
    std::vector< int > objectData;
    objectData.reserve( 2 * m_numberOfObjects );

    // Retrieving the top class
    pool::Collection<MyOtherClass> myOtherClassCollection( dataSvc,
							   "ImplicitCollection",
							   "PFN:" + m_databaseFile,
							   m_container,
							   pool::ICollection::READ );
    pool::Collection<MyOtherClass>::Iterator myOtherClassIterator = myOtherClassCollection.select();
    
    std::cout << "Collecting the data looping over the top level objects" << std::endl;
    int numberOfObjectsReadBack = 0;
    while ( myOtherClassIterator.next() ) {
      if ( myOtherClassIterator->otherData != 3*numberOfObjectsReadBack ) {
	throw std::runtime_error( "Unexpected data in the read object" );
      }

      pool::Ref<MyBaseClass>& refBase = myOtherClassIterator->refToBase;
      if ( ! refBase.token() ) {
	throw std::runtime_error( "Token not valid for the (base) reference" );
      }
      tokens.push_back( const_cast< pool::Token* >( refBase.token() ) );
      tokens.back()->addRef();
      if ( refBase->data != numberOfObjectsReadBack + 1111 ) {
	throw std::runtime_error( "Unexpected data for the (base) referenced object" );
      }
      objectData.push_back( refBase->data );
      
      pool::Ref<MyBaseClass>& refDerived = myOtherClassIterator->refToDerived;
      if ( ! refDerived.token() ) {
	throw std::runtime_error( "Token not valid for the (derived) reference" );
      }
      tokens.push_back( const_cast< pool::Token* >( refDerived.token() ) );
      tokens.back()->addRef();
      if ( refDerived->data != 2 * numberOfObjectsReadBack + 2222 ) {
	throw std::runtime_error( "Unexpected data for the (derived) referenced object" );
      }
      objectData.push_back( refDerived->data );
      
      ++numberOfObjectsReadBack;
    };
    
    if ( numberOfObjectsReadBack != m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of object read back" );
    }

    // Retrieving the base class
    pool::Collection<MyBaseClass> myBaseClassCollection( dataSvc,
							 "ImplicitCollection",
							 "PFN:" + m_databaseFile,
							 m_container,
							 pool::ICollection::READ );  
    pool::Collection<MyBaseClass>::Iterator myBaseClassIterator = myBaseClassCollection.select();
    
    // Looping over the top class objects
    numberOfObjectsReadBack = 0;
    while ( myBaseClassIterator.next() ) {
      if ( numberOfObjectsReadBack >= static_cast<int>( tokens.size() ) ) {
	throw std::runtime_error( "Too many objects are read back." );
      }
      
      if ( myBaseClassIterator->data != objectData[numberOfObjectsReadBack] ) {
	std::cerr << "Data from the embedded ref : " <<  objectData[numberOfObjectsReadBack] << std::endl;
	std::cerr << "Data from the implicit collection : " << myBaseClassIterator->data << std::endl;
	throw std::runtime_error( "Different data retrieved!!!" );
      }

      // Comparing the tokens
      pool::Ref<MyBaseClass> ref = myBaseClassIterator.ref();
      if ( ! ref.token() ) {
	throw std::runtime_error( "Could not retrieve the token from the ref" );
      }
      if ( !( *( ref.token() ) == *( tokens[numberOfObjectsReadBack] ) ) ) {
	std::cerr << "Token string from the embedded ref : " << tokens[numberOfObjectsReadBack]->toString() << std::endl;
	std::cerr << "Token string from the implicit collection : " << ref.token()->toString() << std::endl;
	throw std::runtime_error( "Different tokens retrieved!!!" );
      }
      
      ++numberOfObjectsReadBack;
    }
    
    if ( numberOfObjectsReadBack != 2 * m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of objects read back from the implicit collection" );
    }

    for ( std::vector< pool::Token* >::iterator iToken = tokens.begin();
	  iToken != tokens.end(); ++iToken ) {
      (*iToken)->release();
    }

    if ( ! dataSvc->transaction().commit() ) {
      throw std::runtime_error( "Could not commit the transaction" );
    }
    fc->commit();

  }

  delete dataSvc;
  delete fc;
}


void
TestDriver::readDataInParallel()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }
  else {

    fc->start();
    if ( ! dataSvc->transaction().start( pool::ITransaction::READ ) ) {
      throw std::runtime_error( "Could not start the transaction" );
    }

    // Retrieving the top class
    pool::Collection<MyOtherClass> myOtherClassCollection( dataSvc,
							   "ImplicitCollection",
							   "PFN:" + m_databaseFile,
							   m_container,
							   pool::ICollection::READ );
    pool::Collection<MyOtherClass>::Iterator myOtherClassIterator = myOtherClassCollection.select();

    // Retrieving the base class
    pool::Collection<MyBaseClass> myBaseClassCollection( dataSvc,
							 "ImplicitCollection",
							 "PFN:" + m_databaseFile,
							 m_container,
							 pool::ICollection::READ );
    pool::Collection<MyBaseClass>::Iterator myBaseClassIterator = myBaseClassCollection.select();
    
    // Looping over the top class objects
    int numberOfObjectsReadBack = 0;
    while ( myOtherClassIterator.next() ) {
      
      if ( myOtherClassIterator->otherData != 3*numberOfObjectsReadBack ) {
	throw std::runtime_error( "Unexpected data in the read object" );
      }
      
      // Fetching the base object
      if ( ! myBaseClassIterator.next() ) {
	throw std::runtime_error( "Could not retrieve the referenced (base) object" );
      }
      
      if ( myBaseClassIterator->data != numberOfObjectsReadBack + 1111 ) {
	throw std::runtime_error( "Unexpected data for the base object" );
      }
      // Comparing the data of the referenced object
      if ( myOtherClassIterator->refToBase->data != myBaseClassIterator->data ) {
	throw std::runtime_error( "Different data are retrieved for the same (base) object" );
      }
      
      // Comparing the tokens
      const pool::Token* tokenBase1 = myOtherClassIterator->refToBase.token();
      pool::Ref<MyBaseClass> refBase = myBaseClassIterator.ref();
      const pool::Token* tokenBase2 = refBase.token();
      if ( ! ( *tokenBase1 == *tokenBase2 ) ) {
	throw std::runtime_error( "Different tokens for the same persistent (base) object" );
      }
      
      // Fetching the derived object
      if ( ! myBaseClassIterator.next() ) {
	throw std::runtime_error( "Could not retrieve the referenced (derived) object" );
      }
      if ( myBaseClassIterator->data != 2*numberOfObjectsReadBack + 2222 ) {
	throw std::runtime_error( "Unexpected data for the base object" );
      }
      // Comparing the data of the referenced object
      if ( myOtherClassIterator->refToDerived->data != myBaseClassIterator->data ) {
	throw std::runtime_error( "Different data are retrieved for the same (derived) object" );
      }
      
      // Comparing the tokens
      const pool::Token* tokenDerived1 = myOtherClassIterator->refToDerived.token();
      pool::Ref<MyBaseClass> refDerived = myBaseClassIterator.ref();
      const pool::Token* tokenDerived2 = refDerived.token();
      if ( ! ( *tokenDerived1 == *tokenDerived2 ) ) {
	throw std::runtime_error( "Different tokens for the same persistent (base) object" );
      }
      
      ++numberOfObjectsReadBack;
    }
    
    if ( numberOfObjectsReadBack != m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of object read back" );
    }
    
    std::cout << "Committing..." << std::endl;
    if ( ! dataSvc->transaction().commit() ) {
      throw std::runtime_error( "Could not commit the transaction" );
    }
    fc->commit();
  }

  delete dataSvc;
  delete fc;
}
