#include <stdexcept>

#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"

#include "POOLCore/POOLContext.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"

#include "StorageSvc/DbType.h"

#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"

#include "DataSvc/IDataSvc.h"
#include "DataSvc/ICacheSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/Ref.h"

#include "Collection/Collection.h"

#include "TestDictionary/SimpleTestClass.h"

#include "TestDriver.h"


namespace pool  {
  int numTokenInstances();
}


TestDriver::TestDriver():
  m_databaseFile( "pool_test.root" ),
  m_container( "theContainer" ),
  m_fileCatalog( "FileCatalog.xml" ),
  m_numberOfObjects( 1000 )
{
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

  std::cout << "Loading the dictionaries" << std::endl;

  try {
    seal::SharedLibrary::load( seal::SharedLibrary::libname( "test_TestDictionaryDict" ) );
  }
  catch ( seal::SharedLibraryError *error) {
    throw std::runtime_error( error->explainSelf().c_str() );
  }
}


TestDriver::~TestDriver()
{
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}


void
TestDriver::writeData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }

  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  dataSvc->session().setDefaultConnectionPolicy( policy );

  fc->start();
  if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
    throw std::runtime_error( "Could not start the transaction" );
  }

  pool::Placement placementHint;
  placementHint.setDatabase( m_databaseFile, pool::DatabaseSpecification::PFN );
  placementHint.setContainerName( m_container );
  placementHint.setTechnology( pool::ROOTKEY_StorageType.type() );


  for ( int i = 0; i < m_numberOfObjects; ++i ) {
    SimpleTestClass* data = new SimpleTestClass;
    data->data = i;
    pool::Ref< SimpleTestClass > myRef( dataSvc, data );
    myRef.markWrite( placementHint );
  }

  if ( ! dataSvc->transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction" );
  }
  fc->commit();

  std::cout << "[OVAL] Number of cached items : " << dataSvc->cacheSvc().cacheSize() << std::endl;
  delete dataSvc;
  delete fc;
}


void
TestDriver::readData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }
  else {

    fc->start();
    if ( ! dataSvc->transaction().start( pool::ITransaction::READ ) ) {
      throw std::runtime_error( "Could not start the transaction" );
    }

    pool::Collection<SimpleTestClass> collection( dataSvc,
                                                  "ImplicitCollection",
                                                  "PFN:" + m_databaseFile,
                                                  m_container,
                                                  pool::ICollection::READ );

    std::cout << "Selecting all the objects." << std::endl;
    pool::Collection<SimpleTestClass>::Iterator collectionIterator = collection.select();

    std::cout << "Looping all over the objects." << std::endl;
    int numberOfObjectsReadBack = 0;
    while ( collectionIterator.next() ) {
      if ( collectionIterator->data != numberOfObjectsReadBack ) {
        throw std::runtime_error( "Unexpected data in the read object" );
      }
      ++numberOfObjectsReadBack;
    }

    if ( numberOfObjectsReadBack != m_numberOfObjects ) {
      std::cout << "Read " << numberOfObjectsReadBack << " instead of " << m_numberOfObjects << std::endl;
      throw std::runtime_error( "Invalid number of object read back" );
    }
    
    std::cout << "Done reading the objects." << std::endl;
    
    if ( ! dataSvc->transaction().commit() ) {
      throw std::runtime_error( "Could not commit the transaction" );
    }
    fc->commit();
  }

  delete dataSvc;
  delete fc;
}


void
TestDriver::appendReadData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }
  else {

    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::RAISE_ERROR );
    policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::UPDATE );
    dataSvc->session().setDefaultConnectionPolicy( policy );

    fc->start();
    {
      if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
        throw std::runtime_error( "Could not start the transaction" );
      }
      
      pool::Placement placementHint;
      placementHint.setDatabase( m_databaseFile, pool::DatabaseSpecification::PFN );
      placementHint.setContainerName( m_container );
      placementHint.setTechnology( pool::ROOTKEY_StorageType.type() );
    
      // appending the new data
      for ( int i = 0; i < m_numberOfObjects; ++i ) {
        SimpleTestClass* data = new SimpleTestClass;
        data->data = i + m_numberOfObjects;
        pool::Ref< SimpleTestClass > myRef( dataSvc, data );
        myRef.markWrite( placementHint );
      }
      
      // Reading back the data.
      pool::Collection<SimpleTestClass> collection( dataSvc,
                                                    "ImplicitCollection",
                                                    "PFN:" + m_databaseFile,
                                                    m_container,
                                                    pool::ICollection::READ );
      
      std::cout << "[OVAL] *** 1 Selecting all the objects." << std::endl;
      pool::Collection<SimpleTestClass>::Iterator collectionIterator = collection.select();
      
      std::cout << "[OVAL] *** 1 Looping all over the objects." << std::endl;
      int numberOfObjectsReadBack = 0;
      while ( collectionIterator.next() ) {
        if ( collectionIterator->data != numberOfObjectsReadBack ) {
          std::cout << "Read " << collectionIterator->data << " Should be:" << numberOfObjectsReadBack << std::endl;
          throw std::runtime_error( "Unexpected data in the read object" );
        }
        ++numberOfObjectsReadBack;
      }
      
      if ( numberOfObjectsReadBack != 2 * m_numberOfObjects ) {
        std::cout << "Read " << numberOfObjectsReadBack << " instead of " << 2*m_numberOfObjects << std::endl;
        throw std::runtime_error( "Invalid number of object read back" );
      }
    
      std::cout << "[OVAL] *** 1 Done reading back " << 2*m_numberOfObjects << " objects." << std::endl;
      
      // Committing the transaction and the file catalog
      if ( ! dataSvc->transaction().commit() ) {
        throw std::runtime_error( "Could not commit the transaction" );
      }
    }

    // ===========================================================================================
    // Delete every second object starting with the first one: sequence should then be:
    // 1, 3, 5, 7, ...
    // ===========================================================================================
    {
      if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
        throw std::runtime_error( "Could not start the transaction" );
      }
      // Reading back the data.
      pool::Collection<SimpleTestClass> collection( dataSvc,
                                                    "ImplicitCollection",
                                                    "PFN:" + m_databaseFile,
                                                    m_container,
                                                    pool::ICollection::READ );
      
      std::cout << "[OVAL] *** 2 Selecting all the objects and delete every second..." << std::endl;
      pool::Collection<SimpleTestClass>::Iterator collectionIterator = collection.select();
      
      std::cout << "[OVAL] *** 2 Looping all over the objects." << std::endl;
      int numberOfObjectsReadBack = 0;
      while ( collectionIterator.next() ) {
        if ( collectionIterator->data != numberOfObjectsReadBack ) {
          std::cout << "Read " << collectionIterator->data << " Should be:" << numberOfObjectsReadBack << std::endl;
          throw std::runtime_error( "Unexpected data in the read object" );
        }
        if ( (numberOfObjectsReadBack%2) == 0 ) {
          collectionIterator.ref().markDelete();
        }
        ++numberOfObjectsReadBack;
      }
      
      if ( numberOfObjectsReadBack != 2 * m_numberOfObjects ) {
        std::cout << "*** 2 Read " << numberOfObjectsReadBack << " instead of " << 2*m_numberOfObjects << std::endl;
        throw std::runtime_error( "Invalid number of object read back" );
      }
      
      std::cout << "[OVAL] *** 2 Done reading back the objects. Commit delete operation" << std::endl;
      
      // Committing the transaction and the file catalog
      if ( ! dataSvc->transaction().commit() ) {
        throw std::runtime_error( "Could not commit the transaction" );
      }
    }
    // ===========================================================================================
    // Read back objects after the delete cycle.
    // The object data sequence now should be 1,3,5,7,...
    // ===========================================================================================
    {
      if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
        throw std::runtime_error( "Could not start the transaction" );
      }
      // Reading back the data.
      pool::Collection<SimpleTestClass> collection( dataSvc,
                                                    "ImplicitCollection",
                                                    "PFN:" + m_databaseFile,
                                                    m_container,
                                                    pool::ICollection::READ );
      
      std::cout << "[OVAL] *** 3 Selecting all the objects." << std::endl;
      pool::Collection<SimpleTestClass>::Iterator collectionIterator = collection.select();
      
      std::cout << "[OVAL] *** 3 Looping all over the objects." << std::endl;
      int numberOfObjectsReadBack = 0;
      while ( collectionIterator.next() ) {
        if ( collectionIterator->data != (1+2*(numberOfObjectsReadBack)) ) {
          std::cout << "Read " << collectionIterator->data 
                    << " Should be:" << (1+(2*numberOfObjectsReadBack)) << std::endl;
          throw std::runtime_error( "Unexpected data in the read object" );
        }
        ++numberOfObjectsReadBack;
      }
      
      if ( numberOfObjectsReadBack != m_numberOfObjects ) {
        std::cout << "*** 3 Read " << numberOfObjectsReadBack << " instead of " << m_numberOfObjects << std::endl;
        throw std::runtime_error( "Invalid number of object read back" );
      }
      
      std::cout << "[OVAL] *** 3 Done reading back " << m_numberOfObjects << " objects." << std::endl;
      
      // Committing the transaction and the file catalog
      if ( ! dataSvc->transaction().commit() ) {
        throw std::runtime_error( "Could not commit the transaction" );
      }
    }

    fc->commit();
  }
  std::cout << "[OVAL] Number of cached items : " << dataSvc->cacheSvc().cacheSize() << std::endl;
  delete dataSvc;
  delete fc;
}
