#include <stdexcept>

#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"

#include "POOLCore/POOLContext.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"

#include "StorageSvc/DbType.h"

#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"

#include "DataSvc/IDataSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/Ref.h"

#include "Collection/Collection.h"

#include "MyTestClasses.h"
#include "TestDriver.h"

namespace pool  {
  int numTokenInstances();
}

TestDriver::TestDriver():
  m_databaseFile( "pool_test.root" ),
  m_container( "theContainer" ),
  m_fileCatalog( "FileCatalog.xml" ),
  m_numberOfObjects( 1000 )
{
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

  std::cout << "Loading the dictionary" << std::endl;

  const std::string dictlibrary = "test_ImplicitCollection_MixedContainerDict";

  try {
    seal::SharedLibrary::load( seal::SharedLibrary::libname( dictlibrary ) );
  }
  catch ( seal::SharedLibraryError *error) {
    throw std::runtime_error( error->explainSelf().c_str() );
  }
}


TestDriver::~TestDriver()
{
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}


void
TestDriver::writeData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }

  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  dataSvc->session().setDefaultConnectionPolicy( policy );

  fc->start();
  if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
    throw std::runtime_error( "Could not start the transaction" );
  }

  pool::Placement placementHint;
  placementHint.setDatabase( m_databaseFile, pool::DatabaseSpecification::PFN );
  placementHint.setContainerName( m_container );
  placementHint.setTechnology( pool::ROOTKEY_StorageType.type() );

  for ( int i = 0; i < m_numberOfObjects; ++i ) {
    MyBaseClass* myBaseClass = new MyBaseClass;
    myBaseClass->data = i;
    pool::Ref< MyBaseClass > myBaseClassRef( dataSvc, myBaseClass );
    myBaseClassRef.markWrite( placementHint );

    MyDerivedClass* myDerivedClass = new MyDerivedClass;
    myDerivedClass->data = 2*i;
    myDerivedClass->extraData = 2*i + 1;
    pool::Ref< MyDerivedClass > myDerivedClassRef( dataSvc, myDerivedClass );
    myDerivedClassRef.markWrite( placementHint );

    MyOtherClass* myOtherClass = new MyOtherClass;
    myOtherClass->otherData = 3*i;
    pool::Ref< MyOtherClass > myOtherClassRef( dataSvc, myOtherClass );
    myOtherClassRef.markWrite( placementHint );
  }

  if ( ! dataSvc->transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction" );
  }
  fc->commit();

  dataSvc->clearCache();
  delete dataSvc;
  delete fc;
}


void
TestDriver::readData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }
  else {

    fc->start();
    if ( ! dataSvc->transaction().start( pool::ITransaction::READ ) ) {
      throw std::runtime_error( "Could not start the transaction" );
    }

    pool::Collection<MyOtherClass> myOtherClassCollection( dataSvc,
							   "ImplicitCollection",
							   "PFN:" + m_databaseFile,
							   m_container,
							   pool::ICollection::READ );

    pool::Collection<MyOtherClass>::Iterator myOtherClassIterator = myOtherClassCollection.select();
    
    int numberOfObjectsReadBack = 0;
    while ( myOtherClassIterator.next() ) {
      if ( myOtherClassIterator->otherData != 3*numberOfObjectsReadBack ) {
	throw std::runtime_error( "Unexpected data in the read object" );
      }
      ++numberOfObjectsReadBack;
    }
    
    if ( numberOfObjectsReadBack != m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of object read back" );
    }

    // Retrieving now the base class
    pool::Collection<MyBaseClass> myBaseClassCollection( dataSvc,
							 "ImplicitCollection",
							 "PFN:" + m_databaseFile,
							 m_container,
							 pool::ICollection::READ );
  
    pool::Collection<MyBaseClass>::Iterator myBaseClassIterator = myBaseClassCollection.select();
    
    numberOfObjectsReadBack = 0;
    while ( myBaseClassIterator.next() ) {
      if ( myBaseClassIterator->data != ( numberOfObjectsReadBack % 2  + 1 ) * ( numberOfObjectsReadBack / 2 ) ) {
	throw std::runtime_error( "Unexpected data in the read object" );
      }
      ++numberOfObjectsReadBack;
    }
    
    if ( numberOfObjectsReadBack != 2 * m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of object read back" );
    }
    
    // Retrieving now the derived class
    pool::Collection<MyDerivedClass> myDerivedClassCollection( dataSvc,
							       "ImplicitCollection",
							       "PFN:" + m_databaseFile,
							       m_container,
							       pool::ICollection::READ );
    
    pool::Collection<MyDerivedClass>::Iterator myDerivedClassIterator = myDerivedClassCollection.select();
    
    numberOfObjectsReadBack = 0;
    while ( myDerivedClassIterator.next() ) {
      if ( myDerivedClassIterator->data != 2 * numberOfObjectsReadBack || 
	   myDerivedClassIterator->extraData != 2 * numberOfObjectsReadBack + 1 ) {
	throw std::runtime_error( "Unexpected data in the read object" );
      }
      ++numberOfObjectsReadBack;
    }
    
    if ( numberOfObjectsReadBack != m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of object read back" );
    }
    
    if ( ! dataSvc->transaction().commit() ) {
      throw std::runtime_error( "Could not commit the transaction" );
    }
    fc->commit();
  }

  delete dataSvc;
  delete fc;
}
