#include <stdexcept>

#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"

#include "POOLCore/POOLContext.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"

#include "StorageSvc/DbType.h"

#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/Placement.h"

#include "DataSvc/IDataSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/Ref.h"

#include "Collection/CollectionProxy.h"

#include "TestDriver.h"
#include "MyTestClasses.h"

namespace pool  {
  int numTokenInstances();
}

TestDriver::TestDriver():
  m_databaseFile( "pool_test.root" ),
  m_container( "theContainer" ),
  m_fileCatalog( "FileCatalog.xml" ),
  m_numberOfObjects( 1000 )
{
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

  std::cout << "Loading the dictionary" << std::endl;

  const std::string dictlibrary = "test_ImplicitCollection_UntypedDict";

  try {
    seal::SharedLibrary::load( seal::SharedLibrary::libname( dictlibrary ) );
  }
  catch ( seal::SharedLibraryError *error) {
    throw std::runtime_error( error->explainSelf().c_str() );
  }
}


TestDriver::~TestDriver()
{
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}


void
TestDriver::writeData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }

  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  dataSvc->session().setDefaultConnectionPolicy( policy );

  fc->start();
  if ( ! dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) {
    throw std::runtime_error( "Could not start the transaction" );
  }

  pool::Placement placementHint;
  placementHint.setDatabase( m_databaseFile, pool::DatabaseSpecification::PFN );
  placementHint.setContainerName( m_container );
  placementHint.setTechnology( pool::ROOTKEY_StorageType.type() );

  for ( int i = 0; i < m_numberOfObjects; ++i ) {
    MyBaseClass* myBaseClass = new MyBaseClass;
    myBaseClass->data = i;
    pool::Ref< MyBaseClass > myBaseClassRef( dataSvc, myBaseClass );
    myBaseClassRef.markWrite( placementHint );

    MyDerivedClass* myDerivedClass = new MyDerivedClass;
    myDerivedClass->data = 2*i;
    myDerivedClass->extraData = 2*i + 1;
    pool::Ref< MyDerivedClass > myDerivedClassRef( dataSvc, myDerivedClass );
    myDerivedClassRef.markWrite( placementHint );

    MyOtherClass* myOtherClass = new MyOtherClass;
    myOtherClass->otherData = 3*i;
    pool::Ref< MyOtherClass > myOtherClassRef( dataSvc, myOtherClass );
    myOtherClassRef.markWrite( placementHint );
  }

  if ( ! dataSvc->transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction" );
  }
  fc->commit();

  dataSvc->clearCache();
  delete dataSvc;
  delete fc;
}


void
TestDriver::readData()
{
  std::cout << "Creating a file catalog object" << std::endl;
  pool::IFileCatalog* fc = new pool::IFileCatalog;
  if ( ! fc ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  fc->setWriteCatalog( p.contactstring() );
  fc->connect();

  pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fc );
  if ( ! dataSvc ) {
    throw std::runtime_error( "Could not get an IDataSvc object." );
  }
  else {

    fc->start();
    if ( ! dataSvc->transaction().start( pool::ITransaction::READ ) ) {
      throw std::runtime_error( "Could not start the transaction" );
    }

    pool::CollectionProxy myCollection( dataSvc,
					"ImplicitCollection",
					"PFN:" + m_databaseFile,
					m_container,
					pool::ICollection::READ );

    pool::CollectionProxy::Iterator myIterator = myCollection.select();
    
    int numberOfObjectsReadBack = 0;
    while ( myIterator.next() ) {
      int subType = numberOfObjectsReadBack % 3;
      switch ( subType ) {

      case 0:
	{
	  if ( ! myIterator.isCurrentObjectOfType< MyBaseClass >() ) {
	    throw std::runtime_error( "Object is not of type MyBaseClass" );
	  }
	  if ( myIterator.isCurrentObjectOfType< MyDerivedClass >() ) {
	    throw std::runtime_error( "Object is of type MyDerivedClass" );
	  }
	  if ( myIterator.isCurrentObjectOfType< MyOtherClass >() ) {
	    throw std::runtime_error( "Object is of type MyOtherClass" );
	  }
	  pool::Ref< MyBaseClass > myRef = myIterator.ref< MyBaseClass >();
	  if ( myRef->data != numberOfObjectsReadBack / 3 ) {
	    throw std::runtime_error( "Unexpected data retrieved" );
	  }
	  break;
	}
      case 1:
	{
	  if ( ! myIterator.isCurrentObjectOfType< MyBaseClass >() ) {
	    throw std::runtime_error( "Derived object is not of type MyBaseClass" );
	  }
	  if ( ! myIterator.isCurrentObjectOfType< MyDerivedClass >() ) {
	    throw std::runtime_error( "Object is not of type MyDerivedClass" );
	  }
	  if ( myIterator.isCurrentObjectOfType< MyOtherClass >() ) {
	    throw std::runtime_error( "Object is of type MyOtherClass ! " );
	  }
	  pool::Ref< MyBaseClass > myRef = myIterator.ref< MyBaseClass >();
	  if ( myRef->data != 2 * ( numberOfObjectsReadBack / 3  ) ) {
	    throw std::runtime_error( "Unexpected data retrieved" );
	  }
	  pool::Ref< MyDerivedClass > myRefD = myIterator.ref< MyDerivedClass >();
	  if ( myRefD->data != 2 * ( numberOfObjectsReadBack / 3  ) ) {
	    throw std::runtime_error( "Unexpected data retrieved" );
	  }
	  if ( myRefD->extraData != 2 * ( numberOfObjectsReadBack / 3  ) + 1 ) {
	    throw std::runtime_error( "Unexpected data retrieved" );
	  }
	  break;
	}
      case 2:
	{
	  if ( myIterator.isCurrentObjectOfType< MyBaseClass >() ) {
	    throw std::runtime_error( "Object is of type MyBaseClass !!!" );
	  }
	  if ( myIterator.isCurrentObjectOfType< MyDerivedClass >() ) {
	    throw std::runtime_error( "Object is of type MyDerivedClass !!!" );
	  }
	  if ( ! myIterator.isCurrentObjectOfType< MyOtherClass >() ) {
	    throw std::runtime_error( "Object is not of type MyOtherClass" );
	  }
	  pool::Ref< MyOtherClass > myRef = myIterator.ref< MyOtherClass >();
	  int expectedData = numberOfObjectsReadBack / 3;
	  if ( myRef->otherData != 3 * expectedData ) {
	    throw std::runtime_error( "Unexpected data retrieved" );
	  }
	  break;
	}
      };
      ++numberOfObjectsReadBack;
    }
    
    if ( numberOfObjectsReadBack != 3 * m_numberOfObjects ) {
      throw std::runtime_error( "Invalid number of object read back" );
    }

    if ( ! dataSvc->transaction().commit() ) {
      throw std::runtime_error( "Could not commit the transaction" );
    }
    fc->commit();
  }

  delete dataSvc;
  delete fc;
}
