#ifndef POOL_FCPERFORMANCETESTBASE_H
#define POOL_FCPERFORMANCETESTBASE_H

#include <string>
#include <pthread.h>
#include <exception>

typedef long long TimeT;

namespace pool  {

  /// Forward class declaration
  class URIParser;
  class seal::Exception;

  /**
     Class that is the base for the multithreaded FileCatalog performance tests;
  */
  class FCPerformanceTestBase
  {

  public:

    /// Constructor - as arguments takes number command line parameters and table with the parameters
    FCPerformanceTestBase( int argc, char* argv[] );

    /// Destructor
    virtual ~FCPerformanceTestBase();

    /// Static method that returns number of nanoseconds since the Epoc
    static TimeT universalTime();

    /// Static method that prints info how to run the test. As an argument takes test's name.
    static void printUserInfo( char* appName );

    /// Static method that prints SEAL exception messages. As an argument takes reference to the exception object
    static void printException( seal::Exception& e );
    
    /// Exception thrown when there are wrong command line arguments
    class BadArgumentsException : public std::exception
    {
      public:
        BadArgumentsException() throw ()
        {}
        ~BadArgumentsException() throw()
        {}
        const char* what() const throw()
        { return ""; }
    };

    /// Exception thrown where the POOL_CATALOG env. variable contains unknown catalog's connection string
    class BadCatalogTypeException : public std::exception
    {
      public:
        BadCatalogTypeException( std::string prefix ) throw ()
                 : m_prefix( prefix )
        {}
        ~BadCatalogTypeException() throw()
        {}
        const char* what() const throw()
        { return m_prefix.c_str(); }
      private:
        std::string m_prefix;
    };
    
    /// Exception thrown when there are problems during thread creation
    class ThreadCreationFalseException : public std::exception
    {
      public:
        ThreadCreationFalseException( int number ) throw ()
                 : m_number( "" )
        { char s[10]; sprintf(s, "%d", number); m_number = std::string( s ); }
        ~ThreadCreationFalseException() throw()
        {}
        const char* what() const throw()
        { return m_number.c_str(); }
      private:
        std::string m_number;
    };

    /// Exception thrown when there are problems during thread termination
    class ThreadTerminationFalseException : public std::exception
    {
      public:
        ThreadTerminationFalseException( int number ) throw ()
                 : m_number( "" )
        { char s[10]; sprintf(s, "%d", number); m_number = std::string( s ); }
        ~ThreadTerminationFalseException() throw()
        {}
        const char*  what() const throw()
        { return m_number.c_str(); }
      private:
        std::string m_number;
    };

  protected:
    /**
        Method that prints results collected in the table passed as third parameter.
        Two first parameters specify title and name of the action the results concern
     */
    void printSingleResults( std::string title, std::string action, TimeT* timeTable[] );

    /// Method that prepares file catalog to the test. It is used to avoid errors if the catalog is not empty.
    virtual void cleanCatalog() = 0;
    /// Method that prints all results collected during execution of the test driver.
    virtual void printAllResults() = 0;
    /// Method that runs the test
    virtual void run() = 0;

  protected:

    std::string m_appName;              /// name of the application
    int m_noi;                          /// number of iterations
    int m_noth;                         /// number of threads
    std::string m_nodeName;             /// name of the node where the test driver is executed
    URIParser m_parser;                 /// URIParser handler
    const std::string m_filePrefix;     /// Prefix of used PFNs and LFNs
    pthread_t* m_threads;               /// table of thread handlers
    int m_threadID;                     /// Variable used by threads to get their id number

  private:

    /// Method that parsing command line arguments. Returns 'true' if everything is ok and 'false' in the other case.
    bool parseArguments( int argc, char* argv[] );

  };


}

#endif // POOL_FCPERFORMANCETESTBASE_H

