#include <iostream>
#include <memory>

#include "TestBase/TestBase.h"

#include <stdexcept>
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/RefException.h"
#include "DataSvc/SharedAnyPtr.h"
#include "FileCatalog/IFileCatalog.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ISession.h"

#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFCAction.h"
#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"

void pool::TestBase::init_(const std::string& catalogFileName){
  try{
    std::cout << "Loading the SEAL message service" << std::endl;
    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
    std::cout << "Setting the verbosity level to Info" << std::endl;
    pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );

    std::cout << "initializing catalog" <<std::endl;
    // initialize catalog
    std::string contStr = "file:"+catalogFileName;
    if (!catalogFileName.length()) contStr = contStr+"test_catalog.xml";
    URIParser pars(contStr);
    pars.parse();
    m_fileCatalog = new pool::IFileCatalog;
    try {
       m_fileCatalog->setWriteCatalog( pars.contactstring() );
    } catch (const seal::Exception& e){
      std::cout << "ERROR: " << e.what() << std::endl;
    } catch (const std::exception& e){
      std::cout << "ERROR: " << e.what() << std::endl;
    } 
    m_fileCatalog->connect();
    m_fileCatalog->start();
    // initialize dataSvc
    m_dataSvc = DataSvcFactory::instance( m_fileCatalog );
    m_init_test = true;
    pool::DatabaseConnectionPolicy policy;
    policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
    policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
    m_dataSvc->session().setDefaultConnectionPolicy( policy );
  } catch (const std::exception& e){
    std::cerr << "test ERROR:" << e.what() << "\n";
    addError();
  }
}

/// Standard Constructor
pool::TestBase::TestBase(const std::string& catalogFileName) : TestKernel(), m_dataSvc(0),m_fileCatalog(0),m_init_test(false){
  init_(catalogFileName); 
}

/// Standard Constructor
pool::TestBase::TestBase() : TestKernel(), m_dataSvc(0),m_fileCatalog(0),m_init_test(false){
  init_(""); 
}

/// Standard Destructor
pool::TestBase::~TestBase(){
  delete m_dataSvc;
  m_fileCatalog->commit();
  m_fileCatalog->disconnect();
  delete m_fileCatalog;
}
    
    /// Utility function to check for errors during execution
void pool::TestBase::checkError(seal::Status sc, const std::string& msg)    {
  if ( !sc.isSuccess() )  {
    std::cout << "test Error: " << msg << std::endl;
    addError();
  }
}

//// Utility function to print exception messages
void pool::TestBase::printException(const pool::Exception& e){
  std::cout << "Exception: " << e.what() << "\n";
  addError();
}

std::string pool::TestBase::fileId(const std::string& pfn,
                                   const pool::DbType& technology){
  std::string fid, tech;
  // try to re-use an already existing file
  pool::IFCAction action;
  m_fileCatalog->setAction( action );
  action.lookupFileByPFN(pfn, fid, tech);
  // did not exist - so try register the file
  if(fid.length()==0){
    pool::DbType majorType(technology.majorType());
    action.registerPFN(pfn,majorType.storageName(),fid);
    if ( fid.length() == 0 )    
      throw std::runtime_error("failed to register file `" + pfn +"'");
  }
  return fid;
}

void pool::TestBase::init( const std::vector< std::string >& libraries ){

  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
	iLibrary != libraries.end(); ++iLibrary ) {
    std::string library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;
    try {  // This is temporary. SEAL should get better in this.
      seal::SharedLibrary::load( seal::SharedLibrary::libname( library ) );
    }
    catch ( const seal::SharedLibraryError& error) 
    {
      std::cout << "test ERROR: "<< error.explainSelf() << std::endl;
    }
    catch (const seal::Exception &e) 
    {
      std::cout << "test ERROR: "<< e.what() << std::endl;
    }
    catch (const std::exception &e) 
    {
      std::cout << "test ERROR: "<< e.what() << std::endl;
    }
  }
}

void pool::TestBase::setDbLogin( const std::string& userName,const std::string passwd){
  pool::POOLContext::loadComponent( "POOL/Services/EnvironmentAuthenticationService" );
  const std::string userNameEnv = "POOL_AUTH_USER=" + userName;
  ::putenv( const_cast<char*>( userNameEnv.c_str() ) );
  const std::string passwdEnv = "POOL_AUTH_PASSWORD=" + passwd;
  ::putenv( const_cast<char*>( passwdEnv.c_str() ) );
}


namespace pool  {
  int numTokenInstances();
}

void pool::TestBase::run(){
  if(m_init_test){
    try   {
      execute();    
      // disconnect 
      m_dataSvc->session().disconnectAll();
      std::cout << OVALTAG << "Number of token instances: " << 
      pool::numTokenInstances() << std::endl;
      unsigned int nobj = pool::SharedAnyPtr::numberOfInstances();
      std::cout << OVALTAG << "Number of object instances: " 
                << nobj << std::endl;
      if(nobj){
        pool::SharedAnyPtr::dumpRegister();
      }
    }
    catch(const seal::Exception& e)   {
      std::cout << "The test has thrown a (SEAL) exception:" << e.what() << std::endl;
      addError();
    }
    catch(const std::runtime_error& e)   {
      std::cout << "The test has thrown a (runtime) exception:" << e.what() << std::endl;
      addError();
    }
    catch(const std::exception& e)   {
      std::cout << "The test has thrown a (std) exception:" << e.what() << std::endl;
      addError();
    }
    catch(...)    {
      std::cout << "The test has thrown an unknown exception." << std::endl;
      addError();
    }
  }
  dumpError();
}

void pool::TestBase::setMessageVerbosityLevel( seal::Msg::Level level ){
  pool::POOLContext::setMessageVerbosityLevel( level );
}

