#include "ObjectCallBack.h"
#include "StorageSvc/DataHandler.h"
#include "StorageSvc/Shape.h"
#include "StorageSvc/DbColumn.h"
#include "PersistentOriginObject.h"
#include "OriginObject.h"
#include "ITokenCache.h"
#include <vector>


ObjectCallBack::ObjectCallBack( ITokenCache& cache ):
  m_cache( cache ),
  m_transientType( ROOT::Reflex::Type::ByName( "OriginObject" ) ),
  m_dataHandler( 0 ),
  m_shape( 0 ),
  m_transientObjectData( 0 ),
  m_persistentObjectData( 0 )
{}

ObjectCallBack::~ObjectCallBack()
{}

pool::DataCallBack*
ObjectCallBack::clone() const
{
  ObjectCallBack* p = new ObjectCallBack( m_cache );
  p->m_dataHandler = m_dataHandler;
  p->m_shape = m_shape;
  return static_cast< pool::DataCallBack* >( p );
}


ROOT::Reflex::Type
ObjectCallBack::transientType() const
{
  return m_transientType;
}


void
ObjectCallBack::setObject( const void* ptr )
{
  m_transientObjectData = ptr;
}

const void*
ObjectCallBack::object() const
{
  return m_transientObjectData;
}

void
ObjectCallBack::setShape( const pool::Shape* shape )
{
  m_shape = shape;
}

const pool::Shape*
ObjectCallBack::shape() const
{
  return m_shape;
}

void
ObjectCallBack::setHandler( pool::DataHandler* pSvc )
{
  m_dataHandler = pSvc;
}

pool::DataHandler*
ObjectCallBack::handler() const
{
  return m_dataHandler;
}


pool::DbStatus
ObjectCallBack::start( pool::DataCallBack::CallType action_type,
		       void* data,
		       void** context)
{
  switch(action_type) {
  case GET:
    {
      // Do not allocate space for the persistent data buffer.
      // The back end implementation will do that.
      m_persistentObjectData = 0;
    }
    break;
  case PUT:
    // Check if the data corresponds to the transient data pointer
    if ( data != m_transientObjectData ) {
      // Input data is different from the transient data
      return pool::Error;
    }
    else {
      // allocate space for the persistent data buffer
      PersistentOriginObject* pc = new PersistentOriginObject();
      if ( ! pc ) return pool::Error;
      // Copy the relevant data from the transient class to the persistent one.
      const OriginObject* tc = static_cast< const OriginObject* >( m_transientObjectData );
      pc->label = tc->label;
      for ( std::vector< DataLink >::const_iterator iData = tc->data.begin();
	    iData != tc->data.end(); ++iData ) {
	const DataLink& dataLink = *iData;
	DataRef ref;
	ref.m_key = dataLink.key();
	pc->data.push_back( ref );
      }
      m_persistentObjectData = pc;
    }
    break;
  default:
    return pool::Error;
  };
  *context = m_persistentObjectData;
  return pool::Success;
}


pool::DbStatus
ObjectCallBack::bind( pool::DataCallBack::CallType action_type,
		      const pool::DbColumn* col_identifier,
		      int /* col_number */,
		      void* context,
		      void** data )
{
  switch(action_type)  {
  case GET:
    *data = &m_persistentObjectData;
    break;
  case PUT:
    /* This works for ROOT Trees but not for ROOT keys !!!
    if ( context != m_persistentObjectData ) {
      std::cerr << "Context different than persistent data" << std::endl;
      return pool::Error;
    }
    */
    *data = static_cast< char* >( context ) + col_identifier->offset();
    break;
  default:
    return pool::Error;
  };
  return pool::Success;
}


pool::DbStatus
ObjectCallBack::end( pool::DataCallBack::CallType action_type,
		     void* /* context */ )
{
  switch(action_type)  {
  case GET:
    {
      // Create and fill the corresponding transient object
      OriginObject* tc = new OriginObject();
      PersistentOriginObject* pc = static_cast< PersistentOriginObject* >( m_persistentObjectData );
      tc->label = pc->label;
      for ( std::vector< DataRef >::iterator iData = pc->data.begin();
	    iData != pc->data.end(); ++iData ) {
	DataRef& dataRef = *iData;
	m_cache.registerKey( dataRef.m_ref.index(), dataRef.m_key );
	DataLink dataLink;
	dataLink.setLink( &m_cache, dataRef.m_key );
	tc->data.push_back( dataLink );
      }
      m_transientObjectData = tc;
      // free the persistent data buffer
      delete pc;
      m_persistentObjectData = 0;
      break;
    }
  case PUT:
    // free the persistent data buffer
    delete static_cast< PersistentOriginObject* >( m_persistentObjectData );
    m_persistentObjectData = 0;
    break;
  default:
    return pool::Error;
  };
  return pool::Success;
}


pool::DbStatus
ObjectCallBack::notify( pool::DataCallBack::CallType action_type,
			int data_type,
			ROOT::Reflex::Type reflection,
			void* object_ptr,
			void* data )
{
  if ( m_dataHandler ) {
    switch( action_type )   {
    case GET:
      return m_dataHandler->onLoad(data_type, reflection, object_ptr, data);
    case PUT:
      return m_dataHandler->onSave(data_type, reflection, object_ptr, data);
    default:
      return pool::Error;
    };
  }
  return pool::Success;
}
