#include "TestDriver.h"
#include "TestClasses.h"
#include "POOLCore/POOLContext.h"
#include "RelationalAccess/IConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITransaction.h"
#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "Collection/Collection.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/poolDb.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/ISession.h"
#include "DataSvc/Ref.h"
#include "DataSvc/RefException.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/IFCContainer.h"
#include "FileCatalog/FCEntry.h"
#include "FileCatalog/FCException.h"
#include <cstdlib>
#include <stdexcept>
#include "ObjectRelationalAccess/ObjectRelationalMappingUtilities.h"
static std::string testClassLibrary("test_RelationalStorageService_BlobsWriteReadDict");
static std::string inputFile("input.xml");
static std::string outputFile("output.xml");

pool::TestDriver::TestDriver( const std::string& conn):TestBase(),
                                                        m_connection(conn){
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );  
  std::vector<std::string> libs;
  libs.push_back(testClassLibrary);
  init(libs);
  pool::POOLContext::loadComponent( "POOL/TestServices/BlobStreamingService" );  
}

pool::TestDriver::~TestDriver(){
}

void pool::TestDriver::runCustomMapping(){
  std::cout << "##### processing custom mapping " << std::endl;
  pool::ObjectRelationalMappingUtilities utility;
  utility.connect(m_connection,coral::Update);
  utility.session()->transaction().start();
  utility.buildAndMaterializeMapping(inputFile,outputFile, false);
  utility.session()->transaction().commit();
  std::cout << OVALTAG <<  "Mapping built and stored." <<std::endl; 
}

void pool::TestDriver::createDatabase(unsigned int nobjects){
  std::cout << "##### creating and filling POOL Database " << std::endl;  
  
  pool::Placement place1, place2;
  place1.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place1.setContainerName("Cont1");
  place1.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());
  place2.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place2.setContainerName("Cont2");
  place2.setTechnology(pool::POOL_RDBMS_POLYMORPHIC_StorageType .type());
  pool::Placement place3;
  place3.setDatabase(m_connection, pool::DatabaseSpecification::PFN );
  place3.setContainerName("Cont3");
  place3.setTechnology(pool::POOL_RDBMS_HOMOGENEOUS_StorageType.type());

  std::cout << "Starting UPDATE transaction..." << std::endl;
  // Start an update transaction
  if ( ! ( m_dataSvc->transaction().start( pool::ITransaction::UPDATE ) ) ) {
    printError("test ERROR: Could not start an UPDATE transaction");
  }  

  unsigned int i;
  for (i = 0; i < nobjects; ++i )   {
    pool::Ref<SimpleClass> simple(m_dataSvc,new SimpleClass(i));
    pool::Ref<OtherClass> other(m_dataSvc,new OtherClass(i));
    pool::Ref<DataClass> data(m_dataSvc,new DataClass(i));
    pool::Ref<SiStripNoises> sstrip(m_dataSvc,new SiStripNoises(i));

    try{
      simple.markWrite(place1);
    } catch (const pool::RefException& e){
      printException(e);        
    }    
    try{
      other.markWrite(place2);
    } catch (const pool::RefException& e){
      printException(e);
    }
    try{
      data.markWrite(place2);
    } catch (const pool::RefException& e){
      printException(e);
    }
    try{
      sstrip.markWrite(place3);
    } catch (const pool::RefException& e){
      printException(e);
    }
  }
      
  std::cout << "Commit transaction..." << std::endl;
  if ( ! m_dataSvc->transaction().commit() ) {
     printError("test ERROR: Could not commit the WRITE transaction." );
  }
  std::cout << OVALTAG << "Written on disk "<< i << " objects\n";

  m_dataSvc->session().disconnectAll();

}

void pool::TestDriver::cleanUp(){
  std::cout << "##### cleaning up existing schema " << std::endl;

  seal::Context* ctx = pool::POOLContext::context();
  seal::IHandle<coral::IConnectionService> serviceHandle = ctx->query<coral::IConnectionService>( "CORAL/Services/ConnectionService" );
  if ( ! serviceHandle ) {
    seal::Handle<seal::ComponentLoader> loader = ctx->component<seal::ComponentLoader>();
    if(!loader) loader = new seal::ComponentLoader( ctx );
    loader->load( "CORAL/Services/ConnectionService" );  
    serviceHandle = ctx->query<coral::IConnectionService>( "CORAL/Services/ConnectionService" );
  }
  if ( ! serviceHandle ) {
    printError("TEST ERROR: Could not load ConnectionService.");
    return; 
  }
  coral::ISessionProxy* session = serviceHandle->connect( m_connection, coral::Update );
  coral::ISchema& schema = session->nominalSchema();
  session->transaction().start();
  schema.dropIfExistsTable( "POOL_OR_MAPPING_COLUMNS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_ELEMENTS" );
  schema.dropIfExistsTable( "POOL_OR_CLASS_VERSIONS" );
  schema.dropIfExistsTable( "POOL_OR_MAPPING_VERSIONS" );
  schema.dropIfExistsTable( "POOL_CONT_OTHER" );
  schema.dropIfExistsTable( "POOL_CONT___LINKS" );
  schema.dropIfExistsTable( "POOL_CONT___PARAMS" );
  schema.dropIfExistsTable( "POOL_CONT___SHAPES" );
  schema.dropIfExistsTable( "POOL_RSS_CONTAINERS" );
  schema.dropIfExistsTable( "POOL_RSS_DB" );
  schema.dropIfExistsTable( "POOL_RSS_SEQ" );
  schema.dropIfExistsTable( "SIMPLECLASS_M_ARRA_1" );
  schema.dropIfExistsTable( "SIMPLECLASS_M_NEST_2" );
  schema.dropIfExistsTable( "SIMPLECLASS_M_NEST_1" );
  schema.dropIfExistsTable( "SIMPLECLASS_M_ENTR_1" );
  schema.dropIfExistsTable( "SIMPLECLASS" );
  schema.dropIfExistsTable( "OTHERCLASS_M_ARRAY_1" );
  schema.dropIfExistsTable( "OTHERCLASS" );
  schema.dropIfExistsTable( "DATACLASS" );
  schema.dropIfExistsTable( "T_SISTRIPNOISES_INDEX" );
  schema.dropIfExistsTable( "SISTRIPNOISES" );
  schema.dropIfExistsTable( "POOL_CONT_CONT2" );
  session->transaction().commit();
  delete session;
  std::cout << OVALTAG <<  "All POOL Database Tables removed." <<std::endl; 
}

void pool::TestDriver::readBackData(unsigned int expected){

  std::cout << "##### reading back data from POOL database " << std::endl;

  // Start a read transaction
  m_dataSvc->transaction().start( pool::ITransaction::READ );
  pool::Collection< SimpleClass > simpleCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connection,
                                                   "Cont1",
                                                   pool::ICollection::READ );
  pool::Collection< SimpleClass >::Iterator simpleObj = simpleCollection.select();

  unsigned int numberOfObj = 0;
  while ( simpleObj.next() ) {
    int seed = simpleObj.ref()->m_intData;
    SimpleClass r(seed);
    if(simpleObj.ref()->m_arrayData.size()!=r.m_arrayData.size())
    {
      printError("TEST ERROR: SimpleClass objects read back different for vector content.");
      std::cout <<"Size of SimpleClass obj readout="<< simpleObj.ref()->m_arrayData.size()<<std::endl;
      std::cout <<"Size of SimpleClass obj expected="<< r.m_arrayData.size()<<std::endl;
    }
    if(simpleObj.ref()->operator!=(r)){
      printError("TEST ERROR: SimpleClass object read back different from expected.");
      //break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type SimpleClass" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of SimpleClass objects read back different from expected.");
  }
  
  pool::Collection< OtherClass > otherCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connection,
                                                   "Cont2",
                                                   pool::ICollection::READ );
  pool::Collection< OtherClass >::Iterator otherObj = otherCollection.select();

  numberOfObj = 0;
  while ( otherObj.next() ) {
    int seed = otherObj.ref()->m_intData;
    OtherClass r(seed);
    if(otherObj.ref()->m_arrayData.size()!=r.m_arrayData.size())
    {
      printError("TEST ERROR: OtherClass objects read back different for vector content.");
      std::cout <<"Size of OtherClass obj readout="<< otherObj.ref()->m_arrayData.size()<<std::endl;
      std::cout <<"Size of OtherClass obj expected="<< r.m_arrayData.size()<<std::endl;
    }
    if(otherObj.ref()->operator!=(r)){
      printError("TEST ERROR: OtherClass object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type OtherClass" <<std::endl;
                                
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of OtherClass objects read back different from expected.");
  }

  pool::Collection< DataClass > dataCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connection,
                                                   "Cont2",
                                                   pool::ICollection::READ );
  pool::Collection< DataClass >::Iterator dataObj = dataCollection.select();

  numberOfObj = 0;
  while ( dataObj.next() ) {
    int seed = dataObj.ref()->m_i;
    DataClass r(seed);
    if(dataObj.ref()->operator!=(r)){
      printError("TEST ERROR: DataClass object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type DataClass" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of DataClass objects read back different from expected.");
  }

  pool::Collection< SiStripNoises > siStripCollection( m_dataSvc,
                                                   "ImplicitCollection",
                                                   "PFN:" + m_connection,
                                                   "Cont3",
                                                   pool::ICollection::READ );
  pool::Collection< SiStripNoises >::Iterator siStripObj = siStripCollection.select();

  numberOfObj = 0;
  while ( siStripObj.next() ) {
    int seed = siStripObj.ref()->indexes.back().detid;
    SiStripNoises r(seed);
    if(siStripObj.ref()->operator!=(r)){
      printError("TEST ERROR: SiStripNoises object read back different from expected.");
      break;
    }
    numberOfObj++;
  }
  std::cout << OVALTAG << "Read back " <<numberOfObj<< " object of type SiStripNoises" <<std::endl;
                   
  if(numberOfObj!=expected){
    printError("TEST ERROR: number of SiStripNoises objects read back different from expected.");
  }
  // commit the transaction
  m_dataSvc->transaction().commit();
  m_dataSvc->session().disconnectAll();

}
#ifdef _WIN32
#include <windows.h>
#endif

void pool::TestDriver::execute(){ 
  cleanUp();
  runCustomMapping();
  createDatabase(20);
#ifdef _WIN32
  ::Sleep(1000);
#else
  sleep(1);
#endif
  readBackData(20);
  cleanUp();
}
