#include "TestDriver.h"
#include "TestDictionary/SimpleTestClass.h"

#include "PluginManager/PluginManager.h"
#include "SealKernel/Exception.h"
#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"
#include "Reflex/Type.h"

#include "POOLCore/Guid.h"
#include "POOLCore/Token.h"
#include "POOLCore/POOLContext.h"

#include "StorageSvc/Shape.h"
#include "StorageSvc/poolDb.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageSvcFactory.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DatabaseConnection.h"
#include "StorageSvc/FileDescriptor.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbObjectCallBack.h"

#include <stdexcept>
#include <iostream>
#include <sstream>
#include <memory>

namespace pool { int numTokenInstances(); }

static const std::string file = "pool_test.root";
static const std::string container = "container";
static const int nObjects = 100;


TestDriver::TestDriver()
{
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel( seal::Msg::Info );
}

TestDriver::~TestDriver()
{
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}

void
TestDriver::loadLibraries( const std::vector<std::string>& libraries )
{
  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
	iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;
    try {
      seal::SharedLibrary::load( seal::SharedLibrary::libname( library ) );
    }
    catch ( seal::SharedLibraryError *error) 
    {
      throw std::runtime_error( error->explainSelf().c_str() );
    }
  }
}

void
TestDriver::testWriting()
{
  pool::IStorageSvcFactory* fac = pool::IStorageSvcFactory::get();
  if ( ! fac ) {
    throw std::runtime_error( "Could not retrieve a StorageSvcFactory" );
  }
  pool::IStorageSvc* storSvc = fac->create("pool_DbStorageSvc", 0, "StorageSvc");
  if ( ! storSvc ) {
    throw std::runtime_error( "Could not create a StorageSvc object" );
  }
  storSvc->addRef();
  pool::Session* sessionHandle = 0;
  if ( ! ( storSvc->startSession( poolDb::CREATE, pool::ROOT_StorageType.type(), sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a session." );
  }

  pool::FileDescriptor* fd = new pool::FileDescriptor( file, file );
  if ( ! ( storSvc->connect( sessionHandle, poolDb::CREATE, *fd ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a connection." );
  }
  pool::DatabaseConnection* connection = fd->dbc();
  pool::Transaction* transaction = 0;
  if ( ! ( storSvc->startTransaction( connection, transaction ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a transaction." );
  }

  // Retrieve the dictionary
  ROOT::Reflex::Type class_SimpleTestClass = ROOT::Reflex::Type::ByName( "SimpleTestClass" );
  if ( ! class_SimpleTestClass ) {
    throw std::runtime_error( "Could not retrieve the dictionary for class SimpleTestClass" );
  }
  //std::string guid_nam = class_SimpleTestClass->propertyList()->getProperty( "ClassID" );
  //if ( guid_nam.empty() ) {
  //  std::ostringstream error;
  //  error << "There is no ClassID property for class \"SimpleTestClass\"" << std::ends;
  //  throw std::runtime_error( error.str() );
  //}
  //pool::Guid guid(guid_nam);

  std::vector< SimpleTestClass* > myObjects;
  for ( int i = 0; i < nObjects; ++i ) {
    myObjects.push_back( new SimpleTestClass() );
    SimpleTestClass* myObject = myObjects.back();
    myObject->data = i;

    // Creating a callback object
    std::auto_ptr<pool::DataCallBack> cb( new pool::DbObjectCallBack( class_SimpleTestClass ) );
    cb->setObject( myObject );

    // Creating the persistent shape.
    const pool::Shape* shape = 0;
    if ( storSvc->getShape( *fd,class_SimpleTestClass , shape ) == pool::IStorageSvc::SHAPE_NOT_AVAILIBLE ) {
      storSvc->createShape( *fd, container, class_SimpleTestClass, shape );
    }
    if ( ! shape ) {
      throw std::runtime_error( "Could not create a persistent shape." );
    }
  
    // Writing the object.
    pool::Token* token;
    if ( ! ( storSvc->allocate( transaction, *fd,
				container, pool::ROOTTREE_StorageType.type(),
				shape, cb.get(), token ).isSuccess() ) ) {
      throw std::runtime_error( "Could not write an object" );
    }
    //token->setClassID( guid );
    cb.release();
    delete token;
  }

  // Closing the transaction.
  if ( ! ( storSvc->endTransaction( transaction, pool::Transaction::TRANSACT_COMMIT ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end a transaction." );
  }

  // Clearing the cache
  for ( std::vector< SimpleTestClass* >::iterator iObject = myObjects.begin();
	iObject != myObjects.end(); ++iObject ) delete *iObject;
  myObjects.clear();

  if ( ! ( storSvc->disconnect( *fd ).isSuccess() ) ) {
    throw std::runtime_error( "Could not disconnect." );
  }
  delete fd;
  if ( ! ( storSvc->endSession( sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end correctly the session." );
  }
  storSvc->release();
}


void
TestDriver::testReading()
{
  pool::IStorageSvcFactory* fac = pool::IStorageSvcFactory::get();
  if ( ! fac ) {
    throw std::runtime_error( "Could not retrieve a StorageSvcFactory" );
  }
  pool::IStorageSvc* storSvc = fac->create("pool_DbStorageSvc", 0, "StorageSvc");
  if ( ! storSvc ) {
    throw std::runtime_error( "Could not create a StorageSvc object" );
  }
  storSvc->addRef();
  void* pVoid = 0;
  seal::Status sc = storSvc->queryInterface( pool::IStorageExplorer::interfaceID(), &pVoid );
  pool::IStorageExplorer* storageExplorer = (pool::IStorageExplorer*)pVoid;
  if ( !( sc == seal::Status::SUCCESS && storageExplorer ) ) {
    storSvc->release();
    throw std::runtime_error( "Could not retrieve a IStorageExplorer interface" );
  }

  pool::Session* sessionHandle = 0;
  if ( ! ( storSvc->startSession( poolDb::READ, pool::ROOT_StorageType.type(), sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not start a session." );
  }

  pool::FileDescriptor* fd = new pool::FileDescriptor( file, file );
  sc = storSvc->connect( sessionHandle, poolDb::READ, *fd );
  if ( sc != seal::Status::SUCCESS ) {
    throw std::runtime_error( "Could not start a connection." );
  }

  // Fetch the containers
  std::vector<const pool::Token*> containerTokens;
  storageExplorer->containers( *fd, containerTokens );
  if ( containerTokens.size() != 1 ) {
    throw std::runtime_error( "Unexpected number of containers" );
  }
  const pool::Token* containerToken = containerTokens.front();
  const std::string containerName = containerToken->contID();
  if ( containerName != container ) {
    throw std::runtime_error( "Container name read is different from the container name written" );
  }

  // Fetch the objects in the container.
  pool::DbSelect selectionObject("");
  sc = storageExplorer->select( *fd, containerToken->contID(), selectionObject );
  int iObject = 0;
  if ( sc.isSuccess() ) {
    pool::Token* objectToken = 0;
    while ( storageExplorer->next( selectionObject, objectToken ).isSuccess() ) {
      const pool::Guid& guid = objectToken->classID();
      const pool::Shape* shape = 0;
      if ( storSvc->getShape( *fd, guid, shape ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE ) {
	throw std::runtime_error( "Could not fetch the persistent shape" );
      }
      ROOT::Reflex::Type classType = pool::DbReflex::forGuid(guid);
      if(!classType){
	throw std::runtime_error( "Could not resolve the class by guid" );
      }
      pool::DataCallBack* cb = new pool::DbObjectCallBack( classType );
      cb->setHandler( 0 );
      ROOT::Reflex::Type classH;
      if ( ! ( storSvc->read( *fd, *objectToken, cb, shape, classH ) ).isSuccess() ) {
	throw std::runtime_error( "failed to read an object back from the persistency" );
      }

      if ( classH.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) != "SimpleTestClass" ) {
	throw std::runtime_error( "read wrong class type." );
      }

      SimpleTestClass* object = static_cast< SimpleTestClass* >( const_cast<void*>( cb->object() ) );
      if ( object->data != iObject ) {
	throw std::runtime_error( "Object read different from object written" );
      }
      delete object;
      delete cb;
      ++iObject;
      objectToken->release();
    }
  }
  if ( iObject != nObjects ) {
    throw std::runtime_error( "Objects read different from objects written" );
  }


  if ( ! ( storSvc->disconnect( *fd ).isSuccess() ) ) {
    throw std::runtime_error( "Could not disconnect." );
  }
  delete fd;

  std::cout << "Closing the session" << std::endl;
  if ( ! ( storSvc->endSession( sessionHandle ).isSuccess() ) ) {
    throw std::runtime_error( "Could not end correctly the session." );
  }
  storageExplorer->release();
  storSvc->release();
}
