/**
 * @file CollAppend.cpp
 * @brief Utility to append POOL collections to another POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollAppend.cpp,v 1.34 2008/02/12 16:58:04 mnowak Exp $ 
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/ICollectionDataEditor.h"

#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"
#include "CmdLineArgs.h"

#include "CoralBase/Attribute.h"

#include <iostream>
#include <string>
#include <map>

using namespace std;
using namespace pool;


int main(int argc, const char *argv[])
{
   try
   {
      bool noAttrib = false;
      std::vector<std::string> srcCollNameVec;
      std::vector<std::string> srcCollTypeVec;
      std::vector<std::string> srcCollConnectVec;
      std::vector<std::string> dstCollNameVec;
      std::vector<std::string> dstCollTypeVec;
      std::vector<std::string> dstCollConnectVec;
      std::vector<bool> dstCollExistVec;
      std::vector<std::string> collCatalogReadVec;
      std::vector<int> srcCountVec;

      CmdLineArgs cmdLineArgs(argc, argv);

      if( cmdLineArgs.hasQual("-help") ||
	  !cmdLineArgs.hasQual("-src") ||
	  !cmdLineArgs.hasQual("-dst") )
      {
	 std::cerr << "Usage: CollAppend "<< endl
		   << "<-src [+]<<input collection name> <input collection type>> > "
		   << "(List of name-type pairs is space separated; "
		   << "Choices for collection types are RootCollection, "
		   << "RelationalCollection, "
		   << "PHYSICAL_NAME, LOGICAL_NAME or GUID) "<< endl
		   << "<-dst [+]<<output collection name> <output collection type>> > "
		   << "(NOTE: output collections must already exist) "
		   << "Optional arguments: "<< endl
		   << "[-srcconnect [Input database connection string]] "
		   << "(Note: The default value is \"\" and this argument MUST "
		   << "be specified for RelationalCollection. Also, the environment " 
		   << "variables POOL_AUTH_USER and POOL_AUTH_PASSWORD "
		   << "must be set, or CORAL authentication must be used) "<< endl
		   << "[-dstconnect [Output database connection string]] "
		   << "(Note: The default value is \"\" and for "
		   << "RelationalCollection this will be set to the value of "
		   << "the input database connection string if no value is "
		   << "specified. In that case, the input database connection "
		   << "string MUST be specified and satisfy the conditions "
		   << "described above) "
		   << "[-query [predicate string (DEFAULT = \"\")]] " << endl
		   << "[-queryopt <comma separated attribute list> "
		   << "(DEFAULT = * ) "<< endl
		   << "[-ccread [+][read collection catalog connection string "
		   << "(DEFAULT = \"\")]] "<< endl
		   << "[-noattrib] (no attributes copied, only event tokens) " << endl
		   << "[-nevtpercommit [Max. number of events to process between "
		   << "collection commits (DEFAULT = 10000)]] " << endl
		   << "[-nevtperprint [Number of events to process before "
		   << "each print of status to the screen. \"all\" => "
		   << "skip printing for all events (DEFAULT = \"all\")]]"<< endl
		   << "[-nevtcached [size of the insert buffer for bulk operations "
		   << "DEFAULT = 0 (no bulk operations)]]" << endl
		   << "[-ignoreschemadiff] - attempts to copy collections even if schema seems incompatible (use at your own risk)"
		   << std::endl;
	 return 1;
      }

      std::string inputConnection = cmdLineArgs.getArg( "-srcconnect" );
      // ---
      // Input is -src <name> <type>
      // ==> check that input is specified
      // ---
      std::string srcCollName = "";
      srcCollName = cmdLineArgs.getArg( "-src" );
      if( !srcCollName.size() || srcCollName[0] == '-' )  {
	 std::cerr << "CollAppend: ERROR No input collection specified on " 
		   << "command line." << std::endl;
	 return 1;
      }
      // ---
      // Now check the type argument
      // ---
      std::string srcCollType = "";
      std::string srcCollConnect = "";
      while( srcCollName.size() && srcCollName[0] != '-' )  {
	 srcCollType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
         // First check if it is a RelationalCollection
         //   or an old, deprecated, not to be long supported type of rdb format
	 if( srcCollType == "MySQLCollection" || 
	     srcCollType == "MySQLltCollection" ||
	     srcCollType == "RelationalCollection" )
	 {
	    if( inputConnection != "" ) {
	       srcCollConnect = inputConnection;
	    } else {
	       std::cerr << "CollAppend: ERROR Must specify an input database "
			 << "connection string for collections of type " 
			 << "MySQL or Oracle." << std::endl;
	       return 1;
	    }
	 }
         // Or is it a file based or logical type collection
	 else if( srcCollType == "RootCollection" || 
		  srcCollType == "PHYSICAL_NAME" ||
		  srcCollType == "LOGICAL_NAME" || 
		  srcCollType == "GUID" )
	 {
	    srcCollConnect = "";
	 }
         // Or is it gibberish, and they need to try again.
	 else {
	    std::cerr << "CollAppend: ERROR Collection type " << srcCollType 
		      << " is not presently handled by this program." 
		      << std::endl;
	    return 1;
	 }
	 srcCollNameVec.push_back( srcCollName );
	 srcCollTypeVec.push_back( srcCollType );
	 srcCollConnectVec.push_back( srcCollConnect );
	 srcCollName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      std::string outputConnection = cmdLineArgs.getArg( "-dstconnect" );

      // ---
      // Now that we have the input settled, let's look for the output, -dst, target
      // ---
      std::string dstCollName = "";
      dstCollName = cmdLineArgs.getArg( "-dst" );
      if( !dstCollName.size() || dstCollName[0] == '-' )  {
	 std::cerr << "CollAppend: ERROR No output collection specified on " 
		   << "command line." << std::endl;
	 return 1;
      }
      // ---
      // Now check the type argument
      // ---
      std::string dstCollType = "";
      std::string dstCollConnect = "";
      while( dstCollName.size() && dstCollName[0] != '-' )  {
	 dstCollType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
         // First check if it is a RelationalCollection
         //   or an old, deprecated, not to be long supported type of rdb format
	 if( dstCollType == "MySQLCollection" || 
	     dstCollType == "MySQLltCollection" ||
	     dstCollType == "RelationalCollection" )
	 {
	    if( outputConnection != "" ) {
	       dstCollConnect = outputConnection;
	    }
	    else {
	       if( inputConnection != "" ) {
		  dstCollConnect = inputConnection;
	       }
	       else {
		  std::cerr << "CollAppend: ERROR Must specify an output database "
			    << "connection string for collections of type " 
			    << "MySQL or Oracle unless an input database " 
			    << "connection string was already specified." 
			    << std::endl;
		  return 1;
	       }
	    }
	 }
         // Or is it a file based or logical type collection
	 else if ( dstCollType == "RootCollection" || 
		   dstCollType == "PHYSICAL_NAME" ||
		   dstCollType == "LOGICAL_NAME" || 
		   dstCollType == "GUID" )
	 {
	    dstCollConnect = "";
	 }
         // Or is it gibberish, and they need to try again.
	 else {
	    std::cerr << "CollAppend: ERROR Collection type " << dstCollType 
		      << " is not presently handled by this program." 
		      << std::endl;
	    return 1;
	 }
	 dstCollNameVec.push_back( dstCollName );
	 dstCollTypeVec.push_back( dstCollType );
	 dstCollConnectVec.push_back( dstCollConnect );
	 dstCollExistVec.push_back( true );
	 dstCollName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      // ---
      // Is there a query predicate associated
      // ---
      std::string query = cmdLineArgs.getArg( "-query" );

      std::string queryOptions = "";
      if( cmdLineArgs.hasQual( "-queryopt" ) ) {
	 queryOptions = cmdLineArgs.getArg("-queryopt");
	 if( queryOptions[0] == '-' )  {
	    queryOptions = "";
	 }
      }

      // ---
      // Is there a collection catalog specified
      // ---
      std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
      while( collCatalogRead.size() && collCatalogRead[0] != '-' )  {
	 collCatalogReadVec.push_back( collCatalogRead );
	 collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      if ( cmdLineArgs.hasQual("-noattrib") ) {
	 noAttrib = true;
      }

      // ---
      // EXPERT OPTION
      // For tuning purposes the number of events between commits can be specified
      // ---
      unsigned int numEventsPerCommit = 10000;
      if( cmdLineArgs.hasQual( "-nevtpercommit" ) ) {
	 std::string numEventsPerCommitStr = cmdLineArgs.getArg("-nevtpercommit");
	 if( numEventsPerCommitStr.size() && numEventsPerCommitStr[0] != '-' )  {
	    numEventsPerCommit = atoi( numEventsPerCommitStr.c_str() );
	 }
      }

      int numEventsPerPrint = -1;
      if ( cmdLineArgs.hasQual( "-nevtperprint" ) ) {
	 std::string numEventsPerPrintStr = cmdLineArgs.getArg("-nevtperprint");
	 if( numEventsPerPrintStr == "all" ) {
	    numEventsPerPrint = -1;
	 }
	 else if( numEventsPerPrintStr.size() && numEventsPerPrintStr[0] != '-' )  {
	    numEventsPerPrint = atoi( numEventsPerPrintStr.c_str() );
	 }
      }

      int numRowsCached = 0;
      if( cmdLineArgs.hasQual( "-nevtcached" ) ) {
	 std::string numRowsCachedStr = cmdLineArgs.getArg("-nevtcached");
	 if( numRowsCachedStr.size() && numRowsCachedStr[0] != '-' )  {
	    numRowsCached = atoi( numRowsCachedStr.c_str() );
	 }
      }
      
      bool ignoreSchemaDiff = false;
      if( cmdLineArgs.hasQual( "-ignoreschemadiff" ) ) {
	 ignoreSchemaDiff = true;
      }
     

      // MN: this line prevents hang on exit when only RootCollections are used
      // no idea why
      pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );

      // MN: without loading MessageService here, the ouput level set below for some reason does not affect CORAL
      pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
      
      if(!::getenv( "POOL_OUTMSG_LEVEL" ))      {
	 pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Info);
      }
      else {
	 pool::PoolMessageStream pms("get threshold");
	 pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
      }
      pool::PoolMessageStream log( "CollAppend", seal::Msg::Debug );
      log << "creating CollectionService" << seal::endmsg;
      
      pool::ICollectionService *collectionService( new pool::CollectionService( pool::POOLContext::context() ) );

      //collectionService->setDefaultCollectionCatalog();  // this should create the catalog
      // FIX - does the next line connect to the catalog by itself?
      for( unsigned int i=0; i<collCatalogReadVec.size(); i++ )      {
	 collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
      }

/// INTRO COMPLETE

/// BEGIN PROCESSING

      std::map<unsigned int, pool::ICollection*> dstCollMap;
      // ---
      // Now loop over destination collections
      //   - possibly moot since this is normally one destination
      // ---
      for( unsigned int i=0; i<dstCollNameVec.size(); i++ ) {
	 log << "Testing collection '" << dstCollNameVec[i]
	     << "' of type " << dstCollTypeVec[i] <<  seal::endmsg;
	 dstCollMap[i] = 0;
	 // testing for collection existence
	 try {
	    bool readOnly( true );
	    std::auto_ptr<pool::ICollection> dstCollection(collectionService->handle( dstCollNameVec[i], dstCollTypeVec[i], dstCollConnectVec[i], readOnly ) );
	    
	    std::auto_ptr<pool::ICollectionQuery> query( dstCollection->newQuery() );
	    pool::ICollectionCursor& cursor = query->execute();
	    if( !cursor.next() ) {
	       log << seal::Msg::Info
		   << "Destination collection `" + dstCollNameVec[i] + "' is empty."
		   <<  seal::endmsg;
	    }
	 }
	 catch ( pool::Exception& ) {
	    log << seal::Msg::Warning << "Destination Collection `" << dstCollNameVec[i]
		<< "' does not exist; A new one will be created; "
		<< seal::endmsg;
	    dstCollExistVec[i] = false;
	 }
      }

      unsigned int evtCounterTotal = 0;
      srcCountVec.clear();
      std::map<int, pool::ICollection*> srcCollMap;
      pool::CollectionDescription	destDescription( srcCollNameVec[0], "type" );
      
      // ---
      // Now create ICollection for source collection
      // *AND* check whether the schemas are equivalent
      // IF the schemas differ issue a warning but DO NOT stop processing
      // ---
      bool SrcSchemaSame;
      bool DstEqualsSrc;
      //bool DstIncludesSrc;
      bool schemaOK = true;
      size_t k=0;
      for ( unsigned int i=0; i<srcCollNameVec.size(); i++ )
      {
	 log << seal::Msg::Debug
	     << "Opening source collection " << srcCollNameVec[i]
	     << " of type " << srcCollTypeVec[i] <<  seal::endmsg;
	 bool readOnly( true );
	 pool::ICollection *srcCollection = collectionService->handle( srcCollNameVec[i], srcCollTypeVec[i], srcCollConnectVec[i], readOnly );
	 srcCollMap[i] = srcCollection;
         srcCountVec.push_back(0);
	 log << "Opened source collection " << srcCollNameVec[i] <<  seal::endmsg;

         // Get schema, define first as element 0
	 const pool::ICollectionDescription &firstDesc = srcCollMap[0]->description();
         // Get schema for current collection defined by index i
	 const pool::ICollectionDescription &nextDesc = srcCollection->description();

         SrcSchemaSame = nextDesc.equals(firstDesc);
 
	 log << "Creating query for the source collection" <<  seal::endmsg;
	 auto_ptr<ICollectionQuery>  srcQuery( srcCollection->newQuery() );
	 srcQuery->setCondition( query );
	 if( queryOptions.size() ) {
	    srcQuery->addToOutputList( queryOptions );
	 } else {
	    srcQuery->selectAll();
	 }
	 log << "Executing query for the source collection" <<  seal::endmsg;
	 pool::ICollectionCursor& cursor = srcQuery->execute();

	 if( i == 0 ) { 
	    if( !queryOptions.size() ) {
	       // cout << " ---------  Copy source Coll Desc to Dst" << endl;
	       destDescription = firstDesc;
	    } else {
	       const TokenList &tokens =  cursor.currentRow().tokenList();
	       for( TokenList::const_iterator token = tokens.begin();
		    token != tokens.end();  ++token ) {
		  const string& name = token.tokenName();	
		  const ICollectionColumn& column = firstDesc.column( name );
		  if( name != firstDesc.eventReferenceColumnName() )
		     destDescription.insertTokenColumn( name, column.annotation(),
							column.collectionFragmentName() );
	       }	       
	       if ( !noAttrib ) {
		  const coral::AttributeList &attribs = cursor.currentRow().attributeList();
		  for( coral::AttributeList::const_iterator attrib = attribs.begin();
		       attrib != attribs.end();  ++attrib ) {
		     const string& name = attrib->specification().name();
		     const ICollectionColumn& column = firstDesc.column( name );
		     destDescription.insertColumn( name, column.type(), column.annotation(),
						   column.collectionFragmentName(),
						   column.maxSize(), column.sizeIsFixed() );
		  }
	       }
	    }
	 }
	 
	 // open/create destination collections
	 for ( unsigned int i=0; i<dstCollNameVec.size(); i++ )	 {
	    if( !dstCollExistVec[i] ) {
	       // create a new dst collection
	       pool::CollectionDescription newDestDesc( destDescription );
	       newDestDesc.setName( dstCollNameVec[i] );
	       newDestDesc.setType( dstCollTypeVec[i] ); 
	       newDestDesc.setConnection( dstCollConnectVec[i] );
	       log << seal::Msg::Info <<"Creating destination collection "
		   << dstCollNameVec[i]
		   << " of type " << dstCollTypeVec[i] <<  seal::endmsg;
	       dstCollMap[i] = collectionService->create( newDestDesc, true );
	       dstCollExistVec[i] = true;
	    } else if( !dstCollMap[i] ) {
	       // open an existing dst collection if not opened yet
	       bool readOnly(false);
	       dstCollMap[i] = collectionService->handle( dstCollNameVec[i], dstCollTypeVec[i], dstCollConnectVec[i], readOnly );
	    }

	    if( numRowsCached )
	       dstCollMap[i]->dataEditor().setRowCacheSize( numRowsCached );

	    const pool::ICollectionDescription& dstDesc = dstCollMap[i]->description();

            DstEqualsSrc  = nextDesc.equals(dstDesc);
	    if( dstCollExistVec[i] && !noAttrib && ( !SrcSchemaSame || !DstEqualsSrc) ) {
	       std::cout << "CollAppend: WARNING Schemas of two or more collections "
			 << "are different and attributes have not been requested to "
			 << "be excluded from append. By definition, Token attribute "
			 << "schema of equivalent collections of type "
			 << "RootCollection and RelationalCollection "
			 << "are expected to differ slightly and will be handled "
			 << "properly here. Other types of cases could lead to "
			 << "unexpected errors." << std::endl;
                if (!SrcSchemaSame) {
                   std::cout << "SRC-SRC SCHEMA DIFFERENCE" << std::endl;
                }
                if (!DstEqualsSrc) {
                   std::cout << "SRC-DST SCHEMA DIFFERENCE" << std::endl;
                }
                schemaOK = false;
	    }
	 }
      
	 if( (query != "" || queryOptions != "") && !nextDesc.equals(firstDesc) ) {
	    std::cout << "CollAppend: WARNING The schemas of one or more input "
		      << "collections are different and a query has been "
		      << "requested. This may lead to unexpected errors."
		      << std::endl;
	 }
         
         // ---
         // Now to the meat of the program
         // - get an iterator to the input
         // - loop over it
         // - fill the output
         // ---
	       
	 log << "Iterating over the source collection" <<  seal::endmsg;
	 unsigned int evtCounter = 0;
         const std::string& srcRefName = srcCollection->description().eventReferenceColumnName();
         
	 while( (schemaOK || ignoreSchemaDiff) && cursor.next() ) {
	    evtCounter++;
	    evtCounterTotal++;

	    for ( unsigned int i=0; i<dstCollNameVec.size(); i++ )
	    {
	       // Get row buffer for adding data to collection.
	       pool::CollectionRowBuffer &rowBuffer = dstCollMap[i]->dataEditor().rowBuffer();
	       if ( !noAttrib ) {
		  rowBuffer.attributeList() = cursor.currentRow().attributeList();
	       }

               const pool::TokenList&  tokens = cursor.currentRow().tokenList();
               const std::string& dstRefName = dstCollMap[i]->description().eventReferenceColumnName();
               for( pool::TokenList::const_iterator ti = tokens.begin(); ti != tokens.end(); ++ti ) {
                  if( ti.tokenName() == srcRefName ) {
                     ti->setData( &rowBuffer.tokenList()[ dstRefName ] );
                  } else {
                     ti->setData( &rowBuffer.tokenList()[ ti.tokenName() ] );
                  }
               }
	       dstCollMap[i]->dataEditor().insertRow( rowBuffer );

	       if( !(evtCounterTotal % numEventsPerCommit) ) {
		  dstCollMap[i]->commit();
	       }

	       if ( (numEventsPerPrint >= 0) && !(evtCounterTotal % numEventsPerPrint) ) {
		  cout << "CollAppend: " << evtCounterTotal
		      << " events appended" <<  endl;
	       }
	    }
	 }
         if (k < srcCountVec.size()) srcCountVec[k]=evtCounter;
         k++;

	 // Copy over the metadata
	 // Existing metadata with the same key will be overwritten

	 
      }

/// FINISHED PROCESSING

/// CLEANUP

      for( size_t i=0; i<srcCollNameVec.size(); i++ ) {
	 srcCollMap[i]->close();
	 delete srcCollMap[i];
      }
      for( size_t i=0; i<dstCollNameVec.size(); i++ ) {
	 dstCollMap[i]->commit();
	 dstCollMap[i]->close();
	 delete dstCollMap[i];
      }

      // List individual collections (with evt counts)
      cout << "CollAppend: Finished appending input collection(s): " << endl;
      for( unsigned int i=0; i<srcCollNameVec.size(); i++ ) {
	 cout << i+1 << ": `" << srcCollNameVec[i] << ":" 
                           << srcCollTypeVec[i] << "' (" 
                           << srcCountVec[i]    << " events) "
                           << endl;
      }
      
      // List individual output collections
      cout << " to output collection(s): " << endl;
      for( unsigned int i=0; i<dstCollNameVec.size(); i++ ) {
	 cout << i+1 << ": `" << dstCollNameVec[i] << ":" 
	                   << dstCollTypeVec[i] << "' "
                           << endl;
      }
      // Note if there was a selection involved
      if (query.size() > 0) cout << "QUALified by " << query << endl;
      // Give total for output
      cout << endl;
      cout << " ===> " << evtCounterTotal << " total events appended to each output" << endl;

      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::exception: " << exception.what() << std::endl;;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception caught in main()." << std::endl;;
      return 1;
   }
}
