/**
 * @file CollListFileGUID.cpp
 * @brief Utility to list the file GUID's used by a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollListFileGUID.cpp,v 1.19 2008/01/31 16:25:23 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"

#include "FileCatalog/IFileCatalog.h"
//#include "FileCatalog/IFCAction.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"

#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ITransaction.h"

#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "CmdLineArgs.h"

#include "SealKernel/Status.h"

#include <iostream>
using namespace std;

const std::string FILE_CATALOG_CONNECTION = 
   "xmlcatalog_file:PoolFileCatalog.xml";
const std::string COLLECTION_CATALOG_CONNECTION =
   "xmlcatalog_file:CollectionCatalog.xml";


int main(int argc, const char *argv[])
{
   try
   {
      seal::Status sc = seal::Status::SUCCESS;
      std::vector<std::string> collNameVec;
      std::vector<std::string> collTypeVec;
      std::vector<std::string> collConnectVec;
      std::vector<std::string> fileCatalogReadVec;
      std::vector<std::string> collCatalogReadVec;

      CmdLineArgs cmdLineArgs(argc, argv);

      if( cmdLineArgs.hasQual("-help") ||
	  !cmdLineArgs.hasQual("-src") )
      {
	 std::cerr << "Usage: CollListFileGUID "
		   << "<-src [+]<<input collection name> <input collection type>> > "
		   << "(List of name-type pairs is space separated; "
		   << "Choices for collection types are RootCollection, "
		   << "MySQLCollection, MySQLltCollection, RelationalCollection, "
		   << "PHYSICAL_NAME, LOGICAL_NAME or GUID) "
		   << "Optional arguments: "
		   << "[-srcconnect [Input database connection string]] "
		   << "(Note: The default is \"\" and this argument MUST be "
		   << "specified for collections of type MySQLCollection, "
		   << "MySQLltCollection and RelationalCollection. Also, for "
		   << "collections of type RelationalCollection the environment " 
		   << "variables POOL_AUTH_USER and POOL_AUTH_PASSWORD "
		   << "must be set) "
		   << "[-query [predicate string (DEFAULT = \"\")]] "
		   << "[-queryopt [Choices are space separated list of the "
		   << "following: \"FetchOne\" (no caching) or \"FetchAll\" "
		   << "(caching) and/or \"SELECT <comma separated attribute list>\" "
		   << "(DEFAULT = \"FetchOne SELECT *\")]] "
		   << "[-fcread [+][read file catalog connection string "
		   << "(DEFAULT = \"\")]] "
		   << "[-ccread [+][read collection catalog connection string "
		   << "(DEFAULT = \"\")]]"
		   << std::endl;
	 return 1;
      }

      std::string connection = cmdLineArgs.getArg( "-srcconnect" );

      std::string collName = "";
      collName = cmdLineArgs.getArg( "-src" );
      if( (collName == "") ||
	  (collName == "-src") ||
	  (collName == "-query") ||
	  (collName == "-queryopt") ||
	  (collName == "-srcconnect") ||
	  (collName == "-fcread") ||
	  (collName == "-ccread") )
      {
	 std::cerr << "CollListFileGUID: ERROR No input collection specified on " 
		   << "command line." << std::endl;
	 return 1;
      }

      std::string collType = "";
      std::string collConnect = "";
      while( (collName != "") &&
	     (collName != "-src") &&
	     (collName != "-query") &&
	     (collName != "-queryopt") &&
	     (collName != "-srcconnect") &&
	     (collName != "-fcread") &&
	     (collName != "-ccread") )
      {
	 collType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
	 if( collType == "MySQLCollection" ||
	     collType == "MySQLltCollection" ||
	     collType == "RelationalCollection" )
	 {
	    if ( connection != "" )
	    {
	       collConnect = connection;
	    }
	    else
	    {
	       std::cerr << "CollListFileGUID: ERROR Must specify a database "
			 << "connection string for collections of type " 
			 << "MySQL or Oracle." << std::endl;
	       return 1;
	    }
	 }
	 else if ( collType == "RootCollection" ||
		   collType == "PHYSICAL_NAME" ||
		   collType == "LOGICAL_NAME" ||
		   collType == "GUID" )
	 {
	    collConnect = "";
	 }
	 else
	 {
	    std::cerr << "CollListFileGUID: ERROR Collection type " << collType 
		      << " is not presently handled by this program." 
		      << std::endl;
	    return 1;
	 }
	 collNameVec.push_back( collName );
	 collTypeVec.push_back( collType );
	 collConnectVec.push_back( collConnect );
	 collName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      std::string query = cmdLineArgs.getArg( "-query" );

      std::string queryOptions = "";
      if ( cmdLineArgs.hasQual( "-queryopt" ) )
      {
	 queryOptions = cmdLineArgs.getArg("-queryopt");
	 if( (queryOptions == "-src") ||
	     (queryOptions == "-query") ||
	     (queryOptions == "-queryopt") ||
	     (queryOptions == "-srcconnect") ||
	     (queryOptions == "-fcread") ||
	     (queryOptions == "-ccread") )
	 {
	    std::string queryOptions = "";
	 }
      }

      std::string fileCatalogRead = cmdLineArgs.getArg( "-fcread" );
      if( (fileCatalogRead == "") ||
	  (fileCatalogRead == "-src") ||
	  (fileCatalogRead == "-query") ||
	  (fileCatalogRead == "-queryopt") ||
	  (fileCatalogRead == "-srcconnect") ||
	  (fileCatalogRead == "-fcread") ||
	  (fileCatalogRead == "-ccread") )
      {
	 fileCatalogRead = FILE_CATALOG_CONNECTION;
      }
      while( (fileCatalogRead != "") &&
	     (fileCatalogRead != "-src") &&
	     (fileCatalogRead != "-query") &&
	     (fileCatalogRead != "-queryopt") &&
	     (fileCatalogRead != "-srcconnect") &&
	     (fileCatalogRead != "-fcread") &&
	     (fileCatalogRead != "-ccread") )
      {
	 fileCatalogReadVec.push_back( fileCatalogRead );
	 fileCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
      if( (collCatalogRead == "") ||
	  (collCatalogRead == "-src") ||
	  (collCatalogRead == "-query") ||
	  (collCatalogRead == "-queryopt") ||
	  (collCatalogRead == "-srcconnect") ||
	  (collCatalogRead == "-fcread") ||
	  (collCatalogRead == "-ccread") )
      {
	 collCatalogRead = COLLECTION_CATALOG_CONNECTION;
      }
      while( (collCatalogRead != "") &&
	     (collCatalogRead != "-src") &&
	     (collCatalogRead != "-query") &&
	     (collCatalogRead != "-queryopt") &&
	     (collCatalogRead != "-srcconnect") &&
	     (collCatalogRead != "-fcread") &&
	     (collCatalogRead != "-ccread") )
      {
	 collCatalogReadVec.push_back( collCatalogRead );
	 collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      // MN: this line prevents hang on exit when only RootCollections are used
      // no idea why
      pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );

      pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
      if(!::getenv( "POOL_OUTMSG_LEVEL" ))
      {
	 pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Warning);
      }
      else
      {
	 pool::PoolMessageStream pms("get threshold");
	 pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
      }

      pool::PoolMessageStream log( "CollListFileGUID", seal::Msg::Nil );
      log << seal::Msg::Info << "creating CollectionService" << seal::endmsg;
      pool::ICollectionService *collectionService( new pool::CollectionService( pool::POOLContext::context() ) );

      for( unsigned int i=0; i<collCatalogReadVec.size(); i++ )      {
	 collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
      }
      // MN: FIX the catalog connection
      // collectionService->getCollectionCatalog()->connect();
 

      for( unsigned int i=0; i<fileCatalogReadVec.size(); i++ )      {
	 collectionService->addReadFileCatalog( fileCatalogReadVec[i] );
      }
      //collectionService->getFileCatalog()->connect();
      //collectionService->getFileCatalog()->start();


      pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( collectionService->getFileCatalog() );
      pool::DatabaseConnectionPolicy policy;
      policy.setReadMode( pool::DatabaseConnectionPolicy::READ );
      dataSvc->session().setDefaultConnectionPolicy( policy );
      sc = dataSvc->transaction().start( pool::ITransaction::READ );
      if( !sc.isSuccess() )
      {
	 std::cerr << "CollListFileGUID: ERROR Cannot start database "
		   << "transaction." << std::endl;
	 return 1;
      }

      std::map<int, pool::ICollection*> collMap;
      std::map< std::string, int > fileIdMap;
      int fileCounter = 0;
      for( unsigned int i=0; i<collNameVec.size(); i++ ) {

	 bool readOnly( true );
	 pool::ICollection* collection = collectionService->handle( collNameVec[i], collTypeVec[i], collConnectVec[i], readOnly );
	 //&( dataSvc->session() ),   MN: FIX - how to implement that?

	 collMap[i] = collection;
	 
	 cout << ">> Creating query for the collection" << endl;
	 pool::ICollectionQuery *collQuery = collection->newQuery();
	 collQuery->setCondition( query );
	 if( queryOptions.size() ) {
	    collQuery->addToOutputList( queryOptions );
	 } else {
	    collQuery->selectAllTokens();
	 }
	 cout << ">> Executing the query " << endl;
	 pool::ICollectionCursor& cursor = collQuery->execute();

	 while( cursor.next() ) {
	    const pool::TokenList &tokens = cursor.currentRow().tokenList();
	    for( pool::TokenList::const_iterator iter = tokens.begin();
		 iter != tokens.end(); ++iter ) {
	       std::string fileId = iter->dbID();
	       std::map< std::string, int >::iterator mapIter;
	       mapIter = fileIdMap.find( fileId );
	       if( mapIter == fileIdMap.end() )	       {
		  fileIdMap[ fileId ] = fileCounter++;
		  std::cout << fileId << std::endl;
	       }
	    }
	 }
      }

      for( unsigned int i=0; i<collNameVec.size(); i++ )
      {      
	 collMap[i]->close();
      }

      sc = dataSvc->transaction().commit();
      if( !sc.isSuccess() )
      {
	 std::cerr << "CollListFileGUID: ERROR Cannot commit database "
		   << "transaction." << std::endl;
	 return 1;
      }
      delete dataSvc;
      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::Exception: " << exception.what() << std::endl;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception. " << std::endl;
      return 1;
   }
}
