/**
 * @file CollListMetadata.cpp
 * @brief Utility to print metadata from a POOL collection
 * @author Marcin.Nowak@cern.ch
 */
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionMetadata.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"

#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Exception.h"

#include "CmdLineArgs.h"

#include "SealKernel/Status.h"

#include <iostream>
using namespace std;

int main(int argc, const char *argv[])
{
  try
  {
    seal::Status sc = seal::Status::SUCCESS;
    std::vector<std::string> collNameVec;
    std::vector<std::string> collTypeVec;
    std::vector<std::string> collConnectVec;
    std::vector<std::string> collCatalogReadVec;

    CmdLineArgs cmdLineArgs(argc, argv);

    if( cmdLineArgs.hasQual("-help") ||
        !cmdLineArgs.hasQual("-src") )
    {
      std::cerr
	 << "Usage: " << endl
	 << "------ " << endl
	 << "CollListMetadata <-src [+]<<input collection name> <input collection type>> > "
	 << " (List of name-type pairs is space separated; " << endl
	 << " Choices for collection types are: RelationalCollection, "
	 << "RootCollection, PHYSICAL_NAME, LOGICAL_NAME or GUID) " << endl
	 << "Optional arguments: " << endl
	 << "[-srcconnect [database connection string]] "
	 << " (MUST be specified for Relational Collections) " << endl
	 << " The connection string will be looked up in $CORAL_DBLOOKUP_PATH/dblookup.xml" << endl
	 << " Database user authentication is done using CORAL_AUTH_PATH/authentication.xml" << endl
	 << " or POOL_AUTH_USER and POOL_AUTH_PASSWORD variables" << endl
	 << "[-ccread [+][collection catalog connection string "
	 << "(DEFAULT = \"\")]]"
	 << std::endl;
      return 1;
    }

    std::string connection = cmdLineArgs.getArg( "-srcconnect" );

    std::string collName = "";
    collName = cmdLineArgs.getArg( "-src" );
    if( !collName.size() || collName[0] == '-' )  {
       std::cerr << "CollListMetadata: ERROR No input collection specified on " 
		 << "the command line." << std::endl;
       return 1;
    }

    while( collName.size() && collName[0] != '-' )  {
       std::string collConnect = "";
       string collType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
       if( collType.empty() || collType[0] == '-' ) {
	  // collType = (connection.empty()? "RootCollection" : "RelationalCollection" );
	  collType = "RelationalCollection";
       }
       if( collType == "RelationalCollection" ) {
	  if( connection.empty() ) {
	     std::cerr << "CollListMetadata: ERROR Must specify a database "
		       << "connection string for collections of type " 
		       << "MySQL or Oracle." << std::endl;
	     return 1;
	  }
	  collConnect = connection;
       }
       else if( collType != "RootCollection" &&
	      collType != "PHYSICAL_NAME" &&
	      collType != "LOGICAL_NAME" &&
	      collType != "GUID" ) 
       {
	  std::cerr << "CollListMetadata: ERROR Collection type " << collType 
		    << " is not presently handled by this program." 
		    << std::endl;
	  return 1;
       }
       collNameVec.push_back( collName );
       collTypeVec.push_back( collType );
       collConnectVec.push_back( collConnect );
       collName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
    while( collCatalogRead.size() && collCatalogRead[0] != '-' )  {
       collCatalogReadVec.push_back( collCatalogRead );
       collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    // MN: this line prevents hang on exit when only RootCollections are used
    // no idea why
    pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );
    
    pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
    if(!::getenv( "POOL_OUTMSG_LEVEL" )) {
      pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Warning);
    }
    else {
      pool::PoolMessageStream pms("get threshold");
      pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
    }

    pool::PoolMessageStream log( "CollListMetadata", seal::Msg::Nil );
    log << seal::Msg::Info << "creating CollectionService" << seal::endmsg;

//    pool::POOLContext::loadComponent( "POOL/Services/CollectionService" );
//    vector< seal::IHandle< pool::ICollectionService > > loadedCollectionServices;
//    pool::POOLContext::context()->query( loadedCollectionServices );
//    seal::IHandle< pool::ICollectionService > collectionService = loadedCollectionServices.front();

    pool::ICollectionService *collectionService( new pool::CollectionService( pool::POOLContext::context() ) );

    // FIX - does the next line connect to the catalog by itself?
    for( unsigned int i=0; i<collCatalogReadVec.size(); i++ )      {
       collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
    }
    // collCatalog->connect();
    

    std::map<int, pool::ICollection*> collMap;
    std::map< std::string, int > pfnMap; 
    for( unsigned int i=0; i<collNameVec.size(); i++ )
    {
       cout<< "Printing metadata from collection '" << collNameVec[i]
	   << "', type: " << collTypeVec[i] 
	   << endl;

       bool readOnly( true );
       pool::ICollection* collection = collectionService->handle( collNameVec[i], collTypeVec[i], collConnectVec[i], readOnly );
       pool::ICollectionMetadata::const_iterator  mdIter = collection->metadata().begin();
       while( mdIter != collection->metadata().end() ) {
	  cout << "  Metadata Key= " << mdIter.key();
	  cout << ",  Value= " << mdIter.value() << endl;
	  ++mdIter;
       }
    }

    cout << "CollListMetadata: Finished" << endl; 
    return 0;

  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}





