/**
 * @file CollListToken.cpp
 * @brief Utility to list the event references of a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollListToken.cpp,v 1.16 2008/01/31 16:25:23 mnowak Exp $ 
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"

#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "CmdLineArgs.h"

#include "SealKernel/Status.h"

#include <iostream>
using namespace std;


int main(int argc, const char *argv[])
{
   try
   { 
      seal::Status sc = seal::Status::SUCCESS;
      std::vector<std::string> collNameVec;
      std::vector<std::string> collTypeVec;
      std::vector<std::string> collConnectVec;
      std::vector<std::string> collCatalogReadVec;

      CmdLineArgs cmdLineArgs(argc, argv);

      if( cmdLineArgs.hasQual("-help") ||
	  !cmdLineArgs.hasQual("-src") )
      {
	 std::cerr << "Usage: CollListToken "
		   << "<-src [+]<<input collection name> <input collection type>> > "
		   << "(List of name-type pairs is space separated; "
		   << "Choices for collection types are MySQLCollection, "
		   << "RootCollection, MySQLltCollection, RelationalCollection, "
		   << "PHYSICAL_NAME, LOGICAL_NAME or GUID) "
		   << "Optional arguments: "
		   << "[-srcconnect [Input database connection string]] "
		   << "(Note: The default is \"\" and this argument MUST be "
		   << "specified for collections of type MySQLCollection, "
		   << "MySQLltCollection and RelationalCollection. Also, for "
		   << "collections of type RelationalCollection the environment " 
		   << "variables POOL_AUTH_USER and POOL_AUTH_PASSWORD "
		   << "must be set) "
		   << "[-nevtdmp [Number of events starting from first event for "
		   << "which to dump the Token values. \"all\" => "
		   << "dump for all events (DEFAULT = \"\")]] "
		   << "[-query [predicate string (DEFAULT = \"\")]] "
		   << "[-queryopt [Choices are space separated list of the "
		   << "following: \"FetchOne\" (no caching) or \"FetchAll\" "
		   << "(caching) and/or \"SELECT <comma separated attribute list>\" "
		   << "(DEFAULT = \"FetchOne SELECT *\")]] "
		   << "[-ccread [+][read collection catalog connection string "
		   << "(DEFAULT = \"\")]]"
		   << std::endl;
	 return 1;
      }

      std::string connection = cmdLineArgs.getArg( "-srcconnect" );

      std::string collName = "";
      collName = cmdLineArgs.getArg( "-src" );
      if( (collName == "") ||
	  (collName == "-src") ||
	  (collName == "-query") ||
	  (collName == "-queryopt") ||
	  (collName == "-srcconnect") ||
	  (collName == "-ccread") ||
	  (collName == "-nevtdmp") )
      {
	 std::cerr << "CollListToken: ERROR No input collection specified on " 
		   << "command line." << std::endl;
	 return 1;
      }

      std::string collType = "";
      std::string collConnect = "";
      while( (collName != "") &&
	     (collName != "-src") &&
	     (collName != "-query") &&
	     (collName != "-queryopt") &&
	     (collName != "-srcconnect") &&
	     (collName != "-ccread") &&
	     (collName != "-nevtdmp") )
      {
	 collType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
	 if( collType == "MySQLCollection" || 
	     collType == "MySQLltCollection" ||
	     collType == "RelationalCollection" )
	 {
	    if ( connection != "" )
	    {
	       collConnect = connection;
	    }
	    else
	    {
	       std::cerr << "CollListToken: ERROR Must specify a database "
			 << "connection string for collections of type " 
			 << "MySQL or Oracle." << std::endl;
	       return 1;
	    }
	 }
	 else if ( collType == "RootCollection" || 
		   collType == "PHYSICAL_NAME" ||
		   collType == "LOGICAL_NAME" ||
		   collType == "GUID" )
	 {
	    collConnect = "";
	 }
	 else
	 {
	    std::cerr << "CollListToken: ERROR Collection type " << collType 
		      << " is not presently handled by this program." 
		      << std::endl;
	    return 1;
	 }
	 collNameVec.push_back( collName );
	 collTypeVec.push_back( collType );
	 collConnectVec.push_back( collConnect );
	 collName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      std::string query = cmdLineArgs.getArg( "-query" );

      std::string queryOptions = "";
      if ( cmdLineArgs.hasQual( "-queryopt" ) )
      {
	 queryOptions = cmdLineArgs.getArg("-queryopt");
	 if ( (queryOptions == "-src") &&
	      (queryOptions == "-query") &&
	      (queryOptions == "-queryopt") &&
	      (queryOptions == "-srcconnect") &&
	      (queryOptions == "-ccread") &&
	      (queryOptions == "-nevtscan") &&
	      (queryOptions == "-nevtread") &&
	      (queryOptions == "-nevtstat") &&
	      (queryOptions == "-nevtdmp") &&
	      (queryOptions == "-nevtperprint") &&
	      (queryOptions == "-quiet") &&
	      (queryOptions == "-timing") )
	 {
	    std::string queryOptions = "";
	 }
      }

      std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
      while( collCatalogRead.size() && collCatalogRead[0] != '-' )  {
	 collCatalogReadVec.push_back( collCatalogRead );
	 collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      int numEventsDump = -1;
      if ( cmdLineArgs.hasQual( "-nevtdmp" ) )
      {
	 std::string numEventsDumpStr = cmdLineArgs.getArg("-nevtdmp");
	 if ( numEventsDumpStr == "all" )
	 {
	    numEventsDump = -1;
	 }
	 else if ( (numEventsDumpStr != "") &&
		   (numEventsDumpStr != "-src") &&
		   (numEventsDumpStr != "-query") &&
		   (numEventsDumpStr != "-queryopt") &&
		   (numEventsDumpStr != "-srcconnect") &&
		   (numEventsDumpStr != "-ccread") &&
		   (numEventsDumpStr != "-nevtdmp") )
	 {
	    numEventsDump = atoi( numEventsDumpStr.c_str() );
	 }
      }

      // MN: this line prevents hang on exit when only RootCollections are used
      // no idea why
      pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );

      pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
      if(!::getenv( "POOL_OUTMSG_LEVEL" ))
      {
	 pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Warning);
      }
      else
      {
	 pool::PoolMessageStream pms("get threshold");
	 pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
      }


      pool::PoolMessageStream log( "CollListToken", seal::Msg::Nil );
      log << seal::Msg::Info << "creating CollectionService" << seal::endmsg;
      pool::ICollectionService *collectionService( new pool::CollectionService( pool::POOLContext::context() ) );

      // FIX - does the next line connect to the catalog by itself?
      for( unsigned int i=0; i<collCatalogReadVec.size(); i++ )      {
	 collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
      }
      // collCatalog->connect();


      std::map<int, pool::ICollection*> collMap;
      unsigned int collCounter = 0;
      for ( unsigned int i=0; i<collNameVec.size(); i++ )
      {
	 collCounter++;
	 
	 bool readOnly( true );
	 pool::ICollection* collection = collectionService->handle( collNameVec[i], collTypeVec[i], collConnectVec[i], readOnly );
	 collMap[i] = collection;

	 const pool::ICollectionDescription &description = collection->description();

	 std::cout << std::endl;
	 std::cout << "Collection name: " << description.name() << std::endl;
	 std::cout << "Collection type: " << description.type() << std::endl;
	 std::cout << "Number of collection fragments: " << description.numberOfCollectionFragments() << std::endl;
	 std::cout << "Number of tokens: " << description.numberOfTokenColumns() << std::endl;
	 std::cout << "Tokens are: " << std::endl;

	 unsigned  maxNameSize = 0;
	 for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	    for( int c=0; c < description.numberOfTokenColumns( f ); c++ ) {
	       const pool::ICollectionColumn&	column = description.tokenColumn(c, f);
	       if( maxNameSize < column.name().size() ) 
		  maxNameSize = column.name().size();
	    }
	 }
	 for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	    for( int c=0; c < description.numberOfTokenColumns( f ); c++ ) {
	       const pool::ICollectionColumn&	column = description.tokenColumn(c, f);
	       std::cout.setf(std::ios::left);
	       std::cout << "NAME: ";
	       std::cout.width(maxNameSize+5);
	       std::cout << column.name();
	       std::cout << "INFO: ";
	       std::cout << column.annotation();
	       std::cout << std::endl;
	    }
	 }

	 log << seal::Msg::Debug
	     << ">> Creating query for the collection" << seal::endmsg;
	 pool::ICollectionQuery *collQuery = collection->newQuery();
	 collQuery->setCondition( query );
	 if( queryOptions.size() ) {
	    collQuery->addToOutputList( queryOptions );
	 } else {
	    collQuery->selectAllTokens();
	 }
	 log << seal::Msg::Debug
	     << ">> Executing the query " << seal::endmsg;
	 pool::ICollectionCursor& cursor = collQuery->execute();	       

	 int evtCounter = 0;
	 while( cursor.next() ) {
	    evtCounter++;

	    if ( (numEventsDump < 0) || (evtCounter <= numEventsDump) )
	    {
	       std::cout << "Number of events counted: " << evtCounter << std::endl;
	       std::cout << "Tokens: " << std::endl;

	       const pool::TokenList &tokens = cursor.currentRow().tokenList();
	       for( pool::TokenList::const_iterator iter = tokens.begin();
		    iter != tokens.end(); ++iter ) {
		  std::cout.setf(std::ios::left);
		  std::cout << "NAME: ";
		  std::cout.width(maxNameSize+5);
		  std::cout << iter.tokenName();
		  std::cout << "VALUE: " << iter->toString();
		  std::cout << std::endl;
	       }
	    }
	    else
	    {
	       break;
	    }
	 }
      }

      for( unsigned int i=0; i<collNameVec.size(); i++ ) {      
	 collMap[i]->close();
      }

      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::exception: " << exception.what() << std::endl;;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception caught in main()." << std::endl;;
      return 1;
   }
}
