//$Id: FCaddMetaData.cpp,v 1.14.2.1 2007/12/20 10:15:59 frankb Exp $
/**FCaddMetaData.cpp -- FileCatalog command line tool to add metadata
  @author Zhen Xie
  @date: 02/03/2005 
  set default logging to Warning if no POOL_OUTMSG_LEVEL is set; 
  separate logging stream to std::cerr, output stream to std::cout.
  @date: 07/04/2005
  adopt to split catalog interface
*/
#include <cstdlib>
#include "FileCatalog/CommandLine.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/FCException.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/FCEntry.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/Exception.h"
#include "POOLCore/PoolMessageStream.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"
#include <memory>
using namespace pool;
void printUsage(){
  std::cout<<"usage: FCaddMetadata [-m metadata -p pfname -l lfname -u contactstring -h]" <<std::endl;
  std::cout<<"format of metadata: "<<std::endl;
  std::cout<<"(jobid,value),(owner,value),(jobtype,value)"<<std::endl;
}

class Metadataparser{
public:
  Metadataparser(){}
  Metadataparser(const std::string& attr):m_attr(attr){}
  void getAttr(MetaDataEntry& meta) const{
    std::string delim("(,) ");
    std::string::size_type begIdx,endIdx;
    while( (begIdx=m_attr.find_first_not_of(delim))!= m_attr.npos ){
      endIdx=m_attr.find_first_of(')');
      std::string elem=m_attr.substr(begIdx,endIdx-begIdx);
      std::string::size_type colIdx=elem.find_last_of(',');
      if(colIdx != elem.npos){
        std::string colname=elem.substr(0,colIdx);
        //std::cout<<colname<<std::endl;
        std::string colval=elem.substr(colIdx+1);
        //std::cout<<colval<<std::endl;
        std::string coltype=meta.spec()[colname].type_name();
        if( coltype=="string" ){          
          meta.setAttributeValue(colname, colval);
        }else if( coltype=="int" ){
          meta.setAttributeValue(colname,(atoi(colval.c_str())));
        }else if( coltype=="float" ){
          meta.setAttributeValue(colname,(float)atof(colval.c_str()));
        }
      }
      m_attr=m_attr.substr(endIdx+1);
    }
  }
  
  ~Metadataparser(){
  }
private:
    mutable std::string m_attr;
};

int main(int argc, char** argv) {
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  if(!::getenv( "POOL_OUTMSG_LEVEL" )){ //if not set, default to warning
    pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Warning);
  }else{
    pool::PoolMessageStream pms("get threshold");
    pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
  }
  seal::IHandle<seal::IMessageService> mesgsvc =
    pool::POOLContext::context()->query<seal::IMessageService>("SEAL/Services/MessageService");
  if( mesgsvc ){
    //all logging go to cerr
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Nil );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Verbose );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Debug );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Info );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Fatal );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Error );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Warning );
  } 
  std::string myuri;
  std::string mypfn;
  std::string mylfn;
  std::string metadata;
  try{
    CommandLine commands(argc,argv);
    if( commands.Exists("u") ){
      myuri=commands.GetByName("u");
    } 
    if( commands.Exists("m") ){
      metadata=commands.GetByName("m");
    }     
    if( commands.Exists("p") ){
      mypfn=commands.GetByName("p");
    }
    if( commands.Exists("l") ){
      mylfn=commands.GetByName("l");
    }
    if( commands.Exists("h") ){
      printUsage();
      exit(0);
    }
  }catch(std::string& strError){
    std::cerr << "Error: command parsing error "<<strError<<std::endl;
    exit(0);
  }
  if( mypfn.empty()&&mylfn.empty() ){
    printUsage();
    std::cerr << "Error: must specify pfname using -p or lfname using -l" << std::endl;
    exit(0);
  }
  try{
    std::auto_ptr<IFileCatalog> mycatalog(new IFileCatalog);
    mycatalog->setWriteCatalog(myuri);
    FCregister r;
    mycatalog->setAction(r);
    mycatalog->connect();
    mycatalog->start();
    MetaDataEntry mymeta; 
    r.getMetaDataSpec(mymeta);
    Metadataparser metaparser(metadata);
    metaparser.getAttr(mymeta);
    FileCatalog::FileID fid;
    if( !mylfn.empty() ){
      r.lookupFileByLFN(mylfn,fid);
    }else if( !mypfn.empty() ){
      std::string ftype;
      r.lookupFileByPFN(mypfn,fid,ftype);
    }
    r.registerMetaData(fid,mymeta);
    mycatalog->commit();  
    mycatalog->disconnect();
  }catch (const pool::Exception& er){
    //er.printOut(std::cerr);
    //std::cerr << std::endl;
    std::cerr<<er.what()<<std::endl;
    exit(1);
  }catch (const std::exception& er){
    std::cerr<<er.what()<<std::endl;
    exit(1);
  }
}

  
