//$Id: FClistMetaData.cpp,v 1.16.2.1 2007/12/20 10:15:59 frankb Exp $
/**FClistMetaData.cpp -- FileCatalog command line tool to list the selected MetaData entries
  @author: Zhen Xie
  @date: 02/03/2005 Z.X.
  set default logging to Warning if no POOL_OUTMSG_LEVEL is set; 
  separate logging stream to std::cerr, output stream to std::cout
  @date: 07/04/2005 Z.X.
  adopt to split catalog
*/
#include "FileCatalog/CommandLine.h"
#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/FCException.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/IFCContainer.h"
#include "POOLCore/POOLContext.h"
#include "POOLCore/Exception.h"
#include "POOLCore/PoolMessageStream.h"
#include "SealKernel/Context.h"
#include "SealKernel/MessageStream.h"
#include <memory>
#include <vector>
#include <string>
#include <cstdlib>
using namespace pool;
void printUsage(){
  std::cout<<"usage: FClistMetaData [-l lfname -p pfname -q query -u contactstring -m maxcache(default 1000) -h]" <<std::endl; 
}
class contactParser{
public:
  explicit contactParser(const std::string& contact):m_contact(contact){}
  void getContacts(std::vector<std::string>& uris){
    std::string delim(" ");
    std::string::size_type begIdx,endIdx;
    std::string elem;
    while( (begIdx=m_contact.find_first_not_of(delim))!= m_contact.npos ){
      endIdx=m_contact.find_first_of(delim);
      if(endIdx != m_contact.npos ){
        elem=m_contact.substr(begIdx,endIdx-begIdx);
        m_contact=m_contact.substr(endIdx+1);
        uris.push_back(elem);
      }else{
        elem=m_contact;
        uris.push_back(elem);
        break;
      }
    }
  }
  ~contactParser(){}
private:
  std::string m_contact;
};

int main(int argc, char** argv) {
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  if(!::getenv( "POOL_OUTMSG_LEVEL" )){ //if not set, default to warning
    pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Warning);
  }else{
    pool::PoolMessageStream pms("get threshold");
    pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
  }
  seal::IHandle<seal::IMessageService> mesgsvc =
    pool::POOLContext::context()->query<seal::IMessageService>("SEAL/Services/MessageService");
  if( mesgsvc ){
    //all logging go to cerr
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Nil );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Verbose );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Debug );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Info );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Fatal );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Error );
    mesgsvc->setOutputStream( std::cerr, seal::Msg::Warning );
  } 
  std::string myuri;
  std::string mylfn;
  std::string mypfn;
  unsigned int maxcache=1000;
  std::string query; 
  try{
    CommandLine commands(argc,argv);
    if( commands.Exists("u") ){
      myuri=commands.GetByName("u");
    }else{
      myuri=std::string(::getenv("POOL_CATALOG"));
    }
    if( commands.Exists("l") ){
      mylfn=commands.GetByName("l");
    }
    if( commands.Exists("p") ){
      mypfn=commands.GetByName("p");
    }
    if( commands.Exists("q") ){
      query=commands.GetByName("q");
    }
    if( commands.Exists("h") ){
      printUsage();
      exit(0);
    }
    if( commands.Exists("m") ){
      maxcache=atoi(commands.GetByName("m").c_str());
    }
  }catch(std::string& strError){
    std::cerr << "Error: command parsing error "<<strError<<std::endl;
    exit(0);
  }
  if(!query.empty()&&(!mypfn.empty()||!mylfn.empty())){
    std::cerr << "list metadata as query result"<<std::endl;
    mypfn=""; mylfn="";
  }
  if(!mypfn.empty()&&!mylfn.empty()){
    std::cerr << "WARNING: list metadata by PFN "<< mypfn<<std::endl;
  }
  try{
    std::auto_ptr<IFileCatalog> mycatalog(new IFileCatalog);
    if(myuri.empty()){
      mycatalog->addReadCatalog(myuri);
    }else{
      contactParser parser(myuri);
      std::vector<std::string> uris;
      parser.getContacts(uris);
      for(std::vector<std::string>::iterator i=uris.begin(); i!=uris.end(); ++i){
        mycatalog->addReadCatalog(*i);
      }
    }
    FClookup l;
    mycatalog->setAction(l);
    mycatalog->connect();
    mycatalog->start();
    MetaDataContainer metas( mycatalog.get(), maxcache );
    if( !mypfn.empty() ){
      FileCatalog::FileID fid;
      std::string ftype;
      l.lookupFileByPFN(mypfn, fid, ftype);
      l.lookupMetaDataByQuery(std::string("guid='"+fid+"'"),metas);
    }else if( !mylfn.empty() ){
      FileCatalog::FileID fid;
      l.lookupFileByLFN(mylfn, fid);
      l.lookupMetaDataByQuery(std::string("guid='"+fid+"'"),metas);
    }else{
      l.lookupMetaDataByQuery(query, metas);
    }
    while(metas.hasNext()){
      MetaDataEntry attrs=metas.Next();
      AttributeListSpecification spec=attrs.spec();
      AttributeListSpecification::iterator i;
      for(i=spec.begin(); i!=spec.end(); ++i){
        std::string colname=(*i).name();
        std::string coltype=(*i).type_name();
        if( colname != "guid" ){
          std::string colvalue=attrs.getAttributeValueAsString(colname);
          if( colvalue.empty()){
            colvalue="NULL";
          }
          if(i != (spec.end()-1)){
            std::cout<<"("<<colname<<","<<colvalue<<"),";
          }else{
            std::cout<<"("<<colname<<","<<colvalue<<")"<<std::endl;
          }
        }
      }
    }
    mycatalog->commit();  
    mycatalog->disconnect();
  }catch (const pool::Exception& er){
    //er.printOut(std::cerr);
    //std::cerr << std::endl;
    std::cerr<<er.what()<<std::endl;
    exit(1);
  }catch (const std::exception& er){
    std::cerr<<er.what()<<std::endl;
    exit(1);
  }
}


