#ifndef LFCCATALOG_QUERY_EVALUATOR_H
#define LFCCATALOG_QUERY_EVALUATOR_H

#include <vector>
#include <string>
#include <memory>

#include "LFCCatalog/LFCCatalog.h"
#include "FileCatalog/FCException.h"
#include "LFCCatalog/LFCParser.h"

#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"

namespace pool {

  /* LFCcounted_ptr class is based on Yonat Sharon's counted_ptr class, http://ootips.org/yonat/4dev/ */
  template <class X> class LFCcounted_ptr {
  public:

    explicit LFCcounted_ptr(X* p = 0) : itsCounter(0) {if (p) itsCounter = new counter(p);}

    ~LFCcounted_ptr() {release();}

    LFCcounted_ptr(const LFCcounted_ptr& r) throw() {acquire(r.itsCounter);}

    LFCcounted_ptr& operator=(const LFCcounted_ptr& r) {
      if (this != &r) {
        release();
        acquire(r.itsCounter);
      }
      return *this;
    }

    X& operator*()  const throw()   {return *itsCounter->ptr;}
    X* operator->() const throw()   {return itsCounter->ptr;}
    X* get()        const throw()   {return itsCounter ? itsCounter->ptr : 0;}
    bool unique()   const throw()   {return (itsCounter ? itsCounter->count == 1 : true);}

  private:

    struct counter {
        counter(X* p = 0, unsigned c = 1) : ptr(p), count(c) { }
        X*          ptr;
        unsigned    count;
    }* itsCounter;

    void acquire(counter* c) throw() {
      // increment the count
      itsCounter = c;
      if (c) ++c->count;
    }

    void release() {
      // decrement the count, delete if it is 0
      if (itsCounter) {
        if (--itsCounter->count == 0) {
          delete itsCounter->ptr;
          delete itsCounter;
        }
        itsCounter = 0;
      }
    }
  };

  /* LFCQueryRequest class */

  template <class X> class LFCQueryRequest {
  public:

    LFCQueryRequest( LFCcounted_ptr<LFCQueryRequest<X> >& a, const std::string &join_type ) {
      PoolMessageStream fclog( "LFCQueryRequest::LFCQueryRequest", seal::Msg::Nil);

      if (join_type == "DISTINCT") {
        m_type = DISTINCT;
      } else {
        fclog<<seal::Msg::Warning<< "Unrecognised type of query object requested" <<seal::endmsg;
        throw LFCQueryException("LFCQueryRequest::LFCQueryRequest",
                                "Unrecognised type of query object requested");
      }
      m_a = a;
      if (m_a->m_type != LIST) {
        fclog<<seal::Msg::Warning<< "Attempt to place a distinct object below a non LIST object" <<seal::endmsg;
        throw FCbackendException("LFCQueryRequest::LFCQueryRequest","Attempt to place a distinct object below a non LIST object");
      }

      size_t bufsize = m_a->maxsize();
      
      // create another, identical, list object
      LFCcounted_ptr<LFCQueryRequest<X> > p(new LFCQueryRequest<X>(m_a->m_cat, m_a->m_qfunc, m_a->m_matchtype, m_a->m_sarg, bufsize));
      m_b = p;

      // create our own buffer
      LFCcounted_ptr<FCBuf<X> > p2(new FCBuf<X>(bufsize));
      m_pointer = p2;

      // copy partent's cat instance
      m_cat = m_a->m_cat;

      LFCCatalog::constraint_list cl;
      rewind(NULL, true, cl, true);
    }

    LFCQueryRequest( LFCcounted_ptr<LFCQueryRequest<X> >& a,  LFCcounted_ptr<LFCQueryRequest<X> >& b, const std::string &join_type, const size_t &bufsize ) {
      PoolMessageStream fclog( "LFCQueryRequest::LFCQueryRequest", seal::Msg::Nil);
      if (join_type == "AND") {
        m_type = AND;
        m_state = NO_STATE;
      } else if (join_type == "OR") {
        fclog<<seal::Msg::Warning<< "The LFCCatalog interface does not yet support the OR operator" <<seal::endmsg;
        throw LFCQueryException("LFCQueryRequest::LFCQueryRequest","The LFCCatalog interface does not yet support the OR operator");
      } else {
        fclog<<seal::Msg::Warning<< "Unrecognised type of query object requested" <<seal::endmsg;
        throw LFCQueryException("LFCQueryRequest::LFCQueryRequest",
                                "Unrecognised type of query object requested");
      }

      m_a = a;
      m_b = b;
      LFCcounted_ptr<FCBuf<X> > p(new FCBuf<X>(bufsize));
      m_pointer = p;

      // copy one of partents' cat instance
      m_cat = m_a->m_cat;
    }

    LFCQueryRequest( LFCCatalog *cat_instance, void (LFCCatalog::*func)(LFC_LIST_FUNC_TYPE(X)), const LFCCatalog::match_types_t &matchtype, const std::string &s, const size_t &bufsize) {
      PoolMessageStream fclog( "LFCQueryRequest::LFCQueryRequest", seal::Msg::Nil);
      m_type = LIST;
      m_qfunc = func;
      m_cat = cat_instance;
      m_sarg = s;
      LFCcounted_ptr<FCBuf<X> > p(new FCBuf<X>(bufsize));
      m_pointer = p;
      m_matchtype = matchtype;

      // gather data from list function using the LIST_CONSTRAINTS parameter
      LFCCatalog::distinct_t dist;
      LFCCatalog::constraint_list cl;
      // get constraint profile: non NULL set list
      (m_cat->*m_qfunc)(*m_pointer,m_privdata, m_sarg, m_matchtype, m_pointer.get(), dist, cl, LFCCatalog::LIST_CONSTRAINTS);
      m_need_set = (m_privdata == LFC_NEED_SET) ? true : false;
      m_list_distinct_with_set = dist;
      m_list_constraints = cl;
      if (!m_need_set) {
        // get constraint profile: with NULL set list
        (m_cat->*m_qfunc)(*m_pointer,m_privdata, m_sarg, m_matchtype, NULL, dist, cl, LFCCatalog::LIST_CONSTRAINTS);
        m_list_distinct_without_set = dist;
      }
    }

    ~LFCQueryRequest() {
      switch(m_type) {
        case LIST:
          if (m_state == CONTINUE) {
            LFCCatalog::distinct_t dist;
            LFCCatalog::constraint_list cl;
            m_state = END;
            (m_cat->*m_qfunc)(*m_pointer,m_privdata, m_sarg, m_matchtype, NULL, dist, cl, LFCCatalog::LIST_END);
          }
          break;

        default:
          break;
      }
    }

    /* allow query for size of our cache */
    size_t maxsize() const {
      return m_pointer->maxsize();
    }

    const X& next() {
      PoolMessageStream fclog( "LFCQueryRequest::next", seal::Msg::Nil);

      if (m_offset-m_bufstart < m_pointer->size()) {
        size_t off = m_offset++;
        off -= m_bufstart;
        return (*m_pointer)[off];
      }
      fclog<<seal::Msg::Warning<< "No next object available" <<seal::endmsg;
      throw LFCQueryException("LFCQueryRequest::next",
                              "No next object available");
    }

    void info(bool &need_set, LFCCatalog::distinct_t &with_set, LFCCatalog::distinct_t &without_set) const {
      PoolMessageStream fclog( "LFCQueryRequest::need_set", seal::Msg::Nil);

      switch(m_type) {
        case LIST:
          need_set = m_need_set;
          with_set = m_list_distinct_with_set;
          without_set = m_list_distinct_without_set;
          break;

        case DISTINCT:
          m_a->info(need_set,with_set,without_set);
          if (m_try_distinct) {
            with_set = LFCCatalog::YES;
            if (!need_set) without_set = LFCCatalog::YES;
          }
          break;

        case AND:
          { bool a_need,b_need;
            LFCCatalog::distinct_t a_with,a_without,b_with,b_without;
            m_a->info(a_need,a_with,a_without);
            m_b->info(b_need,b_with,b_without);
            need_set = (a_need && b_need) ? true : false;

            if (a_need && b_need) {
              if (a_with == LFCCatalog::YES || b_with == LFCCatalog::YES) {
                with_set = LFCCatalog::YES;
                return;
              }
              if (a_with == LFCCatalog::MAYBE && b_with == LFCCatalog::MAYBE) {
                with_set = LFCCatalog::MAYBE;
                return;
              }
              with_set = LFCCatalog::NO;
              return;
            }

            if (a_need && !b_need) {
              LFCCatalog::constraint_list cl_b;
              m_b->constraints(cl_b);

              if (a_with == LFCCatalog::YES || b_with == LFCCatalog::YES) {
                with_set = LFCCatalog::YES;
              } else if (a_with == LFCCatalog::MAYBE && cl_b.n()>0) {
                with_set = LFCCatalog::MAYBE;
              } else if (a_with == LFCCatalog::MAYBE && b_with == LFCCatalog::MAYBE) {
                with_set = LFCCatalog::MAYBE;
              } else {
                with_set = LFCCatalog::NO;
              }
              if (a_with == LFCCatalog::YES) {
                without_set = LFCCatalog::YES;
              } else if (a_with == LFCCatalog::NO) {
                without_set = LFCCatalog::NO;
              } else {
                without_set = b_without;
              }
              return;
            }

            if (!a_need && b_need) {
              LFCCatalog::constraint_list cl_a;
              m_a->constraints(cl_a);

              if (a_with == LFCCatalog::YES || b_with == LFCCatalog::YES) {
                with_set = LFCCatalog::YES;
              } else if (b_with == LFCCatalog::MAYBE && cl_a.n()>0) {
                with_set = LFCCatalog::MAYBE;
              } else if (a_with == LFCCatalog::MAYBE && b_with == LFCCatalog::MAYBE) {
                with_set = LFCCatalog::MAYBE;
              } else {
                with_set = LFCCatalog::NO;
              }
              if (b_with == LFCCatalog::YES) {
                without_set = LFCCatalog::YES;
              } else if (b_with == LFCCatalog::NO) {
                without_set = LFCCatalog::NO;
              } else {
                without_set = a_without;
              }
              return;
            }

            if (!a_need && !b_need) {
              LFCCatalog::constraint_list cl_a,cl_b;
              m_a->constraints(cl_a);
              m_b->constraints(cl_b);
  
              if (a_with == LFCCatalog::YES || b_with == LFCCatalog::YES) {
                without_set = LFCCatalog::YES;
              } else if ((a_without == LFCCatalog::YES && cl_b.n()>0) ||
                         (b_without == LFCCatalog::YES && cl_a.n()>0)) {
                without_set = LFCCatalog::YES;
              } else if (a_without == LFCCatalog::YES && b_with == LFCCatalog::MAYBE) {
                without_set = LFCCatalog::YES;
              } else if (a_with == LFCCatalog::MAYBE && b_without == LFCCatalog::YES) {
                without_set = LFCCatalog::YES;
              } else {
                without_set = LFCCatalog::NO;
              }

              if (a_with == LFCCatalog::YES || b_with == LFCCatalog::YES) {
                with_set = LFCCatalog::YES;
              } else if ((a_with == LFCCatalog::MAYBE && cl_b.n()>0) ||
                         (b_with == LFCCatalog::MAYBE && cl_a.n()>0)) {
                with_set = LFCCatalog::MAYBE;
              } else if (a_with == LFCCatalog::MAYBE && b_with == LFCCatalog::MAYBE) {
                with_set = LFCCatalog::MAYBE;
              } else {
                with_set = LFCCatalog::NO;
              }

              return;
            }
          }
          break;

        default:
          fclog<<seal::Msg::Warning<< "Unknown object type" <<seal::endmsg;
          throw FCbackendException("LFCQueryRequest::info","Unknown object type");
          break;
      }
    };

    bool isOutputDistinct() const {
      PoolMessageStream fclog( "LFCQueryRequest::isOutputDistinct", seal::Msg::Nil);

      bool output;

      bool need_set;
      LFCCatalog::distinct_t distinct_with_set, distinct_without_set;
      info(need_set,distinct_with_set,distinct_without_set);

      if (m_setp == NULL) {
        if (need_set) {
          fclog<<seal::Msg::Warning<< "Need set, but no input set" <<seal::endmsg;
          throw FCbackendException("LFCQueryRequest::isOutputDistinct","Need set, but no input set");
        }
        if (distinct_without_set == LFCCatalog::MAYBE) {
          fclog<<seal::Msg::Warning<< "Distinct without set is not definite" <<seal::endmsg;
          throw FCbackendException("LFCQueryRequest::isOutputDistinct","Distinct without set is not definite");
        }
        output = (distinct_without_set == LFCCatalog::YES) ? true : false;
      } else {
        if (distinct_with_set == LFCCatalog::MAYBE) {
          output = m_input_set_distinct;
        } else {
          output = (distinct_with_set == LFCCatalog::YES) ? true : false;
        }
      }

      return output;
    };

    void constraints(LFCCatalog::constraint_list &cl) {
      PoolMessageStream fclog( "LFCQueryRequest::constraints", seal::Msg::Nil);

      cl.clear();

      switch(m_type) {
        case LIST:
          cl = m_list_constraints;
          break;

        case DISTINCT:
          m_a->constraints(cl);
          break;

        case AND:
          { LFCCatalog::constraint_list a,b;
            m_a->constraints(a);
            m_b->constraints(b);
            if (a.n()>0 && b.n()>0) {
              cl = a;
              cl.add(b);
            }
          }
          break;

        default:
          fclog<<seal::Msg::Warning<< "Unknown object type" <<seal::endmsg;
          throw FCbackendException("LFCQueryRequest::constraints","Unknown object type");
          break;
      }
    }

    void rewind(FCBuf<X> *input_set, const bool &input_is_distinct, const LFCCatalog::constraint_list &c, const bool &try_distinct) {
      PoolMessageStream fclog( "LFCQueryRequest::rewind", seal::Msg::Nil);

      m_setp = input_set;
      m_try_distinct = try_distinct;
      m_input_set_distinct = input_is_distinct;
      m_cl = c;

      switch(m_type) {
        case LIST:
          if (m_state == CONTINUE) {
            LFCCatalog::distinct_t dist;
            LFCCatalog::constraint_list cl;
            m_state = END;
            (m_cat->*m_qfunc)(*m_pointer,m_privdata, m_sarg, m_matchtype, m_setp, dist, cl, LFCCatalog::LIST_END);
          }
          m_offset = 0;
          m_bufstart = 0;
          m_pointer->clear();
          m_state = BEGIN;
          break;

        case DISTINCT:
          { m_offset = 0;
            m_bufstart = 0;
            m_pointer->clear();

            m_a->rewind(m_setp, m_input_set_distinct, m_cl, m_try_distinct);

            bool need_set;
            LFCCatalog::distinct_t distinct_with_set, distinct_without_set;
            m_a->info(need_set,distinct_with_set,distinct_without_set);

            if (m_try_distinct==false ||
                ((m_setp != NULL && (distinct_with_set == LFCCatalog::YES ||
                            (distinct_with_set == LFCCatalog::MAYBE && m_input_set_distinct))) ||
                 (m_setp == NULL && distinct_without_set == LFCCatalog::YES))) {
              m_state = NOP_DISTINCT;
            } else {
              m_state = MAKE_DISTINCT;
            }
          }
          break;

        case AND:
          { m_state = NO_STATE;

            bool a_need_set,b_need_set;
            LFCCatalog::distinct_t a_distinct_with_set, b_distinct_with_set;
            LFCCatalog::distinct_t a_distinct_without_set, b_distinct_without_set;

            m_a->info(a_need_set,a_distinct_with_set,a_distinct_without_set);
            m_b->info(b_need_set,b_distinct_with_set,b_distinct_without_set);

            LFCCatalog::constraint_list cl_a,cl_b;
            m_a->constraints(cl_a);
            m_b->constraints(cl_b);
            if (m_setp == NULL) {
              if (a_need_set && b_need_set) {
                fclog<<seal::Msg::Warning<< "Rewind of AND object without input set, while both parents require a set" <<seal::endmsg;
                throw LFCQueryException("LFCQueryRequest::rewind",
                                        "Rewind of AND object without input set, while both parents require a set");
              } else if (a_need_set && !b_need_set) {
                if (cl_a.n()>0) {
                  m_state = AND_B_WITH_CONSTRAINTS;
                } else {
                  m_state = AND_B_FIRST;
                }
              } else if (!a_need_set && b_need_set) {
                if (cl_b.n()>0) {
                  m_state = AND_A_WITH_CONSTRAINTS;
                } else {
                  m_state = AND_A_FIRST;
                }
              } else { // neither a nor b need an input set
                if (m_try_distinct) {
                  if (a_distinct_with_set == LFCCatalog::MAYBE &&
                      a_distinct_without_set == LFCCatalog::NO &&
                      b_distinct_without_set == LFCCatalog::YES) {
                    m_state = AND_B_FIRST;
                  } else if (b_distinct_with_set == LFCCatalog::MAYBE && 
                             b_distinct_without_set == LFCCatalog::NO &&
                             a_distinct_without_set == LFCCatalog::YES) {
                    m_state = AND_A_FIRST;
                  } else if (b_distinct_with_set == LFCCatalog::YES &&
                             b_distinct_without_set == LFCCatalog::NO &&
                             a_distinct_with_set == LFCCatalog::NO) {
                    m_state = AND_A_FIRST;
                  } else if (a_distinct_with_set == LFCCatalog::YES &&
                             a_distinct_without_set == LFCCatalog::NO &&
                             b_distinct_with_set == LFCCatalog::NO) {
                    m_state = AND_B_FIRST;
                  }
                }
              }
            } else { // i.e. set is not null
              if (a_need_set && b_need_set) {
                if (cl_a.n_literal() >= cl_b.n_literal()) {
                  m_state = AND_A_FIRST; // if equal start with A by convention
                } else {
                  m_state = AND_B_FIRST;
                }
              } else {
                if (m_try_distinct) {
                  if (m_input_set_distinct) {
                    if (a_distinct_with_set == LFCCatalog::YES &&
                        b_distinct_with_set == LFCCatalog::NO ) {
                      if (cl_b.n()>0) {
                        m_state = AND_A_WITH_CONSTRAINTS;
                      } else {
                        m_state = AND_B_FIRST;
                      }
                    } else if (b_distinct_with_set == LFCCatalog::YES &&
                               a_distinct_with_set == LFCCatalog::NO ) {
                      if (cl_a.n()>0) {
                        m_state = AND_B_WITH_CONSTRAINTS;
                      } else {
                        m_state = AND_A_FIRST;
                      }
                    }
                  } else { // i.e. input set is not distinct
                    if (a_distinct_with_set == LFCCatalog::YES &&
                        b_distinct_with_set != LFCCatalog::YES ) {
                      if (cl_b.n()>0) {
                        m_state = AND_A_WITH_CONSTRAINTS;
                      } else {
                        m_state = AND_B_FIRST;
                      }
                    } else if (b_distinct_with_set == LFCCatalog::YES &&
                               a_distinct_with_set != LFCCatalog::YES ) {
                      if (cl_a.n()>0) {
                        m_state = AND_B_WITH_CONSTRAINTS;
                      } else {
                        m_state = AND_A_FIRST;
                      }
                    }
                  }
                }
              }
            }

            if (m_state == NO_STATE) {
              if (cl_a.n()>0 && cl_b.n()==0) {
                m_state = AND_B_WITH_CONSTRAINTS;
              } else if (cl_a.n()==0 && cl_b.n()>0) {
                m_state = AND_A_WITH_CONSTRAINTS;
              } else if (cl_a.n()==0 && cl_b.n() == 0) {
                m_state = AND_A_FIRST; // by convention
              } else if (cl_a.n_literal() >= cl_b.n_literal()) {
                m_state = AND_A_WITH_CONSTRAINTS; // if count equal use A by convention
              } else {
                m_state = AND_B_WITH_CONSTRAINTS;
              }
            }

            if (!m_try_distinct) {
              m_a_try_dist = false;
              m_b_try_dist = false;
            } else {
              switch(m_state) {
                case AND_A_FIRST:
                  m_a_try_dist = true;
                  m_b_try_dist = false;
                  break;

                case AND_B_FIRST:
                  m_b_try_dist = true;
                  m_a_try_dist = false;
                  break;

                case AND_A_WITH_CONSTRAINTS:
                  m_a_try_dist = true;
                  break;

                case AND_B_WITH_CONSTRAINTS:
                  m_b_try_dist = true;
                  break;

                default:
                  fclog<<seal::Msg::Warning<< "Unknown state" <<seal::endmsg;
                  throw LFCQueryException("LFCQueryRequest::rewind",
                                          "Unknown state");
                  break;
              }
            }

            m_offset = 0;
            m_bufstart = 0;
            m_pointer->clear();

            switch(m_state) {
              case AND_A_FIRST:
                m_a->rewind(m_setp, m_input_set_distinct, m_cl, m_a_try_dist);
                break;

              case AND_B_FIRST:
                m_b->rewind(m_setp, m_input_set_distinct, m_cl, m_b_try_dist);
                break;

              case AND_A_WITH_CONSTRAINTS:
                m_cl.add(cl_b);
                m_a->rewind(m_setp, m_input_set_distinct, m_cl, m_a_try_dist);
                break;

              case AND_B_WITH_CONSTRAINTS:
                m_cl.add(cl_a);
                m_b->rewind(m_setp, m_input_set_distinct, m_cl, m_b_try_dist);
                break;

              default:
                fclog<<seal::Msg::Warning<< "Unknown state in AND object" <<seal::endmsg;
                throw LFCQueryException("LFCQueryRequest::rewind",
                                        "Unknown state in AND object");
                break;
            }
          }
          break;

        default:
          fclog<<seal::Msg::Warning<< "Not implimented for this type of query object" <<seal::endmsg;
          throw LFCQueryException("LFCQueryRequest::rewind",
                                  "Not implimented for this type of query object");
          break;
      }
    }

    bool hasNext() {
      PoolMessageStream fclog( "LFCQueryRequest::hasNext", seal::Msg::Nil);

      m_cat->keepalive();

      if (m_offset-m_bufstart < m_pointer->size()) {
        return true;
      }

      m_bufstart += m_pointer->size();
      m_pointer->clear();

      switch(m_type) {
        case LIST:
          { LFCCatalog::distinct_t dist = LFCCatalog::DIST_UNSET;

            if (m_state == END) return false;

            if (m_state == BEGIN) {
              m_state = CONTINUE;

              (m_cat->*m_qfunc)(*m_pointer,m_privdata,m_sarg,m_matchtype,m_setp,dist,m_cl, LFCCatalog::LIST_BEGIN);
             } else {
              (m_cat->*m_qfunc)(*m_pointer,m_privdata,m_sarg,m_matchtype,m_setp,dist,m_cl, LFCCatalog::LIST_CONTINUE);
             }

            if (m_pointer->empty()) {
              m_state = END;
              (m_cat->*m_qfunc)(*m_pointer,m_privdata,m_sarg,m_matchtype,m_setp,dist,m_cl,LFCCatalog::LIST_END);
              return false;
            }

            bool returned_dist,verify_dist;
            if (dist == LFCCatalog::MAYBE) {
              if (m_setp == NULL) {
                fclog<<seal::Msg::Warning<< "LIST object returned maybe distinct for list without set" <<seal::endmsg;
                throw FCbackendException("LFCQueryRequest::hasNext","LIST object returned maybe distinct for list without set");
              }
              returned_dist = m_input_set_distinct;
            } else {
              returned_dist = (dist == LFCCatalog::YES) ? true : false;
            }
            verify_dist = isOutputDistinct();
            if (returned_dist != verify_dist) {
              fclog<<seal::Msg::Warning<< "LIST returned distinct status different from the calculated one" <<seal::endmsg;
              throw FCbackendException("LFCQueryRequest::hasNext","LIST returned distinct status different from the calculated one");
            }

            return true;
          }
          break;

        case DISTINCT:
          switch(m_state) {
            case NOP_DISTINCT:
              while(!m_pointer->full() && m_a->hasNext()) {
                const X &item = m_a->next();
                m_pointer->push_back(item);
              }
              break;

            case MAKE_DISTINCT:
              { FCBuf<X> *a_buf;

                while( !m_pointer->full() && ((a_buf = m_a->getCurrBuf()) != NULL || m_a->hasNext()) ) {

                  size_t s_start,s_len;
                  s_start = m_a->m_offset - m_a->m_bufstart;

                  size_t max_select = m_pointer->maxsize() - m_pointer->size();

                  if (a_buf == NULL) {
                    a_buf = m_a->getCurrBuf();
                    if (a_buf == NULL) {
                      fclog<<seal::Msg::Warning<< "Got NULL buf during MAKE_DISTINCT" <<seal::endmsg;
                      throw FCbackendException("LFCQueryRequest::hasNext","Got NULL buf during MAKE_DISTINCT");
                    }
                  }

                  s_len = a_buf->size() - s_start;
                  if (s_len > max_select) s_len=max_select;

                  if (s_len == 0) break;

                  // for keeping track of which items are found to be duplicates
                  if (m_bool_buf.get() != NULL && m_bool_buf->maxsize() >= s_len) {
                    for(size_t i=0;i<s_len;++i) (m_bool_buf->getData())[i]=false;
                  } else {
                    LFCcounted_ptr<FCBuf<bool> > p(new FCBuf<bool>(s_len));
                    m_bool_buf = p;
                    for(size_t i=0;i<s_len;++i) m_bool_buf->push_back(false);
                  }

                  if (m_a->m_offset > 0) {
                    m_b->rewind(m_setp, m_input_set_distinct, m_cl, m_try_distinct);
                  }
                    
                  bool use_only_a = false;
                  size_t b_offset = 0;

                  while(b_offset < (m_a->m_offset + s_len - 1)) {

                    const X *itemp;
                    if (use_only_a) {
                      size_t off = b_offset - m_a->m_bufstart;
                      itemp = &((*a_buf)[off]);
                    } else {
                      if ( (b_offset >= m_a->m_bufstart) || !m_b->hasNext() ) {
                        // have queried all entries in B
                        // now start using only a
                        b_offset = m_a->m_bufstart;
                        use_only_a = true;
                        continue;
                      }
                      itemp = &m_b->next();
                    }
                    ++b_offset;

                    size_t st = 0;
                    if (b_offset > m_a->m_offset) st = (b_offset - m_a->m_offset);

                    // try to record only first occurance of entry
                    // for each entry need to know if it has been seen
                    // in a position before the current one

                    for(size_t i=st;i<s_len;++i) {
                      if ((m_bool_buf->getData())[i]) continue;
                      size_t off=i+s_start;
                      if (LFCCatalog::compare((*a_buf)[off], *itemp)) {
                        (m_bool_buf->getData())[i] = true;
                      }
                    }
                  }

                  // finished scan. Advance m_a
                  for(size_t i=0;i<s_len;++i) (void)m_a->next();

                  // add distinct entries to output
                  for(size_t i=0;i<s_len;++i) {
                    size_t off = i+s_start;
                    if (!(m_bool_buf->getData())[i]) m_pointer->push_back((*a_buf)[off]);
                  }
                }
              }
              break;

            default:
              fclog<<seal::Msg::Warning<< "Unknown object state for DISTINCT object" <<seal::endmsg;
              throw FCbackendException("LFCQueryRequest::hasNext","Unknown object state for DISTINCT object");
              break;
          }
          break;

        case AND:
          switch(m_state) {
            case AND_A_FIRST:
            case AND_B_FIRST:
              { LFCcounted_ptr<LFCQueryRequest<X> > first;
                LFCcounted_ptr<LFCQueryRequest<X> > second;
                bool *first_try_distinct;
                bool *second_try_distinct;

                if (m_state == AND_A_FIRST) {
                  first = m_a;
                  second = m_b;
                  first_try_distinct = &m_a_try_dist;
                  second_try_distinct = &m_b_try_dist;
                } else {
                  first = m_b;
                  second = m_a;
                  first_try_distinct = &m_b_try_dist;
                  second_try_distinct = &m_a_try_dist;
                }

                FCBuf<X> *first_buf;

                while( !m_pointer->full() && ((first_buf = first->getCurrBuf()) != NULL || first->hasNext()) ) {

                  size_t s_start,s_len;
                  s_start = first->m_offset - first->m_bufstart;

                  if (first_buf == NULL) {
                    first_buf = first->getCurrBuf();
                    if (first_buf == NULL) {
                      fclog<<seal::Msg::Warning<< "Got NULL buf during AND_A/B_FIRST" <<seal::endmsg;
                      throw FCbackendException("LFCQueryRequest::hasNext","Got NULL buf during AND_A/B_FIRST");
                    }

                    s_len = first_buf->size() - s_start;

                    // by keeping track of which items from first
                    // are selected, we require only the first input
                    // to be distinct in order that the result is distinct
                    if (m_bool_buf.get() != NULL && m_bool_buf->maxsize() == s_len) {
                      for(size_t i=0;i<s_len;++i) (m_bool_buf->getData())[i]=false;
                    } else {
                      LFCcounted_ptr<FCBuf<bool> > p(new FCBuf<bool>(s_len));
                      m_bool_buf = p;
                      for(size_t i=0;i<s_len;++i) m_bool_buf->push_back(false);
                    }
                  }

                  s_len = first_buf->size() - s_start;

                  bool out_distinct = first->isOutputDistinct();
                  if (*second_try_distinct) {
                    if (!out_distinct) {
                      fclog<<seal::Msg::Warning<< "Unexpected result when testing for result uniqueness" <<seal::endmsg;
                      throw FCbackendException("LFCQueryRequest::hasNext","Unexpected result when testing for result uniqueness");
                    }
                  }
                  second->rewind(first_buf, out_distinct, m_cl, *second_try_distinct);

                  size_t max_select = m_pointer->maxsize() - m_pointer->size();
                  if (max_select > s_len) max_select = s_len;

                  size_t selected = 0;
                  bool fin = false;

                  while(selected<max_select) {
                    if (second->hasNext()) {
                      const X &second_item = second->next();

                      for(size_t i=0;i<s_len;++i) {
                        if ((m_bool_buf->getData())[i]) continue;
                        size_t off = i+s_start;
                        if (LFCCatalog::compare((*first_buf)[off], second_item)) {
                          m_pointer->push_back(second_item);
                          (m_bool_buf->getData())[i] = true;
                          ++selected;
                          break;
                        }
                      }
                    } else {
                      // end of second
                      fin = true;
                      break;
                    }
                  }

                  if (fin) {
                    // finished scan of all second. Advance first
                    for(size_t i=0;i<s_len;++i) (void)first->next();
                  }
                }
              }
              break;

            case AND_A_WITH_CONSTRAINTS:
            case AND_B_WITH_CONSTRAINTS:
              { LFCcounted_ptr<LFCQueryRequest<X> > first;

                if (m_state == AND_A_WITH_CONSTRAINTS) {
                  first = m_a;
                } else {
                  first = m_b;
                }

                while(!m_pointer->full() && first->hasNext()) {
                  const X &item = first->next();
                  m_pointer->push_back(item);
                }
              }
              break;
              
            default:
              fclog<<seal::Msg::Warning<< "Unknown object state for AND object" <<seal::endmsg;
              throw FCbackendException("LFCQueryRequest::hasNext","Unknown object state for AND object");
              break;
          }
          break;

        default:
          fclog<<seal::Msg::Warning<< "Not implimented for this type of query object" <<seal::endmsg;
          throw LFCQueryException("LFCQueryRequest::hasNext",
                                  "Not implimented for this type of query object");
          break;
      }

      if (m_pointer->empty()) return false;
      return true;
    } /* hasNext() */

  private:

    FCBuf<X> *getCurrBuf() {
      if (m_offset-m_bufstart < m_pointer->size()) {
        return m_pointer.get();
      }
      return NULL;
    }

    // member vars

    enum RequestType { LIST, AND, OR, DISTINCT } m_type;
    void (LFCCatalog::*m_qfunc)(LFC_LIST_FUNC_TYPE(X));
    LFCCatalog *m_cat;
    void *m_privdata;
    std::string m_sarg;
    LFCcounted_ptr<FCBuf<X> > m_pointer;
    size_t m_offset;
    size_t m_bufstart;
    enum RequestState { NO_STATE,
                        BEGIN, CONTINUE, END,
                        AND_A_FIRST, AND_B_FIRST, AND_A_WITH_CONSTRAINTS, AND_B_WITH_CONSTRAINTS,
                        NOP_DISTINCT, MAKE_DISTINCT
                      } m_state;

    LFCCatalog::match_types_t m_matchtype;
    LFCCatalog::constraint_list m_cl;
    FCBuf<X> *m_setp;

    bool m_need_set;
    LFCCatalog::distinct_t m_list_distinct_with_set, m_list_distinct_without_set;
    LFCCatalog::constraint_list m_list_constraints;
    bool m_try_distinct;
    bool m_input_set_distinct;
    bool m_a_try_dist;
    bool m_b_try_dist;

    LFCcounted_ptr<FCBuf<bool> > m_bool_buf;
    LFCcounted_ptr<LFCQueryRequest<X> > m_a, m_b;
  };

  /* LFCQueryEvaluator class */

  template <class Item>
  class LFCQueryEvaluator {
  public:

    typedef class LFCQueryRequest<pool::LFCCatalog::udata> query_request_t;
    typedef class LFCcounted_ptr<query_request_t> base_value_t;

    typedef class std::vector<base_value_t> values_t;
    typedef class std::vector<LFCQueryPhrase> queries_t;

    typedef queries_t::iterator query_itr_t;
    typedef values_t::iterator value_itr_t; /* */
    
    typedef class std::vector<std::pair<query_itr_t, query_itr_t> > query_ranges_t;
    typedef class std::vector<std::pair<value_itr_t, value_itr_t> > value_ranges_t;

    typedef query_ranges_t::iterator query_range_itr_t;
    typedef value_ranges_t::iterator value_range_itr_t; /* */

    typedef class std::vector<values_t> expressions_list_t;
    typedef class std::vector<value_itr_t> composite_value_itrs_t;

    typedef expressions_list_t::iterator expressions_list_itr_t; /* */
    typedef composite_value_itrs_t::iterator composite_value_itrs_itr_t; /* */

      //// Constructor
    LFCQueryEvaluator(LFCCatalog *cat, queries_t &query_results,
                         void (LFCCatalog::*f1)(LFC_LIST_FUNC_TYPE(Item)),
                         void (LFCCatalog::*f2)(LFC_LIST_FUNC_TYPE(Item)),
                         void (LFCCatalog::*f3)(LFC_LIST_FUNC_TYPE(Item)))
    { m_query_results = query_results; m_cat = cat; m_f1 = f1; m_f2 = f2; m_f3 = f3; };
    
    /// destructor
    virtual ~LFCQueryEvaluator() {};
    
    base_value_t build_tree(const bool &);

    void setup_list_objs(FCBuf<Item> &buf);

    struct retrieveData_t {
      base_value_t m_root;
      size_t m_offset;
      std::string m_query;
      LFCCatalog *m_cat;
      
      void fill(LFCCatalog *cat, const std::string &q, const size_t &start, FCBuf<Item> &buf,
                 void (LFCCatalog::*f1)(LFC_LIST_FUNC_TYPE(Item)),
                 void (LFCCatalog::*f2)(LFC_LIST_FUNC_TYPE(Item)),
                 void (LFCCatalog::*f3)(LFC_LIST_FUNC_TYPE(Item))) {
        PoolMessageStream fclog( "LFCQueryEvaluator::fill", seal::Msg::Nil);

        if (m_cat != cat || m_query != q || m_offset > start) {
          m_cat = cat;
          m_offset = 0;
          m_query = q;
          LFCParser parser;
          queries_t result = parser.parse(m_query);
          LFCQueryEvaluator evaluator(m_cat,result,f1,f2,f3);
          evaluator.setup_list_objs(buf);
          m_root = evaluator.build_tree(m_cat->returns_distinct());
        }

        while(!buf.full() && m_root->hasNext()) {
          pool::LFCCatalog::udata udata = m_root->next();
          Item item;
          udata.get_item(item);
          ++m_offset;
          if (m_offset>start) {
            m_cat->L_unqualify(item);
            buf.push_back(item);
          }
        }
      }
    };

  private:

    void find_joins(std::string,
                    expressions_list_t &,
                    composite_value_itrs_t&);

    void merge_values(expressions_list_t&,
                      composite_value_itrs_t&,
                      std::string);

    values_t m_phrase_values;
    queries_t m_query_results;
    LFCCatalog * m_cat;

    void (LFCCatalog::*m_f1)(LFC_LIST_FUNC_TYPE(Item));
    void (LFCCatalog::*m_f2)(LFC_LIST_FUNC_TYPE(Item));
    void (LFCCatalog::*m_f3)(LFC_LIST_FUNC_TYPE(Item));

  };

  template <class Item>
  void LFCQueryEvaluator<Item>::find_joins(std::string join_type,
                                           expressions_list_t& expressions_list,
                                           composite_value_itrs_t& composite_value_itrs)
  {
    PoolMessageStream fclog( "LFCQueryEvaluator::find_joins", seal::Msg::Nil);

    query_ranges_t query_ranges;
    value_ranges_t value_ranges;

    query_itr_t qstart, query_itr, prev_query_itr;
    value_itr_t vstart, value_itr, prev_value_itr;

    bool found = false;

    expressions_list.clear();
    composite_value_itrs.clear();

    prev_query_itr = m_query_results.end();
    prev_value_itr = m_phrase_values.end();

    for(query_itr=m_query_results.begin(),value_itr=m_phrase_values.begin();
        query_itr != m_query_results.end();
        prev_query_itr = query_itr, ++query_itr) {

      if (!query_itr->isType(LFCQueryPhrase::JOIN)) {
        if (value_itr == m_phrase_values.end()) {
          fclog<<seal::Msg::Warning<< "List of value sets short compared with query parse results" <<seal::endmsg;
          throw LFCQueryException("LFCQueryEvaluator::find_joins",
                                  "List of value sets short compared with query parse results");
        }

        prev_value_itr = value_itr++;
        continue;
      }

      /* current query phrase is a join */
      if (!found) {
        if (query_itr->getOp() == join_type) {
          found = true;

          if (prev_query_itr == m_query_results.end() || prev_value_itr == m_phrase_values.end()) {
            fclog<<seal::Msg::Warning<< "Unable to find the starting query result or value set for set of joins" <<seal::endmsg;
            throw LFCQueryException("LFCQueryEvaluator::find_joins",
                                    "Unable to find the starting query result or value set for set of joins");
          }

          qstart = prev_query_itr;
          vstart = prev_value_itr;

        }
      } else {
        if (query_itr->getOp() != join_type) {
          found = false;
          query_ranges.push_back(std::make_pair(qstart,query_itr));
          value_ranges.push_back(std::make_pair(vstart,value_itr));
        }
      }
    }

    if (query_itr != m_query_results.end() || value_itr != m_phrase_values.end()) {
      fclog<<seal::Msg::Warning<< "Query to values length mismatch" <<seal::endmsg;
      throw LFCQueryException("LFCQueryEvaluator::find_joins",
                              "Query to values length mismatch");
    }

    // record final range
    if (found) {
      query_ranges.push_back(std::make_pair(qstart,query_itr));
      value_ranges.push_back(std::make_pair(vstart,value_itr));
    }

    // create new vectors, with the above ranges removed
    queries_t new_query;
    values_t new_phrase_values;

    // ensure the new vector for phrase values doesn't
    // need to be resized as it is being filled, as this
    // would invalidate any previous iterators taken on
    // the new elements
    new_phrase_values.reserve(m_phrase_values.size());

    // save the expression value evaluators first
    // keep the removed groups of evulators and prepare
    // the new vector

    vstart = m_phrase_values.begin();
    for(value_range_itr_t value_range_itr=value_ranges.begin();
        value_range_itr != value_ranges.end();
        ++value_range_itr) {
      values_t expressions;

      for(value_itr=value_range_itr->first;
          value_itr != value_range_itr->second;
          ++value_itr) {
        expressions.push_back(*value_itr);
      }
      expressions_list.push_back(expressions);

      while(vstart != value_range_itr->first) new_phrase_values.push_back(*vstart++);
      while(vstart != value_range_itr->second) vstart++;

      // insert the empty evaluator as place holder to receive later the merged evaluator
      base_value_t empty;
      composite_value_itrs.push_back( new_phrase_values.insert(new_phrase_values.end(), empty) );

    }
    while(vstart != m_phrase_values.end()) new_phrase_values.push_back(*vstart++);

    // now fill the new vector for the query results

    qstart = m_query_results.begin();
    for(query_range_itr_t query_range_itr=query_ranges.begin();
        query_range_itr != query_ranges.end();
        ++query_range_itr) {

      while(qstart != query_range_itr->first) new_query.push_back(*qstart++);
      while(qstart != query_range_itr->second) qstart++;

      // add in the composite/place holder into the new vectors
      LFCQueryPhrase empty_phrase(LFCQueryPhrase::COMPOSITE);
      new_query.push_back(empty_phrase);
    }
    while(qstart != m_query_results.end()) new_query.push_back(*qstart++);
    
    // swap in the new vectors

    m_query_results = new_query;
    m_phrase_values = new_phrase_values;

    // rewrite composite_value_itrs to refer to m_phrase_values
    composite_value_itrs_itr_t ci;
    composite_value_itrs_t c;
    for(value_itr=m_phrase_values.begin(), vstart = new_phrase_values.begin(), ci=composite_value_itrs.begin();
        ci != composite_value_itrs.end();
        ++value_itr, ++vstart) {
      if (*ci == vstart) {
        c.push_back(value_itr);
        ++ci;
      }
    }
    composite_value_itrs = c;
  }

  template <class Item>
  void LFCQueryEvaluator<Item>::merge_values(expressions_list_t& expressions_list,
                                             composite_value_itrs_t &composite_value_itrs,
                                             std::string join_type)
  {
    PoolMessageStream fclog( "LFCQueryEvaluator::merge_values", seal::Msg::Nil);

    expressions_list_itr_t expressions_list_itr;
    composite_value_itrs_itr_t composite_value_itrs_itr;

    for(expressions_list_itr = expressions_list.begin(),
          composite_value_itrs_itr = composite_value_itrs.begin();
        expressions_list_itr != expressions_list.end();
        ++composite_value_itrs_itr, ++expressions_list_itr) {

      if (composite_value_itrs_itr == composite_value_itrs.end()) {
        fclog<<seal::Msg::Warning<< "List of composite values set iterators short compared with list of individual values sets" <<seal::endmsg;
        throw LFCQueryException("LFCQueryEvaluator::merge_values",
                                "List of composite values set iterators short compared with list of individual values sets");
      }

      values_t &expressions = *expressions_list_itr;

      if (expressions.size() == 0) {
        fclog<<seal::Msg::Warning<< "Found an element of the individual values set with no values objects" <<seal::endmsg;
        throw LFCQueryException("LFCQueryEvaluator::merge_values",
                                "Found an element of the individual values set with no values objects");
      }

      value_itr_t start_itr = expressions.begin();
      base_value_t internal(*start_itr++);

      for(value_itr_t value_itr=start_itr;
          value_itr != expressions.end();
          ++value_itr) {

        // take the smaller of the buffer sizes of the parents
        size_t bufsize = internal->maxsize();
        if ((*value_itr)->maxsize() < bufsize) bufsize = (*value_itr)->maxsize();

        base_value_t internal_new(new query_request_t(internal,*value_itr,join_type,bufsize));

        internal = internal_new;
      }
      **composite_value_itrs_itr = internal;
    }
  }

  template<class Item>
  void LFCQueryEvaluator<Item>::setup_list_objs(FCBuf<Item> &buf)
  {
    PoolMessageStream fclog( "LFCQueryEvaluator::setup_list_objs", seal::Msg::Nil);

    queries_t &result = m_query_results;
    values_t &phrase_values = m_phrase_values;

    query_itr_t itr;

    //size_t bufsize = buf.maxsize() / ( 1 + result.size() * result.size() / 2);
    size_t bufsize = buf.maxsize();
    if (bufsize< 1000) bufsize=1000;

    for(itr=result.begin(); itr != result.end(); ++itr) {
      base_value_t value;

      if (itr->isType(LFCQueryPhrase::PFN)) {
        if (itr->getOp() == "=") {
          base_value_t p(new query_request_t(m_cat,m_f1, LFCCatalog::LITERAL, itr->getValue(), bufsize));
          value = p;
        } else if (itr->getOp() == "LIKE") {
          base_value_t p(new query_request_t(m_cat,m_f1, LFCCatalog::WILDCARD, itr->getValue(), bufsize));
          value = p;
        } else {
          fclog<<seal::Msg::Warning<< "Query comparator not implemented" <<seal::endmsg;
          throw FCbackendException("LFCCatalog::retrieveGuid","Query comparator not implemented");
        }

      } else if (itr->isType(LFCQueryPhrase::GUID)) {

        if (itr->getOp() == "=") {
          base_value_t p(new query_request_t(m_cat,m_f2, LFCCatalog::LITERAL, itr->getValue(), bufsize));
          value = p;
        } else if (itr->getOp() == "LIKE") {
          base_value_t p(new query_request_t(m_cat,m_f2, LFCCatalog::WILDCARD, itr->getValue(), bufsize));
          value = p;
        } else {
          fclog<<seal::Msg::Warning<< "Query comparator not implemented" <<seal::endmsg;
          throw FCbackendException("LFCCatalog::retrieveGuid","Query comparator not implemented");
        }

      } else if (itr->isType(LFCQueryPhrase::LFN)) {

        std::string lf_unqual = itr->getValue();
        std::string lf_qual;

        m_cat->L_qualify(lf_unqual, lf_qual);

        if (itr->getOp() == "=") {
          base_value_t p(new query_request_t(m_cat,m_f3, LFCCatalog::LITERAL, lf_qual, bufsize));
          value = p;
        } else if (itr->getOp() == "LIKE") {
          base_value_t p(new query_request_t(m_cat,m_f3, LFCCatalog::WILDCARD, lf_qual, bufsize));
          value = p;
        } else {
          fclog<<seal::Msg::Warning<< "Query comparator not implemented" <<seal::endmsg;
          throw FCbackendException("LFCCatalog::retrieveGuid","Query comparator not implemented");
        }

      } else if (itr->isType(LFCQueryPhrase::JOIN)) {
        /* No value to store for the JOIN phrase */
      } else {
        fclog<<seal::Msg::Warning<< "Unknown token in query" <<seal::endmsg;
        throw FCbackendException("LFCCatalog::retrieveGuid","Unknown token in query");
      }
      if (value.get() != NULL) {
        // add 'distinct' objects below all the list objects
        base_value_t p(new query_request_t(value, "DISTINCT"));
        value = p;
        phrase_values.push_back(value);
      }
    }
  }

  template <class Item>
  typename LFCQueryEvaluator<Item>::base_value_t LFCQueryEvaluator<Item>::build_tree(const bool &want_distinct)
  {
    PoolMessageStream fclog( "LFCQueryEvaluator::build_tree", seal::Msg::Nil);

    queries_t &query = m_query_results;
    values_t &phrase_values = m_phrase_values;

    {
      expressions_list_t expressions_list;
      composite_value_itrs_t composite_value_itrs;

      /* evalute the groups of values which should be ANDed together */
      find_joins("AND", expressions_list, composite_value_itrs);

      /* create the heirachy of LFCQueryRequest objects */
      merge_values(expressions_list, composite_value_itrs, "AND");
    } 

    {
      expressions_list_t expressions_list;
      composite_value_itrs_t composite_value_itrs;

      /* find groups of values for ORing together */
      find_joins("OR", expressions_list, composite_value_itrs);

      /* before merging the OR items, check that the query list should
         now comprise of only 1 (non join) entry: */

      if (query.size() != 1 || query[0].isType(LFCQueryPhrase::JOIN)) {
        fclog<<seal::Msg::Warning<< "Not left with value result after processing all AND and OR joins" <<seal::endmsg;
        throw LFCQueryException("LFCQueryEvaluator::build_tree",
                                "Not left with value result after processing all AND and OR joins");
      }

      /* create the heirachy of LFCQueryRequest objects */
      merge_values(expressions_list, composite_value_itrs, "OR");
    }

    /* check count of values */
    if (phrase_values.size() != 1) {
        fclog<<seal::Msg::Warning<< "Not left with single set of values after processing all AND and OR joins" <<seal::endmsg;
        throw LFCQueryException("LFCQueryEvaluator::build_tree",
                                "Not left with single set of values after processing all AND and OR joins");
    }

    base_value_t v = phrase_values[0];

    bool need_set;
    LFCCatalog::distinct_t dist_with,dist_without;
    v->info(need_set,dist_with,dist_without);

    if (need_set) {
      // lfname like '%'
      std::string lf_qual;
      m_cat->L_qualify( "%", lf_qual);
      base_value_t p(new query_request_t(m_cat,m_f3, LFCCatalog::WILDCARD, lf_qual, v->maxsize()));
      base_value_t p2(new query_request_t(p, "DISTINCT"));

      // (v) and lfname like '%'
      base_value_t internal_new(new query_request_t(v,p2,"AND",v->maxsize()));

      v = internal_new;
      v->info(need_set,dist_with,dist_without);
    }

    LFCCatalog::constraint_list c;
    v->rewind(NULL,true,c,want_distinct);

    if (need_set || (want_distinct && dist_without != LFCCatalog::YES)) {
      fclog<<seal::Msg::Warning<< "Unable to build parse tree for this query" <<seal::endmsg;
      throw LFCQueryException("LFCQueryEvaluator::build_tree",
                              "Unable to build parse tree for this query");
    }

    return v;
  }
}
#endif
