/* 
 * Concrete implementation of the POOL to LFC Catalog interface
 * Author : James Casey and David Smith
 */

static const char clientid[] = "$Id: LFCCatalog.cpp,v 1.20 2006/03/17 10:59:27 dhsmith Exp $";

extern "C" {
// These are to allow us to call the Cglobals & Cmutex init routines, which
// are from the common routines (in the lfc api library)
extern void Cmutex_init(int (*)(void *, int),int (*)(void *));
extern void Cglobals_init(int (*)(int *, void **),int (*)(int *, void *),int (*)(void));
extern int (*local_getspec)(int *, void **);
extern int (*local_setspec)(int *, void *);
extern int (*local_mutex_lock)(void *, int);
extern int (*local_mutex_unlock)(void *);
}

#include "LFCCatalog/LFCCatalog.h"

#include "LFCCatalog/LFCQueryEvaluator.h"

#include "FileCatalog/FCException.h"

#include "POOLCore/POOLContext.h"
#include "POOLCore/PoolMessageStream.h"

namespace pool {

  // static members for Cglobals handling
  pthread_mutex_t LFCCatalog::thread_data_mutex = PTHREAD_MUTEX_INITIALIZER;
  int LFCCatalog::threads_installed = 0;
  bool LFCCatalog::tids_key_created = false;
  int LFCCatalog::next_tid = 0;
  pthread_mutex_t LFCCatalog::my_cmutex_data_mutex = PTHREAD_MUTEX_INITIALIZER;
  int LFCCatalog::my_cmutex_installed = 0;
  bool LFCCatalog::gave_interfere_warning = false;

  std::map<int *, pthread_key_t> LFCCatalog::gkey_2_pkey;
  pthread_key_t LFCCatalog::tids_key;
  std::map<int, bool> LFCCatalog::tid_map;
  std::map<const LFCCatalog *, pthread_t> LFCCatalog::cat_2_thread;
  std::map<pthread_t, const LFCCatalog *> LFCCatalog::thread_2_cat;
  std::map<void *, LFCCatalog::my_cmutex_s> LFCCatalog::my_cmutex_data;

  // static variables for delayed PFNs count
  pthread_mutex_t LFCCatalog::delayed_t::numb_mutex =  PTHREAD_MUTEX_INITIALIZER;
  size_t LFCCatalog::delayed_t::numb = 0;

  LFCCatalog::LFCCatalog() : m_errbuf(new char[LFC_ERRBUFSZ+1])
  {
    m_retLFN = NULL;
    m_retPFN = NULL;
    m_retGUID = NULL;

    my_init_object_tid();
    my_init_object_keys();

    my_threadspecific_install();
    my_cmutex_install();
  }

  LFCCatalog::~LFCCatalog()
  {
    // rollback if we are still in a transaction
    if ( m_transactionsta.test( 0 ) ) {
      try {
        this->rollback();
      } catch(...) {
        // ignore
      }
    }
    // disconnect if we are still connected
    if ( m_transactionsta.test( 1 ) ) {
      this->disconnect();
    }

    // clear out bulk query buffers
    destroyRetData();

    delete m_errbuf;

    my_free_object_keys();
    my_release_tid(m_tid);
  }

  void LFCCatalog::my_init_object_tid() {
    lfc_take_mutex lk(&thread_data_mutex);
    m_tid = my_calc_new_tid();
  }

  /* STATIC member function */
  int LFCCatalog::my_calc_new_tid() {
    // EXPECTS THREAD_DATA_MUTEX TO BE LOCKED
    int i=next_tid;
    bool found = false;
    do {
      if (tid_map.find(i) == tid_map.end()) {
        found = true;
        break;
      }
    } while(++i != next_tid);

    if (!found) {
      throw FCbackendException("LFCCatalog::my_calc_new_tid", "Could not find any unused thread id");
    }

    next_tid = i+1;
    tid_map[i] = true;
    return i;
  }

  void LFCCatalog::my_threadspecific_install() {
    {
      lfc_take_mutex lk(&thread_data_mutex);
      if (!threads_installed) {
        if (local_getspec != NULL || local_setspec != NULL) {
          threads_installed = -1;
          return;
        } else {
          threads_installed = 1;
          /* fall through to do Cglobals_init() */
        }
      } else {
        if (threads_installed>1) {
          /* fine, we're completely installed */
        }
        return;
      }
    }

    Cglobals_init(&my_get_thread_specific, &my_set_thread_specific, &my_get_tid);

    {
      lfc_take_mutex lk(&thread_data_mutex);
      threads_installed = 2;
    }

  }

  /* STATIC member function */
  int LFCCatalog::my_get_tid() {
    lfc_take_mutex lk(&thread_data_mutex);
    int ret_tid;
    thread_2_cat_itr itr;
    pthread_t pself = pthread_self();
    itr = thread_2_cat.find(pself);
    if (itr != thread_2_cat.end()) {
      ret_tid = itr->second->m_tid;
    } else {
      void *val;
      if (tids_key_created) {
        val = pthread_getspecific(tids_key);
      } else {
        if (pthread_key_create(&tids_key, &my_tid_destructor)) {
          throw FCbackendException("LFCCatalog::my_get_tid", "Could not create thread id key");
        }
        tids_key_created = true;
        val = NULL;
      }
      if (val == NULL) {
        val = new int;
        if (pthread_setspecific(tids_key, val)) {
          throw FCbackendException("LFCCatalog::my_get_tid", "Could not set thread id key value");
        }
        ret_tid = my_calc_new_tid();
        *((int *)val) = ret_tid;
      } else {
        ret_tid = *((int *)val);
      }
    }
    return ret_tid;
  }

  /* STATIC member function */
  void LFCCatalog::my_tid_destructor(void *arg) {
    int *ip = (int *)arg;

    my_release_tid(*ip);
    delete ip;
  }

  /* STATIC member function */
  void LFCCatalog::my_release_tid(int tid) {
    lfc_take_mutex lk(&thread_data_mutex);
    tid_map.erase(tid);
  }

  /* STATIC member function */
  int LFCCatalog::my_get_thread_specific(int *key, void **val) {
    if (key == NULL || val == NULL) {
      throw FCbackendException("LFCCatalog::my_get_thread_specific", "Invalid argument");
    }
    lfc_take_mutex lk(&thread_data_mutex);
    thread_2_cat_itr itr;
    pthread_t pself = pthread_self();
    itr = thread_2_cat.find(pself);
    if (itr != thread_2_cat.end()) {
      std::map<int *, void *> &gkey_to_vp = itr->second->m_gkey_to_vp;
      if (gkey_to_vp.find(key) == gkey_to_vp.end()) {
        gkey_to_vp[key] = NULL;
      }
      *val = gkey_to_vp[key];
    } else {
      gkey_2_pkey_itr itr2;
      pthread_key_t pkey;
      itr2 = gkey_2_pkey.find(key);
      if (itr2 != gkey_2_pkey.end()) {
        pkey = itr2->second;
        my_key_val_pair *myp = (my_key_val_pair *)pthread_getspecific(pkey);
        if (myp != NULL) {
          *val = myp->val;
        }
      } else {
        if (pthread_key_create(&pkey, &my_key_destructor)) {
          throw FCbackendException("LFCCatalog::my_get_thread_specific", "Could not create key");
        }
        gkey_2_pkey[key] = pkey;
        my_key_val_pair *myp = new my_key_val_pair;
        if (pthread_setspecific(pkey, (void *)myp)) {
          delete myp;
          throw FCbackendException("LFCCatalog::my_get_thread_specific", "Could not set key value");
        }
        *val = NULL;
        myp->val = val;
        myp->key = key;
      }
    }
    return 0;
  }

  /* STATIC member function */
  int LFCCatalog::my_set_thread_specific(int *key, void *val) {
    if (key == NULL) {
      throw FCbackendException("LFCCatalog::my_set_thread_specific", "Invalid argument");
    }
    lfc_take_mutex lk(&thread_data_mutex);
    thread_2_cat_itr itr;
    pthread_t pself = pthread_self();
    itr = thread_2_cat.find(pself);
    if (itr != thread_2_cat.end()) {
      std::map<int *, void *> &gkey_to_vp = itr->second->m_gkey_to_vp;
      if (gkey_to_vp.find(key) == gkey_to_vp.end()) {
        /* according to the Cglobals man page getspecfic should already have been called */
        errno = EINVAL;
        return -1;
      }
      gkey_to_vp[key] = val;
    } else {
      gkey_2_pkey_itr itr2;
      pthread_key_t pkey;
      itr2 = gkey_2_pkey.find(key);
      if (itr2 != gkey_2_pkey.end()) {
        pkey = itr2->second;
        my_key_val_pair *myp = (my_key_val_pair *)pthread_getspecific(pkey);
        if (myp == NULL) {
          throw FCbackendException("LFCCatalog::my_set_thread_specific", "Invalid key value pair");
        }
        myp->val = val;
      } else {
        /* according to the Cglobals man page getspecfic should already have been called */
        errno = EINVAL;
        return -1;
      }
    }
    return 0;
  }

  /* STATIC member function */
  void LFCCatalog::my_key_destructor(void *arg) {
    my_key_val_pair *myp = (my_key_val_pair *)arg;
    if (myp == NULL) return;
    if (myp->key == NULL) {
      throw FCbackendException("LFCCatalog::my_key_destructor", "Invalid key value");
    }
    lfc_take_mutex lk(&thread_data_mutex);
    gkey_2_pkey.erase(myp->key);
    if (myp->val != NULL) {
      free(myp->val);
    }
    delete myp;
  }

  void LFCCatalog::my_init_object_keys() {
    lfc_take_mutex lk(&thread_data_mutex);
    m_thread_enter_count = 0;
    m_thread_acquire_count = 0;
    if (pthread_cond_init(&m_thread_cond, NULL)) {
      throw FCbackendException("LFCCatalog::my_init_object_keys", "Failure to initalise condition variable");
    }
    cat_2_thread[this] = pthread_self();
  }

  void LFCCatalog::my_free_object_keys() {
    lfc_take_mutex lk(&thread_data_mutex);
    gkey_to_vp_itr itr;
    for(itr=m_gkey_to_vp.begin(); itr != m_gkey_to_vp.end(); ++itr) {
      void *vp = itr->second;
      if (vp != NULL) {
        free(vp);
      }
    }
    m_gkey_to_vp.clear();
    cat_2_thread_itr itr2 = cat_2_thread.find(this);
    if (itr2 != cat_2_thread.end()) {
      if (m_thread_enter_count != 0 || m_thread_acquire_count != 0) {
        throw FCbackendException("LFCCatalog::my_free_object_keys", "Invalid thread enter or acquire count");
      }
      if (pthread_cond_destroy(&m_thread_cond)) {
        throw FCbackendException("LFCCatalog::my_free_object_keys", "Failure to destroy condition variable");
      }
      cat_2_thread.erase(this);
    } else {
      throw FCbackendException("LFCCatalog::my_free_object_keys", "Current catalog object not found");
    }
  }

  /* STATIC member function */
  int LFCCatalog::my_cmutex_lockroutine(void *addr, int timeout) {
    lfc_take_mutex lk(&my_cmutex_data_mutex);
    pthread_t pself = pthread_self();
    my_cmutex_data_itr itr;
    itr = my_cmutex_data.find(addr);
    if (itr == my_cmutex_data.end()) {
      my_cmutex_s data;
      data.locked = true;
      data.owner_id = pself;
      if (pthread_cond_init(&data.wakeup_cond,NULL)) {
        throw FCbackendException("LFCCatalog::my_cmutex_lockroutine", "Failure to initalise condition variable");
      }
      data.number_waiting = 0;
      my_cmutex_data[addr] = data;
      return 0;
    }
    my_cmutex_s &data = itr->second;
    if (!data.locked) {
      data.locked = true;
      data.owner_id = pself;
      return 0;
    }
    if (data.owner_id == pself) {
      serrno = EDEADLK;
      return -1;
    }
    if (timeout == 0) {
      serrno = EBUSY;
      return -1;
    }
    data.number_waiting++;

    struct timeval now;
    struct timespec to;

    if (timeout>0) {
      gettimeofday(&now, NULL);
      to.tv_sec = now.tv_sec + timeout;
      to.tv_nsec = now.tv_usec * 1000;
    }

    int retcode;
    do {
      if (timeout>0) {
        retcode = pthread_cond_timedwait(&data.wakeup_cond, &my_cmutex_data_mutex, &to);
      } else {
        retcode = pthread_cond_wait(&data.wakeup_cond, &my_cmutex_data_mutex);
      }
    } while(data.locked && retcode == EINTR);

    data.number_waiting--;
    if (data.locked) {
      if (retcode == ETIMEDOUT)
        serrno = SETIMEDOUT;
      else
        serrno = SEINTERNAL;
      return -1;
    }
    data.locked = true;
    data.owner_id = pself;
    return 0;
  }
      
  /* STATIC member function */
  int LFCCatalog::my_cmutex_unlockroutine(void *addr) {
    lfc_take_mutex lk(&my_cmutex_data_mutex);
    pthread_t pself = pthread_self();
    my_cmutex_data_itr itr;
    itr = my_cmutex_data.find(addr);
    if (itr == my_cmutex_data.end()) {
      serrno = EINVAL;
      return -1;
    }
    my_cmutex_s &data = itr->second;
    if (!data.locked || data.owner_id != pself) {
      serrno = EPERM;
      return -1;
    }
    data.locked = false;
    if (data.number_waiting>0) {
      if (pthread_cond_signal(&data.wakeup_cond)) {
        throw FCbackendException("LFCCatalog::my_cmutex_unlockroutine", "Failure to signal on condition variable");
      }
    } else {
      if (pthread_cond_destroy(&data.wakeup_cond)) {
        throw FCbackendException("LFCCatalog::my_cmutex_unlockroutine", "Failure to destroy condition variable");
      }
      my_cmutex_data.erase(addr);
    }
    return 0;
  }

  void LFCCatalog::my_cmutex_install() {
    {
      lfc_take_mutex lk(&my_cmutex_data_mutex);
      if (my_cmutex_installed) return;
      my_cmutex_installed = 1;
      if (local_mutex_lock != NULL || local_mutex_unlock != NULL) {
        my_cmutex_installed = -1;
        return;
      }
    }

    Cmutex_init(&my_cmutex_lockroutine, &my_cmutex_unlockroutine);

    {
      lfc_take_mutex lk(&my_cmutex_data_mutex);
      my_cmutex_installed = 2;
    }
  }

  void
  LFCCatalog::connect() 
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::connect", seal::Msg::Nil);

    if(m_url.empty()){
      fclog<<seal::Msg::Warning<< "empty url" <<seal::endmsg;
      throw FCillegalContactStringException("LFCCatalog::connect","empty url");
    }

    std::string url = m_url;

    //
    //look up protocol - host separator and process protocol
    //  
    size_t colpos=url.find("://");
    std::string protocol;
    if(colpos != url.npos ){
      protocol=url.substr(0,colpos);
      // remove protocol and separator from the url
      url = url.substr(colpos+3, url.size() - colpos - 3);
    }

    if (protocol == "lfc") {
      m_want_session = false;
    } else if (protocol == "lfcsess") {
      m_want_session = true;
    } else {
      fclog<<seal::Msg::Warning<< "unknown protocol" <<seal::endmsg;
      throw FCillegalContactStringException("LFCCatalog::connect","unknown protocol");
    }

    // check for final paths after host/port
    std::string paths;
    colpos = url.find("/");
    if (colpos != url.npos) {
      paths = url.substr(colpos, url.length()-colpos);
      // remove the paths from the url
      url = url.substr(0, colpos);
    }

    // process the paths
    std::string lfc_home, lfc_generated;

    lfc_home = paths;

    // separate off the generated path, if specified
    colpos=lfc_home.find(":");
    if (colpos != lfc_home.npos) {
      lfc_generated = lfc_home.substr(colpos+1,lfc_home.length()-colpos-1);
      lfc_home = lfc_home.substr(0,colpos);
    }

    // setup the path prefix to use for relative LFNs
    if (m_lfc_home.empty()) {
      const char *p;
      if (!lfc_home.empty() && lfc_home != "/") {
        p = lfc_home.c_str();
      } else {
        p = getenv("LFC_HOME");
      }
      char *ns = NULL;
      if (p != NULL) {
        ns = strdup(p);
      }
      m_lfc_home = "/";
      if (ns != NULL) {
        m_lfc_home += std::string(ns);
        L_regularise_path(m_lfc_home);
        free(ns);
      }
    }

    // if there is a path use it for the generated path
    // (nb depends on m_lfc_home being defined)
    if (m_gen_path.empty()) {
      const char *p;
      std::string qual_path;
      if (!lfc_generated.empty()) {
        this->L_qualify(lfc_generated,qual_path);
        p = qual_path.c_str();
      } else {
        p = getenv("LFC_POOL_GENERATED");
      }
      std::string new_name;
      if (p != NULL && strlen(p)>0) {
        new_name.assign(p);
        L_regularise_path(new_name);
      } else {
        new_name.assign("/grid/dteam/generated/pool");
      }
      L_qualify(new_name, m_gen_path);
    }

    // check for host/port specification
    std::string host, port;
    host = url;
    colpos=host.find(":");
    if (colpos != host.npos) {
      port=host.substr(colpos+1,host.length()-colpos-1);
      // remove the port from the host name
      host=host.substr(0,colpos);
    }

    if (host.empty()) {
      fclog<<seal::Msg::Warning<< "bad host specifier" <<seal::endmsg;
      throw FCillegalContactStringException("LFCCatalog::connect","bad host specifier");
    }

    m_host = host;
    m_port = port;

    m_transactionsta.set( 1 );//set connection flag to 1  

    // do we want to try to ensure unique results for bulk queries?
    const char *p = getenv("LFC_POOL_ALLOW_DUPLICATES");
    if (p != NULL) {
      m_want_distinct = false;
    } else {
      p = getenv("LFC_POOL_REMOVE_DUPLICATES");
      if (p != NULL) {
        m_want_distinct = true;
      } else {
        m_want_distinct = false;
      }
    }

   // do we want to attempt to avoid leaving dangling symlinks after deletes?
   p = getenv("LFC_POOL_ALLOW_DANGLING_LINKS_ON_DELETE");
   if (p != NULL) {
     m_allow_dangling_lfns = true;
   } else {
     m_allow_dangling_lfns = false;
   }
  }

  void
  LFCCatalog::disconnect() const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::disconnect", seal::Msg::Nil);

    if ( m_transactionsta.test( 1 ) ) {
      // connected
      if ( m_transactionsta.test(0)) {
        // still in a transaction
         this->rollback();
      }
      m_transactionsta.reset( 1 );//set connection flag to 0
    } else {
      fclog<<seal::Msg::Warning<< "Not connected to a catalog - cannot disconnect" <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::disconnect",
				    "Not connected to a catalog - cannot disconnect" );
    }
    m_host.clear();
    m_port.clear();
  }
  
  void
  LFCCatalog::start() const 
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::start", seal::Msg::Nil);

    if ( ! m_transactionsta.test( 1 ) ) {
      fclog<<seal::Msg::Warning<< "Not yet connected to a database. Cannot start a new transaction." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::start",
				    "Not yet connected to a database. Cannot start a new transaction." );
    }
    if (  m_transactionsta.test( 0 ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has already started" <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::start",
				    "A transaction has already started" );
    }

    // sets a time after which something should have been sent on the transaction connection
    (void)time(&m_timek);

    {
      std::string hostandport;
      hostandport = m_host;
      if (!m_port.empty()) {
        hostandport += std::string(":") + m_port;
      }
      if (!m_want_session) {
        if( lfc_starttrans((char *)(hostandport.c_str()), (char *)clientid) < 0) {
          std::string s = L_build_serr("Could not start a new transaction");
          this->L_transactionDone();
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCTransactionException( "LFCCatalog::start",s);
        }
      } else {
        if( lfc_startsess((char *)(hostandport.c_str()), (char *)clientid) < 0) {
          std::string s = L_build_serr("Could not start a new session");
          this->L_transactionDone();
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCTransactionException( "LFCCatalog::start",s);
        }
      }
    }
    m_transactionsta.set( 0 );

    // clear out bulk query buffers
    destroyRetData();

    // no saved working directories yet
    m_cwds.clear();

    // make sure we have an inital dir saved in the cd stack;
    // also causes current directory server to be set in thip in LFC api
    if (!push_cd(m_host+":"+m_port+":"+m_lfc_home)) {
      // infact if push_cd failes it will throw as the chdir will have aborted the transaction
      std::string s = L_build_serr("Could not chdir to " + m_lfc_home); 
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::start",s);
    }

    // no delayed PFN registration yet
    m_delayed_pfn.clear();
  }
  
  
  void
  LFCCatalog::commit( FileCatalog::CommitMode ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::commit", seal::Msg::Nil);

    if ( ! m_transactionsta.test( 0 ) ) {
      return;
      //throw FCTransactionException( "LFCCatalog::commit",
      //                              "Not in a transaction state. Cannot commit." );
    }

    commitAllDelayedPFN();

    if (!m_want_session) {
      if(lfc_endtrans() < 0) {
        std::string s = L_build_serr("Could not commit transaction");
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCTransactionException( "LFCCatalog::commit",s);
      }
    } else {
      if(lfc_endsess() < 0) {
        std::string s = L_build_serr("Could not end session");
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCTransactionException( "LFCCatalog::commit",s);
      }
    }

    m_transactionsta.reset( 0 );
    this->L_transactionDone();
  }

  
  void
  LFCCatalog::rollback() const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::rollback", seal::Msg::Nil);

    if ( ! m_transactionsta.test( 0 ) ) {
      return;
      //throw FCTransactionException( "LFCCatalog::rollback",
      //                              "Not in a transaction state. Cannot roll back." );
    }

    if (!m_want_session) {
      if ( lfc_aborttrans() < 0) {
        std::string s = L_build_serr("Could not rollback a transaction");
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCTransactionException( "LFCCatalog::rollback",s);
      }
    } else {
      if ( lfc_endsess() < 0) {
        std::string s = L_build_serr("Could not end session");
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCTransactionException( "LFCCatalog::rollback",s);
      }
    }

    m_transactionsta.reset( 0 );
    this->L_transactionDone();
  }

  // called when we expect transaction to have finished
  // ie. after commit or rollback OR after certain database operations failed
  // preserve error numbers
  void
  LFCCatalog::L_transactionDone() const {
    int serr = serrno;
    int err = errno;

    PoolMessageStream fclog( "LFCCatalog::L_transactionDone", seal::Msg::Nil);

    if (m_transactionsta.test( 0 )) {
      // we haven't done an explicit committ or rollback
      if (!m_want_session) {
        // the lfc api should already have closed connection
        // if we try again it will lookup desintation host in the enviornment
        // (void)lfc_aborttrans();
      } else {
        // session will still be effect, so close it
        (void)lfc_endsess();
      }
      m_transactionsta.reset( 0 );
    }

    serrno = serr;
    errno = err;
  }

  // preserve error numbers
  std::string
  LFCCatalog::L_build_err(const std::string &s) const {
    std::string serr(strerror(errno));

    return s + ": " + serr;
  }

  // preserve error numbers
  std::string
  LFCCatalog::L_build_serr(const std::string &s) const {
    std::string serr(sstrerror(serrno));
    errno = ECOMM;
    if (serrno < 1000)
      errno = serrno;

    return s + ": " + serr + " (from lfc api)";
  }

  /* STATIC memeber function */
  void LFCCatalog::L_regularise_path(std::string &path) {

    if (path.empty()) return;

    size_t pos;

    /* find any repeated backslash and replace with single backslash */
    while( (pos=path.find("//")) != path.npos ) {
      path.replace(pos, 2, "/");
    }

    /* if we're a single backslash return */
    if (path == "/") return;

    /* remove any trailing backslash */
    if (path[path.size()-1] == '/') path.erase(path.size()-1,1);
  }

  void
  LFCCatalog::L_qualify(const std::string &unqual, std::string &qual) const {

    qual = unqual;
    L_regularise_path(qual);

    if (!qual.empty() && qual[0] == '/') 
      return;

    if (m_lfc_home.empty() || m_lfc_home[m_lfc_home.size()-1] != '/') {
      qual = std::string("/") + qual;
    }

    qual = m_lfc_home + qual;

    return;
  }

  void
  LFCCatalog::L_unqualify(LFNEntry &lfn) const {
    std::string out;
    L_unqualify(lfn.lfname(), out);
    LFNEntry new_lfn(out, lfn.guid());
    lfn=new_lfn;
  }

  void
  LFCCatalog::L_unqualify(PFNEntry __attribute__((unused)) &pfn) const { return; }

  void
  LFCCatalog::L_unqualify(FileCatalog::FileID __attribute__((unused)) &guid) const { return; }

  void
  LFCCatalog::L_unqualify(const std::string &in, std::string &out) const {

    out = in;
    L_regularise_path(out);

    if (out.size()==0 || out.substr(0,1) != "/")
      return;

    if (m_lfc_home.empty() || m_lfc_home == "/")
      return;

    if (out.find(m_lfc_home) == 0) {
      size_t off = m_lfc_home.size();
      size_t len = out.size();

      if (len>(off+1) && out[off] == '/') {
        out = out.substr(off+1,len-off-1);
      }
    }

    return;
  }

  LFNEntry
  LFCCatalog::generateLFN(const FileCatalog::FileID &fid) const {
    PoolMessageStream fclog( "LFCCatalog::generateLFN", seal::Msg::Nil);

    if (m_gen_path.empty()) {
      fclog<<seal::Msg::Warning<< "No base pathname for generated LFNs" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::generateLFN", "No base pathname for generated LFNs");
    }

    if (fid.empty()) {
      LFNEntry lfn;
      generateLFN(lfn);
      return lfn;
    }

    std::string new_name = m_gen_path;

    new_name += std::string("/") + fid;
    LFNEntry lfn(new_name, fid);
    return lfn;
  }

  void
  LFCCatalog::generateLFN(LFNEntry &my_lfn) const {
    FileCatalog::FileID fid = my_lfn.guid();
    if (fid.empty()) {
      Guid guid;
      Guid::create(guid);
      std::string ng = guid.toString();
      fid = ng;
    }

    LFNEntry new_lfn(my_lfn.lfname(), fid);
    
    if (new_lfn.lfname().empty()) {
      new_lfn =  generateLFN(fid);
    }

    my_lfn = new_lfn;
  }

  // private
  // (m_delayed_pfn is mutable)
  void
  LFCCatalog::commitAllDelayedPFN() const {
    if (m_delayed_pfn.empty())
      return;

    // for a delayed PFN registration we know
    // the guid and pfn doesn't yet exist

    delayed_t::pfns_t pfns;
    while((pfns = m_delayed_pfn.get_any_pfns()).size()>0) {
      bool first = true;
      for(delayed_t::pfns_t::const_iterator it2=pfns.begin();
          it2 != pfns.end(); ++it2) {
        this->insertPFN(*it2, (first) ? 2 : 1, 2);
        first = false;
      }
    }
    if (!m_delayed_pfn.empty())
      throw FCbackendException("LFCCatalog::commitAllDelayedPFN", "Still some delayed PFNs after commit");
  }

  void
  LFCCatalog::registerPFN( const std::string& pfn,
			   const std::string& filetype,
			   FileCatalog::FileID& fid ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::registerPFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::registerPFN",
				    "A transaction has not been started." );
    }

    std::string ft;
    this->lookupFileByPFN( pfn, fid, ft );

    if (fid.empty()) {
      // Need a new GUID
      LFNEntry lfn;
      generateLFN(lfn);
      fid = lfn.guid();
    } else {
      //fclog<<seal::Msg::Warning<< "Duplicate PFN: " << pfn <<seal::endmsg;
      throw FCduplicatePFNException("LFCCatalog::registerPFN");
    }

    PFNEntry entry(pfn, fid, filetype);
    this->insertPFN(entry, 0, 0);
  }

  
  void
  LFCCatalog::registerLFN( const std::string& pfn,
			   const std::string& lfn_unqual ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::registerLFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::registerLFN",
				    "A transaction has not been started." );
    }

    std::string lfn;
    this->L_qualify(lfn_unqual, lfn);

    FileCatalog::FileID fid;
    std::string filetype;

    if (lfn.empty()) {
      fclog<<seal::Msg::Warning<< "No LFN name supplied" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::registerLFN", "No LFN name supplied");
    }

    // check to see if the LFN already exists
    {
      struct lfc_filestat stat; 
      if(lfc_lstat(lfn.c_str(), &stat) < 0 ) {
        if (serrno != ENOENT) {
          std::string s = L_build_serr("Could not lstat LFN: " + lfn);
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::registerLFN",s);
        }
      } else {
        //fclog<<seal::Msg::Warning<< "Duplicate LFN: " << lfn <<seal::endmsg;
        throw FCduplicateLFNException("LFCCatalog::registerLFN");
      }
    }

    // see if we're registering an LFN to a delayed PFN
    delayed_t::pfns_t delayed = m_delayed_pfn.find("",pfn);
    if (delayed.size()==1) {
      fid = delayed[0].guid();
      LFNEntry l(lfn,fid);
      // insertLFN will add the delayed PFNs
      this->insertLFN(l);
      return;
    }
    // the PFN 'pfn' doesn't exist in the delayed PFN list

    if ( pfn.empty() ) {
       // PFN not specified. Throw
       fclog<<seal::Msg::Warning<< "No PFN specified" <<seal::endmsg;
       throw FCnonexistentFileException( "LFCCatalog::registerLFN", "No PFN specified" );
    }

    this->lookupFileByPFN( pfn, fid, filetype );
    if (fid.empty()) {

      // PFN defined but does not exist. Throw

      //fclog<<seal::Msg::Warning<< "No such PFN: " << pfn <<seal::endmsg;
      throw FCnonexistentFileException( "LFCCatalog::registerLFN", pfn );
    }

    LFNEntry l(lfn,fid);
    this->insertLFN(l);
  }   

  void
  LFCCatalog::addReplicaPFN( const std::string& pf,
			     const std::string& rpf ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::addReplicaPFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::addReplicaPFN",
				    "A transaction has not been started." );
    }

    FileCatalog::FileID fid;
    std::string ftype;

    this->lookupFileByPFN( pf, fid, ftype );
    if ( fid.empty() ) {
      //fclog<<seal::Msg::Warning<< "No such PFN: " << pf <<seal::endmsg;
      throw FCnonexistentFileException( "LFCCatalog::addReplicaPFN", pf );
    }
    PFNEntry entry(rpf, fid, ftype);
    this->insertPFN(entry, 3, -1);
  }


  void
  LFCCatalog::addPFNtoGuid( const FileCatalog::FileID& guid,
			    const std::string& pf,
			    const std::string& filetype ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::addPFNtoGuid", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::addPFNtoGuid",
				    "A transaction has not been started." );
    }

    if (m_delayed_pfn.has_pfname(pf)) {
      throw FCduplicatePFNException("LFCCatalog::addPFNtoGuid");
    }

    delayed_t::pfns_t delayed = m_delayed_pfn.find(guid,"");
    if (delayed.size()>0) {
      PFNEntry entry(pf, guid, filetype);
      this->insertPFN(entry, 2, 4);
      return;
    }
    // neither guid nor pf is in the delayed PFN list

    int guid_exists;
    if ((guid_exists=lfc_guid_exists(guid.c_str(), m_errbuf, LFC_ERRBUFSZ)) < 0) {
      std::string s = L_build_err("Error while checking for the existance of GUID : " + guid);
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::addPFNtoGuid", s);
    }

    if (!guid_exists) {
      //fclog<<seal::Msg::Warning<< "No such GUID: " << guid <<seal::endmsg;
      throw FCnonexistentFileException("LFCCatalog::addPFNtoGuid", guid);
    }

    PFNEntry entry(pf, guid, filetype);
    this->insertPFN(entry, 5, 4);
  }


  void
  LFCCatalog::renamePFN( const std::string& pf,
			 const std::string& newpf ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::renamePFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::renamePFN",
				    "A transaction has not been started." );
    }

    if (m_delayed_pfn.has_pfname(newpf)) {
      throw FCduplicatePFNException("LFCCatalog::renamePFN");
    }

    delayed_t::pfns_t delayed = m_delayed_pfn.find("",pf);
    if (delayed.size()==1) {
      m_delayed_pfn.remove_pfname(pf);
      PFNEntry entry(newpf, delayed[0].guid(), delayed[0].filetype());
      m_delayed_pfn.insert_pfn(entry);
      return;
    }
    // neither the pf nor newpf are in the delayed PFN list
    // at the moment we don't pass this information to methods below

    addReplicaPFN(pf, newpf);
    deletePFN(pf);
  }
  

  void
  LFCCatalog::lookupFileByPFN( const std::string& pf,
			       FileCatalog::FileID& fid,
			       std::string& ftype ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::lookupFileByPFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::lookupFileByPFN",
				    "A transaction has not been started." );
    }

    delayed_t::pfns_t delayed = m_delayed_pfn.find("",pf);
    if (delayed.size()) {
      fid = delayed[0].guid();
      ftype = delayed[0].filetype();
      return;
    }
    // pf is not in the delayed PFN list

    // JC : We don't store filetype
    ftype="";

    char *guid;
    if((guid = lfc_guidforpfn(pf.c_str(), m_errbuf, LFC_ERRBUFSZ)) == NULL) {
      if(errno == ENOENT) {
	fid=""; 
	return;
      }
      fclog<<seal::Msg::Warning<< "Could not lookup File" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::lookupFileByPFN",
			       "Could not lookup File");    
    }
    fid = std::string(guid);
    free (guid);
  }

  void
  LFCCatalog::lookupFileByLFN( const std::string& lf_unqual,
			       FileCatalog::FileID& fid) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::lookupFileByLFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::lookupFileByLFN",
				    "A transaction has not been started." );
    }

    std::string lf;
    this->L_qualify(lf_unqual, lf);

    char *guid;
    if((guid = lfc_guidfromlfn(lf.c_str(), m_errbuf, LFC_ERRBUFSZ)) == NULL) {
      int err = errno;
      std::string s = L_build_err("Could not lookup file " + lf);
      fid = "";
      if (err != ENOENT) {
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::lookupFileByLFN",s);
      }
      return;
    }
    fid = std::string(guid);
    free (guid);
  }

  void
  LFCCatalog::lookupBestPFN( const FileCatalog::FileID& fid,
			     const FileCatalog::FileOpenMode&, 
			     const FileCatalog::FileAccessPattern&,
			     std::string& pf,
			     std::string& filetype ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::lookupBestPFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::lookupFileByPFN",
				    "A transaction has not been started." );
    }

    // no filetype
    filetype.clear();
    pf.clear();
    char **surls;

    delayed_t::pfns_t delayed = m_delayed_pfn.find(fid,"");
    if (delayed.size()>0) {
      surls = (char **)malloc(sizeof(char *) * (delayed.size() + 1));
      surls[delayed.size()] = NULL;
      delayed_t::pfns_t::const_iterator itr;
      int i;
      for(i=0,itr=delayed.begin(); itr != delayed.end(); ++itr, ++i) {
        surls[i] = strdup(itr->pfname().c_str());
      }
    } else {
      if((surls = lfc_surlsfromguid(fid.c_str(),m_errbuf,LFC_ERRBUFSZ)) == NULL) {
        fclog<<seal::Msg::Warning<< "Could not lookup list of PFNs" <<seal::endmsg;
        throw FCbackendException("LFCCatalog::lookupBestPFN","Could not lookup list of PFNs");
      }
    }

    int num=0;
    while(surls[num] != NULL) num++;

    char *result = getbestfile(surls, num, m_errbuf, LFC_ERRBUFSZ);
    char **cp;

    for(cp = surls; *cp != NULL; ++cp) {
      if(*cp != result) {
        free (*cp);
      }
    }
    free (surls);

    if (result != NULL) {
      pf.assign(result);
      free(result);
    }
  }

  void
  LFCCatalog::insertPFN( PFNEntry& pentry) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::insertPFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::insertPFN",
				    "A transaction has not been started." );
    }

    this->insertPFN(pentry, -1, -1);
  }

  // private method:
  //   ge GUID exists hint
  //   pe PFN exists hint
  //   HINTS: -1 == DON'T KNOW ANYTHING
  //           0 == DOES NOT EXIST IN CATALOG OR DELAYED LIST
  //           1 == DOES EXIST IN BOTH CATALOG AND DELAYED LIST
  //           2 == DOES EXIST IN DELAYED LIST AND NOT IN CATALOG
  //           3 == EXISTS IN AT LEAST 1 OF CATALOG OR DELAYED LIST
  //           4 == DOES NOT EXIST IN DELAYED LIST MAY EXIST IN CATALOG
  //           5 == DOES EXIST IN CATALOG AND NOT IN DELAYED LIST
  //
  // A duplicate PFN exception will be thrown if the pfn already exists in the
  // catalog or in the delayed list unless the pe hint is 1, 2 or 3.
  //
  // in this routine we do not assume or enforce that entries in the delayed list
  // have any implications for entries in the catalog, other than those given in the
  // hints. The more general rule that pfns exist only in one list or the other and
  // that PFNs of the same GUID are not split between delayed list and catalog are
  // enforced in the higher level methods. Upon the delayed list becoming full we
  // do however need to ensure that at least all GUIDs in the list that are shared
  // with the current request are removed.
  void
  LFCCatalog::insertPFN(const PFNEntry& pentry, const int &ge, const int &pe ) const
  {
    PoolMessageStream fclog( "LFCCatalog::insertPFN", seal::Msg::Nil);

    if ( (ge == 0 && (pe == 1 || pe == 2 || pe == 3 || pe == 5)) ||
         (ge == 2 && (pe == 1 || pe == 5)) ||
         (ge == 4 && (pe == 1 || pe == 2)) ||
         (ge == 5 && (pe == 1 || pe == 2)) ) {
      throw FCbackendException("LFCCatalog::insertPFN","Internal error: Inconsistent existence hints");
    }

    FileCatalog::FileID guid = pentry.guid();

    int guid_exists_del;
    if (ge == -1 || ge == 3) {
      if (m_delayed_pfn.has_guid(guid)) {
        guid_exists_del = 1;
      } else {
        guid_exists_del = 0;
      }
    } else if (ge == 0 || ge == 4 || ge == 5) {
      guid_exists_del = 0;
    } else {
      guid_exists_del = 1;
    }

    int pfn_exists_del;
    if (guid_exists_del == 0) {
      pfn_exists_del = 0;
    } else if (pe == -1 || pe == 3) {
      if (m_delayed_pfn.has_pfname(pentry.pfname())) {
        pfn_exists_del = 1;
      } else {
        pfn_exists_del = 0;
      }
    } else if (pe == 0 || pe == 4 || pe == 5) {
      pfn_exists_del = 0;
    } else {
      pfn_exists_del = 1;
    }

    int guid_exists_cat;
    if ((ge == -1) ||
        (ge == 3 && guid_exists_del == 1) ||
        (ge == 4)) {
      if ((guid_exists_cat=lfc_guid_exists(guid.c_str(), m_errbuf, LFC_ERRBUFSZ)) < 0) {
        std::string s = L_build_err("Error while checking for the existance of GUID : " + guid);
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::insertPFN", s);
      }
    } else if ( (ge == 0) || (ge == 2) ) {
      guid_exists_cat = 0;
    } else {
      guid_exists_cat = 1;
    }

    int pfn_exists_cat;
    if (guid_exists_cat == 0) {
      pfn_exists_cat = 0;
    } else if ((pe == -1) ||
        (pe == 3 && pfn_exists_del == 1) ||
        (pe == 4)) {
      std::string ft;
      FileCatalog::FileID existing_pfn_guid;
      this->lookupFileByPFN(pentry.pfname(), existing_pfn_guid, ft);
      pfn_exists_cat = existing_pfn_guid.empty() ? 0 : 1;
    } else if ( (pe == 0) || (pe == 2) ) {
      pfn_exists_cat = 0;
    } else {
      pfn_exists_cat = 1;
    }

    if (pfn_exists_cat || (pfn_exists_del && pe != 1 && pe != 2 && pe != 3)) {
      // already exists: throw even if it is possibly already registered with the desired guid
      //fclog<<seal::Msg::Warning<< "Duplicate PFN: " << pentry.pfname() <<seal::endmsg;
      throw FCduplicatePFNException("LFCCatalog::insertPFN");
    }

    if (guid_exists_cat) {
      if(lfc_register_pfn(guid.c_str(), pentry.pfname().c_str(), NULL, 0) < 0) {
        std::string s = L_build_err("Could not register the PFN (" + pentry.pfname() + ") to existing GUID: " + guid);
        this->L_transactionDone();
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::insertPFN",s);
      } else {
        // attempt to register the pfn was ok
        if (pfn_exists_del)
          m_delayed_pfn.remove_pfname(pentry.pfname());
        return;
      }
    } else {
      if (!pfn_exists_del) {
        m_delayed_pfn.insert_pfn(pentry);
        if (m_want_session || m_delayed_pfn.total_size() > MAX_DELAYED_PFNS) {
          // if delayed list is full just commit them all
          commitAllDelayedPFN();
        }
        return;
      }

      // Generate a new LFN/GUID, using specified guid

      LFNEntry lfn("",guid);
      generateLFN(lfn);

      if(lfc_create_alias(guid.c_str(), lfn.lfname().c_str(), 0, m_errbuf, LFC_ERRBUFSZ) <0 ) {
        std::string s =  L_build_err("Could not register a new generated LFN: " + lfn.lfname());
        this->L_transactionDone();
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::insertPFN",s);
      }

      // now register the pfn
      insertPFN(pentry, 1, 2);
      // potentialy there are other PFNs with this guid still in the delayed list
      // but they must be added by repeated calls to this method from higher level routines
      return;
    }
  }
  

  void
  LFCCatalog::insertLFN( LFNEntry& lentry_unqual) const 
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::insertLFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::insertLFN",
				    "A transaction has not been started." );
    }

    if (lentry_unqual.lfname().empty()) {
      std::string s = "No LFN name specified";
      fclog<<seal::Msg::Warning<< s << seal::endmsg;
      throw FCbackendException("LFCCatalog::insertLFN",s);
    }

    // if the fid is empty we generate a new one
    bool using_new_fid = false;
    if (lentry_unqual.guid().empty()) {
      generateLFN(lentry_unqual);
      using_new_fid = true;
    }

    LFNEntry lentry;
    {
      std::string lfn;
      this->L_qualify(lentry_unqual.lfname(), lfn);
      LFNEntry l(lfn, lentry_unqual.guid());
      lentry = l;
    }

    // check to see if the LFN (name) already exists
    {
      struct lfc_filestat stat; 
      if(lfc_lstat(lentry.lfname().c_str(), &stat) < 0 ) {
        if (serrno != ENOENT) {
          std::string s = L_build_serr("Could not lstat LFN: " + lentry.lfname());
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::insertLFN",s);
        }
      } else {
        //fclog<<seal::Msg::Warning<< "Duplicate LFN: " << lentry.lfname() <<seal::endmsg;
        throw FCduplicateLFNException("LFCCatalog::insertLFN");
      }
    }

    FileCatalog::FileID fid = lentry.guid();

    if (!using_new_fid) {
      delayed_t::pfns_t delayed = m_delayed_pfn.find(fid,"");
      if (delayed.size()>0) {

        // the guid of the LFN matches the guid of one or more delayed PFN registrations
        // by definitaion of the delayed PFN we know the PFNs and GUID don't exist
        // and have already checked the LFN name doesn't exist

        if(lfc_create_alias(fid.c_str(), lentry.lfname().c_str(), 0, m_errbuf, LFC_ERRBUFSZ) < 0 ) {
          std::string s = L_build_err("Could not add LFN: " + lentry.lfname());
          this->L_transactionDone();
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::insertLFN",s);
        }

        for (delayed_t::pfns_t::const_iterator it = delayed.begin(); it != delayed.end(); ++it) {
          this->insertPFN(*it, 1, 2);
        }
        return;
      }
    }
    // GUID of LFN is not in the delayed PFN list

    // see if there is a generated LFN matching with our guid
    if (!using_new_fid)
    {
      // try to rename an existing single 'generated' LFN for this GUID
      LFNEntry gen_lfn("",fid);
      generateLFN(gen_lfn);

      // for a little extra efficienty use stat not statg: so cann't cross check the guid
      struct lfc_filestat statbuf;
      if (lfc_stat(gen_lfn.lfname().c_str(), &statbuf)<0) {
        if (serrno != ENOENT) {
          std::string s = L_build_serr("Could not gstat LFN: " + gen_lfn.lfname());
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::insertLFN",s);
        }
      } else {

        if (statbuf.filemode & S_IFREG) {
          if (gen_lfn.lfname() == lentry.lfname()) return;

          if (lfc_rename(gen_lfn.lfname().c_str(), lentry.lfname().c_str()) < 0) {
            std::string s = L_build_serr("Could not rename: " + gen_lfn.lfname() + " to " + lentry.lfname());
            this->L_transactionDone();
            fclog<<seal::Msg::Warning<< s <<seal::endmsg;
            throw FCbackendException("LFCCatalog::insertLFN",s);
          } else {
            // OK
            return;
          }
        }
      }
    }

    // first try to register alias (add a sym link). It will
    // fail with ENOENT if the guid doesn't exist (but not abort trans)
    if(lfc_register_alias(fid.c_str(), lentry.lfname().c_str(), m_errbuf, LFC_ERRBUFSZ) >= 0 ) {
      return;
    } else if (errno == ENOENT) {
      if(lfc_create_alias(fid.c_str(), lentry.lfname().c_str(), 0, m_errbuf, LFC_ERRBUFSZ) >=0 ) {
        return;
      }
    }

    std::string s = L_build_err("Could not add LFN: " + lentry.lfname());
    this->L_transactionDone();
    fclog<<seal::Msg::Warning<< s <<seal::endmsg;
    throw FCbackendException("LFCCatalog::insertLFN",s);
  }
  
  
  void
  LFCCatalog::deletePFN( const std::string& pf ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::deletePFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::deletePFN",
				    "A transaction has not been started." );
    }

    if (m_delayed_pfn.has_pfname(pf)) {
      m_delayed_pfn.remove_pfname(pf);
      return;
    }
    // pf is not in delayed PFN list

    std::string filetype;
    FileCatalog::FileID guid;
    this->lookupFileByPFN(pf, guid, filetype);

    if (guid.empty()) {
      //fclog<<seal::Msg::Warning<< "No such PFN: " << pf <<seal::endmsg;
      throw FCnonexistentFileException( "LFCCatalog::deletePFN", pf );
    }

    // and now unregister the pfn
    if(lfc_unregister_pfn(guid.c_str(), pf.c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
      std::string s = L_build_err("Could not unregister PFN: " + pf);
      this->L_transactionDone();
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::deletePFN",s);
    }

    // and now check if there are any more replicas - if not maybe delete the
    // guid and lfns
    int replicas;
    if((replicas = lfc_replica_exists(guid.c_str(), m_errbuf, LFC_ERRBUFSZ)) < 0) {
      std::string s = L_build_err("Could not check for existing replicas");
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::deletePFN",s);
    }
    // return if we have replicas or 'allow dangling lfns'
    if(replicas || m_allow_dangling_lfns) {
      return;
    }

    // no replicas, list all the LFNs
    // don't ignore ENOENT - (should be at least 1 LFN - better to compain if something unexpected is happening)
    std::vector<LFNEntry> lfn_vec;
    fid_to_entry(guid, lfn_vec, false);

    // remove all the LFNs
    std::vector<LFNEntry>::const_iterator lfn_vec_itr;

    for(lfn_vec_itr = lfn_vec.begin(); lfn_vec_itr != lfn_vec.end(); ++lfn_vec_itr) {
      if(lfc_unregister_alias(lfn_vec_itr->guid().c_str(), lfn_vec_itr->lfname().c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
        std::string s = L_build_err("Could not unregister an LFN: " + lfn_vec_itr->lfname());
        this->L_transactionDone();
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::deletePFN",s);
      }
    }
  }

  /// If there is only 1, primary, LFN: if there are no replicas, just
  /// delete the LFN. Otherwise rename the LFN to a generated one
  /// (this is to preserve the symantics of allowing a GUID-PFNs to exist
  ///  with no LFNs)
  ///
  ///     If there is more than 1 LFN:
  ///       If it is the primary LFN and we don't want danling LFNs:
  ///         If there are replicas:
  ///           rename to a symlink and change all the other smylinks to
  ///           point to the new primary
  ///         else
  ///           delete all the LFNS
  ///       else if it is the primary LFN and we don't mind dangling LFNs:
  ///        If there are replicas
  ///          rename it to a new LFN
  ///        else
  ///          just delete the LFN
  ///       else It is an alias:
  ///         if there are replicas:
  ///           just delete the LFN alias
  ///         else
  ///           if we don't mind dangling LFNs:
  ///             just delete the LFN alias
  ///           else
  ///             delete all the LFNs
  void
  LFCCatalog::deleteLFN( const std::string& lf_unqual ) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::deleteLFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::deleteLFN",
				    "A transaction has not been started." );
    }

    std::string lf;
    this->L_qualify(lf_unqual, lf);

    std::vector<LFNEntry> lfn_vec;
    std::vector<PFNEntry> pfn_vec;

    FileCatalog::FileID fid;

    translate_entry(lf, fid, pfn_vec, true);

    if (fid.empty()) {
      // the LFN didn't exist or was a dangling symlink.

      struct lfc_filestat stat; 
  
      if(lfc_lstat(lf.c_str(), &stat) < 0 ) {
        if (serrno == ENOENT) {
          //fclog<<seal::Msg::Warning<< "No such LFN: " << lf <<seal::endmsg;
          throw FCnonexistentFileException( "LFCCatalog::deleteLFN", lf );
        }
        std::string s = L_build_serr("Could not lstat LFN: " + lf);
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::deleteLFN",s);
      }

      if (!(stat.filemode & S_IFLNK)) {
        std::string s = std::string("Not an LFN or a dangling symlink: ") + lf;
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::deleteLFN",s);
      }

      if(lfc_unregister_alias(NULL, lf.c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
        std::string s = L_build_err("Error removing a dangling symlink: " + lf);
        this->L_transactionDone();
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::deleteLFN",s);
      }
      return;
    }

    // get all the LFNs for our GUID (should be at least one), so don't ignore ENOENT
    fid_to_entry(fid, lfn_vec, false);

    if (lfn_vec.empty()) {
      // This doesn't make sense, if there was a fid there must be at least 1 LFN
      std::string s = std::string("Internal error: found no LFNs for GUID : ") + fid;
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException( "LFCCatalog::deleteLFN", s);
    }

    // see if there are any replicas registered
    int replicas;
    if((replicas = lfc_replica_exists(fid.c_str(), m_errbuf, LFC_ERRBUFSZ)) < 0) {
      fclog<<seal::Msg::Warning<< "Could not check for existing replicas" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::deleteLFN",
			       "Could not check for existing replicas");
    }

    if (lfn_vec.size()==1) {
      if (!replicas) {
        // just delete the LFN
        if(lfc_unregister_alias(fid.c_str(), lf.c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
          std::string s = L_build_err("Error removing a single LFN with no replicas: " + lf);
          this->L_transactionDone();
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::deleteLFN",s);
        }
      } else {
        // rename the LFN to a new generated name
        LFNEntry new_lfn("",fid);
        generateLFN(new_lfn);
        if (lf == new_lfn.lfname()) return;

        // see if the new target LFN already exists
        FileCatalog::FileID fd;
        this->lookupFileByLFN(new_lfn.lfname(), fd);
        if (!fd.empty()) {
          //fclog<<seal::Msg::Warning<< "Duplicate LFN: " << new_lfn.lfname() <<seal::endmsg;
          throw FCduplicateLFNException("LFCCatalog::deleteLFN");
        }

        if (lfc_rename(lf.c_str(), new_lfn.lfname().c_str()) < 0) {
          std::string s =  L_build_serr("Could not rename single LFN (with replicas) to generated form: " +
                                            lf + " to " + new_lfn.lfname());
          this->L_transactionDone();
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::deleteLFN",s);
        }
      }
      return;
    }

    // we have more than 1 LFN
    // find if we are the primary LFN or not

    LFNEntry entry( lf, fid);
    bool found = false;
    size_t i=0;

    while(i<lfn_vec.size()) {
      if (compare(lfn_vec[i], entry)) {
        found = true;
        break;
      }
      ++i;
    }

    if (!found) {
      fclog<<seal::Msg::Warning<< "Interal error: did not find LFN in explicit list of LFNs" <<seal::endmsg;
      throw FCbackendException( "LFCCatalog::deleteLFN",
                                "Interal error: did not find LFN in explicit list of LFNs");
    }

    if (i > 0) {
      if (replicas || m_allow_dangling_lfns) {
        // it is an alias, with PFNs or we don't care about dangling LFNs - just delete it
        if(lfc_unregister_alias(entry.guid().c_str(), entry.lfname().c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
          std::string s =  L_build_err("Error removing an alias: " + entry.lfname());
          this->L_transactionDone();
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::deleteLFN",s);
        }
      } else {
        // delete all the LFN entries
        deleteEntry(entry.guid());
      }
    } else {
      // it's the primary LFN
      if (!m_allow_dangling_lfns) {
        if (replicas) {
          // rename this primary LFN to one of the aliases, and then
          // make all the other aliases point to the new primary lfn

          // first, try to delete all the aliases
          size_t j=1;
          while(j<lfn_vec.size()) {
            LFNEntry &e = lfn_vec[j];
            if(lfc_unregister_alias(e.guid().c_str(), e.lfname().c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
              std::string s =  L_build_err("Error while removing all aliases for LFN: " + e.lfname());
              this->L_transactionDone();
              fclog<<seal::Msg::Warning<< s <<seal::endmsg;
              throw FCbackendException("LFCCatalog::deleteLFN",s);
            }
            ++j;
          }
  
          LFNEntry &new_prim = lfn_vec[1];
          // try to rename the primary LFN to the name of one of the old aliases
          if (lfc_rename(entry.lfname().c_str(), new_prim.lfname().c_str()) < 0) {
            std::string s =  L_build_serr("Could not rename primary LFN into place of one of the aliases: " +
                                             entry.lfname() + " to " + new_prim.lfname());
            this->L_transactionDone();
            fclog<<seal::Msg::Warning<< s <<seal::endmsg;
            throw FCbackendException("LFCCatalog::deleteLFN",s);
          }

          // now recreate the other old aliases (if any), pointing to the new primary lfn
          j = 2;
          while(j<lfn_vec.size()) {
            LFNEntry &e = lfn_vec[j];
            if (lfc_symlink(new_prim.lfname().c_str(), e.lfname().c_str()) < 0) {
              std::string s =  L_build_serr("Could not recreate symlinks to point to the new primary LFN: " +
                                               new_prim.lfname() + " to " + e.lfname());
              this->L_transactionDone();
              fclog<<seal::Msg::Warning<< s <<seal::endmsg;
              throw FCbackendException("LFCCatalog::deleteLFN",s);
            }
            ++j;
          }
        } else {
          // delete all the LFNs
          deleteEntry(entry.guid());
        }
      } else {
        if (replicas) {
          // rename the LFN to a new generated name
          LFNEntry new_lfn("",entry.guid());
          generateLFN(new_lfn);

          // see if the new target LFN already exists
          FileCatalog::FileID fd;
          this->lookupFileByLFN(new_lfn.lfname(), fd);
          if (!fd.empty()) {
            //fclog<<seal::Msg::Warning<< "Duplicate LFN: " << new_lfn.lfname() <<seal::endmsg;
            throw FCduplicateLFNException("LFCCatalog::deleteLFN");
          }

          if (lfc_rename(entry.lfname().c_str(), new_lfn.lfname().c_str()) < 0) {
            std::string s =  L_build_serr("Could not rename primary LFN into generated form, leaving dangling aliases: " +
                                             entry.lfname() + " to " + new_lfn.lfname());
            this->L_transactionDone();
            fclog<<seal::Msg::Warning<< s <<seal::endmsg;
            throw FCbackendException("LFCCatalog::deleteLFN",s);
          }
        } else {
          // just delete the LFN
          if(lfc_unregister_alias(entry.guid().c_str(), entry.lfname().c_str(), m_errbuf, LFC_ERRBUFSZ) < 0) {
            std::string s =  L_build_err("Error removing the primary LFN (" + entry.lfname() + "), leaving dangling aliases");
            this->L_transactionDone();
            fclog<<seal::Msg::Warning<< s <<seal::endmsg;
            throw FCbackendException("LFCCatalog::deleteLFN",s);
          }
        }
      }
    }
  }
  
  /// Delete all pfns and lfns for this guid
  void
  LFCCatalog::deleteEntry(const FileCatalog::FileID& guid) const
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::deleteEntry", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::deleteEntry",
				    "A transaction has not been started." );
    }

    if (m_delayed_pfn.has_guid(guid)) {
      m_delayed_pfn.remove_guid(guid);
      return;
    }
    // guid not in list of delayed PFNs

    bool fail=false;
    std::string fail_item;
    
    // first get the list of PFNs
    char **pfns;
    pfns = lfc_surlsfromguid(guid.c_str(), m_errbuf, LFC_ERRBUFSZ);
    if(pfns != NULL) {
      int i = 0;
      while(pfns[i]) {
	if (!fail) {
          if (lfc_unregister_pfn(guid.c_str(), pfns[i], m_errbuf, LFC_ERRBUFSZ)<0) {
            fail = true;
            fail_item = pfns[i];
          }
        }
	free(pfns[i++]);
      }
      free(pfns);
    } else if (errno == ENOENT) {
      // cann't throw here as it breaks FCdeleteEntry with query on files with more than 1 replica
      // no such guid, silently return
      return;
    }

    if (fail) {
      std::string s = L_build_err("Could not remove all the PFNs: Failed on " + fail_item);
      this->L_transactionDone();
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::deleteEntry",s);
    }

    // and now the LFNs
    char **lfns;
    if((lfns = lfc_lfnsforguid(guid.c_str(), m_errbuf, LFC_ERRBUFSZ)) == NULL) {
      std::string s = L_build_err("Could not get LFNs for guid: " + guid);
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::deleteEntry",s);
    }

    int i = 0;
    while(lfns[i]) {
      if (!fail) {
        if (lfc_unregister_alias( guid.c_str(), lfns[i], m_errbuf, LFC_ERRBUFSZ)<0) {
          fail = true;
          fail_item = lfns[i];
        }
      }
      free(lfns[i++]);
    }
    free(lfns);

    if (fail) {
      std::string s = L_build_err("Could not remove all the LFNs: Failed on " + fail_item);
      this->L_transactionDone();
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::deleteEntry",s);
    }
  }

  
  bool
  LFCCatalog::isReadOnly() const
  {
    lfc_enter_routine routine(this);

    return false;
  }


  bool
  LFCCatalog::retrievePFN( const std::string& in_query,
			   FCBuf<PFNEntry>& buf,
			   const size_t& start )
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::retrievePFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::retrievePFN",
				    "A transaction has not been started." );
    }

    commitAllDelayedPFN();

    buf.clear();
    std::string query = in_query;

    if(query.empty()) {
      query = "lfname like %";
    }

    if (m_retPFN == NULL) {
      m_retPFN = (void *)(new LFCQueryEvaluator<PFNEntry>::retrieveData_t);
    }

    MRET(PFNEntry,m_retPFN)->fill(this,query,start,buf,
                                   &pool::LFCCatalog::L_get_pfnentry_from_pfn,
                                   &pool::LFCCatalog::L_get_pfnentry_from_guid,
                                   &pool::LFCCatalog::L_get_pfnentry_from_lfn);

    if(buf.empty()) return false;
    return true;
  }
  
  bool
  LFCCatalog::retrieveLFN( const std::string& in_query,
			   FCBuf<LFNEntry>& buf,
			   const size_t& start )
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::retrieveLFN", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::retrieveLFN",
				    "A transaction has not been started." );
    }

    commitAllDelayedPFN();

    buf.clear();
    std::string query = in_query;

    if(query.empty()) {
      query = "lfname like %";
    }

    if (m_retLFN == NULL) m_retLFN = (void *)(new LFCQueryEvaluator<LFNEntry>::retrieveData_t);
    MRET(LFNEntry,m_retLFN)->fill(this,query,start,buf,
                                   &pool::LFCCatalog::L_get_lfnentry_from_pfn,
                                   &pool::LFCCatalog::L_get_lfnentry_from_guid,
                                   &pool::LFCCatalog::L_get_lfnentry_from_lfn);

    if(buf.empty()) return false;
    return true;
  }
  

  bool
  LFCCatalog::retrieveGuid( const std::string& in_query,
			    FCBuf<FileCatalog::FileID>& buf,
			    const size_t& start )
  {
    lfc_enter_routine routine(this);

    PoolMessageStream fclog( "LFCCatalog::retrieveGuid", seal::Msg::Nil);

    if ( ! ( m_transactionsta.test( 0 ) && m_transactionsta.test( 1 ) ) ) {
      fclog<<seal::Msg::Warning<< "A transaction has not been started." <<seal::endmsg;
      throw FCTransactionException( "LFCCatalog::retrieveGuid",
					  "A transaction has not been started." );
    }

    commitAllDelayedPFN();

    buf.clear();
    std::string query = in_query;

    if(query.empty()) {
      query = "lfname like %";
    }

    if (m_retGUID == NULL) m_retGUID = (void *)(new LFCQueryEvaluator<FileCatalog::FileID>::retrieveData_t);
    MRET(FileCatalog::FileID,m_retGUID)->fill(this,query,start,buf,
                                   &pool::LFCCatalog::L_get_fid_from_pfn,
                                   &pool::LFCCatalog::L_get_fid_from_guid,
                                   &pool::LFCCatalog::L_get_fid_from_lfn);

    if(buf.empty()) return false;
    return true;
  }

  /* STATIC member function */
  bool LFCCatalog::compare(const FileCatalog::FileID &a, const FileCatalog::FileID &b) {
    if (a == b) return true;
    return false;
  }

  /* STATIC member function */
  bool LFCCatalog::compare(const LFNEntry &a, const LFNEntry &b) {
    if (a.guid() == b.guid() &&
        a.lfname() == b.lfname()) return true;
    return false;
  }

  /* STATIC member function */
  bool LFCCatalog::compare(const PFNEntry &a, const PFNEntry &b) {
    // don't require equality of filetype
    if (a.guid() == b.guid() &&
        a.pfname() == b.pfname()) return true;
    return false;
  }

  /* STATIC member function */
  bool LFCCatalog::compare(const udata &u1, const udata &u2) {
    return (u1 == u2);
  }

  void LFCCatalog::fid_to_entry(const FileCatalog::FileID &fid, std::vector<PFNEntry> &pfn_vec, const bool ignore_enoent) const
  {
    PoolMessageStream fclog( "LFCCatalog::fid_to_entry", seal::Msg::Nil);

    char **pfns;

    pfn_vec.clear();

    if((pfns = lfc_surlsfromguid(fid.c_str(), m_errbuf, LFC_ERRBUFSZ)) == NULL) {
      if (!ignore_enoent || errno != ENOENT) {
        std::string s = L_build_err(std::string("Could not lookup PFNs for guid : ") + fid);
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::fid_to_entry",s);
      }
    } else {
      size_t i=0;
      while(pfns[i]) {
        std::string filetype;
        PFNEntry entry(pfns[i],fid,filetype);
        pfn_vec.push_back(entry);
        free(pfns[i++]);
      }
      free(pfns);
      if (!ignore_enoent && i==0) {
        errno = ENOENT;
        std::string s = L_build_err(std::string("Could not find any PFNs for guid : ") + fid);
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::fid_to_entry",s);
      }
    }
  }

  void LFCCatalog::fid_to_entry(const FileCatalog::FileID &fid, std::vector<LFNEntry> &lfn_vec, const bool ignore_enoent) const
  {
    PoolMessageStream fclog( "LFCCatalog::fid_to_entry", seal::Msg::Nil);

    char **lfns;

    lfn_vec.clear();

    if((lfns = lfc_lfnsforguid(fid.c_str(), m_errbuf, LFC_ERRBUFSZ)) == NULL) {
      if (!ignore_enoent || errno != ENOENT) {
        std::string s = L_build_err(std::string("Could not get LFNs for guid : ") + fid);
        fclog<<seal::Msg::Warning<< s <<seal::endmsg;
        throw FCbackendException("LFCCatalog::fid_to_entry",s);
      }
    } else {
      size_t i=0;
      while(lfns[i]) {
        LFNEntry entry(lfns[i],fid);
        lfn_vec.push_back(entry);
        free(lfns[i++]);
      }
      free(lfns);
    }
  }

  void LFCCatalog::translate_entry(const std::string value, FileCatalog::FileID &fid, std::vector<PFNEntry> &pfn_vec, const bool ignore_enoent) const
  {
    PoolMessageStream fclog( "LFCCatalog::translate_entry", seal::Msg::Nil);

    lookupFileByLFN( value.c_str(), fid );

    if (!ignore_enoent && fid.empty()) {
      std::string s = std::string("Could not get GUID for LFN : ") + value;
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::translate_entry",s);
    }

    if (fid.empty()) return;

    // may be no PFNs for this LFN, so ignore ENOENT
    fid_to_entry(fid, pfn_vec, true);
  }

  void LFCCatalog::translate_entry(const std::string value, FileCatalog::FileID &fid, std::vector<LFNEntry> &lfn_vec, const bool ignore_enoent) const
  {
    PoolMessageStream fclog( "LFCCatalog::translate_entry", seal::Msg::Nil);

    std::string filetype;
    lookupFileByPFN( value.c_str(), fid, filetype );

    if (!ignore_enoent && fid.empty()) {
      std::string s = std::string("Could not get GUID for PFN : ") + value;
      fclog<<seal::Msg::Warning<< s <<seal::endmsg;
      throw FCbackendException("LFCCatalog::translate_entry",s);
    }

    if (fid.empty()) return;

    // there should always be at least 1 LFN for this guid, so don't ignore ENOENT
    fid_to_entry(fid, lfn_vec, false);
  }

  /* STATIC member function */
  bool LFCCatalog::has_wildcard(const std::string &s) {
    bool found = false;
    for(size_t i=0;i<s.size() && !found;++i) {
      if (s[i] == '_' || s[i] == '%') found=true;
    }
    return found;
  }

  /* STATIC member function */
  const char *LFCCatalog::wild_underscore_strstr(const char *haystack, const char *needle) {
    const char *p;
    bool found = false;

    p = haystack;
    size_t len = strlen(p);
    size_t needle_len = strlen(needle);
    while(p != NULL && *p != '\0') {
      if (len < needle_len) break;
      bool m=true;
      for(size_t i=0;m && i<needle_len;++i) {
        if (needle[i] == '_') continue;
        if (needle[i] != p[i]) m = false;
      }
      if (m) {
        found=true;
        break;
      }
      ++p; --len;
    }
    if (!found) return NULL;

    return p;
  }

  /* STATIC member function */
  bool LFCCatalog::string_wildmatch(const std::string &s, const std::string &match) {
    size_t i,j;

    for(i=0,j=0;j<match.size();) {

      // count the minimum number chars remaining to be matched: mr
      size_t k=j,mr=0;
      while(match[k] != '\0') {
        if (match[k] != '%') ++mr;
        ++k;
      }

      // number of string chars left: sr
      size_t sr = s.size() - i;

      // if the minimum number of chars to match is larger than the number remaining, fail
      if (mr>sr) return false;

      const char *p, *p1,*p2;
      p1 = match.c_str();
      p1 += j+1;

      // find the next %
      p2 = strstr(p1, "%");

      std::string next_fixed_block;
      if (p2 != NULL) {
        next_fixed_block = match.substr(j+1,p2-p1);
      } else {
        next_fixed_block = match.substr(j+1,strlen(p1));
      }

      // now:  p1 is match string AFTER current char
      //       p2 is next % after or at p1
      //       next_fixed_block is largest fixed sized 'match' segment after next match char

      if (match[j] == '%') {

        // replace the '%' with a number of underscores and iterate

        std::vector<size_t> n_underscores;

        if (p1 == p2) {

          // i.e. next char is %

          n_underscores.push_back(0);
        } else if (next_fixed_block.empty()) {
          // i.e. no more match, this is a trailing %
          // so relace with as many _ as there are characters remaining in the string

          n_underscores.push_back(sr);
        } else {

          // some constant sized block after the %

          p = next_fixed_block.c_str();
          p1 = s.c_str();
          p1 += i;

          p2 = wild_underscore_strstr(p1, p);

          // there is no occurance of the next fixed part of the match, so we cann't match
          if (p2 == NULL) return false;

          n_underscores.push_back(p2 - p1);

          const char *p3 = p2;
          while(1) {
            const char *pp;
            pp = wild_underscore_strstr(p3+1, p);
            if (pp == NULL) break;
            p3 = pp;
            n_underscores.push_back(p3 - p1);
          }
        }

        std::vector<size_t>::reverse_iterator rev_itr;
        for(rev_itr = n_underscores.rbegin(); rev_itr != n_underscores.rend(); ++rev_itr) {

          size_t k = 1 + *rev_itr; 
          std::string r;
          for(size_t l=1;l<k;++l) r+='_';
          std::string m = match;
          m.replace(0,j+1,r);
          r = s;
          r.erase(0,i);
          if (string_wildmatch(r,m)) return true;
        }
        return false;
      }

      // this match char isn't a variable length one, so add it to next_fixed_block
      next_fixed_block = match[j] + next_fixed_block;

      p = next_fixed_block.c_str();
      p1 = s.c_str();
      p1 += i;

      p2 = wild_underscore_strstr(p1, p);

      if (p2 != p1) return false;

      j += next_fixed_block.size();
      i += next_fixed_block.size();
    }
    if (i!=s.size()) return false; // did not use up all the string

    return true;
  }

  /* STATIC member function */
  void LFCCatalog::breakup_path(const std::string path, std::vector<std::string> &components) {
    PoolMessageStream fclog( "LFCCatalog::breakup_path", seal::Msg::Nil);

    const char *p = path.c_str();
    size_t i=0,j;

    while(p[i]) {
      if (p[i] != '/') {
        fclog<<seal::Msg::Warning<< "invalid path" <<seal::endmsg;
        throw FCbackendException("LFCCatalog::breakup_path","invalid path");
      }
      j=i+1;
      while(p[j] && p[j] != '/') ++j;
      if (j-i > 1) {
        char *c = (char *)malloc(j-i);
        strncpy(c,&p[i+1],j-i-1);
        c[j-i-1] = '\0';
        std::string n(c);
        free(c);
        components.push_back(n);
      } else {
        if (!p[j]) {
          fclog<<seal::Msg::Warning<< "last path element a directory" <<seal::endmsg;
          throw FCbackendException("LFCCatalog::breakup_path","last path element a directory");
        }
      }
      i = j;
    }
  }

  /* STATIC member function */
  bool LFCCatalog::lfn_wildmatch(const LFNEntry &lfn, const std::string &match) {
    std::vector<std::string> lfn_elements, match_elements;

    breakup_path(lfn.lfname(), lfn_elements);
    breakup_path(match, match_elements);

    if (lfn_elements.size() == 0 || (lfn_elements.size() != match_elements.size())) return false;

    size_t i;
    for(i=0;i<lfn_elements.size();++i) {
      if (!string_wildmatch(lfn_elements[i],match_elements[i])) return false;
    }

    return true;
  }

  /* STATIC member function */
  bool LFCCatalog::lfn_wildmatch(const LFNEntry &lfn, const std::vector<std::string> &match_elements) {
    std::vector<std::string> lfn_elements;

    breakup_path(lfn.lfname(), lfn_elements);

    if (lfn_elements.size() == 0 || (lfn_elements.size() != match_elements.size())) return false;

    size_t i;
    for(i=0;i<lfn_elements.size();++i) {
      if (!string_wildmatch(lfn_elements[i],match_elements[i])) return false;
    }

    return true;
  }

  /* STATIC member function */
  bool LFCCatalog::pfn_wildmatch(const PFNEntry &pfn, const std::string &match) {

    return string_wildmatch(pfn.pfname(), match);
  }

  /* STATIC member function */
  bool LFCCatalog::guid_wildmatch(const FileCatalog::FileID &fid, const std::string &match) {

    return string_wildmatch(fid, match);
  }

  void LFCCatalog::L_get_fid_from_pfn(LFC_LIST_FUNC_TYPE(FileCatalog::FileID))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_fid_from_pfn", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      std::string filetype;
      constraint_checker c;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.pfns_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
            set_distinct = MAYBE; // output will be distinct if the universal set was
        } else {
            // not supported
        }
        priv_data = LFC_NEED_SET;
      } else {
        set_distinct = YES;
        priv_data = LFC_NO_NEED_SET;
      }
      return;
    }

    /* cannot do wildcard searches for PFNs without a given universal set,
       so in this case throw */
    if (universal_set == NULL && match_type == WILDCARD) {
      fclog<<seal::Msg::Warning<< "Cannot make wildcard list" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::L_get_fid_from_pfn","Cannot make wildcard list");
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL) {
        lookupFileByPFN( value, pd->fid, pd->filetype );
      } else {
        pd->c.pfns_add(match_type, value);
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      if (match_type == LITERAL) {
        if (pd->off>0) return;
        for(size_t i=0;i<universal_set->size();++i) {
          if (compare( (*universal_set)[i], pd->fid ) && pd->c.check((*universal_set)[i])) {
            buf.push_back((*universal_set)[i]);
            set_distinct = YES; // only 1 element
            break;
          }
        }
        ++pd->off;
      } else { /* wildcard match */
        while(!buf.full() && pd->off < universal_set->size()) {
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // output will be distinct if the universal set was
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      if (pd->off>0) return;
      if (pd->fid.empty()) return;
      if (pd->c.check(pd->fid)) {
        buf.push_back(pd->fid);
        set_distinct = YES; // only 1 element
      }
      ++pd->off;
    }

    return;
  }

  void LFCCatalog::L_get_fid_from_guid(LFC_LIST_FUNC_TYPE(FileCatalog::FileID))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_fid_from_guid", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      constraint_checker c;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.fids_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE; // output will be distinct if the universal set was
        } else {
          // not supported
        }
        priv_data = LFC_NEED_SET;
      } else {
        set_distinct = YES;
        priv_data = LFC_NO_NEED_SET;
      }
      return;
    }

    /* cannot do wildcard searches for GUIDs without a given universal set,
       so in this case throw */
    if (universal_set == NULL && match_type == WILDCARD) {
      fclog<<seal::Msg::Warning<< "Cannot make wildcard list" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::L_get_fid_from_guid","Cannot make wildcard list");
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL && universal_set == NULL) {
        int guid_exists;
        if ((guid_exists=lfc_guid_exists(value.c_str(), m_errbuf, LFC_ERRBUFSZ)) < 0) {
          std::string s = L_build_err(std::string("Could not check for existance of guid: ") + value);
          fclog<<seal::Msg::Warning<< s <<seal::endmsg;
          throw FCbackendException("LFCCatalog::L_get_fid_from_guid",s);
        }
        if (guid_exists) {
          pd->fid = value;
        }
      } else if (match_type == WILDCARD) {
        pd->c.fids_add(match_type, value);
      }
    }
        
    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      if (match_type == LITERAL) {
        if (pd->off>0) return;
        for(size_t i=0;i<universal_set->size();++i) {
          if (compare( (*universal_set)[i], value ) && pd->c.check((*universal_set)[i])) {
            buf.push_back((*universal_set)[i]);
            set_distinct = YES; // only 1 element
            break;
          }
        }
        ++pd->off;
      } else { /* wildcard match */
        while(!buf.full() && pd->off < universal_set->size()) {
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // output will be distinct if the universal set was
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      if (pd->off>0) return;
      if (pd->fid.empty()) return;
      if (pd->c.check(pd->fid)) {
        buf.push_back(pd->fid);
        set_distinct = YES; // only 1 element
      }
      ++pd->off;
    }

    return;
  }

  void LFCCatalog::L_get_fid_from_lfn(LFC_LIST_FUNC_TYPE(FileCatalog::FileID))
  {
    struct priv_data_t {
      path_list_t path_lister;
      FileCatalog::FileID fid;
      constraint_checker c;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.lfns_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          set_distinct = NO;
        }
      } else {
        set_distinct = YES;
      }
      priv_data = LFC_NO_NEED_SET; // means that we don't require a set
      return;
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL) {
        lookupFileByLFN( value, pd->fid );
      } else {
        /* setup the search over the LFNs */
        if (universal_set == NULL) {
          pd->path_lister.init(this,value);
        } else {
          pd->c.lfns_add(match_type, value);
        }
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      if (match_type == LITERAL) {
        if (pd->off>0) return;
        for(size_t i=0;i<universal_set->size();++i) {
          if (compare( (*universal_set)[i], pd->fid ) && pd->c.check( (*universal_set)[i])) {
            buf.push_back((*universal_set)[i]);
            set_distinct = YES; // only 1 element
            break;
          }
        }
        ++pd->off;
      } else { /* wildcard match */
        while(!buf.full() && pd->off < universal_set->size()) {
          if (pd->c.check( (*universal_set)[pd->off] )) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // output will be distinct if the universal set was
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      if (match_type == LITERAL) {
        if (pd->off > 0) return;
        if (pd->fid.empty()) return;
        if (pd->c.check(pd->fid)) {
          buf.push_back(pd->fid);
          set_distinct = YES; // only 1 element
        }
        ++pd->off;
      } else { /* wildcard match, no set specified so expand path */
        while(!buf.full()) {
          udata data = pd->path_lister.next().guid(0);
          if (data.guid().empty()) return;
          if (pd->c.check(data)) {
            buf.push_back(data);
            set_distinct = NO; // may scan LFN aliases, giving identical guids
          }
        }
      }
    }

    return;
  }

  void LFCCatalog::L_get_lfnentry_from_pfn(LFC_LIST_FUNC_TYPE(LFNEntry))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_lfnentry_from_pfn", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      constraint_checker c;
      std::vector<LFNEntry> lfns;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.pfns_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          // not allowed
        }
        priv_data = LFC_NEED_SET; // means that we need a set
      } else {
        set_distinct = YES;
        priv_data = LFC_NO_NEED_SET; // means that we don't require a set
      }
      return;
    }

    /* cannot do wildcard searches for PFNs without a given universal set,
       so in this case throw */
    if (universal_set == NULL && match_type == WILDCARD) {
      fclog<<seal::Msg::Warning<< "Cannot make wildcard list" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::L_get_lfnentry_from_pfn","Cannot make wildcard list");
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL) {

        // lookup FID and LFNs for this PFN. Ignore ENOENT
        translate_entry(value, pd->fid, pd->lfns, true);

      } else {
        pd->c.pfns_add(match_type, value);
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      while(!buf.full() && pd->off < universal_set->size()) {

        if (match_type == LITERAL) {
          std::vector<LFNEntry>::iterator lfns_itr;
          for(lfns_itr=pd->lfns.begin();lfns_itr!=pd->lfns.end();++lfns_itr) {
            if (compare( (*universal_set)[pd->off], *lfns_itr ) &&
                pd->c.check( (*universal_set)[pd->off] )) {
              buf.push_back((*universal_set)[pd->off]);
              set_distinct = YES; // 'lfns' list must be distinct, therefore output is too
              break;
            }
          }
        } else { /* wildcard match */
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // set distinct only if the universal set was
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      while(!buf.full() && pd->off < pd->lfns.size()) {
        if (pd->c.check(pd->lfns[pd->off])) {
          buf.push_back(pd->lfns[pd->off]);
          set_distinct = YES; // 'lfns' list must be distinct
        }
        ++pd->off;
      }
    }

    return;
  }

  void LFCCatalog::L_get_lfnentry_from_guid(LFC_LIST_FUNC_TYPE(LFNEntry))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_lfnentry_from_guid", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      constraint_checker c;
      std::vector<LFNEntry> lfns;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.fids_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          // not allowed
        }
        priv_data = LFC_NEED_SET;
      } else {
        set_distinct = YES;
        priv_data = LFC_NO_NEED_SET;
      }
      return;
    }

    /* cannot do wildcard searches for GUIDs without a given universal set,
       so in this case throw */
    if (universal_set == NULL && match_type == WILDCARD) {
      fclog<<seal::Msg::Warning<< "Cannot make wildcard list" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::L_get_lfnentry_from_guid","Cannot make wildcard list");
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL) {

        pd->fid = value;

        // find the LFNs for the GUID, ignore ENOENT
        fid_to_entry(pd->fid, pd->lfns, true);

      } else {
        pd->c.fids_add(match_type, value);
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      while(!buf.full() && pd->off < universal_set->size()) {

        if (match_type == LITERAL) {
          std::vector<LFNEntry>::iterator lfns_itr;
          for(lfns_itr=pd->lfns.begin();lfns_itr!=pd->lfns.end();++lfns_itr) {
            if (compare( (*universal_set)[pd->off], *lfns_itr ) &&
                pd->c.check( (*universal_set)[pd->off] )) {
              buf.push_back((*universal_set)[pd->off]);
              set_distinct = YES; // 'lfns' list must be distinct, therefore output is too
              break;
            }
          }
        } else { /* wildcard match */
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // output only distinct if universal set is too
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      while(!buf.full() && pd->off < pd->lfns.size()) {
        if (pd->c.check(pd->lfns[pd->off])) {
          buf.push_back(pd->lfns[pd->off]);
          set_distinct = YES; // 'lfns' list must be distinct
        }
        ++pd->off;
      }
    }

    return;
  }

  void LFCCatalog::L_get_lfnentry_from_lfn(LFC_LIST_FUNC_TYPE(LFNEntry))
  {
    struct priv_data_t {
      FileCatalog::FileID fid;
      path_list_t path_lister;
      constraint_checker c;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.lfns_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          set_distinct = YES;
        }
      } else {
        set_distinct = YES;
      }
      priv_data = LFC_NO_NEED_SET;
      return;
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL && universal_set == NULL) {
        lookupFileByLFN( value, pd->fid );
      } else if (match_type == WILDCARD) {
        if (universal_set == NULL) {
          pd->path_lister.init(this,value);
        } else {
          pd->c.lfns_add(match_type, value);
        }
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      if (match_type == LITERAL) {
        if (pd->off>0) return;
        for(size_t i=0;i<universal_set->size();++i) {
          LFNEntry lfn(value, (*universal_set)[i].guid());
          if (compare( (*universal_set)[i], lfn ) && pd->c.check( (*universal_set)[i] )) {
            buf.push_back((*universal_set)[i]);
            set_distinct = YES; // only 1 element
            break;
          }
        }
        ++pd->off;
      } else { /* wildcard match */
        while(!buf.full() && pd->off < universal_set->size()) {
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // set distinct only if the universal set was too
          }
          ++pd->off;
        }
      }
    } else { /* no universal set */
      if (match_type == LITERAL) {
        if (pd->off>0) return;
        if (pd->fid.empty()) return;
        LFNEntry lfn(value, pd->fid);
        if (pd->c.check(lfn)) {
          buf.push_back(lfn);
          set_distinct = YES; // only 1 element
        }
        ++pd->off;
      } else { /* wildcard match, no set specified so expand path */
        while(!buf.full()) {
          udata data = pd->path_lister.next();
          if (data.lfn().lfname().empty()) return;
          if (pd->c.check(data)) {
            buf.push_back(data);
            set_distinct = YES; // the LFNEntries from the path_lister are unique
          }
        }
      }
    }

    return;
  }

  void LFCCatalog::L_get_pfnentry_from_pfn(LFC_LIST_FUNC_TYPE(PFNEntry))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_pfnentry_from_pfn", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      std::string filetype;
      constraint_checker c;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.pfns_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          // not allowed
        }
        priv_data = LFC_NEED_SET;
      } else {
        set_distinct = YES;
        priv_data = LFC_NO_NEED_SET;
      }
      return;
    }

    /* cannot do wildcard searches for PFNs without a given universal set,
       so in this case throw */
    if (universal_set == NULL && match_type == WILDCARD) {
      fclog<<seal::Msg::Warning<< "Cannot make wildcard list" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::L_get_pfnentry_from_pfn","Cannot make wildcard list");
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL && universal_set == NULL) {
        lookupFileByPFN( value, pd->fid, pd->filetype );
      } else if (match_type == WILDCARD) {
        pd->c.pfns_add(match_type, value);
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      if (match_type == LITERAL) {
        if (pd->off>0) return;
        for(size_t i=0;i<universal_set->size();++i) {
          std::string ft;
          PFNEntry pfn(value, (*universal_set)[i].guid(), ft);
          if (compare( (*universal_set)[i], pfn ) && pd->c.check( (*universal_set)[i] )) {
            buf.push_back((*universal_set)[i]);
            set_distinct = YES; // only 1 element
            break;
          }
        }
        ++pd->off;
      } else { /* wildcard match */
        while(!buf.full() && pd->off < universal_set->size()) {
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // set distinct only if universal set is too
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      if (pd->off>0) return;
      if (pd->fid.empty()) return;
      PFNEntry pfn(value, pd->fid, pd->filetype);
      if (pd->c.check(pfn)) {
        buf.push_back(pfn);
        set_distinct = YES; // only 1 element
      }
      ++pd->off;
    }

    return;
  }

  void LFCCatalog::L_get_pfnentry_from_guid(LFC_LIST_FUNC_TYPE(PFNEntry))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_pfnentry_from_guid", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      constraint_checker c;
      std::vector<PFNEntry> pfns;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.fids_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          // not allowed
        }
        priv_data = LFC_NEED_SET;
      } else {
        set_distinct = YES;
        priv_data = LFC_NO_NEED_SET;
      }
      return;
    }

    /* cannot do wildcard searches for GUIDs without a given universal set,
       so in this case throw */
    if (universal_set == NULL && match_type == WILDCARD) {
      fclog<<seal::Msg::Warning<< "Cannot make wildcard list" <<seal::endmsg;
      throw FCbackendException("LFCCatalog::L_get_pfnentry_from_guid","Cannot make wildcard list");
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL) {

        pd->fid = value;

        // look PFNs for this guid, ignore ENOENT
        fid_to_entry(pd->fid, pd->pfns, true);

      } else {
        pd->c.fids_add(match_type, value);
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      while(!buf.full() && pd->off < universal_set->size()) {

        if (match_type == LITERAL) {
          std::vector<PFNEntry>::iterator pfns_itr;
          for(pfns_itr=pd->pfns.begin();pfns_itr!=pd->pfns.end();++pfns_itr) {
            if (compare( (*universal_set)[pd->off], *pfns_itr ) &&
                pd->c.check( (*universal_set)[pd->off] )) {
              buf.push_back((*universal_set)[pd->off]);
              set_distinct = YES; // 'pfns' list must be distinct, so output will be too
              break;
            }
          }
        } else { /* wildcard match */
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // output only distinct if universal set was too
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      while(!buf.full() && pd->off < pd->pfns.size()) {
        if (pd->c.check(pd->pfns[pd->off])) {
          buf.push_back(pd->pfns[pd->off]);
          set_distinct = YES; // 'pfns' list must be distinct
        }
        ++pd->off;
      }
    }

    return;
  }

  void LFCCatalog::L_get_pfnentry_from_lfn(LFC_LIST_FUNC_TYPE(PFNEntry))
  {
    PoolMessageStream fclog( "LFCCatalog::L_get_pfnentry_from_lfn", seal::Msg::Nil);

    struct priv_data_t {
      FileCatalog::FileID fid;
      path_list_t path_lister;
      constraint_checker c;
      udata item;
      std::vector<PFNEntry> pfns;
      size_t off;
    } *pd;

    if (flags == LIST_CONSTRAINTS) {
      cl.lfns_add(match_type, value);
      if (match_type == WILDCARD) {
        if (universal_set != NULL) {
          set_distinct = MAYBE;
        } else {
          set_distinct = NO;
        }
      } else {
        set_distinct = YES;
      }
      priv_data = LFC_NO_NEED_SET;
      return;
    }

    if (flags == LIST_BEGIN) {
      pd = new(priv_data_t);
      priv_data = (void *)pd;

      pd->off = 0;
      pd->c.init(this,cl);

      if (match_type == LITERAL) {

        // lookup LFN and find corresponding PFNs. Ignore ENOENT
        translate_entry(value, pd->fid, pd->pfns, true);

      } else if (match_type == WILDCARD) {
        if (universal_set == NULL) {
          pd->path_lister.init(this,value);
        } else {
          pd->c.lfns_add(match_type, value);
        }
      }
    }

    pd = (priv_data_t *)priv_data;

    if (flags == LIST_END) {
      delete pd;
      return;
    }

    if (universal_set != NULL) {
      while(!buf.full() && pd->off < universal_set->size()) {

        if (match_type == LITERAL) {
          std::vector<PFNEntry>::iterator pfns_itr;
          for(pfns_itr=pd->pfns.begin();pfns_itr!=pd->pfns.end();++pfns_itr) {
            if (compare( (*universal_set)[pd->off], *pfns_itr ) &&
                pd->c.check( (*universal_set)[pd->off] )) {
              buf.push_back((*universal_set)[pd->off]);
              set_distinct = YES; // 'pfns' list must be distinct, so output will be too
              break;
            }
          }
        } else { /* wildcard match */
          if (pd->c.check((*universal_set)[pd->off])) {
            buf.push_back((*universal_set)[pd->off]);
            set_distinct = MAYBE;
            // output only distinct if universal set was too
          }
        }
        ++pd->off;
      }
    } else { /* no universal set */
      if (match_type == LITERAL) {
        while(!buf.full() && pd->off < pd->pfns.size()) {
          if (pd->c.check(pd->pfns[pd->off])) {
            buf.push_back(pd->pfns[pd->off]);
            set_distinct = YES; // 'pfns' list must be distinct
          }
          ++pd->off;
        }
      } else { /* wildcard match, no set specified so expand path */
        while(!buf.full()) {

          if (pd->off >= pd->pfns.size()) {
            pd->item = pd->path_lister.next();
            if (pd->item.guid().empty()) return;

            pd->pfns.clear();
            pd->off = 0;

            if (pd->c.check(pd->item)) {
              std::vector<PFNEntry> pfns;
              if (pd->item.need_pfns()) {
                // lookup this GUID and get list of PFNs, ignore ENOENT
                fid_to_entry(pd->item.guid(), pfns, true);
                pd->item.set_full(pfns);
              } else {
                pfns = pd->item.pfns();
              }
              if (pfns.empty()) continue;
              pd->pfns = pfns;
            } else {
              continue;
            }
          }

          udata it = pd->item.pfns(pd->off);
          if (pd->c.check(it)) {
            buf.push_back(it);
            set_distinct = NO; // may scan LFN aliases, which will give duplicate PFNs
          }
          ++pd->off;
        }
      }
    }

    return;
  }

  void LFCCatalog::destroyRetData() const {
    if (m_retPFN != NULL) {
      delete MRET(PFNEntry,m_retPFN);
      m_retPFN=NULL;
    }
    if (m_retLFN != NULL) {
      delete MRET(LFNEntry,m_retLFN);
      m_retLFN=NULL;
    }
    if (m_retGUID != NULL) {
      delete MRET(FileCatalog::FileID,m_retGUID);
      m_retGUID=NULL;
    }
  }
}
