// $Id: FuncTest.cpp,v 1.10 2006/03/17 10:59:28 dhsmith Exp $
/*
  CPPUnit test of LHC File Catalog.
  Author: James Casey
*/
#include <cppunit/extensions/HelperMacros.h>

#include "LFCCatalog/LFCCatalog.h"
#include "LFCCatalog/gfal_api.h" // for lfc_mkdirp
#include "FileCatalog/FCCommon.h"
#include "FileCatalog/FCBuf.h"
#include "FileCatalog/FCException.h"
#include "POOLCore/Guid.h"

#define LFC_HOSTNAME "lfc-dteam-test.cern.ch"
#define LFN_PREFIX std::string("/grid/dteam/")
#define PFN_PREFIX std::string("srm://host.example.com/path/to/file/")

class FuncTest : public CppUnit::TestFixture{
  CPPUNIT_TEST_SUITE( FuncTest );
  CPPUNIT_TEST( testRegisterPFN );
  CPPUNIT_TEST( testInsertPFN );
  CPPUNIT_TEST( testInsertPFNNonExistentGuid );
  CPPUNIT_TEST( testRegisterLFN );
  CPPUNIT_TEST( testRegisterLFNBadPFN );
  CPPUNIT_TEST( testAddReplicaPFN );
  CPPUNIT_TEST( testAddReplicaPFNBadPFN );
  CPPUNIT_TEST( testInsertLFN );
  CPPUNIT_TEST( testInsertLFNNonExistentGuid );

  CPPUNIT_TEST_SUITE_END();

private:
  void helper_remove_by_lfn(const std::string& lfn) {
    pool::FileCatalog::FileID guid;
    
    m_catalog.lookupFileByLFN(lfn, guid);
    if (guid != "") {
      m_catalog.deleteEntry(guid);
    }
  }
  
  void helper_remove_by_pfn(const std::string& pfn) {
    pool::FileCatalog::FileID guid;
    std::string ftype;
    m_catalog.lookupFileByPFN(pfn, guid, ftype);
    if (guid != "") {
      m_catalog.deleteEntry(guid);
    }
  }

  void helper_create_test_root(const std::string& root) {
    if(lfc_mkdirp(root.c_str(), 0777, errbuf, 1023) <0 ) {
      CPPUNIT_FAIL("Can't make test root");
    }
  }

  void helper_remove_test_root(const std::string& root) {
    (void) lfc_rmdir(root.c_str());
  }

public:
  void setUp(){ 

    {
      std::string s = std::string("LFC_HOST=")+LFC_HOSTNAME;
      char *p = (char *)malloc(s.size()+1);
      strcpy(p,s.c_str());
      putenv(p);
    }

    std::string name("catalog_tests");

    m_catalog.setURL(std::string("lfc://")+LFC_HOSTNAME);
    m_catalog.setType("lfccatalog");
    m_catalog.connect();

    instance_dir = "pool-testdir-" + make_guid() + "/";

    helper_create_test_root(LFN_PREFIX + instance_dir);

    m_catalog.start();

    // now add a sample lfn and guid for testing replica functions
    file_guid = make_guid();
    file_path = make_lfn(name);
    file_pfn = make_pfn(name);

    pool::PFNEntry epfn(file_pfn, file_guid, "");
    pool::LFNEntry elfn(file_path, file_guid);

    try {
      m_catalog.insertPFN(epfn);
      m_catalog.insertLFN(elfn);
    } catch(...) {
      CPPUNIT_FAIL("Can't create sample lfn/guid/surls");
    }
  }
  
  void tearDown(){
    try {
      m_catalog.rollback(); // don't need to commit any of this testing stuff in the DB
      m_catalog.disconnect();
    } catch(...) {
    }
    helper_remove_test_root(LFN_PREFIX + instance_dir);
    {
      char *p=getenv("LFC_HOST");
      unsetenv("LFC_HOST");
      if (p!=NULL)
        free(p);
    }
  }

  /** Test that we can add a new guid -> pfn mapping */
  void testRegisterPFN(){
    std::cout<<"[OVAL]TEST --> testRegisterPFN"<<std::endl;

    std::string guid(make_guid());
    std::string pfn(make_pfn("registerPFN"));
    
    try {
      m_catalog.registerPFN(pfn, "", guid);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    // and if we remove it we should be able to add it again
    helper_remove_by_pfn(pfn);

    try {
      m_catalog.registerPFN(pfn, "", guid);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    bool ok = false;
    
    // now we can try again, and we expect an exception
    try {
      m_catalog.registerPFN(pfn, "", guid);
    } catch (const pool::FCduplicatePFNException &e) {
      ok = true;
    } catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    if (!ok) CPPUNIT_FAIL("Expected Exception on duplicate PFN insertion");
  }

/** check we can add a pfn to an existing guid */
  void testInsertPFN(){
    std::cout<<"[OVAL]TEST --> testInsertPFN"<<std::endl;

    std::string pfn(make_pfn("insertPFN"));
    
    pool::PFNEntry pentry(pfn, file_guid, "");

    try {
      m_catalog.insertPFN(pentry);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
    
    bool ok=false;

    // now try again - we expect an exception
    try {
      m_catalog.insertPFN(pentry);
    } catch(const pool::FCduplicatePFNException &e) {
      ok = true;
    } catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    if (!ok) CPPUNIT_FAIL("Expected Exception on duplicate PFN insertion");
  }

  /** If we try and do an insertPFN when the guid isn't already there,
      the guid should be created. This test used to throw an exception */
  void testInsertPFNNonExistentGuid() {
    std::cout<<"[OVAL]TEST --> testInsertPFNNonExistentGuid"<<std::endl;

    std::string pfn(make_pfn("insertPFNNoGuid"));
    std::string no_guid(make_guid());

    pool::PFNEntry pentry(pfn, no_guid, "");

    try {
      m_catalog.insertPFN(pentry);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
  }

  /** check we can register an LFN onto an existing PFN */
  void testRegisterLFN() {
    std::cout<<"[OVAL]TEST --> testRegisterLFN"<<std::endl;

    std::string pfn(make_pfn("registerLFN"));
    std::string lfn(make_lfn("registerLFN"));
    std::string guid;
 
    // setup - create the PFN
    try {
      m_catalog.registerPFN(pfn, "", guid);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    // test - register the lfn
    try {
      m_catalog.registerLFN(pfn, lfn);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    bool ok = false;
    // try registing the LFN to the PFN again - expect an exception
    try {
      m_catalog.registerLFN(pfn, lfn);
    }catch(const pool::FCduplicateLFNException &e) {
      ok = true;
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    if (!ok) CPPUNIT_FAIL("Expected Exception on duplicate LFN insertion");
  }

  /** check we get an exception if we register an LFN onto an non-existant
      PFN */
  void testRegisterLFNBadPFN() {
    std::cout<<"[OVAL]TEST --> testRegisterLFNBadPFN"<<std::endl;

    std::string pfn(make_pfn("willNeverEverExist"));
    std::string lfn(make_lfn("registerLFNBadPFN"));
    std::string guid(make_guid());
 
    // try to register - we expect an exception
    try {
      m_catalog.registerLFN(pfn, lfn);
    }catch(const pool::FCnonexistentFileException &e) {
      return;
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
    CPPUNIT_FAIL("Expected Exception on LFN insertion to non-existent file");
  }

  /* test that we can add another PFN to an existing one */
  void testAddReplicaPFN() {
    std::cout<<"[OVAL]TEST --> testAddReplicaPFN"<<std::endl;
    std::string pfn(make_pfn("addReplicaPFN"));
    std::string extrapfn(make_pfn("addReplicaPFNSecondOne"));
    std::string guid;
    
    // setup - create the PFN
    try {
      m_catalog.registerPFN(pfn, "", guid);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
    
    // test - register the extra pfn to the first PFN
    try {
      m_catalog.addReplicaPFN(pfn, extrapfn);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    bool ok=false;

    // try registering the extra PFN to the PFN again - we expect an exception
    try {
      m_catalog.addReplicaPFN(pfn, extrapfn);
    }catch(const pool::FCduplicatePFNException &e) {
      ok = true;
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    if (!ok) CPPUNIT_FAIL("Expected Exception on duplicate PFN insertion");
  }

  void testAddReplicaPFNBadPFN() {
    std::cout<<"[OVAL]TEST --> testAddReplicaPFNBadPFN"<<std::endl;
    std::string pfn(make_pfn("ThisPfnDoesNotExist"));
    std::string extrapfn(make_pfn("addReplicaPFNSecondOne"));
    std::string guid(make_guid());
    
    // test - add the replica pfn - we should get an exception
    try {
      m_catalog.addReplicaPFN(pfn, extrapfn);
    }catch(const pool::FCnonexistentFileException &e) {
      return;
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
    CPPUNIT_FAIL("Expected Exception on adding replica to nonexistent PFN");
  }

  void testInsertLFN() {
    std::cout<<"[OVAL]TEST --> testInsertLFN"<<std::endl;    
    std::string pfn(make_pfn("insertLFN"));
    std::string lfn(make_lfn("insertLFN"));
    std::string guid,guid2;
 
    // setup - create the PFN
    try {
      m_catalog.registerPFN(pfn, "", guid);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }

    // test - register the lfn with the first GUID
    pool::LFNEntry lentry(lfn, guid);
    try {
      m_catalog.insertLFN(lentry);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
      
    bool ok=false;

    // try registering the LFN to the PFN/GUID again - should give an exception
    try {
      m_catalog.insertLFN(lentry);
    }catch(const pool::FCduplicateLFNException &e) {
      ok = true;
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
    if (!ok) CPPUNIT_FAIL("Expected Exception on duplicate LFN insertion");
  }

  // This used to throw an exception - but now we allow insertLFN with
  // non existant GUID.
  void testInsertLFNNonExistentGuid() {
    std::cout<<"[OVAL]TEST --> testInsertLFNNonExistentGuid"<<std::endl;
    std::string lfn(make_lfn("insertLFNNonExistentGuid"));
    std::string guid(make_guid());
 
    // test - register the lfn
    pool::LFNEntry lentry(lfn, guid);
    try {
      m_catalog.insertLFN(lentry);
    }catch (const pool::Exception &e) {
      CPPUNIT_FAIL(std::string("Caught Exception : ") + e.what());
    }
  }

private:
  std::string make_guid() {
    pool::Guid g;
    pool::Guid::create(g);
    return g.toString();
  }

  std::string make_lfn(const std::string path) {
    return LFN_PREFIX + instance_dir + path;
  }

  std::string make_pfn(const std::string path) {
    return PFN_PREFIX + path;
  }

private:
  pool::LFCCatalog m_catalog;
  std::string instance_dir;
  std::string file_guid;
  std::string file_pfn;
  std::string file_path;
  char errbuf[1024];
};

CPPUNIT_TEST_SUITE_REGISTRATION(FuncTest);

#include<CppUnit_testdriver.cpp>
