#ifndef OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGPERSISTENCY_H
#define OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGPERSISTENCY_H

#include <map>
#include <vector>
#include <string>
#include <set>

namespace coral {

  class AttributeList;
  class ISchema;
  class ITableDataEditor;
}

namespace pool 
{
  
  class ObjectRelationalMapping;
  class ObjectRelationalMappingElement;

  /**
   * @class ObjectRelationalMappingPersistency ObjectRelationalMappingPersistency.h ObjectRelationalAccess/ObjectRelationalMappingPersistency.h
   *
   * Class respondible for the I/O of object/relational mappings with a relational database.
   */
  
  class ObjectRelationalMappingPersistency
  {
  public:
    /// Constructor
    explicit ObjectRelationalMappingPersistency( coral::ISchema& schema );

    /// Destructor
    ~ObjectRelationalMappingPersistency();

    void initialize() const;

    bool upgradeSchema();
    
    /**
     *
     *
     */
    bool existsMappingDatabase();

    /**
     * Stores a mapping structure in the database.
     *
     * @param mapping The mapping structure
     *        
     */
    bool storeMapping( const ObjectRelationalMapping& mapping, const std::string& classVersion, std::string containerName="" );

    /**
     * Retrieves the versions of the mappings stored in the database
     */
    std::map<std::string,std::string> retrieveMappingVersions() const;
    
    /**
     * Retrieves the versions of the mappings stored in the database
     */
    std::map<std::pair<std::string,std::string>,std::string> retrieveMappingForClassVersions() const;

    /**
     * Retrieves the mapping version (if any) corresponding to the specified class version and container name.
     */
    std::string selectMappingVersion( const std::string& classVersion, const std::string& containerName) const;

    
    std::string containerForMappingVersion( const std::string& mappingVersion) const;

    std::vector<std::string> mappingsForContainer( const std::string& containerName ) const;

    /**
     * Delivers a map of the table usage by containers
     */
    std::map<std::string, std::string> containerTableMap();

    void selectMapping(const std::string& version, std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct) const;
    
    /**
     * Retrieves a mapping of a given version
     *
     * @param version The version of the mapping
     */
    const ObjectRelationalMapping& retrieveMapping( const std::string& version ) const;

    /**
     * Retrieves a mapping of a given version
     *
     * @param version The version of the mapping
     */
    ObjectRelationalMapping* exportMapping( const std::string& version ) const;

    /**
     * Removes a mapping from the database given its version
     *
     * @param version The version of the mapping
     */
    bool removeMapping( const std::string& version );

    /**
     * Removes the tables materialized from the database given their mapping version
     *
     * @param version The version of the mapping
     */
    bool removeMaterializedTables(const std::string& mappingVersion, bool dryRun=false );

    /**
     * Creates all the tables needed for the storage of objects corresponding to a mapping.
     *
     * @param version The version of the mapping. It should be an existing one.
     * @param evolve Boolean flag which if set allows the altering of existing tables to accomodate the mapping.
     */
    bool materializeMapping( const std::string& version,
                             bool evolve, bool dryRun=false );

    /**
     * Creates all the tables needed for the storage of objects corresponding to a mapping.
     *
     * @param mapping The mapping to materialize.
     * @param evolve Boolean flag which if set allows the altering of existing tables to accomodate the mapping.
     */
    bool materializeMapping( const pool::ObjectRelationalMapping& mapping,
                             bool evolve, bool dryRun=false );
    
    bool assignMapping( const std::string& classVersion, const std::string& mappingVersion, const std::string& containerName);
    private:
    
    struct MappingDescription {
        MappingDescription():
          variableType(""),
          elementType(""),
          tableName(""),
          columns(){}
        MappingDescription(const MappingDescription& rhs):
          variableType(rhs.variableType),
          elementType(rhs.elementType),
          tableName(rhs.tableName),
          columns(rhs.columns){}
        MappingDescription& operator==(const MappingDescription& rhs){
          variableType = rhs.variableType;
          elementType = rhs.elementType;
          tableName = rhs.tableName;
          columns = rhs.columns; 
          return *this;
        }
        std::string variableType;
        std::string elementType;
        std::string tableName;
        std::vector<std::string> columns;
    }; 
    
    private:
    /**
     * Reference to the relational schema in use
     */
    coral::ISchema& m_schema;

    /**
     * The map of the mappings in the database
     */
    std::map< std::string, ObjectRelationalMapping* > m_storedMappings;

    bool m_versionsLoaded;
    
    /**
     *  Temporary cache for the loading of the mapping from db into the proper structures
     */
    std::map< std::string, std::map<std::string, MappingDescription> > m_tempMappingTable;

  private:
    /// Creates if necessary the mapping tables
    void createIfNecessaryTheMappingTables();
    /// Loads the mapping versions
    void loadMappingVersions();
    /// Loads the mapping of a specific version
    void loadMapping( const std::string& version );
    /// Loads the mapping of a specific version, retrieving data with a loop of queries
    void loadMappingLoop( const std::string& version, pool::ObjectRelationalMapping* map );
    /// Loads the mapping of a specific version, retrieving data with a single query
    void loadMappingInOneGo( const std::string& version, pool::ObjectRelationalMapping* map );
    
    /// Loads the mapping elements under a given element
    void loadMappingElement( const std::string& version,
                             ObjectRelationalMappingElement& parentElement );
    /// Loads the columns of a mapping element
    void loadMappingElementColumns( const std::string& version,
                                    ObjectRelationalMappingElement& element );

    void makeTableList( const ObjectRelationalMappingElement& element,
                        std::set<std::string>& tableList );

    void recursiveStoreMappingElement( ObjectRelationalMappingElement& mappingElement, const std::string& scopeName );
    
  private:
    /// Stores a mapping element
    static void storeMappingElement( coral::ITableDataEditor& dataEditor,
                                     coral::AttributeList& dataBuffer,
                                     coral::ITableDataEditor& columnsDataEditor,
                                     coral::AttributeList& columnsDataBuffer,
                                     const pool::ObjectRelationalMappingElement& element );
  };
}

#endif // OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGPERSISTENCY_H
