#include "MappingBuilder.h"
#include "XMLDatabaseManagerDriver.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h"
#include "ObjectRelationalAccess/ObjectRelationalTableNameRationalizer.h"
#include "ObjectRelationalAccess/ObjectRelationalTableVariableNameRationalizer.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalDatabaseDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include <sstream>
#include <fstream>
#include "Reflex/Type.h"
#include "SealKernel/Context.h"
#include "SealKernel/ComponentLoader.h"
#include "SealKernel/MessageStream.h"
#include "CoralBase/Exception.h"

pool::MappingBuilder::MappingBuilder( seal::Context* c, coral::ISchema& schema ):
  m_initialized(false),
  m_builtMappings(),
  m_containerMappings(),
  m_mappingVersions(),
  m_context(c),
  m_schema(schema),
  m_mappingPersistency(schema){
}

pool::MappingBuilder::~MappingBuilder(){
  deleteMappings();
}

void pool::MappingBuilder::init(){
  if(!m_initialized){
    m_mappingPersistency.initialize();
    m_mappingVersions.clear();
    std::map<std::string,std::string> vers = m_mappingPersistency.retrieveMappingVersions();
    for(std::map<std::string,std::string>::const_iterator iV=vers.begin();
        iV!=vers.end();++iV){
      m_mappingVersions.insert(std::make_pair(iV->first,true));
    }
    m_initialized = true;
  }
}


void pool::MappingBuilder::deleteMappings(){
  for(std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::iterator iM=m_builtMappings.begin();
      iM!=m_builtMappings.end();++iM) {
    delete iM->second.first;
  }
  m_builtMappings.clear();
}

void pool::MappingBuilder::buildMapping(const ObjectRelationalMappingDescriptor* mappingDescriptor){
  this->init();
  std::auto_ptr<ObjectRelationalTableNameRationalizer> rationalizer(new ObjectRelationalTableNameRationalizer( m_schema ));
  pool::ObjectRelationalTableVariableNameRationalizer vrationalizer;
  
  const std::vector<ObjectRelationalMappingDescriptorElement*>& classDescriptors =  mappingDescriptor->classDescriptors();
  std::vector<ObjectRelationalMappingDescriptorElement*>::const_iterator iClassDescriptor;
  ObjectRelationalDefaultMappingBuilder defaultBuilder;
  
  std::map<std::string,std::string> containerTableMap = m_mappingPersistency.containerTableMap();
  seal::MessageStream log( m_context, "MappingBuilder",seal::Msg::Verbose);
  for(iClassDescriptor=classDescriptors.begin();iClassDescriptor!=classDescriptors.end();iClassDescriptor++){

    std::string className = (*iClassDescriptor)->name();
    
    ROOT::Reflex::Type dict = ROOT::Reflex::Type::ByName( className );
    if ( ! dict ) {
      throw ObjectRelationalException( "Dictionary information for class \"" + className + "\" not found.",
                                       "MappingBuilder::buildMapping");
    }

    std::string classVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::defaultClassVersion( className );
    ROOT::Reflex::PropertyList classProps = dict.Properties();
    if(classProps.HasProperty(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary())){
      classVersion = classProps.PropertyAsString(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary());
    }

    bool buildMapping = true;
    std::string containerName("");
    try{
      containerName = (*iClassDescriptor)->attribute(pool::ObjectRelationalMappingDescriptorElement::poolContainerNameAttr());
    } catch (const ObjectRelationalException&){
    }
    if(!containerName.empty()){
      std::string useMapping("");
      try{
        useMapping = (*iClassDescriptor)->attribute(ObjectRelationalMappingDescriptorElement::useMappingAttr());
      } catch (const ObjectRelationalException&){
      }
      if(!useMapping.empty()){
        buildMapping = false;
        std::map<std::string,bool>::const_iterator iM=m_mappingVersions.find(useMapping);
        if(iM==m_mappingVersions.end()){
          throw ObjectRelationalException( "Mapping \""+ useMapping +"\" specified for container \""+containerName+"\" does not exist in the database.",
                                           "MappingBuilder::buildMapping");
        } else {
          m_containerMappings.insert(std::make_pair(containerName,std::make_pair(classVersion,useMapping)));
        }
      }
      
    }

    if(buildMapping){

      std::string mappingVersion("");
      try{
        mappingVersion = (*iClassDescriptor)->attribute(ObjectRelationalMappingDescriptorElement::mappingVersionAttr());
      } catch (const ObjectRelationalException&){
      }

      if(mappingVersion.empty()){
        for ( int iteration = 0;; ++iteration ) {
          mappingVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::newMappingVersionForClass( className, iteration );
          bool found = false;
          for ( std::map<std::string,bool>::reverse_iterator iVersion = m_mappingVersions.rbegin();
                iVersion != m_mappingVersions.rend(); ++iVersion ) {
            if ( iVersion->first == mappingVersion ) {
              found = true;
              break;
            }
          }
          if ( ! found ) break;
        }
      } else {
        std::map<std::string,bool>::const_iterator iVer = m_mappingVersions.find(mappingVersion);
        if(iVer!=m_mappingVersions.end()){
          throw ObjectRelationalException( "Mapping \""+ mappingVersion +"\" already exists in the database.",
                                           "MappingBuilder::buildMapping");          
        }
        
      }
      
      m_mappingVersions.insert(std::make_pair(mappingVersion,false));

      std::string baseMappingVersion("");
      bool useBase = false;
      try{
        baseMappingVersion = (*iClassDescriptor)->attribute(ObjectRelationalMappingDescriptorElement::baseMappingAttr());
      } catch (const ObjectRelationalException&){
      }
      if(!baseMappingVersion.empty()){
        std::map<std::string,bool>::const_iterator iBm = m_mappingVersions.find( baseMappingVersion );
        if(iBm == m_mappingVersions.end() || !iBm->second){
          throw ObjectRelationalException( "Base Mapping \""+ baseMappingVersion +"\" does not exist in the database.",
                                           "MappingBuilder::buildMapping");
        } else {
          useBase = true;
        }
      }
      
      ObjectRelationalMapping* defaultMapping = 0;
      std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::iterator iM = m_builtMappings.find(mappingVersion);
      if(iM==m_builtMappings.end()){
        defaultMapping = new ObjectRelationalMapping(mappingVersion);
        ClassMetadata mdata = std::make_pair(classVersion,containerName);
        m_builtMappings.insert(std::make_pair(mappingVersion,std::make_pair(defaultMapping,mdata)));
      } else {
        defaultMapping = iM->second.first;
      }

      ROOT::Reflex::Type dictionaryEntry = dict;
      className = dictionaryEntry.Name(ROOT::Reflex::SCOPED);
      log << seal::Msg::Debug << "Processing class " << className << seal::flush;
      ObjectRelationalMapping::iterator iDefaultMappingElement = defaultMapping->find( className );
      if(iDefaultMappingElement==defaultMapping->end()){
        if ( !defaultBuilder.processClass(*defaultMapping, dictionaryEntry) ){
          throw ObjectRelationalException( "Processing of default mapping for class \"" + className + "\" failed.",
                                           "MappingBuilder::buildMapping" );
        }
      }
      iDefaultMappingElement = defaultMapping->find( className );
      if(iDefaultMappingElement==defaultMapping->end()){
        throw ObjectRelationalException( "Mapping for class \"" + className + "\" not found.",
                                         "MappingBuilder::buildMapping");
      }

      // rationalize mapping element
      rationalizer->rationalizeMappingElement(iDefaultMappingElement->second );
      vrationalizer.rationalizeMappingElement(iDefaultMappingElement->second );
      if(useBase){
        ObjectRelationalMapping* baseMapping = m_mappingPersistency.exportMapping( baseMappingVersion );
        if(baseMapping) {
          std::auto_ptr<ObjectRelationalMapping> baseM(baseMapping);
          ObjectRelationalMapping::iterator iBaseMappingElement = baseM->find(className);
          if(iBaseMappingElement!=baseM->end()){
            iDefaultMappingElement->second.override(iBaseMappingElement->second);
          }
        } else {
          throw ObjectRelationalException( "Could not retrieve the version \"" + baseMappingVersion + "\" as base mapping.",
                                           "MappingBuilder::buildMapping" );          
        }
      }
      (*iClassDescriptor)->alter(iDefaultMappingElement->second,vrationalizer);
      //check table usage
      std::set<std::string> involvedTables = iDefaultMappingElement->second.involvedTables();
      for(std::set<std::string>::const_iterator iT=involvedTables.begin();
          iT!=involvedTables.end();++iT){
        std::map<std::string,std::string>::const_iterator iCt = containerTableMap.find(*iT);
        if(iCt==containerTableMap.end()){
          containerTableMap.insert(std::make_pair(*iT,containerName));
        } else {
          if(iCt->second!=containerName){
            throw ObjectRelationalException( "Table \""+*iT+"\" involved in mapping of class \"" + className +
                                             "\" for container \"" + containerName+"\" is already used by container \""+iCt->second+"\".",
                                             "MappingBuilder::buildMapping" );            
          }
        }
      }
      
    }
    
  }
}

bool pool::MappingBuilder::exportMapping(const std::string& fileName){
  if(m_builtMappings.size()){
    std::auto_ptr<ObjectRelationalDatabaseDescriptor> databaseDescriptor(new ObjectRelationalDatabaseDescriptor);
    for(std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::iterator iM=m_builtMappings.begin();
        iM!=m_builtMappings.end();++iM){
      pool::ObjectRelationalContainerDescriptor& contDescriptor = databaseDescriptor->addContainer(iM->second.second.second,"");
      ObjectRelationalMapping::iterator ime=iM->second.first->begin();
      if(ime!=iM->second.first->end()){
        pool::ObjectRelationalClassDescriptor* classDescriptor = contDescriptor.addClassStore(ime->second.variableType(),iM->first);
        ObjectRelationalMappingDescriptor* mappDescr =  classDescriptor->newMapping();
        mappDescr->import(*(iM->second.first));
      }
    }
    XMLDatabaseManagerDriver driver;
    std::ofstream outFile;
    outFile.open (fileName.c_str(), std::fstream::out);
    driver.write(*databaseDescriptor,outFile);
    outFile.close();
    return true;
  } else {
    return false;
  }
}

bool pool::MappingBuilder::retrieveMapping(const std::string& version,const std::string& fileName){
  if(!m_mappingPersistency.existsMappingDatabase()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMapping");
  }
  const ObjectRelationalMapping& mapping = m_mappingPersistency.retrieveMapping( version );
  std::string containerName = m_mappingPersistency.containerForMappingVersion( version );
  
  std::auto_ptr<ObjectRelationalDatabaseDescriptor> databaseDescriptor(new ObjectRelationalDatabaseDescriptor);
  pool::ObjectRelationalContainerDescriptor& contDescriptor = databaseDescriptor->addContainer(containerName,"");
  ObjectRelationalMapping::const_iterator iM=mapping.begin();
  if(iM!=mapping.end()){
    pool::ObjectRelationalClassDescriptor* classDescriptor = contDescriptor.addClassStore(iM->second.variableType(),version);
    ObjectRelationalMappingDescriptor* mappDescr =  classDescriptor->newMapping();
    mappDescr->import(mapping);
  }
  XMLDatabaseManagerDriver driver;
  std::ofstream outFile;
  outFile.open (fileName.c_str(), std::fstream::out);
  driver.write(*databaseDescriptor,outFile);
  outFile.close();
  return true;
}

std::map<std::string,std::string >
pool::MappingBuilder::mappingVersions(){
  if(!m_mappingPersistency.existsMappingDatabase()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMappingVersions");
  }
  return m_mappingPersistency.retrieveMappingVersions();
}

std::map<std::pair<std::string,std::string>, std::string >
pool::MappingBuilder::mappingForClassVersions(){
  if(!m_mappingPersistency.existsMappingDatabase()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMappingVersions");
  }
  return m_mappingPersistency.retrieveMappingForClassVersions();  
}

std::map<std::string, std::string>
pool::MappingBuilder::containerTableMap(){
  if(!m_mappingPersistency.existsMappingDatabase()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::retrieveMappingVersions");
  }
  return m_mappingPersistency.containerTableMap();  
}

void
pool::MappingBuilder::selectMapping(const std::string& version, std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct){
  m_mappingPersistency.selectMapping(version,mappingStruct);
}

bool 
pool::MappingBuilder::removeMapping( const std::string& version, bool removeTables, bool dryRun ){
  if(!m_mappingPersistency.existsMappingDatabase()){
    throw ObjectRelationalException( "Mapping database not found.",
                                     "MappingBuilder::removeMapping");
  }
  bool ret = true;
  if( removeTables ) ret = m_mappingPersistency.removeMaterializedTables( version, dryRun );
  if( ret && !dryRun ) ret = m_mappingPersistency.removeMapping(version);
  return ret;
}

void pool::MappingBuilder::finalizeMapping(bool evolve, bool dryRun){
  if(m_containerMappings.size() || m_builtMappings.size()){
    for(std::map<std::string,std::pair<ObjectRelationalMapping*,ClassMetadata> >::const_iterator iM=m_builtMappings.begin();
        iM!=m_builtMappings.end();++iM){
      if(!dryRun) {
        if(!m_mappingPersistency.storeMapping( *(iM->second.first),iM->second.second.first,iM->second.second.second)){
          throw ObjectRelationalException( "Could not store the mapping.","MappingBuilder::finalizeMapping");
        }
      }
      if(!m_mappingPersistency.materializeMapping( *(iM->second.first), evolve, dryRun )){
        throw ObjectRelationalException( "Could not materialize the mapping.","MappingBuilder::finalizeMapping");
      }
    }
    for(std::map<std::string,std::pair<std::string,std::string> >::const_iterator iC=m_containerMappings.begin();
        iC!=m_containerMappings.end();++iC){
      if(!m_mappingPersistency.assignMapping( iC->second.first, iC->second.second, iC->first)){
        throw ObjectRelationalException( "Could not assign the mapping.","MappingBuilder::finalizeMapping");
      }
    }
  } else {
    seal::MessageStream log( m_context, "MappingBuilder",seal::Msg::Verbose);
    log << seal::Msg::Debug << "No mapping to finalize..." << seal::flush;
  }
}




