#include "MappingMaterialization.h"
#include "TableDescriptionStack.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "CoralBase/Blob.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/AttributeListSpecification.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/IForeignKey.h"
#include "RelationalAccess/IIndex.h"
#include "RelationalAccess/IUniqueConstraint.h"
#include "RelationalAccess/IPrimaryKey.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IView.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/SchemaException.h"
#include "RelationalAccess/TableDescription.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "POOLCore/POOLContext.h"
#include "SealKernel/MessageStream.h"
#include <algorithm>
#include <sstream>

pool::ObjectRelationalAccess::MappingMaterialization::MappingMaterialization( coral::ISchema& schema ):
  m_schema( schema ),
  m_dryRun(false),
  m_tableDescriptionStack( 0 ),
  m_tableMap(),
  m_mappingToProcess(true),
  m_operationList(){}

pool::ObjectRelationalAccess::MappingMaterialization::~MappingMaterialization()
{
  if ( m_tableDescriptionStack ) delete m_tableDescriptionStack;
}

bool
pool::ObjectRelationalAccess::MappingMaterialization::materialize( const pool::ObjectRelationalMapping& mapping, bool evolve, bool dryRun )
{
  m_dryRun = dryRun;
  
  if ( m_tableDescriptionStack ) delete m_tableDescriptionStack;
  m_tableDescriptionStack = new pool::ObjectRelationalAccess::TableDescriptionStack;

  if(!this->processMapping(mapping, evolve)) return false;
  
    // check if the existing columns in the tables are all mapped to the object attributes 
  for(std::map<std::string,std::set<std::string> >::const_iterator iT=m_tableMap.begin();
      iT!=m_tableMap.end();++iT){
    if( m_schema.existsTable( iT->first )){

      coral::ITable& table = m_schema.tableHandle( iT->first );
      seal::MessageStream log( pool::POOLContext::context(),
                               pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
      log << seal::Msg::Debug << "Checking existing columns in table \"" << iT->first <<"\"."
          << seal::flush;
      int nColumns = table.description().numberOfColumns();
      for(int i=0;i<nColumns;++i){
        const coral::IColumn& column = table.description().columnDescription(i);
        std::set< std::string >::const_iterator iC = iT->second.find(column.name());
        if(iC==iT->second.end()){
          log << seal::Msg::Info << "The column named \"" << column.name() << "\" in table \"" << iT->first
              << "\" is not mapped to any class attribute."
              << seal::flush;
          if(column.isNotNull()){
            if ( ! evolve ) {
              log << seal::Msg::Error << "The column named \"" << column.name() << "\" in table \"" << iT->first
                  << "\", has been found with a NOT NULL constraint, and automatic schema evolution has been disabled."
                  << seal::flush;
              return false;
            }
            else {
              std::ostringstream message;
              message << "Unsetting NOT NULL constraint on column \"" << column.name() << "\" in table \"" << iT->first <<"\"";
              this->logOperation( message.str() );
              if(!m_dryRun) table.schemaEditor().setNotNullConstraint( column.name(), false );
            }
          }
        }
      }
    }
  }

  // create the tables with the given description set.
  for ( pool::ObjectRelationalAccess::TableDescriptionStack::const_iterator iTableName = m_tableDescriptionStack->begin();
        iTableName != m_tableDescriptionStack->end(); ++iTableName ) {
    std::ostringstream message;
    message << "Creating table \"" << *iTableName <<"\"";
    this->logOperation( message.str() );
    if(!m_dryRun) m_schema.createTable( m_tableDescriptionStack->description( *iTableName ) ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
        
  }
  
  delete m_tableDescriptionStack;
  m_tableDescriptionStack = 0;

  if(m_mappingToProcess){
    if(!this->processMapping(mapping, evolve)) return false;
  }
  
  return true;
}

const std::set<std::string>&
pool::ObjectRelationalAccess::MappingMaterialization::operationList() const {
  return m_operationList;
}

bool
pool::ObjectRelationalAccess::MappingMaterialization::processMapping(const pool::ObjectRelationalMapping& mapping, bool evolve){
  m_mappingToProcess = false;
  // Loop over the elements of the mapping.
  for ( pool::ObjectRelationalMapping::const_iterator iElement = mapping.begin();
        iElement != mapping.end(); ++iElement ) {
    const pool::ObjectRelationalMappingElement& element = iElement->second;
    if ( ! this->processElement( element, evolve ) ) {
      delete m_tableDescriptionStack;
      m_tableDescriptionStack = 0;
      return false;
    }
  }
  return true;
}

bool
pool::ObjectRelationalAccess::MappingMaterialization::isTableEmpty(const std::string& tableName){
  coral::ITable& table = m_schema.tableHandle( tableName );
  std::auto_ptr< coral::IQuery > query( table.newQuery() );
  query->limitReturnedRows( 1, 0 );
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) return false;
  else return true;
}

bool
pool::ObjectRelationalAccess::MappingMaterialization::processElement( const pool::ObjectRelationalMappingElement& element,
                                                                      bool evolve )
{
  if ( ! ( this->processObjectElement( element, evolve ) ) ) return false;

  //  compile the table map
  std::map<std::string, std::set<std::string> >::iterator iT = m_tableMap.find(element.tableName());
  if(iT==m_tableMap.end()) {
    std::set<std::string> cols;
    iT = m_tableMap.insert(std::make_pair(element.tableName(),cols)).first;
  }
  for(std::vector<std::string>::const_iterator iC=element.columnNames().begin();
      iC!=element.columnNames().end();++iC){
    iT->second.insert(*iC);
  }

  //  loop over sub-elements.
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    const pool::ObjectRelationalMappingElement& subElement = iElement->second;
    if ( ! this->processSubElement( element, subElement, evolve ) ) return false;
  }

  return true;
}

bool
pool::ObjectRelationalAccess::MappingMaterialization::processSubElement( const pool::ObjectRelationalMappingElement& parentElement,
                                                                         const pool::ObjectRelationalMappingElement& element,
                                                                         bool evolve )
{
  pool::ObjectRelationalMappingElement::ElementType elementType = element.elementType();
  switch ( elementType ) {
  case pool::ObjectRelationalMappingElement::Object :
    if ( !( this->processObjectElement( element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::Primitive :
    if ( !( this->processPrimitiveElement( element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::Blob :
    if ( !( this->processBlobElement( element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::Array :
    if ( !( this->processArrayElement( parentElement, element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::CArray :
    if ( !( this->processArrayElement( parentElement, element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::InlineCArray :
    if ( !( this->processObjectElement( element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::PoolReference :
    if ( !( this->processReferenceElement( element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::Reference :
  case pool::ObjectRelationalMappingElement::Pointer :
    if ( !( this->processPointerElement( element, evolve ) ) ) return false;
    break;
  case pool::ObjectRelationalMappingElement::Undefined :
    // This should never appear
    break;
  };

  //  compile the table map
  std::map<std::string, std::set<std::string> >::iterator iT = m_tableMap.find(element.tableName());
  if(iT==m_tableMap.end()) {
    std::set<std::string> cols;
    iT = m_tableMap.insert(std::make_pair(element.tableName(),cols)).first;
  }
  for(std::vector<std::string>::const_iterator iC=element.columnNames().begin();
      iC!=element.columnNames().end();++iC){
    iT->second.insert(*iC);
  }

  //  loop over sub-elements.
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    const pool::ObjectRelationalMappingElement& subElement = iElement->second;
    if ( ! this->processSubElement( element, subElement, evolve ) ) return false;
  }

  return true;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::existsUniqueIndex( const coral::ITableDescription& description,
                                                                         const std::vector< std::string >& columnNames ) const
{
  if ( description.hasPrimaryKey() ) {
    const std::vector<std::string>& columns = description.primaryKey().columnNames();
    if ( columns.size() == columnNames.size() ) {
      bool same = true;
      for ( size_t i = 0; i < columns.size(); ++i ) {
        if ( columns[i] != columnNames[i] ) {
          same = false;
          break;
        }
      }
      if ( same ) return true;
    }
  }

  int numberOfUniqueConstraints = description.numberOfUniqueConstraints();
  for ( int iConstraint = 0; iConstraint < numberOfUniqueConstraints; ++iConstraint ) {
    const std::vector<std::string>& columns = description.uniqueConstraint( iConstraint ).columnNames();
    if ( columns.size() == columnNames.size() ) {
      bool same = true;
      for ( size_t i = 0; i < columns.size(); ++i ) {
        if ( columns[i] != columnNames[i] ) {
          same = false;
          break;
        }
      }
      if ( same ) return true;
    }
  }

  int numberOfIndices = description.numberOfIndices();
  for ( int iIndex = 0; iIndex < numberOfIndices; ++iIndex ) {
    if ( ! description.index( iIndex ).isUnique() ) continue;
    const std::vector<std::string>& columns = description.index( iIndex ).columnNames();
    if ( columns.size() == columnNames.size() ) {
      bool same = true;
      for ( size_t i = 0; i < columns.size(); ++i ) {
        if ( columns[i] != columnNames[i] ) {
          same = false;
          break;
        }
      }
      if ( same ) return true;
    }
  }

  return false;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::processObjectElement( const pool::ObjectRelationalMappingElement& element,
                                                                            bool evolve )
{
  if ( ! this->processObject( element.tableName(), element.columnNames(), false, evolve ) ) return false;
  // Process the sub elements
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    const pool::ObjectRelationalMappingElement& subElement = iElement->second;
    if ( ! this->processSubElement( element, subElement, evolve ) ) return false;
  }

  return true;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::processBlobElement( const pool::ObjectRelationalMappingElement& element,
                                                                          bool evolve )
{
  // Chech if the table exists.
  const std::string& tableName = element.tableName();
  const std::string& columnName = element.columnNames()[0];
  std::string variableType = coral::AttributeSpecification::typeNameForId( typeid(coral::Blob) );
  if ( m_schema.existsTable( tableName ) ) {
    coral::ITable& table = m_schema.tableHandle( tableName );
    // Check if the column exists in the table
    try {
      const coral::IColumn& column = table.description().columnDescription( columnName );
      if ( ! column.isNotNull() && isTableEmpty(tableName)) {
        if ( ! evolve ) {
          seal::MessageStream log( pool::POOLContext::context(),
                                   pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "The column named \"" << columnName << "\" in table \""
              << tableName
              << "\" has been found without a NOT NULL constraint while processing a mapping element, and automatic schema evolution has been disabled."
              << seal::flush;
          return false;
        } else {
          std::ostringstream message;
          message << "Setting NOT NULL constraint on column \"" << columnName << "\" in table \"" << tableName << "\"";
          this->logOperation( message.str() );
          if(!m_dryRun) table.schemaEditor().setNotNullConstraint( columnName );
        }
      }
      if ( column.type() != variableType ) {
        if(isTableEmpty(tableName)){
          if ( ! evolve ) {
            seal::MessageStream log( pool::POOLContext::context(),
                                     pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
            log << seal::Msg::Error << "The column named \"" << columnName << "\" in table \""
                << tableName
                << "\" has been found without a BLOB variable while processing a mapping element, and automatic schema evolution has been disabled."
                << seal::flush;
            return false;
          } else {
            std::ostringstream message;
            message << "Changing type of column \"" << columnName << "\" in table \"" << tableName <<" to \"" << variableType << "\"";
            this->logOperation( message.str() );
            if(!m_dryRun) table.schemaEditor().changeColumnType( columnName, variableType );
          }
        } else {
          seal::MessageStream log( pool::POOLContext::context(),
                                     pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "The column named \"" << columnName << "\" in table \""
              << tableName
              << "\" has been found without a BLOB variable while processing a mapping element, and the table is not empty."
              << seal::flush;
          return false;
        }
      
      }
      
    }
    catch (const coral::InvalidColumnNameException&) {
      if ( ! evolve ) {
        seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
        log << seal::Msg::Error << "A column named \"" << columnName << "\" has not been found in table \""
            << tableName << "\" while processing a mapping element, and automatic schema evolution has been disabled." << seal::flush;
        return false;
      } else {
        std::ostringstream message;
        message << "Inserting column \"" << columnName << "\" of type \""<< variableType<<"\" in table \"" << tableName <<"\"";
        this->logOperation( message.str() );
        if(!m_dryRun) {
          table.schemaEditor().insertColumn( columnName,
                                             variableType );
        }
        if( isTableEmpty( tableName)){
          std::ostringstream message;
          message << "Setting NOT NULL constraint on column \"" << columnName <<"\" in table \"" << tableName <<"\"";
          this->logOperation( message.str() );
          if(!m_dryRun) table.schemaEditor().setNotNullConstraint( columnName );
        }
      }
    }
  } // Check if the view exists in the schema
  else if ( m_schema.existsView( tableName ) ) {
    coral::IView& view = m_schema.viewHandle( tableName );
    bool found = false;
    int numberOfColumns = view.numberOfColumns();
    for ( int iColumn = 0; iColumn < numberOfColumns; ++iColumn ) {
      const coral::IColumn& column = view.column( iColumn );
      if ( column.name() != columnName ) continue;
      found = true;
      if ( ! column.isNotNull() ) {
        seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
        log << seal::Msg::Error << "The column named \"" << columnName << "\" in view \""
            << tableName
            << "\" has been found without a NOT NULL constraint."
            << seal::flush;
        return false;
      }
      if ( column.type() != variableType ) {
        seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
        log << seal::Msg::Error << "The column named \"" << columnName << "\" in view \""
            << tableName
            << "\" has been found without a BLOB variable."
            << seal::flush;
        return false;
      }
      break;
    }
    if ( ! found ) {
      seal::MessageStream log( pool::POOLContext::context(),
                               pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
      log << seal::Msg::Error << "A column named \"" << columnName << "\" has not been found in view \""
          << tableName << "\" while processing a mapping element." << seal::flush;
      return false;
    }
  }
  else {
    coral::TableDescription& description = m_tableDescriptionStack->description( tableName );
    try {
      description.columnDescription(columnName);
    }
    catch (const coral::InvalidColumnNameException&) {
      description.insertColumn( columnName,
                                variableType );
      description.setNotNullConstraint( columnName );
    }
  }

  return true;
}

bool
pool::ObjectRelationalAccess::MappingMaterialization::processPrimitiveElement( const pool::ObjectRelationalMappingElement& element,
                                                                               bool evolve )
{
  // Chech if the table exists.
  const std::string& tableName = element.tableName();
  const std::string& columnName = element.columnNames()[0];
  const std::string variableType = element.variableType();
  if ( m_schema.existsTable( tableName ) ) {
    coral::ITable& table = m_schema.tableHandle( tableName );
    // Check if the column exists in the table
    try {      
      const coral::IColumn& column = table.description().columnDescription( columnName );
      if ( ! column.isNotNull() && isTableEmpty(tableName) ) {
        if ( ! evolve ) {
          seal::MessageStream log( pool::POOLContext::context(),
                                   pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "The column named \"" << columnName << "\" in table \""
              << tableName
              << "\" has been found without a NOT NULL constraint while processing a mapping element, and automatic schema evolution has been disabled."
              << seal::flush;
          return false;
        } else {
          std::ostringstream message;
          message << "Setting NOT NULL constraint on column \"" << columnName << "\" in table \"";
          this->logOperation( message.str() );
          if(!m_dryRun) table.schemaEditor().setNotNullConstraint( columnName );
        }
      }
    }
    catch (const coral::InvalidColumnNameException&) {
      if ( ! evolve ) {
        seal::MessageStream log( pool::POOLContext::context(),
                               pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
        log << seal::Msg::Error << "A column named \"" << columnName << "\" has not been found in table \""
            << tableName << "\" while processing a mapping element, and automatic schema evolution has been disabled." << seal::flush;
        return false;
      } else {
        std::ostringstream message;
        message << "Inserting column \"" << columnName << "\" of type \""<< variableType<<"\" in table \"" << tableName <<"\"";
        this->logOperation( message.str() );
        if(!m_dryRun){
          table.schemaEditor().insertColumn( columnName,
                                             variableType );
        }
        if( isTableEmpty( tableName )){
          std::ostringstream message;
          message << "Setting NOT NULL constraint on column \"" << columnName <<"\" in table \"" << tableName <<"\"";
          this->logOperation( message.str() );
         if(!m_dryRun) table.schemaEditor().setNotNullConstraint( columnName );
        }
      }
    }
  } // Check if the view exists in the schema
  else if ( m_schema.existsView( tableName ) ) {
    coral::IView& view = m_schema.viewHandle( tableName );
    bool found = false;
    int numberOfColumns = view.numberOfColumns();
    for ( int iColumn = 0; iColumn < numberOfColumns; ++iColumn ) {
      const coral::IColumn& column = view.column( iColumn );
      if ( column.name() != columnName ) continue;
      found = true;
      if ( ! column.isNotNull() ) {
        seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
        log << seal::Msg::Error << "The column named \"" << columnName << "\" in view \""
            << tableName
            << "\" has been found without a NOT NULL constraint."
            << seal::flush;
        return false;
      }
      break;
    }
    if ( ! found ) {
      seal::MessageStream log( pool::POOLContext::context(),
                               pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
      log << seal::Msg::Error << "A column named \"" << columnName << "\" has not been found in view \""
          << tableName << "\" while processing a mapping element." << seal::flush;
      return false;
    }
  }
  else {
    coral::TableDescription& description = m_tableDescriptionStack->description( tableName );
    try {
      description.columnDescription(columnName);
    }
    catch (const coral::InvalidColumnNameException&) {
      description.insertColumn( columnName,
                                variableType );
      description.setNotNullConstraint( columnName );
    }
  }

  return true;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::processArrayElement( const pool::ObjectRelationalMappingElement& parentElement,
                                                                           const pool::ObjectRelationalMappingElement& element,
                                                                           bool evolve )
{
  const std::string& parentTableName = parentElement.tableName();
  const std::string& tableName = element.tableName();
  // Retrieve the columns of the parent table.
  std::vector< std::string > parentIdColumns = parentElement.columnNames();
  if ( parentElement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    parentIdColumns.pop_back();
    parentIdColumns.pop_back();
  }

  const std::vector< std::string >& columnNames = element.columnNames();
  bool isRef = ( element.elementType() == pool::ObjectRelationalMappingElement::PoolReference );
  
  // To set up the array table.
  if ( ! ( this->processObject( tableName, columnNames, isRef, evolve ) ) ) return false;
    
  // Process the sub elements
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    const pool::ObjectRelationalMappingElement& subElement = iElement->second;
    if ( ! this->processSubElement( element, subElement, evolve ) ) return false;
  }

  // Make sure that there is a foreign key constraint id columns
  std::vector< std::string > idColumns( columnNames.size() - 1 );
  for ( size_t i = 0; i < idColumns.size(); ++i ) idColumns[i] = columnNames[i+1];
  
  // Reverse the columns.
  std::reverse( parentIdColumns.begin(), parentIdColumns.end() );
  std::reverse( idColumns.begin(), idColumns.end() );
  return this->checkOrSetForeignKey( tableName, idColumns, parentTableName, parentIdColumns, evolve );
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::processReferenceElement( const pool::ObjectRelationalMappingElement& element,
                                                                               bool evolve )
{
  if ( ! this->processObject( element.tableName(), element.columnNames(), true, evolve ) ) return false;
  // Process the sub elements
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    const pool::ObjectRelationalMappingElement& subElement = iElement->second;
    if ( ! this->processSubElement( element, subElement, evolve ) ) return false;
  }

  return true;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::processPointerElement( const pool::ObjectRelationalMappingElement& element,
                                                                             bool evolve )
{
  // Chech if the table exists.
  const std::string& tableName = element.tableName();
  for ( std::vector<std::string>::const_iterator iColumnName = element.columnNames().begin();
        iColumnName != element.columnNames().end(); ++iColumnName ) {
    const std::string& columnName = *iColumnName;
    if ( m_schema.existsTable( tableName ) ) {
      coral::ITable& table = m_schema.tableHandle( tableName );
      // Check if the column exists in the table
      try
      {
        const coral::IColumn& column = table.description().columnDescription( columnName );
        if ( ! column.isNotNull() && isTableEmpty( tableName ) ) {
          if ( ! evolve ) {
            seal::MessageStream log( pool::POOLContext::context(),
                                     pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
            log << seal::Msg::Error << "The column named \"" << columnName << "\" in table \""
                << tableName
                << "\" has been found without a NOT NULL constraint while processing a mapping element, and automatic schema evolution has been disabled."
                << seal::flush;
            return false;
          } else {
            std::ostringstream message;
            message << "Setting NOT NULL constraint on column  \"" << columnName << "\" in table \"";
            this->logOperation( message.str() );
            if(!m_dryRun) table.schemaEditor().setNotNullConstraint( columnName );
          }
        }
      } catch (const coral::InvalidColumnNameException&) {
        if ( ! evolve ) {
          seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "A column named \"" << columnName << "\" has not been found in table \""
              << tableName << "\" while processing a mapping element, and automatic schema evolution has been disabled." << seal::flush;
          return false;
        }
        else {
          std::ostringstream message;
          message << "Inserting column \"" << columnName << "\" of type \""
                  << coral::AttributeSpecification::typeNameForType<std::string>()
                  <<"\" in table \"" << tableName <<"\"";
          this->logOperation( message.str() );
          if(!m_dryRun){
            table.schemaEditor().insertColumn( columnName,
                                               coral::AttributeSpecification::typeNameForType<std::string>() );
          }
          if( isTableEmpty( tableName )){
            std::ostringstream message;
            message << "Setting NOT NULL constraint on column \"" << columnName <<"\" in table \"" << tableName <<"\"";
            this->logOperation( message.str() );
            if(!m_dryRun)table.schemaEditor().setNotNullConstraint( columnName );
          }
        }
      }
    }
    else if ( m_schema.existsView( tableName ) ) {
      coral::IView& view = m_schema.viewHandle( tableName );
      bool found = false;
      int numberOfColumns = view.numberOfColumns();
      for ( int iColumn = 0; iColumn < numberOfColumns; ++iColumn ) {
        const coral::IColumn& column = view.column( iColumn );
        if ( column.name() != columnName ) continue;
        found = true;
        if ( ! column.isNotNull() ) {
          seal::MessageStream log( pool::POOLContext::context(),
                                   pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "The column named \"" << columnName << "\" in view \""
              << tableName
              << "\" has been found without a NOT NULL constraint."
              << seal::flush;
          return false;
        }
        break;
      }
      if ( ! found ) {
        seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
        log << seal::Msg::Error << "A column named \"" << columnName << "\" has not been found in view \""
            << tableName << "\" while processing a mapping element." << seal::flush;
        return false;
      }
    }
    else {
      coral::TableDescription& description = m_tableDescriptionStack->description( tableName );
      try
      {
        description.columnDescription(columnName);
      } catch (const coral::InvalidColumnNameException& )
      {
        description.insertColumn( columnName,
                                  coral::AttributeSpecification::typeNameForType<std::string>() );
        description.setNotNullConstraint( columnName );
      }
    }
  }
  return true;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::processObject( const std::string& tableName,
                                                                     const std::vector< std::string >& columnNames,
                                                                     bool isRef,
                                                                     bool evolve )
{
  // Check if the table exists.
  if ( m_schema.existsTable( tableName ) ) {
    coral::ITable& table = m_schema.tableHandle( tableName );
    
    bool allColumnsExist = true;
    size_t idCol = 0;
    for ( std::vector< std::string >::const_iterator iColumnName = columnNames.begin();
          iColumnName != columnNames.end(); ++iColumnName, ++idCol ) {
      try
      {
        const coral::IColumn&  column = table.description().columnDescription(*iColumnName);  
        if ( ! column.isNotNull() && isTableEmpty( tableName ) ) {
          if ( ! evolve ) {
            seal::MessageStream log( pool::POOLContext::context(),
                                   pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
            log << seal::Msg::Error << "The column named \"" << *iColumnName << "\" in table \"" << tableName
                << "\" has been found without a NOT NULL constraint while processing a mapping element, and automatic schema evolution has been disabled."
                << seal::flush;
            return false;
          }
          else {
            std::ostringstream message;
            message << "Setting NOT NULL constraint on column \"" << *iColumnName << "\" in table \"" << tableName << "\"";
            this->logOperation( message.str() );
            if(!m_dryRun) table.schemaEditor().setNotNullConstraint( *iColumnName );
          }
        }
      }catch (const coral::InvalidColumnNameException&)
      {
        if ( ! evolve ) {
          seal::MessageStream log( pool::POOLContext::context(),
                                   pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "A column named \"" << *iColumnName << "\" has not been found in table \""
              << tableName << "\" while processing a mapping element, and automatic schema evolution has been disabled." << seal::flush;
          return false;
        }
        else {
          std::ostringstream message;
          message << "Inserting column \"" << *iColumnName << "\" of type \""
                  << coral::AttributeSpecification::typeNameForType<unsigned long>()
                  <<"\" in table \"" << tableName <<"\"";
          this->logOperation( message.str() );
          if(!m_dryRun) {
            table.schemaEditor().insertColumn( *iColumnName,
                                               coral::AttributeSpecification::typeNameForType<unsigned long>());
            // an integer type for the identity column by default
          }
          if( isTableEmpty( tableName)){
            std::ostringstream message;
            message << "Setting NOT NULL constraint on column \"" << *iColumnName <<"\" in table \"" << tableName <<"\"";
            this->logOperation( message.str() );
            if(!m_dryRun) table.schemaEditor().setNotNullConstraint( *iColumnName );
          }
          if ( ( ! isRef ) || ( isRef && idCol < columnNames.size() - 2 ) ) allColumnsExist = false;
        }        
      }
    }

    std::vector<std::string> columnsForIndex = columnNames;
    if ( isRef ) {
      columnsForIndex.pop_back();
      columnsForIndex.pop_back();
    }

    // Reverse the order of the columns
    std::reverse( columnsForIndex.begin(), columnsForIndex.end() );

    if ( ! allColumnsExist ) { // define a primary key or a unique index
      if ( ! table.description().hasPrimaryKey() ) {
        std::ostringstream message;
        message << "Setting PRIMARY KEY constraint on columns ";
        size_t cols = columnsForIndex.size();
        unsigned int i=0;
        while(i<cols) {
          message << "\""<< columnsForIndex[i] << "\"";
          if(i<cols-1) message<<",";
          i++;
        }
        message <<"\" in table \"" << tableName <<"\"";
        this->logOperation( message.str() );
        if(!m_dryRun) table.schemaEditor().setPrimaryKey( columnsForIndex );
      } else {
        std::ostringstream message;
        message << "Creating INDEX \""<<pool::ObjectRelationalAccess::ObjectRelationalNames::indexNameForIdentity( tableName )<<"\""
            << " on columns ";
        size_t cols = columnsForIndex.size();
        unsigned int i=0;
        while(i<cols) {
          message << "\""<< columnsForIndex[i] << "\"";
          if(i<cols-1) message<<",";
          i++;
        }
        message <<"\" in table \"" << tableName <<"\"";
        this->logOperation( message.str() );
        if(!m_dryRun) table.schemaEditor().createIndex( pool::ObjectRelationalAccess::ObjectRelationalNames::indexNameForIdentity( tableName ),
                                                        columnsForIndex, true );
      }
    }
    else {
      if ( !( this->existsUniqueIndex( table.description(), columnsForIndex ) ) ) {
        if ( ! evolve ) {
          seal::MessageStream log( pool::POOLContext::context(),
                                   pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
          log << seal::Msg::Error << "A proper unique index has not been found in table \""
              << tableName << "\" while processing a mapping element, and automatic schema evolution has been disabled." << seal::flush;
          return false;
        } else {
          if ( ! table.description().hasPrimaryKey() ) {
            std::ostringstream message;
            message << "Setting PRIMARY KEY constraint on columns ";
            size_t cols = columnsForIndex.size();
            unsigned int i=0;
            while(i<cols) {
              message << "\""<< columnsForIndex[i] << "\"";
              if(i<cols-1) message<<",";
              i++;
            }
            message <<"\" in table \"" << tableName <<"\"";
            this->logOperation( message.str() );
            if(!m_dryRun) table.schemaEditor().setPrimaryKey( columnsForIndex );
          } else {
            std::ostringstream message;
            message << "Creating INDEX \""<<pool::ObjectRelationalAccess::ObjectRelationalNames::indexNameForIdentity( tableName )<<"\""
                    << " on columns ";
            size_t cols = columnsForIndex.size();
            unsigned int i=0;
            while(i<cols) {
              message << "\""<< columnsForIndex[i] << "\"";
              if(i<cols-1) message<<",";
              i++;
            }
            message <<"\" in table \"" << tableName <<"\"";
            this->logOperation( message.str() );
            if(!m_dryRun) table.schemaEditor().createIndex( pool::ObjectRelationalAccess::ObjectRelationalNames::indexNameForIdentity( tableName ),
                                                            columnsForIndex, true );
          }
        }
      }
    }
  }
  else if ( m_schema.existsView( tableName ) ) { // Check if a corresponding view exists instead.

    coral::IView& view = m_schema.viewHandle( tableName );
    std::map< std::string, int > columnNameToIndex;
    int numberOfColumns = view.numberOfColumns();
    for ( int iColumn = 0; iColumn < numberOfColumns; ++iColumn ) {
      const coral::IColumn& column = view.column( iColumn );
      columnNameToIndex.insert( std::make_pair( column.name(), iColumn ) );
    }
    for ( std::vector< std::string >::const_iterator iColumnName = columnNames.begin();
          iColumnName != columnNames.end(); ++iColumnName ) {
      std::map< std::string, int >::const_iterator iColumnIndex = columnNameToIndex.find( *iColumnName );
      seal::MessageStream log( pool::POOLContext::context(),
                                 pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
      if ( iColumnIndex == columnNameToIndex.end() ) {
        log << seal::Msg::Error << "A column named \"" << *iColumnName << "\" has not been found in view \""
            << tableName << "\" while processing a mapping element." << seal::flush;
        return false;
      } else {
        if ( ! view.column( iColumnIndex->second ).isNotNull() ) {
          log << seal::Msg::Error << "The column named \"" << *iColumnName << "\" in view \"" << tableName
              << "\" has been found without a NOT NULL constraint while processing a mapping element."
              << seal::flush;
          return false;
        }
      }
    }

    // No check for index in this case...

  }
  else {
    coral::TableDescription& description = m_tableDescriptionStack->description( tableName );
    for ( std::vector< std::string >::const_iterator iColumnName = columnNames.begin();
          iColumnName != columnNames.end(); ++iColumnName ) {
      try
      {
        description.columnDescription(*iColumnName);
      }
      catch(const coral::InvalidColumnNameException&)
      {
        description.insertColumn( *iColumnName,
                                  coral::AttributeSpecification::typeNameForType<unsigned long>() ); // an integer type for the identity column by default
        description.setNotNullConstraint( *iColumnName );
      }
    }

    std::vector<std::string> columnsForIndex = columnNames;
    if ( isRef ) {
      columnsForIndex.pop_back();
      columnsForIndex.pop_back();
    }
    
    // Reverse the order of the columns
    std::reverse( columnsForIndex.begin(), columnsForIndex.end() );

    if ( !( this->existsUniqueIndex( description, columnsForIndex ) ) ) {
      if ( ! description.hasPrimaryKey() ) {
        description.setPrimaryKey( columnsForIndex );
      }
      else {
        description.createIndex( pool::ObjectRelationalAccess::ObjectRelationalNames::indexNameForIdentity( tableName ),
                                 columnsForIndex, true );
      }
    }
  }

  return true;
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::checkOrSetForeignKey( const std::string& tableName,
                                                                            const std::vector< std::string >& columnNames,
                                                                            const std::string& referencedTable,
                                                                            const std::vector< std::string >& referencedColumnNames,
                                                                            bool evolve )
{
  std::string fkName = pool::ObjectRelationalAccess::ObjectRelationalNames::fkNameForIdentity( tableName );
  if ( m_schema.existsTable( tableName ) ) {
    if( !m_schema.existsTable( referencedTable )){
      // the processing has to be re-run after the creation of all of the tables
      m_mappingToProcess = true;
      return true;
    }
    
    const coral::ITableDescription& description = m_schema.tableHandle( tableName ).description();
    
    if ( this->existsForeignKey( description, columnNames, referencedTable, referencedColumnNames ) ) return true;

    if ( ! evolve ) {
      seal::MessageStream log( pool::POOLContext::context(),
                               pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
      log << seal::Msg::Error
          << "A proper foreign key has not been found in table \"" << tableName
          << "\" referencing table \"" << referencedTable
          << "\" while processing a mapping element, and automatic schema evolution has been disabled." << seal::flush;
      return false;
    }

    // examine the existing foreign keys to drop the current ones
    int numberOfForeignKeys = description.numberOfForeignKeys();
    std::set<std::string> fkeys;
    for ( int iForeignKey = 0; iForeignKey < numberOfForeignKeys; ++iForeignKey ) {
      const coral::IForeignKey& foreignKey = description.foreignKey( iForeignKey );
      const std::vector< std::string >& fkColumns = foreignKey.columnNames();
      for(std::vector<std::string>::const_iterator ifk=fkColumns.begin();
            ifk!=fkColumns.end();++ifk){
        bool found = false;
        for(std::vector<std::string>::const_iterator iC=columnNames.begin();
              iC!=columnNames.end()  && !found ;++iC){
          if(*ifk==*iC) found = true;
        }
        if(found) fkeys.insert(foreignKey.name());
      }
      
    }
    for(std::set<std::string>::const_iterator ifk=fkeys.begin();
        ifk!=fkeys.end();++ifk){
      std::ostringstream message;
      message << "Dropping FOREIGN KEY constraint \""<<*ifk<<"\"";
      this->logOperation( message.str() );
      if(!m_dryRun) m_schema.tableHandle( tableName ).schemaEditor().dropForeignKey( *ifk );
    }
    

  try{
      std::ostringstream message;
      message << "Setting FOREIGN KEY constraint \""<<fkName<<"\" on columns ";
      size_t cols = columnNames.size();
      unsigned int i=0;
      while(i<cols) {
        message << "\""<< columnNames[i] << "\"";
        if(i<cols-1) message<<",";
        i++;
      }
      message <<"\" in table \"" << tableName <<"\""
              << " referencing columns ";
      cols = referencedColumnNames.size();
      i=0;
      while(i<cols) {
        message << "\""<< referencedColumnNames[i] << "\"";
        if(i<cols-1) message<<",";
        i++;
      }
      message <<"\" in table \"" << referencedTable << "\"";
      this->logOperation( message.str() );
      if(!m_dryRun) m_schema.tableHandle( tableName ).schemaEditor().createForeignKey( fkName, columnNames, referencedTable, referencedColumnNames );
      return true;
    }catch (const coral::InvalidForeignKeyIdentifierException& e)
    {
      seal::MessageStream log( pool::POOLContext::context(),
                               pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
      log << seal::Msg::Error << e.what()<<std::endl;
      return false;
    }
  }
  else if ( m_schema.existsView( tableName ) ) {
    return true;
  }
  else {
    if ( m_schema.existsView( referencedTable ) )
      return true;
    coral::TableDescription& description = m_tableDescriptionStack->description( tableName );
    if ( this->existsForeignKey( description, columnNames, referencedTable, referencedColumnNames ) ) return true;
    else {
      try
      {
        description.createForeignKey( fkName, columnNames, referencedTable, referencedColumnNames );
        return true;
      }
      catch (const coral::InvalidForeignKeyIdentifierException&)
      {
        return false;
      }
    }
  }
}


bool
pool::ObjectRelationalAccess::MappingMaterialization::existsForeignKey( const coral::ITableDescription& description,
                                                                        const std::vector< std::string >& columnNames,
                                                                        const std::string& referencedTableName,
                                                                        const std::vector< std::string >& referencedColumnNames )
{
  int numberOfForeignKeys = description.numberOfForeignKeys();
  for ( int iForeignKey = 0; iForeignKey < numberOfForeignKeys; ++iForeignKey ) {
    const coral::IForeignKey& foreignKey = description.foreignKey( iForeignKey );
    if ( foreignKey.referencedTableName() == referencedTableName ) {
      const std::vector< std::string >& fkColumns = foreignKey.columnNames();
      const std::vector< std::string >& fkReferencedColumns = foreignKey.referencedColumnNames();
      if ( columnNames.size() == fkColumns.size() && referencedColumnNames.size() == fkReferencedColumns.size() ) {
        bool same = true;
        for ( size_t i = 0; i < columnNames.size(); ++i ) {
          if ( columnNames[i] != fkColumns[i] ) {
            same = false;
            break;
          }
        }
        if ( ! same ) continue;
        for ( size_t i = 0; i < referencedColumnNames.size(); ++i ) {
          if ( referencedColumnNames[i] != fkReferencedColumns[i] ) {
            same = false;
            break;
          }
        }
        if ( same ) {
          return true;
        }
      }
    }
  }
  return false;
}

void
pool::ObjectRelationalAccess::MappingMaterialization::logOperation(const std::string& message){
  if(!m_dryRun) {
    seal::MessageStream log( pool::POOLContext::context(),
                             pool::ObjectRelationalAccess::ObjectRelationalNames::moduleName() );
    log << seal::Msg::Warning << message << seal::flush;
  } else {
    m_operationList.insert( message );
  }
}

