#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalElementTypeNames.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "CoralBase/AttributeSpecification.h"
#include "POOLCore/PoolMessageStream.h"
#include "POOLCore/Reference.h"
#include "Reflex/Reflex.h"
#include <sstream>

pool::ObjectRelationalDefaultMappingBuilder::ObjectRelationalDefaultMappingBuilder():
  m_tablesInvolved(),
  m_log( *(new pool::PoolMessageStream("ObjectRelationalDefaultMappingBuilder")) ),
  m_arraysInBlobs( false )
{}

pool::ObjectRelationalDefaultMappingBuilder::~ObjectRelationalDefaultMappingBuilder()
{
  delete &m_log;
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processClass( pool::ObjectRelationalMapping& mapping,
                                                           const ROOT::Reflex::Type& classDictionary,
                                                           bool arraysInBlobs )
{
  m_arraysInBlobs = arraysInBlobs;

  std::string className = classDictionary.Name(ROOT::Reflex::SCOPED);
  size_t sz = sizeInColumns( classDictionary );
  m_log << seal::Msg::Info << "Class  \""<< className << "\" requires "<<sz<<" columns."<< seal::flush;
  if(sz > MaxColumnsPerTable){
    m_log << seal::Msg::Error << "Cannot process default mapping for class \""<<className<<"\" to the relational schema: number of columns ("<<sz<<") required exceedes maximum="<< MaxColumnsPerTable << seal::flush;
    return false;    
  }

  std::string tableName = pool::ObjectRelationalDefaultMappingRules::tableNameForClass( className );
  m_tablesInvolved.push_back( tableName );
  m_log << seal::Msg::Debug << "Processing mapping for class \""<< className << "\"." << seal::flush;

  // Define the top level element
  pool::ObjectRelationalMappingElement& classElement = mapping.appendElement( className, tableName );

  // Set the id of the class
  classElement.setColumnNames( std::vector< std::string >( 1, pool::ObjectRelationalDefaultMappingRules::columnNameForId() ) );

  
  return this->processObject( classElement, classDictionary );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::buildCArrayElementTree( ObjectRelationalMappingElement& topElement ){
  if( topElement.elementType()!=ObjectRelationalMappingElement::CArray ){
    return false;
  }
  ROOT::Reflex::Type elementType = ROOT::Reflex::Type::ByName(topElement.variableType());
  elementType = ObjectRelationalClassUtils::resolvedType( elementType );
  ROOT::Reflex::Type arrayElementType = elementType.ToType();
  std::string contentTypeName = arrayElementType.Name();
  return processItem( topElement, contentTypeName, arrayElementType);
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processPrimitive( pool::ObjectRelationalMappingElement& parentelement,
                                                               const std::string& attributeName,
                                                               const ROOT::Reflex::Type& attributeType )
{
  const std::type_info* attrType = &attributeType.TypeInfo();
  if(attributeType.IsEnum()) attrType = &typeid(int);
  std::string typeName = coral::AttributeSpecification::typeNameForId(*attrType);
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( pool::ObjectRelationalElementTypeNames::primitiveMappingElementType(), attributeName, typeName, parentelement.tableName() );
  me.setColumnNames( std::vector< std::string >( 1, pool::ObjectRelationalDefaultMappingRules::columnNameForVariable( attributeName, me.scopeForColumnName() ) ) );
  return true;
}



bool
pool::ObjectRelationalDefaultMappingBuilder::processBlob( pool::ObjectRelationalMappingElement& parentelement,
                                                          const std::string& attributeName,
                                                          const ROOT::Reflex::Type& attributeType )
{
  std::string className = attributeType.Name(ROOT::Reflex::SCOPED);
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( pool::ObjectRelationalElementTypeNames::blobMappingElementType(),
                                                                             attributeName, className, parentelement.tableName() );
  me.setColumnNames( std::vector< std::string >( 1, pool::ObjectRelationalDefaultMappingRules::columnNameForVariable( attributeName, me.scopeForColumnName() ) ) );
  return true;
}

bool 
pool::ObjectRelationalDefaultMappingBuilder::processEmbeddedClass( pool::ObjectRelationalMappingElement& parentelement,
                                                                   const std::string& attributeName,
                                                                   const ROOT::Reflex::Type& attributeType,
                                                                   bool isRef )
{
  std::string className = attributeType.Name(ROOT::Reflex::SCOPED);
  std::string elementType = ( isRef ) ? pool::ObjectRelationalElementTypeNames::poolReferenceMappingElementType() : pool::ObjectRelationalElementTypeNames::objectMappingElementType();
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( elementType, attributeName, className, parentelement.tableName() );

  std::vector< std::string > idColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    idColumns.pop_back();
    idColumns.pop_back();
  }
  if ( isRef ) {
    idColumns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForOID_1( attributeName, me.scopeName() ) );
    idColumns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForOID_2( attributeName, me.scopeName() ) );
  }

  me.setColumnNames( idColumns );

  return this->processObject( me, attributeType );
}


bool
pool::ObjectRelationalDefaultMappingBuilder::processObject( pool::ObjectRelationalMappingElement& mappingElement,
                                                            const ROOT::Reflex::Type& objType )
{
  std::string className = objType.Name(ROOT::Reflex::SCOPED);
  std::string tableName = mappingElement.tableName();
  
  // resolve possible typedef chains
  ROOT::Reflex::Type objectType = ObjectRelationalClassUtils::resolvedType(objType);
  // loop over the data members (including the ones from the base classes)
  objectType.UpdateMembers();
  for ( size_t i=0; i< objectType.DataMemberSize(); i++){

    ROOT::Reflex::Member objectMember = objectType.DataMemberAt(i);
    // Skip the transient ones
    if ( objectMember.IsTransient() ) continue;

    // Retrieve the field type
    ROOT::Reflex::Type type = ObjectRelationalClassUtils::resolvedType( objectMember.TypeOf() );
    
    // check if the member is from a class in the inheritance tree
    std::string scope("");
    ROOT::Reflex::Type declaringType = ObjectRelationalClassUtils::resolvedType( objectMember.DeclaringType());
    if( declaringType != objectType ){
      scope = declaringType.Name(ROOT::Reflex::SCOPED);
    }
    // Retrieve the field name
    std::string objectMemberName = pool::ObjectRelationalMappingElement::scopedVariableName( objectMember.Name(), scope );
    
    // Check for the existence of the dictionary information
    if ( !type ){
      m_log << seal::Msg::Error << "Field \"" <<objectMemberName << "\" in class \""<<className <<"\" is unknown type"<< seal::flush;
      return false;
    }

    if ( ! this->processItem( mappingElement, objectMemberName, type ) ) return false;
  }
  return true;
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processItem( pool::ObjectRelationalMappingElement& parentElement,
                                                          const std::string& attributeName,
                                                          const ROOT::Reflex::Type& attributeType )
{

  ROOT::Reflex::Type resType = ObjectRelationalClassUtils::resolvedType( attributeType );
  // primitive and string
  if ( resType.IsFundamental() || resType.IsEnum() || resType.TypeInfo() == typeid(std::string) ) {
    // Correct the string type name
    return this->processPrimitive( parentElement,
                                   attributeName,
                                   resType );
  }
  else if ( resType.IsArray() ){
    
    if ( m_arraysInBlobs ){
      return this->processBlob( parentElement,
                                attributeName,
                                resType );
    }
    return this->processCArray( parentElement,
                                attributeName,
                                resType);
  }
  else if ( pool::ObjectRelationalClassUtils::isTypeContainer( resType ) ) {

    if ( m_arraysInBlobs ){
      return this->processBlob( parentElement,
                                attributeName,
                                resType );
    }
    return this->processArray( parentElement,
                               attributeName,
                               resType);
  }
  else if ( resType.IsPointer() || resType.IsReference() ){
    m_log << seal::Msg::Error << "Pointers or references are not supported" << seal::flush;
    return false;
  }
  else { // embeddedobject
    bool isRef = ( resType.TypeInfo() == typeid(pool::Reference) ||
                   resType.HasBase( ROOT::Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) ) );
    return this->processEmbeddedClass( parentElement,
                                       attributeName,
                                       attributeType,
                                       isRef );
  }
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processArray( pool::ObjectRelationalMappingElement& parentelement,
                                                           const std::string& attributeName,
                                                           const ROOT::Reflex::Type& attributeType ){
  std::ostringstream scopedAttributeName;
  if( parentelement.scopeName()!=""){
    scopedAttributeName << parentelement.variableName() << "::";
  }
  scopedAttributeName  << attributeName;
  std::string arrayTable = pool::ObjectRelationalDefaultMappingRules::tableNameForClass( scopedAttributeName.str(), parentelement.tableName() );
  m_tablesInvolved.push_back( arrayTable );
  
  std::string className = attributeType.Name(ROOT::Reflex::SCOPED);

  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( pool::ObjectRelationalElementTypeNames::arrayMappingElementType(),
                                                                             attributeName, className, arrayTable );
  std::vector<std::string> parentColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    parentColumns.pop_back();
    parentColumns.pop_back();
  }
  std::vector<std::string> columns;
  columns.reserve( parentColumns.size() + 1 );
  columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForPosition() );
  for ( std::vector<std::string>::const_iterator iColumn = parentColumns.begin();
        iColumn != parentColumns.end(); ++iColumn ) {
    columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForId() + "_" + *iColumn );
  }
  me.setColumnNames( columns );
  
  bool singleItemContainer =  pool::ObjectRelationalClassUtils::isTypeNonAssociativeContainer(attributeType);
  bool associativeContainer =  pool::ObjectRelationalClassUtils::isTypeAssociativeContainer(attributeType);

  ROOT::Reflex::Type contentType;
  ROOT::Reflex::Type keyType;
  
  if( singleItemContainer ){
    contentType = ObjectRelationalClassUtils::containerValueType(attributeType);
  }
  else if ( associativeContainer ) { // This is an associative container type
    contentType = ObjectRelationalClassUtils::containerDataType( attributeType );
    keyType = ObjectRelationalClassUtils::containerKeyType( attributeType );
    if( !keyType || !ObjectRelationalClassUtils::resolvedType(keyType) ){
      m_log << seal::Msg::Error << "cannot decide the type of the key item of the container \"" << attributeType.Name(ROOT::Reflex::SCOPED) << "\"." << seal::flush;
      return false;
    }
  }
  else { // Not supported container
    m_log << seal::Msg::Error << "Unsupported container type=\""<<attributeType.Name(ROOT::Reflex::SCOPED) <<"\"."<< seal::flush;
    return false;
  }

  if( !contentType || !ObjectRelationalClassUtils::resolvedType(contentType) ){
    m_log << seal::Msg::Error << "cannot decide the type of the content item of the container \"" << attributeType.Name(ROOT::Reflex::SCOPED) << "\"." << seal::flush;
    return false;
  }
  if ( keyType ) {
    std::string keyTypeName = keyType.Name();
    if ( !( this->processItem( me, keyTypeName, keyType ) ) )
      return false;
  }
  std::string contentTypeName = contentType.Name();
  return this->processItem( me, contentTypeName, contentType );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processCArray( ObjectRelationalMappingElement& parentelement,
                                                            const std::string& attributeName,
                                                            const ROOT::Reflex::Type& attributeType ){
  ROOT::Reflex::Type arrayElementType = attributeType.ToType();
  if( !arrayElementType || !ObjectRelationalClassUtils::resolvedType( arrayElementType ) ){
    m_log << seal::Msg::Error << "cannot decide the type of the content of the array \"" << attributeType.Name(ROOT::Reflex::SCOPED) << "\"." << seal::flush;
    return false;
  }
  size_t arraySizeInColumns = attributeType.ArrayLength();
  if( arraySizeInColumns <= DefaultSizeForInlineCArrays ) {
    arraySizeInColumns = sizeInColumns( attributeType );
    if( arraySizeInColumns <= MaxSizeForInlineCArrays ){
      m_log << seal::Msg::Debug << "Attribute \""<< parentelement.variableType()<<"::"<<attributeName<<"\" requires "<<arraySizeInColumns<<" column(s)."<<seal::flush;
      return processInlineCArrayItem(parentelement,attributeName,attributeType, arrayElementType);
    }
  }
  m_log << seal::Msg::Debug << "Attribute \""<< parentelement.variableType()<<"::"<<attributeName<<"\" requires "<<arraySizeInColumns<<" column(s)."<<seal::flush;
  return processCArrayItem(parentelement,attributeName,attributeType,arrayElementType );
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processCArrayItem( ObjectRelationalMappingElement& parentelement,
                                                                const std::string& attributeName,
                                                                const ROOT::Reflex::Type& attributeType,
                                                                const ROOT::Reflex::Type& arrayElementType ){
  std::string className = attributeType.Name(ROOT::Reflex::SCOPED);
  std::ostringstream scopedAttributeName;
  if( parentelement.scopeName()!=""){
    scopedAttributeName << parentelement.variableName() << "::";
  }
  scopedAttributeName  << attributeName;
  std::string arrayTable = pool::ObjectRelationalDefaultMappingRules::tableNameForClass( scopedAttributeName.str(), parentelement.tableName() );
  m_tablesInvolved.push_back( arrayTable );
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( pool::ObjectRelationalElementTypeNames::CArrayMappingElementType(),
                                                                             attributeName, className, arrayTable );
  std::vector<std::string> parentColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    parentColumns.pop_back();
    parentColumns.pop_back();
  }
  std::vector<std::string> columns;
  columns.reserve( parentColumns.size() + 1 );
  columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForPosition() );
  for ( std::vector<std::string>::const_iterator iColumn = parentColumns.begin();
        iColumn != parentColumns.end(); ++iColumn ) {
    columns.push_back( pool::ObjectRelationalDefaultMappingRules::columnNameForId() + "_" + *iColumn );
  }
  me.setColumnNames( columns );
  std::string contentTypeName = arrayElementType.Name();
  return this->processItem( me, contentTypeName, arrayElementType );  
}

bool
pool::ObjectRelationalDefaultMappingBuilder::processInlineCArrayItem( ObjectRelationalMappingElement& parentelement,
                                                                      const std::string& attributeName,
                                                                      const ROOT::Reflex::Type& attributeType,
                                                                      const ROOT::Reflex::Type& arrayElementType ){
  std::string className = attributeType.Name(ROOT::Reflex::SCOPED);
  std::string mappingElementType = pool::ObjectRelationalElementTypeNames::inlineCArrayMappingElementType();
  pool::ObjectRelationalMappingElement& me = parentelement.appendSubElement( mappingElementType, attributeName, className, parentelement.tableName() );
  std::vector< std::string > idColumns = parentelement.columnNames();
  if ( parentelement.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
    idColumns.pop_back();
    idColumns.pop_back();
  }
  me.setColumnNames( idColumns );
  for(size_t i=0;i<attributeType.ArrayLength();i++){
    if ( ! this->processItem( me, ObjectRelationalMappingElement::variableNameForArray(attributeName,i), arrayElementType ) ) return false;
  }
  return true;  
}

const std::vector<std::string>&
pool::ObjectRelationalDefaultMappingBuilder::tablesInvolved() const
{
  return m_tablesInvolved;
}

/*still missing or not tested:
  -pointer
  -reference
*/

size_t pool::ObjectRelationalDefaultMappingBuilder::MaxColumnsPerTable = 101;

size_t pool::ObjectRelationalDefaultMappingBuilder::MaxSizeForInlineCArrays = 100;

size_t pool::ObjectRelationalDefaultMappingBuilder::DefaultSizeForInlineCArrays = 10;

size_t
pool::ObjectRelationalDefaultMappingBuilder::sizeInColumns(const ROOT::Reflex::Type& topLevelClassType){
  size_t sz=0;
  _sizeInColumns(topLevelClassType,sz);
  return sz;
}

void
pool::ObjectRelationalDefaultMappingBuilder::_sizeInColumns(const ROOT::Reflex::Type& topLevelClassType,
                                                            size_t& sz){

  // resolve possible typedef chains
  ROOT::Reflex::Type typ = ObjectRelationalClassUtils::resolvedType( topLevelClassType );

  // primitive and string
  if ( typ.IsFundamental() || typ.TypeInfo() == typeid(std::string) || typ.IsEnum() ) {
    ++sz;
  } else if (typ.IsArray()){
    size_t arraySize = typ.ArrayLength();
    if(arraySize <= DefaultSizeForInlineCArrays){
      ROOT::Reflex::Type arrayType = typ.ToType();
      size_t arrayElementSize = 0;
      _sizeInColumns(arrayType, arrayElementSize);
      if( arrayElementSize <= MaxSizeForInlineCArrays) sz += arraySize*arrayElementSize;
    }   
  } else {
  
    bool isContainer =  pool::ObjectRelationalClassUtils::isTypeNonAssociativeContainer(typ) ||
      pool::ObjectRelationalClassUtils::isTypeAssociativeContainer(typ);
    if( !isContainer ){
      
      // loop over the data members
      typ.UpdateMembers();
      for ( size_t i=0; i< typ.DataMemberSize(); i++){
        ROOT::Reflex::Member objMember = typ.DataMemberAt(i);

        // Skip the transient ones
        if ( objMember.IsTransient() ) continue;

        // Retrieve the field type
        ROOT::Reflex::Type objMemberType = objMember.TypeOf();

        _sizeInColumns(objMemberType,sz);
      }
    }
  }
}

