#include "ObjectRelationalAccess/ObjectRelationalUtilities.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "POOLCore/POOLContext.h"
#include "StringTokenize.h"
#include "SealKernel/Context.h"
#include "PluginManager/ModuleDef.h"
#include "SealKernel/MessageStream.h"
#include "SealBase/SharedLibrary.h"
#include "SealBase/SharedLibraryError.h"
#include "SealKernel/ComponentFactory.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/IConnectionService.h"
#include "CoralCommon/AuthenticationCredentials.h"
#include <stdexcept>

DEFINE_SEAL_COMPONENT ( pool::ObjectRelationalUtilities::LocalAuthenticationService, "POOL/Services/ObjectRelationalUtilities/LocalAuthenticationService")
  
pool::ObjectRelationalUtilities::LocalAuthenticationService::LocalAuthenticationService( seal::Context* c )
  : seal::Service( c, classContextKey() ),
    m_credentials( 0 )
{
  m_credentials = new coral::AuthenticationCredentials( this->name() );
}

pool::ObjectRelationalUtilities::LocalAuthenticationService::LocalAuthenticationService( seal::Context* c,
                                                                                          const std::string& key )
  : seal::Service( c, key ),
    m_credentials( 0 )
{
  m_credentials = new coral::AuthenticationCredentials( this->name() );
}

pool::ObjectRelationalUtilities::LocalAuthenticationService::~LocalAuthenticationService()
{
  if ( m_credentials ) delete m_credentials;
}


const coral::IAuthenticationCredentials&
pool::ObjectRelationalUtilities::LocalAuthenticationService::credentials( const std::string& ) const
{
  return *m_credentials;
}

const coral::IAuthenticationCredentials&
pool::ObjectRelationalUtilities::LocalAuthenticationService::credentials( const std::string&, const std::string& ) const
{
  return *m_credentials;
}

void pool::ObjectRelationalUtilities::LocalAuthenticationService::setCredentials(const std::string& userName, const std::string& password)
{
  m_credentials->registerItem( "user", std::string( userName ) );
  m_credentials->registerItem( "password", std::string( password ) );
}


pool::ObjectRelationalUtilities::ObjectRelationalUtilities() : m_properties(),m_flags(),m_session(0),m_sessionOwned(false),m_verbosityLevel(seal::Msg::Warning){ 
  pool::POOLContext::loadComponent( "SEAL/Services/MessageService" );
  pool::POOLContext::setMessageVerbosityLevel(m_verbosityLevel);
}

pool::ObjectRelationalUtilities::ObjectRelationalUtilities(coral::ISessionProxy* session): m_properties(),m_flags(),m_session(session),m_sessionOwned(false){

}

pool::ObjectRelationalUtilities::~ObjectRelationalUtilities(){
  if(m_session && m_sessionOwned) {
    delete m_session;
    m_session = 0;
  }
}

void pool::ObjectRelationalUtilities::connect(const std::string& connectionString, coral::AccessMode accessMode){
  seal::IHandle<coral::IConnectionService> serviceHandle = POOLContext::context()->query<coral::IConnectionService>( "CORAL/Services/ConnectionService" );
  if ( ! serviceHandle ) {
    throw pool::ObjectRelationalException("Could not load the ConnectionService.","ObjectRelationalUtilities::connect");
  }
  if(m_session && m_sessionOwned) {
    delete m_session;
    m_session = 0;
  }
  m_sessionOwned = true;
  m_session = serviceHandle->connect( connectionString, accessMode);
  if(!m_session) throw pool::ObjectRelationalException("Could not connect to "+connectionString,"ObjectRelationalUtilities::connect");
}

coral::ISessionProxy* pool::ObjectRelationalUtilities::session(){
  if(!m_session) throw pool::ObjectRelationalException("Connection to database not available.","ObjectRelationalUtilities::session");
  return m_session;
}

void pool::ObjectRelationalUtilities::setProperty(const std::string& propertyName,const std::string& propertyValue){
  if(m_properties.find(propertyName)!=m_properties.end())
  {
    m_properties.erase(propertyName);
  }
  m_properties.insert(std::make_pair(propertyName,propertyValue));
}

void pool::ObjectRelationalUtilities::setFlag(const std::string& flagName,bool flag){
  if(m_flags.find(flagName)!=m_flags.end())
  {
    m_flags.erase(flagName);
  }
  m_flags.insert(std::make_pair(flagName,flag));  
}

std::string pool::ObjectRelationalUtilities::getProperty(const std::string& propertyName) const
{
  std::string ret = "";
  std::map<std::string,std::string>::const_iterator iProp =m_properties.find(propertyName);
  if(iProp!=m_properties.end())
  {
    ret = iProp->second;
  }
  return ret;
}

bool pool::ObjectRelationalUtilities::getFlag(const std::string& flagName) const
{
  bool ret = false;
  std::map<std::string,bool>::const_iterator iFlag =m_flags.find(flagName);
  if(iFlag!=m_flags.end())
  {
    ret = iFlag->second;
  }
  return ret;
}

void pool::ObjectRelationalUtilities::loadDictionaries(const std::string& dictionaries){
  if(!dictionaries.empty())
  { 
     std::vector<std::string> dictionaryList;
     stringTokenize(dictionaries,dictionaryList);
     loadLibraries(dictionaryList);
  }
}

void pool::ObjectRelationalUtilities::loadLibraries(std::vector<std::string> libraries){ 
 for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin(); iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    seal::MessageStream log( pool::POOLContext::context(), "POOL/ObjectRelationalAccess/ObjectRelationalUtilities",seal::Msg::Verbose);
    try {  // This is temporary. SEAL should get better in this.
      log << seal::Msg::Debug << "Loading library "<<library<<seal::flush;
      seal::SharedLibrary::load( "lib" + library + ".so" );
    } catch ( const seal::SharedLibraryError& error) {
      std::stringstream msg;
      msg <<  "Could not load library " << library << " - " << error.explainSelf();
      throw pool::ObjectRelationalException(msg.str(),"ObjectRelationalUtilities::loadLibraries");
    } catch(const std::runtime_error& e)   {
      std::stringstream msg;
      msg <<  "Runtime error while loading library " << library << " - " << e.what();
      throw pool::ObjectRelationalException(msg.str(),"ObjectRelationalUtilities::loadLibraries");
    } catch(const std::exception& e)   {
      std::stringstream msg;
      msg <<  "Exception loading library " << library << " - " << e.what();
      throw pool::ObjectRelationalException(msg.str(),"ObjectRelationalUtilities::loadLibraries");
    } catch(...)    {
      std::stringstream msg;
      msg <<  "Unexpected error loading library " << library;
      throw pool::ObjectRelationalException(msg.str(),"ObjectRelationalUtilities::loadLibraries");
    }
  }
}

void pool::ObjectRelationalUtilities::setMessageVerbosityLevel(seal::Msg::Level level )
{
  pool::POOLContext::setMessageVerbosityLevel(level);
  m_verbosityLevel = level;
}

seal::Msg::Level pool::ObjectRelationalUtilities::messageVerbosityLevel() const {
  return m_verbosityLevel;
}

void pool::ObjectRelationalUtilities::initialize()
{
  std::string verbosityLevel = getProperty(verbosityProperty());
  if(!verbosityLevel.empty())
  {
    if(verbosityLevel==debugVerbosityPropertyValue()) setMessageVerbosityLevel(seal::Msg::Debug);
    else if(verbosityLevel==infoVerbosityPropertyValue()) setMessageVerbosityLevel(seal::Msg::Info);
  }
  POOLContext::loadComponent("CORAL/Services/ConnectionService"); 
  std::string user = getProperty(userNameProperty());
  std::string password = getProperty(passwordProperty());
  if(!user.empty() && !password.empty()){
    POOLContext::loadComponent("POOL/Services/ObjectRelationalUtilities/LocalAuthenticationService");
    seal::IHandle<coral::IAuthenticationService> serviceHandle =
      POOLContext::context()->query<coral::IAuthenticationService>( "POOL/Services/ObjectRelationalUtilities/LocalAuthenticationService" );
    dynamic_cast< LocalAuthenticationService* >( serviceHandle.get() )->setCredentials(user,password);
  } 
}

const std::string& pool::ObjectRelationalUtilities::connectionStringProperty(){
  static std::string s_connectionStringProperty = "ConnectionString";
  return s_connectionStringProperty;
}

const std::string& pool::ObjectRelationalUtilities::userNameProperty(){
  static std::string s_userNameProperty = "UserName";
  return s_userNameProperty;
}

const std::string& pool::ObjectRelationalUtilities::passwordProperty(){
  static std::string s_passwordProperty = "Password";
  return s_passwordProperty;
}

const std::string& pool::ObjectRelationalUtilities::authenticationProperty(){
  static std::string s_authenticationProperty = "Authentication";
  return s_authenticationProperty;
}

const std::string& pool::ObjectRelationalUtilities::dictionaryProperty(){
  static std::string s_dictionaryProperty = "Dictionary";
  return s_dictionaryProperty;
}

const std::string& pool::ObjectRelationalUtilities::verbosityProperty(){
  static std::string s_verbosityProperty = "Verbosity";
  return s_verbosityProperty;
}

const std::string& pool::ObjectRelationalUtilities::debugVerbosityPropertyValue(){
  static std::string s_debugVerbosityPropertyValue = "Debug";
  return s_debugVerbosityPropertyValue;
}

const std::string& pool::ObjectRelationalUtilities::infoVerbosityPropertyValue(){
  static std::string s_infoVerbosityPropertyValue = "Info";
  return s_infoVerbosityPropertyValue;
}

DEFINE_SEAL_MODULE();

DEFINE_SEAL_PLUGIN( seal::ComponentFactory, pool::ObjectRelationalUtilities::LocalAuthenticationService,
                    pool::ObjectRelationalUtilities::LocalAuthenticationService::classContextLabel() );
 
 
