#include "RelationalCollectionMetadata.h"
#include "RelationalCollectionNames.h"
#include "RelationalCollectionBindVariables.h"

#include "POOLCore/Exception.h"

#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableSchemaEditor.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/TableDescription.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Attribute.h"


#include <iostream>
using namespace std;

namespace pool {
   namespace RelationalCollection { 

      RelationalCollectionMetadata::RelationalCollectionMetadata() {
      }

      RelationalCollectionMetadata::~RelationalCollectionMetadata() {
      }
        
      void
      RelationalCollectionMetadata::initialize( const std::string& collectionName,
                                                coral::ISessionProxy& session,
                                                pool::ICollection::OpenMode mode )
      {
         m_mode = mode;
         m_collectionName = collectionName; 
         coral::ISchema& nominalSchema = session.nominalSchema();
         bool metadataTableExists = nominalSchema.existsTable( RelationalCollectionNames::metadataTable() );
         
         if(  !metadataTableExists ) {
            std::cout << " ****** RelationalCollection::CollectionMetadata - no metadata table found" << std::endl;
            if( mode == ICollection::READ || mode == ICollection::UPDATE ) {
               // if( mode == ICollection::READ )   for now ignoring update mode case
               return;
            } else {
               // create mode
            
               coral::TableDescription description( RelationalCollectionNames::collectionTypeName() );
               description.setName( RelationalCollectionNames::metadataTable() );
               description.insertColumn( RelationalCollectionNames::collectionNameColumn(),
                                         coral::AttributeSpecification::typeNameForType<std::string>() );
               description.insertColumn( RelationalCollectionNames::metadataKeyColumn(),
                                         coral::AttributeSpecification::typeNameForType<std::string>() );
               description.insertColumn( RelationalCollectionNames::metadataValueColumn(),
                                         coral::AttributeSpecification::typeNameForType<std::string>() );
               
               std::vector<std::string> primaryKeyColumns;
               primaryKeyColumns.push_back( RelationalCollectionNames::collectionNameColumn() );
               primaryKeyColumns.push_back( RelationalCollectionNames::metadataKeyColumn() );

               /*  MN:  getting Oracle error on this ... ??
               cout << " Primary key def: " << primaryKeyColumns[0] << ", " << primaryKeyColumns[1] << endl;
               description.setPrimaryKey( primaryKeyColumns );
               */
               
               nominalSchema.createTable( description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
            
               std::cout << " ****** RelationalCollection::CollectionMetadata - created metadata table" << std::endl;
            }
         }
         
         m_metadataTable = &nominalSchema.tableHandle( RelationalCollectionNames::metadataTable() );
         
         // Define bind variables value holder for querying metadata table 
         m_whereDataForMetadata.extend<std::string>( RelationalCollectionNames::collectionNameColumn() );
         m_whereDataForMetadata[0].data<std::string>() = m_collectionName;

         m_whereDataForMetadataKey.extend<std::string>( RelationalCollectionNames::collectionNameColumn() );
         m_whereDataForMetadataKey[0].data<std::string>() = m_collectionName;
         m_whereDataForMetadataKey.extend<std::string>( RelationalCollectionNames::metadataKeyColumn() );

  
         
//          std::string errorMsg( "hmmm" );
//          std::cout << " ****** RelationalCollection::CollectionMetadata - " << errorMsg << std::endl;
//          throw pool::Exception( errorMsg,
//                                 "RelationalCollectionMetadata::initialize", 
//                                 "RelationalCollection");
      }

        

      
      bool
      RelationalCollectionMetadata::existsKey( const std::string& key ) const
      {
         return getValueForKey( key ) != 0;
      }

      
      const char*
      RelationalCollectionMetadata::getValueForKey( const std::string& key ) const
      {
         std::auto_ptr<coral::IQuery> query( m_metadataTable->newQuery() );
         m_whereDataForMetadataKey[1].data<std::string>() = key;
         query->setCondition( RelationalCollectionBindVariables::whereClauseForMetadataKey(), m_whereDataForMetadataKey );
         query->limitReturnedRows( 1, 0 );
         coral::ICursor& cursor = query->execute();
         if( !cursor.next() )
            return 0;
         return cursor.currentRow()[2].data<std::string>().c_str();
      }

        
      void
      RelationalCollectionMetadata::setValueForKey( const std::string& key,
                                                    const std::string& value )
      {
         if( m_mode == ICollection::READ ) {
            throw pool::Exception( "Metadata update not allowed in Read-only mode",
                                   "RelationalCollectionMetadata::setValueForKey", 
                                   "RelationalCollection");
         }
         if( key.empty() ) {
            throw pool::Exception( "Attempt to set metadata with an empty Key",
                                   "RelationalCollectionMetadata::setValueForKey", 
                                   "RelationalCollection");
         } 
         if( existsKey(key) ) {
            //update
            std::string setClause = RelationalCollectionNames::metadataValueColumn() + " = '" + value + "'";
            m_whereDataForMetadataKey[1].data<std::string>() = key;
            m_metadataTable->dataEditor().updateRows(
               setClause,
               RelationalCollectionBindVariables::whereClauseForMetadataKey(),
               m_whereDataForMetadataKey );
            std::cout << " Updated metadata entry: (" << key <<", " << value << ") " << std::endl;
         } else {
            // instert
            coral::AttributeList rowBuffer;
            rowBuffer.extend<std::string>( RelationalCollectionNames::collectionNameColumn() );
            rowBuffer.extend<std::string>( RelationalCollectionNames::metadataKeyColumn() );
            rowBuffer.extend<std::string>( RelationalCollectionNames::metadataValueColumn() );
            coral::AttributeList::iterator iAttribute = rowBuffer.begin();
            iAttribute->data< std::string >() = m_collectionName;
            ++iAttribute;
            iAttribute->data< std::string >() = key;
            ++iAttribute;
            iAttribute->data< std::string >() = value;
            m_metadataTable->dataEditor().insertRow( rowBuffer );
            
            std::cout << " Inserted new metadata entry: (" << key <<", " << value << ") " << std::endl;
         }
         
      }


      RelationalCollectionMetadata::RelationalCollectionMetadataIterator::
      RelationalCollectionMetadataIterator( coral::IQuery *query )
            : m_query( query ),
              m_cursor( 0 ),
              m_valid( false )
      {
         if( query ) {
            m_cursor = &query->execute();
            ++*this;
         }
      }

 
      RelationalCollectionMetadata::RelationalCollectionMetadataIterator::
      ~RelationalCollectionMetadataIterator( )
      {
         delete m_query;
      }
      
     
      bool
      RelationalCollectionMetadata::RelationalCollectionMetadataIterator::
       operator==( const ICollectionMetadataIterator& _rhs ) const
      {
         const RelationalCollectionMetadataIterator* rhs
            = dynamic_cast<const RelationalCollectionMetadataIterator*>( &_rhs );
         if( !rhs )
            throw pool::Exception( "Incompatible iterator comparison",           
                                   "RelationalCollectionMetadataItorator::operator==", 
                                   "RelationalCollection");
         return key() == rhs->key();
      }


      void
      RelationalCollectionMetadata::RelationalCollectionMetadataIterator::
      operator++()
      {
         if( m_cursor && m_cursor->next() ) {
            m_valid = true;
         } else {
            m_valid = false;
         }
      }


      const std::string&
      RelationalCollectionMetadata::RelationalCollectionMetadataIterator::
      key() const
      {
         static std::string        emptyKey;
         return ( m_valid? m_cursor->currentRow()[1].data<std::string>() : emptyKey );
      }

      
      const char *
      RelationalCollectionMetadata::RelationalCollectionMetadataIterator::
      value() const
      {
         return ( m_valid? m_cursor->currentRow()[2].data<std::string>().c_str() : 0 );
      }


      
      ICollectionMetadata::const_iterator
      RelationalCollectionMetadata::begin() const
      {
         std::auto_ptr<coral::IQuery> query( m_metadataTable->newQuery() );
         query->setCondition( RelationalCollectionBindVariables::whereClauseForMetadata(),
                              m_whereDataForMetadata );
         return ICollectionMetadata::const_iterator(
            new RelationalCollectionMetadataIterator( query.release() ) );
      }

        
      ICollectionMetadata::const_iterator
      RelationalCollectionMetadata::end() const
      {
         return ICollectionMetadata::const_iterator(
            new RelationalCollectionMetadataIterator( 0 ) );
      }
        

   } // end namespace
}
