#include "MappingSequence.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "SealKernel/MessageStream.h"
#include <memory>

pool::RelationalStorageService::MappingSequence::MappingSequence( coral::ISchema& schema,
                                                                  seal::Context* serviceContext,
                                                                  const std::string& serviceName ):
  m_schema( schema ),
  m_serviceContext( serviceContext ),
  m_serviceName( serviceName ),
  m_tableToOid(),
  m_sequenceTableExists( schema.existsTable( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ) ),
  m_whereClause( pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName() + " =:" +
                 pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName() ),
  m_whereData( 0 ),
  m_setClause( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() + " = " +
               pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() + " + 1" )
{
  m_whereData = new coral::AttributeList; 
  m_whereData->extend<std::string>(pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName());
  
}


pool::RelationalStorageService::MappingSequence::~MappingSequence()
{
  delete m_whereData;
}


int
pool::RelationalStorageService::MappingSequence::incrementOid( const std::string& tableName )
{
  std::map< std::string, int >::iterator iSequence = m_tableToOid.find( tableName );

  if ( iSequence == m_tableToOid.end() ) {
    // Make sure that the sequence table exists.
    if ( ! m_sequenceTableExists ) {
      this->createTheSequencesTable();
      m_sequenceTableExists = true;
    }

    // Lock the entry in the table.
    int lastOidUsed = 0;
    if ( ! ( this->lockEntry( tableName, lastOidUsed ) ) ) {
      // Create the entry in the table if it does not exist.
      coral::AttributeList rowData;
      rowData.extend<std::string>(pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName());
      rowData.extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() );
      coral::AttributeList::iterator iAttribute = rowData.begin();
      iAttribute->data< std::string >() = tableName;
      ++iAttribute;
      int startingOidValue = 0;
      iAttribute->data< int >() = startingOidValue;
      m_schema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ).dataEditor().insertRow( rowData );
      m_tableToOid.insert( std::make_pair( tableName, startingOidValue ) );
      return startingOidValue;
    }

    // Add the entry into the map.
    iSequence = m_tableToOid.insert( std::make_pair( tableName, lastOidUsed ) ).first;
  }

  // Increment the oid transiently
  int& oid = iSequence->second;
  ++oid;

  // Increment the oid in the database as well
  m_whereData->begin()->data<std::string>() = tableName;
  m_schema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ).dataEditor().updateRows( m_setClause,
                                                                                                                                       m_whereClause,
                                                                                                                                       *m_whereData );
  return oid;
}

void
pool::RelationalStorageService::MappingSequence::updateOid( const std::string& tableName, int lastOid ){
  // Make sure that the sequence table exists.
  if ( ! m_sequenceTableExists ) {
    this->createTheSequencesTable();
    m_sequenceTableExists = true;
  }

  bool update = false;
  coral::IQuery* query = m_schema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ).newQuery();
  query->limitReturnedRows( 1, 0 );
  query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() );
  query->defineOutputType( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName(), coral::AttributeSpecification::typeNameForType<int>() );
  m_whereData->begin()->data< std::string >() = tableName;
  query->setCondition( m_whereClause, *m_whereData );
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) {
    update = true;
  }
  delete query;

  coral::AttributeList rowData;
  rowData.extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() );
  rowData.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName() );
  coral::AttributeList::iterator iAttribute = rowData.begin();
  iAttribute->data< int >() = lastOid;
  ++iAttribute;
  iAttribute->data< std::string >() = tableName;

  if ( update ) {
    // Update the entry in the table
    std::string setClause( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() + " =: " + pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName());
    m_schema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ).dataEditor().updateRows( setClause,m_whereClause,rowData );
    m_tableToOid.erase( tableName );

  } else {

    m_schema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ).dataEditor().insertRow( rowData );
  }

  m_tableToOid.insert( std::make_pair( tableName, lastOid ) );
}

void
pool::RelationalStorageService::MappingSequence::clear()
{
  m_tableToOid.clear();
}


void
pool::RelationalStorageService::MappingSequence::createTheSequencesTable()
{
  seal::MessageStream log( m_serviceContext, m_serviceName );
  coral::TableDescription description( "CORAL" );
  description.setName(pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName());

  description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName(),
                            coral::AttributeSpecification::typeNameForType<std::string>() );
  description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName() );

  description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName(),
                            coral::AttributeSpecification::typeNameForType<int>() );
  description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() );

  description.setPrimaryKey( std::vector< std::string >( 1, pool::RelationalStorageService::RelationalStorageServiceNames::nameVariableInSequencesTableName() ) );

  m_schema.createTable( description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
}


bool
pool::RelationalStorageService::MappingSequence::lockEntry( const std::string& sequenceName,
                                                            int& lastOid )
{
  std::auto_ptr< coral::IQuery > query( m_schema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::sequencesTableName() ).newQuery() );
  query->limitReturnedRows( 1, 0 );
  query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName() );
  query->defineOutputType( pool::RelationalStorageService::RelationalStorageServiceNames::valueVariableInSequencesTableName(), coral::AttributeSpecification::typeNameForType<int>() );
  query->setForUpdate();
  m_whereData->begin()->data< std::string >() = sequenceName;
  query->setCondition( m_whereClause, *m_whereData );
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) {
    lastOid = cursor.currentRow().begin()->data< int >();
    return true;
  }
  else return false;
}
