#include "ObjectReader.h"
#include "ObjectRelationalReadDataBuffer.h"
#include "IContainerHandler.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeListSpecification.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "POOLCore/Reference.h"
#include "POOLCore/Token.h"
#include "Reflex/Member.h"
#include <memory>

namespace pool {
  namespace RelationalStorageService {
    union U_Primitives {
      bool b;
      char c;
      unsigned char uc;
      short s;
      unsigned short us;
      int i;
      unsigned int ui;
      long l;
      unsigned long ul;
      float f;
      double d;
      long double ld;
    };
  }
}


pool::RelationalStorageService::ObjectReader::ObjectReader( coral::ISchema& schema,
                                                            const ROOT::Reflex::Type& dictionary,
                                                            const pool::ObjectRelationalMappingElement& mapping,
                                                            pool::DbDatabase* databaseHandle ):
  m_dataBuffer( new pool::RelationalStorageService::ObjectRelationalReadDataBuffer ),
  m_refDict( ROOT::Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) ),
  m_databaseContext( databaseHandle )
{
  m_dataBuffer->build( schema, mapping, dictionary );
}



pool::RelationalStorageService::ObjectReader::ObjectReader( coral::ISchema& schema,
                                                            const pool::ObjectRelationalMappingElement& mapping,
                                                            pool::DbDatabase* databaseHandle ):
  m_dataBuffer( new pool::RelationalStorageService::ObjectRelationalReadDataBuffer ),
  m_refDict( ROOT::Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) ),
  m_databaseContext( databaseHandle )
{
  const ROOT::Reflex::Type dictionary = ROOT::Reflex::Type::ByName( mapping.variableType() );
  if ( ! dictionary ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for \"" +
                                                                             mapping.variableType() + "\"",
                                                                             "ObjectReader" );

  }
  m_dataBuffer->build( schema, mapping, dictionary );
}



pool::RelationalStorageService::ObjectReader::~ObjectReader()
{
  delete m_dataBuffer;
}



bool
pool::RelationalStorageService::ObjectReader::readObject( void* objectData,
                                                          int oid,
                                                          bool anyObject,
                                                          pool::DataCallBack& callback )
{
  std::auto_ptr< coral::IQuery > query( m_dataBuffer->table().newQuery() );
  const coral::AttributeListSpecification& outputSpecification = m_dataBuffer->dataSpecification();
  for ( coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
        iSpec != outputSpecification.end(); ++iSpec ) {
    query->addToOutputList( iSpec->name() );
    query->defineOutputType( iSpec->name(),iSpec->typeName());
  }

  coral::AttributeList& oidData = m_dataBuffer->whereDataForId();
  oidData.begin()->data< int >() = oid;
  std::string idColumn = oidData.begin()->specification().name();
  std::string whereClause = idColumn + " ";
  if ( anyObject ) whereClause += ">";
  else whereClause += "=";
  whereClause += ":" + idColumn;
  if ( anyObject ) {
    query->addToOrderList( idColumn );
    query->limitReturnedRows( 1, 0 );
  }
  query->setCondition( whereClause, oidData );

  coral::ICursor& cursor = query->execute();

  if ( cursor.next() ) {

    const coral::AttributeList& row = cursor.currentRow();

    // Get the OID
    coral::AttributeList::const_iterator iAttribute = row.begin();
    oid = iAttribute->data< int >();
    ++iAttribute;

    // Copy the data
    std::vector<int>::const_iterator iOffset = m_dataBuffer->offsets().begin();
    while ( iAttribute != row.end() ) {
      int offset = *iOffset;
      if ( offset < 0 ) {
        pool::Token::OID_t oid;
        oid.first = iAttribute->data<int>();
        ++iAttribute;
        ++iOffset;
        oid.second = iAttribute->data<int>();
        void* refAddress = static_cast< char* >( objectData ) - ( 1 + offset );
        // Call here the ref handler to get the oid...
        pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
        callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, refAddress, &oid );
      }
      else {
        if( iAttribute->specification().type() == typeid(coral::Blob) )
        {
          m_dataBuffer->readBlob( *iOffset, iAttribute->data<coral::Blob>(), static_cast< char* >( objectData ) + *iOffset );
        } else 
        {
          iAttribute->copyValueToAddress( static_cast< char* >( objectData ) + *iOffset );
        }
      }
      ++iAttribute;
      ++iOffset;
    }

    // Loop over the arrays now...
    std::vector< int > identity( 1, oid );
    for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = m_dataBuffer->begin();
          iBuffer != m_dataBuffer->end(); ++iBuffer ) {
      (*iBuffer)->containerHandler().clear(static_cast< char* >( objectData ) + (*iBuffer)->offsetToParent());
      if ( ! this->readArray( identity, **iBuffer,
                              static_cast< char* >( objectData ) + (*iBuffer)->offsetToParent(),
                              callback ) ) return false;
    }

    return true;
  }
  else {
    return false;
  }
}


bool
pool::RelationalStorageService::ObjectReader::readArray( const std::vector<int>& identity,
                                                         ObjectRelationalReadDataBuffer& buffer,
                                                         void* address,
                                                         pool::DataCallBack& callback )
{
  std::auto_ptr< coral::IQuery > query( buffer.table().newQuery() );
  const coral::AttributeListSpecification& outputSpecification = buffer.dataSpecification();
  for (coral::AttributeListSpecification::const_iterator iSp=outputSpecification.begin(); iSp != outputSpecification.end(); ++iSp ) {
    query->addToOutputList( iSp->name() );
    query->defineOutputType( iSp->name(),iSp->typeName());
  }

  coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
  query->addToOrderList( iSpec->name() ); // Order by pos.

  // Fill-in the identities.
  coral::AttributeList& whereData = buffer.whereDataForId();
  coral::AttributeList::iterator iData = whereData.begin();
  for ( std::vector<int>::const_iterator id = identity.begin();
        id != identity.end(); ++id ) {
    iData->data<int>() = *id;
    ++iData;
  }

  // Set the condition
  query->setCondition( buffer.whereClauseForId(), whereData );

  // Prefetch rows.
  query->setRowCacheSize( 100 ); // We should better define this value !!!

  // Create the identity for the sub-elements
  std::vector<int> identityForSubElements( 1 + identity.size(), 0 );

  // Check if the container is an associative one
  ROOT::Reflex::Type keyType;
  ROOT::Reflex::Type dictionary = buffer.dictionary();
  ROOT::Reflex::Type iteratorDereferenceReturnType = buffer.containerHandler().iteratorReturnType();
  pool::RelationalStorageService::U_Primitives primitiveStub;
  int keyItems = buffer.keyItems();
  
  ROOT::Reflex::Member firstField;
  if ( keyItems > 0 ) {
    keyType = dictionary.TemplateArgumentAt(0);
    if(keyType){
      firstField = iteratorDereferenceReturnType.MemberByName( "first" );
      if ( ! firstField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException(
          "Could not retrieve the \"first\" field for the class \"" +
          iteratorDereferenceReturnType.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) + "\"",
          "ObjectReader::readArray" );
      }
    }
  }
  
  // Issue the query...
  coral::ICursor& cursor = query->execute();

  bool isArray = dictionary.IsArray();
  bool isElementFundamental = iteratorDereferenceReturnType.IsFundamental();

  size_t arraySize = dictionary.ArrayLength();
  
  unsigned int i=0;
  while ( cursor.next() ) {

    const coral::AttributeList& row = cursor.currentRow();
    
    // Get the Identity
    coral::AttributeList::const_iterator iAttribute = row.begin();
    int arrayIndex = iAttribute->data< int >();
    for ( std::vector<int>::iterator idout = identityForSubElements.begin();
          idout != identityForSubElements.end(); ++idout ) {
      *idout = iAttribute->data< int >();
      ++iAttribute;
    }

    // Create a new element for the array
    void* objectData = 0;
    if( isArray ){
      if(arrayIndex < (int)arraySize){
        // the memory has been allocated already!
        objectData = static_cast<char*>(address)+arrayIndex*iteratorDereferenceReturnType.SizeOf();
        if(!isElementFundamental){
          // in this case the initialization is required: use default constructor...
          iteratorDereferenceReturnType.Construct(ROOT::Reflex::Type(0,0),std::vector< void* >(),objectData);
        }
      } else {
        // not sure what to do in this case... exception or ignore???
        continue;
      }
    } else {
      if(isElementFundamental){
        objectData = &primitiveStub;
      } else {
        objectData = iteratorDereferenceReturnType.Construct().Address();
      }
    }

    void* componentData = objectData;
    void* keyData = 0;

    // Copy the data
    std::vector<int>::const_iterator iOffset = buffer.offsets().begin();

    if ( keyType ) { // treat the key object first
      keyData = static_cast< char* >( objectData ) + firstField.Offset();
      for ( int i = 0; i < keyItems; ++i ) {
        int offset = *iOffset;
        if ( offset < 0 ) {
          pool::Token::OID_t oid;
          oid.first = iAttribute->data<int>();
          ++iAttribute;
          ++iOffset;
          oid.second = iAttribute->data<int>();
          void* refAddress = static_cast< char* >( keyData ) - ( 1 + offset );
          // Call here the ref handler to get the oid...
          pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
          callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, refAddress, &oid );
        }
        else {
          if( iAttribute->specification().type() == typeid(coral::Blob) ){
            buffer.readBlob( *iOffset, iAttribute->data<coral::Blob>(), static_cast< char* >( keyData ) + *iOffset );
          } else {
            iAttribute->copyValueToAddress( static_cast< char* >( keyData ) + *iOffset );
          }
        }
        ++iAttribute;
        ++iOffset;
      }
      const ROOT::Reflex::Member secondField = iteratorDereferenceReturnType.MemberByName( "second" );
      if ( ! secondField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the \"second\" field for the class \"" +
                                                                                   iteratorDereferenceReturnType.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) + "\"",
                                                                                 "ObjectReader::readArray" );
      }
      componentData = static_cast< char* >( objectData ) + secondField.Offset();
    }

    while ( iAttribute != row.end() ) {
      int offset = *iOffset;
      if ( offset < 0 ) {
        pool::Token::OID_t oid;
        oid.first = iAttribute->data<int>();
        ++iAttribute;
        ++iOffset;
        oid.second = iAttribute->data<int>();
        void* refAddress = static_cast< char* >( componentData ) - ( 1 + offset );
        // Call here the ref handler to get the oid...
        pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
        callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, refAddress, &oid );
      }
      else {
        if( iAttribute->specification().type() == typeid(coral::Blob) ){
          buffer.readBlob( *iOffset, iAttribute->data<coral::Blob>(), static_cast< char* >( componentData ) + *iOffset );
        } else {
          iAttribute->copyValueToAddress( static_cast< char* >( componentData ) + *iOffset );
        }
      }
      ++iAttribute;
      ++iOffset;
    }

    // Loop over the arrays now...
    for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = buffer.begin();
          iBuffer != buffer.end(); ++iBuffer ) {
      void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->readArray( identityForSubElements, **iBuffer,
                                static_cast< char* >( theAddress ) + (*iBuffer)->offsetToParent(),
                              callback ) ) return false;
    }

    if(!isArray) {
      size_t prevSize = buffer.containerHandler().size( address );
      buffer.containerHandler().appendNewElement( address, objectData );
      bool inserted = buffer.containerHandler().size( address )>prevSize;
      if ( ! ( iteratorDereferenceReturnType.IsFundamental() ) ) {
        iteratorDereferenceReturnType.Destruct( objectData );
      }
      if ( !inserted ) return false;
    }
    ++i;
  }

  return true;
}
