#include "RelationalPoolContainer.h"
#include "RelationalPoolDatabase.h"
#include "PoolDatabaseManager.h"
#include "PoolBaseContainerManager.h"
#include "StorageSvc/DbInstanceCount.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbColumn.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbTransaction.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/SessionException.h"
#include "SealKernel/MessageStream.h"
#include "Reflex/Object.h"

pool::RelationalStorageService::RelationalPoolContainer::RelationalPoolContainer( pool::IOODatabase* idb,
                                                                                  const std::string& name,
                                                                                  seal::Context* context ):
  pool::DbContainerImp( idb ),
  m_serviceName( name ),
  m_context( context ),
  m_readOnly( false ),
  m_databaseManager( 0 ),
  m_databaseContext( 0 )
{
  pool::DbInstanceCount::increment(this);
  m_canDestroy = true;
  m_canUpdate  = true;
}


pool::RelationalStorageService::RelationalPoolContainer::~RelationalPoolContainer()
{
  pool::DbInstanceCount::decrement(this);
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::isShapeSupported( const pool::DbTypeInfo* /* type */ ) const
{
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::open( const pool::DbDatabase& dbH,
                                                               const std::string& name,
                                                               const pool::DbTypeInfo* info,
                                                               pool::DbAccessMode mode )
{
  if ( name.empty() ) {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << "Invalid container name supplied at opening." << seal::flush;
    return pool::Error;
  }

  if ( ! dbH.isValid() ) {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << "Invalid database handle supplied when opening the \"" << name << "\" container" << seal::flush;
    return pool::Error;
  }

  pool::IDbDatabase* idb = dbH.info();

  try 
  {
    // Check here if a transaction is active in case of read-only mode !!!
    if ( mode == pool::READ ) {
      coral::ISessionProxy& session = dynamic_cast< pool::RelationalStorageService::RelationalPoolDatabase*>( idb )->session();
      coral::ITransaction& transaction = session.transaction();
      if ( ! ( transaction.isActive() ) ) {
        transaction.start( true ); // Implicitly start a read-only transaction
      }
    }    

    m_databaseManager = &( dynamic_cast< pool::RelationalStorageService::RelationalPoolDatabase*>( idb )->databaseManager() );
    pool::RelationalStorageService::PoolDatabaseManager& dbManager = *m_databaseManager;

    if ( dbManager.existsContainer( name ) ) {
      if ( mode == pool::CREATE ) {
        seal::MessageStream log( m_context, m_serviceName );
        log << seal::Msg::Error << "Cannot create a container named \"" << name << "\". It already exists!" << seal::flush;
        return pool::Error;
      }
      if ( mode == pool::READ ) {
        m_readOnly = true;
      }
      else m_readOnly = false;
    }
    else {
      if ( mode == pool::READ ) {
        seal::MessageStream log( m_context, m_serviceName );
        log << seal::Msg::Error << "Cannot open for reading a non-existing container named \"" << name << "\"." << seal::flush;
        return pool::Error;
      }

      // Support only for objects of a single column
      if ( info->columns().size() != 1 ) {
        seal::MessageStream log( m_context, m_serviceName );
        log << seal::Msg::Error << "Invalid database handle supplied when opening the \"" << name << "\" container" << seal::flush;
        return pool::Error;
      }

      this->createContainerManager( dbManager, name, info );
      m_readOnly = false;
    }
  } catch (const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << "Failure while opening container named \"" << name << "\"." << seal::flush;
    log << seal::Msg::Error << e.what() << seal::flush;
    return pool::Error;
  }

  // Set the protected data member of the base class !!!
  m_name = name;

  // Store the database context
  m_databaseContext = &( const_cast<pool::DbDatabase&>( dbH ) );

  this->containerManager().id(); // for debug!!
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::close() {
  m_databaseManager->freeContainerHandle( m_name );
  return DbContainerImp::close();
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::select(  pool::DbSelect& /* criteria */ )
{
  // Temporary implementation
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::fetch( pool::DbSelect& selection )
{
  // Temporary implementation !!!
  if ( selection.criteria().length() == 0 || selection.criteria() == "*" )  {
    selection.link().second++;
    return this->fetch( selection.link(), selection.link() );
  }
  return pool::Error;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::fetch( const pool::DbLink& linkH,
                                                                pool::DbLink& stmt )
{
  try
  {
    pool::RelationalStorageService::PoolBaseContainerManager& containerManager = m_databaseManager->containerHandle( m_name );
    ROOT::Reflex::Type dictionary;
    if ( containerManager.existsOid( linkH.second, stmt.second, dictionary ) ) {
      return pool::Success;
    }
  }
  catch(const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << e.what() << seal::flush;
  }
  return pool::Error;
}


long long
pool::RelationalStorageService::RelationalPoolContainer::size()
{
  try
  {
    pool::RelationalStorageService::PoolBaseContainerManager& containerManager = m_databaseManager->containerHandle( m_name );
    return containerManager.size() + pool::DbContainerImp::size();
  }catch (const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << e.what() << seal::flush;    
  }
  return 0;
  
}


long long
pool::RelationalStorageService::RelationalPoolContainer::nextRecordId()
{
  try
  {
    pool::RelationalStorageService::PoolBaseContainerManager& containerManager = m_databaseManager->containerHandle( m_name );
    return containerManager.numberOfObjectsWritten() + pool::DbContainerImp::size();
  }catch (const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << e.what() << seal::flush;    
  }
  return 0;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::loadAnObject( pool::DataCallBack& callback,
                                                                       int& oid,
                                                                       bool any_next )
{
  try 
  {
    
    // Checks the state of the database manager
    pool::RelationalStorageService::PoolBaseContainerManager& containerManager = m_databaseManager->containerHandle( m_name );

    // Retrieve the persistent shape
    const pool::DbTypeInfo* persistentType = dynamic_cast< const pool::DbTypeInfo* >( callback.shape() );
    if ( ! persistentType )   {
      seal::MessageStream log( m_context, m_serviceName );
      log << seal::Msg::Error << "No type information present when reading an object!" << seal::flush;
      return pool::Error;
    }

    if ( persistentType->columns().size() != 1 ) {
      seal::MessageStream log( m_context, m_serviceName );
      log << seal::Msg::Error << "Objects of more than one column are not yet supported" << seal::flush;
      return pool::Error;
    }
    else {
      // Get the pointer to the persistent object
      const pool::DbColumn* column = *( persistentType->columns().begin() );

      void* context = 0;
      if ( ! callback.start( pool::DataCallBack::GET, 0, &context ).isSuccess() ) {
        seal::MessageStream log( m_context, m_serviceName );
        log << "Could not initialize the persistent->transient shape transformation" << seal::flush;
        return pool::Error;
      }

      void* userPointer = 0;
      if ( ! callback.bind( pool::DataCallBack::GET, column, 0, context, &userPointer ).isSuccess() ) {
        seal::MessageStream log( m_context, m_serviceName );
        log << seal::Msg::Error << "Could not retrieve the address of the data to be written" << seal::flush;
        return pool::Error;
      }

      void** pointerToAddressToData = static_cast< void** >( userPointer );

      // Instantiate the persistent object if necessary
      if ( context == 0 ) {
        if ( ! ( column->type() ) ) {
          seal::MessageStream log( m_context, m_serviceName );
          log << seal::Msg::Error << "No dictionary found for " << column->typeName() << seal::flush;
          return pool::Error;
        }
        *pointerToAddressToData = column->type().Construct().Address();
      }
      else {
        pointerToAddressToData = &userPointer;
      }

      // Read the object
      if ( any_next ) {
        if ( ! containerManager.readAnyObject( pointerToAddressToData,
                                               oid,
                                               callback,
                                               column->typeName(),
                                               column->type(),
                                               m_databaseContext ) ) {
          seal::MessageStream log( m_context, m_serviceName );
          log << seal::Msg::Debug << "Could not retrieve an object with address equal or higher than " << oid << seal::flush;
          return pool::Error;
        }
      }
      else {
        if ( ! containerManager.readObject( pointerToAddressToData,
                                            oid,
                                            callback,
                                            column->typeName(),
                                            column->type(),
                                            m_databaseContext ) ) {
          seal::MessageStream log( m_context, m_serviceName );
          log << seal::Msg::Error << "Could not read an object with an oid equal to " << oid << seal::flush;
          return pool::Error;
        }
      }

      if ( ! callback.end( pool::DataCallBack::GET, context ) ) {
        seal::MessageStream log( m_context, m_serviceName );
        log << seal::Msg::Error <<"Could not perform the persistent->transient shape transformation" << seal::flush;
        return pool::Error;
      }

    }
  }
  catch(const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error <<"Failure in loading object with oid="<< oid << seal::flush;
    log << seal::Msg::Error <<e.what() << seal::flush;
    return pool::Error;
  }
  

  return pool::Success;
}



pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::load( pool::DataCallBack* call,
                                                               const pool::DbLink& lnkH, 
                                                               pool::DbLink&       oid,
                                                               pool::DbAccessMode,
                                                               bool any_next )
{
  oid.second = lnkH.second;
  return this->loadAnObject( *call, oid.second, any_next );
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::loadObject( pool::DataCallBack* call,
                                                                     pool::DbLink& oid,
                                                                     pool::DbAccessMode /* mode */)
{
  return this->loadAnObject( *call, oid.second, false );
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::destroyObject(  pool::DbContainerImp::TransactionStack::value_type& entry )
{
  try 
  {
    
    // Checks the state of the database manager
    pool::RelationalStorageService::PoolBaseContainerManager& containerManager = m_databaseManager->containerHandle( m_name );
    if ( containerManager.deleteObject( entry.link.second ) ) return pool::Success;
    else {
      seal::MessageStream log( m_context, m_serviceName );
      log << seal::Msg::Error <<"Could not destroy an object in the container " << m_name << " with an OID " << entry.link.second << seal::flush;
      return pool::Error;
    }
  }catch (const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error <<"Could not destroy an object in the container " << m_name << " with an OID " << entry.link.second << seal::flush;
    log <<seal::Msg::Error<<e.what()  << seal::flush;
    return pool::Error;    
  }
  
  
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::writeOrUpdateObject( pool::DbContainerImp::TransactionStack::value_type& entry,
                                                                              bool write )
{
  // Checks the state of the database manager
  pool::RelationalStorageService::PoolBaseContainerManager& containerManager = m_databaseManager->containerHandle( m_name );

  pool::DataCallBack& callback = *( entry.call );
  const void* inputData = entry.objH;
  if ( ! inputData ) inputData = callback.object();
  void* context = 0;

  // Perform the transient->persistent shape transformation
  if ( ! callback.start( pool::DataCallBack::PUT,
                         const_cast< void * >( inputData ),
                         &context ).isSuccess() ) {
    seal::MessageStream log( m_context, m_serviceName );
    log << "Could not perform the transient->persistent shape transformation" << seal::flush;
    return pool::Error;
  }

  // Retrieve the persistent shape
  const pool::DbTypeInfo* persistentType = dynamic_cast< const pool::DbTypeInfo* >( callback.shape() );
  if ( ! persistentType )   {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << "No type information present when writing an object!" << seal::flush;
    return pool::Error;
  }

  if ( persistentType->columns().size() != 1 ) {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << "Objects of more than one column are not yet supported" << seal::flush;
    return pool::Error;
  }
  else {
    // Get the pointer to the persistent object
    const pool::DbColumn* column = *( persistentType->columns().begin() );
    void* dataToBeWritten = 0;
    if ( ! callback.bind( pool::DataCallBack::PUT, column, 0, context, &dataToBeWritten).isSuccess() ) {
      seal::MessageStream log( m_context, m_serviceName );
      log << seal::Msg::Error << "Could not retrieve the address of the data to be written" << seal::flush;
      return pool::Error;
    }

    try 
    {
      
      if ( write ) { // Write a new object
        if ( ! containerManager.writeObject( dataToBeWritten,
                                             entry.link.second,
                                             callback,
                                             column->typeName(),
                                             column->type(),
                                             m_databaseContext ) ) {
          seal::MessageStream log( m_context, m_serviceName );
          log << seal::Msg::Error << "Could not write an object" << seal::flush;
          return pool::Error;
        }
      }
      else { // Update an existing object
        if ( ! containerManager.updateObject( dataToBeWritten,
                                              entry.link.second,
                                              callback,
                                              column->typeName(),
                                              column->type(),
                                              m_databaseContext ) ) {
          seal::MessageStream log( m_context, m_serviceName );
          log << seal::Msg::Error << "Could not update an object" << seal::flush;
          return pool::Error;
        }
      }
    }
    catch (const coral::SessionException& e)
    {
      seal::MessageStream log( m_context, m_serviceName );
      log << seal::Msg::Error << "Could not update an object" << seal::flush;
      log << seal::Msg::Error << e.what() << seal::flush;
      return pool::Error;
    }
    
  }
  
  // Release the internal persistent buffer
  return callback.end( pool::DataCallBack::PUT, context );
}


pool::RelationalStorageService::PoolBaseContainerManager&
pool::RelationalStorageService::RelationalPoolContainer::containerManager()
{
  return m_databaseManager->containerHandle( m_name );
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::updateObject( pool::DbContainerImp::TransactionStack::value_type& entry )
{
  return this->writeOrUpdateObject( entry, false );
}



pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::writeObject( pool::DbContainerImp::TransactionStack::value_type& entry )
{
  return this->writeOrUpdateObject( entry, true );
}


pool::DbStatus 
pool::RelationalStorageService::RelationalPoolContainer::updateContent(const std::vector<std::pair<std::string,std::string> >& classInfo, bool recreate, bool dry){
  pool::DbStatus status = pool::Success;
  try
  {
    if(! m_databaseManager->containerHandle( m_name ).updateContent( classInfo, recreate, dry )){
      status = pool::Error;
    }
  }
  catch (const coral::SessionException& e)
  {
    seal::MessageStream log( m_context, m_serviceName );
    log << seal::Msg::Error << "Could not update content of container" << seal::flush;
    log << seal::Msg::Error << e.what() << seal::flush;
    return pool::Error;    
  }
  
  return status; 
}

pool::DbStatus
pool::RelationalStorageService::RelationalPoolContainer::endTransaction( pool::DbTransaction& refTr )
{
  pool::DbStatus status = pool::Success;
  switch( refTr.state() ) {
  case Transaction::TRANSACT_START:
    break;
  case Transaction::TRANSACT_ROLLBACK:
  case Transaction::TRANSACT_COMMIT:
  case Transaction::TRANSACT_FLUSH:
    m_databaseManager->containerHandle( m_name ).endTransaction();
    //m_databaseManager->freeContainerHandle( m_name );
    m_databaseManager->endOfContainerTransaction( m_name );
    break;
  default:
    status = pool::Error;
    break;
  }
  return status;
}

std::string
pool::RelationalStorageService::RelationalPoolContainer::mappingVersion(){
  return m_databaseManager->containerHandle( m_name ).mappingVersion();
}

std::string
pool::RelationalStorageService::RelationalPoolContainer::className(){
  return m_databaseManager->containerHandle( m_name ).className();
}
