#include "RelationalStorageServiceException.h"
#include "RelationalPoolDatabase.h"
#include "PoolDatabaseManager.h"
#include "PoolBaseContainerManager.h"
#include "StorageSvc/DbInstanceCount.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbTransaction.h"
#include "POOLCore/Guid.h"
#include "POOLCore/POOLContext.h"
#include "RelationalAccess/IConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "CoralBase/Exception.h"
#include "SealKernel/MessageStream.h"
#include "SealKernel/Context.h"
#include <memory>

static std::string serviceName = "POOL/RelationalStorageService";

pool::RelationalStorageService::RelationalPoolDatabase::RelationalPoolDatabase( pool::IOODatabase* idb,
                                                                                const std::string& name,
                                                                                seal::Context* context ):
  pool::DbDatabaseImp( idb ),
  m_formatVersion( "1.0" ),
  m_name( name ),
  m_context( context ),
  m_session( 0 ),
  m_readOnlyMode( false ),
  m_databaseManager( 0 ),
  m_pfnName( "" )
{
  pool::DbInstanceCount::increment( this );
}


pool::RelationalStorageService::RelationalPoolDatabase::~RelationalPoolDatabase()
{
  if ( m_databaseManager ) delete m_databaseManager;
  if ( m_session ) {
    delete m_session;
  }
  pool::DbInstanceCount::decrement( this );
}


pool::RelationalStorageService::PoolDatabaseManager&
pool::RelationalStorageService::RelationalPoolDatabase::databaseManager()
{
  return *m_databaseManager;
}


bool
pool::RelationalStorageService::RelationalPoolDatabase::exists( const std::string& name,
                                                                seal::Context* context )
{
  // Retrieve an IConnectionService
  std::vector< seal::IHandle< coral::IConnectionService > > loadedServices;
  context->query( loadedServices );
  if ( loadedServices.empty() ) {
    // Load the ConnectionService into the top-lecel POOL context
    pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );
    context->query( loadedServices );
  }
  seal::IHandle< coral::IConnectionService > connectionService = loadedServices.front();

  bool existsTopLevelTable = false;
  coral::ISessionProxy* tempSession = 0;
  try{
    tempSession = connectionService->connect( name, coral::ReadOnly );
    if ( ! tempSession ) return false;
    coral::ITransaction& transaction = tempSession->transaction();
    bool activeTransaction = transaction.isActive();
    if ( ! activeTransaction ) transaction.start( true );
    pool::RelationalStorageService::PoolDatabaseManager manager( tempSession, context );
    existsTopLevelTable = manager.existsDatabase();
    if ( ! activeTransaction ) transaction.commit();
  }
  catch (const coral::Exception& e)
  {
    seal::MessageStream log( context, serviceName );
    log << seal::Msg::Error << "Could not connect to \"" << name << "\"" << seal::flush;
    log << seal::Msg::Error << e.what()<< seal::flush;
    if(tempSession) delete tempSession;
    return false;
  }

  delete tempSession;
  return existsTopLevelTable;
}


coral::ISessionProxy&
pool::RelationalStorageService::RelationalPoolDatabase::session()
{
  return *m_session;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::onOpen( const pool::DbDatabase& dbH,
                                                                pool::DbAccessMode mode )
{
  std::string par_val = m_formatVersion;
  if ( !dbH.param( "FORMAT_VSN", par_val ).isSuccess() ) {
    if ( mode&pool::CREATE || mode&pool::UPDATE ) {
      dbH.addParam( "FORMAT_VSN", m_formatVersion );
    }
    else {
      seal::MessageStream log( m_context, dbH.name() );
      log << seal::Msg::Warning << "No data format parameter present" << seal::endmsg;
    }
  }
  else {
    m_formatVersion = par_val;
  }

  try{
    coral::ITransaction& transaction = m_session->transaction();
    if ( transaction.isActive() )
      transaction.commit();
  }
  catch (const coral::Exception& e)
  {
    if(m_session) delete m_session;
    m_session = 0;
    seal::MessageStream log( m_context, m_name );
    log << seal::Msg::Error << "Cannot commit transaction - RelationalPoolDatabase::onOpen"<<seal::flush;
    log << seal::Msg::Error << e.what() << seal::flush;
    return pool::Error;
  }
  
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::open( const pool::DbDomain&, 
                                                              const std::string& name, 
                                                              pool::DbAccessMode mode )
{
  if ( m_databaseManager ) delete m_databaseManager;
  if ( m_session ) {
    delete m_session;
    m_session = 0;
  }

  // Connect to the database

  // Retrieve an IConnectionService
  std::vector< seal::IHandle< coral::IConnectionService > > loadedServices;
  m_context->query( loadedServices );
  if ( loadedServices.empty() ) {
    // Load the ConnectionService into the top-lecel POOL context
    pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );
    m_context->query( loadedServices );
  }
  seal::IHandle< coral::IConnectionService > connectionService = loadedServices.front();

  try {
    m_session = connectionService->connect( name,
                                            ( mode == pool::READ ) ? coral::ReadOnly : coral::Update );
  }
  catch (const coral::Exception& e) {
    seal::MessageStream log( m_context, m_name );
    log << seal::Msg::Error << "Cannot start the user session" << seal::endmsg;
    log << seal::Msg::Error << e.what() << seal::endmsg;
    if(m_session) delete m_session;
    m_session = 0;
    return pool::Error;
  }

  coral::ITransaction& transaction = m_session->transaction();
  bool activeTransaction = transaction.isActive();
  if ( ! activeTransaction ) transaction.start( true );
  m_databaseManager = new pool::RelationalStorageService::PoolDatabaseManager( m_session, m_context );
  if ( ! activeTransaction ) transaction.commit();
  
  // Check for the existence of the pool database
  bool existsDatabase = pool::RelationalStorageService::RelationalPoolDatabase::exists( name, m_context );

  try
  {
      // Take the relevant actions depending on the opening mode and the existence of the database
    if ( existsDatabase ) {
      if ( mode == pool::CREATE ) {
        delete m_databaseManager;
        m_databaseManager = 0;
        if(m_session) delete m_session;
        m_session = 0;
        seal::MessageStream log( m_context, serviceName );
        log << seal::Msg::Error << "Cannot create a database in \"" << name << "\". It already exists !!!" << seal::flush;
        return pool::Error;
      }
      else if ( mode == pool::RECREATE ) {
        if ( ! activeTransaction ) transaction.start();
        m_databaseManager->dropDatabase();
        m_databaseManager->createDatabase();        
        if ( ! activeTransaction ) transaction.commit();
      }
    }
    else {
      if ( mode == pool::READ ) {
        delete m_databaseManager;
        m_databaseManager = 0;
        if(m_session)delete m_session;
        m_session = 0;
        seal::MessageStream log( m_context, m_name );
        log << seal::Msg::Error << "Cannot open the non-existing database in \"" << name << "\" for reading" << seal::flush;
        return pool::Error;
      }
      if ( ! activeTransaction ) transaction.start();
      m_databaseManager->createDatabase();
      if ( ! activeTransaction ) transaction.commit();
    }
    // Set if necessary the read-only flag
    if ( mode == pool::READ ) m_readOnlyMode = true;
    else m_readOnlyMode = false;

    if ( ! activeTransaction ) transaction.start( m_readOnlyMode );
    // I need to start the transaction here because the StorageSvc opens the special containers.
    // The transaction will be committed at the onOpen method.
  }
  catch (const coral::Exception& e)
  {
    delete m_databaseManager;
    m_databaseManager = 0;
    if(m_session) delete m_session;
    m_session = 0;
    seal::MessageStream log( m_context, m_name );
    log << seal::Msg::Error << "Cannot open the database in \"" << name << "\"." << seal::flush;
    log << seal::Msg::Error << e.what() << seal::flush;
    return pool::Error;
    
  }

  m_pfnName = name;
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::reopen( pool::DbAccessMode mode )
{
  if ( mode == pool::READ ) m_readOnlyMode = true;
  else m_readOnlyMode = false;

  try
  {     
    if ( m_session ) {
      delete m_session;
      m_session = 0;
    }

    std::vector< seal::IHandle< coral::IConnectionService > > loadedServices;
    m_context->query( loadedServices );
    seal::IHandle< coral::IConnectionService > connectionService = loadedServices.front();
    
    try {
      m_session = connectionService->connect( m_pfnName,
                                              m_readOnlyMode ? coral::ReadOnly : coral::Update );
    }
    catch (const coral::Exception& e) {
      seal::MessageStream log( m_context, m_name );
      log << seal::Msg::Error << "Cannot start the user session" << seal::endmsg;
      log << seal::Msg::Error << e.what() << seal::endmsg;
      if(m_session) delete m_session;
      m_session = 0;
      return pool::Error;
    }

    coral::ITransaction& transaction = m_session->transaction();
    bool activeTransaction = transaction.isActive();
    if ( ! activeTransaction ) transaction.start( m_readOnlyMode );

    m_databaseManager->reopen( m_session );

  }
  catch(const coral::Exception& e)
  {
    if(m_session) delete m_session;
    m_session = 0;
    seal::MessageStream log( m_context, m_name );
    log << seal::Msg::Error << "Cannot reopen the database in \"" << m_name << "\"." << seal::flush;
    log << seal::Msg::Error << e.what() << seal::flush;
    return pool::Error;    
  }
  
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::close( pool::DbAccessMode /* mode */ )
{
  try 
  {
    coral::ITransaction& transaction = m_session->transaction();
    if ( transaction.isActive() ) {
      if ( transaction.isReadOnly() ) {
        transaction.commit();
      }
      else {
        transaction.rollback();
      }
    }
  }
  catch(const coral::Exception&)
  {
  }
  
  if ( m_databaseManager ) delete m_databaseManager;
  m_databaseManager = 0;
  if ( m_session ) delete m_session;
  m_session = 0;
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::transAct( pool::DbTransaction& refTransaction )
{
  try 
  {
    pool::Transaction::Action action = refTransaction.state();
    coral::ITransaction& transaction = m_session->transaction();
    bool activeTransaction = transaction.isActive();
    switch ( action ) {
    case pool::Transaction::TRANSACT_START:
    case pool::Transaction::TRANSACT_ACTIVE:
      if ( ! activeTransaction )
        transaction.start( m_readOnlyMode );
      break;
    case pool::Transaction::TRANSACT_COMMIT: // This is the PersistencySvc commitAndHold()
      if ( ! activeTransaction ) {
        seal::MessageStream log( m_context, m_name );
        log << seal::Msg::Error << "There is no active transaction to commit." << seal::flush;
        return pool::Error;
      }
      if ( m_databaseManager ) m_databaseManager->signalEndOfTransaction();
      transaction.commit();
      if ( ! activeTransaction ) transaction.start( m_readOnlyMode );
      break;
    case pool::Transaction::TRANSACT_FLUSH:
      if ( ! activeTransaction ) {
        seal::MessageStream log( m_context, m_name );
        log << seal::Msg::Error << "There is no active transaction to commit." << seal::flush;
        return pool::Error;
      }
      if ( m_databaseManager ) m_databaseManager->signalEndOfTransaction();
      transaction.commit();
      break;
    case pool::Transaction::TRANSACT_ROLLBACK:
    case pool::Transaction::TRANSACT_ENDED:
      if ( activeTransaction ) {
        if ( m_databaseManager ) m_databaseManager->signalEndOfTransaction();
        transaction.rollback();
      }
      break;
    };
  }
  catch(const coral::Exception& e)
  {
    if(m_session) delete m_session;
    m_session = 0;
    seal::MessageStream log( m_context, m_name );
    log << seal::Msg::Error << "Cannot terminate the transaction. "<< seal::flush;
    log << seal::Msg::Error << e.what() << seal::flush;
    return pool::Error;
  }
  
  return pool::Success;
}

void
pool::RelationalStorageService::RelationalPoolDatabase::dropContainer( const std::string& databaseName,
                                                                       const std::string& containerName,
                                                                       bool removeDataTables ){
  // Retrieve an IConnectionService
  std::vector< seal::IHandle< coral::IConnectionService > > loadedServices;
  pool::POOLContext::context()->query( loadedServices );
  if ( loadedServices.empty() ) {
    // Load the ConnectionService into the top-lecel POOL context
    pool::POOLContext::loadComponent( "CORAL/Services/ConnectionService" );
    pool::POOLContext::context()->query( loadedServices );
  }
  seal::IHandle< coral::IConnectionService > connectionService = loadedServices.front();

  coral::ISessionProxy* session = 0;
  try{
    session = connectionService->connect( databaseName, coral::Update );
    if ( ! session ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException("Could not connect to \""+databaseName+"\"",
                                                                              "RelationalPoolDatabaseUtilities::dropContainer");
    }
    coral::ITransaction& transaction = session->transaction();
    transaction.start();

    pool::RelationalStorageService::PoolDatabaseManager manager( session, pool::POOLContext::context()  );
    if(!manager.existsDatabase()){
      transaction.rollback();
      delete session;
      throw pool::RelationalStorageService::RelationalStorageServiceException("No POOL Database found in "+databaseName+".",
                                                                              "RelationalPoolDatabaseUtilities::dropContainer");
    }

    // when requested, first remove all the materialized tables
    if(removeDataTables){
      pool::ObjectRelationalMappingPersistency mappingPersistency(session->nominalSchema());
      std::vector<std::string> mappingsForContainer = mappingPersistency.mappingsForContainer(containerName);

      for(std::vector<std::string>::const_iterator iM=mappingsForContainer.begin();
          iM!=mappingsForContainer.end();++iM){
        mappingPersistency.removeMaterializedTables(*iM);
        mappingPersistency.removeMapping(*iM);
      }
    }

    // delete the container from the pool database 
    manager.dropContainer( containerName );
    
    transaction.commit();
    delete session;
    session = 0;
  }
  catch (const coral::Exception& /* e */)
  {
    if(session) delete session;
    throw pool::RelationalStorageService::RelationalStorageServiceException("Could not connect to \""+databaseName+"\"",
                                                                            "RelationalPoolDatabaseUtilities::dropContainer");
  }
  
  if(session) delete session;
}

